/*!
 *Last modified: 2024-07-10 22:17:31
 *Filename: ax.js
 *Description: Global JS
 *Version: v2.1.2
 *Website:www.axui.cn
 *Contact:3217728223@qq.com
 *Author:Michael
 */


//////////////////////////////////////////// 版块分割线 ////////////////////////////////////////////
/*!
 * 全局变量，至于ax.js头部，可全站使用
 */
'use strict';
/**
 * 定义常用错误提示 
 */
const axError = {
    formatError: 'This is not a valid data format!',
    paraError: 'The parameters do not meet the requirements!',
    arrayRequired: 'The value must be an array type!',
    noNode: 'This page can not find the node!',
    required: 'Required field is missing!',
    notExist: 'The program cannot be executed because the object does not exist!',
}
/**
 * 定义判断是否是移动端函数，返回true或false 
 */
const axIsMobi = () => ('ontouchstart' in document.documentElement);
/**
 * 获得移动端触摸事件和pc端鼠标事件的映射
*/
const axEventMap = axIsMobi() ? ['touchstart', 'touchmove', 'touchend', 'touchcancel'] : ['mousedown', 'mousemove', 'mouseup', 'mouseleave'];

/**
 * 判断字符串是否为日期格式，返回true或false 
 * 2021-2-5,2021/2/5,2021,2021-2,2-5
 */
const axIsDate = (str) => (isNaN(str) && !isNaN(Date.parse(str)));
/**
 * 从css获得终端边界像素
 */
const axCeil = {
    phone: parseInt(getComputedStyle(document.documentElement).getPropertyValue('--ceil-phone')),
    pad: parseInt(getComputedStyle(document.documentElement).getPropertyValue('--ceil-pad')),
    flip: parseInt(getComputedStyle(document.documentElement).getPropertyValue('--ceil-flip')),
    pro: parseInt(getComputedStyle(document.documentElement).getPropertyValue('--ceil-pro')),
    pc: parseInt(getComputedStyle(document.documentElement).getPropertyValue('--ceil-pc')),
}
/**
 * 定义客户端函数，通过屏幕宽度判断是什么终端
 */
const axClient = () => {
    //let screenWidth = window.screen.width;
    let width = document.body.clientWidth;
    if (width < axCeil.phone) {
        return 'phone';
    } else if (width > axCeil.phone && width < axCeil.pad) {
        return 'pad';
    } else if (width > axCeil.pad && width < axCeil.flip) {
        return 'flip';
    } else if (width > axCeil.flip && width < axCeil.pro) {
        return 'pro';
    } else {
        return 'pc';
    }
}
/**
 * 获得对象真正的类型
 * 可检测出Array,Object,Function,String,Number,Boolean,Date,Null,Undefined,HTMLDivElement(Dom节点都包含HTML)等
 */
const axType = (obj) => Object.prototype.toString.call(obj).slice(8, -1);
/**
 * 定义判断为空函数，可判断function(){}，0，''，[]，{}，undefined，null
 */
const axIsEmpty = (data) => {
    let type = axType(data);
    if (!data) {
        //0,'',undefined,null
        return true;
    } else if (type === 'Function') {
        //function(){}
        return (data.toString().replace(/\s+/g, '').match(/{.*}/g)[0] == '{}');
    } else if (type === 'Array') {
        //[null],[undefined],[''],[""]
        return data.join('') == '';
    } else if (type === 'Object') {
        //[],{}
        return (Object.keys(data).length === 0)
    } else {
        //node节点类型等返回false
        return false;
    }
}
/**
 * 非数字转为数字
*/
const axToNumber = (value) => isNaN(value) ? parseFloat(value) : value;

/**
 * 设置项目颜色主题
*/
const axTheme = {
    set: (value, name = '') => {
        //字符串value=#ebebeb,rgba(255,255,255,1),hsla(0,0%,0%,1)
        //对象value={primary:[h,s,l,a],text:'',error:'',success:'',info:'',warning:'',question:''}
        if (axIsEmpty(value)) {
            return false;
        }
        let hexToRgbArr = (val) => {
            val = val.trim();
            // 16进制颜色值的正则
            let reg = /^#([0-9a-fA-f]{3}|[0-9a-fA-f]{6})$/,
                colorStr = val.toLowerCase(),
                result = [];
            if (reg.test(colorStr)) {
                // 如果只有三位的值，需变成六位，如：#fff => #ffffff
                //#abc === #aabbcc
                if (colorStr.length === 4) {
                    let temp = '#';
                    for (let k = 1; k < 4; k += 1) {
                        temp += colorStr.slice(k, k + 1).concat(colorStr.slice(k, k + 1));
                    }
                    colorStr = temp;
                }
                //此时颜色值已经标准处理了，以16进制进行计算并转为10进制结果（取整）
                for (let k = 1; k < 7; k += 2) {
                    result.push(parseInt("0x" + colorStr.slice(k, k + 2)));
                }
            }
            return result;
        },
            rgbToHsl = (arr) => {
                let r = arr[0] / 255,
                    g = arr[1] / 255,
                    b = arr[2] / 255,
                    min = Math.min(r, g, b),
                    max = Math.max(r, g, b),
                    h, s, l = (min + max) / 2,
                    diff = max - min;
                if (max === min) {
                    h = 0;
                    s = 0;
                } else {
                    s = l > 0.5 ? diff / (2 - max - min) : diff / (max + min);
                    switch (max) {
                        case r: h = (g - b) / diff + (g < b ? 6 : 0); break;
                        case g: h = 2 + (b - r) / diff; break;
                        case b: h = 4 + (r - g) / diff; break;
                    }
                    h = Math.round(h * 60);
                }
                //转换成百分比的形式
                s = Math.round(s * 100);
                l = Math.round(l * 100);
                //保留a通道
                return arr.length === 4 ? [h, s, l, arr[3]] : [h, s, l, 1];
            },
            getHslaArr = (val) => {
                let valType = axType(val), result = [];
                if (!axIsEmpty(val)) {
                    if (valType === 'String') {
                        val = val.trim().toLowerCase();
                        //#fff或rgba|rbg(255,255,255)或hsl|hsla(0,0%，0%,1)
                        if (val.startsWith('hsl')) {
                            //hsla格式
                            result = val.replace(/^(hsl|hsla)\(/, '').replace(/\)$/, '').replace(/\s/g, '').split(',');
                        } else {
                            let temp = []
                            if (val.startsWith('#')) {
                                //16进制格式
                                temp = hexToRgbArr(val);
                            } else if (val.startsWith('rgb')) {
                                //rgba格式
                                temp = val.replace(/^(rgb|rgba)\(/, '').replace(/\)$/, '').replace(/\s/g, '').split(',').map(Number);
                            }
                            result = rgbToHsl(temp);
                        }
                    } else if (valType === 'Array') {
                        //hsl数组
                        result = val;
                    }
                }
                //补全a通道
                result.length === 3 ? result.push(1) : null;
                return result;
            },
            valueType = axType(value);
        let hslaObj = { primary: [], text: [], error: [], success: [], info: [], warning: [], question: [] };

        if (valueType === 'Object') {
            for (let k in value) {
                if (value.hasOwnProperty(k) && value[k].length > 2) {
                    hslaObj[k] = getHslaArr(value[k]);
                }
            }
        } else {
            hslaObj.primary = getHslaArr(value);
        }
        //保存原始主题色
        if (axIsEmpty(axLocalStorage.get('ax-theme-raw'))) {
            axLocalStorage.set('ax-theme-raw', axGetColors());
        }
        //此时已经获得标准hsla数组，根据该数组对css中的变量进行重写
        let rootDom = document.querySelector(":root");
        for (let k in hslaObj) {
            if (hslaObj.hasOwnProperty(k) && hslaObj[k].length > 2) {
                rootDom.style.setProperty(`--h-${k}`, hslaObj[k][0]);
                rootDom.style.setProperty(`--s-${k}`, typeof hslaObj[k][1] === 'string' && hslaObj[k][1].includes('%') ? hslaObj[k][1] : hslaObj[k][1] + '%');
                rootDom.style.setProperty(`--l-${k}`, typeof hslaObj[k][2] === 'string' && hslaObj[k][2].includes('%') ? hslaObj[k][2] : hslaObj[k][2] + '%');
                rootDom.style.setProperty(`--a-${k}`, hslaObj[k][3]);
            }
        }
        //保存风格
        name ? axLocalStorage.set(name, hslaObj) : null;
    },
    get: () => {
        return axGetColors();
    },
    reset: () => { }
}
axTheme.reset = () => {
    let themeRaw = axLocalStorage.get('ax-theme-raw');
    axTheme.set(themeRaw);
}
/**
 * 获得浏览器滚轮对象，兼容firefox和chrome
*/
const axGetScrollObj = () => {
    return navigator.userAgent.indexOf("Firefox") > 0 ? { event: 'DOMMouseScroll', detail: 'detail' } : { event: 'mousewheel', detail: 'wheelDelta' };
}
/**
 * 获得终端client对象（clientX和clientY），移动端将取targetTouches第一项，但是可能为null
*/
const axGetClientObj = (event) => {
    let client = axIsMobi() ? event.targetTouches[0] : event;
    return client ? { x: client.clientX, y: client.clientY } : null;
}

/**
 * 判断页面是否在上卷，兼容firefox和chrome
*/
const axIsScrollUp = (e, detail = axGetScrollObj().detail) => {
    let mouseDetail = e[detail], isUp;
    if (navigator.userAgent.indexOf("Firefox") > 0) {
        isUp = mouseDetail > 0 ? true : false;
    } else {
        isUp = mouseDetail < 0 ? true : false;
    }
    return isUp;
}
/**
 * 判断是否是虚拟节点，也就是仅存在内存当中并没有append到页面里
*/
const axIsVnode = (elem) => {
    return !getComputedStyle(elem).display;
}
/**
 * 保留两位小数，不补零
*/
const axNumber = (data, opt) => {
    //opt={places:'',mode:''}
    //mode='round'四舍五入，floor向下取小，ceil向上取大
    let options = Object.assign({ places: 2, mode: 'round', zero: false }, opt),
        result = 0;
    if (!data || !Number(data)) {
        return 0;
    } else {
        let number = Number(data),
            //四舍五入需要使用Number.EPSILON
            precise = number + Number.EPSILON,
            places = Number(`1${'0'.repeat(options.places)}`),
            tempNumber = number * places,
            tempPrecise = precise * places;
        //根据模式取舍整数部分
        if (options.mode === 'floor') {
            result = Math.floor(tempNumber) / places;
        } else if (options.mode === 'ceil') {
            result = Math.ceil(tempNumber) / places;
        } else {
            result = Math.round(tempPrecise) / places;
        }
    }
    //如果需要补零，那么result将转为string类型
    if (options.zero) {
        result = result.toFixed(options.places)
    }
    return result;
}
/**
 * 去掉字符串的头尾空格，中间如果有空格则只保留一个
*/
const axTrim = (str, all = false) => {
    return all ? str.replace(/\s+/g, '') : str.trim().replace(/\s+/ig, ' ');
};
/**
 * 字符串根据分隔符转成数组
*/
const axSplitToArr = (data, separator = ',', space = 'all') => {
    let dataType = axType(data),
        spaceStr = '',
        dataArr = [];
    if (dataType === 'String') {
        if (space === 'one') {
            //中间保留一个空格
            spaceStr = axTrim(data, false)
        } else if (space === 'none') {
            //中间不保留空格
            spaceStr = axTrim(data, true)
        } else {
            //不处理中间的空格
            spaceStr = data.trim();
        }
        dataArr = spaceStr.split(separator);
    } else if (dataType === 'Array') {
        dataArr = data;
    }
    return dataArr;
};
/**
 * 阻止默认事件
*/
const axPreventDefault = (event) => {
    if (event.preventDefault) {
        event.preventDefault();
    } else {
        event.returnValue = false;
    }
}
/**
 * 取字符串，可向前取或向后取
*/
const axStrEnd = (str, key, type = 'afterend', inside = true) => {
    if (!str || !key) {
        return false;
    }
    str = str.trim();
    if (str.includes(key)) {
        let index, result, len = key.length;
        if (type === 'beforebegin') {
            index = str.indexOf(key);
            inside ? index += len : null;
            result = str.substring(0, index);
        } else if (type === 'afterbegin') {
            index = str.indexOf(key);
            !inside ? index += len : null;
            result = str.substring(index, str.length);
        } else if (type === 'beforeend') {
            index = str.lastIndexOf(key);
            inside ? index += len : null;
            result = str.substring(0, index);
        } else if (type === 'afterend') {
            index = str.lastIndexOf(key);
            !inside ? index += len : null;
            result = str.substring(index, str.length);
        } else {
            result = '';
        }
        return result;
    } else {
        return '';
    }
}
/**
 * 判断选择器类型
*/
const axStrType = (str) => {
    if (typeof str !== 'string') {
        console.error('Argument must be a string!')
        return false;
    }
    str = str.trim();
    let isUpperCase = (letter) => {
        if (letter >= 'A' && letter <= 'Z') {
            return true;
        }
    },
        type = '';
    if (str) {
        if (str.includes(' ')) {
            //字符串有空格，是父子选择器
            type = 'level';
        } else {
            if (str.startsWith('#')) {
                //以#开头则判断为id名
                type = 'id';
            } else if (str.startsWith('.')) {
                //以.开头则判断为样式类名
                type = 'class';
            } else if (str.startsWith('[') && str.endsWith(']')) {
                //以[开头，以]结尾则判断为表单域name名
                type = 'name';
            } else if (['H1', 'H2', 'H3', 'H4', 'H5', 'H6'].includes(str) || [...str].every(i => isUpperCase(i))) {
                //如果都是大写则判断为node节点名称
                type = 'node';
            }
        }
    }
    return type;
}
/**
 * 将字符串转成DOM对象，诸如<div>...</div>的字符串
 */
const axStrToDom = (str) => {
    if (typeof str != 'string') {
        //必须是字符串类型，否则不执行
        return false;
    }
    str = str.trim();
    let wrapper;
    if (str.startsWith("<tr") || str.endsWith("</tr>")) {
        //tr+td字符串
        wrapper = document.createElement('tr');
        wrapper.innerHTML = str;
        return wrapper;
    } else {
        wrapper = document.createElement('div');
        wrapper.innerHTML = str;
        return wrapper.firstElementChild;
    }
}
/**
 * 将#id转成DOM，插件中常用
 */
const axIdToDom = (elem) => {
    let type = axType(elem);
    if (!elem) {
        console.warn('Node selector is empty!');
        return false;
    } else if (type === 'String') {
        //是字符串
        let trim = elem.trim();
        if (axStrType(trim)) {
            if (axStrType(trim) === 'name' && !trim.includes('=')) {
                //trim=['xxx']转成：trim=[name='xxx']
                trim = trim.slice(0, 1) + 'name=' + trim.slice(1);
            }
            return document.querySelector(trim);
        } else {
            console.warn('No node is found with this string!');
            return false;
        }
    } else if (type.includes('HTML')) {
        //本身就是节点
        return elem;
    } else {
        return false;
    }
}
/**
 * 将多种形式转成DOM数组，插件中常用
 */
const axAllToDoms = (data) => {
    if (axIsEmpty(data)) {
        return [];
    }
    let doms = [],
        type = axType(data);
    if (type.includes('HTML')) {
        //是一个节点
        doms.push(data);
    } else if (type === 'String') {
        //多值字符串
        let str = axTrim(data),
            separator = ',';
        if (str.includes(' ')) {
            separator = ' ';
        }
        str.split(separator).forEach(k => {
            let dom = axIdToDom(k);
            dom ? doms.push(dom) : null;
        });
    } else if (type === 'Array') {
        //数组
        data.forEach(k => {
            let dom = axIdToDom(k);
            dom ? doms.push(dom) : null;
        });
    } else if (type === 'NodeList') {
        //节点伪数组
        doms = [...data];
    }
    return doms;
}
/**
 * 将多种形态的字符串，对象等都转成数组格式
 * obj='a'/'a,b,c'/'a b c'/{}/0/1/[]
 */
const axValueToArr = (obj) => {
    let arr = [];
    if (!axIsEmpty(obj) || obj === 0) {
        //不为空才执行
        if (Array.isArray(obj)) {
            //数组
            arr = obj;
        } else if (typeof obj === 'string') {
            //字符串
            let str = obj.trim(),
                separator = obj.includes(',') ? ',' : ' ';
            if (str.includes(separator)) {
                arr = str.split(separator);
            } else {
                arr.push(obj);
            }
        } else {
            //数字或对象
            arr.push(obj);
        }
    }
    return arr;
}
/**
 * 从对象数组中去重，[{id:0,label:'a'},{id:0,label:'b'}]转为[{id:0,label:'a'}]
 * 如果prop为空则，对普通数组去重，[0,0,1,2]转为[0,1,2]
 */
const axUnique = (arr, prop) => {
    if (prop) {
        let map = new Map();
        for (let k of arr) {
            if (!map.has(k[prop])) {
                map.set(k[prop], k);
            }
        }
        return [...map.values()];
    } else {
        return [...new Set(arr)]
    }
}
/**
 * 对DOM节点进行class操作，支持get、add和remove三个方法
 */
const axClass = {
    get: function (classes) {
        return axValueToArr(classes);
    },
    add: function (obj, classes, before) {
        let dom = axIdToDom(obj),
            arr = this.get(classes);
        //没有节点或arr为空则不再执行
        if (!dom || arr.length === 0) {
            return false;
        }
        arr.forEach(k => {
            let flag = before ? before(k) : true;
            flag ? dom.classList.add(k) : null;
        });
        return this;
    },
    del: function (obj, classes, before) {
        let dom = axIdToDom(obj),
            arr = this.get(classes);
        //没有节点或arr为空则不再执行
        if (!dom || arr.length === 0) {
            return false;
        }
        arr.forEach(k => {
            let flag = before ? before(k) : true;
            flag ? dom.classList.remove(k) : null;
        });
        return this;
    }
}

/**
 * 将字符串转成数组继而转成Dom，诸如"#xxx,,yyy,DIV"这类参数型的字符串
 */
const axArrToDom = (str, parent) => {
    let arr = str.split(","),
        doms = [],
        reg = /^[0-9]*$/,
        _parent = '';
    if ((typeof parent) === 'string') {
        _parent = document.querySelector(parent);
    } else {
        _parent = parent;
    }
    arr.forEach(item => {
        if (reg.test(item.substr(0, 1))) {
            return;
        } else if (axStrType(item) == 'class') {
            if (_parent) {
                doms = doms.concat([..._parent.querySelectorAll(item)]);
            } else {
                doms = doms.concat([...document.querySelectorAll(item)]);
            }
        } else if (axStrType(item) == 'id') {
            doms.push(document.querySelector(item));
        } else if (item) {
            let nodes = document.querySelectorAll(item);
            if (nodes) {
                if (_parent) {
                    doms = doms.concat([..._parent.querySelectorAll(item)]);
                } else {
                    doms = doms.concat([...document.querySelectorAll(item)]);
                }
            }
        } else {
            return doms;
        }
    })
    return doms;
}
/**
 * 将带children属性的多维数组拍平为一维数组
*/
const axArrToFlat = (arr) => {
    let result = [],
        data = [...arr];
    data.forEach(i => {
        if (i.children) {
            result = [...result, i, ...axArrToFlat(i.children)];
        } else {
            result.push(i);
        }
    });
    return result;
}
/**
 * 根据现有数组获得新项目的ID（数组最大id或索引+1）
*/
const axIncreaseId = (data, hasId = true) => {
    let ids = [],
        newId;
    if (hasId) {
        data.forEach(i => {
            ids.push(~~i.id);
        });
        //自动+1
        newId = Math.max(...ids) + 1;
    } else {
        //当前数组长度
        newId = data.length;
    }
    return newId;
}
/**
 * 从表单控件取值，用于校验判断，适合的控件：checkbox、radio、select-multiple和file
*/
const axMultiValues = (name, type = 'checkbox', format = 'array', form, separator = ',') => {
    let inputs,
        values = [],
        parent = form ? axIdToDom(form) : '';

    if (axType(name) == 'NodeList' || axType(name) == 'Array') {
        //节点数组
        if (type.includes('select') || type == 'file') {
            //select-multiple,select-one,select均可
            //对于select和file不允许是节点数组
            console.warn('Can not be an node array in this type!');
            return false;
        }
        //如果是input节点数组
        inputs = name;
    } else if (typeof name == 'object' || (axStrType(name) && axStrType(name) !== 'name')) {
        //一个节点
        let dom = axIdToDom(name);
        if (type.includes('select') || name.nodeName == 'SELECT') {
            //select-multiple,select-one,select均可
            //如果是父节点或节点本身
            type = 'select';
            inputs = dom.options;
        } else if (type == 'file' || dom.type == 'file') {
            //如果是父节点或节点本身
            type = 'file';
            //转成数组，便于下方循环inputs
            inputs = [dom];
        } else {
            //如果是父节点（对于checkbox和radio）
            inputs = [...dom.querySelectorAll('input')];
        }
        //inputs是input或options的节点数组
    } else {
        //如果是字符串：name名或[name]
        let selector = '';
        //支持name以[name]形式传入
        if (axStrType(name) === 'name') {
            selector = name.replace('[', '[name=');
        } else {
            selector = `[name=${name}]`;
        }

        if (type.includes('select')) {
            //select-multiple,select-one,select均可
            //取第一个select的选项
            inputs = parent ? parent.querySelector(selector).options : document.querySelector(selector).options;
        } else if (type == 'file') {
            //节点数组取第一个
            inputs = parent ? [parent.querySelector(selector)] : [document.querySelector(selector)];
        } else {
            inputs = parent ? [...parent.querySelectorAll(selector)] : [...document.querySelectorAll(selector)];
        }
    }
    //循环获取checked状态的input值
    for (let i = 0; i < inputs.length; i++) {
        let condition,
            item = inputs[i];
        if (type.includes('select')) {
            //select-multiple,select-one,select均可
            condition = item.selected;
        } else {
            condition = item.checked;
        }
        //值格式:['a','b','c']
        if (type == 'file') {
            //对于file
            values = [...item.files];
        } else {
            if (condition) {
                values.push(item.value);
            }
        }
    }
    //值格式：a,b,c
    if (format == 'string') {
        values = values.join(separator)
    }
    return values;
}
/**
 * 防抖函数
 * 使用表现：持续输入后延迟执行
*/
const axDebounce = (fn, delay = 200, hold = false) => {
    let timer = null,
        handler = function (e) {
            if (hold) {
                e.stopPropagation();
                e.preventDefault();
            }
            if (timer) {
                clearTimeout(timer);
            };
            timer = setTimeout(() => {
                fn.apply(this, arguments);
            }, delay);
        }
    //取消处理
    handler.cancel = function () {
        if (timer) clearTimeout(timer);
    }
    return handler;
}
/**
 * 节流函数
 * 使用表现：持续点击多次也只会按固定间隔时间执行
*/
const axThrottle = (fn, interval = 500, hold = false) => {
    let last = 0,
        timer = null,
        handler = function (e) {
            let now = new Date().getTime();
            if (now - last > interval) {
                if (timer) {
                    clearTimeout(timer);
                    timer = null;
                }
                fn.apply(this, arguments);
                last = now;
            } else {
                if (hold) {
                    e.stopPropagation();
                    e.preventDefault();
                }
                /*  if (timer === null) {
                     //只是最后一次
                     timer = setTimeout(() => {
                         timer = null;
                         fn.apply(this, arguments);
                     }, interval);
                 } */
            }
        }
    handler.cancel = function () {
        clearTimeout(timer);
        timer = null;
    }
    return handler;
}
/**
 * 数组元素移动位置
 */
const axMoveArr = (arr, from, to) => {
    if (from > arr.length || to > arr.length || from < 0 || to < 0) {
        console.error('Exceeded array length!');
        return false;
    }
    let obj = arr[from];
    arr.splice(from, 1);
    arr.splice(to, 0, obj);
    return arr;
}
/**
 * 判断元素以及其父元素是否可见
 */
const axIsVisible = (element) => {
    let elem = axIdToDom(element),
        loopable = true,
        visible = getComputedStyle(elem).display != 'none' && getComputedStyle(elem).visibility != 'hidden';
    while (loopable && visible) {
        elem = elem.parentNode;
        if (elem && elem != document.body) {
            visible = getComputedStyle(elem).display != 'none' && getComputedStyle(elem).visibility != 'hidden';
        } else {
            loopable = false;
        }
    }
    return visible;
}
/**
 * 将值限制在范围内
*/
const axClampValue = (options = { value, min, max }) => {
    if (!options.hasOwnProperty('min') || !options.hasOwnProperty('max')) {
        console.error('Must include max and min properties!');
        return false;
    }
    //先转为数值
    options.value = axToNumber(options.value);
    options.min = axToNumber(options.min);
    options.max = axToNumber(options.max);
    return options.value > options.max ? options.max : options.value < options.min ? options.min : options.value;
}
/**
 * 判断在元素在其父元素中是否可见，要求父元素position必须为absolute/fixed/relative，否则会判断出错
 */
const axChildVisible = (child, parent) => {
    let parentDom = axIdToDom(parent),
        childDom = axIdToDom(child),
        parentScroll = parentDom.scrollTop,
        parentHeight = parentDom.getBoundingClientRect().height;
    return (childDom.offsetTop + childDom.offsetHeight > parentScroll) && (childDom.offsetTop < parentScroll + parentHeight);
};
//将Ul+Li的Dom树转成对象数组格式：[{id:'',label:'',value:''},{id:'',label:'',value:''}...]
const axUlToArr = (element, start = 0, replace = true) => {
    let elem = axIdToDom(element),
        newId = start;
    if (!element) { console.error(axError.notExist); return false; }
    let getChild = () => {
        let data = [];
        [...elem.children].forEach(item => {
            data.push(getJson(item, newId++));
        });
        return data;
    },
        getAttrs = (node, obj) => {
            //label和value另外单独设定
            let attrs = [...node.attributes].filter(k => k.name !== 'label' && k.name !== 'value');
            attrs.forEach(k => {
                if (['selected', 'disabled', 'readonly', 'active', 'checked', 'expanded'].includes(k.name)) {
                    //状态属性
                    obj[k.name] = true;
                } else {
                    //brief='xxx'有值属性
                    obj[k.name] = k.value.trim();
                }
            });
        };

    let getJson = (elem, index) => {
        let node = elem.firstElementChild,
            obj = { id: index, node: node };
        obj.label = node.textContent.trim();
        if (replace) {
            //如果允许替换
            obj.value = node.hasAttribute('value') ? node.getAttribute('value') : obj.label;
        } else {
            node.hasAttribute('value') ? obj.value = node.getAttribute('value') : null;
        }

        //获得属性值
        getAttrs(node, obj);
        //有值属性：'sub','tip', 'legend', 'image', 'avatar', 'brief','rank','time','editor','source','count','keyword','state','flag','type','visit','review'
        //无值属性：'selected', 'disabled','readonly', 'active', 'checked', 'expanded'

        let parent = elem.querySelector('ul') || elem.querySelector('ol');
        if (parent) {
            //有ul节点，说明是父节点
            obj.children = [];
            [...parent.children].forEach(item => {
                obj.children.push(getJson(item, newId++));
            });
        }
        return obj;
    }
    return { data: getChild(), index: newId };
}
/**
 * 将下拉菜单Dom树转成对象数组格式：[{id:'',label:'',value:''},{id:'',label:'',value:''}...]
 */
const axSelectToArr = (element, start = 0, replace = true) => {
    let elem = axIdToDom(element),
        newId = start;
    if (!elem) { console.error(axError.notExist); return false; }
    let getChild = () => {
        let data = [];
        [...elem.children].forEach(item => {
            data.push(getJson(item, newId++));
        });
        return data;
    },
        getAttrs = (node, obj) => {
            //label和value另外单独设定
            let attrs = [...node.attributes].filter(k => k.name !== 'label' && k.name !== 'value');
            attrs.forEach(k => {
                if (['selected', 'disabled', 'readonly', 'active', 'checked', 'expanded'].includes(k.name)) {
                    //状态属性
                    obj[k.name] = true;
                } else {
                    //brief='xxx'有值属性
                    obj[k.name] = k.value.trim();
                }
            });
        };

    let getJson = (elem, index) => {
        let obj = { id: index, node: elem },
            label = elem.textContent.trim(),
            value = elem.hasAttribute('value') ? elem.getAttribute('value').trim() : '';
        //获得属性值
        getAttrs(elem, obj);
        //有值属性：'sub','tip', 'legend', 'image', 'avatar', 'brief','rank','time','editor','source','count','keyword','state','flag','type','visit','review'
        //无值属性：'selected', 'disabled','readonly', 'active', 'checked', 'expanded'

        //获得label和value
        if (elem.nodeName == 'OPTGROUP') {
            obj.label = elem.getAttribute('label').trim();
            obj.value = obj.label;
            //追加子元素
            obj.children = [];
            [...elem.children].forEach(item => {
                obj.children.push(getJson(item, newId++));
            });
        } else {
            //闭合标签label是为空的，此时应该是datalist
            obj.label = label ? label : value.trim();
            if (replace) {
                //如果允许替换，则将label赋值给value
                obj.value = value ? value : obj.label;
            } else {
                obj.value = value;
            }
        }
        return obj;
    }
    return { data: getChild(), index: newId };
}
/**
 * 从table数据转成二维数组格式：[['','',...],...]
 * element为table标签
 */
const axTableToJson = (element) => {
    let elem = axIdToDom(element),
        tbody = elem.querySelector('tbody'),
        rows = tbody.rows,
        rowArr = [];
    [...rows].forEach((i) => {
        let tds = i.cells,
            tdArr = [];
        [...tds].forEach((k) => {
            tdArr.push(k.innerText.trim());
        });
        rowArr.push(tdArr);
    });
    return rowArr;
}
/**
 * 一维数组转tree数组，root为顶级id，通常是-1或0
 */
const axArrToTree = (arr) => {
    let root = arr[0].pId;
    //循环给元素找到子元素并赋予children
    arr.forEach(k => {
        let children = arr.filter(i => k.id == i.pId);
        if (children.length) {
            k.children = children;
        }
    });
    //过滤掉已经被折叠为children的对象
    return arr.filter(k => k.pId == root);
}
/**
 * 找到节点数组
 */
const axGetDoms = (data) => {
    //data可以是字符串或节点
    let dataType = axType(data),
        arr = [],
        getData = (obj) => {
            let objType = axType(obj), result;
            if (objType === 'String' && axStrType(obj)) {
                result = [...this.targetDom.querySelectorAll(obj)];
            } else if (objType.includes('HTML')) {
                result = [obj];
            }
            return result;
        };
    if (dataType === 'Array') {
        arr = data.map(k => getData(k)).flat(Infinity);
    } else {
        arr = getData(data);
    }
    return arr.filter(Boolean);
}
/**
 * 使用任意值找到一项，可以是一个对象、数字、字符串或节点
 */
const axFindItem = (obj, data, map) => {
    let type = axType(obj),
        propMap = Object.assign({ string: 'label', number: 'id', dom: 'headerDom' }, map),
        result,
        objFind = (value, list) => {
            //比对每一个key
            let keys = Object.keys(value),
                item = list.find(k => {
                    let flag = true;
                    for (let v of keys) {
                        if (k[v] != value[v]) {
                            flag = false;
                            break;
                        }
                    }
                    return flag
                });
            return item;
        };
    if (type === 'Object') {
        //优先直接从数据中比对，其次再比对每一个属性
        result = data.find(k => k === obj) || objFind(obj, data);
    } else if (type === 'Array') {
        //obj是数组
        result = data.find(k => k === obj);
    } else if (type === 'Number') {
        //是数字
        if (data[0].hasOwnProperty(propMap.number)) {
            //有id属性则找到对应ID的项，id可能是字符串
            result = data.find(k => String(k[propMap.number]) === String(obj));
        } else {
            //如果没有id属性则取索引
            result = data[obj];
        }
    } else if (type === 'String') {
        if (obj.startsWith('#')) {
            //以#开头则找对应节点
            result = data.find(k => '#' + k[propMap.dom].id === obj);
        } else {
            //是字符串则找对label一致的项；可能数据中的id也是字符串所以还是先判断id
            result = data.find(k => String(k[propMap.number]) === obj || String(k[propMap.string]) === obj);
        }
    } else if (type.includes('HTML')) {
        //是label节点则找到有对应labelDom的项
        result = data.find(k => k[propMap.dom] === obj);
    } else {
        //返回false
        result = false;
    }
    return result;
}
/**
 * 将创建tree的数据对象push到数组里和节点里
 */
const axTreeMethod = {
    //tree数据带上id、pid、floor和path，以便于查找父子链条
    addPath: (data, rootStart = -1, idStart = 0) => {
        if (!Array.isArray(data) || data.length === 0) {
            //非数组，数组为空，数组没有id属性都不再执行
            return false;
        }
        let floorMax = 0,
            idMax = 0,
            //判断是否有id，针对没有id的数组
            hasId = data[0].hasOwnProperty('id') ? true : false,
            index = idStart,
            each = (data, floor, path = rootStart) => {
                //路径都以rootStart>开头,rootStart默认为0或-1
                data.forEach(item => {
                    if (!hasId) {
                        item.id = index;
                        //循环+1
                        index++;
                    }
                    item.floor = floor;
                    //追加本身path路径
                    item.pId = item.pId || item.pId === 0 ? item.pId : rootStart;
                    //追加路径
                    item.path = path.toString() + '>' + item.id.toString();
                    //获得最大层级
                    floor >= floorMax ? floorMax = floor : null;
                    //获得最大id
                    if (!isNaN(item.id)) {
                        item.id >= idMax ? idMax = item.id : null;
                    }
                    //追加子节点path路径
                    if (item.children && item.children.length > 0) {
                        item.children.forEach(k => {
                            //设置pId
                            k.pId = item.id;
                        });
                        //进入循环
                        each(item.children, floor + 1, item.path);
                    }
                    if (!item.path) {
                        //第一层还没有path，直接为pId > id
                        item.path = item.pId + '>' + item.id;
                    }
                });
            }
        //data:原始数据，1:开始层级
        each(data, 1);
        return { data, floorMax, idMax };
    },
    //从flat数据中删除（与tree数据联动），每一项都有pid、id、floor、path和wrapperDom属性
    remove: (item, flat, tree, callback) => {
        //floor表示是tree的第一层
        if (item.floor !== 1) {
            //path构成“0>1>5”，0是rootStart，pId=rootStart表明这个item是第一层
            let parent = flat.find(k => k.id == item.pId),
                children = parent.children,
                index = children.indexOf(item);
            children.splice(index, 1);
        } else {
            //可选
            tree ? tree = tree.filter(k => k != item) : null;
        }
        //从flat中删除自身对象以及子对象
        flat = flat.filter(k => k !== item && !k.path.includes('>' + item.id + '>'));

        //删除物理节点
        item.wrapperDom.remove();
        //回调
        callback && callback(item, flat);
        return { flat, tree };
    },
    //创建tree的数据对象（还未push到数组里，也未生成节点）
    createData: (source, target, flat, isChild) => {
        let obj = {},
            sourceType = axType(source),
            newId = axIncreaseId(flat),
            //如果source是字符串，那么直接填为label；source为空或者是其他类型则自动生成label
            newLabel = sourceType === 'String' ? source : '新节点' + newId;
        if (sourceType === 'Object') {
            //source为带pId和id的对象（一维的tree）
            let attrs = isChild ? {
                path: target.path + '>' + source.id,
                floor: target.floor + 1
            } : {
                //如果是添加兄弟元素则target的path去掉target的id换成source的id便是source的path
                path: target.path.replace(new RegExp('(.*)' + target.id), '$1' + source.id),
                floor: target.floor
            };
            //source是一个对象，但是如果没有id属性则用自动id
            !source.hasOwnProperty('id') ? source.id = newId : null;
            //合并属性
            obj = Object.assign(source, attrs);
        } else {
            //source为字符串
            obj = isChild ? {
                id: newId,
                label: newLabel,
                pId: target.id,
                path: target.path + '>' + newId,
                floor: target.floor + 1
            } : {
                id: newId,
                label: newLabel,
                pId: target.pId,
                //如果是添加兄弟元素则target的path去掉target的id换成source的id便是source的path
                path: target.path.replace(new RegExp('(.*)' + target.id), '$1' + newId),
                floor: target.floor
            };
        }
        return obj;
    },
    //如果子元素存在（ul存在）则push到children（ul）里
    existChildPush: (source, target, isFront = true) => {
        //source={id,pId,wrapperDom,labelDom,label,...}
        if (isFront) {
            target.children.unshift(source);
            target.bodyDom.insertAdjacentElement('afterBegin', source.wrapperDom);
        } else {
            target.children.push(source);
            target.bodyDom.insertAdjacentElement('beforeEnd', source.wrapperDom);
        }
    },
    //如果子元素不存在（ul不存在）则将target转成有ul的元素并push到children（ul）里
    noExistChildPush: (source, target, childToParent) => {
        //source={id,pId,wrapperDom,labelDom,label,...}
        //新增children属性
        target.children = [];
        target.children.push(source);
        //将target转成有ul的节点，增加bodyDom属性
        target = childToParent(target);
        //修改target的wrapper节点
        target.bodyDom.appendChild(source.wrapperDom);
        target.wrapperDom.appendChild(target.bodyDom);
    },
    //从路径获取数据
    getItemsFromPath(opt) {
        let options = Object.assign({ connector: '/', prop: 'label' }, opt),
            paths = options.path.split('>'),
            items = paths.map(k => options.data.find(i => i.id == k)).filter(Boolean),
            path = items.map(k => k[options.prop]).join(options.connector);
        return { items, path };
    },
    //获取完整的他tree数据
    getFullData: async function (opt = {}) {
        //options={source:'',async:'',idStart:'',rootStart:'',replace:'',instance:'',xhrName:'',ajaxData:'',ajaxType:'',before:'',success:'',addPath:'',callback:''}
        //data是必填项
        if (axIsEmpty(opt.source)) {
            return false;
        }
        let dataType = axType(opt.source),
            options = Object.assign({ idStart: 0, rootStart: -1, replace: false, ajaxData: {}, ajaxType: 'post', async: false, addPath: true }, opt),
            treeData = [],
            result = {},
            getFromNode = (node) => {
                let nodeName = node.nodeName, arr = [];
                if (nodeName === 'SELECT' || nodeName === 'DATALIST') {
                    arr = axSelectToArr(node, options.idStart, options.replace).data;
                } else if (nodeName === 'UL' || nodeName === 'OL') {
                    arr = axUlToArr(node, options.idStart, options.replace).data;
                }
                return arr;
            },
            getFromArray = (data) => {
                let arr = [], copyData = axClone(data);
                if (copyData[0].hasOwnProperty('pId') && copyData[0].hasOwnProperty('id')) {
                    //一维数组
                    arr = axArrToTree(copyData);
                } else {
                    //多维数组
                    arr = copyData;
                }
                return arr;
            };

        //如果填了ajaxData，则async自动为true
        !axIsEmpty(options.ajaxData) ? options.async = true : null;

        if (options.async) {
            //异步获取数据
            //source={url:'',data:{},loadDom:'',xhrName:'',before:''}
            let loadDom = options.source.loadDom ? axIdToDom(options.source.loadDom) : null;
            await axAjax({
                //根据参数进行数据扩展
                url: options.source,
                data: !axIsEmpty(options.ajaxData) ? options.ajaxData : '',
                type: options.ajaxType,
                opened: (response) => {
                    if (options.opened) {
                        //有opened回调
                        options.opened(response);
                    } else {
                        //给实例追加xhrName
                        options.instance && options.xhrName ? options.instance[options.xhrName] = response.xhr : null;
                    }
                },
                before: (response) => {
                    if (options.before) {
                        //有before回调
                        options.before(response);
                    } else {
                        //如果提供了加载的loadDom
                        loadDom ? loadDom.setAttribute('loading', '') : null;
                    }
                },
                success: (response) => {
                    if (axIsEmpty(response.content)) {
                        console.warn('No data obtained!');
                        return false;
                    }
                    //清除laoding属性
                    loadDom && loadDom.hasAttribute('loading') ? loadDom.removeAttribute('loading') : null;
                    //有success回调
                    options.success && options.success(response);
                    //获取值
                    treeData = getFromArray(response.content);
                }
            });
        }
        else {
            if (dataType.includes('HTML')) {
                //ul+li、ol+li、select、datalist物理节点，
                treeData = getFromNode(options.source);
            } else if (dataType === 'String') {
                //#id、.className、nodeName
                if (axStrType(options.source)) {
                    //为true说明节点存在
                    treeData = getFromNode(axIdToDom(options.source));
                }
            } else if (dataType === 'Array') {
                //原始数组，可能是树形也可能是一维带pId数组
                treeData = getFromArray(options.source);
            }
        }

        //获得输出值
        result = options.addPath ? this.addPath(treeData, options.rootStart, options.idStart) : treeData;
        //回调
        options.callback && options.callback(result);
        return result;
    }
}
/**
 * 日期相关工具
 */
const axDateMethod = {
    //短横字符串转斜杆字符串
    toSlash: function (dateObj) {
        //2022-1-23 21:12:16转成2022/1/23 21:12:16 
        return dateObj.replace('-', '/');
    },
    //2021-2-23或{YYYY:'',MM:'',DD:''}格式日期转成new Date()格式
    toDate: function (dateObj, type) {
        //都转new Date()格式
        let dateType = axType(dateObj),
            date;
        if (dateType === 'Date') {
            //克隆
            date = new Date(dateObj.valueOf());
        } else if (dateType === 'String') {
            //转成斜杆分隔符
            date = new Date(this.toSlash(dateObj));
        } else if (dateType === 'Object') {
            //是{YYYY:'',MM:''}对象
            date = new Date();
            dateObj.hasOwnProperty('YYYY') ? date.setFullYear(dateObj.YYYY) : null;
            dateObj.hasOwnProperty('MM') ? date.setMonth(dateObj.MM) : null;
            dateObj.hasOwnProperty('DD') ? date.setDate(dateObj.DD) : null;
            dateObj.hasOwnProperty('hh') ? date.setHours(dateObj.hh) : null;
            dateObj.hasOwnProperty('mm') ? date.setMinutes(dateObj.mm) : null;
            dateObj.hasOwnProperty('ss') ? date.setSeconds(dateObj.ss) : null;
        } else {
            //今天
            date = new Date();
        }
        //归零
        if (type === 'date') {
            date.setHours(0);
            date.setMinutes(0);
            date.setSeconds(0);
        } else if (type === 'month') {
            date.setDate(1);
            date.setHours(0);
            date.setMinutes(0);
            date.setSeconds(0);
        } else if (type === 'year') {
            date.setMonth(0);
            date.setDate(1);
            date.setHours(0);
            date.setMinutes(0);
            date.setSeconds(0);
        }
        return date;
    },
    //2021-2-23或Date格式日期转成{YYYY:'',MM:'',DD:''}对象格式
    toObj: function (obj, fillTime = false, type) {
        let objType = axType(obj),
            date,
            result = '';
        if (objType === 'Object') {
            //obj={YYYY:2033,MM:5}
            result = obj;
        } else if (objType === 'String' && !obj.trim().includes(' ') && obj.includes(':')) {
            //obj='2:5:6'
            let timeStrArr = obj.split(':');
            result = { hh: ~~timeStrArr[0], mm: ~~timeStrArr[1], ss: ~~timeStrArr[2] };
        } else {
            date = this.toDate(obj);
            result = { YYYY: date.getFullYear(), MM: date.getMonth(), DD: date.getDate(), hh: date.getHours(), mm: date.getMinutes(), ss: date.getSeconds() }
        }
        //补充time为0
        if (fillTime) {
            !result.hasOwnProperty('hh') ? result.hh = 0 : null;
            !result.hasOwnProperty('mm') ? result.mm = 0 : null;
            !result.hasOwnProperty('ss') ? result.ss = 0 : null;
        }

        //此时result是规范的daytime对象

        //年和月模式不需要时间
        if (type === 'year') {
            Reflect.deleteProperty(result, 'MM');
            Reflect.deleteProperty(result, 'DD');
            Reflect.deleteProperty(result, 'hh');
            Reflect.deleteProperty(result, 'mm');
            Reflect.deleteProperty(result, 'ss');
        } else if (type === 'month') {
            Reflect.deleteProperty(result, 'DD');
            Reflect.deleteProperty(result, 'hh');
            Reflect.deleteProperty(result, 'mm');
            Reflect.deleteProperty(result, 'ss');
        } else if (type === 'date') {
            Reflect.deleteProperty(result, 'hh');
            Reflect.deleteProperty(result, 'mm');
            Reflect.deleteProperty(result, 'ss');
        }
        return result;
    },

    //时间转成对象'12:00:12'=>{hh:12,mm:0,ss:12}
    strToTimeObj: function (str) {
        //去掉空格
        str = str.replace(' ', '');
        let arr = str.split(':');
        return { hh: ~~arr[0] || 0, mm: ~~arr[1] || 0, ss: ~~arr[2] || 0 }
    },
    //获得当前日期所在月份的天数
    getDays: function (dateObj) {
        let curDate = this.toDate(dateObj),
            curYear = curDate.getFullYear(),
            curMonth = curDate.getMonth();
        return new Date(curYear, curMonth + 1, 0).getDate();
    },
    //根据当前日期推算出昨天和明天
    getChain: function (dateObj, options = { YYYY: 0, MM: 0, DD: 0, hh: 0, mm: 0, ss: 0 }) {
        let curDate = this.toDate(dateObj);
        //重置日期
        options.hasOwnProperty('ss') ? curDate.setDate(curDate.getSeconds() + options.ss) : null;
        options.hasOwnProperty('mm') ? curDate.setDate(curDate.getMinutes() + options.mm) : null;
        options.hasOwnProperty('hh') ? curDate.setDate(curDate.getHours() + options.hh) : null;
        options.hasOwnProperty('DD') ? curDate.setDate(curDate.getDate() + options.DD) : null;
        options.hasOwnProperty('MM') ? curDate.setMonth(curDate.getMonth() + options.MM) : null;
        options.hasOwnProperty('YYYY') ? curDate.setFullYear(curDate.getFullYear() + options.YYYY) : null;
        return curDate;
    },
    //不足10则补零
    fillZero: function (obj) {
        //先转数字
        let num = ~~obj;
        return num < 10 ? '0' + num : num;
    },
    //将日期格式或日期对象根据format要求填充为相应的字符串
    fillFormat: function (date, format) {
        //date可以是一个日期new Date()或日期对象{YYYY:'',MM:''}
        let YYYY, MM, DD, hh, mm, ss, WW, weeks = ['日', '一', '二', '三', '四', '五', '六'];
        if (date instanceof Date) {
            YYYY = date.getFullYear();
            MM = date.getMonth() + 1;
            DD = date.getDate();
            hh = date.getHours();
            mm = date.getMinutes();
            ss = date.getSeconds();
            WW = date.getDay();
        } else {
            YYYY = date.YYYY;
            MM = date.MM + 1;
            DD = date.DD;
            hh = date.hh;
            mm = date.mm;
            ss = date.ss;
            date.hasOwnProperty('DD') ? WW = new Date(date.YYYY, date.MM, date.DD).getDay() : null;
        }
        //年
        format.includes('YYYY') ? format = format.replace('YYYY', YYYY) : null;
        //月
        format.includes('MM') ? format = format.replace('MM', this.fillZero(MM)) : null;
        //日
        format.includes('DD') ? format = format.replace('DD', this.fillZero(DD)) : null;
        //时
        format.includes('hh') ? format = format.replace('hh', this.fillZero(hh)) : null;
        //分
        format.includes('mm') ? format = format.replace('mm', this.fillZero(mm)) : null;
        //秒
        format.includes('ss') ? format = format.replace('ss', this.fillZero(ss)) : null;
        //周几
        format.includes('WW') && WW !== undefined ? format = format.replace('WW', weeks[WW]) : null;
        return format;
    },
    //对日期排序
    sort: function (data) {
        return data.sort((a, b) => { return this.toDate(a) - this.toDate(b) });
    },
    //去掉头尾空格，中间多个空格转为1个空格
    trim: function (str) {
        return str.trim().replace(/\s+/ig, ' ');
    },
    //根据日期字符串格式判断面板类型（'2012-2-3 12:12:00 => daytime'）
    getType: function (format) {
        let type = '';
        if (RegExp(/^(?=.*YYYY)(?=.*MM)(?=.*DD)(?=.*hh).*$/).test(format)) {
            //标准面板，包含年月日时间
            type = 'datetime';
        } else if (RegExp(/^(?=.*YYYY)(?=.*MM)(?=.*DD).*$/).test(format)) {
            //日期面板，包含年月日
            type = 'date';
        } else if (RegExp(/^(?=.*YYYY)(?=.*MM).*$/).test(format)) {
            //月份面板，包含年月
            type = 'month';
        } else if (RegExp(/^(?=.*YYYY).*$/).test(format)) {
            //年份面板，只包含年
            type = 'year';
        } else if (RegExp(/^(?=.*hh).*$/).test(format)) {
            //时间面板，只包含时间
            type = 'time';
        }
        return type;
    },
    isSameDay: function (a, b) {
        let x = { ...a },
            y = { ...b };
        Reflect.deleteProperty(x, 'hh');
        Reflect.deleteProperty(x, 'mm');
        Reflect.deleteProperty(x, 'ss');
        Reflect.deleteProperty(y, 'hh');
        Reflect.deleteProperty(y, 'mm');
        Reflect.deleteProperty(y, 'ss');
        return JSON.stringify(x) == JSON.stringify(y);
    }
};

/**
 * 判断是否在元素矩形区域之外
 */
const axIsOutside = (ev, id, ox, oy) => {
    //元素为空或不存在则直接返回true
    if (!id) { return true }
    let elem = axIdToDom(id);
    if (!elem) { return true }
    let x = ev.clientX,
        y = ev.clientY,
        x1, y1, x2, y2;
    ox ? x1 = ox : x1 = elem.getBoundingClientRect().left;
    oy ? y1 = oy : y1 = elem.getBoundingClientRect().top;
    x2 = x1 + elem.offsetWidth;
    y2 = y1 + elem.offsetHeight;
    return (x < x1 || x > x2 || y < y1 || y > y2) ? true : false;
}
/**
 * 封装getComputedStyle方法
 */
const axStyle = (elem, pseudo) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem);
    return document.defaultView.getComputedStyle(element, pseudo);
}
/**
 * 根据动画执行时间判断动画完成与否
 * 代替settimeout
 */
const axListenAnimation = (options = { duration, doing, done }) => {
    if (!options.duration) {
        return false;
    }
    let deadline = new Date().getTime() + options.duration,
        listen = () => {
            let newTime = new Date().getTime(), loop;
            if (newTime >= deadline) {
                options.done && options.done(listen);
                //取消动画请求
                loop ? cancelAnimationFrame(listen) : null;
            } else {
                options.doing && options.doing(listen);
                //继续循环动画
                loop = requestAnimationFrame(listen);
            }
        }
    listen();
    return listen;
}
/**
 * 从style中获得transform值 
 * 返回值格式：{rotate:0,scale:{x:1,y:1},translate:{x:0,y:0},skew:{x:10,y:15}}
 */
const axGetStyleTransform = (elem, types = ['translate', 'scale', 'rotate', 'skew'], instead = false) => {
    let dom = axIdToDom(elem),
        style = axStyle(dom),
        value = dom.style.transform,
        result = {
            rotate: 0,
            scale: { x: 1, y: 1 },
            translate: { x: 0, y: 0 },
            skew: { x: 0, y: 0 }
        }
    if (value) {
        let getReg = (type) => new RegExp(`${type}\\(([^\\)]+)\\)`),
            getText = (type) => value.match(getReg(type));
        //获得translate（整数）
        if (types.includes('translate') && value.includes('translate') && getText('translate')) {
            if (getText('translate')[1].includes(',')) {
                let arr = getText('translate')[1].split(',').map(k => parseInt(k));
                result.translate = { x: arr[0], y: arr[1] };
            } else {
                result.translate = { x: parseInt(getText('translate')[1]), y: 0 };
            }
        }
        //获得scale（浮点）
        if (types.includes('scale') && value.includes('scale') && getText('scale')) {
            if (getText('scale')[1].includes(',')) {
                let arr = getText('scale')[1].split(',').map(k => parseFloat(k));
                result.scale = { x: arr[0], y: arr[1] };
            } else {
                result.scale = { x: parseFloat(getText('scale')[1]), y: parseFloat(getText('scale')[1]) };
            }
        }
        //获得rotate（浮点）
        if (types.includes('rotate') && value.includes('rotate') && getText('rotate')) {
            result.rotate = parseFloat(getText('rotate')[1]);
        }
        //获得skew（浮点）
        if (types.includes('skew') && value.includes('skew') && getText('skew')) {
            if (getText('skew')[1].includes(',')) {
                let arr = getText('skew')[1].split(',').map(k => parseFloat(k));
                result.skew = { x: arr[0], y: arr[1] };
            } else {
                result.skew = { x: parseFloat(getText('skew')[1]), y: 0 };
            }
        }
    }
    //用position的left和top代替translate的x和y
    if (instead) {
        result.translate = { x: parseInt(style.left), y: parseInt(style.top) };
    }
    return result;
}
/**
 * 设置style的transform值，可设置scale、rotate和translate值
 * data格式：{rotate:0,scale:{x:1,y:1}||1,translate:{x:0,y:0}||0,skew:{x:10,y:15}||10}
 */
const axSetStyleTransform = (elem, data, instead = false) => {
    //data={translate:{x:0,y:0},scale:{x:0,y:0},rotate:0}
    let dom = axIdToDom(elem),
        transformTxt = dom.style.transform,
        getReg = (type) => { return new RegExp(`${type}\\(([^\\)]+)\\)`, 'gmi') };
    //没有数据则不再执行
    if (axIsEmpty(data)) {
        return false;
    }
    for (let k in data) {
        if (data.hasOwnProperty(k)) {
            let value = '';
            //设置transform值，优先设置translate
            if (k === 'translate') {
                if (instead) {
                    //修改position的left和top值来代替修改translateX和translateY值
                    !isNaN(data[k]) ?
                        (dom.style.left = `${data[k]}px`) :
                        (dom.style.left = `${data[k].x}px`, dom.style.top = `${data[k].y}px`);
                } else {
                    //传入是一个数值，则为translate(30px)，等同于translateX(30px)和translate(30px,0px)
                    value = !isNaN(data[k]) ? `${k}(${data[k]}px)` : `${k}(${data[k].x}px,${data[k].y}px)`;
                    if (transformTxt.includes(k)) {
                        if (transformTxt.startsWith('translate')) {
                            //如果是translate开头则直接替换
                            transformTxt = transformTxt.replace(getReg(k), value);
                        } else {
                            //如果不是则将translate新值追加到最前面
                            transformTxt = transformTxt.replace(getReg(k), '');
                            transformTxt = `${value} ${transformTxt}`;
                        }
                    } else {
                        transformTxt = `${value} ${transformTxt}`;
                    }
                }

            } else {
                if (k === 'scale') {
                    //传入是一个数值，则为scale(2)，等同于scale(2,2)
                    value = !isNaN(data[k]) ? `${k}(${data[k]})` : `${k}(${data[k].x},${data[k].y})`;
                } else if (k === 'rotate') {
                    value = `${k}(${data[k]}deg)`;
                } else if (k === 'skew') {
                    //传入是一个数值，则为skew(30deg)，等同于skew(30deg,0deg)
                    value = !isNaN(data[k]) ? `${k}(${data[k]}deg)` : `${k}(${data[k].x}deg,${data[k].y}deg)`;
                }
                if (transformTxt.includes(k)) {
                    transformTxt = transformTxt.replace(getReg(k), value);
                } else {
                    transformTxt += ` ${value}`;
                }
            }

        }
    }
    //设置
    dom.style.transform = transformTxt;
    return transformTxt;
}
/**
 * 定义模板引擎
 */
const axTplEngine = (html, data) => {
    if (!html) { console.error('No html!'); return false; }

    let re = /<#([^#>]+)?#>/g,
        reExp = /(^( )?(if|for|else|switch|case|break|{|}))(.*)?/g,
        code = 'let r=[];\n',
        cursor = 0;
    let match;
    let add = (line, js) => {
        js ? (code += line.match(reExp) ? line + '\n' : 'r.push(' + line + ');\n') :
            (code += line != '' ? 'r.push("' + line.replace(/"/g, '\\"') + '");\n' : '');
        return add;
    }
    while (match = re.exec(html)) {
        add(html.slice(cursor, match.index))(match[1], true);
        cursor = match.index + match[0].length;
    }
    add(html.substr(cursor, html.length - cursor));
    code += 'return r.join("");';
    return new Function(code.replace(/[\r\t\n]/g, '')).apply(data);
}
/**
 * 定义深度拷贝函数
 */
const axClone = (data) => {
    let cloneObj = (obj) => {
        if (obj === null) {
            return null;
        } else if (typeof obj !== 'object') {
            return obj;
        } else if (obj.constructor === Date) {
            return new Date(obj);
        } else {
            let newObj = {};
            //遍历属性
            for (let k in obj) {
                if (obj[k] == null) {
                    newObj[k] = null;
                } else if ((typeof obj[k]) == 'object' && !obj[k].nodeType) {
                    newObj[k] = axClone(obj[k]);
                    if (obj[k] instanceof Array) {
                        let newArray = [];
                        //数组里如果有对象
                        for (let i of obj[k]) {
                            newArray.push(axClone(i));
                        }
                        newObj[k] = newArray;
                    }
                } else {
                    newObj[k] = obj[k];
                }
            }
            return newObj;
        }
    }
    //
    if (Array.isArray(data)) {
        //是数组
        return data.map(k => cloneObj(k));
    } else {
        //非数组
        return cloneObj(data);
    }
};
/**
 * 定义合并属性递归函数
 */
const axMerge = (target, source, option = { arrAppend: false, clone: false, attrAppend: true }) => {
    if (typeof target != 'object' || typeof source != 'object') {
        return source;
    }
    if (!source) {
        return target;
    }
    //如果是复制方法，则先复制target
    let obj = option.copy ? axClone(target) : target;

    for (let k in source) {
        if (source.hasOwnProperty(k) && obj.hasOwnProperty(k)) {
            if (!obj[k]) {
                //如果原对象值为null，undefined或''，则直接用新数据覆盖
                obj[k] = source[k];
            } else {
                if (axType(source[k]) != 'Array') {
                    //如果遇上非数组
                    obj[k] = axMerge(obj[k], source[k], option);
                } else {
                    //如果遇上数组
                    if (option.arrAppend) {
                        //如果arrAppend为true追加模式，则在尾部追加数组
                        obj[k].push(...source[k]);
                    } else {
                        //否则直接覆盖
                        obj[k] = source[k];
                    }
                }
            }

        } else if (source.hasOwnProperty(k) && !obj.hasOwnProperty(k) && option.attrAppend) {
            //如果source有属性，obj没有该属性，但是option允许追加属性则直接赋值
            obj[k] = source[k];
        }
    }

    return obj;
}
/**
 * 字符串转成数组，'[]'，'{"":""},{"":""}'，'a,b,c'均转成数组
 */
const axStrToArr = (string, dftAttr, dftValue) => {
    let array = [],
        dft = {};
    //根据默认属性和值创建默认对象，用于扩展数组
    if (!axIsEmpty(dftAttr)) {
        dftAttr.forEach((item, index) => {
            dft[item] = !axIsEmpty(dftValue) ? dftValue[index] : '';
        })
    }

    if (string.substr(0, 1) == '[' && string.substr(-1) == ']') {
        //如果是'[]'格式，直接转数组
        array = JSON.parse(string);
    } else if (string.replace(/\s*/g, "").indexOf('},{') != -1) {
        //如果是'{"":""},{"":""}'格式，则转为数组格式
        array = JSON.parse('[' + string + ']');
    } else if (string.replace(/\s*/g, "").substr(0, 1) == '{' && string.replace(/\s*/g, "").substr(-1) == '}') {
        //如果是'{"":""}'格式，则转为length=1的数组
        array = JSON.parse('[' + string + ']');
    } else {
        //如果是'a,b,c'格式，使用英文逗号分割为数组
        array = string.split(',');

        array.forEach((item, index) => {
            //去掉左右空格
            item = item.trim();
            if (typeof item == 'string' && !axIsEmpty(dftAttr)) {
                let obj = {};
                dftAttr.forEach((str, i) => {
                    if (i == 0) {
                        obj[str] = item;
                    } else {
                        obj[str] = '';
                    }
                });
                //文本替换为带media属性的对象
                array.splice(index, 1, obj)
            }
        });
    }
    //根据默认属性和默认值统一格式
    array.forEach(item => {
        if (!axIsEmpty(dft)) {
            for (let k in dft) {
                if (!item.hasOwnProperty(k)) {
                    item[k] = dft[k] ? dft[k] : '';
                } else if (item.hasOwnProperty(k) && !item[k]) {
                    item[k] = dft[k];
                }
            }
        }

    })
    return array;
}
/**
 * HTML模板转成数组，支持列表循环
 */
const axTplToArr = (template, roles, keys, itemNode, dft) => {
    let tplDom = axIdToDom(template),
        data = [];
    if (itemNode) {
        //如果是列表形式
        let items = tplDom.querySelectorAll(itemNode);
        items.forEach(item => {
            //遍历子元素
            let obj = {};
            roles.forEach((i, index) => {
                //遍历roles并将与roles对应keys的值赋予obj
                let child = item.querySelector('[' + i + ']');
                if (child) {
                    obj[keys[index]] = child.getAttribute(i) ? child.getAttribute(i).trim() : child.innerHTML.trim();
                } else {
                    obj[keys[index]] = dft[index] || '';
                }
            })
            //追加入data
            data.push(obj);
        })

    } else {
        //如果是非列表形式
        let obj = {};
        roles.forEach((i, index) => {
            //遍历roles并将与roles对应keys的值赋予obj
            let child = tplDom.querySelector('[' + i + ']');
            if (child) {
                obj[keys[index]] = child.getAttribute(i) ? child.getAttribute(i).trim() : child.innerHTML.trim();
            } else {
                obj[keys[index]] = dft[index] || '';
            }
        })
        //追加入data
        data.push(obj);
    }
    return data;
}
/**
 * 将元素data-*属性与对象合并为新对象；
 * 1、先将对象obj复制保存为copy；
 * 2、将复制的对象name全转成小写；
 * 3、将data-*属性存为三层结构的对象，name均为小写并与copy合并；
 * 4、将合并后的copy与原obj对照还原大写name，并最终输出合并后的obj
 */
const axDataset = (element, obj) => {
    let elem = element ? axIdToDom(element) : null;
    let integer = /^\d+(\.\d+)?$/; //整数正则
    let float = /^(-(([0-9]+\.[0-9]*[1-9][0-9]*)|([0-9]*[1-9][0-9]*\.[0-9]+)|([0-9]*[1-9][0-9]*)))$/; //浮点正则

    //元素存在并且存在data属性
    if (elem) {
        let attrs = [...elem.attributes].filter((child) => {//获取元素的data属性集
            return child.name.indexOf('data-') == 0;
        });
        if (!axIsEmpty(attrs)) {
            //拷贝原参数并将属性名转为小写
            let lowerDft = axClone(obj);
            //循环转成小写属性
            let loopLow = (arg) => {
                for (let k in arg) {
                    let lower = k.toLowerCase();
                    arg[lower] = arg[k];
                    if (axType(arg[k]) === 'Object') {
                        loopLow(arg[k]);
                    }
                    //删除被原驼峰的属性
                    if (lower != k) {
                        delete arg[k];
                    }
                }
            }
            loopLow(lowerDft);
            //将data-*值赋予小写属性
            for (let k = 0; k < attrs.length; k++) {
                let value = attrs[k].value.trim(),
                    name = attrs[k].name;
                //将获得true和false字符串转成布尔值
                if (value == 'true') {
                    value = JSON.parse('true');
                } else if (value == 'false') {
                    value = JSON.parse('false');
                } else if (value == 'null') {
                    value = null;//null转成空对象
                } else if (value.substr(0, 1) == '[' && value.substr(-1) == ']') {
                    //如果是[{},{}]数字格式，直接转数组
                    value = JSON.parse(value);
                } else if (value.replace(/\s*/g, "").indexOf('},{') != -1) {
                    //如果是{},{},{}格式，则转为数组格式
                    value = JSON.parse('[' + value + ']');
                }
                //判断是否是数字字符串，是则转成数字
                if (typeof value == 'string' && (integer.test(value) || float.test(value))) {
                    value = Number(value);
                }
                //获得data-*的属性值
                let split = name.split('-');

                if (split.length - 1 == 1) {
                    lowerDft[split[1]] = value
                } else if (split.length - 1 == 2) {
                    lowerDft[split[1]][split[2]] = value
                } else if (split.length - 1 == 3) {
                    lowerDft[split[1]][split[2]][split[3]] = value
                } else {
                    console.error('Three levels are supported!');
                }
            }

            //将转小写的属性值赋予原参数
            let loopBack = (arg, opt) => {
                for (let k in arg) {
                    let lower = k.toLowerCase();
                    if (axType(arg[k]) === 'Object') {
                        loopBack(arg[k], opt[lower]);
                    } else {
                        arg[k] = opt[lower];
                    }
                }
            }
            loopBack(obj, lowerDft);

        }

    }

    return obj;
}
/**
 * 将元素ax*属性值得字符串形式转为对象形式
 */
const axAttrToJson = (elem, attr) => {
    let dom = axIdToDom(elem);
    if (!dom) {
        //节点不存在则不再执行下去
        console.warn('The attribute node does not exist!');
        return {};
    }
    let domAttr = dom.getAttribute(attr);
    if (!attr || !domAttr) {
        //没有填属性或属性值为空，直接返回空对象
        return {};
    }
    if (!domAttr.includes(':')) {
        //如果不是键值对，直接返回空对象
        return {};
    }
    let text = domAttr.trim();
    //开头没有{则两侧都加上大括号
    if (!text.startsWith('{')) {
        text = '{' + text + '}';
    }
    return new Function('return ' + text)();
}

//以下是标准对象格式，支持无限子集
/* let obj = {
    name: 'Mick',
    sex: 'male',
    school: {
        no: 'NO.5665',
        city: 'shenzhen',
        tel: '13800138000',
        zip: '518000'
    },
    teachers: {
        mathematics: {
            name: 'Lily',
            age: 29,
            sex: 'female',
        },
        physics: {
            name: 'Bob',
            age: 33,
            sex: 'male',
        },
    },
    callback: () => { },
} */
/**
 * 定义对象扩展
 * user属性可覆盖dft属性，并自动合并为新的dft 
 * DOM的data-*属性级别最高，使用短横“-”区分属性层级，最多允许使用三个短横
 * 举例说明： <button data-name="" data-school="" data-teachers-mathematics-name=""></button>
 */
const axExtend = (dft, user, element, attr) => {
    if (!dft) { console.error(axError.notExist); return false; }
    let elem = element ? axIdToDom(element) : null;
    //递归合并属性
    axMerge(dft, user);
    //合并data-*属性和dft属性
    //axDataset(elem, dft);
    elem ? axMerge(dft, axAttrToJson(elem, attr)) : null;
    return dft;
}
/**
 * 定义监听事件函数(观察者模式)
 */
//新增，data可为一个字符串，或以空格或逗号分开的字符串，或字符串数组
const axAddPlan = (data, handler, instance) => {
    let dataType = axType(data),
        dataArr = [],
        handleFun = (str) => {
            //使用hasOwnProperty，避免读取对象原型链
            if (!instance.handlers.hasOwnProperty(str)) {
                instance.handlers[str] = [];
            }
            //shown,hidden,show,hide,close,clear,confirm...
            let strHandle = handler.toString().replace(/\s/g, '').replace(/\;/g, '').match(/{(\S*)}/)[1],
                strHandles = instance.handlers[str].toString().replace(/\s/g, '').replace(/\;/g, '');
            strHandles.indexOf(strHandle) > -1 ? null : instance.handlers[str].push(handler);
        }
    //
    if (dataType === 'String') {
        dataArr = data.includes(',') ? data.split(',') : data.split(' ')
    } else if (dataType === 'Array') {
        dataArr = data;
    }
    if (dataArr.length > 0) {
        for (let k of dataArr) {
            handleFun(k);
        }
    }
}
//执行
const axExePlan = (data, instance, ...params) => {
    let dataType = axType(data),
        dataArr = [],
        handleFun = (str) => {
            //使用hasOwnProperty，避免读取对象原型链
            if (!instance.handlers.hasOwnProperty(str)) {
                return console.warn(`The ${str} event is unregistered or unusable!`);
            }
            instance.handlers[str].forEach(handler => {
                handler.call(instance, ...params);
            });
        }
    //
    if (dataType === 'String') {
        dataArr = data.includes(',') ? data.split(',') : data.split(' ')
    } else if (dataType === 'Array') {
        dataArr = data;
    }
    if (dataArr.length > 0) {
        for (let k of dataArr) {
            handleFun(k);
        }
    }

}
//删除
const axDelPlan = (data, handler, instance) => {
    let dataType = axType(data),
        dataArr = [],
        handleFun = (str) => {
            //使用hasOwnProperty，避免读取对象原型链
            if (!instance.handlers.hasOwnProperty(str)) {
                return console.warn(`The ${str} event is invalid or unusable!`);
            }
            if (!handler) {
                delete instance.handlers[str];
            } else {
                let index = instance.handlers[str].findIndex(elem => elem === handler)
                if (index === undefined) {
                    return console.warn(`The ${str} event is not bound!`);
                }
                instance.handlers[str].splice(index, 1);
                if (instance.handlers[str].length === 0) {
                    delete instance.handlers[str];
                }
            }
        }
    //
    if (dataType === 'String') {
        dataArr = data.includes(',') ? data.split(',') : data.split(' ')
    } else if (dataType === 'Array') {
        dataArr = data;
    }
    if (dataArr.length > 0) {
        for (let k of dataArr) {
            handleFun(k);
        }
    }
}
/**
 * 新增一个节点元素
 */
const axAddElem = (node, attrs, content) => {
    //统一大小写
    let nodeName = node.toLowerCase().trim(),
        nodeDom = document.createElement(nodeName),
        loop = function (dom, data) {
            if (data === '' || data === null || data === undefined) {
                //为空、未定义则不再执行
                return false;
            }
            //template是DocumentFragment类型不能用insertAdjacentHTML
            if (nodeName === 'template') {
                dom.innerHTML = data;
            } else {
                if (Array.isArray(data)) {
                    //节点数组
                    if (data.length > 0) {
                        data.forEach(i => {
                            loop(dom, i);
                        });
                    }
                } else {
                    if (typeof data == 'object') {
                        //节点
                        dom.appendChild(data);
                    } else {
                        //字符串数字等
                        dom.insertAdjacentHTML('beforeEnd', data);
                    }
                }
            }
        }
    //添加属性
    if (attrs) {
        for (let k in attrs) {
            attrs.hasOwnProperty(k) ? nodeDom.setAttribute(k, attrs[k]) : null;
        }
    }
    //执行循环创建子节点
    loop(nodeDom, content);
    return nodeDom;
}
/**
 * 字符串转筛选语句返回筛选结果
 */
const axStrToFilter = (str, source, callback) => {
    let text = `source.filter(item =>${str})`,
        result = new Function('source', 'return ' + text)(source);
    callback && callback(result);
    return result;
}
/**
 * 替换字符串某位置的字符
 */
const axRangeReplace = (originText, replaceText, indexStart, indexStop, n) => {
    if (typeof originText !== 'string') {
        originText = JSON.stringify(originText);
    }
    if (typeof replaceText !== 'string') {
        originText = JSON.stringify(replaceText);
    }
    let result = '';
    if (!indexStart && !indexStop) {
        //如果没有起止位置，那么就完全替换
        result = replaceText;
    } else if (!indexStop) {
        //如果没有截止位置，则从indexStart之后都替换掉
        if (n) {
            result = originText.substring(0, indexStart) + replaceText + originText.substring(indexStart + n);
        } else {
            result = originText.substring(0, indexStart) + replaceText;
        }
    } else if (!indexStart) {
        //如果没有起始位置，则从indexStop之前都替换掉
        if (n) {
            result = originText.substring(indexStop - n) + replaceText + originText.substring(0, indexStart);
        } else {
            result = replaceText + originText.substring(indexStop);
        }
    } else {
        //有起止位置，则替换起止中间的内容
        result = originText.substring(0, indexStart) + replaceText + originText.substring(indexStop);
    }
    return result;
}
//将数组进行完全排列组合
const axArrange = (origin, n, index = 0, splice = [], result = []) => {
    let arrayLen = origin.length
    //可选数量小于项所需元素的个数，则递归终止
    if (index + n > arrayLen) {
        return;
    }
    for (let i = index; i < arrayLen; i++) {
        //n === 1的时候，说明splice在添加一个元素，就能生成一个新的完整项了。
        //debugger
        if (n === 1) {
            //再增加一个元素就能生成一个完整项，再加入到结果集合中
            result.push([...splice, origin[i]])
            //继续下一个元素生成一个新的完整项
            i + 1 < arrayLen && axArrange(origin, n, i + 1, splice, result)
            break;
        }
        //执行到这，说明n > 2，splice还需要两个以上的元素，才能生成一个新的完整项。则递归，往splice添加元素
        axArrange(origin, n - 1, i + 1, [...splice, origin[i]], result)
    }
    return result;
}
/**
 * 定义AJAX函数，异步获取内容
 */
const axAjax = (option, elem, callback) => {
    if (!option) { console.error('There is no options!'); return false; }
    let label = axAddElem('SPAN', { 'ajax-message': '' }),
        dom = elem ? axIdToDom(elem) : null;
    dom ? dom.appendChild(label) : null;
    let dft = {
        url: '',
        selector: '',
        type: 'post',
        async: true,
        data: {},
        delay: 60000,
        contentType: '',
        headers: {},
        responseType: '',
        catchable: false,
        abort: () => {
            dom ? (label.innerHTML = '<i class="ax-color-warning">中止了请求！</i>', dom.setAttribute('ajax-result', 'abort')) : console.error('中止了请求！');
        },
        timeout: () => {
            dom ? (label.innerHTML = '<i class="ax-color-error">请求超时了！</i>', dom.setAttribute('ajax-result', 'timeout')) : console.error('请求超时了！');
        },
        loading: '<span class="ax-loading"><i></i></span>',
        opened: '',
        before: (response) => {
            dom ? (label.innerHTML = response.content, dom.setAttribute('ajax-result', 'before')) : null;
        },
        downloading: () => { },
        uploading: () => { },
        success: (response) => {
            dom ? (dom.innerHTML = response.content, dom.setAttribute('ajax-result', 'success')) : null;
        },
        error: (response) => {
            dom ? (label.innerHTML = `<i class="ax-color-error">错误状态：${response.status}</i>`, dom.setAttribute('ajax-result', 'error')) : console.error(`错误状态：${response.status}`);
        },
    };
    //初始化参数
    for (let k in option) {
        dft[k] = option[k];
    }

    //创建XMLHttpRequest
    let xhr = '';
    if (window.XMLHttpRequest) {
        xhr = new XMLHttpRequest();
    }
    else {
        xhr = new ActiveXObject("Microsoft.XMLHTTP");
    }

    //设置发送数据和预设请求头
    let params = '';
    if (!axIsEmpty(dft.data)) {
        if (axType(dft.data) == 'FormData') {
            //如果是new FormData格式，直接相等
            params = dft.data;
            //强行修正请求头，会发生意外
            //dft.contentType = 'multipart/form-data';
        } else if (axType(dft.data) == 'Object') {
            //如果是对象格式{name:'',age:''}
            if (dft.contentType.includes('json')) {
                //Content-Type=application/json
                params = JSON.stringify(dft.data);
            } else {
                //Content-Type=application/x-www-form-urlencoded
                for (let k in dft.data) {
                    params += '&' + k + '=' + dft.data[k];
                }
                dft.contentType = 'application/x-www-form-urlencoded';
            }
        } else {
            //如果是name=''&age=''字符串
            dft.data = dft.data.trim();
            //name=''&age=''或&name=''&age=''统一去掉第一个&
            params = dft.data.startsWith('&') || dft.data.startsWith('?') ? dft.data.slice(1) : dft.data;
            dft.contentType = 'application/x-www-form-urlencoded';
        }
    }
    //设置请求头
    if (dft.contType) {
        dft.headers['Content-Type'] = dft.contType;
        if (!dft.contType.includes('urlencoded') && dft.type.toLowerCase() === 'get') {
            console.error('Please request data with post!');
            return false;
        }
    }

    let setHeader = () => {
        //headers={'':'','':''}
        for (let k in dft.headers) {
            dft.headers[k] ? xhr.setRequestHeader(k, dft.headers[k]) : null;

        }
    },
        timeoutFun;
    //设置超时时间
    xhr.timeout = dft.delay;
    //设置接收数据类型
    xhr.responseType = dft.responseType;
    //定义取消函数
    let abort = () => {
        //解除timeout超时
        timeoutFun ? xhr.removeEventListener('timeout', timeoutFun) : null;
        //中止请求
        xhr.abort();
        //回调
        !axIsEmpty(dft.abort) ? dft.abort(dom, xhr) : null;
    }
    //定义promise
    let promise = new Promise((resolve, reject) => {
        //定义进度函数
        let progress = (e, callback) => {
            let current = e.loaded,
                total = e.total,
                ratio = 0,
                percent = 0,
                result = '0%',
                time = (new Date(e.timeStamp)).getTime();
            if (e.lengthComputable) {
                ratio = current / total;
                percent = Math.round(ratio * 100);
                result = percent + '%';
                callback && callback({ result, percent, current, total, ratio, time, dom });
            }
        }
        //监听上传进度
        xhr.upload.onprogress = (e) => {
            progress(e, (data) => {
                !axIsEmpty(dft.uploading) ? dft.uploading(data) : null;
            });
        }
        //监听下载进度
        xhr.onprogress = (e) => {
            progress(e, (data) => {
                !axIsEmpty(dft.downloading) ? dft.downloading(data) : null;
            });
        }
        //监听报错
        xhr.onerror = (e) => {
            let response = { content: xhr.response, status: xhr.status, dom, xhr }
            !axIsEmpty(dft.error) ? dft.error(response) : null;
            //reject只能接受一个参数
            dft.catchable ? reject(response) : null;
        }
        //监听超时
        timeoutFun = () => {
            //修改ajax状态
            let response = { content: xhr.response, status: xhr.status, dom, xhr }
            dom ? dom.setAttribute('ajax-result', 'timeout') : null;
            !axIsEmpty(dft.timeout) ? dft.timeout(response) : null;
            //reject只能接受一个参数
            dft.catchable ? reject(response) : null;
        };
        xhr.addEventListener('timeout', timeoutFun);
        //状态判断
        xhr.onreadystatechange = function () {
            if (xhr.readyState < 4) {
                //0-1-2-3的状态，使用加载器
                let response = { content: dft.loading, status: xhr.status, dom, xhr, abort }
                if (xhr.readyState === 1) {
                    //建立请求，但是还未发送
                    !axIsEmpty(dft.opened) ? dft.opened(response) : null;
                }
                !axIsEmpty(dft.before) ? dft.before(response) : null;
                callback && callback(response);
            } else {
                //4，已经获得返回数据
                if ((xhr.status >= 200 && xhr.status < 300) || xhr.status == 304) {
                    if (!dft.responseType || xhr.responseType == 'text') {
                        //返回文本类型
                        let div = document.createElement('div'),
                            content = '',
                            reg = /(<head\b[^<]*(?:(?!<\/head>)<[^<]*)*<\/head>)|(<\/?html.*?>)|(<\!DOCTYPE.*?html.*?>)|(<\/?body.*?>)/gi,
                            strEnd = axStrEnd(dft.url, '#'),
                            selector = dft.selector || strEnd;
                        div.innerHTML = xhr.responseText;
                        //如果有异步内容中能找到选择器，则调取选择器的内容
                        if (selector && div.querySelector(selector)) {
                            content = div.querySelector(selector).innerHTML;
                        } else if (selector && !div.querySelector(selector)) {
                            //如果有异步内容中能找不到选择器，则返回一个警告和整个异步页面内容
                            console.warn(`The node of "${selector}" is not exist!`);
                            //过滤掉head标签的内容并去掉html和body标签，这样子content将可获得body以内的内容（不包含body标签）
                            content = xhr.responseText.replace(reg, '').trim();
                        } else {
                            //没有设定选择器，则直接返回整个异步页面内容
                            //通过判断开头字符是{或[来确定异步页面是否是JSON内容，如果是则转成JSON对象
                            if (xhr.responseText.trim().substr(0, 1) == "{" || xhr.responseText.trim().substr(0, 1) == "[") {
                                content = JSON.parse(xhr.responseText.trim());
                            } else {
                                content = xhr.responseText.replace(reg, '').trim();
                            }
                        }
                        //修改ajax状态
                        dom ? dom.setAttribute('ajax-result', 'success') : null;
                        let response = { content, status: xhr.status, dom, xhr }
                        !axIsEmpty(dft.success) ? dft.success(response) : null;
                        //resolve只能接受一个参数
                        resolve(response);
                    } else {
                        //修改ajax状态
                        dom ? dom.setAttribute('ajax-result', 'success') : null;
                        //返回json、blob类型等
                        let response = { content: xhr.response, status: xhr.status, dom, xhr }
                        !axIsEmpty(dft.success) ? dft.success(response) : null;
                        resolve(response);
                    }

                }
                else {
                    let response = { content: xhr.response, status: xhr.status, dom, xhr }
                    !axIsEmpty(dft.error) ? dft.error(response) : null;
                    dft.catchable ? reject(response) : null;

                }
            }
        };

        //发送异步请求
        if (dft.type.toLowerCase() == 'get') {
            xhr.open('get', dft.url + '?now=' + new Date().getTime() + '&' + params, dft.async);
            setHeader();
            xhr.send();
        } else if (dft.type.toLowerCase() == 'post') {
            xhr.open('post', dft.url, dft.async);
            setHeader();
            xhr.send(params);
        }
    });
    //附加上xhr变量和取消方法
    promise.xhr = xhr;
    promise.abort = abort;
    return promise;
};
/**
 * ajax提交表单封装
 * url是异步提交地址，必填项
 * outer是form表单节点，必填项
 * target是点击的按钮，选填项
 * loading是缓冲加载显示位置，button是指在按钮上转圈（要求填写target），parent是指在form表单上转圈，选填项
 * type是指异步表单提交方式，默认post，可选择get，选填项
 * delay是指多久超时，单位毫秒，默认60秒
 * header是指设定请求头，数组格式，支持两种格式：['content-Type','']和[{name:'',value:''},...]
 * before是成功返回前的回调函数，选填项
 * success是成功返回后的回调函数，选填项
 */
const axAjaxSubmit = (url, outer, target, options = { type, success, loading, delay, before, error, opended, other }) => {
    if (!url) { console.error('Url is required!'); return false; }
    if (!outer) { console.error('Form node is required!'); return false; }
    options.type = options.type || 'post';
    options.other = options.other || {};
    let parentDom = axIdToDom(outer),
        targetDom = target ? axIdToDom(target) : null,
        targetText = targetDom ? targetDom.innerHTML : '',
        mode = options.loading || 'button',
        opt = Object.assign(options.other, {
            formClass: 'ax-form-submitting',
            buttonClass: 'ax-button-submitting',
            buttonText: '正在提交',
            overlayClass: 'ax-form-overlay',
            overlayShowClass: 'ax-show',
            successText: '恭喜！提交成功！',
            errorText: '对不起，提交失败！',
        }),
        overlayDom = parentDom.querySelector('.ax-form-overlay') ? parentDom.querySelector('.ax-form-overlay') : axAddElem('div', { class: 'ax-form-overlay' });
    //如果正在提交则禁止再次提交
    if (parentDom.classList.contains(opt.formClass)) {
        return false;
    }
    if (targetDom && targetDom.classList.contains(opt.buttonClass)) {
        return false;
    }
    //获得要传递的值
    let formData = {}, promise;
    //获得表单元素的值并转成对象，格式：{name:value,name:value}
    parentDom.querySelectorAll('[name]').forEach(function (item) {
        formData[item.getAttribute('name')] = item.value.trim();
    });
    promise = axAjax({
        url: url,
        data: formData,
        type: options.type,
        opened: (response) => {
            options.opened && options.opened(response);
        },
        before: (response) => {
            //给正在提交的表单增加样式
            parentDom.classList.add(opt.formClass);
            targetDom ? targetDom.classList.add(opt.buttonClass) : null;

            if (mode == 'button' && targetDom) {
                targetDom.innerHTML = opt.buttonText;
                targetDom.insertAdjacentHTML('afterbegin', '<i class="ax-btn-loading"></i>');
            } else {
                overlayDom.innerHTML = response.content;
                overlayDom.classList.add(opt.overlayClass);
                //页面没有该节点则直接在页面追加
                if (!parentDom.querySelector('.ax-form-overlay')) {
                    parentDom.appendChild(overlayDom);
                }
            }
            //追加回调
            options.before && options.before(response);
        },
        success: (response) => {
            setTimeout(() => {
                //取消表单样式
                parentDom.classList.remove(opt.formClass);
                targetDom ? targetDom.classList.remove(opt.buttonClass) : null;

                if (mode == 'button' && targetDom) {
                    //按钮恢复原样
                    targetDom.innerHTML = targetText;
                } else {
                    overlayDom.classList.remove(opt.overlayClass);
                }

                //信息窗告知成功
                new axMessage({
                    content: opt.successText,
                    result: 'success',
                    iconShow: true,
                });
                //追加回调
                options.success && options.success(response);
            }, options.delay === 0 ? 0 : options.delay || 1000);
        },
        error: (response) => {
            if (mode == 'button' && targetDom) {
                //按钮恢复原样
                targetDom.innerHTML = targetText;
            } else {
                overlayDom.classList.remove(opt.overlayClass);
            }
            new axMessage({
                content: opt.errorText,
                result: 'error',
                iconShow: true,
            });
            //追加回调
            options.error && options.error(response);
        }
    });
    return promise;
}
/**
 * 获取距离body的上偏移和左偏移
 */
const axOffset = (elem) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let totalLeft,
        totalTop,
        scrollLeft = document.documentElement.scrollLeft || window.pageXOffset || document.body.scrollLeft,
        scrollTop = document.documentElement.scrollTop || window.pageYOffset || document.body.scrollTop,
        //elem为'#ID'转成DOM
        targetDom = axIdToDom(elem);
    totalLeft = targetDom.getBoundingClientRect().left + scrollLeft;
    totalTop = targetDom.getBoundingClientRect().top + scrollTop;
    return {
        left: totalLeft,
        top: totalTop
    }
}
/**
 * 获得元素真实宽度，宽度构成包括：margin、padding、borderWidth、width
 */
const axWidth = (elem, type) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let _elem = axIdToDom(elem);
    let { width, paddingLeft, paddingRight, marginLeft, marginRight, borderLeftWidth, borderRightWidth } = getComputedStyle(_elem);
    width = parseInt(width, 10);
    paddingLeft = parseInt(paddingLeft, 10);
    paddingRight = parseInt(paddingRight, 10);
    marginLeft = parseInt(marginLeft, 10);
    marginRight = parseInt(marginRight, 10);
    borderLeftWidth = parseInt(borderLeftWidth, 10);
    borderRightWidth = parseInt(borderRightWidth, 10);
    if (type == 'inner') {
        return width + paddingLeft + paddingRight;
    } else if (type == 'outer') {
        return width + paddingLeft + paddingRight + marginLeft + marginRight + borderLeftWidth + borderRightWidth;
    } else {
        return width;
    }
}
/**
 * 获得元素真实高度，宽度构成包括：margin、padding、borderWidth、height
 */
const axHeight = (elem, type) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let _elem = axIdToDom(elem);
    let { height, paddingTop, paddingBottom, marginTop, marginBottom, borderTopWidth, borderBottomWidth } = getComputedStyle(_elem);
    height = parseInt(height, 10);
    paddingTop = parseInt(paddingTop, 10);
    paddingBottom = parseInt(paddingBottom, 10);
    marginTop = parseInt(marginTop, 10);
    marginBottom = parseInt(marginBottom, 10);
    borderTopWidth = parseInt(borderTopWidth, 10);
    borderBottomWidth = parseInt(borderBottomWidth, 10);
    if (type == 'inner') {
        return height + paddingTop + paddingBottom;
    } else if (type == 'outer') {
        return height + paddingTop + paddingBottom + marginTop + marginBottom + borderTopWidth + borderBottomWidth;
    } else {
        return height;
    }
}
/**
 * 获得元素高度构成部分，包括：height, paddingTop, paddingBottom, marginTop, marginBottom, borderTopWidth, borderBottomWidth
 */
const axArrHeight = (elem) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let _elem = axIdToDom(elem);
    let { height, paddingTop, paddingBottom, marginTop, marginBottom, borderTopWidth, borderBottomWidth } = getComputedStyle(_elem);
    height = parseInt(height, 10);
    paddingTop = parseInt(paddingTop, 10);
    paddingBottom = parseInt(paddingBottom, 10);
    marginTop = parseInt(marginTop, 10);
    marginBottom = parseInt(marginBottom, 10);
    borderTopWidth = parseInt(borderTopWidth, 10);
    borderBottomWidth = parseInt(borderBottomWidth, 10);
    return { height, paddingTop, paddingBottom, marginTop, marginBottom, borderTopWidth, borderBottomWidth };
};
/**
 * 获得元素宽度构成部分，包括：width, paddingLeft, paddingRight, marginLeft, marginRight, borderLeft, borderRight
 */
/* const axArrWidth = (elem) => {
    let { width, paddingLeft, paddingRight, marginLeft, marginRight, borderLeftWidth, borderRightWidth } = getComputedStyle(elem);
    width = parseInt(width, 10);
    paddingLeft = parseInt(paddingLeft, 10);
    paddingRight = parseInt(paddingRight, 10);
    marginLeft = parseInt(marginLeft, 10);
    marginRight = parseInt(marginRight, 10);
    borderLeft = parseInt(borderLeftWidth, 10);
    borderRight = parseInt(borderRightWidth, 10);
    return { width, paddingLeft, paddingRight, marginLeft, marginRight, borderLeft, borderRight };
}; */
/**
 *  判断是否选择了已知元素 ,返回 ture 或 false
 */
const axSelector = (elem, selector) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let _elem = axIdToDom(elem);
    let result = true;
    if (selector) {
        selector = String(selector).replace(/\s+/g, '');
        selector = String(selector).split(',');
        for (let i = 0; i < selector.length; i++) {
            if (!selector[i]) {
                result = true;
                break;
            } else if (axStrType(selector[i]) == 'class') {
                let string = selector[i].replace('.', '');
                if (_elem.classList.contains(string)) {
                    result = true;
                    break;
                } else {
                    result = false;
                }
            } else if (axStrType(selector[i]) == 'id') {
                let string = selector[i].replace('#', '');
                if (_elem.getAttribute('id') == string) {
                    result = true;
                    break;
                } else {
                    result = false;
                }
            } else {
                let string = selector[i].toUpperCase();
                if (_elem.nodeName == string) {
                    result = true;
                    break;
                } else {
                    result = false;
                }
            }
        }
    } else {
        result = true;
    }
    return result;
}
/**
 * 判断是否排除了已知元素 ,返回 ture 或 false
 */
const axNot = (elem, selector) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let _elem = axIdToDom(elem);
    let result = true;
    if (selector) {
        selector = String(selector).replace(/\s+/g, '');
        selector = String(selector).split(',');
        for (let i = 0; i < selector.length; i++) {

            if (axStrType(selector[i]) == 'class') {
                let string = selector[i].replace('.', '');
                if (_elem.classList.contains(string)) {
                    result = true;
                    break;
                } else {
                    result = false;
                }
            }
            else if (axStrType(selector[i]) == 'id') {
                let string = selector[i].replace('#', '');
                if (_elem.getAttribute('id') == string) {
                    result = true;
                    break;
                } else {
                    result = false;
                }
            } else if (selector) {
                let string = selector[i].toUpperCase();
                if (_elem.nodeName == string) {
                    result = true;
                    break;
                } else {
                    result = false;
                }
            } else {
                result = false;
            }
        }
    } else {
        result = false;
    }
    return result;
}
/**
 * 选择兄弟元素
 * element：选择的Dom
 * self：是否包含自己，默认false
 * selector：从指定的选择器中选择兄弟元素
 * ignore：不选择的兄弟元素
 * 举例：axSiblings(elem,'#my','.disabled')
 */
const axSiblings = (element, self, selector, ignore) => {
    if (!element) { console.error(axError.notExist); return false; }
    let elem = axIdToDom(element);
    return [...elem.parentNode.children].filter((child) => {
        if (!self) {
            return child !== elem && axSelector(child, selector) && !axNot(child, ignore);
        } else {
            return axSelector(child, selector) && !axNot(child, ignore);
        }
    });
}
/**
 * 选择子元素
 * element：选择的Dom
 * selector：从指定的选择器中选择兄弟元素
 * ignore：不选择的兄弟元素
 * 举例：axChildren(elem,'.item','.disabled')
 */
const axChildren = (element, selector, ignore) => {
    if (!element) { console.error(axError.notExist); return false; }
    let elem = axIdToDom(element);
    return [...elem.children].filter((child) => {
        return axSelector(child, selector) && !axNot(child, ignore);
    });
}
/**
 * 定义取得嫡系父层函数。
 * start：根节点
 * end：终节点
 * containStart：是否包含根节点
 * containEnd：是否包含终节点
 * parents：父节点数组
 */
const axParents = (start, end, containStart, containEnd, parents = []) => {
    if (!start) { console.error('No node to start!'); return false; }
    //将起止强制转为DOM对象
    let _start = axIdToDom(start),
        _end = end ? axIdToDom(end) : null;
    //父级标签是否是body,是则停止返回集合,反之继续
    if (_start.parentElement.nodeName !== 'BODY') {
        //是否将起点也加入集合
        (containStart && parents.indexOf(_start) < 0) ? parents.push(_start) : null;
        if (_end) {
            //如果有终点节点，根据参数加入
            (containEnd && parents.indexOf(_end) < 0) ? parents.push(_end) : null;
            //放入集合，如果start=end则不加入

            if (_start.parentElement !== _end) {
                parents.push(_start.parentElement);
                //再上一层寻找
                return axParents(_start.parentElement, _end, containStart, containEnd, parents)
            } else {
                //返回集合,结束
                return parents;
            }
        } else {
            //如果没有终点节点直接放入集合
            parents.push(_start.parentElement);
            //再上一层寻找
            return axParents(_start.parentElement, _end, containStart, containEnd, parents)
        }
    } else {
        //返回集合,结束
        return parents;
    }
}
/**
 * 定义动画曲线
 * 举例：axCurves.easeOutIn(0.3)
 */
const axCurves = {
    linear(t) { return t; },
    easeIn(t) { return t * t * t; },
    easeOut(t) { return (t - 1) * (t - 1) * (t - 1) + 1; },
    easeInOut(t) { return t < 0.5 ? 2 * t * t : -2(t - 1) * (t - 1) + 1; },
    easeOutIn(t) { return t < 0.5 ? Math.sqrt(0.5 * t) : -Math.sqrt(0.5 * (-t + 1)) + 1; }
};
/**
 * 定义axSlideDown动画
 * 举例：axSlideDown(elem)
 */
const axSlideDown = (elem, display, callback, duration, curve = 'easeOut') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem),
        dftTime = height => height / 3 + 250,
        rawStyle = element.style.cssText;
    display = display || 'block';
    if (window.getComputedStyle(element).display === "none") {
        //calculate height
        element.style.cssText = `${rawStyle}display: ${display};`;
        const { height, paddingTop, paddingBottom, marginTop, marginBottom } = axArrHeight(element);
        //set time
        let time;
        if (duration !== undefined && typeof duration === 'number') {
            time = duration;
        } else {
            time = dftTime(height);
        }
        //set activeClass
        element.style.cssText = `${rawStyle}display: ${display}; padding-top: 0; padding-bottom: 0;margin-top: 0; margin-bottom: 0;`;
        //animation
        const initTime = new Date().getTime();
        const repeat = () => {
            const newTime = new Date().getTime() - initTime;
            const timestep = newTime / time;
            const timefactor = axCurves[curve](timestep);
            const step = 0 + height * timefactor;
            const stepPaddingT = 0 + (paddingTop * timefactor);
            const stepPaddingB = 0 + (paddingBottom * timefactor);
            const stepMarginT = 0 + (marginTop * timefactor);
            const stepMarginB = 0 + (marginBottom * timefactor);
            if (newTime <= time && Math.ceil(step) < height) {
                element.style.cssText = `${rawStyle}overflow:hidden;display: ${display}; padding-top: ${stepPaddingT}px; padding-bottom: ${stepPaddingB}px;margin-top: ${stepMarginT}px; margin-bottom: ${stepMarginB}px; height: ${step}px`;
            } else {
                element.style.cssText = `${rawStyle}display: ${display};`;
            }
            //console.log(time + ',' + newTime)
            const repeatLoop = requestAnimationFrame(repeat);
            //如果对象是隐藏的（display:none）,会导致height值为NaN，如果是如此一次循环便结束了
            if (Math.ceil(step) >= height || !height) {
                cancelAnimationFrame(repeatLoop);
                callback && callback();
            }
        };
        repeat();
    }
    return element;
};
/**
 * 不判断元素是否隐藏的下拉展开动画
 */
const axHeightDown = (elem, arrHeight, callback, duration, curve = 'easeOut') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem),
        dftTime = height => height / 3 + 250;
    //计算高度
    const height = (arrHeight && arrHeight.height) ? arrHeight.height : axArrHeight(element).height,
        paddingTop = (arrHeight && arrHeight.paddingTop) ? arrHeight.paddingTop : axArrHeight(element).paddingTop,
        paddingBottom = (arrHeight && arrHeight.paddingBottom) ? arrHeight.paddingBottom : axArrHeight(element).paddingBottom,
        marginTop = (arrHeight && arrHeight.marginTop) ? arrHeight.marginTop : axArrHeight(element).marginTop,
        marginBottom = (arrHeight && arrHeight.marginBottom) ? arrHeight.marginBottom : axArrHeight(element).marginBottom;
    //设置time
    let time;
    if (duration !== undefined && typeof duration === 'number') {
        time = duration;
    } else {
        time = dftTime(height);
    }
    element.style.overflow = 'hidden';
    //初始化高度
    element.style.height = 0;
    paddingTop ? element.style.paddingTop = 0 : null;
    paddingBottom ? element.style.paddingBottom = 0 : null;
    marginTop ? element.style.marginTop = 0 : null;
    marginBottom ? element.style.marginBottom = 0 : null;
    //设置动画
    const initTime = new Date().getTime();
    const repeat = () => {
        const newTime = new Date().getTime() - initTime;
        const timestep = newTime / time;
        const timefactor = axCurves[curve](timestep);
        const step = 0 + height * timefactor;
        const stepPaddingT = 0 + (paddingTop * timefactor);
        const stepPaddingB = 0 + (paddingBottom * timefactor);
        const stepMarginT = 0 + (marginTop * timefactor);
        const stepMarginB = 0 + (marginBottom * timefactor);
        if (newTime <= time && Math.ceil(step) < height) {
            element.style.height = step + 'px';
            paddingTop ? element.style.paddingTop = stepPaddingT + 'px' : null;
            paddingBottom ? element.style.paddingBottom = stepPaddingB + 'px' : null;
            marginTop ? element.style.marginTop = stepMarginT + 'px' : null;
            marginBottom ? element.style.marginBottom = stepMarginB + 'px' : null;
        } else {
            element.style.height = height + 'px';
            paddingTop ? element.style.paddingTop = paddingTop + 'px' : null;
            paddingBottom ? element.style.paddingBottom = paddingBottom + 'px' : null;
            marginTop ? element.style.marginTop = marginTop + 'px' : null;
            marginBottom ? element.style.marginBottom = marginBottom + 'px' : null;
        }
        const repeatLoop = requestAnimationFrame(repeat);
        //如果对象是隐藏的（display:none）,会导致height值为NaN，如果是如此一次循环便结束了
        if (Math.ceil(step) >= height || !height) {
            cancelAnimationFrame(repeatLoop);
            callback && callback();
        }
    };
    repeat();

    return element;
};
/**
 * 定义axSlideUp动画
 * 举例：axSlideUp(elem)
 */
const axSlideUp = (elem, display, callback, duration, curve = 'easeOut') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem),
        dftTime = height => height / 3 + 250,
        rawStyle = element.style.cssText;
    display = display || 'block';
    if (window.getComputedStyle(element).display !== "none") {
        //计算高度
        const { height, paddingTop, paddingBottom, marginTop, marginBottom } = axArrHeight(element);
        //设置time
        let time;
        if (duration !== undefined && typeof duration === 'number') {
            time = duration;
        } else {
            time = dftTime(height);
        }
        //初始化
        element.style.cssText = `${rawStyle}display: ${display}; padding-top: 0; padding-bottom: 0;margin-top: 0; margin-bottom: 0;`;
        //动画
        const initTime = new Date().getTime();
        const repeat = () => {
            const newTime = new Date().getTime() - initTime;
            const timestep = newTime / time;
            const timefactor = axCurves[curve](timestep);
            const step = height + -height * timefactor;
            const stepPaddingT = paddingTop + (-paddingTop * timefactor);
            const stepPaddingB = paddingBottom + (-paddingBottom * timefactor);
            const stepMarginT = marginTop + (-marginTop * timefactor);
            const stepMarginB = marginBottom + (-marginBottom * timefactor);
            if (newTime <= time && Math.floor(step) > 0) {
                element.style.cssText = `${rawStyle}overflow:hidden;display: ${display}; padding-top: ${stepPaddingT}px; padding-bottom: ${stepPaddingB}px;margin-top: ${stepMarginT}px; margin-bottom: ${stepMarginB}px; height: ${step}px`;
            } else {
                element.style.cssText = `${rawStyle}display: none`;
            }
            const repeatLoop = requestAnimationFrame(repeat);
            //如果对象是隐藏的（display:none）,会导致height值为NaN，如果是如此一次循环便结束了
            if (Math.floor(step) <= 0 || !height) {
                cancelAnimationFrame(repeatLoop);
                callback && callback();
            }
        };
        repeat();
    }
    return element;
};
/**
 * 不判断元素是否隐藏的上拉隐藏动画
 */
const axHeightUp = (elem, arrHeight, callback, duration, curve = 'easeOut') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem);
    let dftTime = height => height / 3 + 250;
    //计算高度
    const height = (arrHeight && arrHeight.height) ? arrHeight.height : axArrHeight(element).height,
        paddingTop = (arrHeight && arrHeight.paddingTop) ? arrHeight.paddingTop : axArrHeight(element).paddingTop,
        paddingBottom = (arrHeight && arrHeight.paddingBottom) ? arrHeight.paddingBottom : axArrHeight(element).paddingBottom,
        marginTop = (arrHeight && arrHeight.marginTop) ? arrHeight.marginTop : axArrHeight(element).marginTop,
        marginBottom = (arrHeight && arrHeight.marginBottom) ? arrHeight.marginBottom : axArrHeight(element).marginBottom;
    //设置time
    let time;
    if (duration !== undefined && typeof duration === 'number') {
        time = duration;
    } else {
        time = dftTime(height);
    }
    element.style.overflow = 'hidden';
    //动画
    const initTime = new Date().getTime();
    const repeat = () => {
        const newTime = new Date().getTime() - initTime;
        const timestep = newTime / time;
        const timefactor = axCurves[curve](timestep);
        const step = height + -height * timefactor;
        const stepPaddingT = paddingTop + (-paddingTop * timefactor);
        const stepPaddingB = paddingBottom + (-paddingBottom * timefactor);
        const stepMarginT = marginTop + (-marginTop * timefactor);
        const stepMarginB = marginBottom + (-marginBottom * timefactor);
        if (newTime <= time && Math.floor(step) > 0) {
            element.style.height = step + 'px';
            paddingTop ? element.style.paddingTop = stepPaddingT + 'px' : null;
            paddingBottom ? element.style.paddingBottom = stepPaddingB + 'px' : null;
            marginTop ? element.style.marginTop = stepMarginT + 'px' : null;
            marginBottom ? element.style.marginBottom = stepMarginB + 'px' : null;
        } else {
            element.style.height = 0;
            paddingTop ? element.style.paddingTop = 0 : null;
            paddingBottom ? element.style.paddingBottom = 0 : null;
            marginTop ? element.style.marginTop = 0 : null;
            marginBottom ? element.style.marginBottom = 0 : null;
        }
        const repeatLoop = requestAnimationFrame(repeat);
        //如果对象是隐藏的（display:none）,会导致height值为NaN，如果是如此一次循环便结束了
        if (Math.floor(step) <= 0 || !height) {
            cancelAnimationFrame(repeatLoop);
            callback && callback();
        }
    };
    repeat();
    return element;
};
/**
 * 定义axSlideToggle动画
 * 举例：axSlideToggle(elem)
 */
const axSlideToggle = (elem, display, downCallback, upCallback, duration, curve = 'easeOut') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem);
    display = display || 'block';
    if (window.getComputedStyle(element).display === 'none') {
        axSlideDown(element, display, downCallback, duration, curve);
    } else {
        axSlideUp(element, display, upCallback, duration, curve);
    }
    return element;
}
/**
 * 定义axFadeIn动画
 * 举例：axFadeIn(elem)
 */
const axFadeIn = (elem, show, callback, duration = 200, curve = 'linear') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem);
    show = show || 'block';
    if (element) {
        let display = window.getComputedStyle(element).display,
            opacity = window.getComputedStyle(element).opacity;
        if (display === 'none') {
            //隐藏状态，则全部显示
            element.style.display = show;
            element.style.opacity = 0;
            let initTime = new Date().getTime();
            let repeat = () => {
                let newTime = new Date().getTime() - initTime,
                    timestep = newTime / duration,
                    step = axCurves[curve](timestep);
                if (newTime <= duration && step <= 1) {
                    element.style.opacity = step;
                } else {
                    element.style.opacity = 1;
                }
                let repeatLoop = requestAnimationFrame(repeat);
                if (step >= 1) {
                    cancelAnimationFrame(repeatLoop);
                }
            };
            repeat();
        } else if (display !== 'none' && Number(opacity) !== 1) {
            //半透明状态，则全部显示
            let initTime = new Date().getTime();
            let repeat = () => {
                let newTime = new Date().getTime() - initTime,
                    timestep = newTime / duration,
                    step = axCurves[curve](timestep) + Number(opacity);

                if (newTime <= duration && step <= 1) {
                    element.style.opacity = step;
                } else {
                    element.style.opacity = 1;
                }
                let repeatLoop = requestAnimationFrame(repeat);
                if (step >= 1) {
                    cancelAnimationFrame(repeatLoop);
                }
            };
            repeat();
        }
    }
    //回调
    callback && callback();
    return element;
};
/**
 * 定义axFadeOut动画
 * 举例：axFadeOut(elem)
 */
const axFadeOut = (elem, callback, duration = 200, curve = 'linear') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem);
    if (element) {
        let display = window.getComputedStyle(element).display,
            opacity = window.getComputedStyle(element).opacity;
        if (display !== 'none' || Number(opacity) !== 0) {
            element.style.opacity = opacity;
            let initTime = new Date().getTime();
            let repeat = () => {
                let newTime = new Date().getTime() - initTime,
                    timestep = newTime / duration,
                    step = Number(opacity) - axCurves[curve](timestep);
                if (newTime <= duration && step > 0) {
                    element.style.opacity = step;
                } else {
                    element.style.opacity = 0;
                    element.style.display = 'none';
                }
                let repeatLoop = requestAnimationFrame(repeat);
                if (step <= 0) {
                    cancelAnimationFrame(repeatLoop);
                }
            };
            repeat();
        }
    }
    //回调
    callback && callback();
    return element;
};
/**
 * 定义axFadeToggle动画
 * 举例：axFadeToggle(elem)
 */
const axFadeToggle = (elem, show, inCallback, outCallback, duration = 200, curve = 'linear') => {
    if (!elem) { console.error(axError.notExist); return false; }
    let element = axIdToDom(elem);
    if (element) {
        let display = window.getComputedStyle(element).display;
        let opacity = window.getComputedStyle(element).opacity;
        if (display === 'none') {
            axFadeIn(element, show, inCallback, duration, curve);
        } else if (Number(opacity) <= 0.5) {
            axFadeIn(element, show, inCallback, duration, curve);
        } else if (Number(opacity) > 0.5) {
            axFadeOut(element, outCallback, duration, curve);
        } else {
            axFadeOut(element, outCallback, duration, curve);
        }
    }
    return element;
}
/**
 * 定义axShow动画（元素显示），因为有些元素显示方式可以是inline-block、block、visible或box，所以需要使用show来设定，如果不填则优先使用block
 * 举例：axShow(elem,show,callback)
 */
const axShow = (element, show, callback) => {
    if (!element) { console.error(axError.notExist); return false; }
    let elem = axIdToDom(element);
    //根据所使用显示样式来显示Dom
    if (show == 'visible') {
        elem.style.visibility = 'visible';
    } else if (show) {
        //inline-block，grid，flex等
        elem.style.display = show;
    } else {
        elem.style.display = 'block';
    }
    callback && callback();
}
/**
 * 定义axHide动画(元素隐藏)，根据实际情况选择隐藏方式来设置hide参数，可选none和hidden，如果不填则优先使用none
 * 举例：axHide(elem,hide,callback)
 */
const axHide = (element, hide, callback) => {
    if (!element) { console.error(axError.notExist); return false; }
    let elem = axIdToDom(element);
    //根据所使用隐藏样式来隐藏Dom
    if (hide == 'hidden') {
        elem.style.visibility = 'hidden';
    } else {
        elem.style.display = 'none';
    }
    callback && callback();
}
/**
 * 元素显示隐藏切换，show和hide参数如上面两个函数所述
 * 举例: axToggle(elem,show,hide)
 */
const axToggle = (element, show, hide, showCallback, hideCallback) => {
    if (!element) { console.error(axError.notExist); return false; }
    let elem = axIdToDom(element);
    axIsVisible(elem) ? axHide(elem, hide, hideCallback) : axShow(elem, show, showCallback);
}
/**
 * 页面锚记滚动
 * 举例：axScrollTo(elem,target,2000,'easeOut')
 */
const axScrollTo = (element, options = {}) => {
    //已经滚动的距离
    let parentDom = options.parent ? axIdToDom(options.parent) : null,
        scrollParent = parentDom || window,
        scrollOffset = ~~options.offset || 0,
        curve = options.curve || 'easeOut',
        anchorDom = typeof element === 'number' ? '' : axIdToDom(element),
        bodyScrollTop = document.documentElement.scrollTop || window.pageYOffset || document.body.scrollTop,
        offsetScrolled = parentDom ? parentDom.scrollTop : bodyScrollTop,
        //初始时间戳
        initTime = new Date().getTime(),
        offsetClient;
    //父节点子元素计算offsetTop需要父节点有position值
    if (parentDom && axStyle(parentDom).position === 'static') {
        parentDom.style.position = 'relative';
    }
    //元素顶部距离视窗顶部的距离，下方为正值，上方为负值；如果没有element参数则等于已滚动的负值
    if (anchorDom) {
        offsetClient = parentDom ? anchorDom.offsetTop - parentDom.scrollTop : anchorDom.getBoundingClientRect().top;
    } else {
        //element可以是数字
        offsetClient = typeof element === 'number' ? element - offsetScrolled : -offsetScrolled;
    }
    //动画
    let duration = !options.duration || typeof options.duration !== 'number' ? Math.abs(offsetClient) / 4 + 1000 : options.duration,
        repeat = () => {
            let newTime = new Date().getTime() - initTime,
                timestep = newTime / duration,
                //如果没有element参数，offsetScrolled = offsetClient，也就是直接返回页面顶部
                step = offsetScrolled + (offsetClient - scrollOffset) * axCurves[curve](timestep);
            scrollParent.scrollTo(0, step);
            //执行repeat函数
            repeat.timer = requestAnimationFrame(repeat);

            if (newTime >= duration) {
                scrollParent.scrollTo(0, step);
                //取消repeat函数
                cancelAnimationFrame(repeat.timer);
            }
        };
    repeat();
    return element;
};
/**
 * 定义常用图标
 */
const axIcon = {
    font: {
        success: '<i class="ax-iconfont ax-icon-check-o"></i>',
        error: '<i class="ax-iconfont ax-icon-close-o"></i>',
        warning: '<i class="ax-iconfont ax-icon-warning-o"></i>',
        info: '<i class="ax-iconfont ax-icon-info-o"></i>',
        question: '<i class="ax-iconfont ax-icon-question-o"></i>',
        successFill: '<i class="ax-iconfont ax-icon-check-o-f"></i>',
        errorFill: '<i class="ax-iconfont ax-icon-close-o-f"></i>',
        warningFill: '<i class="ax-iconfont ax-icon-warning-o-f"></i>',
        infoFill: '<i class="ax-iconfont ax-icon-info-o-f"></i>',
        questionFill: '<i class="ax-iconfont ax-icon-question-o-f"></i>',
    },
    svg: {
        success: '<svg class="svg-success" xmlns="http://www.w3.org/2000/svg" width="86.6986mm" height="86.6986mm" viewBox="0 0 86.6986 86.6986"><path class="svg-line svg-bg" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-in-1" d="M26.316,42.859L37.9984,54.5414L60.3826,32.1572"></path></svg>',
        error: '<svg class="svg-error" xmlns="http://www.w3.org/2000/svg" width="86.6986mm" height="86.6986mm" viewBox="0 0 86.6986 86.6986"><path class="svg-line svg-bg" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-in-1" d="M28.774,57.9246L57.9247,28.7739"></path><path class="svg-line svg-in-2" d="M57.9246,57.9246L28.7739,28.7739"></path></svg>',
        warning: '<svg class="svg-warning" xmlns="http://www.w3.org/2000/svg" width="86.6986mm" height="86.6986mm" viewBox="0 0 86.6986 86.6986"><path class="svg-line svg-bg" d="M43.4611 7.24c2.8081,0.0924 4.39,1.7 5.3045,3.1159l17.4543 29.9414 17.3445 29.7538c0.5448,1.0193 1.596,4.0544 0.1109,6.4168 -1.4849,2.3626 -3.6815,2.9155 -5.3768,2.992l-34.9082 0.0002 -34.6892 -0.0002c-1.1636,-0.0421 -4.3433,-0.6583 -5.6666,-3.1131 -1.3232,-2.4549 -0.7085,-4.6157 0.0723,-6.1078l17.454 -29.9417 17.3449 -29.7537c0.6185,-0.977 2.7471,-3.396 5.5554,-3.3036z"></path><path class="svg-line svg-out" d="M43.4611 7.24c2.8081,0.0924 4.39,1.7 5.3045,3.1159l17.4543 29.9414 17.3445 29.7538c0.5448,1.0193 1.596,4.0544 0.1109,6.4168 -1.4849,2.3626 -3.6815,2.9155 -5.3768,2.992l-34.9082 0.0002 -34.6892 -0.0002c-1.1636,-0.0421 -4.3433,-0.6583 -5.6666,-3.1131 -1.3232,-2.4549 -0.7085,-4.6157 0.0723,-6.1078l17.454 -29.9417 17.3449 -29.7537c0.6185,-0.977 2.7471,-3.396 5.5554,-3.3036z"></path><path class="svg-line svg-in-1" d="M43.3493,27.8713L43.3493,57.2858"></path><circle class="svg-circle svg-in-2" cx="43.3492" cy="64.3337" r="2.1166"></circle></svg>',
        info: '<svg class="svg-info" xmlns="http://www.w3.org/2000/svg" width="86.6986mm" height="86.6986mm" viewBox="0 0 86.6986 86.6986"><path class="svg-line svg-bg" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-in-1" d="M43.3493,65.0602L43.3493,30.9723"></path><circle class="svg-circle svg-in-2" cx="43.3492" cy="23.5856" r="2.1166"></circle></svg>',
        question: '<svg class="svg-question" xmlns="http://www.w3.org/2000/svg" width="86.6986mm" height="86.6986mm" viewBox="0 0 86.6986 86.6986"><path class="svg-line svg-bg" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path><path class="svg-line svg-in-1" d="M32.3757 35.7255c-0.2203,-11.823 12.5789,-14.1087 18.4056,-9.4189 5.4663,4.3995 4.7426,12.804 -3.1088,17.9938 -3.0015,1.9839 -3.0003,3.8403 -3.0003,10.1707"></path><circle class="svg-circle svg-in-2" cx="44.6612" cy="60.5502" r="2.1166"></circle></svg>',
    },
};
/**
 * 定义常用图片
 */
const axImage = {
    load: getComputedStyle(document.documentElement).getPropertyValue('--load').trim().split('"')[1],
    loadW: getComputedStyle(document.documentElement).getPropertyValue('--load-w').trim().split('"')[1],
    blank: getComputedStyle(document.documentElement).getPropertyValue('--blank').trim().split('"')[1],
    avatar: getComputedStyle(document.documentElement).getPropertyValue('--avatar').trim().split('"')[1],
    empty: getComputedStyle(document.documentElement).getPropertyValue('--empty').trim().split('"')[1],
    none: getComputedStyle(document.documentElement).getPropertyValue('--none').trim().split('"')[1],
}
/**
 * 获得主题颜色
 */
const axGetColors = () => {
    let style = getComputedStyle(document.documentElement),
        getValue = (type) => {
            return [
                style.getPropertyValue(`--h-${type}`),
                style.getPropertyValue(`--s-${type}`),
                style.getPropertyValue(`--l-${type}`),
                style.getPropertyValue(`--a-${type}`)
            ].map(k => {
                let val = k.trim();
                return val.includes('%') ? val : Number(val);
            });
        };
    return {
        primary: getValue('primary'),
        text: getValue('text'),
        success: getValue('success'),
        error: getValue('error'),
        info: getValue('info'),
        warning: getValue('warning'),
        question: getValue('question'),
    }
}
/**
 * 定义结果函数
 */
const axResult = (elem, state) => {
    if (!elem) { console.error(axError.notExist); return false; }
    let dom = axIdToDom(elem),
        svg = '';
    switch (state) {
        case 'success':
            svg = `<path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path>
            <path class="svg-line svg-in-1" d="M26.316,42.859L37.9984,54.5414L60.3826,32.1572"></path>`;
            dom.querySelector('svg').classList.add('svg-success');
            break;
        case 'error':
            svg = `<path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path>
            <path class="svg-line svg-in-1" d="M28.774,57.9246L57.9247,28.7739"></path>
            <path class="svg-line svg-in-2" d="M57.9246,57.9246L28.7739,28.7739"></path>`;
            dom.querySelector('svg').classList.add('svg-error');
            break;
        case 'question':
            svg = `<path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path>
            <path class="svg-line svg-in-1" d="M32.3757 35.7255c-0.2203,-11.823 12.5789,-14.1087 18.4056,-9.4189 5.4663,4.3995 4.7426,12.804 -3.1088,17.9938 -3.0015,1.9839 -3.0003,3.8403 -3.0003,10.1707"></path>
            <circle class="svg-circle svg-in-2" cx="44.6612" cy="60.5502" r="2.1166"></circle>`;
            dom.querySelector('svg').classList.add('svg-question');
            break;
        case 'info':
            svg = `<path class="svg-line svg-out" d="M7.238500000000002,43.3493A36.1108,36.1108 0,1,1 79.4601,43.3493A36.1108,36.1108 0,1,1 7.238500000000002,43.3493"></path>
            <path class="svg-line svg-in-1" d="M43.3493,65.0602L43.3493,30.9723"></path>
            <circle class="svg-circle svg-in-2" cx="43.3492" cy="23.5856" r="2.1166"></circle>`;
            dom.querySelector('svg').classList.add('svg-info');
            break;
    }
    dom.classList.add('ax-result-show');
    if (dom.querySelector('svg').innerHTML.indexOf('svg-in-1') == -1) {
        dom.querySelector('svg').insertAdjacentHTML('beforeend', svg);
    }
}
/**
 * 获取主体内容，适用于dialog|drawer|popup
 */
const axCreateContent = (options, instance, callback) => {
    let content = options.content, promise = {}, result;

    if (options.url) {
        //有内容链接
        if (options.type === 'async') {
            //异步类型
            let ajaxUrl = '', ajaxSelector = '';
            if (axStrType(options.url) == 'id') {
                //如果不仅有异步链接而且也有ID，那么从异步页面获取该ID的DOM的HTML内容
                ajaxUrl = options.url.match(/(\S*)\#/)[1];
                ajaxSelector = '#' + options.url.match(/\#(\S*)/)[1];
            } else {
                //如果只有异步链接没有ID，那么载入整个异步页面
                ajaxUrl = options.url;
            }
            promise = axAjax({
                url: ajaxUrl,
                selector: ajaxSelector,
                type: options.ajaxType || 'post',
                opened: (data) => {
                    //赋予xhr属性，以便于销毁时进行abort操作
                    instance.contentXhr = data.xhr;
                },
                before: (data) => {
                    //loading文本
                    result = data.content;
                },
                success: (data) => {
                    //如果content是函数，那么将异步返回值作为参数传过去，如果不是则直接输出异步内容
                    if (typeof content === 'function') {
                        result = content(data.content);
                    } else {
                        result = data.content;
                    }
                    callback && callback.call(instance, result);
                }
            });
        } else {
            if (options.type === 'iframe') {
                //定义iframe内容
                result = `<iframe src="${options.url}" frameborder="0" width="100%" height="100%"></iframe>`;
                options.footerShow = false;
            } else if (options.type === 'picture') {
                //定义图片内容
                result = `
            <div class="ax-media-container"><img src="${options.url}" /></div>
            <div class="ax-media-title">${options.name}</div>
            ${content}
            `;
            } else if (options.type === 'video') {
                //定义视频内容
                result = `
                <div class="ax-media-container"><video src="${options.url}" controls="controls" ></video></div>
                <div class="ax-media-title">${options.name}</div>
                ${content}
                `;
            } else if (options.type === 'audio') {
                //定义音频内容
                result = `
                <div class="ax-media-container"><audio src="${options.url}" controls="controls" ></audio></div>
                <div class="ax-media-title">${options.name}</div>
                ${content}
                `;
            }
            callback && callback.call(instance, result);
        }
    } else {
        //无内容链接
        if (options.type == 'html') {
            //文本类型
            if (typeof content === 'string' && axStrType(content) == 'id') {
                //如果content以#开头则从dom取得内容
                result = document.querySelector(content).innerHTML;
            } else if (typeof content === 'function') {
                //如果content是不传值得函数，则直接运行获得return结果
                result = content.call(instance);
            } else {
                //是纯文本
                result = content;
            }
        } else {
            //type为其他
            result = content || '没有任何内容！';
        }
        callback && callback.call(instance, result);
    }
    //绑定content属性
    promise.content = result;
    callback && callback(result);
    return promise;
}
/**
 * 终端适配断点
 */
const axBreakpoints = (obj, points) => {
    if (axIsEmpty(obj)) {
        return false;
    }
    let valids = [],
        assign = {},
        width = document.body.clientWidth,
        validFun = (key, value) => {
            if (key.startsWith('ceil')) {
                //字符串，key=ceil-phone、ceil-pad等
                axCeil[key.split('-')[1]] > width ? valids.push(value) : null;
            } else if (key.startsWith('client')) {
                //字符串，key=client-phone、client-pad等
                key.split('-')[1] === axClient() ? valids.push(value) : null;
            } else if (key.startsWith('width')) {
                //字符串，key=width-900、client-500等
                key.split('-')[1] > width ? valids.push(value) : null;
            } else if (key.startsWith('mobile')) {
                //字符串，key=mobile-true和mobile-false
                key.split('-')[1] == axIsMobi() ? valids.push(value) : null;
            } else {
                //数字，key=900,500
                key > width ? valids.push(value) : null;
            }
        };
    for (let k in points) {
        if (points.hasOwnProperty(k)) {
            //删除breakpoint属性，避免循环合并
            Reflect.deleteProperty(points[k], 'breakpoint');
            validFun(k, points[k]);
        }
    }
    if (valids.length === 0) {
        return false;
    }
    assign = Object.assign(...valids);
    axExtend(obj, assign);
}
/**
 * 获取底部内容，适用于dialog|drawer|popup
 */
const axFooterBtns = (options) => {
    let lineBreak = options.footerBreak ? `<div class="ax-break-line"></div>` : '';
    let footer = '';
    options.note ? options.note = `<div class="ax-note">${options.note}</div>` : null
    if (!options.footerShow) {
        footer = '';
    } else if (options.footerType == 'right') {
        let btnCancel = '',
            btnConfirm = '',
            btnDiy = '';
        options.confirm.text ? btnConfirm = `<a href="###" class="ax-btn ax-primary ${options.confirm.classname}"  confirm>${options.confirm.icon}${options.confirm.text}</a>` : null;
        options.cancel.text ? btnCancel = `<a href="###" class="ax-btn ax-plain ax-ignore ${options.cancel.classname}"  cancel>${options.cancel.icon}${options.cancel.text}</a>` : null;
        if (options.diy.text && options.diy.callback) {
            btnDiy = `<a href="###" class="ax-btn ax-plain ${options.diy.classname}"  diy>${options.diy.icon}${options.diy.text}</a>`;
        } else if (options.diy.text && !options.diy.callback) {
            btnDiy = `<a href="${options.diy.url}" target="${options.diy.url}" class="ax-btn ax-plain ${options.diy.classname}"  diy>${options.diy.icon}${options.diy.text}</a>`;
        }
        footer = `
                                        ${lineBreak}
                                        <div class="ax-padding">
                                              <div class="ax-row">
                                                  <div class="ax-col">
                                                  ${options.note}
                                                  </div>
                                                  <div class="ax-btns">${btnDiy}${btnCancel}${btnConfirm}</div>
                                              </div>
                                          </div>
                          `;
    } else if (options.footerType == 'line') {
        let btnCancel = '',
            btnConfirm = '',
            btnDiy = '';
        options.confirm.text ? btnConfirm = `<a href="###" class="ax-col ax-col-12 ${options.confirm.classname}"  confirm>${options.confirm.icon}${options.confirm.text}</a>` : null;
        options.cancel.text ? btnCancel = `<a href="###" class="ax-col ax-col-12 ${options.cancel.classname}"  cancel>${options.cancel.icon}${options.cancel.text}</a><span class="ax-gutter-line"></span>` : null;
        if (options.diy.text && options.diy.callback) {
            btnDiy = `<a href="###" class="ax-col ax-col-12 ${options.diy.classname}"  diy>${options.diy.icon}${options.diy.text}</a><span class="ax-gutter-line"></span>`;
        } else if (options.diy.text && !options.diy.callback) {
            btnDiy = `<a href="${options.diy.url}" target="${options.diy.url}" class="ax-col ax-col-12 ${options.diy.classname}"  diy>${options.diy.icon}${options.diy.text}</a><span class="ax-gutter-line"></span>`;
        }
        footer = `
                          <div class="ax-break-line"></div>
                          <div class="ax-row">${btnDiy}${btnCancel}${btnConfirm}</div>${options.note}
                          `
    } else {
        let btnCancel = '',
            btnConfirm = '',
            btnDiy = '';
        options.confirm.text ? btnConfirm = `<a href="###" class="ax-btn ax-primary ${options.confirm.classname}"  confirm>${options.confirm.icon}${options.confirm.text}</a>` : null;
        options.cancel.text ? btnCancel = `<a href="###" class="ax-btn ax-ignore ${options.cancel.classname}" cancel>${options.cancel.icon}${options.cancel.text}</a>` : null;
        if (options.diy.text && options.diy.callback) {
            btnDiy = `<a href="###" class="ax-btn ${options.diy.classname}" diy>${options.diy.icon}${options.diy.text}</a>`;
        } else if (options.diy.text && !options.diy.callback) {
            btnDiy = `<a href="${options.diy.url}" target="${options.diy.url}" class="ax-btn ${options.diy.classname}" diy>${options.diy.icon}${options.diy.text}</a>`;
        }
        footer = `
                                        ${lineBreak}
                                         <div class="ax-padding ax-align-center">${btnDiy}${btnCancel}${btnConfirm}</div>
                                         <div class="ax-break"></div>${options.note}
                         `
    }
    return footer;

}
/**
 * 日期时间均转成本地时间戳
 * 类似2021-1-12和2021-01-12字符“-”均转成“/”分隔
 * axLocalTime('2070-1-20') - new Date(axLocalTime('2070-1-20')).getTimezoneOffset() * 60000) % 86400000 = 0 （能整除）
 * new Date('2070-01-20 12:11:13').getTime() == new Date('2070-1-20 12:11:13').getTime() == axLocalTime('2070-1-20 12:11:13')
 * new Date('2070/1/20').getTime() == new Date('2070-01-20').getTime() == axLocalTime('2070-1-20',true)
 * new Date('2070-1-20').getTime() == axLocalTime('2070-1-20')
 * axLocalTime('2070-1-20')和new Date(axLocalTime('2070-1-20'))可以互转
*/
const axLocalTime = (date, utc) => {
    let para, localDate, flag, loacalTime, localOffset;
    if (date) {
        date = date.trim();
        localDate = new Date(date);
        flag = localDate instanceof Date && !isNaN(localDate.getTime());
        if (!flag) {
            return console.error('Date format must be "YYYY-MM-DD" or "YYYY/MM/DD"');
        }
        //将日期字符串“-”转成“/”
        para = date.replace(/-/g, "/");
        //获得本地标准时间时间戳，如果传入日期参数，则对此创建日期对象
        loacalTime = new Date(para).getTime();
        if (para.indexOf(':') !== -1) {
            //如果有完整的日期（有时分秒），则不需要计算偏移值
            localOffset = 0
        } else {
            //如果没有时分秒，获得本地标准时间的偏移，返回是分钟，转成毫秒
            localOffset = new Date(para).getTimezoneOffset() * 60000;
        }
    } else {
        //如果没有传入日期字符串，则将当前日期创建为日期对象
        loacalTime = new Date().getTime();
        localOffset = 0;
    }
    if (utc) {
        //如果是utc时间axLocalTime('2070-1-20',true)返回Mon Jan 20 2070 08:00:00 GMT+0800
        return loacalTime - localOffset;
    } else {
        //如果获得本地时间axLocalTime('2070-1-20')返回Mon Jan 20 2070 00:00:00 GMT+0800
        return loacalTime;
    }
}


/**
 * 将元素内容转成数组[{x:'',y:'',z:''}],主要用于axLightbox
 */
const axCreateData = (
    src,
    roles = ['src', 'cover', 'caption', 'type'],
    keys = ['media', 'cover', 'caption', 'type'],
    node = 'figure',
    dft = ['', '', '', 'image'],
    callback
) => {
    let data = [],
        string = '';

    /* 
        //从src转成data数据
        if (axTypeIs('Array', src)) {
            //如果本身就是数组，便直接输出
            //data = src
            string = JSON.stringify(src);
        } else if (axTypeIs('Object', src)) {
            //如果是一个{}对象，那么添加入data
            //data.push(src);
            string = JSON.stringify(src)
        } else if (axTypeIs('String', src)) {
            if (src.substr(0, 1) == '#') {
                //如果是#id，则从dom获取数据,默认格式如下：
                //<div style="display:none" id=""><figure type=""><img src="" type=""/><figcaption caption></figcaption></figure></div>
                data = axTplToArr(src, roles, keys, node, dft)
            } else {
                //'[]'，'{},{}'，'{}'，'a,b,c'转标准数组，将data统一格式为[{media:'',caption:''},{media:'',caption:''}...]
                data = axStrToArr(src, keys, dft);
            }
        } */

    if (axType(src) === 'String' && axStrType(src) == 'id') {
        //如果是#id，则从dom获取数据,默认格式如下：
        //<div style="display:none" id=""><figure type=""><img src="" type=""/><figcaption caption></figcaption></figure></div>
        data = axTplToArr(src, roles, keys, node, dft);
    } else {
        if (axType(src) === 'Array' || axType(src) === 'Object') {
            //如果本身就是数组或对象，则直接转成字符串
            string = JSON.stringify(src);
        } else if (axType(src) === 'String') {
            string = src;
        }
        //都转成字符串之后，进行统一格式处理，最终输出为：[{media:'',caption:''},{media:'',caption:''}...]
        data = axStrToArr(string, keys, dft);
    }
    //抛弃media为空的项
    data = data.filter(i => i[keys[0]]);
    //执行回调，支持data参数
    callback && callback(data);
    return data;
}
/**
 * 数组多重排序，已用在axList和axSheet插件中
*/
const axSort = (data, sorts, type) => {
    //data是数组数据源
    //sorts支持四种数据，支持col和name属性，其中sort是比较属性，一般是value或label
    //[{ col: 'age', attr:'label',order: 'desc' }...]或[{ name: 'age', attr:'label',order: 'desc' }...]
    //{ col: 'age', attr:'label',order: 'desc' }或{ name: 'age', attr:'label',order: 'desc' }
    //统一排序格式
    let sortsArr = [];
    if (!Array.isArray(sorts)) {
        //如果是对象：{ col: 'age',sort:'label', order: 'desc' }
        sorts.order ? sortsArr.push(sorts) : null;
    } else {
        //如果是数组，只保留有order值得项
        sortsArr = sorts.filter(k => k.order);
    }
    let orders = (a, b, fields) => {
        if (fields.length == 0) {
            //如果为传入排序条件，则返回0，即不排序
            return 0;
        }
        let field = fields.shift(),
            col = field.col,
            order = field.order,
            //如果没有填比较属性，那么默认为label，即字面文字，框架的table数组字面属性是label
            attr = field.attr || 'label',
            flag,
            left,
            right;
        if (type == 'table') {
            //table排序的data=[{row:'',dom:'',children:[{index:'',name:'',col:'',dom:''}...]}...]
            //td的index和name是必须的
            if (typeof col == 'number') {
                //如果传入的是列数，比如1
                left = a.children.filter(i => i.index === col)[0][attr];
                right = b.children.filter(i => i.index === col)[0][attr];
            } else {
                //如果传入的是name名，比如col2或title
                left = a.children.find(i => i.name === col)[attr];
                right = b.children.find(i => i.name === col)[attr];
            }
        } else {
            //list排序的data=[{id:'',dom:'',title:''}...]
            left = a[col];
            right = b[col]
        }
        if (isNaN(left - right)) {
            //如果是字母、中文、特殊字符，a[name] - b[name] == NaN
            flag = (order === 'desc') ? right.localeCompare(left, 'zh-CN-u-kf-lower') : left.localeCompare(right, 'zh-CN-u-kf-lower');
        } else {
            //如果是数字字符串
            flag = (order === 'desc') ? right - left : left - right;
        }
        return left === right ? orders(a, b, fields) : flag;
    }
    //shift会改变原数组，所以复制sorts参数
    return data.sort((a, b) => {
        return orders(a, b, [...sortsArr]);
    });
}
/*!
 * 插件：表单序列化；使用方法：new axSerialize('#id')
 * 支持将表单值序列化为以“&”连接的字符串，和对象数组[{name:'',value:''},...]
 */
const axSerialize = (element, type = "string", separator = ',') => {
    let inputs,
        items = [],
        output;
    if (Array.isArray(element)) {
        //如果是节点数组
        inputs = element;
    } else {
        //只取这三类表单域、有name值、name值不能有_ax_alt字符（如果以隐藏域创建radio或checkbox会自动将他们的原input的name加上这个后缀）
        inputs = [...axIdToDom(element).querySelectorAll('[name]')].filter(i => ['INPUT', 'SELECT', 'TEXTAREA'].includes(i.nodeName) && i.name && !i.name.includes('_ax_alt'));
    }
    inputs.forEach(i => {
        //先从items中找出来name相同的项
        let have = items.find(k => k.name == i.name);
        if (have) {
            //如果有相同且是checkbox或radio类型，才追加进dom
            if (i.type == 'checkbox' || i.type == 'radio') {
                have.dom.push(i);
            }
        } else {
            //如果没有则追加新的对象
            items.push({ name: i.name, type: i.type, dom: [i] })
        }
    });
    //给items项追加value属性
    items.forEach(i => {
        if (i.type == 'file') {
            i.value = axMultiValues(i.dom[0], 'file', 'array');
        } else if (i.type == 'checkbox') {
            i.value = axMultiValues(i.dom, 'checkbox', 'string', '', separator);
        } else if (i.type == 'radio') {
            i.value = axMultiValues(i.dom, 'radio', 'string');
        } else if (i.type.includes('select')) {
            i.value = axMultiValues(i.dom[0], 'select', 'string', '', separator);
        } else {
            //其他input
            i.value = i.dom[0].value;
        }
    });
    //创建输出
    if (type == 'string') {
        //输出为字符串：name=Lily&age=18&parent=Piter
        let str = '';
        items.forEach(k => {
            str += '&' + k.name + '=' + k.value;
        });
        output = str;
    } else if (type == 'json') {
        //输出为json：{name:'',age:'',parent:''}
        let obj = {}
        items.forEach(k => {
            obj[k.name] = k.value;
        });
        output = obj;
    } else if (type == 'array') {
        //输出为数组：[{name:'',value:''},...]
        items.forEach(k => {
            //丢掉type和dom
            delete k.type;
            delete k.dom;
        })
        output = items;
    }
    return output;
}
/*!
 * 通过多个关键字从数组中检索数据，获得每一项的权重
 * 支持多关键字检索
 * 支持多属性检索
 * 支持根据权重进行重排
 * 支持输出指定length数组，0表示不截取
 */
const axArrSearch = (keys, props, source, options) => {
    if (axIsEmpty(keys) || axIsEmpty(props) || !source || source.length === 0) {
        return [];
    }
    //ignore是指是否忽略英文大小写
    let searchOpt = Object.assign({ fuzzy: true, separator: ' ', limit: 0, reorder: false, ignore: true }, options),
        keyArr = axSplitToArr(keys, searchOpt.separator),
        propsArr = axSplitToArr(props, ','),
        crudeArr = [],
        //根据keyArr和accurate获得匹配表达式
        regStr = '(' + keyArr.join('|') + ')',
        regExp = new RegExp(regStr, `${searchOpt.ignore ? 'i' : ''}g`);
    //获得数组
    for (let v of source) {
        //使用■不常见字符作为分隔符避免被用户查找
        let valueStr = propsArr.map(k => v[k]).join('■'),
            valueMatch = valueStr.match(regExp),//全局匹配情况下，没有匹配到match会返回null
            valueObj = {};
        //没有匹配到则不输出
        if (valueMatch) {
            valueObj = { source: v, weight: valueMatch.length, keys: [...new Set(valueMatch)] };
            if (!searchOpt.fuzzy) {
                //如果每个关键字都匹配到了(精确)，那么匹配到的关键字数组应该跟原关键字数组length相同
                if ([...new Set(valueMatch)].length === keyArr.length) {
                    crudeArr.push(valueObj);
                }
            } else {
                crudeArr.push(valueObj);
            }
        }
    }
    //数据处理
    if (searchOpt.reorder) {
        //由大到小
        crudeArr.sort((a, b) => { return b.weight - a.weight });
    }
    if (searchOpt.limit && crudeArr.length > searchOpt.limit) {
        //超过裁掉
        crudeArr.splice(searchOpt.limit)
    }
    return crudeArr;
}
/*!
 * 插件：代理监听；使用方法：new axObserve (obj, options)
 * 基于Proxy监听数组和对象的变化
 */
class axObserve {
    constructor(obj, options) {
        this.dataType = axType(obj);
        if (this.dataType !== 'Object' && this.dataType !== 'Array') {
            return { instance: this, proxy: obj };
        }
        //参数
        this.options = axExtend({
            deep: false,
            accept: true,
            once: false,
            methods: ['set', 'deleteProperty'],
            callback: '',
        }, options);
        this.handlers = {};
        this.completeIndex = 0;
        this.reacts = 0;
        this.proxys = [];
        this.data = obj;

        this.fullMethods = {
            get: (target, key, proxy) => {
                let value = target[key];
                this.options.callback && this.options.callback({ target, key, proxy, value, type: 'get' });
                //根据计划执行(发布)
                this.handlers.hasOwnProperty(key) ? this.emit(key, { target, key, proxy, value, type: 'get' }) : null;
                return Reflect.get(target, key, proxy);
            },
            set: (target, key, value, proxy) => {
                //新值与旧值一样则不再监听
                if (JSON.stringify(target[key]) == JSON.stringify(value)) {
                    return true
                }
                //新增数组或对象则监听
                let valueType = axType(value),
                    raw = target[key],
                    handleType = target[key] === undefined ? 'add' : 'edit';
                if ((valueType === 'Object' || valueType === 'Array') && this.options.deep) {
                    value = this.proxy(value);
                }

                if (this.options.accept) {
                    //如果接受则不拦截并直接赋值
                    Reflect.set(target, key, value, proxy);
                }

                //如果不接受表示拦截赋值，此时便需要在callback里手动赋值
                this.options.callback && this.options.callback({ target, key, proxy, value, raw, type: handleType });
                //根据计划执行(发布)
                this.handlers.hasOwnProperty(key) ? this.emit(key, { target, key, proxy, value, raw, type: handleType }) : null;

                //监听执行完毕最终执行complete监听以进行确认操作
                this.reacts++;
                this.complete(this.reacts);
                return true;
            },
            deleteProperty: (target, key) => {
                let value = target[key];
                if (this.options.accept) {
                    //如果接受则不拦截
                    Reflect.deleteProperty(target, key);
                }
                this.options.callback && this.options.callback({ target, key, value, type: 'delete' });
                //根据计划执行(发布)
                this.handlers.hasOwnProperty(key) ? this.emit(key, { target, key, value, type: 'delete' }) : null;
                //监听执行完毕最终执行complete监听以进行确认操作
                this.reacts++;
                this.complete(this.reacts);
                return true;
            }
        }
        this.methods = {};
        //对this.options.methods和this.fullMethods进行映射，获得最终需要监听的方法
        this.options.methods.forEach(k => {
            this.methods[k] = this.fullMethods[k]
        });

        //深度监听
        if (this.options.deep) {
            for (let k in obj) {
                //数组或对象则监听，使用hasOwnProperty方法不遍历原型方法
                if (obj.hasOwnProperty(k)) {
                    let keyType = axType(obj[k]);
                    if (keyType === 'Object' || keyType === 'Array') {
                        obj[k] = this.proxy(obj[k]);
                    }
                }
            }
        }
        //执行
        return { instance: this, proxy: this.proxy(obj) };
    }
    complete(nowReact) {
        setTimeout(() => {
            //所有操作同步结束后触发complete
            if (nowReact === this.reacts) {
                //该代理实例，在同步或异步运行完毕，从始至终只执行一次complete监听
                if (this.options.once && this.completeIndex !== 0) {
                    return false;
                }
                this.options.callback && this.options.callback({ target: this.data, proxy: this.proxys[0], type: 'complete' });
                this.handlers.hasOwnProperty('complete') ? this.emit('complete', { target: this.data, proxy: this.proxys[0], type: 'complete' }) : null;
                this.completeIndex++;
            }
        }, 0);
    }
    proxy(obj) {
        let proxy = new Proxy(obj, this.methods);
        //保存主代理
        if (this.proxys.length === 0) {
            this.proxys.push(proxy);
        }
        return proxy;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
//////////////////////////////////////////// 版块分割线 ////////////////////////////////////////////

/*!
 * 插件：全局实例管理；使用方法：new axInstance('#id',{参数})
 * 支持axInstance.push、axInstance.destroy、axInstance.destroyAll、axInstance.clear等个方法
 */
class AXINSTANCE {
    constructor() {
        this.data = [];
        this.destroyFun = (item) => {
            if (!item) {
                return false;
            }
            //instance.__proto__.destroy是找到实例的this.destroy()方法
            if ((!item.instance.hasOwnProperty('destroyed') || !item.instance.destroyed) && item.instance.__proto__.destroy) {
                item.instance.destroy();
                item.log = `destroy(${new Date().toLocaleString()})`;
            }
        }
        this.initFun = (item) => {
            if (!item) {
                return false;
            }
             //instance.__proto__.init是找到实例的this.init()方法
            if ((!item.instance.hasOwnProperty('destroyed') || item.instance.destroyed) && item.instance.__proto__.init) {
                item.instance.init();
                item.log = `init(${new Date().toLocaleString()})`;
            }
        }
    }
    push(instance, name, type) {
        //instance是必填项
        if (!instance) {
            return false;
        }
        let obj = { name: name, instance: instance, type: type || '', log: `push(${new Date().toLocaleString()})` }
        //实例合集中没有该instance才加入，允许type和name相同
        if (!this.data.some(i => i.instance === instance)) {
            this.data.push(obj);
        }
    }
    find(name, type, destroyed = false) {
        if (!name) {
            //name必填，没有则不再执行
            return false;
        }
        let item;
        if (type) {
            item = this.data.find(i => {
                if (i.instance.hasOwnProperty('destroyed')) {
                    return i.name === name && i.type === type && i.instance.destroyed === destroyed;
                } else {
                    return i.name === name && i.type === type;
                }
            });
        } else {
            item = this.data.find(i => {
                if (i.instance.hasOwnProperty('destroyed')) {
                    return i.name === name && i.instance.destroyed === destroyed;
                } else {
                    return i.name === name;
                }
            });
        }
        return item ? item.instance : item;
    }
    findAll(type, destroyed = false) {
        let items = [];
        if (!type) {
            //type为空则返回所有合集
            if (i.instance.hasOwnProperty('destroyed')) {
                items = this.data.filter(i => i.instance.destroyed === destroyed);
            } else {
                items = this.data;
            }
        } else {
            items = this.data.filter(i => {
                if (i.instance.hasOwnProperty('destroyed')) {
                    return i.type === type && i.instance.destroyed === destroyed;
                } else {
                    return i.type === type;
                }
            });
        }
        return items.length > 0 ? items.map(i => i.instance) : items;
    }
    //销毁某一个实例
    destroy(name, type, callback) {
        if (!name) {
            //name必填，没有则不再执行
            return false;
        }
        let item;
        if (type) {
            item = this.data.find(i => i.name === name && i.type === type);
        } else {
            item = this.data.find(i => i.name === name);
        }
        //执行销毁
        this.destroyFun(item);
        callback && callback(item);
    }
    //销毁一类实例或全部实例
    destroyAll(type, callback) {
        if (!type) {
            //如果type为空或不填则等于销毁所有实例
            this.data.forEach(i => {
                this.destroyFun(i);
            });
            callback && callback(this.data);
            return false;
        } else {
            let items = this.data.filter(i => i.type === type);
            items.forEach(i => {
                this.destroyFun(i);
            });
            callback && callback(items);
        }
    }
    //销毁所有实例并清空合集
    clear(callback) {
        this.data.forEach(i => {
            this.destroyFun(i);
        });
        //清空数组
        this.data.length = 0
        callback && callback();
    }
    //重新初始化
    init(name, type, callback) {
        if (!name) {
            //name必填，没有则不再执行
            return false;
        }
        let item;
        if (type) {
            item = this.data.find(i => i.name === name && i.type === type);
        } else {
            item = this.data.find(i => i.name === name);
        }
        //执行初始化
        this.initFun(item);
        callback && callback(item);
    }
    //重新初始化
    initAll(type, callback) {
        if (!type) {
            //如果type为空或不填则等于初始化所有实例
            this.data.forEach(i => {
                this.initFun(i);
            });
            callback && callback(this.data);
            return false;
        } else {
            let items = this.data.filter(i => i.type === type);
            items.forEach(i => {
                this.initFun(i);
            });
            callback && callback(items);
        }
    }
}
const axInstance = new AXINSTANCE();

/*!
* 插件：元素笼子；使用方法：new axCage({参数})
*/
class axCage {
    constructor(elem, options) {
        if (!elem) { return false; }
        // 始终将elem转成DOM对象
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom || this.targetDom.children.length === 0) { return false; }

        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            init: true,//加载后是否直接初始化，默认true，可选择false手动初始化
            unbound: {
                attr: 'unbound',//未绑定事件的元素被设置的属性，默认unbound
                selector: '',//未绑定事件元素，通常为.className或nodeName等节点选择器，也可以是一个节点或节点数组，未绑定事件元素不可拖拽滚动，可复制文本
            },
            dampingRatio: 0.5,//拖拽出空白区域时的阻尼系数，直观表现为拖拽比较迟滞
            duration: 300,//执行动画基础时间，单位ms
            drift: {
                enable: false,//是否启用漂移功能，默认false不启用，可使用true启用该功能
                timeTHR: 200,//产生甩动漂移的时间阈值，小于该值会有漂移效果（与drift.distTHR配合），单位ms
                distTHR: 20,//产生甩动漂移的抖动阈值，小于该值则不漂移（与drift.timeTHR配合），单位px，避免垂直（相对）方向滑动也漂移
                coef: 2,//漂移系数，值越大滚动距离越远，默认为2，请填一个正整数
            },
            scale: {
                enable: false,//是否启用滚轮或双指放大
                min: 1,//最小倍数，默认1即保持原样，请填写小于max的正整数
                max: 5,//最大倍数，默认5，请填写大于min的正整数
            },
            completedTHR: 20,//拖出空白多长表示完成以便于执行回调，单位px
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，加载完毕执行，无参数
            onCompleted: '',//回调函数，拖出空白长度大于等于completedTHR时执行，支持一个参数direction，即拖拽方向，为left、right、top和bottom
            onStart: '',//回调函数，手指/鼠标按下执行，支持一个参数即当前子对象
            onMove: '',//回调函数，手指/鼠标拖动过程执行，支持一个参数即当前子对象
            onEnd: '',//回调函数，松开手指/鼠标后执行，支持一个参数即当前子对象
            onScaled: '',//回调函数，内容器发生scale缩放时执行，支持两个参数：缩放对象和原缩放值
            onScaledMin: '',//回调函数，缩放到最小时执行，支持两个参数：缩放对象和缩放值
            onScaledMax: '',//回调函数，缩放到最大时执行，支持两个参数：缩放对象和缩放值
            onTranslated: '',//回调函数，内容器发生translate偏移时执行，支持两个参数：移动对象和原移动值
            onRefactored: '',//回调函数，内容器发生节点变化（删除或添加子元素）时执行，无参数
            onResized: '',//回调函数，外容器尺寸变化后执行，支持两个参数即rect值和容器节点（outer或inner）
            onFitted: '',//回调函数，贴合边缘后执行，支持两一个参数即当前贴合对象（来自this.slides数组）
            onSlided: '',//回调函数，完成一次滚动后执行，支持两个参数即原始值（节点）和目标translate值
            onUpdated: '',//回调函数，参数更新后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        let _this = this;
        //回调监听
        this.handlers = {};
        //终端事件数组
        this.isMobi = axIsMobi();
        this.events = this.isMobi ? ['touchstart', 'touchmove', 'touchend'] : ['mousedown', 'mousemove', 'mouseup'];
        //触发状态
        this.started = false;
        //滚轮对象
        this.scrollObj = axGetScrollObj();
        //缩放倍数
        //start函数
        this.startFun = function (e) {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            //阻止默认事件
            if (_this.unbounds.length > 0 && (_this.unbounds.includes(e.target))) {
                //在未绑定事件项里滑动则不滚动，按默认事件执行
                //至少有一个unbound，并且当前鼠标所指节点再其中
                return false;
            }
            e.preventDefault();
            //手指触碰获取坐标值（相对浏览器左上角，超出为负值），_this.startValue存起来用来屏蔽拖拽中触发click事件
            _this.startValue = { ..._this.getCoordValue(e) };
            let item = _this.slides.find(k => k.dom === this),
                coordValue = { ..._this.startValue };
            _this.distStart = { x: coordValue.x - item.last.x, y: coordValue.y - item.last.y };
            //鼠标按下时的时间戳。pc端甩鼠标使用
            _this.startTime = e.timeStamp;
            //更新状态
            _this.started = true;
            //追加active属性
            this.setAttribute('active', '');
            item.active = true;
            this.style.transitionDuration = '0ms';
            //监听回调
            _this.handlers.hasOwnProperty('start') ? _this.emit('start', item) : null;
            _this.options.onStart && _this.options.onStart.call(_this, item);
        };
        //move函数
        this.moveFun = function (e) {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            //PC端并且不是在inner中点击则不再执行，pc端必须有_this.started
            if (!_this.isMobi && !_this.started) {
                return false;
            }
            //阻止默认事件
            if (_this.unbounds.length > 0 && (_this.unbounds.includes(e.target))) {
                //在未绑定事件项里滑动则不滚动，按默认事件执行
                return false;
            }
            e.preventDefault();

            //手指触碰获取坐标值（相对浏览器左上角，超出为负值）
            let coordValue = _this.getCoordValue(e);

            if (JSON.stringify(coordValue) != JSON.stringify(_this.startValue)) {
                //与开始坐标不一致，说明已经拖拽滚动，追加moveing属性
                _this.targetDom.setAttribute('moving', '');
            }

            //拖动方向，true表示下移，false表示上移
            let item = _this.slides.find(k => k.dom === this),
                startOffset = (axis) => {
                    return item.range[axis][0] + (item.now[axis] - item.range[axis][0]) * Math.pow(_this.options.dampingRatio, 2);
                },
                endOffset = (axis) => {
                    return item.range[axis][1] + (item.now[axis] - item.range[axis][1]) * Math.pow(_this.options.dampingRatio, 2);
                };
            item.now = { x: coordValue.x - _this.distStart.x, y: coordValue.y - _this.distStart.y };
            //将滑块限制的外容器内部，超出的回弹
            if (item.now.y < item.range.y[0]) {
                item.now.y = startOffset('y');
                if (Math.abs(item.now.y - item.range.y[0]) >= _this.options.completedTHR) {
                    //监听回调
                    _this.handlers.hasOwnProperty('completed') ? _this.emit('completed', 'top') : null;
                    _this.options.onCompleted && _this.options.onCompleted.call(_this, 'top');
                }
            } else if (item.now.y > item.range.y[1]) {
                item.now.y = endOffset('y');
                if (Math.abs(item.now.y - item.range.y[1]) >= _this.options.completedTHR) {
                    //监听回调
                    _this.handlers.hasOwnProperty('completed') ? _this.emit('completed', 'bottom') : null;
                    _this.options.onCompleted && _this.options.onCompleted.call(_this, 'bottom');
                }
            }
            if (item.now.x < item.range.x[0]) {
                item.now.x = startOffset('x');
                if (Math.abs(item.now.x - item.range.x[0]) >= _this.options.completedTHR) {
                    //监听回调
                    _this.handlers.hasOwnProperty('completed') ? _this.emit('completed', 'left') : null;
                    _this.options.onCompleted && _this.options.onCompleted.call(_this, 'left');
                }
            } else if (item.now.x > item.range.x[1]) {
                item.now.x = endOffset('x');
                if (Math.abs(item.now.x - item.range.x[1]) >= _this.options.completedTHR) {
                    //监听回调
                    _this.handlers.hasOwnProperty('completed') ? _this.emit('completed', 'right') : null;
                    _this.options.onCompleted && _this.options.onCompleted.call(_this, 'right');
                }
            }

            //偏移滚动
            _this.setValues({ item, value: item.now, apply: false });

            //监听回调
            _this.handlers.hasOwnProperty('move') ? _this.emit('move', item) : null;
            _this.options.onMove && _this.options.onMove.call(_this, item);
        }
        //end函数
        this.endFun = function (e) {
            //阻止事件向上（父层）传播
           e.stopPropagation();
            //取消moving属性
            _this.targetDom.removeAttribute('moving');
            //PC端并且不是在inner中点击则不再执行，pc端必须有_this.started
            if ((!_this.isMobi && !_this.started)) {
                return false;
            }
            e.preventDefault();
            //更新状态
            _this.started = false;
            //抖动值，计算三角形斜边长度
            let item = _this.slides.find(k => k.active),
                triangleValue = _this.computeHYP({ ...item.now }, { ...item.last }),
                distJitter = triangleValue.HYP,
                //左上拖动是负值，右下拖动是正值
                sinValue = triangleValue.SIN,
                cosValue = triangleValue.COS;

            //获得历史距离
            item.last = { ...item.now };
            //移除动画时间，在slideTo中会重新赋值该属性
            item.dom.style.cssText = item.dom.style.cssText.replace('transition-duration: 0ms', '');

            //计算漂移
            let timeOffset = e.timeStamp - _this.startTime, value;
            if (_this.options.drift.enable && timeOffset < _this.options.drift.timeTHR && distJitter > _this.options.drift.distTHR) {
                //自动计算，快甩产生漂移，追加滚动：时间差的两倍（根据drift.coef参数）距离
                let dist = (_this.options.drift.timeTHR - timeOffset) * _this.options.drift.coef;
                value = { x: ~~(item.last.x + sinValue * dist), y: ~~(item.last.y + cosValue * dist) }
            } else {
                value = item.now;
            }
            _this.slideTo({ item, value });
            //移除active属性
            item.dom.removeAttribute('active');
            Reflect.deleteProperty(item, 'active');

            //监听回调
            _this.handlers.hasOwnProperty('end') ? _this.emit('end', item) : null;
            _this.options.onEnd && _this.options.onEnd.call(_this, item);
        }
        //滚轮事件
        this.scrollFun = axThrottle(function (e) {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            //阻止默认事件
            e.preventDefault();
            if (_this.unbounds.length > 0 && (_this.unbounds.includes(this))) {
                //在未绑定事件项里滑动则不滚动，按默认事件执行；
                return false;
            }

            let item = _this.slides.find(k => k.dom === this),
                isUp = axIsScrollUp(e, _this.scrollObj.detail);
            //边界设置，确保是在活动范围
            if (isUp) {
                item.scale++;
                item.scale >= _this.options.scale.max ? item.scale = _this.options.scale.max : null;
            } else {
                item.scale--;
                item.scale <= _this.options.scale.min ? item.scale = _this.options.scale.min : null;
            }
            //不超出范围才执行，避免手动设置scale值
            if (item.scale <= _this.options.scale.max || item.scale >= _this.options.scale.min) {
                //取整数倍
                item.scale = axNumber(item.scale, { places: 0, mode: 'floor' });
                this.style.transitionDuration = _this.options.duration + 'ms';
                _this.setValues({ type: 'scale', item, value: item.scale });
                //修正translate偏移量
                setTimeout(() => {
                    _this.updateItem(item);
                }, _this.options.duration);
            }
        }, _this.options.duration, true);
        //动画完成事件
        this.slidedEvent = function (e) {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            _this.response(this);
        }
        //监听子style值变化和子节点变化
        this.mutationObs = new MutationObserver(mutations => {
            //style发生了变化
            mutations.forEach(k => {

                if (k.type === "attributes" && k.target !== this.targetDom) {
                    let item = this.slides.find(i => i.dom === k.target),
                        //节点变化后修正关系
                        matrix = this.getValues(item);
                    //scale发生了变化
                    if (matrix[0] != item.scale || matrix[3] != item.scale) {


                        //监听回调
                        this.handlers.hasOwnProperty('scaled') ? this.emit('scaled', item, matrix[0]) : null;
                        this.options.onScaled && this.options.onScaled.call(this, item, matrix[0]);
                        //监听缩放到最大和最小
                        if (item.scale === this.options.scale.min) {
                            item.boundary = 'min';
                            this.handlers.hasOwnProperty('scaledMin') ? this.emit('scaledMin', item, this.options.scale.min) : null;
                            this.options.onScaledMin && this.options.onScaledMin.call(this, item, this.options.scale.min);
                        } else if (item.scale === this.options.scale.max) {
                            item.boundary = 'max';
                            this.handlers.hasOwnProperty('scaledMax') ? this.emit('scaledMax', item, this.options.scale.max) : null;
                            this.options.onScaledMax && this.options.onScaledMax.call(this, item, this.options.scale.max);
                        } else {
                            Reflect.deleteProperty(item, 'boundary');
                        }
                    }
                    if (matrix[4] != item.now.x || matrix[5] != item.now.y) {


                        //监听回调
                        this.handlers.hasOwnProperty('translated') ? this.emit('translated', item, { x: matrix[4], y: matrix[5] }) : null;
                        this.options.onTranslated && this.options.onTranslated.call(this, item, { x: matrix[4], y: matrix[5] });
                    }
                }
                if (k.type === "childList" && k.target === this.targetDom) {
                    //更新未绑定事件元素
                    this.options.unbound ? this.getUnbounds() : null;
                    //监听回调
                    this.handlers.hasOwnProperty('refactored') ? this.emit('refactored', '') : null;
                    this.options.onRefactored && this.options.onRefactored.call(this);
                }
            });
        });
        //监听子元素尺寸变化
        this.slideDomSizeObserver = new ResizeObserver(axDebounce((entries) => {
            //inner尺寸变化后修正关系
            for (let k of entries) {
                this.response(k.target);
                //监听回调
                this.handlers.hasOwnProperty('resized') ? this.emit('resized', k, k.target) : null;
                this.options.onResized && this.options.onResized.call(this, k, k.target);
            }
        }));
        //监听父元素尺寸变化
        this.targetDomSizeObserver = new ResizeObserver(axDebounce((entries) => {
            //outer尺寸变化后修正关系
            for (let k of entries) {
                this.response();
                //监听回调
                this.handlers.hasOwnProperty('resized') ? this.emit('resized', k, this.targetDom) : null;
                this.options.onResized && this.options.onResized.call(this, k, this.targetDom);
            }
        }));
        //初始化
        this.options.init ? this.init() : this.destroyed = true;
    }
    init(callback) {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'cage');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;
        //存储未绑定事件元素
        this.getUnbounds();
        //获得slides
        this.slides = [...this.targetDom.children].map(k => {
            return { dom: k, last: { x: 0, y: 0 }, now: { x: 0, y: 0 }, difference: { x: 0, y: 0 }, size: { width: 0, height: 0 }, offset: { x: 0, y: 0 }, shape: '', range: { x: [], y: [] }, scale: 1 };
        });

        //设置风格
        this.setAttribute();
        //获得父元素宽高尺寸
        this.updateTarget();
        //手指按下的X/Y轴值
        this.distStart = { x: 0, y: 0 };
        //父绑定监听事件
        this.addTargetEvent();
        //子绑定监听事件
        this.slides.forEach(k => {
            this.addItemEvent(k.dom);
        });

        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        callback && callback.call(this);
        return this;
    }
    computeHYP(now, last) {
        let nowVal = { ...now },
            lastVal = { ...last },
            xOffset = Math.abs(nowVal.x - lastVal.x),
            yOffset = Math.abs(nowVal.y - lastVal.y),
            HYP = Math.sqrt(xOffset * xOffset + yOffset * yOffset),
            SIN = HYP ? (nowVal.x - lastVal.x) / HYP : 0,
            COS = HYP ? (nowVal.y - lastVal.y) / HYP : 0;
        return { HYP, SIN, COS };
    }
    setAttribute() {
        //强制添加ax-cage类
        !this.targetDom.classList.contains('ax-cage') ? this.targetDom.classList.add('ax-cage') : null;
        //外容器强制要求overflow=hidden
        axStyle(this.targetDom).overflow !== 'hidden' ? this.targetDom.style.overflow = 'hidden' : null;
        //要求使用定位，至少是position:relative，否则offset值会出错
        axStyle(this.targetDom).position === 'static' ? this.targetDom.style.position = 'relative' : null;
    }
    getUnbounds() {
        this.unbounds = this.getDomsFromArg(this.options.unbound.selector);
        this.unbounds.forEach(k => k.setAttribute(this.options.unbound.attr, ''));
    }
    getDomsFromArg(arg) {
        let arr = [];
        if (arg) {
            let unboundType = axType(arg);
            if (unboundType === 'String') {
                arr = [...this.targetDom.querySelectorAll(arg)];
            } else if (unboundType === 'Array') {
                arr = arg;
            } else if (unboundType.includes('HTML')) {
                arr.push(arg);
            }
        }
        return arr;
    }
    fitTo(opt) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        let options = Object.assign({ item: this.slides[0], placement: 'left-top', duration: this.options.duration }, opt),
            type = axType(options.item),
            item = type.includes('HTML') ? this.slides.find(k => k.dom === options.item) : type === 'String' ? this.targetDom.querySelector(options.item) : options.item,
            value = {},
            scaleOffset = this.getScaleOffset(item);

        if (!item || this.unbounds.includes(item.dom)) {
            return false;
        }
        //有item则执行以下代码
        switch (options.placement) {
            case 'right-top':
                value = { x: item.difference.x, y: 0 };
                break;
            case 'left-bottom':
                value = { x: 0, y: item.difference.y };
                break;
            case 'right-bottom':
                value = { ...item.difference };
                break;
            case 'left':
                value = { x: 0, y: item.now.y };
                break;
            case 'right':
                value = { x: item.difference.x, y: item.now.y };
                break;
            case 'top':
                value = { x: item.now.x, y: 0 };
                break;
            case 'bottom':
                value = { x: item.now.x, y: item.difference.y };
                break;
            default:
                value = { x: 0, y: 0 };
        }
        //修正scale之后的偏移值
        options.placement.includes('left') || options.placement.includes('right') ? value.x += scaleOffset.x : null;
        options.placement.includes('top') || options.placement.includes('bottom') ? value.y += scaleOffset.y : null;

        //实验之后，使用异步不会跳跃
        //setTimeout(() => {
        item.dom.style.transitionDuration = options.duration + 'ms';
        this.setValues({ value, item });
        // }, 0);
        //回调
        setTimeout(() => {
            //监听回调
            this.handlers.hasOwnProperty('fitted') ? this.emit('fitted', item) : null;
            this.options.onFitted && this.options.onFitted.call(this, item);
            options.callback && options.callback.call(this, item);
        }, options.duration);

    }

    getEdgeValue(data, item) {
        let datatType = axType(data), value;
        if (datatType.includes('HTML') || datatType === 'String') {
            //data是子元素
            let dom = axIdToDom(data);
            dom ? value = { x: -dom.offsetLeft, y: -dom.offsetTop } : { x: 0, y: 0 };
        } else {
            value = { ...data };
        }
        if (!axIsEmpty(value) && value.hasOwnProperty('x') && value.hasOwnProperty('y')) {
            if (value.x < item.range.x[0]) {
                value.x = item.range.x[0];
            } else if (value.x > item.range.x[1]) {
                value.x = item.range.x[1];
            }
            if (value.y < item.range.y[0]) {
                value.y = item.range.y[0];
            } else if (value.y > item.range.y[1]) {
                value.y = item.range.y[1];
            }
        }
        return value;
    }
    slideTo(opt) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //没有value则不再执行，另外如果运动不可干预且还在运动中也不再执行
        let options = Object.assign({ item: this.slides[0] }, opt),
            type = axType(options.item),
            item = type.includes('HTML') ? this.slides.find(k => k.dom === options.item) : type === 'String' ? this.targetDom.querySelector(options.item) : options.item;
        if (!item || this.unbounds.includes(item.dom)) {
            return false;
        }
        //有item则执行以下代码
        let correction = this.getEdgeValue(options.value, item),
            ratio = this.computeHYP(item.now, correction).HYP / this.hypotenuse,
            //要滚动的距离大于了外围尺寸则自然增加动画时间，动画时间只要整数
            duration = options.duration || options.duration == 0 ? `${options.duration}ms` : ratio > 1 ? Math.trunc(this.options.duration * ratio) + 'ms' : `${this.options.duration}ms`;

        if (JSON.stringify(correction) == JSON.stringify(item.now)) {
            //相同位置则不用滚动了
            return false;
        }
        item.dom.style.transitionDuration = duration;
        this.setValues({ value: correction, item });
        //回调
        setTimeout(() => {
            //监听回调
            this.handlers.hasOwnProperty('slided') ? this.emit('slided', { raw: options.value, correction, item }) : null;
            this.options.onSlided && this.options.onSlided.call(this, { raw: options.value, correction, item });
            options.callback && options.callback.call(this, { raw: options.value, correction, item });
        }, parseInt(duration));
    }
    getCoordValue(e) {
        return this.isMobi ? { x: e.changedTouches[0].clientX, y: e.changedTouches[0].clientY } : { x: e.clientX, y: e.clientY };
    }
    updateTarget() {
        let targetStyle = axStyle(this.targetDom),
            height = this.targetDom.getBoundingClientRect().height - (parseInt(targetStyle.borderTopWidth) + parseInt(targetStyle.borderBottomWidth) + parseInt(targetStyle.paddingTop) + parseInt(targetStyle.paddingBottom)),
            width = this.targetDom.getBoundingClientRect().width - (parseInt(targetStyle.borderLeftWidth) + parseInt(targetStyle.borderRightWidth) + parseInt(targetStyle.paddingLeft) + parseInt(targetStyle.paddingRight)),
            hypotenuse = this.computeHYP({ x: 0, y: 0 }, { x: width, y: height }).HYP;
        this.targetHeight != height ? this.targetHeight = height : null;
        this.targetWidth != width ? this.targetWidth = width : null;
        //计算斜边长度
        this.hypotenuse != hypotenuse ? this.hypotenuse = hypotenuse : null;
    }
    updateItem(obj = this.slides[0]) {
        let type = axType(obj),
            item = type.includes('HTML') ? this.slides.find(k => k.dom === obj) : type === 'String' ? this.targetDom.querySelector(obj) : obj;
        if (!item || this.unbounds.includes(item.dom)) {
            return false;
        }
        //更新内外容器的尺寸差值，outer的border、padding不作为高度一部分
        item.size.height = item.dom.getBoundingClientRect().height;
        item.size.width = item.dom.getBoundingClientRect().width;
        item.difference = { x: this.targetWidth - item.size.width, y: this.targetHeight - item.size.height }
        //获取偏移值
        item.offset = { x: item.dom.offsetLeft, y: item.dom.offsetTop };
        //定义滑块形状
        if (item.difference.x > 0 && item.difference.y > 0) {
            item.shape = 'baby';
        } else if (item.difference.y > 0 && item.difference.x <= 0) {
            item.shape = 'river';
        } else if (item.difference.x > 0 && item.difference.y <= 0) {
            item.shape = 'chimney';
        } else {
            item.shape = 'cover';
        }
        this.fixedTranslate(item);
    }
    getScaleOffset(item) {
        if (item.scale > this.options.scale.min) {
            return { x: (item.size.width / item.scale) / 2 * (item.scale - 1), y: (item.size.height / item.scale) / 2 * (item.scale - 1) }
        } else {
            return { x: 0, y: 0 }
        }
    }
    fixedTranslate(item) {
        //初始化校正
        let matrix = this.getValues(item),
            value = { x: matrix[4], y: matrix[5] },
            scaleOffset = this.getScaleOffset(item),
            insideFit = (axis) => {
                //offset不会变，默认是0；如果使用margin或positon，offset则可能不是0，已经产生了平移状态需要加上value[axis]值（未移动是0）
                if (item.offset[axis] - scaleOffset[axis] + value[axis] < 0) {
                    value[axis] = -item.offset[axis] + scaleOffset[axis];
                } else if (item.offset[axis] - scaleOffset[axis] + value[axis] > item.difference[axis]) {
                    value[axis] = item.difference[axis] - item.offset[axis] + scaleOffset[axis];
                }
                //offset值不会随translate或scale变化而变化，scale加一倍等于宽高都加一倍
                item.range[axis] = [0 - item.offset[axis] + scaleOffset[axis], item.difference[axis] - item.offset[axis] + scaleOffset[axis]];
            },
            exceedFit = (axis) => {
                if (value[axis] < item.difference[axis] - item.offset[axis] + scaleOffset[axis]) {
                    value[axis] = item.difference[axis] - item.offset[axis] + scaleOffset[axis];
                } else if (value[axis] > -item.offset[axis] + scaleOffset[axis]) {
                    value[axis] = -item.offset[axis] + scaleOffset[axis];
                }
                item.range[axis] = [item.difference[axis] - item.offset[axis] + scaleOffset[axis], -item.offset[axis] + scaleOffset[axis]];
            }

        switch (item.shape) {
            case 'baby':
                insideFit('x');
                insideFit('y');
                break;
            case 'river':
                insideFit('y');
                exceedFit('x');
                break;
            case 'chimney':
                insideFit('x');
                exceedFit('y');
                break;
            default:
                exceedFit('x');
                exceedFit('y');
        }
        JSON.stringify(value) != JSON.stringify(item.now) ? this.setValues({ value, item }) : null;
    }
    response(target = 'all') {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        if (target === 'all') {
            //全部更新
            this.updateTarget();
            this.slides.forEach(k => {
                this.updateItem(k);
            });
        } else if (target === 'parent') {
            //只更新父容器
            this.updateTarget();
        } else if (target === 'chidren') {
            //只更新子元素
            this.slides.forEach(k => {
                this.updateItem(k);
            });
        } else {
            //更新某个子元素，item是某个节点或选择器
            let type = axType(target),
                item = type.includes('HTML') ? this.slides.find(k => k.dom === target) : type === 'String' ? this.targetDom.querySelector(target) : target;
            if (item) {
                this.updateItem(item);
            }
        }
    }
    getValues(obj = this.slides[0]) {
        let type = axType(obj),
            item = type.includes('HTML') ? this.slides.find(k => k.dom === obj) : type === 'String' ? this.targetDom.querySelector(obj) : obj;
        if (!item) {
            return false;
        }
        //从style的transform中获得偏移值
        let value = axStyle(item.dom).transform;
        if (value === 'none') {
            return  [1, 0, 0, 1, 0, 0];
        } else {
            return  value.replace(/[^0-9.\-,]/g, '').split(',').map(k => Number(k));
        }
    }
    setValues(opt) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        let options = Object.assign({ item: this.slides[0], type: 'translate', apply: true }, opt),
            type = axType(options.item),
            item = type.includes('HTML') ? this.slides.find(k => k.dom === options.item) : type === 'String' ? this.targetDom.querySelector(options.item) : options.item;
        if (!item || this.unbounds.includes(item.dom)) {
            return false;
        }
        let matrix = this.getValues(item),
            arr = [...matrix],
            valueType = axType(options.value),
            obj = {};
        if (valueType === 'Object') {
            //传入一个对象{x:'',y:''}
            obj = { ...options.value };
        } else if (valueType === 'Number') {
            //传入一个值
            obj = { x: options.value, y: options.value };
        } else {
            //其他类型，强制转为数字
            let val = ~~options.value;
            if (!val) {
                val = options.type === 'scale' ? 1 : 0;
            }
            obj = { x: val, y: val };
        }
        //不同类型，矩阵值不同，分开设置
        if (options.type === 'zero') {
            arr[0] = 1;
            arr[3] = 1;
            arr[4] = 0;
            arr[5] = 0;
        } else if (options.type === 'scale') {
            arr[0] = obj.x;
            arr[3] = obj.y;
        } else {
            arr[4] = obj.x;
            arr[5] = obj.y;
        }
        if (JSON.stringify(arr) != JSON.stringify(matrix)) {
            item.dom.style.transform = `matrix(${arr.join()})`;
           
            //修正
            if (options.apply) {
                if (options.type === 'zero') {
                    item.last = { x: 0, y: 0 };
                    item.now = { x: 0, y: 0 };
                    item.scale = 1;
                } else if (options.type === 'scale') {
                    item.scale = obj.x || obj.y;
                } else {
                    item.last = { ...obj };
                    item.now = { ...obj };
                }
            }

        }
    }
    reset() {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.slides.forEach(k => {
            this.setValues({ type: 'zero', item: k });
            //强制移除动画时间0ms
            k.dom.style.cssText = k.dom.style.cssText.replace('transition-duration: 0ms', '');
        });
        //取消moving属性
        this.targetDom.removeAttribute('moving');
        //更新状态，以免没有执行end()事件导致还能触发move()事件
        this.started = false;
    }
    addItemEvent(dom) {
        //手指放下
        dom.addEventListener(this.events[0], this.startFun);
        //手指拖拽
        dom.addEventListener(this.events[1], this.moveFun);
        //手指松开
        dom.addEventListener(this.events[2], this.endFun);
        //绑定鼠标缩放
        !this.isMobi && this.options.scale.enable ? dom.addEventListener(this.scrollObj.event, this.scrollFun) : null;
        //监听style变化
        this.mutationObs.observe(dom, {
            attributes: true,
            attributeFilter: ['style']
        });
        //监听运动完毕执行
        dom.addEventListener("transitionend", this.slidedEvent);
        //监听尺寸变化
        this.slideDomSizeObserver.observe(dom);
    }
    removeItemEvent(dom) {
        //手指放下
        dom.removeEventListener(this.events[0], this.startFun);
        //手指拖拽
        dom.removeEventListener(this.events[1], this.moveFun);
        //手指松开
        dom.removeEventListener(this.events[2], this.endFun);
        //绑定鼠标缩放
        !this.isMobi && this.options.scale.enable ? dom.removeEventListener(this.scrollObj.event, this.scrollFun) : null;
        //监听运动完毕执行
        dom.removeEventListener("transitionend", this.slidedEvent);
        //监听尺寸变化
        this.slideDomSizeObserver.unobserve(dom);
    }
    addTargetEvent() {

        //丢失目标对象事件
        if (!this.isMobi) {
            //离开父层结束事件
            this.targetDom.addEventListener('mouseleave', this.endFun);
        } else {
            //移动端丢失对象结束事件
            this.targetDom.addEventListener('touchcancel', this.endFun);
        }
        //父节点变化监听
        this.mutationObs.observe(this.targetDom, {
            childList: true,
        });
        //父尺寸大小监听
        this.targetDomSizeObserver.observe(this.targetDom);
    }
    removeTargetEvent() {
        //this.targetDom.removeEventListener(this.events[2], this.endFun);
        if (!this.isMobi) {
            this.targetDom.removeEventListener('mouseleave', this.endFun);
        }
        this.targetDomSizeObserver.unobserve(this.targetDom);
    }
    add(data, callback) {
        let arr = axGetDoms(data),
            adds = [];
        if (arr.length === 0) {
            return false;
        }
        adds = arr.filter(Boolean).map(k => {
            let item = this.slides.find(i => i.dom === k),
                obj = null;
            if (item && item.disabled) {
                Reflect.deleteProperty(item, 'disabled');
                item.dom.removeAttribute('disabled');
                obj = k;
                this.addItemEvent(k);
            } else {
                obj = { dom: k, last: { x: 0, y: 0 }, now: { x: 0, y: 0 }, difference: { x: 0, y: 0 }, size: { width: 0, height: 0 }, offset: { x: 0, y: 0 }, shape: '', range: { x: [], y: [] }, scale: 1 };
                this.addItemEvent(k);
                this.slides.push(obj);
            }
            return obj;
        });
        //监听回调
        this.handlers.hasOwnProperty('add') ? this.emit('add', adds) : null;
        this.options.onAdd && this.options.onAdd.call(this, adds);
        callback && callback.call(this, adds);
        return this;
    }
    remove(data, callback) {
        let arr = axGetDoms(data), removes = this.slides.filter(k => !k.disabled && arr.includes(k.dom));
        if (arr.length === 0) {
            return false;
        }
        removes.forEach(k => {
            k.disabled = true;
            k.dom.setAttribute('disabled', '');
            //取消监听
            this.removeItemEvent(k.dom);
        });
        //监听回调
        this.handlers.hasOwnProperty('remove') ? this.emit('remove', removes) : null;
        this.options.onRemove && this.options.onRemove.call(this, removes);
        callback && callback.call(this, removes);
        return this;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //更新参数
        this.options = axExtend(this.options, setting);
        //归零
        this.reset();
        //根据参数初始化实例
        this.options.init ? this.init() : null;
        //监听回调
        this.handlers.hasOwnProperty('updated') ? this.emit('updated', '') : null;
        this.options.onUpdated && this.options.onUpdated.call(this);
        callback && callback.call(this);
        return this;
    }
    destroy(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //归零
        this.reset();
        //取消绑定事件
        this.slides.forEach(k => {
            this.removeItemEvent(k.dom);
        });
        this.removeTargetEvent();
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/**
* 插件：拖拽移动获得移动距离；使用方法：new axTow('#id',{参数})
*/
class axTow {
    constructor(elem, options) {
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            parent: document,//move事件上的父元素，默认document，
            onInit: '',//回调函数，初始化后执行，无参数
            onStart: null,//回调函数，按下/触摸开始执行，支持一个参数即当前数据，包括target、coord和diff属性
            onMove: null,//回调函数，移动中执行，支持一个参数即当前数据，包括target、coord和diff属性
            onEnd: null,//回调函数，结束后执行，支持一个参数即当前数据，包括target、coord和diff属性
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        //父节点
        this.parentDom = axIdToDom(this.options.parent);
        if (!this.parentDom) { return false; }
        //回调监听
        this.handlers = {};
        //核心事件,开始事件
        this.coord = { x: 0, y: 0 };
        this.diff = { x: 0, y: 0 };
        this.startFun = (e) => {
            this.coord = axGetClientObj(e);
            this.parentDom.addEventListener(axEventMap[1], this.moveFun, { passive: false });
            this.parentDom.addEventListener(axEventMap[2], this.endFun, { passive: false });
            let param = { target: e.target, coord: this.coord, diff: this.diff,originEvent:e };
            this.handlers.hasOwnProperty('start') ? this.emit('start', param) : null;
            this.options.onStart && this.options.onStart.call(this, param);
        }
        //移动事件
        this.moveFun = (e) => {
            let clientObj = axGetClientObj(e);
            this.diff = { x: clientObj.x - this.coord.x, y: clientObj.y - this.coord.y };
            let param = { target: e.target, coord: clientObj, diff: this.diff,originEvent:e };
            this.handlers.hasOwnProperty('move') ? this.emit('move', param) : null;
            this.options.onMove && this.options.onMove.call(this, param);
        }
        //结束事件
        this.endFun = (e) => {
            //设置完成
            this.coord = axGetClientObj(e);
            let param = { target: e.target, coord: this.coord, diff: this.diff,originEvent:e };
            this.handlers.hasOwnProperty('end') ? this.emit('end', param) : null;
            this.options.onEnd && this.options.onEnd.call(this, param);
            this.setCompleted();
        }
        //取消事件
        this.cancelFun = (e) => {
            this.setCompleted();
            this.handlers.hasOwnProperty('cancel') ? this.emit('cancel', { target: e.target, coord: this.coord, diff: this.diff }) : null;
            this.options.onCancel && this.options.onCancel.call(this, { target: e.target, coord: this.coord, diff: this.diff });
        }
        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'tow');
        this.destroyed = false;
        //先移除再添加，因为存在多次init的情况，避免重复绑定事件
        this.targetDom.removeEventListener(axEventMap[0], this.startFun);
        this.targetDom.addEventListener(axEventMap[0], this.startFun, { passive: false });
        this.parentDom.addEventListener(axEventMap[3], this.cancelFun, { passive: false });
        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    setCompleted() {
        this.diff = { x: 0, y: 0 };
        this.parentDom.removeEventListener(axEventMap[1], this.moveFun);
        this.parentDom.removeEventListener(axEventMap[2], this.endFun);
    }
    destroy(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //如果有剩余内容则先执行完毕
        this.setCompleted();
        this.targetDom.removeEventListener(axEventMap[0], this.startFun);
        this.parentDom.removeEventListener(axEventMap[3], this.cancelFun);
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
 * 轮播插件axSwipe，使用方法：new axSwipe('#id',{参数})。源自Swiper v8.1.4 from https://swiperjs.com | AXUI做了部分修改
 */
(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
        typeof define === 'function' && define.amd ? define(factory) :
            (global = typeof globalThis !== 'undefined' ? globalThis : global || self, global.axSwiper = factory());
    //对使用axSwipe属性的元素使用本插件
/*     document.querySelectorAll('[axSwiper]').forEach(element => {
        new axSwiper(element,axAttrToJson(element, 'axSwiper'));
    }); */
})(this, (function () {
    'use strict';

    /**
     * SSR Window 4.0.2
     * Better handling for window object in SSR environment
     * https://github.com/nolimits4web/ssr-window
     *
     * Copyright 2021, Vladimir Kharlampidi
     *
     * Licensed under MIT
     *
     * Released on: December 13, 2021
     */

    /* eslint-disable no-param-reassign */
    function isObject$1(obj) {
        return obj !== null && typeof obj === 'object' && 'constructor' in obj && obj.constructor === Object;
    }

    function extend$1(target, src) {
        if (target === void 0) {
            target = {};
        }

        if (src === void 0) {
            src = {};
        }

        Object.keys(src).forEach(key => {
            if (typeof target[key] === 'undefined') target[key] = src[key]; else if (isObject$1(src[key]) && isObject$1(target[key]) && Object.keys(src[key]).length > 0) {
                extend$1(target[key], src[key]);
            }
        });
    }

    const ssrDocument = {
        body: {},

        addEventListener() { },

        removeEventListener() { },

        activeElement: {
            blur() { },

            nodeName: ''
        },

        querySelector() {
            return null;
        },

        querySelectorAll() {
            return [];
        },

        getElementById() {
            return null;
        },

        createEvent() {
            return {
                initEvent() { }

            };
        },

        createElement() {
            return {
                children: [],
                childNodes: [],
                style: {},

                setAttribute() { },

                getElementsByTagName() {
                    return [];
                }

            };
        },

        createElementNS() {
            return {};
        },

        importNode() {
            return null;
        },

        location: {
            hash: '',
            host: '',
            hostname: '',
            href: '',
            origin: '',
            pathname: '',
            protocol: '',
            search: ''
        }
    };

    function getDocument() {
        const doc = typeof document !== 'undefined' ? document : {};
        extend$1(doc, ssrDocument);
        return doc;
    }

    const ssrWindow = {
        document: ssrDocument,
        navigator: {
            userAgent: ''
        },
        location: {
            hash: '',
            host: '',
            hostname: '',
            href: '',
            origin: '',
            pathname: '',
            protocol: '',
            search: ''
        },
        history: {
            replaceState() { },

            pushState() { },

            go() { },

            back() { }

        },
        CustomEvent: function CustomEvent() {
            return this;
        },

        addEventListener() { },

        removeEventListener() { },

        getComputedStyle() {
            return {
                getPropertyValue() {
                    return '';
                }

            };
        },

        Image() { },

        Date() { },

        screen: {},

        setTimeout() { },

        clearTimeout() { },

        matchMedia() {
            return {};
        },

        requestAnimationFrame(callback) {
            if (typeof setTimeout === 'undefined') {
                callback();
                return null;
            }

            return setTimeout(callback, 0);
        },

        cancelAnimationFrame(id) {
            if (typeof setTimeout === 'undefined') {
                return;
            }

            clearTimeout(id);
        }

    };

    function getWindow() {
        const win = typeof window !== 'undefined' ? window : {};
        extend$1(win, ssrWindow);
        return win;
    }

    /**
     * Dom7 4.0.4
     * Minimalistic JavaScript library for DOM manipulation, with a jQuery-compatible API
     * https://framework7.io/docs/dom7.html
     *
     * Copyright 2022, Vladimir Kharlampidi
     *
     * Licensed under MIT
     *
     * Released on: January 11, 2022
     */
    /* eslint-disable no-proto */

    function makeReactive(obj) {
        const proto = obj.__proto__;
        Object.defineProperty(obj, '__proto__', {
            get() {
                return proto;
            },

            set(value) {
                proto.__proto__ = value;
            }

        });
    }

    class Dom7 extends Array {
        constructor(items) {
            if (typeof items === 'number') {
                super(items);
            } else {
                super(...(items || []));
                makeReactive(this);
            }
        }

    }

    function arrayFlat(arr) {
        if (arr === void 0) {
            arr = [];
        }

        const res = [];
        arr.forEach(el => {
            if (Array.isArray(el)) {
                res.push(...arrayFlat(el));
            } else {
                res.push(el);
            }
        });
        return res;
    }

    function arrayFilter(arr, callback) {
        return Array.prototype.filter.call(arr, callback);
    }

    function arrayUnique(arr) {
        const uniqueArray = [];

        for (let i = 0; i < arr.length; i += 1) {
            if (uniqueArray.indexOf(arr[i]) === -1) uniqueArray.push(arr[i]);
        }

        return uniqueArray;
    }


    function qsa(selector, context) {
        if (typeof selector !== 'string') {
            return [selector];
        }

        const a = [];
        const res = context.querySelectorAll(selector);

        for (let i = 0; i < res.length; i += 1) {
            a.push(res[i]);
        }

        return a;
    }

    function $(selector, context) {
        const window = getWindow();
        const document = getDocument();
        let arr = [];

        if (!context && selector instanceof Dom7) {
            return selector;
        }

        if (!selector) {
            return new Dom7(arr);
        }

        if (typeof selector === 'string') {
            const html = selector.trim();

            if (html.indexOf('<') >= 0 && html.indexOf('>') >= 0) {
                let toCreate = 'div';
                if (html.indexOf('<li') === 0) toCreate = 'ul';
                if (html.indexOf('<tr') === 0) toCreate = 'tbody';
                if (html.indexOf('<td') === 0 || html.indexOf('<th') === 0) toCreate = 'tr';
                if (html.indexOf('<tbody') === 0) toCreate = 'table';
                if (html.indexOf('<option') === 0) toCreate = 'select';
                const tempParent = document.createElement(toCreate);
                tempParent.innerHTML = html;

                for (let i = 0; i < tempParent.childNodes.length; i += 1) {
                    arr.push(tempParent.childNodes[i]);
                }
            } else {
                arr = qsa(selector.trim(), context || document);
            } // arr = qsa(selector, document);

        } else if (selector.nodeType || selector === window || selector === document) {
            arr.push(selector);
        } else if (Array.isArray(selector)) {
            if (selector instanceof Dom7) return selector;
            arr = selector;
        }

        return new Dom7(arrayUnique(arr));
    }

    $.fn = Dom7.prototype; // eslint-disable-next-line

    function addClass() {
        for (var _len = arguments.length, classes = new Array(_len), _key = 0; _key < _len; _key++) {
            classes[_key] = arguments[_key];
        }

        const classNames = arrayFlat(classes.map(c => c.split(' ')));
        this.forEach(el => {
            el.classList.add(...classNames);
        });
        return this;
    }

    function removeClass() {
        for (var _len2 = arguments.length, classes = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
            classes[_key2] = arguments[_key2];
        }

        const classNames = arrayFlat(classes.map(c => c.split(' ')));
        this.forEach(el => {
            el.classList.remove(...classNames);
        });
        return this;
    }

    function toggleClass() {
        for (var _len3 = arguments.length, classes = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
            classes[_key3] = arguments[_key3];
        }

        const classNames = arrayFlat(classes.map(c => c.split(' ')));
        this.forEach(el => {
            classNames.forEach(className => {
                el.classList.toggle(className);
            });
        });
    }

    function hasClass() {
        for (var _len4 = arguments.length, classes = new Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
            classes[_key4] = arguments[_key4];
        }

        const classNames = arrayFlat(classes.map(c => c.split(' ')));
        return arrayFilter(this, el => {
            return classNames.filter(className => el.classList.contains(className)).length > 0;
        }).length > 0;
    }

    function attr(attrs, value) {
        if (arguments.length === 1 && typeof attrs === 'string') {
            // Get attr
            if (this[0]) return this[0].getAttribute(attrs);
            return undefined;
        } // Set attrs


        for (let i = 0; i < this.length; i += 1) {
            if (arguments.length === 2) {
                // String
                this[i].setAttribute(attrs, value);
            } else {
                // Object
                for (const attrName in attrs) {
                    this[i][attrName] = attrs[attrName];
                    this[i].setAttribute(attrName, attrs[attrName]);
                }
            }
        }

        return this;
    }

    function removeAttr(attr) {
        for (let i = 0; i < this.length; i += 1) {
            this[i].removeAttribute(attr);
        }

        return this;
    }

    function transform(transform) {
        for (let i = 0; i < this.length; i += 1) {
            this[i].style.transform = transform;
        }

        return this;
    }

    function transition$1(duration) {
        for (let i = 0; i < this.length; i += 1) {
            this[i].style.transitionDuration = typeof duration !== 'string' ? `${duration}ms` : duration;
        }

        return this;
    }

    function on() {
        for (var _len5 = arguments.length, args = new Array(_len5), _key5 = 0; _key5 < _len5; _key5++) {
            args[_key5] = arguments[_key5];
        }

        let [eventType, targetSelector, listener, capture] = args;

        if (typeof args[1] === 'function') {
            [eventType, listener, capture] = args;
            targetSelector = undefined;
        }

        if (!capture) capture = false;

        function handleLiveEvent(e) {
            const target = e.target;
            if (!target) return;
            const eventData = e.target.dom7EventData || [];

            if (eventData.indexOf(e) < 0) {
                eventData.unshift(e);
            }

            if ($(target).is(targetSelector)) listener.apply(target, eventData); else {
                const parents = $(target).parents(); // eslint-disable-line

                for (let k = 0; k < parents.length; k += 1) {
                    if ($(parents[k]).is(targetSelector)) listener.apply(parents[k], eventData);
                }
            }
        }

        function handleEvent(e) {
            const eventData = e && e.target ? e.target.dom7EventData || [] : [];

            if (eventData.indexOf(e) < 0) {
                eventData.unshift(e);
            }

            listener.apply(this, eventData);
        }

        const events = eventType.split(' ');
        let j;

        for (let i = 0; i < this.length; i += 1) {
            const el = this[i];

            if (!targetSelector) {
                for (j = 0; j < events.length; j += 1) {
                    const event = events[j];
                    if (!el.dom7Listeners) el.dom7Listeners = {};
                    if (!el.dom7Listeners[event]) el.dom7Listeners[event] = [];
                    el.dom7Listeners[event].push({
                        listener,
                        proxyListener: handleEvent
                    });
                    el.addEventListener(event, handleEvent, capture);
                }
            } else {
                // Live events
                for (j = 0; j < events.length; j += 1) {
                    const event = events[j];
                    if (!el.dom7LiveListeners) el.dom7LiveListeners = {};
                    if (!el.dom7LiveListeners[event]) el.dom7LiveListeners[event] = [];
                    el.dom7LiveListeners[event].push({
                        listener,
                        proxyListener: handleLiveEvent
                    });
                    el.addEventListener(event, handleLiveEvent, capture);
                }
            }
        }

        return this;
    }

    function off() {
        for (var _len6 = arguments.length, args = new Array(_len6), _key6 = 0; _key6 < _len6; _key6++) {
            args[_key6] = arguments[_key6];
        }

        let [eventType, targetSelector, listener, capture] = args;

        if (typeof args[1] === 'function') {
            [eventType, listener, capture] = args;
            targetSelector = undefined;
        }

        if (!capture) capture = false;
        const events = eventType.split(' ');

        for (let i = 0; i < events.length; i += 1) {
            const event = events[i];

            for (let j = 0; j < this.length; j += 1) {
                const el = this[j];
                let handlers;

                if (!targetSelector && el.dom7Listeners) {
                    handlers = el.dom7Listeners[event];
                } else if (targetSelector && el.dom7LiveListeners) {
                    handlers = el.dom7LiveListeners[event];
                }

                if (handlers && handlers.length) {
                    for (let k = handlers.length - 1; k >= 0; k -= 1) {
                        const handler = handlers[k];

                        if (listener && handler.listener === listener) {
                            el.removeEventListener(event, handler.proxyListener, capture);
                            handlers.splice(k, 1);
                        } else if (listener && handler.listener && handler.listener.dom7proxy && handler.listener.dom7proxy === listener) {
                            el.removeEventListener(event, handler.proxyListener, capture);
                            handlers.splice(k, 1);
                        } else if (!listener) {
                            el.removeEventListener(event, handler.proxyListener, capture);
                            handlers.splice(k, 1);
                        }
                    }
                }
            }
        }

        return this;
    }

    function trigger() {
        const window = getWindow();

        for (var _len9 = arguments.length, args = new Array(_len9), _key9 = 0; _key9 < _len9; _key9++) {
            args[_key9] = arguments[_key9];
        }

        const events = args[0].split(' ');
        const eventData = args[1];

        for (let i = 0; i < events.length; i += 1) {
            const event = events[i];

            for (let j = 0; j < this.length; j += 1) {
                const el = this[j];

                if (window.CustomEvent) {
                    const evt = new window.CustomEvent(event, {
                        detail: eventData,
                        bubbles: true,
                        cancelable: true
                    });
                    el.dom7EventData = args.filter((data, dataIndex) => dataIndex > 0);
                    el.dispatchEvent(evt);
                    el.dom7EventData = [];
                    delete el.dom7EventData;
                }
            }
        }

        return this;
    }

    function transitionEnd$1(callback) {
        const dom = this;

        function fireCallBack(e) {
            if (e.target !== this) return;
            callback.call(this, e);
            dom.off('transitionend', fireCallBack);
        }

        if (callback) {
            dom.on('transitionend', fireCallBack);
        }

        return this;
    }

    function outerWidth(includeMargins) {
        if (this.length > 0) {
            if (includeMargins) {
                const styles = this.styles();
                return this[0].offsetWidth + parseFloat(styles.getPropertyValue('margin-right')) + parseFloat(styles.getPropertyValue('margin-left'));
            }

            return this[0].offsetWidth;
        }

        return null;
    }

    function outerHeight(includeMargins) {
        if (this.length > 0) {
            if (includeMargins) {
                const styles = this.styles();
                return this[0].offsetHeight + parseFloat(styles.getPropertyValue('margin-top')) + parseFloat(styles.getPropertyValue('margin-bottom'));
            }

            return this[0].offsetHeight;
        }

        return null;
    }

    function offset() {
        if (this.length > 0) {
            const window = getWindow();
            const document = getDocument();
            const el = this[0];
            const box = el.getBoundingClientRect();
            const body = document.body;
            const clientTop = el.clientTop || body.clientTop || 0;
            const clientLeft = el.clientLeft || body.clientLeft || 0;
            const scrollTop = el === window ? window.scrollY : el.scrollTop;
            const scrollLeft = el === window ? window.scrollX : el.scrollLeft;
            return {
                top: box.top + scrollTop - clientTop,
                left: box.left + scrollLeft - clientLeft
            };
        }

        return null;
    }

    function styles() {
        const window = getWindow();
        if (this[0]) return window.getComputedStyle(this[0], null);
        return {};
    }

    function css(props, value) {
        const window = getWindow();
        let i;

        if (arguments.length === 1) {
            if (typeof props === 'string') {
                // .css('width')
                if (this[0]) return window.getComputedStyle(this[0], null).getPropertyValue(props);
            } else {
                // .css({ width: '100px' })
                for (i = 0; i < this.length; i += 1) {
                    for (const prop in props) {
                        this[i].style[prop] = props[prop];
                    }
                }

                return this;
            }
        }

        if (arguments.length === 2 && typeof props === 'string') {
            // .css('width', '100px')
            for (i = 0; i < this.length; i += 1) {
                this[i].style[props] = value;
            }

            return this;
        }

        return this;
    }

    function each(callback) {
        if (!callback) return this;
        this.forEach((el, index) => {
            callback.apply(el, [el, index]);
        });
        return this;
    }

    function filter(callback) {
        const result = arrayFilter(this, callback);
        return $(result);
    }

    function html(html) {
        if (typeof html === 'undefined') {
            return this[0] ? this[0].innerHTML : null;
        }

        for (let i = 0; i < this.length; i += 1) {
            this[i].innerHTML = html;
        }

        return this;
    }

    function text(text) {
        if (typeof text === 'undefined') {
            return this[0] ? this[0].textContent.trim() : null;
        }

        for (let i = 0; i < this.length; i += 1) {
            this[i].textContent = text;
        }

        return this;
    }

    function is(selector) {
        const window = getWindow();
        const document = getDocument();
        const el = this[0];
        let compareWith;
        let i;
        if (!el || typeof selector === 'undefined') return false;

        if (typeof selector === 'string') {
            if (el.matches) return el.matches(selector);
            if (el.webkitMatchesSelector) return el.webkitMatchesSelector(selector);
            if (el.msMatchesSelector) return el.msMatchesSelector(selector);
            compareWith = $(selector);

            for (i = 0; i < compareWith.length; i += 1) {
                if (compareWith[i] === el) return true;
            }

            return false;
        }

        if (selector === document) {
            return el === document;
        }

        if (selector === window) {
            return el === window;
        }

        if (selector.nodeType || selector instanceof Dom7) {
            compareWith = selector.nodeType ? [selector] : selector;

            for (i = 0; i < compareWith.length; i += 1) {
                if (compareWith[i] === el) return true;
            }

            return false;
        }

        return false;
    }

    function index() {
        let child = this[0];
        let i;

        if (child) {
            i = 0; // eslint-disable-next-line

            while ((child = child.previousSibling) !== null) {
                if (child.nodeType === 1) i += 1;
            }

            return i;
        }

        return undefined;
    }

    function eq(index) {
        if (typeof index === 'undefined') return this;
        const length = this.length;

        if (index > length - 1) {
            return $([]);
        }

        if (index < 0) {
            const returnIndex = length + index;
            if (returnIndex < 0) return $([]);
            return $([this[returnIndex]]);
        }

        return $([this[index]]);
    }

    function append() {
        let newChild;
        const document = getDocument();

        for (let k = 0; k < arguments.length; k += 1) {
            newChild = k < 0 || arguments.length <= k ? undefined : arguments[k];

            for (let i = 0; i < this.length; i += 1) {
                if (typeof newChild === 'string') {
                    const tempDiv = document.createElement('div');
                    tempDiv.innerHTML = newChild;

                    while (tempDiv.firstChild) {
                        this[i].appendChild(tempDiv.firstChild);
                    }
                } else if (newChild instanceof Dom7) {
                    for (let j = 0; j < newChild.length; j += 1) {
                        this[i].appendChild(newChild[j]);
                    }
                } else {
                    this[i].appendChild(newChild);
                }
            }
        }

        return this;
    }

    function prepend(newChild) {
        const document = getDocument();
        let i;
        let j;

        for (i = 0; i < this.length; i += 1) {
            if (typeof newChild === 'string') {
                const tempDiv = document.createElement('div');
                tempDiv.innerHTML = newChild;

                for (j = tempDiv.childNodes.length - 1; j >= 0; j -= 1) {
                    this[i].insertBefore(tempDiv.childNodes[j], this[i].childNodes[0]);
                }
            } else if (newChild instanceof Dom7) {
                for (j = 0; j < newChild.length; j += 1) {
                    this[i].insertBefore(newChild[j], this[i].childNodes[0]);
                }
            } else {
                this[i].insertBefore(newChild, this[i].childNodes[0]);
            }
        }

        return this;
    }

    function next(selector) {
        if (this.length > 0) {
            if (selector) {
                if (this[0].nextElementSibling && $(this[0].nextElementSibling).is(selector)) {
                    return $([this[0].nextElementSibling]);
                }

                return $([]);
            }

            if (this[0].nextElementSibling) return $([this[0].nextElementSibling]);
            return $([]);
        }

        return $([]);
    }

    function nextAll(selector) {
        const nextEls = [];
        let el = this[0];
        if (!el) return $([]);

        while (el.nextElementSibling) {
            const next = el.nextElementSibling; // eslint-disable-line

            if (selector) {
                if ($(next).is(selector)) nextEls.push(next);
            } else nextEls.push(next);

            el = next;
        }

        return $(nextEls);
    }

    function prev(selector) {
        if (this.length > 0) {
            const el = this[0];

            if (selector) {
                if (el.previousElementSibling && $(el.previousElementSibling).is(selector)) {
                    return $([el.previousElementSibling]);
                }

                return $([]);
            }

            if (el.previousElementSibling) return $([el.previousElementSibling]);
            return $([]);
        }

        return $([]);
    }

    function prevAll(selector) {
        const prevEls = [];
        let el = this[0];
        if (!el) return $([]);

        while (el.previousElementSibling) {
            const prev = el.previousElementSibling; // eslint-disable-line

            if (selector) {
                if ($(prev).is(selector)) prevEls.push(prev);
            } else prevEls.push(prev);

            el = prev;
        }

        return $(prevEls);
    }

    function parent(selector) {
        const parents = []; // eslint-disable-line

        for (let i = 0; i < this.length; i += 1) {
            if (this[i].parentNode !== null) {
                if (selector) {
                    if ($(this[i].parentNode).is(selector)) parents.push(this[i].parentNode);
                } else {
                    parents.push(this[i].parentNode);
                }
            }
        }

        return $(parents);
    }

    function parents(selector) {
        const parents = []; // eslint-disable-line

        for (let i = 0; i < this.length; i += 1) {
            let parent = this[i].parentNode; // eslint-disable-line

            while (parent) {
                if (selector) {
                    if ($(parent).is(selector)) parents.push(parent);
                } else {
                    parents.push(parent);
                }

                parent = parent.parentNode;
            }
        }

        return $(parents);
    }

    function closest(selector) {
        let closest = this; // eslint-disable-line

        if (typeof selector === 'undefined') {
            return $([]);
        }

        if (!closest.is(selector)) {
            closest = closest.parents(selector).eq(0);
        }

        return closest;
    }

    function find(selector) {
        const foundElements = [];

        for (let i = 0; i < this.length; i += 1) {
            const found = this[i].querySelectorAll(selector);

            for (let j = 0; j < found.length; j += 1) {
                foundElements.push(found[j]);
            }
        }

        return $(foundElements);
    }

    function children(selector) {
        const children = []; // eslint-disable-line

        for (let i = 0; i < this.length; i += 1) {
            const childNodes = this[i].children;

            for (let j = 0; j < childNodes.length; j += 1) {
                if (!selector || $(childNodes[j]).is(selector)) {
                    children.push(childNodes[j]);
                }
            }
        }

        return $(children);
    }

    function remove() {
        for (let i = 0; i < this.length; i += 1) {
            if (this[i].parentNode) this[i].parentNode.removeChild(this[i]);
        }

        return this;
    }

    const Methods = {
        addClass,
        removeClass,
        hasClass,
        toggleClass,
        attr,
        removeAttr,
        transform,
        transition: transition$1,
        on,
        off,
        trigger,
        transitionEnd: transitionEnd$1,
        outerWidth,
        outerHeight,
        styles,
        offset,
        css,
        each,
        html,
        text,
        is,
        index,
        eq,
        append,
        prepend,
        next,
        nextAll,
        prev,
        prevAll,
        parent,
        parents,
        closest,
        find,
        children,
        filter,
        remove
    };
    Object.keys(Methods).forEach(methodName => {
        Object.defineProperty($.fn, methodName, {
            value: Methods[methodName],
            writable: true
        });
    });

    function deleteProps(obj) {
        const object = obj;
        Object.keys(object).forEach(key => {
            try {
                object[key] = null;
            } catch (e) {// no getter for object
            }

            try {
                delete object[key];
            } catch (e) {// something got wrong
            }
        });
    }

    function nextTick(callback, delay) {
        if (delay === void 0) {
            delay = 0;
        }

        return setTimeout(callback, delay);
    }

    function now() {
        return Date.now();
    }

    function getComputedStyle$1(el) {
        const window = getWindow();
        let style;

        if (window.getComputedStyle) {
            style = window.getComputedStyle(el, null);
        }

        if (!style && el.currentStyle) {
            style = el.currentStyle;
        }

        if (!style) {
            style = el.style;
        }

        return style;
    }

    function getTranslate(el, axis) {
        if (axis === void 0) {
            axis = 'x';
        }

        const window = getWindow();
        let matrix;
        let curTransform;
        let transformMatrix;
        const curStyle = getComputedStyle$1(el);

        if (window.WebKitCSSMatrix) {
            curTransform = curStyle.transform || curStyle.webkitTransform;

            if (curTransform.split(',').length > 6) {
                curTransform = curTransform.split(', ').map(a => a.replace(',', '.')).join(', ');
            } // Some old versions of Webkit choke when 'none' is passed; pass
            // empty string instead in this case


            transformMatrix = new window.WebKitCSSMatrix(curTransform === 'none' ? '' : curTransform);
        } else {
            transformMatrix = curStyle.MozTransform || curStyle.OTransform || curStyle.MsTransform || curStyle.msTransform || curStyle.transform || curStyle.getPropertyValue('transform').replace('translate(', 'matrix(1, 0, 0, 1,');
            matrix = transformMatrix.toString().split(',');
        }

        if (axis === 'x') {
            // Latest Chrome and webkits Fix
            if (window.WebKitCSSMatrix) curTransform = transformMatrix.m41; // Crazy IE10 Matrix
            else if (matrix.length === 16) curTransform = parseFloat(matrix[12]); // Normal Browsers
            else curTransform = parseFloat(matrix[4]);
        }

        if (axis === 'y') {
            // Latest Chrome and webkits Fix
            if (window.WebKitCSSMatrix) curTransform = transformMatrix.m42; // Crazy IE10 Matrix
            else if (matrix.length === 16) curTransform = parseFloat(matrix[13]); // Normal Browsers
            else curTransform = parseFloat(matrix[5]);
        }

        return curTransform || 0;
    }

    function isObject(o) {
        return typeof o === 'object' && o !== null && o.constructor && Object.prototype.toString.call(o).slice(8, -1) === 'Object';
    }

    function isNode(node) {
        // eslint-disable-next-line
        if (typeof window !== 'undefined' && typeof window.HTMLElement !== 'undefined') {
            return node instanceof HTMLElement;
        }

        return node && (node.nodeType === 1 || node.nodeType === 11);
    }

    function extend() {
        const to = Object(arguments.length <= 0 ? undefined : arguments[0]);
        const noExtend = ['__proto__', 'constructor', 'prototype'];

        for (let i = 1; i < arguments.length; i += 1) {
            const nextSource = i < 0 || arguments.length <= i ? undefined : arguments[i];

            if (nextSource !== undefined && nextSource !== null && !isNode(nextSource)) {
                const keysArray = Object.keys(Object(nextSource)).filter(key => noExtend.indexOf(key) < 0);

                for (let nextIndex = 0, len = keysArray.length; nextIndex < len; nextIndex += 1) {
                    const nextKey = keysArray[nextIndex];
                    const desc = Object.getOwnPropertyDescriptor(nextSource, nextKey);

                    if (desc !== undefined && desc.enumerable) {
                        if (isObject(to[nextKey]) && isObject(nextSource[nextKey])) {
                            if (nextSource[nextKey].__swiper__) {
                                to[nextKey] = nextSource[nextKey];
                            } else {
                                extend(to[nextKey], nextSource[nextKey]);
                            }
                        } else if (!isObject(to[nextKey]) && isObject(nextSource[nextKey])) {
                            to[nextKey] = {};

                            if (nextSource[nextKey].__swiper__) {
                                to[nextKey] = nextSource[nextKey];
                            } else {
                                extend(to[nextKey], nextSource[nextKey]);
                            }
                        } else {
                            to[nextKey] = nextSource[nextKey];
                        }
                    }
                }
            }
        }

        return to;
    }

    function setCSSProperty(el, varName, varValue) {
        el.style.setProperty(varName, varValue);
    }

    function animateCSSModeScroll(_ref) {
        let {
            swiper,
            targetPosition,
            side
        } = _ref;
        const window = getWindow();
        const startPosition = -swiper.translate;
        let startTime = null;
        let time;
        const duration = swiper.params.speed;
        swiper.wrapperEl.style.scrollSnapType = 'none';
        window.cancelAnimationFrame(swiper.cssModeFrameID);
        const dir = targetPosition > startPosition ? 'next' : 'prev';

        const isOutOfBound = (current, target) => {
            return dir === 'next' && current >= target || dir === 'prev' && current <= target;
        };

        const animate = () => {
            time = new Date().getTime();

            if (startTime === null) {
                startTime = time;
            }

            const progress = Math.max(Math.min((time - startTime) / duration, 1), 0);
            const easeProgress = 0.5 - Math.cos(progress * Math.PI) / 2;
            let currentPosition = startPosition + easeProgress * (targetPosition - startPosition);

            if (isOutOfBound(currentPosition, targetPosition)) {
                currentPosition = targetPosition;
            }

            swiper.wrapperEl.scrollTo({
                [side]: currentPosition
            });

            if (isOutOfBound(currentPosition, targetPosition)) {
                swiper.wrapperEl.style.overflow = 'hidden';
                swiper.wrapperEl.style.scrollSnapType = '';
                setTimeout(() => {
                    swiper.wrapperEl.style.overflow = '';
                    swiper.wrapperEl.scrollTo({
                        [side]: currentPosition
                    });
                });
                window.cancelAnimationFrame(swiper.cssModeFrameID);
                return;
            }

            swiper.cssModeFrameID = window.requestAnimationFrame(animate);
        };

        animate();
    }

    let support;

    function calcSupport() {
        const window = getWindow();
        const document = getDocument();
        return {
            smoothScroll: document.documentElement && 'scrollBehavior' in document.documentElement.style,
            touch: !!('ontouchstart' in window || window.DocumentTouch && document instanceof window.DocumentTouch),
            passiveListener: function checkPassiveListener() {
                let supportsPassive = false;

                try {
                    const opts = Object.defineProperty({}, 'passive', {
                        // eslint-disable-next-line
                        get() {
                            supportsPassive = true;
                        }

                    });
                    window.addEventListener('testPassiveListener', null, opts);
                } catch (e) {// No support
                }

                return supportsPassive;
            }(),
            gestures: function checkGestures() {
                return 'ongesturestart' in window;
            }()
        };
    }

    function getSupport() {
        if (!support) {
            support = calcSupport();
        }

        return support;
    }

    let deviceCached;

    function calcDevice(_temp) {
        let {
            userAgent
        } = _temp === void 0 ? {} : _temp;
        const support = getSupport();
        const window = getWindow();
        const platform = window.navigator.platform;
        const ua = userAgent || window.navigator.userAgent;
        const device = {
            ios: false,
            android: false
        };
        const screenWidth = window.screen.width;
        const screenHeight = window.screen.height;
        const android = ua.match(/(Android);?[\s\/]+([\d.]+)?/); // eslint-disable-line

        let ipad = ua.match(/(iPad).*OS\s([\d_]+)/);
        const ipod = ua.match(/(iPod)(.*OS\s([\d_]+))?/);
        const iphone = !ipad && ua.match(/(iPhone\sOS|iOS)\s([\d_]+)/);
        const windows = platform === 'Win32';
        let macos = platform === 'MacIntel'; // iPadOs 13 fix

        const iPadScreens = ['1024x1366', '1366x1024', '834x1194', '1194x834', '834x1112', '1112x834', '768x1024', '1024x768', '820x1180', '1180x820', '810x1080', '1080x810'];

        if (!ipad && macos && support.touch && iPadScreens.indexOf(`${screenWidth}x${screenHeight}`) >= 0) {
            ipad = ua.match(/(Version)\/([\d.]+)/);
            if (!ipad) ipad = [0, 1, '13_0_0'];
            macos = false;
        } // Android


        if (android && !windows) {
            device.os = 'android';
            device.android = true;
        }

        if (ipad || iphone || ipod) {
            device.os = 'ios';
            device.ios = true;
        } // Export object


        return device;
    }

    function getDevice(overrides) {
        if (overrides === void 0) {
            overrides = {};
        }

        if (!deviceCached) {
            deviceCached = calcDevice(overrides);
        }

        return deviceCached;
    }

    let browser;

    function calcBrowser() {
        const window = getWindow();

        function isSafari() {
            const ua = window.navigator.userAgent.toLowerCase();
            return ua.indexOf('safari') >= 0 && ua.indexOf('chrome') < 0 && ua.indexOf('android') < 0;
        }

        return {
            isSafari: isSafari(),
            isWebView: /(iPhone|iPod|iPad).*AppleWebKit(?!.*Safari)/i.test(window.navigator.userAgent)
        };
    }

    function getBrowser() {
        if (!browser) {
            browser = calcBrowser();
        }

        return browser;
    }

    function Resize(_ref) {
        let {
            swiper,
            on,
            emit
        } = _ref;
        const window = getWindow();
        let observer = null;
        let animationFrame = null;

        const resizeHandler = () => {
            if (!swiper || swiper.destroyed || !swiper.initialized) return;
            emit('beforeResize');
            emit('resize');
        };

        const createObserver = () => {
            if (!swiper || swiper.destroyed || !swiper.initialized) return;
            observer = new ResizeObserver(entries => {
                animationFrame = window.requestAnimationFrame(() => {
                    const {
                        width,
                        height
                    } = swiper;
                    let newWidth = width;
                    let newHeight = height;
                    entries.forEach(_ref2 => {
                        let {
                            contentBoxSize,
                            contentRect,
                            target
                        } = _ref2;
                        if (target && target !== swiper.el) return;
                        newWidth = contentRect ? contentRect.width : (contentBoxSize[0] || contentBoxSize).inlineSize;
                        newHeight = contentRect ? contentRect.height : (contentBoxSize[0] || contentBoxSize).blockSize;
                    });

                    if (newWidth !== width || newHeight !== height) {
                        resizeHandler();
                    }
                });
            });
            observer.observe(swiper.el);
        };

        const removeObserver = () => {
            if (animationFrame) {
                window.cancelAnimationFrame(animationFrame);
            }

            if (observer && observer.unobserve && swiper.el) {
                observer.unobserve(swiper.el);
                observer = null;
            }
        };

        const orientationChangeHandler = () => {
            if (!swiper || swiper.destroyed || !swiper.initialized) return;
            emit('orientationchange');
        };

        on('init', () => {
            if (swiper.params.resizeObserver && typeof window.ResizeObserver !== 'undefined') {
                createObserver();
                return;
            }

            window.addEventListener('resize', resizeHandler);
            window.addEventListener('orientationchange', orientationChangeHandler);
        });
        on('destroy', () => {
            removeObserver();
            window.removeEventListener('resize', resizeHandler);
            window.removeEventListener('orientationchange', orientationChangeHandler);
        });
    }

    function Observer(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        const observers = [];
        const window = getWindow();

        const attach = function (target, options) {
            if (options === void 0) {
                options = {};
            }

            const ObserverFunc = window.MutationObserver || window.WebkitMutationObserver;
            const observer = new ObserverFunc(mutations => {
                // The observerUpdate event should only be triggered
                // once despite the number of mutations.  Additional
                // triggers are redundant and are very costly
                if (mutations.length === 1) {
                    emit('observerUpdate', mutations[0]);
                    return;
                }

                const observerUpdate = function observerUpdate() {
                    emit('observerUpdate', mutations[0]);
                };

                if (window.requestAnimationFrame) {
                    window.requestAnimationFrame(observerUpdate);
                } else {
                    window.setTimeout(observerUpdate, 0);
                }
            });
            observer.observe(target, {
                attributes: typeof options.attributes === 'undefined' ? true : options.attributes,
                childList: typeof options.childList === 'undefined' ? true : options.childList,
                characterData: typeof options.characterData === 'undefined' ? true : options.characterData
            });
            observers.push(observer);
        };

        const init = () => {
            if (!swiper.params.observer) return;

            if (swiper.params.observeParents) {
                const containerParents = swiper.$el.parents();

                for (let i = 0; i < containerParents.length; i += 1) {
                    attach(containerParents[i]);
                }
            } // Observe container


            attach(swiper.$el[0], {
                childList: swiper.params.observeSlideChildren
            }); // Observe wrapper

            attach(swiper.$wrapperEl[0], {
                attributes: false
            });
        };

        const destroy = () => {
            observers.forEach(observer => {
                observer.disconnect();
            });
            observers.splice(0, observers.length);
        };

        extendParams({
            observer: false,
            observeParents: false,
            observeSlideChildren: false
        });
        on('init', init);
        on('destroy', destroy);
    }

    /* eslint-disable no-underscore-dangle */
    var eventsEmitter = {
        on(events, handler, priority) {
            const self = this;
            if (!self.eventsListeners || self.destroyed) return self;
            if (typeof handler !== 'function') return self;
            const method = priority ? 'unshift' : 'push';
            events.split(' ').forEach(event => {
                if (!self.eventsListeners[event]) self.eventsListeners[event] = [];
                self.eventsListeners[event][method](handler);
            });
            return self;
        },

        once(events, handler, priority) {
            const self = this;
            if (!self.eventsListeners || self.destroyed) return self;
            if (typeof handler !== 'function') return self;

            function onceHandler() {
                self.off(events, onceHandler);

                if (onceHandler.__emitterProxy) {
                    delete onceHandler.__emitterProxy;
                }

                for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
                    args[_key] = arguments[_key];
                }

                handler.apply(self, args);
            }

            onceHandler.__emitterProxy = handler;
            return self.on(events, onceHandler, priority);
        },

        onAny(handler, priority) {
            const self = this;
            if (!self.eventsListeners || self.destroyed) return self;
            if (typeof handler !== 'function') return self;
            const method = priority ? 'unshift' : 'push';

            if (self.eventsAnyListeners.indexOf(handler) < 0) {
                self.eventsAnyListeners[method](handler);
            }

            return self;
        },

        offAny(handler) {
            const self = this;
            if (!self.eventsListeners || self.destroyed) return self;
            if (!self.eventsAnyListeners) return self;
            const index = self.eventsAnyListeners.indexOf(handler);

            if (index >= 0) {
                self.eventsAnyListeners.splice(index, 1);
            }

            return self;
        },

        off(events, handler) {
            const self = this;
            if (!self.eventsListeners || self.destroyed) return self;
            if (!self.eventsListeners) return self;
            events.split(' ').forEach(event => {
                if (typeof handler === 'undefined') {
                    self.eventsListeners[event] = [];
                } else if (self.eventsListeners[event]) {
                    self.eventsListeners[event].forEach((eventHandler, index) => {
                        if (eventHandler === handler || eventHandler.__emitterProxy && eventHandler.__emitterProxy === handler) {
                            self.eventsListeners[event].splice(index, 1);
                        }
                    });
                }
            });
            return self;
        },

        emit() {
            const self = this;
            if (!self.eventsListeners || self.destroyed) return self;
            if (!self.eventsListeners) return self;
            let events;
            let data;
            let context;

            for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
                args[_key2] = arguments[_key2];
            }

            if (typeof args[0] === 'string' || Array.isArray(args[0])) {
                events = args[0];
                data = args.slice(1, args.length);
                context = self;
            } else {
                events = args[0].events;
                data = args[0].data;
                context = args[0].context || self;
            }

            data.unshift(context);
            const eventsArray = Array.isArray(events) ? events : events.split(' ');
            eventsArray.forEach(event => {
                if (self.eventsAnyListeners && self.eventsAnyListeners.length) {
                    self.eventsAnyListeners.forEach(eventHandler => {
                        eventHandler.apply(context, [event, ...data]);
                    });
                }

                if (self.eventsListeners && self.eventsListeners[event]) {
                    self.eventsListeners[event].forEach(eventHandler => {
                        eventHandler.apply(context, data);
                    });
                }
            });
            return self;
        }

    };

    function updateSize() {
        const swiper = this;
        let width;
        let height;
        const $el = swiper.$el;

        if (typeof swiper.params.width !== 'undefined' && swiper.params.width !== null) {
            width = swiper.params.width;
        } else {
            width = $el[0].clientWidth;
        }

        if (typeof swiper.params.height !== 'undefined' && swiper.params.height !== null) {
            height = swiper.params.height;
        } else {
            height = $el[0].clientHeight;
        }

        if (width === 0 && swiper.isHorizontal() || height === 0 && swiper.isVertical()) {
            return;
        } // Subtract paddings


        width = width - parseInt($el.css('padding-left') || 0, 10) - parseInt($el.css('padding-right') || 0, 10);
        height = height - parseInt($el.css('padding-top') || 0, 10) - parseInt($el.css('padding-bottom') || 0, 10);
        if (Number.isNaN(width)) width = 0;
        if (Number.isNaN(height)) height = 0;
        Object.assign(swiper, {
            width,
            height,
            size: swiper.isHorizontal() ? width : height
        });
    }

    function updateSlides() {
        const swiper = this;

        function getDirectionLabel(property) {
            if (swiper.isHorizontal()) {
                return property;
            } // prettier-ignore


            return {
                'width': 'height',
                'margin-top': 'margin-left',
                'margin-bottom ': 'margin-right',
                'margin-left': 'margin-top',
                'margin-right': 'margin-bottom',
                'padding-left': 'padding-top',
                'padding-right': 'padding-bottom',
                'marginRight': 'marginBottom'
            }[property];
        }

        function getDirectionPropertyValue(node, label) {
            return parseFloat(node.getPropertyValue(getDirectionLabel(label)) || 0);
        }

        const params = swiper.params;
        const {
            $wrapperEl,
            size: swiperSize,
            rtlTranslate: rtl,
            wrongRTL
        } = swiper;
        const isVirtual = swiper.virtual && params.virtual.enabled;
        const previousSlidesLength = isVirtual ? swiper.virtual.slides.length : swiper.slides.length;
        const slides = $wrapperEl.children(`.${swiper.params.slideClass}`);
        const slidesLength = isVirtual ? swiper.virtual.slides.length : slides.length;
        let snapGrid = [];
        const slidesGrid = [];
        const slidesSizesGrid = [];
        let offsetBefore = params.slidesOffsetBefore;

        if (typeof offsetBefore === 'function') {
            offsetBefore = params.slidesOffsetBefore.call(swiper);
        }

        let offsetAfter = params.slidesOffsetAfter;

        if (typeof offsetAfter === 'function') {
            offsetAfter = params.slidesOffsetAfter.call(swiper);
        }

        const previousSnapGridLength = swiper.snapGrid.length;
        const previousSlidesGridLength = swiper.slidesGrid.length;
        let spaceBetween = params.spaceBetween;
        let slidePosition = -offsetBefore;
        let prevSlideSize = 0;
        let index = 0;

        if (typeof swiperSize === 'undefined') {
            return;
        }

        if (typeof spaceBetween === 'string' && spaceBetween.indexOf('%') >= 0) {
            spaceBetween = parseFloat(spaceBetween.replace('%', '')) / 100 * swiperSize;
        }

        swiper.virtualSize = -spaceBetween; // reset margins

        if (rtl) slides.css({
            marginLeft: '',
            marginBottom: '',
            marginTop: ''
        }); else slides.css({
            marginRight: '',
            marginBottom: '',
            marginTop: ''
        }); // reset cssMode offsets

        if (params.centeredSlides && params.cssMode) {
            setCSSProperty(swiper.wrapperEl, '--swiper-centered-offset-before', '');
            setCSSProperty(swiper.wrapperEl, '--swiper-centered-offset-after', '');
        }

        const gridEnabled = params.grid && params.grid.rows > 1 && swiper.grid;

        if (gridEnabled) {
            swiper.grid.initSlides(slidesLength);
        } // Calc slides


        let slideSize;
        const shouldResetSlideSize = params.slidesPerView === 'auto' && params.breakpoints && Object.keys(params.breakpoints).filter(key => {
            return typeof params.breakpoints[key].slidesPerView !== 'undefined';
        }).length > 0;

        for (let i = 0; i < slidesLength; i += 1) {
            slideSize = 0;
            const slide = slides.eq(i);

            if (gridEnabled) {
                swiper.grid.updateSlide(i, slide, slidesLength, getDirectionLabel);
            }

            if (slide.css('display') === 'none') continue; // eslint-disable-line

            if (params.slidesPerView === 'auto') {
                if (shouldResetSlideSize) {
                    slides[i].style[getDirectionLabel('width')] = ``;
                }

                const slideStyles = getComputedStyle(slide[0]);
                const currentTransform = slide[0].style.transform;
                const currentWebKitTransform = slide[0].style.webkitTransform;

                if (currentTransform) {
                    slide[0].style.transform = 'none';
                }

                if (currentWebKitTransform) {
                    slide[0].style.webkitTransform = 'none';
                }

                if (params.roundLengths) {
                    slideSize = swiper.isHorizontal() ? slide.outerWidth(true) : slide.outerHeight(true);
                } else {
                    // eslint-disable-next-line
                    const width = getDirectionPropertyValue(slideStyles, 'width');
                    const paddingLeft = getDirectionPropertyValue(slideStyles, 'padding-left');
                    const paddingRight = getDirectionPropertyValue(slideStyles, 'padding-right');
                    const marginLeft = getDirectionPropertyValue(slideStyles, 'margin-left');
                    const marginRight = getDirectionPropertyValue(slideStyles, 'margin-right');
                    const boxSizing = slideStyles.getPropertyValue('box-sizing');

                    if (boxSizing && boxSizing === 'border-box') {
                        slideSize = width + marginLeft + marginRight;
                    } else {
                        const {
                            clientWidth,
                            offsetWidth
                        } = slide[0];
                        slideSize = width + paddingLeft + paddingRight + marginLeft + marginRight + (offsetWidth - clientWidth);
                    }
                }

                if (currentTransform) {
                    slide[0].style.transform = currentTransform;
                }

                if (currentWebKitTransform) {
                    slide[0].style.webkitTransform = currentWebKitTransform;
                }

                if (params.roundLengths) slideSize = Math.floor(slideSize);
            } else {
                slideSize = (swiperSize - (params.slidesPerView - 1) * spaceBetween) / params.slidesPerView;
                if (params.roundLengths) slideSize = Math.floor(slideSize);

                if (slides[i]) {
                    slides[i].style[getDirectionLabel('width')] = `${slideSize}px`;
                }
            }

            if (slides[i]) {
                slides[i].swiperSlideSize = slideSize;
            }

            slidesSizesGrid.push(slideSize);

            if (params.centeredSlides) {
                slidePosition = slidePosition + slideSize / 2 + prevSlideSize / 2 + spaceBetween;
                if (prevSlideSize === 0 && i !== 0) slidePosition = slidePosition - swiperSize / 2 - spaceBetween;
                if (i === 0) slidePosition = slidePosition - swiperSize / 2 - spaceBetween;
                if (Math.abs(slidePosition) < 1 / 1000) slidePosition = 0;
                if (params.roundLengths) slidePosition = Math.floor(slidePosition);
                if (index % params.slidesPerGroup === 0) snapGrid.push(slidePosition);
                slidesGrid.push(slidePosition);
            } else {
                if (params.roundLengths) slidePosition = Math.floor(slidePosition);
                if ((index - Math.min(swiper.params.slidesPerGroupSkip, index)) % swiper.params.slidesPerGroup === 0) snapGrid.push(slidePosition);
                slidesGrid.push(slidePosition);
                slidePosition = slidePosition + slideSize + spaceBetween;
            }

            swiper.virtualSize += slideSize + spaceBetween;
            prevSlideSize = slideSize;
            index += 1;
        }

        swiper.virtualSize = Math.max(swiper.virtualSize, swiperSize) + offsetAfter;

        if (rtl && wrongRTL && (params.effect === 'slide' || params.effect === 'coverflow')) {
            $wrapperEl.css({
                width: `${swiper.virtualSize + params.spaceBetween}px`
            });
        }

        if (params.setWrapperSize) {
            $wrapperEl.css({
                [getDirectionLabel('width')]: `${swiper.virtualSize + params.spaceBetween}px`
            });
        }

        if (gridEnabled) {
            swiper.grid.updateWrapperSize(slideSize, snapGrid, getDirectionLabel);
        } // Remove last grid elements depending on width


        if (!params.centeredSlides) {
            const newSlidesGrid = [];

            for (let i = 0; i < snapGrid.length; i += 1) {
                let slidesGridItem = snapGrid[i];
                if (params.roundLengths) slidesGridItem = Math.floor(slidesGridItem);

                if (snapGrid[i] <= swiper.virtualSize - swiperSize) {
                    newSlidesGrid.push(slidesGridItem);
                }
            }

            snapGrid = newSlidesGrid;

            if (Math.floor(swiper.virtualSize - swiperSize) - Math.floor(snapGrid[snapGrid.length - 1]) > 1) {
                snapGrid.push(swiper.virtualSize - swiperSize);
            }
        }

        if (snapGrid.length === 0) snapGrid = [0];

        if (params.spaceBetween !== 0) {
            const key = swiper.isHorizontal() && rtl ? 'marginLeft' : getDirectionLabel('marginRight');
            slides.filter((_, slideIndex) => {
                if (!params.cssMode) return true;

                if (slideIndex === slides.length - 1) {
                    return false;
                }

                return true;
            }).css({
                [key]: `${spaceBetween}px`
            });
        }

        if (params.centeredSlides && params.centeredSlidesBounds) {
            let allSlidesSize = 0;
            slidesSizesGrid.forEach(slideSizeValue => {
                allSlidesSize += slideSizeValue + (params.spaceBetween ? params.spaceBetween : 0);
            });
            allSlidesSize -= params.spaceBetween;
            const maxSnap = allSlidesSize - swiperSize;
            snapGrid = snapGrid.map(snap => {
                if (snap < 0) return -offsetBefore;
                if (snap > maxSnap) return maxSnap + offsetAfter;
                return snap;
            });
        }

        if (params.centerInsufficientSlides) {
            let allSlidesSize = 0;
            slidesSizesGrid.forEach(slideSizeValue => {
                allSlidesSize += slideSizeValue + (params.spaceBetween ? params.spaceBetween : 0);
            });
            allSlidesSize -= params.spaceBetween;

            if (allSlidesSize < swiperSize) {
                const allSlidesOffset = (swiperSize - allSlidesSize) / 2;
                snapGrid.forEach((snap, snapIndex) => {
                    snapGrid[snapIndex] = snap - allSlidesOffset;
                });
                slidesGrid.forEach((snap, snapIndex) => {
                    slidesGrid[snapIndex] = snap + allSlidesOffset;
                });
            }
        }

        Object.assign(swiper, {
            slides,
            snapGrid,
            slidesGrid,
            slidesSizesGrid
        });

        if (params.centeredSlides && params.cssMode && !params.centeredSlidesBounds) {
            setCSSProperty(swiper.wrapperEl, '--swiper-centered-offset-before', `${-snapGrid[0]}px`);
            setCSSProperty(swiper.wrapperEl, '--swiper-centered-offset-after', `${swiper.size / 2 - slidesSizesGrid[slidesSizesGrid.length - 1] / 2}px`);
            const addToSnapGrid = -swiper.snapGrid[0];
            const addToSlidesGrid = -swiper.slidesGrid[0];
            swiper.snapGrid = swiper.snapGrid.map(v => v + addToSnapGrid);
            swiper.slidesGrid = swiper.slidesGrid.map(v => v + addToSlidesGrid);
        }

        if (slidesLength !== previousSlidesLength) {
            swiper.emit('slidesLengthChange');
        }

        if (snapGrid.length !== previousSnapGridLength) {
            if (swiper.params.watchOverflow) swiper.checkOverflow();
            swiper.emit('snapGridLengthChange');
        }

        if (slidesGrid.length !== previousSlidesGridLength) {
            swiper.emit('slidesGridLengthChange');
        }

        if (params.watchSlidesProgress) {
            swiper.updateSlidesOffset();
        }

        if (!isVirtual && !params.cssMode && (params.effect === 'slide' || params.effect === 'fade')) {
            const backFaceHiddenClass = `${params.containerModifierClass}backface-hidden`;
            const hasClassBackfaceClassAdded = swiper.$el.hasClass(backFaceHiddenClass);

            if (slidesLength <= params.maxBackfaceHiddenSlides) {
                if (!hasClassBackfaceClassAdded) swiper.$el.addClass(backFaceHiddenClass);
            } else if (hasClassBackfaceClassAdded) {
                swiper.$el.removeClass(backFaceHiddenClass);
            }
        }
    }

    function updateAutoHeight(speed) {
        const swiper = this;
        const activeSlides = [];
        const isVirtual = swiper.virtual && swiper.params.virtual.enabled;
        let newHeight = 0;
        let i;

        if (typeof speed === 'number') {
            swiper.setTransition(speed);
        } else if (speed === true) {
            swiper.setTransition(swiper.params.speed);
        }

        const getSlideByIndex = index => {
            if (isVirtual) {
                return swiper.slides.filter(el => parseInt(el.getAttribute('data-swiper-slide-index'), 10) === index)[0];
            }

            return swiper.slides.eq(index)[0];
        }; // Find slides currently in view


        if (swiper.params.slidesPerView !== 'auto' && swiper.params.slidesPerView > 1) {
            if (swiper.params.centeredSlides) {
                swiper.visibleSlides.each(slide => {
                    activeSlides.push(slide);
                });
            } else {
                for (i = 0; i < Math.ceil(swiper.params.slidesPerView); i += 1) {
                    const index = swiper.activeIndex + i;
                    if (index > swiper.slides.length && !isVirtual) break;
                    activeSlides.push(getSlideByIndex(index));
                }
            }
        } else {
            activeSlides.push(getSlideByIndex(swiper.activeIndex));
        } // Find new height from highest slide in view


        for (i = 0; i < activeSlides.length; i += 1) {
            if (typeof activeSlides[i] !== 'undefined') {
                const height = activeSlides[i].offsetHeight;
                newHeight = height > newHeight ? height : newHeight;
            }
        } // Update Height


        if (newHeight || newHeight === 0) swiper.$wrapperEl.css('height', `${newHeight}px`);
    }

    function updateSlidesOffset() {
        const swiper = this;
        const slides = swiper.slides;

        for (let i = 0; i < slides.length; i += 1) {
            slides[i].swiperSlideOffset = swiper.isHorizontal() ? slides[i].offsetLeft : slides[i].offsetTop;
        }
    }

    function updateSlidesProgress(translate) {
        if (translate === void 0) {
            translate = this && this.translate || 0;
        }

        const swiper = this;
        const params = swiper.params;
        const {
            slides,
            rtlTranslate: rtl,
            snapGrid
        } = swiper;
        if (slides.length === 0) return;
        if (typeof slides[0].swiperSlideOffset === 'undefined') swiper.updateSlidesOffset();
        let offsetCenter = -translate;
        if (rtl) offsetCenter = translate; // Visible Slides

        slides.removeClass(params.slideVisibleClass);
        swiper.visibleSlidesIndexes = [];
        swiper.visibleSlides = [];

        for (let i = 0; i < slides.length; i += 1) {
            const slide = slides[i];
            let slideOffset = slide.swiperSlideOffset;

            if (params.cssMode && params.centeredSlides) {
                slideOffset -= slides[0].swiperSlideOffset;
            }

            const slideProgress = (offsetCenter + (params.centeredSlides ? swiper.minTranslate() : 0) - slideOffset) / (slide.swiperSlideSize + params.spaceBetween);
            const originalSlideProgress = (offsetCenter - snapGrid[0] + (params.centeredSlides ? swiper.minTranslate() : 0) - slideOffset) / (slide.swiperSlideSize + params.spaceBetween);
            const slideBefore = -(offsetCenter - slideOffset);
            const slideAfter = slideBefore + swiper.slidesSizesGrid[i];
            const isVisible = slideBefore >= 0 && slideBefore < swiper.size - 1 || slideAfter > 1 && slideAfter <= swiper.size || slideBefore <= 0 && slideAfter >= swiper.size;

            if (isVisible) {
                swiper.visibleSlides.push(slide);
                swiper.visibleSlidesIndexes.push(i);
                slides.eq(i).addClass(params.slideVisibleClass);
            }

            slide.progress = rtl ? -slideProgress : slideProgress;
            slide.originalProgress = rtl ? -originalSlideProgress : originalSlideProgress;
        }

        swiper.visibleSlides = $(swiper.visibleSlides);
    }

    function updateProgress(translate) {
        const swiper = this;

        if (typeof translate === 'undefined') {
            const multiplier = swiper.rtlTranslate ? -1 : 1; // eslint-disable-next-line

            translate = swiper && swiper.translate && swiper.translate * multiplier || 0;
        }

        const params = swiper.params;
        const translatesDiff = swiper.maxTranslate() - swiper.minTranslate();
        let {
            progress,
            isBeginning,
            isEnd
        } = swiper;
        const wasBeginning = isBeginning;
        const wasEnd = isEnd;

        if (translatesDiff === 0) {
            progress = 0;
            isBeginning = true;
            isEnd = true;
        } else {
            progress = (translate - swiper.minTranslate()) / translatesDiff;
            isBeginning = progress <= 0;
            isEnd = progress >= 1;
        }

        Object.assign(swiper, {
            progress,
            isBeginning,
            isEnd
        });
        if (params.watchSlidesProgress || params.centeredSlides && params.autoHeight) swiper.updateSlidesProgress(translate);

        if (isBeginning && !wasBeginning) {
            swiper.emit('reachBeginning toEdge');
        }

        if (isEnd && !wasEnd) {
            swiper.emit('reachEnd toEdge');
        }

        if (wasBeginning && !isBeginning || wasEnd && !isEnd) {
            swiper.emit('fromEdge');
        }

        swiper.emit('progress', progress);
    }

    function updateSlidesClasses() {
        const swiper = this;
        const {
            slides,
            params,
            $wrapperEl,
            activeIndex,
            realIndex
        } = swiper;
        const isVirtual = swiper.virtual && params.virtual.enabled;
        slides.removeClass(`${params.slideActiveClass} ${params.slideNextClass} ${params.slidePrevClass} ${params.slideDuplicateActiveClass} ${params.slideDuplicateNextClass} ${params.slideDuplicatePrevClass}`);
        let activeSlide;

        if (isVirtual) {
            activeSlide = swiper.$wrapperEl.find(`.${params.slideClass}[data-swiper-slide-index="${activeIndex}"]`);
        } else {
            activeSlide = slides.eq(activeIndex);
        } // Active classes


        activeSlide.addClass(params.slideActiveClass);

        if (params.loop) {
            // Duplicate to all looped slides
            if (activeSlide.hasClass(params.slideDuplicateClass)) {
                $wrapperEl.children(`.${params.slideClass}:not(.${params.slideDuplicateClass})[data-swiper-slide-index="${realIndex}"]`).addClass(params.slideDuplicateActiveClass);
            } else {
                $wrapperEl.children(`.${params.slideClass}.${params.slideDuplicateClass}[data-swiper-slide-index="${realIndex}"]`).addClass(params.slideDuplicateActiveClass);
            }
        } // Next Slide


        let nextSlide = activeSlide.nextAll(`.${params.slideClass}`).eq(0).addClass(params.slideNextClass);

        if (params.loop && nextSlide.length === 0) {
            nextSlide = slides.eq(0);
            nextSlide.addClass(params.slideNextClass);
        } // Prev Slide


        let prevSlide = activeSlide.prevAll(`.${params.slideClass}`).eq(0).addClass(params.slidePrevClass);

        if (params.loop && prevSlide.length === 0) {
            prevSlide = slides.eq(-1);
            prevSlide.addClass(params.slidePrevClass);
        }

        if (params.loop) {
            // Duplicate to all looped slides
            if (nextSlide.hasClass(params.slideDuplicateClass)) {
                $wrapperEl.children(`.${params.slideClass}:not(.${params.slideDuplicateClass})[data-swiper-slide-index="${nextSlide.attr('data-swiper-slide-index')}"]`).addClass(params.slideDuplicateNextClass);
            } else {
                $wrapperEl.children(`.${params.slideClass}.${params.slideDuplicateClass}[data-swiper-slide-index="${nextSlide.attr('data-swiper-slide-index')}"]`).addClass(params.slideDuplicateNextClass);
            }

            if (prevSlide.hasClass(params.slideDuplicateClass)) {
                $wrapperEl.children(`.${params.slideClass}:not(.${params.slideDuplicateClass})[data-swiper-slide-index="${prevSlide.attr('data-swiper-slide-index')}"]`).addClass(params.slideDuplicatePrevClass);
            } else {
                $wrapperEl.children(`.${params.slideClass}.${params.slideDuplicateClass}[data-swiper-slide-index="${prevSlide.attr('data-swiper-slide-index')}"]`).addClass(params.slideDuplicatePrevClass);
            }
        }

        swiper.emitSlidesClasses();
    }

    function updateActiveIndex(newActiveIndex) {
        const swiper = this;
        const translate = swiper.rtlTranslate ? swiper.translate : -swiper.translate;
        const {
            slidesGrid,
            snapGrid,
            params,
            activeIndex: previousIndex,
            realIndex: previousRealIndex,
            snapIndex: previousSnapIndex
        } = swiper;
        let activeIndex = newActiveIndex;
        let snapIndex;

        if (typeof activeIndex === 'undefined') {
            for (let i = 0; i < slidesGrid.length; i += 1) {
                if (typeof slidesGrid[i + 1] !== 'undefined') {
                    if (translate >= slidesGrid[i] && translate < slidesGrid[i + 1] - (slidesGrid[i + 1] - slidesGrid[i]) / 2) {
                        activeIndex = i;
                    } else if (translate >= slidesGrid[i] && translate < slidesGrid[i + 1]) {
                        activeIndex = i + 1;
                    }
                } else if (translate >= slidesGrid[i]) {
                    activeIndex = i;
                }
            } // Normalize slideIndex


            if (params.normalizeSlideIndex) {
                if (activeIndex < 0 || typeof activeIndex === 'undefined') activeIndex = 0;
            }
        }

        if (snapGrid.indexOf(translate) >= 0) {
            snapIndex = snapGrid.indexOf(translate);
        } else {
            const skip = Math.min(params.slidesPerGroupSkip, activeIndex);
            snapIndex = skip + Math.floor((activeIndex - skip) / params.slidesPerGroup);
        }

        if (snapIndex >= snapGrid.length) snapIndex = snapGrid.length - 1;

        if (activeIndex === previousIndex) {
            if (snapIndex !== previousSnapIndex) {
                swiper.snapIndex = snapIndex;
                swiper.emit('snapIndexChange');
            }

            return;
        } // Get real index


        const realIndex = parseInt(swiper.slides.eq(activeIndex).attr('data-swiper-slide-index') || activeIndex, 10);
        Object.assign(swiper, {
            snapIndex,
            realIndex,
            previousIndex,
            activeIndex
        });
        swiper.emit('activeIndexChange');
        swiper.emit('snapIndexChange');

        if (previousRealIndex !== realIndex) {
            swiper.emit('realIndexChange');
        }

        if (swiper.initialized || swiper.params.runCallbacksOnInit) {
            swiper.emit('slideChange');
        }
    }

    function updateClickedSlide(e) {
        const swiper = this;
        const params = swiper.params;
        const slide = $(e).closest(`.${params.slideClass}`)[0];
        let slideFound = false;
        let slideIndex;

        if (slide) {
            for (let i = 0; i < swiper.slides.length; i += 1) {
                if (swiper.slides[i] === slide) {
                    slideFound = true;
                    slideIndex = i;
                    break;
                }
            }
        }

        if (slide && slideFound) {
            swiper.clickedSlide = slide;

            if (swiper.virtual && swiper.params.virtual.enabled) {
                swiper.clickedIndex = parseInt($(slide).attr('data-swiper-slide-index'), 10);
            } else {
                swiper.clickedIndex = slideIndex;
            }
        } else {
            swiper.clickedSlide = undefined;
            swiper.clickedIndex = undefined;
            return;
        }

        if (params.slideToClickedSlide && swiper.clickedIndex !== undefined && swiper.clickedIndex !== swiper.activeIndex) {
            swiper.slideToClickedSlide();
        }
    }

    var update = {
        updateSize,
        updateSlides,
        updateAutoHeight,
        updateSlidesOffset,
        updateSlidesProgress,
        updateProgress,
        updateSlidesClasses,
        updateActiveIndex,
        updateClickedSlide
    };

    function getSwiperTranslate(axis) {
        if (axis === void 0) {
            axis = this.isHorizontal() ? 'x' : 'y';
        }

        const swiper = this;
        const {
            params,
            rtlTranslate: rtl,
            translate,
            $wrapperEl
        } = swiper;

        if (params.virtualTranslate) {
            return rtl ? -translate : translate;
        }

        if (params.cssMode) {
            return translate;
        }

        let currentTranslate = getTranslate($wrapperEl[0], axis);
        if (rtl) currentTranslate = -currentTranslate;
        return currentTranslate || 0;
    }

    function setTranslate(translate, byController) {
        const swiper = this;
        const {
            rtlTranslate: rtl,
            params,
            $wrapperEl,
            wrapperEl,
            progress
        } = swiper;
        let x = 0;
        let y = 0;
        const z = 0;

        if (swiper.isHorizontal()) {
            x = rtl ? -translate : translate;
        } else {
            y = translate;
        }

        if (params.roundLengths) {
            x = Math.floor(x);
            y = Math.floor(y);
        }

        if (params.cssMode) {
            wrapperEl[swiper.isHorizontal() ? 'scrollLeft' : 'scrollTop'] = swiper.isHorizontal() ? -x : -y;
        } else if (!params.virtualTranslate) {
            $wrapperEl.transform(`translate3d(${x}px, ${y}px, ${z}px)`);
        }

        swiper.previousTranslate = swiper.translate;
        swiper.translate = swiper.isHorizontal() ? x : y; // Check if we need to update progress

        let newProgress;
        const translatesDiff = swiper.maxTranslate() - swiper.minTranslate();

        if (translatesDiff === 0) {
            newProgress = 0;
        } else {
            newProgress = (translate - swiper.minTranslate()) / translatesDiff;
        }

        if (newProgress !== progress) {
            swiper.updateProgress(translate);
        }

        swiper.emit('setTranslate', swiper.translate, byController);
    }

    function minTranslate() {
        return -this.snapGrid[0];
    }

    function maxTranslate() {
        return -this.snapGrid[this.snapGrid.length - 1];
    }

    function translateTo(translate, speed, runCallbacks, translateBounds, internal) {
        if (translate === void 0) {
            translate = 0;
        }

        if (speed === void 0) {
            speed = this.params.speed;
        }

        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        if (translateBounds === void 0) {
            translateBounds = true;
        }

        const swiper = this;
        const {
            params,
            wrapperEl
        } = swiper;

        if (swiper.animating && params.preventInteractionOnTransition) {
            return false;
        }

        const minTranslate = swiper.minTranslate();
        const maxTranslate = swiper.maxTranslate();
        let newTranslate;
        if (translateBounds && translate > minTranslate) newTranslate = minTranslate; else if (translateBounds && translate < maxTranslate) newTranslate = maxTranslate; else newTranslate = translate; // Update progress

        swiper.updateProgress(newTranslate);

        if (params.cssMode) {
            const isH = swiper.isHorizontal();

            if (speed === 0) {
                wrapperEl[isH ? 'scrollLeft' : 'scrollTop'] = -newTranslate;
            } else {
                if (!swiper.support.smoothScroll) {
                    animateCSSModeScroll({
                        swiper,
                        targetPosition: -newTranslate,
                        side: isH ? 'left' : 'top'
                    });
                    return true;
                }

                wrapperEl.scrollTo({
                    [isH ? 'left' : 'top']: -newTranslate,
                    behavior: 'smooth'
                });
            }

            return true;
        }

        if (speed === 0) {
            swiper.setTransition(0);
            swiper.setTranslate(newTranslate);

            if (runCallbacks) {
                swiper.emit('beforeTransitionStart', speed, internal);
                swiper.emit('transitionEnd');
            }
        } else {
            swiper.setTransition(speed);
            swiper.setTranslate(newTranslate);

            if (runCallbacks) {
                swiper.emit('beforeTransitionStart', speed, internal);
                swiper.emit('transitionStart');
            }

            if (!swiper.animating) {
                swiper.animating = true;

                if (!swiper.onTranslateToWrapperTransitionEnd) {
                    swiper.onTranslateToWrapperTransitionEnd = function transitionEnd(e) {
                        if (!swiper || swiper.destroyed) return;
                        if (e.target !== this) return;
                        swiper.$wrapperEl[0].removeEventListener('transitionend', swiper.onTranslateToWrapperTransitionEnd);
                        swiper.$wrapperEl[0].removeEventListener('webkitTransitionEnd', swiper.onTranslateToWrapperTransitionEnd);
                        swiper.onTranslateToWrapperTransitionEnd = null;
                        delete swiper.onTranslateToWrapperTransitionEnd;

                        if (runCallbacks) {
                            swiper.emit('transitionEnd');
                        }
                    };
                }

                swiper.$wrapperEl[0].addEventListener('transitionend', swiper.onTranslateToWrapperTransitionEnd);
                swiper.$wrapperEl[0].addEventListener('webkitTransitionEnd', swiper.onTranslateToWrapperTransitionEnd);
            }
        }

        return true;
    }

    var translate = {
        getTranslate: getSwiperTranslate,
        setTranslate,
        minTranslate,
        maxTranslate,
        translateTo
    };

    function setTransition(duration, byController) {
        const swiper = this;

        if (!swiper.params.cssMode) {
            swiper.$wrapperEl.transition(duration);
        }

        swiper.emit('setTransition', duration, byController);
    }

    function transitionEmit(_ref) {
        let {
            swiper,
            runCallbacks,
            direction,
            step
        } = _ref;
        const {
            activeIndex,
            previousIndex
        } = swiper;
        let dir = direction;

        if (!dir) {
            if (activeIndex > previousIndex) dir = 'next'; else if (activeIndex < previousIndex) dir = 'prev'; else dir = 'reset';
        }

        swiper.emit(`transition${step}`);

        if (runCallbacks && activeIndex !== previousIndex) {
            if (dir === 'reset') {
                swiper.emit(`slideResetTransition${step}`);
                return;
            }

            swiper.emit(`slideChangeTransition${step}`);

            if (dir === 'next') {
                swiper.emit(`slideNextTransition${step}`);
            } else {
                swiper.emit(`slidePrevTransition${step}`);
            }
        }
    }

    function transitionStart(runCallbacks, direction) {
        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        const swiper = this;
        const {
            params
        } = swiper;
        if (params.cssMode) return;

        if (params.autoHeight) {
            swiper.updateAutoHeight();
        }

        transitionEmit({
            swiper,
            runCallbacks,
            direction,
            step: 'Start'
        });
    }

    function transitionEnd(runCallbacks, direction) {
        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        const swiper = this;
        const {
            params
        } = swiper;
        swiper.animating = false;
        if (params.cssMode) return;
        swiper.setTransition(0);
        transitionEmit({
            swiper,
            runCallbacks,
            direction,
            step: 'End'
        });
    }

    var transition = {
        setTransition,
        transitionStart,
        transitionEnd
    };

    function slideTo(index, speed, runCallbacks, internal, initial) {
        if (index === void 0) {
            index = 0;
        }

        if (speed === void 0) {
            speed = this.params.speed;
        }

        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        if (typeof index !== 'number' && typeof index !== 'string') {
            throw new Error(`The 'index' argument cannot have type other than 'number' or 'string'. [${typeof index}] given.`);
        }

        if (typeof index === 'string') {
            /**
             * The `index` argument converted from `string` to `number`.
             * @type {number}
             */
            const indexAsNumber = parseInt(index, 10);
            /**
             * Determines whether the `index` argument is a valid `number`
             * after being converted from the `string` type.
             * @type {boolean}
             */

            const isValidNumber = isFinite(indexAsNumber);

            if (!isValidNumber) {
                throw new Error(`The passed-in 'index' (string) couldn't be converted to 'number'. [${index}] given.`);
            } // Knowing that the converted `index` is a valid number,
            // we can update the original argument's value.


            index = indexAsNumber;
        }

        const swiper = this;
        let slideIndex = index;
        if (slideIndex < 0) slideIndex = 0;
        const {
            params,
            snapGrid,
            slidesGrid,
            previousIndex,
            activeIndex,
            rtlTranslate: rtl,
            wrapperEl,
            enabled
        } = swiper;

        if (swiper.animating && params.preventInteractionOnTransition || !enabled && !internal && !initial) {
            return false;
        }

        const skip = Math.min(swiper.params.slidesPerGroupSkip, slideIndex);
        let snapIndex = skip + Math.floor((slideIndex - skip) / swiper.params.slidesPerGroup);
        if (snapIndex >= snapGrid.length) snapIndex = snapGrid.length - 1;

        if ((activeIndex || params.initialSlide || 0) === (previousIndex || 0) && runCallbacks) {
            swiper.emit('beforeSlideChangeStart');
        }

        const translate = -snapGrid[snapIndex]; // Update progress

        swiper.updateProgress(translate); // Normalize slideIndex

        if (params.normalizeSlideIndex) {
            for (let i = 0; i < slidesGrid.length; i += 1) {
                const normalizedTranslate = -Math.floor(translate * 100);
                const normalizedGrid = Math.floor(slidesGrid[i] * 100);
                const normalizedGridNext = Math.floor(slidesGrid[i + 1] * 100);

                if (typeof slidesGrid[i + 1] !== 'undefined') {
                    if (normalizedTranslate >= normalizedGrid && normalizedTranslate < normalizedGridNext - (normalizedGridNext - normalizedGrid) / 2) {
                        slideIndex = i;
                    } else if (normalizedTranslate >= normalizedGrid && normalizedTranslate < normalizedGridNext) {
                        slideIndex = i + 1;
                    }
                } else if (normalizedTranslate >= normalizedGrid) {
                    slideIndex = i;
                }
            }
        } // Directions locks


        if (swiper.initialized && slideIndex !== activeIndex) {
            if (!swiper.allowSlideNext && translate < swiper.translate && translate < swiper.minTranslate()) {
                return false;
            }

            if (!swiper.allowSlidePrev && translate > swiper.translate && translate > swiper.maxTranslate()) {
                if ((activeIndex || 0) !== slideIndex) return false;
            }
        }

        let direction;
        if (slideIndex > activeIndex) direction = 'next'; else if (slideIndex < activeIndex) direction = 'prev'; else direction = 'reset'; // Update Index

        if (rtl && -translate === swiper.translate || !rtl && translate === swiper.translate) {
            swiper.updateActiveIndex(slideIndex); // Update Height

            if (params.autoHeight) {
                swiper.updateAutoHeight();
            }

            swiper.updateSlidesClasses();

            if (params.effect !== 'slide') {
                swiper.setTranslate(translate);
            }

            if (direction !== 'reset') {
                swiper.transitionStart(runCallbacks, direction);
                swiper.transitionEnd(runCallbacks, direction);
            }

            return false;
        }

        if (params.cssMode) {
            const isH = swiper.isHorizontal();
            const t = rtl ? translate : -translate;

            if (speed === 0) {
                const isVirtual = swiper.virtual && swiper.params.virtual.enabled;

                if (isVirtual) {
                    swiper.wrapperEl.style.scrollSnapType = 'none';
                    swiper._immediateVirtual = true;
                }

                wrapperEl[isH ? 'scrollLeft' : 'scrollTop'] = t;

                if (isVirtual) {
                    requestAnimationFrame(() => {
                        swiper.wrapperEl.style.scrollSnapType = '';
                        swiper._swiperImmediateVirtual = false;
                    });
                }
            } else {
                if (!swiper.support.smoothScroll) {
                    animateCSSModeScroll({
                        swiper,
                        targetPosition: t,
                        side: isH ? 'left' : 'top'
                    });
                    return true;
                }

                wrapperEl.scrollTo({
                    [isH ? 'left' : 'top']: t,
                    behavior: 'smooth'
                });
            }

            return true;
        }

        swiper.setTransition(speed);
        swiper.setTranslate(translate);
        swiper.updateActiveIndex(slideIndex);
        swiper.updateSlidesClasses();
        swiper.emit('beforeTransitionStart', speed, internal);
        swiper.transitionStart(runCallbacks, direction);

        if (speed === 0) {
            swiper.transitionEnd(runCallbacks, direction);
        } else if (!swiper.animating) {
            swiper.animating = true;

            if (!swiper.onSlideToWrapperTransitionEnd) {
                swiper.onSlideToWrapperTransitionEnd = function transitionEnd(e) {
                    if (!swiper || swiper.destroyed) return;
                    if (e.target !== this) return;
                    swiper.$wrapperEl[0].removeEventListener('transitionend', swiper.onSlideToWrapperTransitionEnd);
                    swiper.$wrapperEl[0].removeEventListener('webkitTransitionEnd', swiper.onSlideToWrapperTransitionEnd);
                    swiper.onSlideToWrapperTransitionEnd = null;
                    delete swiper.onSlideToWrapperTransitionEnd;
                    swiper.transitionEnd(runCallbacks, direction);
                };
            }

            swiper.$wrapperEl[0].addEventListener('transitionend', swiper.onSlideToWrapperTransitionEnd);
            swiper.$wrapperEl[0].addEventListener('webkitTransitionEnd', swiper.onSlideToWrapperTransitionEnd);
        }

        return true;
    }

    function slideToLoop(index, speed, runCallbacks, internal) {
        if (index === void 0) {
            index = 0;
        }

        if (speed === void 0) {
            speed = this.params.speed;
        }

        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        const swiper = this;
        let newIndex = index;

        if (swiper.params.loop) {
            newIndex += swiper.loopedSlides;
        }

        return swiper.slideTo(newIndex, speed, runCallbacks, internal);
    }

    /* eslint no-unused-vars: "off" */
    function slideNext(speed, runCallbacks, internal) {
        if (speed === void 0) {
            speed = this.params.speed;
        }

        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        const swiper = this;
        const {
            animating,
            enabled,
            params
        } = swiper;
        if (!enabled) return swiper;
        let perGroup = params.slidesPerGroup;

        if (params.slidesPerView === 'auto' && params.slidesPerGroup === 1 && params.slidesPerGroupAuto) {
            perGroup = Math.max(swiper.slidesPerViewDynamic('current', true), 1);
        }

        const increment = swiper.activeIndex < params.slidesPerGroupSkip ? 1 : perGroup;

        if (params.loop) {
            if (animating && params.loopPreventsSlide) return false;
            swiper.loopFix(); // eslint-disable-next-line

            swiper._clientLeft = swiper.$wrapperEl[0].clientLeft;
        }

        if (params.rewind && swiper.isEnd) {
            return swiper.slideTo(0, speed, runCallbacks, internal);
        }

        return swiper.slideTo(swiper.activeIndex + increment, speed, runCallbacks, internal);
    }

    /* eslint no-unused-vars: "off" */
    function slidePrev(speed, runCallbacks, internal) {
        if (speed === void 0) {
            speed = this.params.speed;
        }

        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        const swiper = this;
        const {
            params,
            animating,
            snapGrid,
            slidesGrid,
            rtlTranslate,
            enabled
        } = swiper;
        if (!enabled) return swiper;

        if (params.loop) {
            if (animating && params.loopPreventsSlide) return false;
            swiper.loopFix(); // eslint-disable-next-line

            swiper._clientLeft = swiper.$wrapperEl[0].clientLeft;
        }

        const translate = rtlTranslate ? swiper.translate : -swiper.translate;

        function normalize(val) {
            if (val < 0) return -Math.floor(Math.abs(val));
            return Math.floor(val);
        }

        const normalizedTranslate = normalize(translate);
        const normalizedSnapGrid = snapGrid.map(val => normalize(val));
        let prevSnap = snapGrid[normalizedSnapGrid.indexOf(normalizedTranslate) - 1];

        if (typeof prevSnap === 'undefined' && params.cssMode) {
            let prevSnapIndex;
            snapGrid.forEach((snap, snapIndex) => {
                if (normalizedTranslate >= snap) {
                    // prevSnap = snap;
                    prevSnapIndex = snapIndex;
                }
            });

            if (typeof prevSnapIndex !== 'undefined') {
                prevSnap = snapGrid[prevSnapIndex > 0 ? prevSnapIndex - 1 : prevSnapIndex];
            }
        }

        let prevIndex = 0;

        if (typeof prevSnap !== 'undefined') {
            prevIndex = slidesGrid.indexOf(prevSnap);
            if (prevIndex < 0) prevIndex = swiper.activeIndex - 1;

            if (params.slidesPerView === 'auto' && params.slidesPerGroup === 1 && params.slidesPerGroupAuto) {
                prevIndex = prevIndex - swiper.slidesPerViewDynamic('previous', true) + 1;
                prevIndex = Math.max(prevIndex, 0);
            }
        }

        if (params.rewind && swiper.isBeginning) {
            const lastIndex = swiper.params.virtual && swiper.params.virtual.enabled && swiper.virtual ? swiper.virtual.slides.length - 1 : swiper.slides.length - 1;
            return swiper.slideTo(lastIndex, speed, runCallbacks, internal);
        }

        return swiper.slideTo(prevIndex, speed, runCallbacks, internal);
    }

    /* eslint no-unused-vars: "off" */
    function slideReset(speed, runCallbacks, internal) {
        if (speed === void 0) {
            speed = this.params.speed;
        }

        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        const swiper = this;
        return swiper.slideTo(swiper.activeIndex, speed, runCallbacks, internal);
    }

    /* eslint no-unused-vars: "off" */
    function slideToClosest(speed, runCallbacks, internal, threshold) {
        if (speed === void 0) {
            speed = this.params.speed;
        }

        if (runCallbacks === void 0) {
            runCallbacks = true;
        }

        if (threshold === void 0) {
            threshold = 0.5;
        }

        const swiper = this;
        let index = swiper.activeIndex;
        const skip = Math.min(swiper.params.slidesPerGroupSkip, index);
        const snapIndex = skip + Math.floor((index - skip) / swiper.params.slidesPerGroup);
        const translate = swiper.rtlTranslate ? swiper.translate : -swiper.translate;

        if (translate >= swiper.snapGrid[snapIndex]) {
            // The current translate is on or after the current snap index, so the choice
            // is between the current index and the one after it.
            const currentSnap = swiper.snapGrid[snapIndex];
            const nextSnap = swiper.snapGrid[snapIndex + 1];

            if (translate - currentSnap > (nextSnap - currentSnap) * threshold) {
                index += swiper.params.slidesPerGroup;
            }
        } else {
            // The current translate is before the current snap index, so the choice
            // is between the current index and the one before it.
            const prevSnap = swiper.snapGrid[snapIndex - 1];
            const currentSnap = swiper.snapGrid[snapIndex];

            if (translate - prevSnap <= (currentSnap - prevSnap) * threshold) {
                index -= swiper.params.slidesPerGroup;
            }
        }

        index = Math.max(index, 0);
        index = Math.min(index, swiper.slidesGrid.length - 1);
        return swiper.slideTo(index, speed, runCallbacks, internal);
    }

    function slideToClickedSlide() {
        const swiper = this;
        const {
            params,
            $wrapperEl
        } = swiper;
        const slidesPerView = params.slidesPerView === 'auto' ? swiper.slidesPerViewDynamic() : params.slidesPerView;
        let slideToIndex = swiper.clickedIndex;
        let realIndex;

        if (params.loop) {
            if (swiper.animating) return;
            realIndex = parseInt($(swiper.clickedSlide).attr('data-swiper-slide-index'), 10);

            if (params.centeredSlides) {
                if (slideToIndex < swiper.loopedSlides - slidesPerView / 2 || slideToIndex > swiper.slides.length - swiper.loopedSlides + slidesPerView / 2) {
                    swiper.loopFix();
                    slideToIndex = $wrapperEl.children(`.${params.slideClass}[data-swiper-slide-index="${realIndex}"]:not(.${params.slideDuplicateClass})`).eq(0).index();
                    nextTick(() => {
                        swiper.slideTo(slideToIndex);
                    });
                } else {
                    swiper.slideTo(slideToIndex);
                }
            } else if (slideToIndex > swiper.slides.length - slidesPerView) {
                swiper.loopFix();
                slideToIndex = $wrapperEl.children(`.${params.slideClass}[data-swiper-slide-index="${realIndex}"]:not(.${params.slideDuplicateClass})`).eq(0).index();
                nextTick(() => {
                    swiper.slideTo(slideToIndex);
                });
            } else {
                swiper.slideTo(slideToIndex);
            }
        } else {
            swiper.slideTo(slideToIndex);
        }
    }

    var slide = {
        slideTo,
        slideToLoop,
        slideNext,
        slidePrev,
        slideReset,
        slideToClosest,
        slideToClickedSlide
    };

    function loopCreate() {
        const swiper = this;
        const document = getDocument();
        const {
            params,
            $wrapperEl
        } = swiper; // Remove duplicated slides

        const $selector = $wrapperEl.children().length > 0 ? $($wrapperEl.children()[0].parentNode) : $wrapperEl;
        $selector.children(`.${params.slideClass}.${params.slideDuplicateClass}`).remove();
        let slides = $selector.children(`.${params.slideClass}`);

        if (params.loopFillGroupWithBlank) {
            const blankSlidesNum = params.slidesPerGroup - slides.length % params.slidesPerGroup;

            if (blankSlidesNum !== params.slidesPerGroup) {
                for (let i = 0; i < blankSlidesNum; i += 1) {
                    const blankNode = $(document.createElement('div')).addClass(`${params.slideClass} ${params.slideBlankClass}`);
                    $selector.append(blankNode);
                }

                slides = $selector.children(`.${params.slideClass}`);
            }
        }

        if (params.slidesPerView === 'auto' && !params.loopedSlides) params.loopedSlides = slides.length;
        swiper.loopedSlides = Math.ceil(parseFloat(params.loopedSlides || params.slidesPerView, 10));
        swiper.loopedSlides += params.loopAdditionalSlides;

        if (swiper.loopedSlides > slides.length) {
            swiper.loopedSlides = slides.length;
        }

        const prependSlides = [];
        const appendSlides = [];
        slides.each((el, index) => {
            const slide = $(el);

            if (index < swiper.loopedSlides) {
                appendSlides.push(el);
            }

            if (index < slides.length && index >= slides.length - swiper.loopedSlides) {
                prependSlides.push(el);
            }

            slide.attr('data-swiper-slide-index', index);
        });

        for (let i = 0; i < appendSlides.length; i += 1) {
            $selector.append($(appendSlides[i].cloneNode(true)).addClass(params.slideDuplicateClass));
        }

        for (let i = prependSlides.length - 1; i >= 0; i -= 1) {
            $selector.prepend($(prependSlides[i].cloneNode(true)).addClass(params.slideDuplicateClass));
        }
    }

    function loopFix() {
        const swiper = this;
        swiper.emit('beforeLoopFix');
        const {
            activeIndex,
            slides,
            loopedSlides,
            allowSlidePrev,
            allowSlideNext,
            snapGrid,
            rtlTranslate: rtl
        } = swiper;
        let newIndex;
        swiper.allowSlidePrev = true;
        swiper.allowSlideNext = true;
        const snapTranslate = -snapGrid[activeIndex];
        const diff = snapTranslate - swiper.getTranslate(); // Fix For Negative Oversliding

        if (activeIndex < loopedSlides) {
            newIndex = slides.length - loopedSlides * 3 + activeIndex;
            newIndex += loopedSlides;
            const slideChanged = swiper.slideTo(newIndex, 0, false, true);

            if (slideChanged && diff !== 0) {
                swiper.setTranslate((rtl ? -swiper.translate : swiper.translate) - diff);
            }
        } else if (activeIndex >= slides.length - loopedSlides) {
            // Fix For Positive Oversliding
            newIndex = -slides.length + activeIndex + loopedSlides;
            newIndex += loopedSlides;
            const slideChanged = swiper.slideTo(newIndex, 0, false, true);

            if (slideChanged && diff !== 0) {
                swiper.setTranslate((rtl ? -swiper.translate : swiper.translate) - diff);
            }
        }

        swiper.allowSlidePrev = allowSlidePrev;
        swiper.allowSlideNext = allowSlideNext;
        swiper.emit('loopFix');
    }

    function loopDestroy() {
        const swiper = this;
        const {
            $wrapperEl,
            params,
            slides
        } = swiper;
        $wrapperEl.children(`.${params.slideClass}.${params.slideDuplicateClass},.${params.slideClass}.${params.slideBlankClass}`).remove();
        slides.removeAttr('data-swiper-slide-index');
    }

    var loop = {
        loopCreate,
        loopFix,
        loopDestroy
    };

    function setGrabCursor(moving) {
        const swiper = this;
        if (swiper.support.touch || !swiper.params.simulateTouch || swiper.params.watchOverflow && swiper.isLocked || swiper.params.cssMode) return;
        const el = swiper.params.touchEventsTarget === 'container' ? swiper.el : swiper.wrapperEl;
        el.style.cursor = 'move';
        el.style.cursor = moving ? 'grabbing' : 'grab';
    }

    function unsetGrabCursor() {
        const swiper = this;

        if (swiper.support.touch || swiper.params.watchOverflow && swiper.isLocked || swiper.params.cssMode) {
            return;
        }

        swiper[swiper.params.touchEventsTarget === 'container' ? 'el' : 'wrapperEl'].style.cursor = '';
    }

    var grabCursor = {
        setGrabCursor,
        unsetGrabCursor
    };

    function closestElement(selector, base) {
        if (base === void 0) {
            base = this;
        }

        function __closestFrom(el) {
            if (!el || el === getDocument() || el === getWindow()) return null;
            if (el.assignedSlot) el = el.assignedSlot;
            const found = el.closest(selector);
            return found || __closestFrom(el.getRootNode().host);
        }

        return __closestFrom(base);
    }

    function onTouchStart(event) {
        const swiper = this;
        const document = getDocument();
        const window = getWindow();
        const data = swiper.touchEventsData;
        const {
            params,
            touches,
            enabled
        } = swiper;
        if (!enabled) return;

        if (swiper.animating && params.preventInteractionOnTransition) {
            return;
        }

        if (!swiper.animating && params.cssMode && params.loop) {
            swiper.loopFix();
        }

        let e = event;
        if (e.originalEvent) e = e.originalEvent;
        let $targetEl = $(e.target);

        if (params.touchEventsTarget === 'wrapper') {
            if (!$targetEl.closest(swiper.wrapperEl).length) return;
        }

        data.isTouchEvent = e.type === 'touchstart';
        if (!data.isTouchEvent && 'which' in e && e.which === 3) return;
        if (!data.isTouchEvent && 'button' in e && e.button > 0) return;
        if (data.isTouched && data.isMoved) return; // change target el for shadow root component

        const swipingClassHasValue = !!params.noSwipingClass && params.noSwipingClass !== '';

        if (swipingClassHasValue && e.target && e.target.shadowRoot && event.path && event.path[0]) {
            $targetEl = $(event.path[0]);
        }

        const noSwipingSelector = params.noSwipingSelector ? params.noSwipingSelector : `.${params.noSwipingClass}`;
        const isTargetShadow = !!(e.target && e.target.shadowRoot); // use closestElement for shadow root element to get the actual closest for nested shadow root element

        if (params.noSwiping && (isTargetShadow ? closestElement(noSwipingSelector, e.target) : $targetEl.closest(noSwipingSelector)[0])) {
            swiper.allowClick = true;
            return;
        }

        if (params.swipeHandler) {
            if (!$targetEl.closest(params.swipeHandler)[0]) return;
        }

        touches.currentX = e.type === 'touchstart' ? e.targetTouches[0].pageX : e.pageX;
        touches.currentY = e.type === 'touchstart' ? e.targetTouches[0].pageY : e.pageY;
        const startX = touches.currentX;
        const startY = touches.currentY; // Do NOT start if iOS edge swipe is detected. Otherwise iOS app cannot swipe-to-go-back anymore

        const edgeSwipeDetection = params.edgeSwipeDetection || params.iOSEdgeSwipeDetection;
        const edgeSwipeThreshold = params.edgeSwipeThreshold || params.iOSEdgeSwipeThreshold;

        if (edgeSwipeDetection && (startX <= edgeSwipeThreshold || startX >= window.innerWidth - edgeSwipeThreshold)) {
            if (edgeSwipeDetection === 'prevent') {
                axPreventDefault(event);//AXUI
            } else {
                return;
            }
        }

        Object.assign(data, {
            isTouched: true,
            isMoved: false,
            allowTouchCallbacks: true,
            isScrolling: undefined,
            startMoving: undefined
        });
        touches.startX = startX;
        touches.startY = startY;
        data.touchStartTime = now();
        swiper.allowClick = true;
        swiper.updateSize();
        swiper.swipeDirection = undefined;
        if (params.threshold > 0) data.allowThresholdMove = false;

        if (e.type !== 'touchstart') {
            let preventDefault = true;

            if ($targetEl.is(data.focusableElements)) {
                preventDefault = false;

                if ($targetEl[0].nodeName === 'SELECT') {
                    data.isTouched = false;
                }
            }

            if (document.activeElement && $(document.activeElement).is(data.focusableElements) && document.activeElement !== $targetEl[0]) {
                document.activeElement.blur();
            }

            const shouldPreventDefault = preventDefault && swiper.allowTouchMove && params.touchStartPreventDefault;

            if ((params.touchStartForcePreventDefault || shouldPreventDefault) && !$targetEl[0].isContentEditable) {
                axPreventDefault(e);//AXUI
            }
        }

        if (swiper.params.freeMode && swiper.params.freeMode.enabled && swiper.freeMode && swiper.animating && !params.cssMode) {
            swiper.freeMode.onTouchStart();
        }

        swiper.emit('touchStart', e);
    }

    function onTouchMove(event) {
        const document = getDocument();
        const swiper = this;
        const data = swiper.touchEventsData;
        const {
            params,
            touches,
            rtlTranslate: rtl,
            enabled
        } = swiper;
        if (!enabled) return;
        let e = event;
        if (e.originalEvent) e = e.originalEvent;

        if (!data.isTouched) {
            if (data.startMoving && data.isScrolling) {
                swiper.emit('touchMoveOpposite', e);
            }

            return;
        }

        if (data.isTouchEvent && e.type !== 'touchmove') return;
        const targetTouch = e.type === 'touchmove' && e.targetTouches && (e.targetTouches[0] || e.changedTouches[0]);
        const pageX = e.type === 'touchmove' ? targetTouch.pageX : e.pageX;
        const pageY = e.type === 'touchmove' ? targetTouch.pageY : e.pageY;

        if (e.preventedByNestedSwiper) {
            touches.startX = pageX;
            touches.startY = pageY;
            return;
        }

        if (!swiper.allowTouchMove) {
            if (!$(e.target).is(data.focusableElements)) {
                swiper.allowClick = false;
            }

            if (data.isTouched) {
                Object.assign(touches, {
                    startX: pageX,
                    startY: pageY,
                    currentX: pageX,
                    currentY: pageY
                });
                data.touchStartTime = now();
            }

            return;
        }

        if (data.isTouchEvent && params.touchReleaseOnEdges && !params.loop) {
            if (swiper.isVertical()) {
                // Vertical
                if (pageY < touches.startY && swiper.translate <= swiper.maxTranslate() || pageY > touches.startY && swiper.translate >= swiper.minTranslate()) {
                    data.isTouched = false;
                    data.isMoved = false;
                    return;
                }
            } else if (pageX < touches.startX && swiper.translate <= swiper.maxTranslate() || pageX > touches.startX && swiper.translate >= swiper.minTranslate()) {
                return;
            }
        }

        if (data.isTouchEvent && document.activeElement) {
            if (e.target === document.activeElement && $(e.target).is(data.focusableElements)) {
                data.isMoved = true;
                swiper.allowClick = false;
                return;
            }
        }

        if (data.allowTouchCallbacks) {
            swiper.emit('touchMove', e);
        }

        if (e.targetTouches && e.targetTouches.length > 1) return;
        touches.currentX = pageX;
        touches.currentY = pageY;
        const diffX = touches.currentX - touches.startX;
        const diffY = touches.currentY - touches.startY;
        if (swiper.params.threshold && Math.sqrt(diffX ** 2 + diffY ** 2) < swiper.params.threshold) return;

        if (typeof data.isScrolling === 'undefined') {
            let touchAngle;

            if (swiper.isHorizontal() && touches.currentY === touches.startY || swiper.isVertical() && touches.currentX === touches.startX) {
                data.isScrolling = false;
            } else {
                // eslint-disable-next-line
                if (diffX * diffX + diffY * diffY >= 25) {
                    touchAngle = Math.atan2(Math.abs(diffY), Math.abs(diffX)) * 180 / Math.PI;
                    data.isScrolling = swiper.isHorizontal() ? touchAngle > params.touchAngle : 90 - touchAngle > params.touchAngle;
                }
            }
        }

        if (data.isScrolling) {
            swiper.emit('touchMoveOpposite', e);
        }

        if (typeof data.startMoving === 'undefined') {
            if (touches.currentX !== touches.startX || touches.currentY !== touches.startY) {
                data.startMoving = true;
            }
        }

        if (data.isScrolling) {
            data.isTouched = false;
            return;
        }

        if (!data.startMoving) {
            return;
        }

        swiper.allowClick = false;

        if (!params.cssMode && e.cancelable) {
            axPreventDefault(e);//AXUI
        }

        if (params.touchMoveStopPropagation && !params.nested) {
            e.stopPropagation();
        }

        if (!data.isMoved) {
            if (params.loop && !params.cssMode) {
                swiper.loopFix();
            }

            data.startTranslate = swiper.getTranslate();
            swiper.setTransition(0);

            if (swiper.animating) {
                swiper.$wrapperEl.trigger('webkitTransitionEnd transitionend');
            }

            data.allowMomentumBounce = false; // Grab Cursor

            if (params.grabCursor && (swiper.allowSlideNext === true || swiper.allowSlidePrev === true)) {
                swiper.setGrabCursor(true);
            }

            swiper.emit('sliderFirstMove', e);
        }

        swiper.emit('sliderMove', e);
        data.isMoved = true;
        let diff = swiper.isHorizontal() ? diffX : diffY;
        touches.diff = diff;
        diff *= params.touchRatio;
        if (rtl) diff = -diff;
        swiper.swipeDirection = diff > 0 ? 'prev' : 'next';
        data.currentTranslate = diff + data.startTranslate;
        let disableParentSwiper = true;
        let resistanceRatio = params.resistanceRatio;

        if (params.touchReleaseOnEdges) {
            resistanceRatio = 0;
        }

        if (diff > 0 && data.currentTranslate > swiper.minTranslate()) {
            disableParentSwiper = false;
            if (params.resistance) data.currentTranslate = swiper.minTranslate() - 1 + (-swiper.minTranslate() + data.startTranslate + diff) ** resistanceRatio;
        } else if (diff < 0 && data.currentTranslate < swiper.maxTranslate()) {
            disableParentSwiper = false;
            if (params.resistance) data.currentTranslate = swiper.maxTranslate() + 1 - (swiper.maxTranslate() - data.startTranslate - diff) ** resistanceRatio;
        }

        if (disableParentSwiper) {
            e.preventedByNestedSwiper = true;
        } // Directions locks


        if (!swiper.allowSlideNext && swiper.swipeDirection === 'next' && data.currentTranslate < data.startTranslate) {
            data.currentTranslate = data.startTranslate;
        }

        if (!swiper.allowSlidePrev && swiper.swipeDirection === 'prev' && data.currentTranslate > data.startTranslate) {
            data.currentTranslate = data.startTranslate;
        }

        if (!swiper.allowSlidePrev && !swiper.allowSlideNext) {
            data.currentTranslate = data.startTranslate;
        } // Threshold


        if (params.threshold > 0) {
            if (Math.abs(diff) > params.threshold || data.allowThresholdMove) {
                if (!data.allowThresholdMove) {
                    data.allowThresholdMove = true;
                    touches.startX = touches.currentX;
                    touches.startY = touches.currentY;
                    data.currentTranslate = data.startTranslate;
                    touches.diff = swiper.isHorizontal() ? touches.currentX - touches.startX : touches.currentY - touches.startY;
                    return;
                }
            } else {
                data.currentTranslate = data.startTranslate;
                return;
            }
        }

        if (!params.followFinger || params.cssMode) return; // Update active index in free mode

        if (params.freeMode && params.freeMode.enabled && swiper.freeMode || params.watchSlidesProgress) {
            swiper.updateActiveIndex();
            swiper.updateSlidesClasses();
        }

        if (swiper.params.freeMode && params.freeMode.enabled && swiper.freeMode) {
            swiper.freeMode.onTouchMove();
        } // Update progress


        swiper.updateProgress(data.currentTranslate); // Update translate

        swiper.setTranslate(data.currentTranslate);
    }

    function onTouchEnd(event) {
        const swiper = this;
        const data = swiper.touchEventsData;
        const {
            params,
            touches,
            rtlTranslate: rtl,
            slidesGrid,
            enabled
        } = swiper;
        if (!enabled) return;
        let e = event;
        if (e.originalEvent) e = e.originalEvent;

        if (data.allowTouchCallbacks) {
            swiper.emit('touchEnd', e);
        }

        data.allowTouchCallbacks = false;

        if (!data.isTouched) {
            if (data.isMoved && params.grabCursor) {
                swiper.setGrabCursor(false);
            }

            data.isMoved = false;
            data.startMoving = false;
            return;
        } // Return Grab Cursor


        if (params.grabCursor && data.isMoved && data.isTouched && (swiper.allowSlideNext === true || swiper.allowSlidePrev === true)) {
            swiper.setGrabCursor(false);
        } // Time diff


        const touchEndTime = now();
        const timeDiff = touchEndTime - data.touchStartTime; // Tap, doubleTap, Click

        if (swiper.allowClick) {
            const pathTree = e.path || e.composedPath && e.composedPath();
            swiper.updateClickedSlide(pathTree && pathTree[0] || e.target);
            swiper.emit('tap click', e);

            if (timeDiff < 300 && touchEndTime - data.lastClickTime < 300) {
                swiper.emit('doubleTap doubleClick', e);
            }
        }

        data.lastClickTime = now();
        nextTick(() => {
            if (!swiper.destroyed) swiper.allowClick = true;
        });

        if (!data.isTouched || !data.isMoved || !swiper.swipeDirection || touches.diff === 0 || data.currentTranslate === data.startTranslate) {
            data.isTouched = false;
            data.isMoved = false;
            data.startMoving = false;
            return;
        }

        data.isTouched = false;
        data.isMoved = false;
        data.startMoving = false;
        let currentPos;

        if (params.followFinger) {
            currentPos = rtl ? swiper.translate : -swiper.translate;
        } else {
            currentPos = -data.currentTranslate;
        }

        if (params.cssMode) {
            return;
        }

        if (swiper.params.freeMode && params.freeMode.enabled) {
            swiper.freeMode.onTouchEnd({
                currentPos
            });
            return;
        } // Find current slide


        let stopIndex = 0;
        let groupSize = swiper.slidesSizesGrid[0];

        for (let i = 0; i < slidesGrid.length; i += i < params.slidesPerGroupSkip ? 1 : params.slidesPerGroup) {
            const increment = i < params.slidesPerGroupSkip - 1 ? 1 : params.slidesPerGroup;

            if (typeof slidesGrid[i + increment] !== 'undefined') {
                if (currentPos >= slidesGrid[i] && currentPos < slidesGrid[i + increment]) {
                    stopIndex = i;
                    groupSize = slidesGrid[i + increment] - slidesGrid[i];
                }
            } else if (currentPos >= slidesGrid[i]) {
                stopIndex = i;
                groupSize = slidesGrid[slidesGrid.length - 1] - slidesGrid[slidesGrid.length - 2];
            }
        }

        let rewindFirstIndex = null;
        let rewindLastIndex = null;

        if (params.rewind) {
            if (swiper.isBeginning) {
                rewindLastIndex = swiper.params.virtual && swiper.params.virtual.enabled && swiper.virtual ? swiper.virtual.slides.length - 1 : swiper.slides.length - 1;
            } else if (swiper.isEnd) {
                rewindFirstIndex = 0;
            }
        } // Find current slide size


        const ratio = (currentPos - slidesGrid[stopIndex]) / groupSize;
        const increment = stopIndex < params.slidesPerGroupSkip - 1 ? 1 : params.slidesPerGroup;

        if (timeDiff > params.longSwipesMs) {
            // Long touches
            if (!params.longSwipes) {
                swiper.slideTo(swiper.activeIndex);
                return;
            }

            if (swiper.swipeDirection === 'next') {
                if (ratio >= params.longSwipesRatio) swiper.slideTo(params.rewind && swiper.isEnd ? rewindFirstIndex : stopIndex + increment); else swiper.slideTo(stopIndex);
            }

            if (swiper.swipeDirection === 'prev') {
                if (ratio > 1 - params.longSwipesRatio) {
                    swiper.slideTo(stopIndex + increment);
                } else if (rewindLastIndex !== null && ratio < 0 && Math.abs(ratio) > params.longSwipesRatio) {
                    swiper.slideTo(rewindLastIndex);
                } else {
                    swiper.slideTo(stopIndex);
                }
            }
        } else {
            // Short swipes
            if (!params.shortSwipes) {
                swiper.slideTo(swiper.activeIndex);
                return;
            }

            const isNavButtonTarget = swiper.navigation && (e.target === swiper.navigation.nextEl || e.target === swiper.navigation.prevEl);

            if (!isNavButtonTarget) {
                if (swiper.swipeDirection === 'next') {
                    swiper.slideTo(rewindFirstIndex !== null ? rewindFirstIndex : stopIndex + increment);
                }

                if (swiper.swipeDirection === 'prev') {
                    swiper.slideTo(rewindLastIndex !== null ? rewindLastIndex : stopIndex);
                }
            } else if (e.target === swiper.navigation.nextEl) {
                swiper.slideTo(stopIndex + increment);
            } else {
                swiper.slideTo(stopIndex);
            }
        }
    }

    function onResize() {
        const swiper = this;
        const {
            params,
            el
        } = swiper;
        if (el && el.offsetWidth === 0) return; // Breakpoints

        if (params.breakpoints) {
            swiper.setBreakpoint();
        } // Save locks


        const {
            allowSlideNext,
            allowSlidePrev,
            snapGrid
        } = swiper; // Disable locks on resize

        swiper.allowSlideNext = true;
        swiper.allowSlidePrev = true;
        swiper.updateSize();
        swiper.updateSlides();
        swiper.updateSlidesClasses();

        if ((params.slidesPerView === 'auto' || params.slidesPerView > 1) && swiper.isEnd && !swiper.isBeginning && !swiper.params.centeredSlides) {
            swiper.slideTo(swiper.slides.length - 1, 0, false, true);
        } else {
            swiper.slideTo(swiper.activeIndex, 0, false, true);
        }

        if (swiper.autoplay && swiper.autoplay.running && swiper.autoplay.paused) {
            swiper.autoplay.run();
        } // Return locks after resize


        swiper.allowSlidePrev = allowSlidePrev;
        swiper.allowSlideNext = allowSlideNext;

        if (swiper.params.watchOverflow && snapGrid !== swiper.snapGrid) {
            swiper.checkOverflow();
        }
    }

    function onClick(e) {
        const swiper = this;
        if (!swiper.enabled) return;

        if (!swiper.allowClick) {
            if (swiper.params.preventClicks){
                axPreventDefault(e);//AXUI
            }

            if (swiper.params.preventClicksPropagation && swiper.animating) {
                e.stopPropagation();
                e.stopImmediatePropagation();
            }
        }
    }

    function onScroll() {
        const swiper = this;
        const {
            wrapperEl,
            rtlTranslate,
            enabled
        } = swiper;
        if (!enabled) return;
        swiper.previousTranslate = swiper.translate;

        if (swiper.isHorizontal()) {
            swiper.translate = -wrapperEl.scrollLeft;
        } else {
            swiper.translate = -wrapperEl.scrollTop;
        } // eslint-disable-next-line


        if (swiper.translate === 0) swiper.translate = 0;
        swiper.updateActiveIndex();
        swiper.updateSlidesClasses();
        let newProgress;
        const translatesDiff = swiper.maxTranslate() - swiper.minTranslate();

        if (translatesDiff === 0) {
            newProgress = 0;
        } else {
            newProgress = (swiper.translate - swiper.minTranslate()) / translatesDiff;
        }

        if (newProgress !== swiper.progress) {
            swiper.updateProgress(rtlTranslate ? -swiper.translate : swiper.translate);
        }

        swiper.emit('setTranslate', swiper.translate, false);
    }

    let dummyEventAttached = false;

    function dummyEventListener() { }

    const events = (swiper, method) => {
        const document = getDocument();
        const {
            params,
            touchEvents,
            el,
            wrapperEl,
            device,
            support
        } = swiper;
        const capture = !!params.nested;
        const domMethod = method === 'on' ? 'addEventListener' : 'removeEventListener';
        const swiperMethod = method; // Touch Events

        if (!support.touch) {
            el[domMethod](touchEvents.start, swiper.onTouchStart, false);
            document[domMethod](touchEvents.move, swiper.onTouchMove, capture);
            document[domMethod](touchEvents.end, swiper.onTouchEnd, false);
        } else {
            const passiveListener = touchEvents.start === 'touchstart' && support.passiveListener && params.passiveListeners ? {
                passive: true,
                capture: false
            } : false;
            el[domMethod](touchEvents.start, swiper.onTouchStart, passiveListener);
            el[domMethod](touchEvents.move, swiper.onTouchMove, support.passiveListener ? {
                passive: false,
                capture
            } : capture);
            el[domMethod](touchEvents.end, swiper.onTouchEnd, passiveListener);

            if (touchEvents.cancel) {
                el[domMethod](touchEvents.cancel, swiper.onTouchEnd, passiveListener);
            }
        } // Prevent Links Clicks


        if (params.preventClicks || params.preventClicksPropagation) {
            el[domMethod]('click', swiper.onClick, true);
        }

        if (params.cssMode) {
            wrapperEl[domMethod]('scroll', swiper.onScroll);
        } // Resize handler


        if (params.updateOnWindowResize) {
            swiper[swiperMethod](device.ios || device.android ? 'resize orientationchange observerUpdate' : 'resize observerUpdate', onResize, true);
        } else {
            swiper[swiperMethod]('observerUpdate', onResize, true);
        }
    };

    function attachEvents() {
        const swiper = this;
        const document = getDocument();
        const {
            params,
            support
        } = swiper;
        swiper.onTouchStart = onTouchStart.bind(swiper);
        swiper.onTouchMove = onTouchMove.bind(swiper);
        swiper.onTouchEnd = onTouchEnd.bind(swiper);

        if (params.cssMode) {
            swiper.onScroll = onScroll.bind(swiper);
        }

        swiper.onClick = onClick.bind(swiper);

        if (support.touch && !dummyEventAttached) {
            document.addEventListener('touchstart', dummyEventListener);
            dummyEventAttached = true;
        }

        events(swiper, 'on');
    }

    function detachEvents() {
        const swiper = this;
        events(swiper, 'off');
    }

    var _events = {//AXUI,原events$1
        attachEvents,
        detachEvents
    };

    const isGridEnabled = (swiper, params) => {
        return swiper.grid && params.grid && params.grid.rows > 1;
    };

    function setBreakpoint() {
        const swiper = this;
        const {
            activeIndex,
            initialized,
            loopedSlides = 0,
            params,
            $el
        } = swiper;
        const breakpoints = params.breakpoints;
        if (!breakpoints || breakpoints && Object.keys(breakpoints).length === 0) return; // Get breakpoint for window width and update parameters

        const breakpoint = swiper.getBreakpoint(breakpoints, swiper.params.breakpointsBase, swiper.el);
        if (!breakpoint || swiper.currentBreakpoint === breakpoint) return;
        const breakpointOnlyParams = breakpoint in breakpoints ? breakpoints[breakpoint] : undefined;
        const breakpointParams = breakpointOnlyParams || swiper.originalParams;
        const wasMultiRow = isGridEnabled(swiper, params);
        const isMultiRow = isGridEnabled(swiper, breakpointParams);
        const wasEnabled = params.enabled;

        if (wasMultiRow && !isMultiRow) {
            $el.removeClass(`${params.containerModifierClass}grids ${params.containerModifierClass}grid-column`);//AXUI
            swiper.emitContainerClasses();
        } else if (!wasMultiRow && isMultiRow) {
            $el.addClass(`${params.containerModifierClass}grids`);//AXUI

            if (breakpointParams.grid.fill && breakpointParams.grid.fill === 'column' || !breakpointParams.grid.fill && params.grid.fill === 'column') {
                $el.addClass(`${params.containerModifierClass}grid-column`);
            }

            swiper.emitContainerClasses();
        }

        const directionChanged = breakpointParams.direction && breakpointParams.direction !== params.direction;
        const needsReLoop = params.loop && (breakpointParams.slidesPerView !== params.slidesPerView || directionChanged);

        if (directionChanged && initialized) {
            swiper.changeDirection();
        }

        extend(swiper.params, breakpointParams);
        const isEnabled = swiper.params.enabled;
        Object.assign(swiper, {
            allowTouchMove: swiper.params.allowTouchMove,
            allowSlideNext: swiper.params.allowSlideNext,
            allowSlidePrev: swiper.params.allowSlidePrev
        });

        if (wasEnabled && !isEnabled) {
            swiper.disable();
        } else if (!wasEnabled && isEnabled) {
            swiper.enable();
        }

        swiper.currentBreakpoint = breakpoint;
        swiper.emit('_beforeBreakpoint', breakpointParams);

        if (needsReLoop && initialized) {
            swiper.loopDestroy();
            swiper.loopCreate();
            swiper.updateSlides();
            swiper.slideTo(activeIndex - loopedSlides + swiper.loopedSlides, 0, false);
        }

        swiper.emit('breakpoint', breakpointParams);
    }

    function getBreakpoint(breakpoints, base, containerEl) {
        if (base === void 0) {
            base = 'window';
        }

        if (!breakpoints || base === 'container' && !containerEl) return undefined;
        let breakpoint = false;
        const window = getWindow();
        const currentHeight = base === 'window' ? window.innerHeight : containerEl.clientHeight;
        const points = Object.keys(breakpoints).map(point => {
            if (typeof point === 'string' && point.indexOf('@') === 0) {
                const minRatio = parseFloat(point.substr(1));
                const value = currentHeight * minRatio;
                return {
                    value,
                    point
                };
            }

            return {
                value: point,
                point
            };
        });
        points.sort((a, b) => parseInt(a.value, 10) - parseInt(b.value, 10));

        for (let i = 0; i < points.length; i += 1) {
            const {
                point,
                value
            } = points[i];

            if (base === 'window') {
                if (window.matchMedia(`(min-width: ${value}px)`).matches) {
                    breakpoint = point;
                }
            } else if (value <= containerEl.clientWidth) {
                breakpoint = point;
            }
        }

        return breakpoint || 'max';
    }

    var breakpoints = {
        setBreakpoint,
        getBreakpoint
    };

    function prepareClasses(entries, prefix) {
        const resultClasses = [];
        entries.forEach(item => {
            if (typeof item === 'object') {
                Object.keys(item).forEach(classNames => {
                    if (item[classNames]) {
                        resultClasses.push(prefix + classNames);
                    }
                });
            } else if (typeof item === 'string') {
                resultClasses.push(prefix + item);
            }
        });
        return resultClasses;
    }

    function addClasses() {
        const swiper = this;
        const {
            classNames,
            params,
            rtl,
            $el,
            device,
            support
        } = swiper; // prettier-ignore

        const suffixes = prepareClasses(['initialized', params.direction, {
            'pointer-events': !support.touch
        }, {
                'free-mode': swiper.params.freeMode && params.freeMode.enabled
            }, {
                'autoheight': params.autoHeight
            }, {
                'rtl': rtl
            }, {
                'grids': params.grid && params.grid.rows > 1//AXUI
            }, {
                'grid-column': params.grid && params.grid.rows > 1 && params.grid.fill === 'column'
            }, {
                'android': device.android
            }, {
                'ios': device.ios
            }, {
                'css-mode': params.cssMode
            }, {
                'centered': params.cssMode && params.centeredSlides
            }, {
                'watch-progress': params.watchSlidesProgress
            }], params.containerModifierClass);
        classNames.push(...suffixes);
        $el.addClass([...classNames].join(' '));
        swiper.emitContainerClasses();
    }

    function removeClasses() {
        const swiper = this;
        const {
            $el,
            classNames
        } = swiper;
        $el.removeClass(classNames.join(' '));
        swiper.emitContainerClasses();
    }

    var classes = {
        addClasses,
        removeClasses
    };

    function loadImage(imageEl, src, srcset, sizes, checkForComplete, callback) {
        const window = getWindow();
        let image;

        function onReady() {
            if (callback) callback();
        }

        const isPicture = $(imageEl).parent('picture')[0];

        if (!isPicture && (!imageEl.complete || !checkForComplete)) {
            if (src) {
                image = new window.Image();
                image.onload = onReady;
                image.onerror = onReady;

                if (sizes) {
                    image.sizes = sizes;
                }

                if (srcset) {
                    image.srcset = srcset;
                }

                if (src) {
                    image.src = src;
                }
            } else {
                onReady();
            }
        } else {
            // image already loaded...
            onReady();
        }
    }

    function preloadImages() {
        const swiper = this;
        swiper.imagesToLoad = swiper.$el.find('img');

        function onReady() {
            if (typeof swiper === 'undefined' || swiper === null || !swiper || swiper.destroyed) return;
            if (swiper.imagesLoaded !== undefined) swiper.imagesLoaded += 1;

            if (swiper.imagesLoaded === swiper.imagesToLoad.length) {
                if (swiper.params.updateOnImagesReady) swiper.update();
                swiper.emit('imagesReady');
            }
        }

        for (let i = 0; i < swiper.imagesToLoad.length; i += 1) {
            const imageEl = swiper.imagesToLoad[i];
            swiper.loadImage(imageEl, imageEl.currentSrc || imageEl.getAttribute('src'), imageEl.srcset || imageEl.getAttribute('srcset'), imageEl.sizes || imageEl.getAttribute('sizes'), true, onReady);
        }
    }

    var images = {
        loadImage,
        preloadImages
    };

    function checkOverflow() {
        const swiper = this;
        const {
            isLocked: wasLocked,
            params
        } = swiper;
        const {
            slidesOffsetBefore
        } = params;

        if (slidesOffsetBefore) {
            const lastSlideIndex = swiper.slides.length - 1;
            const lastSlideRightEdge = swiper.slidesGrid[lastSlideIndex] + swiper.slidesSizesGrid[lastSlideIndex] + slidesOffsetBefore * 2;
            swiper.isLocked = swiper.size > lastSlideRightEdge;
        } else {
            swiper.isLocked = swiper.snapGrid.length === 1;
        }

        if (params.allowSlideNext === true) {
            swiper.allowSlideNext = !swiper.isLocked;
        }

        if (params.allowSlidePrev === true) {
            swiper.allowSlidePrev = !swiper.isLocked;
        }

        if (wasLocked && wasLocked !== swiper.isLocked) {
            swiper.isEnd = false;
        }

        if (wasLocked !== swiper.isLocked) {
            swiper.emit(swiper.isLocked ? 'lock' : 'unlock');
        }
    }

    var checkOverflow$1 = {
        checkOverflow
    };

    var defaults = {
        init: true,
        direction: 'horizontal',
        touchEventsTarget: 'wrapper',
        initialSlide: 0,
        speed: 300,
        cssMode: false,
        updateOnWindowResize: true,
        resizeObserver: true,
        nested: false,
        createElements: false,
        enabled: true,
        focusableElements: 'input, select, option, textarea, button, video, label',
        // Overrides
        width: null,
        height: null,
        //
        preventInteractionOnTransition: false,
        // ssr
        userAgent: null,
        url: null,
        // To support iOS's swipe-to-go-back gesture (when being used in-app).
        edgeSwipeDetection: false,
        edgeSwipeThreshold: 20,
        // Autoheight
        autoHeight: false,
        // Set wrapper width
        setWrapperSize: false,
        // Virtual Translate
        virtualTranslate: false,
        // Effects
        effect: 'slide',
        // 'slide' or 'fade' or 'cube' or 'coverflow' or 'flip'
        // Breakpoints
        breakpoints: undefined,
        breakpointsBase: 'window',
        // Slides grid
        spaceBetween: 0,
        slidesPerView: 1,
        slidesPerGroup: 1,
        slidesPerGroupSkip: 0,
        slidesPerGroupAuto: false,
        centeredSlides: false,
        centeredSlidesBounds: false,
        slidesOffsetBefore: 0,
        // in px
        slidesOffsetAfter: 0,
        // in px
        normalizeSlideIndex: true,
        centerInsufficientSlides: false,
        // Disable swiper and hide navigation when container not overflow
        watchOverflow: true,
        // Round length
        roundLengths: false,
        // Touches
        touchRatio: 1,
        touchAngle: 45,
        simulateTouch: true,
        shortSwipes: true,
        longSwipes: true,
        longSwipesRatio: 0.5,
        longSwipesMs: 300,
        followFinger: true,
        allowTouchMove: true,
        threshold: 0,
        touchMoveStopPropagation: false,
        touchStartPreventDefault: true,
        touchStartForcePreventDefault: false,
        touchReleaseOnEdges: false,
        // Unique Navigation Elements
        uniqueNavElements: true,
        // Resistance
        resistance: true,
        resistanceRatio: 0.85,
        // Progress
        watchSlidesProgress: false,
        // Cursor
        grabCursor: false,
        // Clicks
        preventClicks: true,
        preventClicksPropagation: true,
        slideToClickedSlide: false,
        // Images
        preloadImages: true,
        updateOnImagesReady: true,
        // loop
        loop: false,
        loopAdditionalSlides: 0,
        loopedSlides: null,
        loopFillGroupWithBlank: false,
        loopPreventsSlide: true,
        // rewind
        rewind: false,
        // Swiping/no swiping
        allowSlidePrev: true,
        allowSlideNext: true,
        swipeHandler: null,
        // '.swipe-handler',
        noSwiping: true,
        noSwipingClass: 'ax-no-swiping',//AXUI,如果禁止拖拽滚动，则给slide加上这个class
        noSwipingSelector: null,
        // Passive Listeners
        passiveListeners: true,
        maxBackfaceHiddenSlides: 10,
        // NS
        containerModifierClass: 'ax-',
        // NEW
        slideClass: 'ax-slide',
        slideBlankClass: 'ax-slide-invisible-blank',
        slideActiveClass: 'ax-slide-active',
        slideDuplicateActiveClass: 'ax-slide-duplicate-active',
        slideVisibleClass: 'ax-slide-visible',
        slideDuplicateClass: 'ax-slide-duplicate',
        slideNextClass: 'ax-slide-next',
        slideDuplicateNextClass: 'ax-slide-duplicate-next',
        slidePrevClass: 'ax-slide-prev',
        slideDuplicatePrevClass: 'ax-slide-duplicate-prev',
        wrapperClass: 'ax-wrapper',
        // Callbacks
        runCallbacksOnInit: true,
        // Internals
        _emitClasses: false
    };

    function moduleExtendParams(params, allModulesParams) {
        return function extendParams(obj) {
            if (obj === void 0) {
                obj = {};
            }

            const moduleParamName = Object.keys(obj)[0];
            const moduleParams = obj[moduleParamName];

            if (typeof moduleParams !== 'object' || moduleParams === null) {
                extend(allModulesParams, obj);
                return;
            }

            if (['navigation', 'pagination', 'scrollbar'].indexOf(moduleParamName) >= 0 && params[moduleParamName] === true) {
                params[moduleParamName] = {
                    auto: true
                };
            }

            if (!(moduleParamName in params && 'enabled' in moduleParams)) {
                extend(allModulesParams, obj);
                return;
            }

            if (params[moduleParamName] === true) {
                params[moduleParamName] = {
                    enabled: true
                };
            }

            if (typeof params[moduleParamName] === 'object' && !('enabled' in params[moduleParamName])) {
                params[moduleParamName].enabled = true;
            }

            if (!params[moduleParamName]) params[moduleParamName] = {
                enabled: false
            };
            extend(allModulesParams, obj);
        };
    }

    /* eslint no-param-reassign: "off" */
    const prototypes = {
        eventsEmitter,
        update,
        translate,
        transition,
        slide,
        loop,
        grabCursor,
        events: _events,//AXUI,原events$1
        breakpoints,
        checkOverflow: checkOverflow$1,
        classes,
        images
    };
    const extendedDefaults = {};

    class Swiper {
        constructor() {
            let el;
            let params;

            for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
                args[_key] = arguments[_key];
            }

            if (args.length === 1 && args[0].constructor && Object.prototype.toString.call(args[0]).slice(8, -1) === 'Object') {
                params = args[0];
            } else {
                [el, params] = args;
            }

            if (!params) params = {};
            params = extend({}, params);
            if (el && !params.el) params.el = el;

            if (params.el && $(params.el).length > 1) {
                const swipers = [];
                $(params.el).each(containerEl => {
                    const newParams = extend({}, params, {
                        el: containerEl
                    });
                    swipers.push(new Swiper(newParams));
                });
                return swipers;
            } // Swiper Instance


            const swiper = this;
            swiper.__swiper__ = true;
            swiper.support = getSupport();
            swiper.device = getDevice({
                userAgent: params.userAgent
            });
            swiper.browser = getBrowser();
            swiper.eventsListeners = {};
            swiper.eventsAnyListeners = [];
            swiper.modules = [...swiper.__modules__];

            if (params.modules && Array.isArray(params.modules)) {
                swiper.modules.push(...params.modules);
            }

            const allModulesParams = {};
            swiper.modules.forEach(mod => {
                mod({
                    swiper,
                    extendParams: moduleExtendParams(params, allModulesParams),
                    on: swiper.on.bind(swiper),
                    once: swiper.once.bind(swiper),
                    off: swiper.off.bind(swiper),
                    emit: swiper.emit.bind(swiper)
                });
            }); // Extend defaults with modules params

            const swiperParams = extend({}, defaults, allModulesParams); // Extend defaults with passed params

            swiper.params = extend({}, swiperParams, extendedDefaults, params);

            //swiper.params = axDataset(params.el, swiper.params);//AXUI,追加DOM的data属性值

            //swiper.params =  axMerge(swiper.params, axAttrToJson(params.el, 'axSwiper'));//axSwiper="{}"//AXUI

            
           
            //AXUI,从dom中取class。如果通过ID创建实例，那么params.el是ID，如果通过axSwiper创建实例，那么params.el是DOM
            if (typeof params.el == 'string') {
                if (swiper.params.navigation.prevEl == '.ax-prev') {
                    swiper.params.navigation.prevEl = document.querySelector(params.el + ' .ax-prev');
                }
                if (swiper.params.navigation.nextEl == '.ax-next') {
                    swiper.params.navigation.nextEl = document.querySelector(params.el + ' .ax-next');
                }
            } else {
                if (swiper.params.navigation.prevEl == '.ax-prev') {
                    swiper.params.navigation.prevEl = params.el.querySelector('.ax-prev');
                }
                if (swiper.params.navigation.nextEl == '.ax-next') {
                    swiper.params.navigation.nextEl = params.el.querySelector('.ax-next');
                }
            }


            swiper.originalParams = extend({}, swiper.params);

            swiper.passedParams = extend({}, params); // add event listeners

            if (swiper.params && swiper.params.on) {
                Object.keys(swiper.params.on).forEach(eventName => {
                    swiper.on(eventName, swiper.params.on[eventName]);
                });
            }

            if (swiper.params && swiper.params.onAny) {
                swiper.onAny(swiper.params.onAny);
            } // Save Dom lib


            swiper.$ = $; // Extend Swiper

            Object.assign(swiper, {
                enabled: swiper.params.enabled,
                el,
                // Classes
                classNames: [],
                // Slides
                slides: $(),
                slidesGrid: [],
                snapGrid: [],
                slidesSizesGrid: [],

                // isDirection
                isHorizontal() {
                    return swiper.params.direction === 'horizontal';
                },

                isVertical() {
                    return swiper.params.direction === 'vertical';
                },

                // Indexes
                activeIndex: 0,
                realIndex: 0,
                //
                isBeginning: true,
                isEnd: false,
                // Props
                translate: 0,
                previousTranslate: 0,
                progress: 0,
                velocity: 0,
                animating: false,
                // Locks
                allowSlideNext: swiper.params.allowSlideNext,
                allowSlidePrev: swiper.params.allowSlidePrev,
                // Touch Events
                touchEvents: function touchEvents() {
                    const touch = ['touchstart', 'touchmove', 'touchend', 'touchcancel'];
                    const desktop = ['pointerdown', 'pointermove', 'pointerup'];
                    swiper.touchEventsTouch = {
                        start: touch[0],
                        move: touch[1],
                        end: touch[2],
                        cancel: touch[3]
                    };
                    swiper.touchEventsDesktop = {
                        start: desktop[0],
                        move: desktop[1],
                        end: desktop[2]
                    };
                    return swiper.support.touch || !swiper.params.simulateTouch ? swiper.touchEventsTouch : swiper.touchEventsDesktop;
                }(),
                touchEventsData: {
                    isTouched: undefined,
                    isMoved: undefined,
                    allowTouchCallbacks: undefined,
                    touchStartTime: undefined,
                    isScrolling: undefined,
                    currentTranslate: undefined,
                    startTranslate: undefined,
                    allowThresholdMove: undefined,
                    // Form elements to match
                    focusableElements: swiper.params.focusableElements,
                    // Last click time
                    lastClickTime: now(),
                    clickTimeout: undefined,
                    // Velocities
                    velocities: [],
                    allowMomentumBounce: undefined,
                    isTouchEvent: undefined,
                    startMoving: undefined
                },
                // Clicks
                allowClick: true,
                // Touches
                allowTouchMove: swiper.params.allowTouchMove,
                touches: {
                    startX: 0,
                    startY: 0,
                    currentX: 0,
                    currentY: 0,
                    diff: 0
                },
                // Images
                imagesToLoad: [],
                imagesLoaded: 0
            });
           
            swiper.emit('_swiper'); // Init

            if (swiper.params.init) { 
                swiper.init();
            } // Return app instance

            

            return swiper;
        }

        enable() {
            const swiper = this;
            if (swiper.enabled) return;
            swiper.enabled = true;

            if (swiper.params.grabCursor) {
                swiper.setGrabCursor();
            }

            swiper.emit('enable');
        }

        disable() {
            const swiper = this;
            if (!swiper.enabled) return;
            swiper.enabled = false;

            if (swiper.params.grabCursor) {
                swiper.unsetGrabCursor();
            }

            swiper.emit('disable');
        }

        setProgress(progress, speed) {
            const swiper = this;
            progress = Math.min(Math.max(progress, 0), 1);
            const min = swiper.minTranslate();
            const max = swiper.maxTranslate();
            const current = (max - min) * progress + min;
            swiper.translateTo(current, typeof speed === 'undefined' ? 0 : speed);
            swiper.updateActiveIndex();
            swiper.updateSlidesClasses();
        }

        emitContainerClasses() {
            const swiper = this;
            if (!swiper.params._emitClasses || !swiper.el) return;
            const cls = swiper.el.className.split(' ').filter(className => {
                return className.indexOf('ax') === 0 || className.indexOf(swiper.params.containerModifierClass) === 0;
            });
            swiper.emit('_containerClasses', cls.join(' '));
        }

        getSlideClasses(slideEl) {
            const swiper = this;
            if (swiper.destroyed) return '';
            return slideEl.className.split(' ').filter(className => {
                return className.indexOf('ax-slide') === 0 || className.indexOf(swiper.params.slideClass) === 0;
            }).join(' ');
        }

        emitSlidesClasses() {
            const swiper = this;
            if (!swiper.params._emitClasses || !swiper.el) return;
            const updates = [];
            swiper.slides.each(slideEl => {
                const classNames = swiper.getSlideClasses(slideEl);
                updates.push({
                    slideEl,
                    classNames
                });
                swiper.emit('_slideClass', slideEl, classNames);
            });
            swiper.emit('_slideClasses', updates);
        }

        slidesPerViewDynamic(view, exact) {
            if (view === void 0) {
                view = 'current';
            }

            if (exact === void 0) {
                exact = false;
            }

            const swiper = this;
            const {
                params,
                slides,
                slidesGrid,
                slidesSizesGrid,
                size: swiperSize,
                activeIndex
            } = swiper;
            let spv = 1;

            if (params.centeredSlides) {
                let slideSize = slides[activeIndex].swiperSlideSize;
                let breakLoop;

                for (let i = activeIndex + 1; i < slides.length; i += 1) {
                    if (slides[i] && !breakLoop) {
                        slideSize += slides[i].swiperSlideSize;
                        spv += 1;
                        if (slideSize > swiperSize) breakLoop = true;
                    }
                }

                for (let i = activeIndex - 1; i >= 0; i -= 1) {
                    if (slides[i] && !breakLoop) {
                        slideSize += slides[i].swiperSlideSize;
                        spv += 1;
                        if (slideSize > swiperSize) breakLoop = true;
                    }
                }
            } else {
                // eslint-disable-next-line
                if (view === 'current') {
                    for (let i = activeIndex + 1; i < slides.length; i += 1) {
                        const slideInView = exact ? slidesGrid[i] + slidesSizesGrid[i] - slidesGrid[activeIndex] < swiperSize : slidesGrid[i] - slidesGrid[activeIndex] < swiperSize;

                        if (slideInView) {
                            spv += 1;
                        }
                    }
                } else {
                    // previous
                    for (let i = activeIndex - 1; i >= 0; i -= 1) {
                        const slideInView = slidesGrid[activeIndex] - slidesGrid[i] < swiperSize;

                        if (slideInView) {
                            spv += 1;
                        }
                    }
                }
            }

            return spv;
        }

        update() {
            const swiper = this;
            if (!swiper || swiper.destroyed) return;
            const {
                snapGrid,
                params
            } = swiper; // Breakpoints

            if (params.breakpoints) {
                swiper.setBreakpoint();
            }

            swiper.updateSize();
            swiper.updateSlides();
            swiper.updateProgress();
            swiper.updateSlidesClasses();

            function setTranslate() {
                const translateValue = swiper.rtlTranslate ? swiper.translate * -1 : swiper.translate;
                const newTranslate = Math.min(Math.max(translateValue, swiper.maxTranslate()), swiper.minTranslate());
                swiper.setTranslate(newTranslate);
                swiper.updateActiveIndex();
                swiper.updateSlidesClasses();
            }

            let translated;

            if (swiper.params.freeMode && swiper.params.freeMode.enabled) {
                setTranslate();

                if (swiper.params.autoHeight) {
                    swiper.updateAutoHeight();
                }
            } else {
                if ((swiper.params.slidesPerView === 'auto' || swiper.params.slidesPerView > 1) && swiper.isEnd && !swiper.params.centeredSlides) {
                    translated = swiper.slideTo(swiper.slides.length - 1, 0, false, true);
                } else {
                    translated = swiper.slideTo(swiper.activeIndex, 0, false, true);
                }

                if (!translated) {
                    setTranslate();
                }
            }

            if (params.watchOverflow && snapGrid !== swiper.snapGrid) {
                swiper.checkOverflow();
            }

            swiper.emit('update');
        }

        changeDirection(newDirection, needUpdate) {
            if (needUpdate === void 0) {
                needUpdate = true;
            }

            const swiper = this;
            const currentDirection = swiper.params.direction;

            if (!newDirection) {
                // eslint-disable-next-line
                newDirection = currentDirection === 'horizontal' ? 'vertical' : 'horizontal';
            }

            if (newDirection === currentDirection || newDirection !== 'horizontal' && newDirection !== 'vertical') {
                return swiper;
            }

            swiper.$el.removeClass(`${swiper.params.containerModifierClass}${currentDirection}`).addClass(`${swiper.params.containerModifierClass}${newDirection}`);
            swiper.emitContainerClasses();
            swiper.params.direction = newDirection;
            swiper.slides.each(slideEl => {
                if (newDirection === 'vertical') {
                    slideEl.style.width = '';
                } else {
                    slideEl.style.height = '';
                }
            });
            swiper.emit('changeDirection');
            if (needUpdate) swiper.update();
            return swiper;
        }

        mount(el) {
            const swiper = this;
            if (swiper.mounted) return true; // Find el

            const $el = $(el || swiper.params.el);
            el = $el[0];

            if (!el) {
                return false;
            }

            el.swiper = swiper;

            const getWrapperSelector = () => {
                return `.${(swiper.params.wrapperClass || '').trim().split(' ').join('.')}`;
            };

            const getWrapper = () => {
                if (el && el.shadowRoot && el.shadowRoot.querySelector) {
                    const res = $(el.shadowRoot.querySelector(getWrapperSelector())); // Children needs to return slot items

                    res.children = options => $el.children(options);

                    return res;
                }

                return $el.children(getWrapperSelector());
            }; // Find Wrapper


            let $wrapperEl = getWrapper();

            if ($wrapperEl.length === 0 && swiper.params.createElements) {
                const document = getDocument();
                const wrapper = document.createElement('div');
                $wrapperEl = $(wrapper);
                wrapper.className = swiper.params.wrapperClass;
                $el.append(wrapper);
                $el.children(`.${swiper.params.slideClass}`).each(slideEl => {
                    $wrapperEl.append(slideEl);
                });
            }

            Object.assign(swiper, {
                $el,
                el,
                $wrapperEl,
                wrapperEl: $wrapperEl[0],
                mounted: true,
                // RTL
                rtl: el.dir.toLowerCase() === 'rtl' || $el.css('direction') === 'rtl',
                rtlTranslate: swiper.params.direction === 'horizontal' && (el.dir.toLowerCase() === 'rtl' || $el.css('direction') === 'rtl'),
                wrongRTL: $wrapperEl.css('display') === '-webkit-box'
            });
            return true;
        }

        init(el) {
            const swiper = this;
            if (swiper.initialized) return swiper;
            const mounted = swiper.mount(el);
            if (mounted === false) return swiper;
            swiper.emit('beforeInit'); // Set breakpoint

            if (swiper.params.breakpoints) {
                swiper.setBreakpoint();
            } // Add Classes


            swiper.addClasses(); // Create loop

            if (swiper.params.loop) {
                swiper.loopCreate();
            } // Update size


            swiper.updateSize(); // Update slides

            swiper.updateSlides();

            if (swiper.params.watchOverflow) {
                swiper.checkOverflow();
            } // Set Grab Cursor


            if (swiper.params.grabCursor && swiper.enabled) {
                swiper.setGrabCursor();
            }

            if (swiper.params.preloadImages) {
                swiper.preloadImages();
            } // Slide To Initial Slide


            if (swiper.params.loop) {
                swiper.slideTo(swiper.params.initialSlide + swiper.loopedSlides, 0, swiper.params.runCallbacksOnInit, false, true);
            } else {
                swiper.slideTo(swiper.params.initialSlide, 0, swiper.params.runCallbacksOnInit, false, true);
            } // Attach events

            swiper.attachEvents(); // Init Flag

            swiper.initialized = true; // Emit

            swiper.emit('init');
            swiper.emit('afterInit');
            return swiper;
        }

        destroy(deleteInstance, cleanStyles) {
            if (deleteInstance === void 0) {
                deleteInstance = true;
            }

            if (cleanStyles === void 0) {
                cleanStyles = true;
            }

            const swiper = this;
            const {
                params,
                $el,
                $wrapperEl,
                slides
            } = swiper;

            if (typeof swiper.params === 'undefined' || swiper.destroyed) {
                return null;
            }

            swiper.emit('beforeDestroy'); // Init Flag

            swiper.initialized = false; // Detach events

            swiper.detachEvents(); // Destroy loop

            if (params.loop) {
                swiper.loopDestroy();
            } // Cleanup styles


            if (cleanStyles) {
                swiper.removeClasses();
                $el.removeAttr('style');
                $wrapperEl.removeAttr('style');

                if (slides && slides.length) {
                    slides.removeClass([params.slideVisibleClass, params.slideActiveClass, params.slideNextClass, params.slidePrevClass].join(' ')).removeAttr('style').removeAttr('data-swiper-slide-index');
                }
            }

            swiper.emit('destroy'); // Detach emitter events

            Object.keys(swiper.eventsListeners).forEach(eventName => {
                swiper.off(eventName);
            });

            if (deleteInstance !== false) {
                swiper.$el[0].swiper = null;
                deleteProps(swiper);
            }

            swiper.destroyed = true;
            return null;
        }

        static extendDefaults(newDefaults) {
            extend(extendedDefaults, newDefaults);
        }

        static get extendedDefaults() {
            return extendedDefaults;
        }

        static get defaults() {
            return defaults;
        }

        static installModule(mod) {
            if (!Swiper.prototype.__modules__) Swiper.prototype.__modules__ = [];
            const modules = Swiper.prototype.__modules__;

            if (typeof mod === 'function' && modules.indexOf(mod) < 0) {
                modules.push(mod);
            }
        }

        static use(module) {
            if (Array.isArray(module)) {
                module.forEach(m => Swiper.installModule(m));
                return Swiper;
            }

            Swiper.installModule(module);
            return Swiper;
        }

    }

    Object.keys(prototypes).forEach(prototypeGroup => {
        Object.keys(prototypes[prototypeGroup]).forEach(protoMethod => {
            Swiper.prototype[protoMethod] = prototypes[prototypeGroup][protoMethod];
        });
    });
    Swiper.use([Resize, Observer]);

    function Virtual(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        extendParams({
            virtual: {
                enabled: false,
                slides: [],
                cache: true,
                renderSlide: null,
                renderExternal: null,
                renderExternalUpdate: true,
                addSlidesBefore: 0,
                addSlidesAfter: 0
            }
        });
        let cssModeTimeout;
        swiper.virtual = {
            cache: {},
            from: undefined,
            to: undefined,
            slides: [],
            offset: 0,
            slidesGrid: []
        };

        function renderSlide(slide, index) {
            const params = swiper.params.virtual;

            if (params.cache && swiper.virtual.cache[index]) {
                return swiper.virtual.cache[index];
            }

            const $slideEl = params.renderSlide ? $(params.renderSlide.call(swiper, slide, index)) : $(`<div class="${swiper.params.slideClass}" data-swiper-slide-index="${index}">${slide}</div>`);
            if (!$slideEl.attr('data-swiper-slide-index')) $slideEl.attr('data-swiper-slide-index', index);
            if (params.cache) swiper.virtual.cache[index] = $slideEl;
            return $slideEl;
        }

        function update(force) {
            const {
                slidesPerView,
                slidesPerGroup,
                centeredSlides
            } = swiper.params;
            const {
                addSlidesBefore,
                addSlidesAfter
            } = swiper.params.virtual;
            const {
                from: previousFrom,
                to: previousTo,
                slides,
                slidesGrid: previousSlidesGrid,
                offset: previousOffset
            } = swiper.virtual;

            if (!swiper.params.cssMode) {
                swiper.updateActiveIndex();
            }

            const activeIndex = swiper.activeIndex || 0;
            let offsetProp;
            if (swiper.rtlTranslate) offsetProp = 'right'; else offsetProp = swiper.isHorizontal() ? 'left' : 'top';
            let slidesAfter;
            let slidesBefore;

            if (centeredSlides) {
                slidesAfter = Math.floor(slidesPerView / 2) + slidesPerGroup + addSlidesAfter;
                slidesBefore = Math.floor(slidesPerView / 2) + slidesPerGroup + addSlidesBefore;
            } else {
                slidesAfter = slidesPerView + (slidesPerGroup - 1) + addSlidesAfter;
                slidesBefore = slidesPerGroup + addSlidesBefore;
            }

            const from = Math.max((activeIndex || 0) - slidesBefore, 0);
            const to = Math.min((activeIndex || 0) + slidesAfter, slides.length - 1);
            const offset = (swiper.slidesGrid[from] || 0) - (swiper.slidesGrid[0] || 0);
            Object.assign(swiper.virtual, {
                from,
                to,
                offset,
                slidesGrid: swiper.slidesGrid
            });

            function onRendered() {
                swiper.updateSlides();
                swiper.updateProgress();
                swiper.updateSlidesClasses();

                if (swiper.lazy && swiper.params.lazy.enabled) {
                    swiper.lazy.load();
                }

                emit('virtualUpdate');
            }

            if (previousFrom === from && previousTo === to && !force) {
                if (swiper.slidesGrid !== previousSlidesGrid && offset !== previousOffset) {
                    swiper.slides.css(offsetProp, `${offset}px`);
                }

                swiper.updateProgress();
                emit('virtualUpdate');
                return;
            }

            if (swiper.params.virtual.renderExternal) {
                swiper.params.virtual.renderExternal.call(swiper, {
                    offset,
                    from,
                    to,
                    slides: function getSlides() {
                        const slidesToRender = [];

                        for (let i = from; i <= to; i += 1) {
                            slidesToRender.push(slides[i]);
                        }

                        return slidesToRender;
                    }()
                });

                if (swiper.params.virtual.renderExternalUpdate) {
                    onRendered();
                } else {
                    emit('virtualUpdate');
                }

                return;
            }

            const prependIndexes = [];
            const appendIndexes = [];

            if (force) {
                swiper.$wrapperEl.find(`.${swiper.params.slideClass}`).remove();
            } else {
                for (let i = previousFrom; i <= previousTo; i += 1) {
                    if (i < from || i > to) {
                        swiper.$wrapperEl.find(`.${swiper.params.slideClass}[data-swiper-slide-index="${i}"]`).remove();
                    }
                }
            }

            for (let i = 0; i < slides.length; i += 1) {
                if (i >= from && i <= to) {
                    if (typeof previousTo === 'undefined' || force) {
                        appendIndexes.push(i);
                    } else {
                        if (i > previousTo) appendIndexes.push(i);
                        if (i < previousFrom) prependIndexes.push(i);
                    }
                }
            }

            appendIndexes.forEach(index => {
                swiper.$wrapperEl.append(renderSlide(slides[index], index));
            });
            prependIndexes.sort((a, b) => b - a).forEach(index => {
                swiper.$wrapperEl.prepend(renderSlide(slides[index], index));
            });
            swiper.$wrapperEl.children('.swiper-slide').css(offsetProp, `${offset}px`);
            onRendered();
        }

        function appendSlide(slides) {
            if (typeof slides === 'object' && 'length' in slides) {
                for (let i = 0; i < slides.length; i += 1) {
                    if (slides[i]) swiper.virtual.slides.push(slides[i]);
                }
            } else {
                swiper.virtual.slides.push(slides);
            }

            update(true);
        }

        function prependSlide(slides) {
            const activeIndex = swiper.activeIndex;
            let newActiveIndex = activeIndex + 1;
            let numberOfNewSlides = 1;

            if (Array.isArray(slides)) {
                for (let i = 0; i < slides.length; i += 1) {
                    if (slides[i]) swiper.virtual.slides.unshift(slides[i]);
                }

                newActiveIndex = activeIndex + slides.length;
                numberOfNewSlides = slides.length;
            } else {
                swiper.virtual.slides.unshift(slides);
            }

            if (swiper.params.virtual.cache) {
                const cache = swiper.virtual.cache;
                const newCache = {};
                Object.keys(cache).forEach(cachedIndex => {
                    const $cachedEl = cache[cachedIndex];
                    const cachedElIndex = $cachedEl.attr('data-swiper-slide-index');

                    if (cachedElIndex) {
                        $cachedEl.attr('data-swiper-slide-index', parseInt(cachedElIndex, 10) + numberOfNewSlides);
                    }

                    newCache[parseInt(cachedIndex, 10) + numberOfNewSlides] = $cachedEl;
                });
                swiper.virtual.cache = newCache;
            }

            update(true);
            swiper.slideTo(newActiveIndex, 0);
        }

        function removeSlide(slidesIndexes) {
            if (typeof slidesIndexes === 'undefined' || slidesIndexes === null) return;
            let activeIndex = swiper.activeIndex;

            if (Array.isArray(slidesIndexes)) {
                for (let i = slidesIndexes.length - 1; i >= 0; i -= 1) {
                    swiper.virtual.slides.splice(slidesIndexes[i], 1);

                    if (swiper.params.virtual.cache) {
                        delete swiper.virtual.cache[slidesIndexes[i]];
                    }

                    if (slidesIndexes[i] < activeIndex) activeIndex -= 1;
                    activeIndex = Math.max(activeIndex, 0);
                }
            } else {
                swiper.virtual.slides.splice(slidesIndexes, 1);

                if (swiper.params.virtual.cache) {
                    delete swiper.virtual.cache[slidesIndexes];
                }

                if (slidesIndexes < activeIndex) activeIndex -= 1;
                activeIndex = Math.max(activeIndex, 0);
            }

            update(true);
            swiper.slideTo(activeIndex, 0);
        }

        function removeAllSlides() {
            swiper.virtual.slides = [];

            if (swiper.params.virtual.cache) {
                swiper.virtual.cache = {};
            }

            update(true);
            swiper.slideTo(0, 0);
        }

        on('beforeInit', () => {
            if (!swiper.params.virtual.enabled) return;
            swiper.virtual.slides = swiper.params.virtual.slides;
            swiper.classNames.push(`${swiper.params.containerModifierClass}virtual`);
            swiper.params.watchSlidesProgress = true;
            swiper.originalParams.watchSlidesProgress = true;

            if (!swiper.params.initialSlide) {
                update();
            }
        });
        on('setTranslate', () => {
            if (!swiper.params.virtual.enabled) return;

            if (swiper.params.cssMode && !swiper._immediateVirtual) {
                clearTimeout(cssModeTimeout);
                cssModeTimeout = setTimeout(() => {
                    update();
                }, 100);
            } else {
                update();
            }
        });
        on('init update resize', () => {
            if (!swiper.params.virtual.enabled) return;

            if (swiper.params.cssMode) {
                setCSSProperty(swiper.wrapperEl, '--swiper-virtual-size', `${swiper.virtualSize}px`);
            }
        });
        Object.assign(swiper.virtual, {
            appendSlide,
            prependSlide,
            removeSlide,
            removeAllSlides,
            update
        });
    }

    /* eslint-disable consistent-return */
    function Keyboard(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        const document = getDocument();
        const window = getWindow();
        swiper.keyboard = {
            enabled: false
        };
        extendParams({
            keyboard: {
                enabled: false,
                onlyInViewport: true,
                pageUpDown: true
            }
        });

        function handle(event) {
            if (!swiper.enabled) return;
            const {
                rtlTranslate: rtl
            } = swiper;
            let e = event;
            if (e.originalEvent) e = e.originalEvent; // jquery fix

            const kc = e.keyCode || e.charCode;
            const pageUpDown = swiper.params.keyboard.pageUpDown;
            const isPageUp = pageUpDown && kc === 33;
            const isPageDown = pageUpDown && kc === 34;
            const isArrowLeft = kc === 37;
            const isArrowRight = kc === 39;
            const isArrowUp = kc === 38;
            const isArrowDown = kc === 40; // Directions locks

            if (!swiper.allowSlideNext && (swiper.isHorizontal() && isArrowRight || swiper.isVertical() && isArrowDown || isPageDown)) {
                return false;
            }

            if (!swiper.allowSlidePrev && (swiper.isHorizontal() && isArrowLeft || swiper.isVertical() && isArrowUp || isPageUp)) {
                return false;
            }

            if (e.shiftKey || e.altKey || e.ctrlKey || e.metaKey) {
                return undefined;
            }

            if (document.activeElement && document.activeElement.nodeName && (document.activeElement.nodeName.toLowerCase() === 'input' || document.activeElement.nodeName.toLowerCase() === 'textarea')) {
                return undefined;
            }

            if (swiper.params.keyboard.onlyInViewport && (isPageUp || isPageDown || isArrowLeft || isArrowRight || isArrowUp || isArrowDown)) {
                let inView = false; // Check that swiper should be inside of visible area of window

                if (swiper.$el.parents(`.${swiper.params.slideClass}`).length > 0 && swiper.$el.parents(`.${swiper.params.slideActiveClass}`).length === 0) {
                    return undefined;
                }

                const $el = swiper.$el;
                const swiperWidth = $el[0].clientWidth;
                const swiperHeight = $el[0].clientHeight;
                const windowWidth = window.innerWidth;
                const windowHeight = window.innerHeight;
                const swiperOffset = swiper.$el.offset();
                if (rtl) swiperOffset.left -= swiper.$el[0].scrollLeft;
                const swiperCoord = [[swiperOffset.left, swiperOffset.top], [swiperOffset.left + swiperWidth, swiperOffset.top], [swiperOffset.left, swiperOffset.top + swiperHeight], [swiperOffset.left + swiperWidth, swiperOffset.top + swiperHeight]];

                for (let i = 0; i < swiperCoord.length; i += 1) {
                    const point = swiperCoord[i];

                    if (point[0] >= 0 && point[0] <= windowWidth && point[1] >= 0 && point[1] <= windowHeight) {
                        if (point[0] === 0 && point[1] === 0) continue; // eslint-disable-line

                        inView = true;
                    }
                }

                if (!inView) return undefined;
            }

            if (swiper.isHorizontal()) {
                if (isPageUp || isPageDown || isArrowLeft || isArrowRight) {
                    axPreventDefault(e);//AXUI
                }

                if ((isPageDown || isArrowRight) && !rtl || (isPageUp || isArrowLeft) && rtl) swiper.slideNext();
                if ((isPageUp || isArrowLeft) && !rtl || (isPageDown || isArrowRight) && rtl) swiper.slidePrev();
            } else {
                if (isPageUp || isPageDown || isArrowUp || isArrowDown) {
                    axPreventDefault(e);//AXUI
                }

                if (isPageDown || isArrowDown) swiper.slideNext();
                if (isPageUp || isArrowUp) swiper.slidePrev();
            }

            emit('keyPress', kc);
            return undefined;
        }

        function enable() {
            if (swiper.keyboard.enabled) return;
            $(document).on('keydown', handle);
            swiper.keyboard.enabled = true;
        }

        function disable() {
            if (!swiper.keyboard.enabled) return;
            $(document).off('keydown', handle);
            swiper.keyboard.enabled = false;
        }

        on('init', () => {
            if (swiper.params.keyboard.enabled) {
                enable();
            }
        });
        on('destroy', () => {
            if (swiper.keyboard.enabled) {
                disable();
            }
        });
        Object.assign(swiper.keyboard, {
            enable,
            disable
        });
    }

    /* eslint-disable consistent-return */
    function Mousewheel(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        const window = getWindow();
        extendParams({
            mousewheel: {
                enabled: false,
                releaseOnEdges: false,
                invert: false,
                forceToAxis: false,
                sensitivity: 1,
                eventsTarget: 'container',
                thresholdDelta: null,
                thresholdTime: null
            }
        });
        swiper.mousewheel = {
            enabled: false
        };
        let timeout;
        let lastScrollTime = now();
        let lastEventBeforeSnap;
        const recentWheelEvents = [];

        function normalize(e) {
            // Reasonable defaults
            const PIXEL_STEP = 10;
            const LINE_HEIGHT = 40;
            const PAGE_HEIGHT = 800;
            let sX = 0;
            let sY = 0; // spinX, spinY

            let pX = 0;
            let pY = 0; // pixelX, pixelY
            // Legacy

            if ('detail' in e) {
                sY = e.detail;
            }

            if ('wheelDelta' in e) {
                sY = -e.wheelDelta / 120;
            }

            if ('wheelDeltaY' in e) {
                sY = -e.wheelDeltaY / 120;
            }

            if ('wheelDeltaX' in e) {
                sX = -e.wheelDeltaX / 120;
            } // side scrolling on FF with DOMMouseScroll


            if ('axis' in e && e.axis === e.HORIZONTAL_AXIS) {
                sX = sY;
                sY = 0;
            }

            pX = sX * PIXEL_STEP;
            pY = sY * PIXEL_STEP;

            if ('deltaY' in e) {
                pY = e.deltaY;
            }

            if ('deltaX' in e) {
                pX = e.deltaX;
            }

            if (e.shiftKey && !pX) {
                // if user scrolls with shift he wants horizontal scroll
                pX = pY;
                pY = 0;
            }

            if ((pX || pY) && e.deltaMode) {
                if (e.deltaMode === 1) {
                    // delta in LINE units
                    pX *= LINE_HEIGHT;
                    pY *= LINE_HEIGHT;
                } else {
                    // delta in PAGE units
                    pX *= PAGE_HEIGHT;
                    pY *= PAGE_HEIGHT;
                }
            } // Fall-back if spin cannot be determined


            if (pX && !sX) {
                sX = pX < 1 ? -1 : 1;
            }

            if (pY && !sY) {
                sY = pY < 1 ? -1 : 1;
            }

            return {
                spinX: sX,
                spinY: sY,
                pixelX: pX,
                pixelY: pY
            };
        }

        function handleMouseEnter() {
            if (!swiper.enabled) return;
            swiper.mouseEntered = true;
        }

        function handleMouseLeave() {
            if (!swiper.enabled) return;
            swiper.mouseEntered = false;
        }

        function animateSlider(newEvent) {
            if (swiper.params.mousewheel.thresholdDelta && newEvent.delta < swiper.params.mousewheel.thresholdDelta) {
                // Prevent if delta of wheel scroll delta is below configured threshold
                return false;
            }

            if (swiper.params.mousewheel.thresholdTime && now() - lastScrollTime < swiper.params.mousewheel.thresholdTime) {
                // Prevent if time between scrolls is below configured threshold
                return false;
            } // If the movement is NOT big enough and
            // if the last time the user scrolled was too close to the current one (avoid continuously triggering the slider):
            //   Don't go any further (avoid insignificant scroll movement).


            if (newEvent.delta >= 6 && now() - lastScrollTime < 60) {
                // Return false as a default
                return true;
            } // If user is scrolling towards the end:
            //   If the slider hasn't hit the latest slide or
            //   if the slider is a loop and
            //   if the slider isn't moving right now:
            //     Go to next slide and
            //     emit a scroll event.
            // Else (the user is scrolling towards the beginning) and
            // if the slider hasn't hit the first slide or
            // if the slider is a loop and
            // if the slider isn't moving right now:
            //   Go to prev slide and
            //   emit a scroll event.


            if (newEvent.direction < 0) {
                if ((!swiper.isEnd || swiper.params.loop) && !swiper.animating) {
                    swiper.slideNext();
                    emit('scroll', newEvent.raw);
                }
            } else if ((!swiper.isBeginning || swiper.params.loop) && !swiper.animating) {
                swiper.slidePrev();
                emit('scroll', newEvent.raw);
            } // If you got here is because an animation has been triggered so store the current time


            lastScrollTime = new window.Date().getTime(); // Return false as a default

            return false;
        }

        function releaseScroll(newEvent) {
            const params = swiper.params.mousewheel;

            if (newEvent.direction < 0) {
                if (swiper.isEnd && !swiper.params.loop && params.releaseOnEdges) {
                    // Return true to animate scroll on edges
                    return true;
                }
            } else if (swiper.isBeginning && !swiper.params.loop && params.releaseOnEdges) {
                // Return true to animate scroll on edges
                return true;
            }

            return false;
        }

        function handle(event) {
            let e = event;
            let disableParentSwiper = true;
            if (!swiper.enabled) return;
            const params = swiper.params.mousewheel;

            if (swiper.params.cssMode) {
                axPreventDefault(e);//AXUI
            }

            let target = swiper.$el;

            if (swiper.params.mousewheel.eventsTarget !== 'container') {
                target = $(swiper.params.mousewheel.eventsTarget);
            }

            if (!swiper.mouseEntered && !target[0].contains(e.target) && !params.releaseOnEdges) return true;
            if (e.originalEvent) e = e.originalEvent; // jquery fix

            let delta = 0;
            const rtlFactor = swiper.rtlTranslate ? -1 : 1;
            const data = normalize(e);

            if (params.forceToAxis) {
                if (swiper.isHorizontal()) {
                    if (Math.abs(data.pixelX) > Math.abs(data.pixelY)) delta = -data.pixelX * rtlFactor; else return true;
                } else if (Math.abs(data.pixelY) > Math.abs(data.pixelX)) delta = -data.pixelY; else return true;
            } else {
                delta = Math.abs(data.pixelX) > Math.abs(data.pixelY) ? -data.pixelX * rtlFactor : -data.pixelY;
            }

            if (delta === 0) return true;
            if (params.invert) delta = -delta; // Get the scroll positions

            let positions = swiper.getTranslate() + delta * params.sensitivity;
            if (positions >= swiper.minTranslate()) positions = swiper.minTranslate();
            if (positions <= swiper.maxTranslate()) positions = swiper.maxTranslate(); // When loop is true:
            //     the disableParentSwiper will be true.
            // When loop is false:
            //     if the scroll positions is not on edge,
            //     then the disableParentSwiper will be true.
            //     if the scroll on edge positions,
            //     then the disableParentSwiper will be false.

            disableParentSwiper = swiper.params.loop ? true : !(positions === swiper.minTranslate() || positions === swiper.maxTranslate());
            if (disableParentSwiper && swiper.params.nested) e.stopPropagation();

            if (!swiper.params.freeMode || !swiper.params.freeMode.enabled) {
                // Register the new event in a variable which stores the relevant data
                const newEvent = {
                    time: now(),
                    delta: Math.abs(delta),
                    direction: Math.sign(delta),
                    raw: event
                }; // Keep the most recent events

                if (recentWheelEvents.length >= 2) {
                    recentWheelEvents.shift(); // only store the last N events
                }

                const prevEvent = recentWheelEvents.length ? recentWheelEvents[recentWheelEvents.length - 1] : undefined;
                recentWheelEvents.push(newEvent); // If there is at least one previous recorded event:
                //   If direction has changed or
                //   if the scroll is quicker than the previous one:
                //     Animate the slider.
                // Else (this is the first time the wheel is moved):
                //     Animate the slider.

                if (prevEvent) {
                    if (newEvent.direction !== prevEvent.direction || newEvent.delta > prevEvent.delta || newEvent.time > prevEvent.time + 150) {
                        animateSlider(newEvent);
                    }
                } else {
                    animateSlider(newEvent);
                } // If it's time to release the scroll:
                //   Return now so you don't hit the preventDefault.


                if (releaseScroll(newEvent)) {
                    return true;
                }
            } else {
                // Freemode or scrollContainer:
                // If we recently snapped after a momentum scroll, then ignore wheel events
                // to give time for the deceleration to finish. Stop ignoring after 500 msecs
                // or if it's a new scroll (larger delta or inverse sign as last event before
                // an end-of-momentum snap).
                const newEvent = {
                    time: now(),
                    delta: Math.abs(delta),
                    direction: Math.sign(delta)
                };
                const ignoreWheelEvents = lastEventBeforeSnap && newEvent.time < lastEventBeforeSnap.time + 500 && newEvent.delta <= lastEventBeforeSnap.delta && newEvent.direction === lastEventBeforeSnap.direction;

                if (!ignoreWheelEvents) {
                    lastEventBeforeSnap = undefined;

                    if (swiper.params.loop) {
                        swiper.loopFix();
                    }

                    let position = swiper.getTranslate() + delta * params.sensitivity;
                    const wasBeginning = swiper.isBeginning;
                    const wasEnd = swiper.isEnd;
                    if (position >= swiper.minTranslate()) position = swiper.minTranslate();
                    if (position <= swiper.maxTranslate()) position = swiper.maxTranslate();
                    swiper.setTransition(0);
                    swiper.setTranslate(position);
                    swiper.updateProgress();
                    swiper.updateActiveIndex();
                    swiper.updateSlidesClasses();

                    if (!wasBeginning && swiper.isBeginning || !wasEnd && swiper.isEnd) {
                        swiper.updateSlidesClasses();
                    }

                    if (swiper.params.freeMode.sticky) {
                        // When wheel scrolling starts with sticky (aka snap) enabled, then detect
                        // the end of a momentum scroll by storing recent (N=15?) wheel events.
                        // 1. do all N events have decreasing or same (absolute value) delta?
                        // 2. did all N events arrive in the last M (M=500?) msecs?
                        // 3. does the earliest event have an (absolute value) delta that's
                        //    at least P (P=1?) larger than the most recent event's delta?
                        // 4. does the latest event have a delta that's smaller than Q (Q=6?) pixels?
                        // If 1-4 are "yes" then we're near the end of a momentum scroll deceleration.
                        // Snap immediately and ignore remaining wheel events in this scroll.
                        // See comment above for "remaining wheel events in this scroll" determination.
                        // If 1-4 aren't satisfied, then wait to snap until 500ms after the last event.
                        clearTimeout(timeout);
                        timeout = undefined;

                        if (recentWheelEvents.length >= 15) {
                            recentWheelEvents.shift(); // only store the last N events
                        }

                        const prevEvent = recentWheelEvents.length ? recentWheelEvents[recentWheelEvents.length - 1] : undefined;
                        const firstEvent = recentWheelEvents[0];
                        recentWheelEvents.push(newEvent);

                        if (prevEvent && (newEvent.delta > prevEvent.delta || newEvent.direction !== prevEvent.direction)) {
                            // Increasing or reverse-sign delta means the user started scrolling again. Clear the wheel event log.
                            recentWheelEvents.splice(0);
                        } else if (recentWheelEvents.length >= 15 && newEvent.time - firstEvent.time < 500 && firstEvent.delta - newEvent.delta >= 1 && newEvent.delta <= 6) {
                            // We're at the end of the deceleration of a momentum scroll, so there's no need
                            // to wait for more events. Snap ASAP on the next tick.
                            // Also, because there's some remaining momentum we'll bias the snap in the
                            // direction of the ongoing scroll because it's better UX for the scroll to snap
                            // in the same direction as the scroll instead of reversing to snap.  Therefore,
                            // if it's already scrolled more than 20% in the current direction, keep going.
                            const snapToThreshold = delta > 0 ? 0.8 : 0.2;
                            lastEventBeforeSnap = newEvent;
                            recentWheelEvents.splice(0);
                            timeout = nextTick(() => {
                                swiper.slideToClosest(swiper.params.speed, true, undefined, snapToThreshold);
                            }, 0); // no delay; move on next tick
                        }

                        if (!timeout) {
                            // if we get here, then we haven't detected the end of a momentum scroll, so
                            // we'll consider a scroll "complete" when there haven't been any wheel events
                            // for 500ms.
                            timeout = nextTick(() => {
                                const snapToThreshold = 0.5;
                                lastEventBeforeSnap = newEvent;
                                recentWheelEvents.splice(0);
                                swiper.slideToClosest(swiper.params.speed, true, undefined, snapToThreshold);
                            }, 500);
                        }
                    } // Emit event


                    if (!ignoreWheelEvents) emit('scroll', e); // Stop autoplay

                    if (swiper.params.autoplay && swiper.params.autoplayDisableOnInteraction) swiper.autoplay.stop(); // Return page scroll on edge positions

                    if (position === swiper.minTranslate() || position === swiper.maxTranslate()) return true;
                }
            }

            axPreventDefault(e);//AXUI
            return false;
        }

        function events(method) {
            let target = swiper.$el;

            if (swiper.params.mousewheel.eventsTarget !== 'container') {
                target = $(swiper.params.mousewheel.eventsTarget);
            }

            target[method]('mouseenter', handleMouseEnter);
            target[method]('mouseleave', handleMouseLeave);
            target[method]('wheel', handle);
        }

        function enable() {
            if (swiper.params.cssMode) {
                swiper.wrapperEl.removeEventListener('wheel', handle);
                return true;
            }

            if (swiper.mousewheel.enabled) return false;
            events('on');
            swiper.mousewheel.enabled = true;
            return true;
        }

        function disable() {
            if (swiper.params.cssMode) {
                swiper.wrapperEl.addEventListener(event, handle);
                return true;
            }

            if (!swiper.mousewheel.enabled) return false;
            events('off');
            swiper.mousewheel.enabled = false;
            return true;
        }

        on('init', () => {
            if (!swiper.params.mousewheel.enabled && swiper.params.cssMode) {
                disable();
            }

            if (swiper.params.mousewheel.enabled) enable();
        });
        on('destroy', () => {
            if (swiper.params.cssMode) {
                enable();
            }

            if (swiper.mousewheel.enabled) disable();
        });
        Object.assign(swiper.mousewheel, {
            enable,
            disable
        });
    }

    function createElementIfNotDefined(swiper, originalParams, params, checkProps) {
        const document = getDocument();

        if (swiper.params.createElements) {
            Object.keys(checkProps).forEach(key => {
                if (!params[key] && params.auto === true) {
                    let element = swiper.$el.children(`.${checkProps[key]}`)[0];

                    if (!element) {
                        element = document.createElement('div');
                        element.className = checkProps[key];
                        swiper.$el.append(element);
                    }

                    params[key] = element;
                    originalParams[key] = element;
                }
            });
        }

        return params;
    }

    function Navigation(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;

        extendParams({
            navigation: {
                nextEl: '.ax-next',//AXUI,原null
                prevEl: '.ax-prev',//AXUI,原null
                hideOnClick: false,
                disabledClass: 'ax-disabled',
                hiddenClass: 'ax-hidden',
                lockClass: 'ax-lock'
            }
        });
        swiper.navigation = {
            nextEl: null,
            $nextEl: null,
            prevEl: null,
            $prevEl: null
        };

        function getEl(el) {
            let $el;

            if (el) {
                $el = $(el);

                if (swiper.params.uniqueNavElements && typeof el === 'string' && $el.length > 1 && swiper.$el.find(el).length === 1) {
                    $el = swiper.$el.find(el);
                }
            }

            return $el;
        }

        function toggleEl($el, disabled) {
            const params = swiper.params.navigation;

            if ($el && $el.length > 0) {
                $el[disabled ? 'addClass' : 'removeClass'](params.disabledClass);
                if ($el[0] && $el[0].tagName === 'BUTTON') $el[0].disabled = disabled;

                if (swiper.params.watchOverflow && swiper.enabled) {
                    $el[swiper.isLocked ? 'addClass' : 'removeClass'](params.lockClass);
                }
            }
        }

        function update() {
            // Update Navigation Buttons
            if (swiper.params.loop) return;
            const {
                $nextEl,
                $prevEl
            } = swiper.navigation;
            toggleEl($prevEl, swiper.isBeginning && !swiper.params.rewind);
            toggleEl($nextEl, swiper.isEnd && !swiper.params.rewind);
        }

        function onPrevClick(e) {
            axPreventDefault(e);//AXUI
            if (swiper.isBeginning && !swiper.params.loop && !swiper.params.rewind) return;
            swiper.slidePrev();
        }

        function onNextClick(e) {
            axPreventDefault(e);//AXUI
            if (swiper.isEnd && !swiper.params.loop && !swiper.params.rewind) return;
            swiper.slideNext();
        }

        function init() {
            const params = swiper.params.navigation;
            swiper.params.navigation = createElementIfNotDefined(swiper, swiper.originalParams.navigation, swiper.params.navigation, {
                nextEl: 'ax-next',//AXUI
                prevEl: 'ax-prev'//AXUI
            });
            if (!(params.nextEl || params.prevEl)) return;
            const $nextEl = getEl(params.nextEl);
            const $prevEl = getEl(params.prevEl);

            if ($nextEl && $nextEl.length > 0) {
                $nextEl.on('click', onNextClick);
            }

            if ($prevEl && $prevEl.length > 0) {
                $prevEl.on('click', onPrevClick);
            }

            Object.assign(swiper.navigation, {
                $nextEl,
                nextEl: $nextEl && $nextEl[0],
                $prevEl,
                prevEl: $prevEl && $prevEl[0]
            });

            if (!swiper.enabled) {
                if ($nextEl) $nextEl.addClass(params.lockClass);
                if ($prevEl) $prevEl.addClass(params.lockClass);
            }
        }

        function destroy() {
            const {
                $nextEl,
                $prevEl
            } = swiper.navigation;

            if ($nextEl && $nextEl.length) {
                $nextEl.off('click', onNextClick);
                $nextEl.removeClass(swiper.params.navigation.disabledClass);
            }

            if ($prevEl && $prevEl.length) {
                $prevEl.off('click', onPrevClick);
                $prevEl.removeClass(swiper.params.navigation.disabledClass);
            }
        }

        on('init', () => {
            init();
            update();
        });
        on('toEdge fromEdge lock unlock', () => {
            update();
        });
        on('destroy', () => {
            destroy();
        });
        on('enable disable', () => {
            const {
                $nextEl,
                $prevEl
            } = swiper.navigation;

            if ($nextEl) {
                $nextEl[swiper.enabled ? 'removeClass' : 'addClass'](swiper.params.navigation.lockClass);
            }

            if ($prevEl) {
                $prevEl[swiper.enabled ? 'removeClass' : 'addClass'](swiper.params.navigation.lockClass);
            }
        });
        on('click', (_s, e) => {
            const {
                $nextEl,
                $prevEl
            } = swiper.navigation;
            const targetEl = e.target;

            if (swiper.params.navigation.hideOnClick && !$(targetEl).is($prevEl) && !$(targetEl).is($nextEl)) {
                if (swiper.pagination && swiper.params.pagination && swiper.params.pagination.clickable && (swiper.pagination.el === targetEl || swiper.pagination.el.contains(targetEl))) return;
                let isHidden;

                if ($nextEl) {
                    isHidden = $nextEl.hasClass(swiper.params.navigation.hiddenClass);
                } else if ($prevEl) {
                    isHidden = $prevEl.hasClass(swiper.params.navigation.hiddenClass);
                }

                if (isHidden === true) {
                    emit('navigationShow');
                } else {
                    emit('navigationHide');
                }

                if ($nextEl) {
                    $nextEl.toggleClass(swiper.params.navigation.hiddenClass);
                }

                if ($prevEl) {
                    $prevEl.toggleClass(swiper.params.navigation.hiddenClass);
                }
            }
        });
        Object.assign(swiper.navigation, {
            update,
            init,
            destroy
        });
    }

    function classesToSelector(classes) {
        if (classes === void 0) {
            classes = '';
        }

        return `.${classes.trim().replace(/([\.:!\/])/g, '\\$1') // eslint-disable-line
            .replace(/ /g, '.')}`;
    }

    function Pagination(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        const pfx = 'ax-pages';
        extendParams({
            pagination: {
                el: '.' + pfx,//AXUI,原null
                bulletElement: 'i',//AXUI,原span
                clickable: true,//AXUI,原false
                hideOnClick: false,
                renderBullet: null,
                renderProgressbar: null,
                renderFraction: null,
                renderCustom: null,
                progressbarOpposite: false,
                type: 'bullets',
                // 'bullets' or 'progressbar' or 'fraction' or 'custom'
                dynamicBullets: false,
                dynamicMainBullets: 1,
                formatFractionCurrent: number => number,
                formatFractionTotal: number => number,
                bulletClass: `${pfx}-bullet`,
                bulletActiveClass: `${pfx}-bullet-active`,
                modifierClass: `${pfx}-`,
                currentClass: `${pfx}-current`,
                totalClass: `${pfx}-total`,
                hiddenClass: `${pfx}-hidden`,
                progressbarFillClass: `${pfx}-progressbar-fill`,
                progressbarOppositeClass: `${pfx}-progressbar-opposite`,
                clickableClass: `${pfx}-clickable`,
                lockClass: `${pfx}-lock`,
                horizontalClass: `${pfx}-horizontal`,
                verticalClass: `${pfx}-vertical`
            }
        });
        swiper.pagination = {
            el: null,
            $el: null,
            bullets: []
        };
        let bulletSize;
        let dynamicBulletIndex = 0;

        function isPaginationDisabled() {
            return !swiper.params.pagination.el || !swiper.pagination.el || !swiper.pagination.$el || swiper.pagination.$el.length === 0;
        }

        function setSideBullets($bulletEl, position) {
            const {
                bulletActiveClass
            } = swiper.params.pagination;
            $bulletEl[position]().addClass(`${bulletActiveClass}-${position}`)[position]().addClass(`${bulletActiveClass}-${position}-${position}`);
        }

        function update() {
            // Render || Update Pagination bullets/items
            const rtl = swiper.rtl;
            const params = swiper.params.pagination;
            if (isPaginationDisabled()) return;
            const slidesLength = swiper.virtual && swiper.params.virtual.enabled ? swiper.virtual.slides.length : swiper.slides.length;
            const $el = swiper.pagination.$el; // Current/Total

            let current;
            const total = swiper.params.loop ? Math.ceil((slidesLength - swiper.loopedSlides * 2) / swiper.params.slidesPerGroup) : swiper.snapGrid.length;

            if (swiper.params.loop) {
                current = Math.ceil((swiper.activeIndex - swiper.loopedSlides) / swiper.params.slidesPerGroup);

                if (current > slidesLength - 1 - swiper.loopedSlides * 2) {
                    current -= slidesLength - swiper.loopedSlides * 2;
                }

                if (current > total - 1) current -= total;
                if (current < 0 && swiper.params.paginationType !== 'bullets') current = total + current;
            } else if (typeof swiper.snapIndex !== 'undefined') {
                current = swiper.snapIndex;
            } else {
                current = swiper.activeIndex || 0;
            } // Types


            if (params.type === 'bullets' && swiper.pagination.bullets && swiper.pagination.bullets.length > 0) {
                const bullets = swiper.pagination.bullets;
                let firstIndex;
                let lastIndex;
                let midIndex;

                if (params.dynamicBullets) {
                    bulletSize = bullets.eq(0)[swiper.isHorizontal() ? 'outerWidth' : 'outerHeight'](true);
                    $el.css(swiper.isHorizontal() ? 'width' : 'height', `${bulletSize * (params.dynamicMainBullets + 4)}px`);

                    if (params.dynamicMainBullets > 1 && swiper.previousIndex !== undefined) {
                        dynamicBulletIndex += current - (swiper.previousIndex - swiper.loopedSlides || 0);

                        if (dynamicBulletIndex > params.dynamicMainBullets - 1) {
                            dynamicBulletIndex = params.dynamicMainBullets - 1;
                        } else if (dynamicBulletIndex < 0) {
                            dynamicBulletIndex = 0;
                        }
                    }

                    firstIndex = Math.max(current - dynamicBulletIndex, 0);
                    lastIndex = firstIndex + (Math.min(bullets.length, params.dynamicMainBullets) - 1);
                    midIndex = (lastIndex + firstIndex) / 2;
                }

                bullets.removeClass(['', '-next', '-next-next', '-prev', '-prev-prev', '-main'].map(suffix => `${params.bulletActiveClass}${suffix}`).join(' '));

                if ($el.length > 1) {
                    bullets.each(bullet => {
                        const $bullet = $(bullet);
                        const bulletIndex = $bullet.index();

                        if (bulletIndex === current) {
                            $bullet.addClass(params.bulletActiveClass);
                        }

                        if (params.dynamicBullets) {
                            if (bulletIndex >= firstIndex && bulletIndex <= lastIndex) {
                                $bullet.addClass(`${params.bulletActiveClass}-main`);
                            }

                            if (bulletIndex === firstIndex) {
                                setSideBullets($bullet, 'prev');
                            }

                            if (bulletIndex === lastIndex) {
                                setSideBullets($bullet, 'next');
                            }
                        }
                    });
                } else {
                    const $bullet = bullets.eq(current);
                    const bulletIndex = $bullet.index();
                    $bullet.addClass(params.bulletActiveClass);

                    if (params.dynamicBullets) {
                        const $firstDisplayedBullet = bullets.eq(firstIndex);
                        const $lastDisplayedBullet = bullets.eq(lastIndex);

                        for (let i = firstIndex; i <= lastIndex; i += 1) {
                            bullets.eq(i).addClass(`${params.bulletActiveClass}-main`);
                        }

                        if (swiper.params.loop) {
                            if (bulletIndex >= bullets.length) {
                                for (let i = params.dynamicMainBullets; i >= 0; i -= 1) {
                                    bullets.eq(bullets.length - i).addClass(`${params.bulletActiveClass}-main`);
                                }

                                bullets.eq(bullets.length - params.dynamicMainBullets - 1).addClass(`${params.bulletActiveClass}-prev`);
                            } else {
                                setSideBullets($firstDisplayedBullet, 'prev');
                                setSideBullets($lastDisplayedBullet, 'next');
                            }
                        } else {
                            setSideBullets($firstDisplayedBullet, 'prev');
                            setSideBullets($lastDisplayedBullet, 'next');
                        }
                    }
                }

                if (params.dynamicBullets) {
                    const dynamicBulletsLength = Math.min(bullets.length, params.dynamicMainBullets + 4);
                    const bulletsOffset = (bulletSize * dynamicBulletsLength - bulletSize) / 2 - midIndex * bulletSize;
                    const offsetProp = rtl ? 'right' : 'left';
                    bullets.css(swiper.isHorizontal() ? offsetProp : 'top', `${bulletsOffset}px`);
                }
            }

            if (params.type === 'fraction') {
                $el.find(classesToSelector(params.currentClass)).text(params.formatFractionCurrent(current + 1));
                $el.find(classesToSelector(params.totalClass)).text(params.formatFractionTotal(total));
            }

            if (params.type === 'progressbar') {
                let progressbarDirection;

                if (params.progressbarOpposite) {
                    progressbarDirection = swiper.isHorizontal() ? 'vertical' : 'horizontal';
                } else {
                    progressbarDirection = swiper.isHorizontal() ? 'horizontal' : 'vertical';
                }

                const scale = (current + 1) / total;
                let scaleX = 1;
                let scaleY = 1;

                if (progressbarDirection === 'horizontal') {
                    scaleX = scale;
                } else {
                    scaleY = scale;
                }

                $el.find(classesToSelector(params.progressbarFillClass)).transform(`translate3d(0,0,0) scaleX(${scaleX}) scaleY(${scaleY})`).transition(swiper.params.speed);
            }

            if (params.type === 'custom' && params.renderCustom) {
                $el.html(params.renderCustom(swiper, current + 1, total));
                emit('paginationRender', $el[0]);
            } else {
                emit('paginationUpdate', $el[0]);
            }

            if (swiper.params.watchOverflow && swiper.enabled) {
                $el[swiper.isLocked ? 'addClass' : 'removeClass'](params.lockClass);
            }
        }

        function render() {
            // Render Container
            const params = swiper.params.pagination;
            if (isPaginationDisabled()) return;
            const slidesLength = swiper.virtual && swiper.params.virtual.enabled ? swiper.virtual.slides.length : swiper.slides.length;
            const $el = swiper.pagination.$el;
            let paginationHTML = '';

            if (params.type === 'bullets') {
                let numberOfBullets = swiper.params.loop ? Math.ceil((slidesLength - swiper.loopedSlides * 2) / swiper.params.slidesPerGroup) : swiper.snapGrid.length;

                if (swiper.params.freeMode && swiper.params.freeMode.enabled && !swiper.params.loop && numberOfBullets > slidesLength) {
                    numberOfBullets = slidesLength;
                }

                for (let i = 0; i < numberOfBullets; i += 1) {
                    if (params.renderBullet) {
                        paginationHTML += params.renderBullet.call(swiper, i, params.bulletClass);
                    } else {
                        paginationHTML += `<${params.bulletElement} class="${params.bulletClass}"></${params.bulletElement}>`;
                    }
                }

                $el.html(paginationHTML);
                swiper.pagination.bullets = $el.find(classesToSelector(params.bulletClass));
            }

            if (params.type === 'fraction') {
                if (params.renderFraction) {
                    paginationHTML = params.renderFraction.call(swiper, params.currentClass, params.totalClass);
                } else {
                    paginationHTML = `<span class="${params.currentClass}"></span>` + ' / ' + `<span class="${params.totalClass}"></span>`;
                }

                $el.html(paginationHTML);
            }

            if (params.type === 'progressbar') {
                if (params.renderProgressbar) {
                    paginationHTML = params.renderProgressbar.call(swiper, params.progressbarFillClass);
                } else {
                    paginationHTML = `<span class="${params.progressbarFillClass}"></span>`;
                }

                $el.html(paginationHTML);
            }

            if (params.type !== 'custom') {
                emit('paginationRender', swiper.pagination.$el[0]);
            }
        }

        function init() {
            swiper.params.pagination = createElementIfNotDefined(swiper, swiper.originalParams.pagination, swiper.params.pagination, {
                el: 'ax-pages'
            });
            const params = swiper.params.pagination;
            if (!params.el) return;
            let $el = $(params.el);
            if ($el.length === 0) return;

            if (swiper.params.uniqueNavElements && typeof params.el === 'string' && $el.length > 1) {
                $el = swiper.$el.find(params.el); // check if it belongs to another nested Swiper

                if ($el.length > 1) {
                    $el = $el.filter(el => {
                        if ($(el).parents('.swiper')[0] !== swiper.el) return false;
                        return true;
                    });
                }
            }

            if (params.type === 'bullets' && params.clickable) {
                $el.addClass(params.clickableClass);
            }

            $el.addClass(params.modifierClass + params.type);
            $el.addClass(swiper.isHorizontal() ? params.horizontalClass : params.verticalClass);

            if (params.type === 'bullets' && params.dynamicBullets) {
                $el.addClass(`${params.modifierClass}${params.type}-dynamic`);
                dynamicBulletIndex = 0;

                if (params.dynamicMainBullets < 1) {
                    params.dynamicMainBullets = 1;
                }
            }

            if (params.type === 'progressbar' && params.progressbarOpposite) {
                $el.addClass(params.progressbarOppositeClass);
            }

            if (params.clickable) {
                $el.on('click', classesToSelector(params.bulletClass), function onClick(e) {
                    axPreventDefault(e);//AXUI
                    let index = $(this).index() * swiper.params.slidesPerGroup;
                    if (swiper.params.loop) index += swiper.loopedSlides;
                    swiper.slideTo(index);
                });
            }

            Object.assign(swiper.pagination, {
                $el,
                el: $el[0]
            });

            if (!swiper.enabled) {
                $el.addClass(params.lockClass);
            }
        }

        function destroy() {
            const params = swiper.params.pagination;
            if (isPaginationDisabled()) return;
            const $el = swiper.pagination.$el;
            $el.removeClass(params.hiddenClass);
            $el.removeClass(params.modifierClass + params.type);
            $el.removeClass(swiper.isHorizontal() ? params.horizontalClass : params.verticalClass);
            if (swiper.pagination.bullets && swiper.pagination.bullets.removeClass) swiper.pagination.bullets.removeClass(params.bulletActiveClass);

            if (params.clickable) {
                $el.off('click', classesToSelector(params.bulletClass));
            }
        }

        on('init', () => {
            init();
            render();
            update();
        });
        on('activeIndexChange', () => {
            if (swiper.params.loop) {
                update();
            } else if (typeof swiper.snapIndex === 'undefined') {
                update();
            }
        });
        on('snapIndexChange', () => {
            if (!swiper.params.loop) {
                update();
            }
        });
        on('slidesLengthChange', () => {
            if (swiper.params.loop) {
                render();
                update();
            }
        });
        on('snapGridLengthChange', () => {
            if (!swiper.params.loop) {
                render();
                update();
            }
        });
        on('destroy', () => {
            destroy();
        });
        on('enable disable', () => {
            const {
                $el
            } = swiper.pagination;

            if ($el) {
                $el[swiper.enabled ? 'removeClass' : 'addClass'](swiper.params.pagination.lockClass);
            }
        });
        on('lock unlock', () => {
            update();
        });
        on('click', (_s, e) => {
            const targetEl = e.target;
            const {
                $el
            } = swiper.pagination;

            if (swiper.params.pagination.el && swiper.params.pagination.hideOnClick && $el.length > 0 && !$(targetEl).hasClass(swiper.params.pagination.bulletClass)) {
                if (swiper.navigation && (swiper.navigation.nextEl && targetEl === swiper.navigation.nextEl || swiper.navigation.prevEl && targetEl === swiper.navigation.prevEl)) return;
                const isHidden = $el.hasClass(swiper.params.pagination.hiddenClass);

                if (isHidden === true) {
                    emit('paginationShow');
                } else {
                    emit('paginationHide');
                }

                $el.toggleClass(swiper.params.pagination.hiddenClass);
            }
        });
        Object.assign(swiper.pagination, {
            render,
            update,
            init,
            destroy
        });
    }

    function Scrollbar(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        const document = getDocument();
        let isTouched = false;
        let timeout = null;
        let dragTimeout = null;
        let dragStartPos;
        let dragSize;
        let trackSize;
        let divider;
        extendParams({
            scrollbar: {
                el: '.ax-scrollbar',//AXUI,原null
                dragSize: 'auto',
                hide: true,//AXUI,原false
                draggable: true,//AXUI,原false
                snapOnRelease: true,
                lockClass: 'ax-scrollbar-lock',
                dragClass: 'ax-scrollbar-drag'
            }
        });
        swiper.scrollbar = {
            el: null,
            dragEl: null,
            $el: null,
            $dragEl: null
        };

        function setTranslate() {
            if (!swiper.params.scrollbar.el || !swiper.scrollbar.el) return;
            const {
                scrollbar,
                rtlTranslate: rtl,
                progress
            } = swiper;
            const {
                $dragEl,
                $el
            } = scrollbar;
            const params = swiper.params.scrollbar;
            let newSize = dragSize;
            let newPos = (trackSize - dragSize) * progress;

            if (rtl) {
                newPos = -newPos;

                if (newPos > 0) {
                    newSize = dragSize - newPos;
                    newPos = 0;
                } else if (-newPos + dragSize > trackSize) {
                    newSize = trackSize + newPos;
                }
            } else if (newPos < 0) {
                newSize = dragSize + newPos;
                newPos = 0;
            } else if (newPos + dragSize > trackSize) {
                newSize = trackSize - newPos;
            }

            if (swiper.isHorizontal()) {
                $dragEl.transform(`translate3d(${newPos}px, 0, 0)`);
                $dragEl[0].style.width = `${newSize}px`;
            } else {
                $dragEl.transform(`translate3d(0px, ${newPos}px, 0)`);
                $dragEl[0].style.height = `${newSize}px`;
            }

            if (params.hide) {
                clearTimeout(timeout);
                $el[0].style.opacity = 1;
                timeout = setTimeout(() => {
                    $el[0].style.opacity = 0;
                    $el.transition(400);
                }, 1000);
            }
        }

        function setTransition(duration) {
            if (!swiper.params.scrollbar.el || !swiper.scrollbar.el) return;
            swiper.scrollbar.$dragEl.transition(duration);
        }

        function updateSize() {
            if (!swiper.params.scrollbar.el || !swiper.scrollbar.el) return;
            const {
                scrollbar
            } = swiper;
            const {
                $dragEl,
                $el
            } = scrollbar;
            $dragEl[0].style.width = '';
            $dragEl[0].style.height = '';
            trackSize = swiper.isHorizontal() ? $el[0].offsetWidth : $el[0].offsetHeight;
            divider = swiper.size / (swiper.virtualSize + swiper.params.slidesOffsetBefore - (swiper.params.centeredSlides ? swiper.snapGrid[0] : 0));

            if (swiper.params.scrollbar.dragSize === 'auto') {
                dragSize = trackSize * divider;
            } else {
                dragSize = parseInt(swiper.params.scrollbar.dragSize, 10);
            }

            if (swiper.isHorizontal()) {
                $dragEl[0].style.width = `${dragSize}px`;
            } else {
                $dragEl[0].style.height = `${dragSize}px`;
            }

            if (divider >= 1) {
                $el[0].style.display = 'none';
            } else {
                $el[0].style.display = '';
            }

            if (swiper.params.scrollbar.hide) {
                $el[0].style.opacity = 0;
            }

            if (swiper.params.watchOverflow && swiper.enabled) {
                scrollbar.$el[swiper.isLocked ? 'addClass' : 'removeClass'](swiper.params.scrollbar.lockClass);
            }
        }

        function getPointerPosition(e) {
            if (swiper.isHorizontal()) {
                return e.type === 'touchstart' || e.type === 'touchmove' ? e.targetTouches[0].clientX : e.clientX;
            }

            return e.type === 'touchstart' || e.type === 'touchmove' ? e.targetTouches[0].clientY : e.clientY;
        }

        function setDragPosition(e) {
            const {
                scrollbar,
                rtlTranslate: rtl
            } = swiper;
            const {
                $el
            } = scrollbar;
            let positionRatio;
            positionRatio = (getPointerPosition(e) - $el.offset()[swiper.isHorizontal() ? 'left' : 'top'] - (dragStartPos !== null ? dragStartPos : dragSize / 2)) / (trackSize - dragSize);
            positionRatio = Math.max(Math.min(positionRatio, 1), 0);

            if (rtl) {
                positionRatio = 1 - positionRatio;
            }

            const position = swiper.minTranslate() + (swiper.maxTranslate() - swiper.minTranslate()) * positionRatio;
            swiper.updateProgress(position);
            swiper.setTranslate(position);
            swiper.updateActiveIndex();
            swiper.updateSlidesClasses();
        }

        function onDragStart(e) {
            const params = swiper.params.scrollbar;
            const {
                scrollbar,
                $wrapperEl
            } = swiper;
            const {
                $el,
                $dragEl
            } = scrollbar;
            isTouched = true;
            dragStartPos = e.target === $dragEl[0] || e.target === $dragEl ? getPointerPosition(e) - e.target.getBoundingClientRect()[swiper.isHorizontal() ? 'left' : 'top'] : null;
            axPreventDefault(e);//AXUI
            e.stopPropagation();
            $wrapperEl.transition(100);
            $dragEl.transition(100);
            setDragPosition(e);
            clearTimeout(dragTimeout);
            $el.transition(0);

            if (params.hide) {
                $el.css('opacity', 1);
            }

            if (swiper.params.cssMode) {
                swiper.$wrapperEl.css('scroll-snap-type', 'none');
            }

            emit('scrollbarDragStart', e);
        }

        function onDragMove(e) {
            const {
                scrollbar,
                $wrapperEl
            } = swiper;
            const {
                $el,
                $dragEl
            } = scrollbar;
            if (!isTouched) return;
            axPreventDefault(e);//AXUI
            setDragPosition(e);
            $wrapperEl.transition(0);
            $el.transition(0);
            $dragEl.transition(0);
            emit('scrollbarDragMove', e);
        }

        function onDragEnd(e) {
            const params = swiper.params.scrollbar;
            const {
                scrollbar,
                $wrapperEl
            } = swiper;
            const {
                $el
            } = scrollbar;
            if (!isTouched) return;
            isTouched = false;

            if (swiper.params.cssMode) {
                swiper.$wrapperEl.css('scroll-snap-type', '');
                $wrapperEl.transition('');
            }

            if (params.hide) {
                clearTimeout(dragTimeout);
                dragTimeout = nextTick(() => {
                    $el.css('opacity', 0);
                    $el.transition(400);
                }, 1000);
            }

            emit('scrollbarDragEnd', e);

            if (params.snapOnRelease) {
                swiper.slideToClosest();
            }
        }

        function events(method) {
            const {
                scrollbar,
                touchEventsTouch,
                touchEventsDesktop,
                params,
                support
            } = swiper;
            const $el = scrollbar.$el;
            const target = $el[0];
            const activeListener = support.passiveListener && params.passiveListeners ? {
                passive: false,
                capture: false
            } : false;
            const passiveListener = support.passiveListener && params.passiveListeners ? {
                passive: true,
                capture: false
            } : false;
            if (!target) return;
            const eventMethod = method === 'on' ? 'addEventListener' : 'removeEventListener';

            if (!support.touch) {
                target[eventMethod](touchEventsDesktop.start, onDragStart, activeListener);
                document[eventMethod](touchEventsDesktop.move, onDragMove, activeListener);
                document[eventMethod](touchEventsDesktop.end, onDragEnd, passiveListener);
            } else {
                target[eventMethod](touchEventsTouch.start, onDragStart, activeListener);
                target[eventMethod](touchEventsTouch.move, onDragMove, activeListener);
                target[eventMethod](touchEventsTouch.end, onDragEnd, passiveListener);
            }
        }

        function enableDraggable() {
            if (!swiper.params.scrollbar.el) return;
            events('on');
        }

        function disableDraggable() {
            if (!swiper.params.scrollbar.el) return;
            events('off');
        }

        function init() {
            const {
                scrollbar,
                $el: $swiperEl
            } = swiper;
            swiper.params.scrollbar = createElementIfNotDefined(swiper, swiper.originalParams.scrollbar, swiper.params.scrollbar, {
                el: 'ax-scrollbar'
            });
            const params = swiper.params.scrollbar;
            if (!params.el) return;
            let $el = $(params.el);

            if (swiper.params.uniqueNavElements && typeof params.el === 'string' && $el.length > 1 && $swiperEl.find(params.el).length === 1) {
                $el = $swiperEl.find(params.el);
            }

            let $dragEl = $el.find(`.${swiper.params.scrollbar.dragClass}`);

            if ($dragEl.length === 0) {
                $dragEl = $(`<div class="${swiper.params.scrollbar.dragClass}"></div>`);
                $el.append($dragEl);
            }

            Object.assign(scrollbar, {
                $el,
                el: $el[0],
                $dragEl,
                dragEl: $dragEl[0]
            });

            if (params.draggable) {
                enableDraggable();
            }

            if ($el) {
                $el[swiper.enabled ? 'removeClass' : 'addClass'](swiper.params.scrollbar.lockClass);
            }
        }

        function destroy() {
            disableDraggable();
        }

        on('init', () => {
            init();
            updateSize();
            setTranslate();
        });
        on('update resize observerUpdate lock unlock', () => {
            updateSize();
        });
        on('setTranslate', () => {
            setTranslate();
        });
        on('setTransition', (_s, duration) => {
            setTransition(duration);
        });
        on('enable disable', () => {
            const {
                $el
            } = swiper.scrollbar;

            if ($el) {
                $el[swiper.enabled ? 'removeClass' : 'addClass'](swiper.params.scrollbar.lockClass);
            }
        });
        on('destroy', () => {
            destroy();
        });
        Object.assign(swiper.scrollbar, {
            updateSize,
            setTranslate,
            init,
            destroy
        });
    }

    function Parallax(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            parallax: {
                enabled: false
            }
        });

        const setTransform = (el, progress) => {
            const {
                rtl
            } = swiper;
            const $el = $(el);
            const rtlFactor = rtl ? -1 : 1;
            const p = $el.attr('data-swiper-parallax') || '0';
            let x = $el.attr('data-swiper-parallax-x');
            let y = $el.attr('data-swiper-parallax-y');
            const scale = $el.attr('data-swiper-parallax-scale');
            const opacity = $el.attr('data-swiper-parallax-opacity');

            if (x || y) {
                x = x || '0';
                y = y || '0';
            } else if (swiper.isHorizontal()) {
                x = p;
                y = '0';
            } else {
                y = p;
                x = '0';
            }

            if (x.indexOf('%') >= 0) {
                x = `${parseInt(x, 10) * progress * rtlFactor}%`;
            } else {
                x = `${x * progress * rtlFactor}px`;
            }

            if (y.indexOf('%') >= 0) {
                y = `${parseInt(y, 10) * progress}%`;
            } else {
                y = `${y * progress}px`;
            }

            if (typeof opacity !== 'undefined' && opacity !== null) {
                const currentOpacity = opacity - (opacity - 1) * (1 - Math.abs(progress));
                $el[0].style.opacity = currentOpacity;
            }

            if (typeof scale === 'undefined' || scale === null) {
                $el.transform(`translate3d(${x}, ${y}, 0px)`);
            } else {
                const currentScale = scale - (scale - 1) * (1 - Math.abs(progress));
                $el.transform(`translate3d(${x}, ${y}, 0px) scale(${currentScale})`);
            }
        };

        const setTranslate = () => {
            const {
                $el,
                slides,
                progress,
                snapGrid
            } = swiper;
            $el.children('[data-swiper-parallax], [data-swiper-parallax-x], [data-swiper-parallax-y], [data-swiper-parallax-opacity], [data-swiper-parallax-scale]').each(el => {
                setTransform(el, progress);
            });
            slides.each((slideEl, slideIndex) => {
                let slideProgress = slideEl.progress;

                if (swiper.params.slidesPerGroup > 1 && swiper.params.slidesPerView !== 'auto') {
                    slideProgress += Math.ceil(slideIndex / 2) - progress * (snapGrid.length - 1);
                }

                slideProgress = Math.min(Math.max(slideProgress, -1), 1);
                $(slideEl).find('[data-swiper-parallax], [data-swiper-parallax-x], [data-swiper-parallax-y], [data-swiper-parallax-opacity], [data-swiper-parallax-scale]').each(el => {
                    setTransform(el, slideProgress);
                });
            });
        };

        const setTransition = function (duration) {
            if (duration === void 0) {
                duration = swiper.params.speed;
            }

            const {
                $el
            } = swiper;
            $el.find('[data-swiper-parallax], [data-swiper-parallax-x], [data-swiper-parallax-y], [data-swiper-parallax-opacity], [data-swiper-parallax-scale]').each(parallaxEl => {
                const $parallaxEl = $(parallaxEl);
                let parallaxDuration = parseInt($parallaxEl.attr('data-swiper-parallax-duration'), 10) || duration;
                if (duration === 0) parallaxDuration = 0;
                $parallaxEl.transition(parallaxDuration);
            });
        };

        on('beforeInit', () => {
            if (!swiper.params.parallax.enabled) return;
            swiper.params.watchSlidesProgress = true;
            swiper.originalParams.watchSlidesProgress = true;
        });
        on('init', () => {
            if (!swiper.params.parallax.enabled) return;
            setTranslate();
        });
        on('setTranslate', () => {
            if (!swiper.params.parallax.enabled) return;
            setTranslate();
        });
        on('setTransition', (_swiper, duration) => {
            if (!swiper.params.parallax.enabled) return;
            setTransition(duration);
        });
    }

    function Zoom(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        const window = getWindow();
        extendParams({
            zoom: {
                enabled: false,
                maxRatio: 3,
                minRatio: 1,
                toggle: true,
                containerClass: 'ax-zoom',
                zoomedSlideClass: 'ax-slide-zoomed'
            }
        });
        swiper.zoom = {
            enabled: false
        };
        let currentScale = 1;
        let isScaling = false;
        let gesturesEnabled;
        let fakeGestureTouched;
        let fakeGestureMoved;
        const gesture = {
            $slideEl: undefined,
            slideWidth: undefined,
            slideHeight: undefined,
            $imageEl: undefined,
            $imageWrapEl: undefined,
            maxRatio: 3
        };
        const image = {
            isTouched: undefined,
            isMoved: undefined,
            currentX: undefined,
            currentY: undefined,
            minX: undefined,
            minY: undefined,
            maxX: undefined,
            maxY: undefined,
            width: undefined,
            height: undefined,
            startX: undefined,
            startY: undefined,
            touchesStart: {},
            touchesCurrent: {}
        };
        const velocity = {
            x: undefined,
            y: undefined,
            prevPositionX: undefined,
            prevPositionY: undefined,
            prevTime: undefined
        };
        let scale = 1;
        Object.defineProperty(swiper.zoom, 'scale', {
            get() {
                return scale;
            },

            set(value) {
                if (scale !== value) {
                    const imageEl = gesture.$imageEl ? gesture.$imageEl[0] : undefined;
                    const slideEl = gesture.$slideEl ? gesture.$slideEl[0] : undefined;
                    emit('zoomChange', value, imageEl, slideEl);
                }

                scale = value;
            }

        });

        function getDistanceBetweenTouches(e) {
            if (e.targetTouches.length < 2) return 1;
            const x1 = e.targetTouches[0].pageX;
            const y1 = e.targetTouches[0].pageY;
            const x2 = e.targetTouches[1].pageX;
            const y2 = e.targetTouches[1].pageY;
            const distance = Math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2);
            return distance;
        } // Events


        function onGestureStart(e) {
            const support = swiper.support;
            const params = swiper.params.zoom;
            fakeGestureTouched = false;
            fakeGestureMoved = false;

            if (!support.gestures) {
                if (e.type !== 'touchstart' || e.type === 'touchstart' && e.targetTouches.length < 2) {
                    return;
                }

                fakeGestureTouched = true;
                gesture.scaleStart = getDistanceBetweenTouches(e);
            }

            if (!gesture.$slideEl || !gesture.$slideEl.length) {
                gesture.$slideEl = $(e.target).closest(`.${swiper.params.slideClass}`);
                if (gesture.$slideEl.length === 0) gesture.$slideEl = swiper.slides.eq(swiper.activeIndex);
                gesture.$imageEl = gesture.$slideEl.find(`.${params.containerClass}`).eq(0).find('picture, img, svg, canvas, .swiper-zoom-target').eq(0);
                gesture.$imageWrapEl = gesture.$imageEl.parent(`.${params.containerClass}`);
                gesture.maxRatio = gesture.$imageWrapEl.attr('data-swiper-zoom') || params.maxRatio;

                if (gesture.$imageWrapEl.length === 0) {
                    gesture.$imageEl = undefined;
                    return;
                }
            }

            if (gesture.$imageEl) {
                gesture.$imageEl.transition(0);
            }

            isScaling = true;
        }

        function onGestureChange(e) {
            const support = swiper.support;
            const params = swiper.params.zoom;
            const zoom = swiper.zoom;

            if (!support.gestures) {
                if (e.type !== 'touchmove' || e.type === 'touchmove' && e.targetTouches.length < 2) {
                    return;
                }

                fakeGestureMoved = true;
                gesture.scaleMove = getDistanceBetweenTouches(e);
            }

            if (!gesture.$imageEl || gesture.$imageEl.length === 0) {
                if (e.type === 'gesturechange') onGestureStart(e);
                return;
            }

            if (support.gestures) {
                zoom.scale = e.scale * currentScale;
            } else {
                zoom.scale = gesture.scaleMove / gesture.scaleStart * currentScale;
            }

            if (zoom.scale > gesture.maxRatio) {
                zoom.scale = gesture.maxRatio - 1 + (zoom.scale - gesture.maxRatio + 1) ** 0.5;
            }

            if (zoom.scale < params.minRatio) {
                zoom.scale = params.minRatio + 1 - (params.minRatio - zoom.scale + 1) ** 0.5;
            }

            gesture.$imageEl.transform(`translate3d(0,0,0) scale(${zoom.scale})`);
        }

        function onGestureEnd(e) {
            const device = swiper.device;
            const support = swiper.support;
            const params = swiper.params.zoom;
            const zoom = swiper.zoom;

            if (!support.gestures) {
                if (!fakeGestureTouched || !fakeGestureMoved) {
                    return;
                }

                if (e.type !== 'touchend' || e.type === 'touchend' && e.changedTouches.length < 2 && !device.android) {
                    return;
                }

                fakeGestureTouched = false;
                fakeGestureMoved = false;
            }

            if (!gesture.$imageEl || gesture.$imageEl.length === 0) return;
            zoom.scale = Math.max(Math.min(zoom.scale, gesture.maxRatio), params.minRatio);
            gesture.$imageEl.transition(swiper.params.speed).transform(`translate3d(0,0,0) scale(${zoom.scale})`);
            currentScale = zoom.scale;
            isScaling = false;
            if (zoom.scale === 1) gesture.$slideEl = undefined;
        }

        function onTouchStart(e) {
            const device = swiper.device;
            if (!gesture.$imageEl || gesture.$imageEl.length === 0) return;
            if (image.isTouched) return;
            if (device.android && e.cancelable){
                axPreventDefault(e);//AXUI
            } 
            image.isTouched = true;
            image.touchesStart.x = e.type === 'touchstart' ? e.targetTouches[0].pageX : e.pageX;
            image.touchesStart.y = e.type === 'touchstart' ? e.targetTouches[0].pageY : e.pageY;
        }

        function onTouchMove(e) {
            const zoom = swiper.zoom;
            if (!gesture.$imageEl || gesture.$imageEl.length === 0) return;
            swiper.allowClick = false;
            if (!image.isTouched || !gesture.$slideEl) return;

            if (!image.isMoved) {
                image.width = gesture.$imageEl[0].offsetWidth;
                image.height = gesture.$imageEl[0].offsetHeight;
                image.startX = getTranslate(gesture.$imageWrapEl[0], 'x') || 0;
                image.startY = getTranslate(gesture.$imageWrapEl[0], 'y') || 0;
                gesture.slideWidth = gesture.$slideEl[0].offsetWidth;
                gesture.slideHeight = gesture.$slideEl[0].offsetHeight;
                gesture.$imageWrapEl.transition(0);
            } // Define if we need image drag


            const scaledWidth = image.width * zoom.scale;
            const scaledHeight = image.height * zoom.scale;
            if (scaledWidth < gesture.slideWidth && scaledHeight < gesture.slideHeight) return;
            image.minX = Math.min(gesture.slideWidth / 2 - scaledWidth / 2, 0);
            image.maxX = -image.minX;
            image.minY = Math.min(gesture.slideHeight / 2 - scaledHeight / 2, 0);
            image.maxY = -image.minY;
            image.touchesCurrent.x = e.type === 'touchmove' ? e.targetTouches[0].pageX : e.pageX;
            image.touchesCurrent.y = e.type === 'touchmove' ? e.targetTouches[0].pageY : e.pageY;

            if (!image.isMoved && !isScaling) {
                if (swiper.isHorizontal() && (Math.floor(image.minX) === Math.floor(image.startX) && image.touchesCurrent.x < image.touchesStart.x || Math.floor(image.maxX) === Math.floor(image.startX) && image.touchesCurrent.x > image.touchesStart.x)) {
                    image.isTouched = false;
                    return;
                }

                if (!swiper.isHorizontal() && (Math.floor(image.minY) === Math.floor(image.startY) && image.touchesCurrent.y < image.touchesStart.y || Math.floor(image.maxY) === Math.floor(image.startY) && image.touchesCurrent.y > image.touchesStart.y)) {
                    image.isTouched = false;
                    return;
                }
            }

            if (e.cancelable) {
                axPreventDefault(e);//AXUI
            }

            e.stopPropagation();
            image.isMoved = true;
            image.currentX = image.touchesCurrent.x - image.touchesStart.x + image.startX;
            image.currentY = image.touchesCurrent.y - image.touchesStart.y + image.startY;

            if (image.currentX < image.minX) {
                image.currentX = image.minX + 1 - (image.minX - image.currentX + 1) ** 0.8;
            }

            if (image.currentX > image.maxX) {
                image.currentX = image.maxX - 1 + (image.currentX - image.maxX + 1) ** 0.8;
            }

            if (image.currentY < image.minY) {
                image.currentY = image.minY + 1 - (image.minY - image.currentY + 1) ** 0.8;
            }

            if (image.currentY > image.maxY) {
                image.currentY = image.maxY - 1 + (image.currentY - image.maxY + 1) ** 0.8;
            } // Velocity


            if (!velocity.prevPositionX) velocity.prevPositionX = image.touchesCurrent.x;
            if (!velocity.prevPositionY) velocity.prevPositionY = image.touchesCurrent.y;
            if (!velocity.prevTime) velocity.prevTime = Date.now();
            velocity.x = (image.touchesCurrent.x - velocity.prevPositionX) / (Date.now() - velocity.prevTime) / 2;
            velocity.y = (image.touchesCurrent.y - velocity.prevPositionY) / (Date.now() - velocity.prevTime) / 2;
            if (Math.abs(image.touchesCurrent.x - velocity.prevPositionX) < 2) velocity.x = 0;
            if (Math.abs(image.touchesCurrent.y - velocity.prevPositionY) < 2) velocity.y = 0;
            velocity.prevPositionX = image.touchesCurrent.x;
            velocity.prevPositionY = image.touchesCurrent.y;
            velocity.prevTime = Date.now();
            gesture.$imageWrapEl.transform(`translate3d(${image.currentX}px, ${image.currentY}px,0)`);
        }

        function onTouchEnd() {
            const zoom = swiper.zoom;
            if (!gesture.$imageEl || gesture.$imageEl.length === 0) return;

            if (!image.isTouched || !image.isMoved) {
                image.isTouched = false;
                image.isMoved = false;
                return;
            }

            image.isTouched = false;
            image.isMoved = false;
            let momentumDurationX = 300;
            let momentumDurationY = 300;
            const momentumDistanceX = velocity.x * momentumDurationX;
            const newPositionX = image.currentX + momentumDistanceX;
            const momentumDistanceY = velocity.y * momentumDurationY;
            const newPositionY = image.currentY + momentumDistanceY; // Fix duration

            if (velocity.x !== 0) momentumDurationX = Math.abs((newPositionX - image.currentX) / velocity.x);
            if (velocity.y !== 0) momentumDurationY = Math.abs((newPositionY - image.currentY) / velocity.y);
            const momentumDuration = Math.max(momentumDurationX, momentumDurationY);
            image.currentX = newPositionX;
            image.currentY = newPositionY; // Define if we need image drag

            const scaledWidth = image.width * zoom.scale;
            const scaledHeight = image.height * zoom.scale;
            image.minX = Math.min(gesture.slideWidth / 2 - scaledWidth / 2, 0);
            image.maxX = -image.minX;
            image.minY = Math.min(gesture.slideHeight / 2 - scaledHeight / 2, 0);
            image.maxY = -image.minY;
            image.currentX = Math.max(Math.min(image.currentX, image.maxX), image.minX);
            image.currentY = Math.max(Math.min(image.currentY, image.maxY), image.minY);
            gesture.$imageWrapEl.transition(momentumDuration).transform(`translate3d(${image.currentX}px, ${image.currentY}px,0)`);
        }

        function onTransitionEnd() {
            const zoom = swiper.zoom;

            if (gesture.$slideEl && swiper.previousIndex !== swiper.activeIndex) {
                if (gesture.$imageEl) {
                    gesture.$imageEl.transform('translate3d(0,0,0) scale(1)');
                }

                if (gesture.$imageWrapEl) {
                    gesture.$imageWrapEl.transform('translate3d(0,0,0)');
                }

                zoom.scale = 1;
                currentScale = 1;
                gesture.$slideEl = undefined;
                gesture.$imageEl = undefined;
                gesture.$imageWrapEl = undefined;
            }
        }

        function zoomIn(e) {
            const zoom = swiper.zoom;
            const params = swiper.params.zoom;

            if (!gesture.$slideEl) {
                if (e && e.target) {
                    gesture.$slideEl = $(e.target).closest(`.${swiper.params.slideClass}`);
                }

                if (!gesture.$slideEl) {
                    if (swiper.params.virtual && swiper.params.virtual.enabled && swiper.virtual) {
                        gesture.$slideEl = swiper.$wrapperEl.children(`.${swiper.params.slideActiveClass}`);
                    } else {
                        gesture.$slideEl = swiper.slides.eq(swiper.activeIndex);
                    }
                }

                gesture.$imageEl = gesture.$slideEl.find(`.${params.containerClass}`).eq(0).find('picture, img, svg, canvas, .swiper-zoom-target').eq(0);
                gesture.$imageWrapEl = gesture.$imageEl.parent(`.${params.containerClass}`);
            }

            if (!gesture.$imageEl || gesture.$imageEl.length === 0 || !gesture.$imageWrapEl || gesture.$imageWrapEl.length === 0) return;

            if (swiper.params.cssMode) {
                swiper.wrapperEl.style.overflow = 'hidden';
                swiper.wrapperEl.style.touchAction = 'none';
            }

            gesture.$slideEl.addClass(`${params.zoomedSlideClass}`);
            let touchX;
            let touchY;
            let offsetX;
            let offsetY;
            let diffX;
            let diffY;
            let translateX;
            let translateY;
            let imageWidth;
            let imageHeight;
            let scaledWidth;
            let scaledHeight;
            let translateMinX;
            let translateMinY;
            let translateMaxX;
            let translateMaxY;
            let slideWidth;
            let slideHeight;

            if (typeof image.touchesStart.x === 'undefined' && e) {
                touchX = e.type === 'touchend' ? e.changedTouches[0].pageX : e.pageX;
                touchY = e.type === 'touchend' ? e.changedTouches[0].pageY : e.pageY;
            } else {
                touchX = image.touchesStart.x;
                touchY = image.touchesStart.y;
            }

            zoom.scale = gesture.$imageWrapEl.attr('data-swiper-zoom') || params.maxRatio;
            currentScale = gesture.$imageWrapEl.attr('data-swiper-zoom') || params.maxRatio;

            if (e) {
                slideWidth = gesture.$slideEl[0].offsetWidth;
                slideHeight = gesture.$slideEl[0].offsetHeight;
                offsetX = gesture.$slideEl.offset().left + window.scrollX;
                offsetY = gesture.$slideEl.offset().top + window.scrollY;
                diffX = offsetX + slideWidth / 2 - touchX;
                diffY = offsetY + slideHeight / 2 - touchY;
                imageWidth = gesture.$imageEl[0].offsetWidth;
                imageHeight = gesture.$imageEl[0].offsetHeight;
                scaledWidth = imageWidth * zoom.scale;
                scaledHeight = imageHeight * zoom.scale;
                translateMinX = Math.min(slideWidth / 2 - scaledWidth / 2, 0);
                translateMinY = Math.min(slideHeight / 2 - scaledHeight / 2, 0);
                translateMaxX = -translateMinX;
                translateMaxY = -translateMinY;
                translateX = diffX * zoom.scale;
                translateY = diffY * zoom.scale;

                if (translateX < translateMinX) {
                    translateX = translateMinX;
                }

                if (translateX > translateMaxX) {
                    translateX = translateMaxX;
                }

                if (translateY < translateMinY) {
                    translateY = translateMinY;
                }

                if (translateY > translateMaxY) {
                    translateY = translateMaxY;
                }
            } else {
                translateX = 0;
                translateY = 0;
            }

            gesture.$imageWrapEl.transition(300).transform(`translate3d(${translateX}px, ${translateY}px,0)`);
            gesture.$imageEl.transition(300).transform(`translate3d(0,0,0) scale(${zoom.scale})`);
        }

        function zoomOut() {
            const zoom = swiper.zoom;
            const params = swiper.params.zoom;

            if (!gesture.$slideEl) {
                if (swiper.params.virtual && swiper.params.virtual.enabled && swiper.virtual) {
                    gesture.$slideEl = swiper.$wrapperEl.children(`.${swiper.params.slideActiveClass}`);
                } else {
                    gesture.$slideEl = swiper.slides.eq(swiper.activeIndex);
                }

                gesture.$imageEl = gesture.$slideEl.find(`.${params.containerClass}`).eq(0).find('picture, img, svg, canvas, .swiper-zoom-target').eq(0);
                gesture.$imageWrapEl = gesture.$imageEl.parent(`.${params.containerClass}`);
            }

            if (!gesture.$imageEl || gesture.$imageEl.length === 0 || !gesture.$imageWrapEl || gesture.$imageWrapEl.length === 0) return;

            if (swiper.params.cssMode) {
                swiper.wrapperEl.style.overflow = '';
                swiper.wrapperEl.style.touchAction = '';
            }

            zoom.scale = 1;
            currentScale = 1;
            gesture.$imageWrapEl.transition(300).transform('translate3d(0,0,0)');
            gesture.$imageEl.transition(300).transform('translate3d(0,0,0) scale(1)');
            gesture.$slideEl.removeClass(`${params.zoomedSlideClass}`);
            gesture.$slideEl = undefined;
        } // Toggle Zoom


        function zoomToggle(e) {
            const zoom = swiper.zoom;

            if (zoom.scale && zoom.scale !== 1) {
                // Zoom Out
                zoomOut();
            } else {
                // Zoom In
                zoomIn(e);
            }
        }

        function getListeners() {
            const support = swiper.support;
            const passiveListener = swiper.touchEvents.start === 'touchstart' && support.passiveListener && swiper.params.passiveListeners ? {
                passive: true,
                capture: false
            } : false;
            const activeListenerWithCapture = support.passiveListener ? {
                passive: false,
                capture: true
            } : true;
            return {
                passiveListener,
                activeListenerWithCapture
            };
        }

        function getSlideSelector() {
            return `.${swiper.params.slideClass}`;
        }

        function toggleGestures(method) {
            const {
                passiveListener
            } = getListeners();
            const slideSelector = getSlideSelector();
            swiper.$wrapperEl[method]('gesturestart', slideSelector, onGestureStart, passiveListener);
            swiper.$wrapperEl[method]('gesturechange', slideSelector, onGestureChange, passiveListener);
            swiper.$wrapperEl[method]('gestureend', slideSelector, onGestureEnd, passiveListener);
        }

        function enableGestures() {
            if (gesturesEnabled) return;
            gesturesEnabled = true;
            toggleGestures('on');
        }

        function disableGestures() {
            if (!gesturesEnabled) return;
            gesturesEnabled = false;
            toggleGestures('off');
        } // Attach/Detach Events


        function enable() {
            const zoom = swiper.zoom;
            if (zoom.enabled) return;
            zoom.enabled = true;
            const support = swiper.support;
            const {
                passiveListener,
                activeListenerWithCapture
            } = getListeners();
            const slideSelector = getSlideSelector(); // Scale image

            if (support.gestures) {
                swiper.$wrapperEl.on(swiper.touchEvents.start, enableGestures, passiveListener);
                swiper.$wrapperEl.on(swiper.touchEvents.end, disableGestures, passiveListener);
            } else if (swiper.touchEvents.start === 'touchstart') {
                swiper.$wrapperEl.on(swiper.touchEvents.start, slideSelector, onGestureStart, passiveListener);
                swiper.$wrapperEl.on(swiper.touchEvents.move, slideSelector, onGestureChange, activeListenerWithCapture);
                swiper.$wrapperEl.on(swiper.touchEvents.end, slideSelector, onGestureEnd, passiveListener);

                if (swiper.touchEvents.cancel) {
                    swiper.$wrapperEl.on(swiper.touchEvents.cancel, slideSelector, onGestureEnd, passiveListener);
                }
            } // Move image


            swiper.$wrapperEl.on(swiper.touchEvents.move, `.${swiper.params.zoom.containerClass}`, onTouchMove, activeListenerWithCapture);
        }

        function disable() {
            const zoom = swiper.zoom;
            if (!zoom.enabled) return;
            const support = swiper.support;
            zoom.enabled = false;
            const {
                passiveListener,
                activeListenerWithCapture
            } = getListeners();
            const slideSelector = getSlideSelector(); // Scale image

            if (support.gestures) {
                swiper.$wrapperEl.off(swiper.touchEvents.start, enableGestures, passiveListener);
                swiper.$wrapperEl.off(swiper.touchEvents.end, disableGestures, passiveListener);
            } else if (swiper.touchEvents.start === 'touchstart') {
                swiper.$wrapperEl.off(swiper.touchEvents.start, slideSelector, onGestureStart, passiveListener);
                swiper.$wrapperEl.off(swiper.touchEvents.move, slideSelector, onGestureChange, activeListenerWithCapture);
                swiper.$wrapperEl.off(swiper.touchEvents.end, slideSelector, onGestureEnd, passiveListener);

                if (swiper.touchEvents.cancel) {
                    swiper.$wrapperEl.off(swiper.touchEvents.cancel, slideSelector, onGestureEnd, passiveListener);
                }
            } // Move image


            swiper.$wrapperEl.off(swiper.touchEvents.move, `.${swiper.params.zoom.containerClass}`, onTouchMove, activeListenerWithCapture);
        }

        on('init', () => {
            if (swiper.params.zoom.enabled) {
                enable();
            }
        });
        on('destroy', () => {
            disable();
        });
        on('touchStart', (_s, e) => {
            if (!swiper.zoom.enabled) return;
            onTouchStart(e);
        });
        on('touchEnd', (_s, e) => {
            if (!swiper.zoom.enabled) return;
            onTouchEnd();
        });
        on('doubleTap', (_s, e) => {
            if (!swiper.animating && swiper.params.zoom.enabled && swiper.zoom.enabled && swiper.params.zoom.toggle) {
                zoomToggle(e);
            }
        });
        on('transitionEnd', () => {
            if (swiper.zoom.enabled && swiper.params.zoom.enabled) {
                onTransitionEnd();
            }
        });
        on('slideChange', () => {
            if (swiper.zoom.enabled && swiper.params.zoom.enabled && swiper.params.cssMode) {
                onTransitionEnd();
            }
        });
        Object.assign(swiper.zoom, {
            enable,
            disable,
            in: zoomIn,
            out: zoomOut,
            toggle: zoomToggle
        });
    }

    function Lazy(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        extendParams({
            lazy: {
                checkInView: false,
                enabled: false,
                loadPrevNext: false,
                loadPrevNextAmount: 1,
                loadOnTransitionStart: false,
                scrollingElement: '',
                elementClass: 'ax-lazy',//AXUI,图片上的样式，原swiper-lazy
                loadingClass: 'ax-lazy-loading',
                loadedClass: 'ax-lazy-loaded',
                preloaderClass: 'ax-loading'//AXUI，缓加载效果Dom，原swiper-lazy-preloader
            }
        });
        swiper.lazy = {};
        let scrollHandlerAttached = false;
        let initialImageLoaded = false;

        function loadInSlide(index, loadInDuplicate) {
            if (loadInDuplicate === void 0) {
                loadInDuplicate = true;
            }

            const params = swiper.params.lazy;
            if (typeof index === 'undefined') return;
            if (swiper.slides.length === 0) return;
            const isVirtual = swiper.virtual && swiper.params.virtual.enabled;
            const $slideEl = isVirtual ? swiper.$wrapperEl.children(`.${swiper.params.slideClass}[data-swiper-slide-index="${index}"]`) : swiper.slides.eq(index);
            const $images = $slideEl.find(`.${params.elementClass}:not(.${params.loadedClass}):not(.${params.loadingClass})`);

            if ($slideEl.hasClass(params.elementClass) && !$slideEl.hasClass(params.loadedClass) && !$slideEl.hasClass(params.loadingClass)) {
                $images.push($slideEl[0]);
            }

            if ($images.length === 0) return;
            $images.each(imageEl => {
                const $imageEl = $(imageEl);
                $imageEl.addClass(params.loadingClass);
                const background = $imageEl.attr('data-background');
                const src = $imageEl.attr('data-src');
                const srcset = $imageEl.attr('data-srcset');
                const sizes = $imageEl.attr('data-sizes');
                const $pictureEl = $imageEl.parent('picture');
                swiper.loadImage($imageEl[0], src || background, srcset, sizes, false, () => {
                    if (typeof swiper === 'undefined' || swiper === null || !swiper || swiper && !swiper.params || swiper.destroyed) return;

                    if (background) {
                        $imageEl.css('background-image', `url("${background}")`);
                        $imageEl.removeAttr('data-background');
                    } else {
                        if (srcset) {
                            $imageEl.attr('srcset', srcset);
                            $imageEl.removeAttr('data-srcset');
                        }

                        if (sizes) {
                            $imageEl.attr('sizes', sizes);
                            $imageEl.removeAttr('data-sizes');
                        }

                        if ($pictureEl.length) {
                            $pictureEl.children('source').each(sourceEl => {
                                const $source = $(sourceEl);

                                if ($source.attr('data-srcset')) {
                                    $source.attr('srcset', $source.attr('data-srcset'));
                                    $source.removeAttr('data-srcset');
                                }
                            });
                        }

                        if (src) {
                            $imageEl.attr('src', src);
                            $imageEl.removeAttr('data-src');
                        }
                    }

                    $imageEl.addClass(params.loadedClass).removeClass(params.loadingClass);
                    $slideEl.find(`.${params.preloaderClass}`).remove();

                    if (swiper.params.loop && loadInDuplicate) {
                        const slideOriginalIndex = $slideEl.attr('data-swiper-slide-index');

                        if ($slideEl.hasClass(swiper.params.slideDuplicateClass)) {
                            const originalSlide = swiper.$wrapperEl.children(`[data-swiper-slide-index="${slideOriginalIndex}"]:not(.${swiper.params.slideDuplicateClass})`);
                            loadInSlide(originalSlide.index(), false);
                        } else {
                            const duplicatedSlide = swiper.$wrapperEl.children(`.${swiper.params.slideDuplicateClass}[data-swiper-slide-index="${slideOriginalIndex}"]`);
                            loadInSlide(duplicatedSlide.index(), false);
                        }
                    }

                    emit('lazyImageReady', $slideEl[0], $imageEl[0]);

                    if (swiper.params.autoHeight) {
                        swiper.updateAutoHeight();
                    }
                });
                emit('lazyImageLoad', $slideEl[0], $imageEl[0]);
            });
        }

        function load() {
            const {
                $wrapperEl,
                params: swiperParams,
                slides,
                activeIndex
            } = swiper;
            const isVirtual = swiper.virtual && swiperParams.virtual.enabled;
            const params = swiperParams.lazy;
            let slidesPerView = swiperParams.slidesPerView;

            if (slidesPerView === 'auto') {
                slidesPerView = 0;
            }

            function slideExist(index) {
                if (isVirtual) {
                    if ($wrapperEl.children(`.${swiperParams.slideClass}[data-swiper-slide-index="${index}"]`).length) {
                        return true;
                    }
                } else if (slides[index]) return true;

                return false;
            }

            function slideIndex(slideEl) {
                if (isVirtual) {
                    return $(slideEl).attr('data-swiper-slide-index');
                }

                return $(slideEl).index();
            }

            if (!initialImageLoaded) initialImageLoaded = true;

            if (swiper.params.watchSlidesProgress) {
                $wrapperEl.children(`.${swiperParams.slideVisibleClass}`).each(slideEl => {
                    const index = isVirtual ? $(slideEl).attr('data-swiper-slide-index') : $(slideEl).index();
                    loadInSlide(index);
                });
            } else if (slidesPerView > 1) {
                for (let i = activeIndex; i < activeIndex + slidesPerView; i += 1) {
                    if (slideExist(i)) loadInSlide(i);
                }
            } else {
                loadInSlide(activeIndex);
            }

            if (params.loadPrevNext) {
                if (slidesPerView > 1 || params.loadPrevNextAmount && params.loadPrevNextAmount > 1) {
                    const amount = params.loadPrevNextAmount;
                    const spv = slidesPerView;
                    const maxIndex = Math.min(activeIndex + spv + Math.max(amount, spv), slides.length);
                    const minIndex = Math.max(activeIndex - Math.max(spv, amount), 0); // Next Slides

                    for (let i = activeIndex + slidesPerView; i < maxIndex; i += 1) {
                        if (slideExist(i)) loadInSlide(i);
                    } // Prev Slides


                    for (let i = minIndex; i < activeIndex; i += 1) {
                        if (slideExist(i)) loadInSlide(i);
                    }
                } else {
                    const nextSlide = $wrapperEl.children(`.${swiperParams.slideNextClass}`);
                    if (nextSlide.length > 0) loadInSlide(slideIndex(nextSlide));
                    const prevSlide = $wrapperEl.children(`.${swiperParams.slidePrevClass}`);
                    if (prevSlide.length > 0) loadInSlide(slideIndex(prevSlide));
                }
            }
        }

        function checkInViewOnLoad() {
            const window = getWindow();
            if (!swiper || swiper.destroyed) return;
            const $scrollElement = swiper.params.lazy.scrollingElement ? $(swiper.params.lazy.scrollingElement) : $(window);
            const isWindow = $scrollElement[0] === window;
            const scrollElementWidth = isWindow ? window.innerWidth : $scrollElement[0].offsetWidth;
            const scrollElementHeight = isWindow ? window.innerHeight : $scrollElement[0].offsetHeight;
            const swiperOffset = swiper.$el.offset();
            const {
                rtlTranslate: rtl
            } = swiper;
            let inView = false;
            if (rtl) swiperOffset.left -= swiper.$el[0].scrollLeft;
            const swiperCoord = [[swiperOffset.left, swiperOffset.top], [swiperOffset.left + swiper.width, swiperOffset.top], [swiperOffset.left, swiperOffset.top + swiper.height], [swiperOffset.left + swiper.width, swiperOffset.top + swiper.height]];

            for (let i = 0; i < swiperCoord.length; i += 1) {
                const point = swiperCoord[i];

                if (point[0] >= 0 && point[0] <= scrollElementWidth && point[1] >= 0 && point[1] <= scrollElementHeight) {
                    if (point[0] === 0 && point[1] === 0) continue; // eslint-disable-line

                    inView = true;
                }
            }

            const passiveListener = swiper.touchEvents.start === 'touchstart' && swiper.support.passiveListener && swiper.params.passiveListeners ? {
                passive: true,
                capture: false
            } : false;

            if (inView) {
                load();
                $scrollElement.off('scroll', checkInViewOnLoad, passiveListener);
            } else if (!scrollHandlerAttached) {
                scrollHandlerAttached = true;
                $scrollElement.on('scroll', checkInViewOnLoad, passiveListener);
            }
        }

        on('beforeInit', () => {
            if (swiper.params.lazy.enabled && swiper.params.preloadImages) {
                swiper.params.preloadImages = false;
            }
        });
        on('init', () => {
            if (swiper.params.lazy.enabled) {
                if (swiper.params.lazy.checkInView) {
                    checkInViewOnLoad();
                } else {
                    load();
                }
            }
        });
        on('scroll', () => {
            if (swiper.params.freeMode && swiper.params.freeMode.enabled && !swiper.params.freeMode.sticky) {
                load();
            }
        });
        on('scrollbarDragMove resize _freeModeNoMomentumRelease', () => {
            if (swiper.params.lazy.enabled) {
                if (swiper.params.lazy.checkInView) {
                    checkInViewOnLoad();
                } else {
                    load();
                }
            }
        });
        on('transitionStart', () => {
            if (swiper.params.lazy.enabled) {
                if (swiper.params.lazy.loadOnTransitionStart || !swiper.params.lazy.loadOnTransitionStart && !initialImageLoaded) {
                    if (swiper.params.lazy.checkInView) {
                        checkInViewOnLoad();
                    } else {
                        load();
                    }
                }
            }
        });
        on('transitionEnd', () => {
            if (swiper.params.lazy.enabled && !swiper.params.lazy.loadOnTransitionStart) {
                if (swiper.params.lazy.checkInView) {
                    checkInViewOnLoad();
                } else {
                    load();
                }
            }
        });
        on('slideChange', () => {
            const {
                lazy,
                cssMode,
                watchSlidesProgress,
                touchReleaseOnEdges,
                resistanceRatio
            } = swiper.params;

            if (lazy.enabled && (cssMode || watchSlidesProgress && (touchReleaseOnEdges || resistanceRatio === 0))) {
                load();
            }
        });
        Object.assign(swiper.lazy, {
            load,
            loadInSlide
        });
    }

    /* eslint no-bitwise: ["error", { "allow": [">>"] }] */
    function Controller(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            controller: {
                control: undefined,
                inverse: false,
                by: 'slide' // or 'container'

            }
        });
        swiper.controller = {
            control: undefined
        };

        function LinearSpline(x, y) {
            const binarySearch = function search() {
                let maxIndex;
                let minIndex;
                let guess;
                return (array, val) => {
                    minIndex = -1;
                    maxIndex = array.length;

                    while (maxIndex - minIndex > 1) {
                        guess = maxIndex + minIndex >> 1;

                        if (array[guess] <= val) {
                            minIndex = guess;
                        } else {
                            maxIndex = guess;
                        }
                    }

                    return maxIndex;
                };
            }();

            this.x = x;
            this.y = y;
            this.lastIndex = x.length - 1; // Given an x value (x2), return the expected y2 value:
            // (x1,y1) is the known point before given value,
            // (x3,y3) is the known point after given value.

            let i1;
            let i3;

            this.interpolate = function interpolate(x2) {
                if (!x2) return 0; // Get the indexes of x1 and x3 (the array indexes before and after given x2):

                i3 = binarySearch(this.x, x2);
                i1 = i3 - 1; // We have our indexes i1 & i3, so we can calculate already:
                // y2 := ((x2−x1) × (y3−y1)) ÷ (x3−x1) + y1

                return (x2 - this.x[i1]) * (this.y[i3] - this.y[i1]) / (this.x[i3] - this.x[i1]) + this.y[i1];
            };

            return this;
        } // xxx: for now i will just save one spline function to to


        function getInterpolateFunction(c) {
            if (!swiper.controller.spline) {
                swiper.controller.spline = swiper.params.loop ? new LinearSpline(swiper.slidesGrid, c.slidesGrid) : new LinearSpline(swiper.snapGrid, c.snapGrid);
            }
        }

        function setTranslate(_t, byController) {
            const controlled = swiper.controller.control;
            let multiplier;
            let controlledTranslate;
            const Swiper = swiper.constructor;

            function setControlledTranslate(c) {
                // this will create an Interpolate function based on the snapGrids
                // x is the Grid of the scrolled scroller and y will be the controlled scroller
                // it makes sense to create this only once and recall it for the interpolation
                // the function does a lot of value caching for performance
                const translate = swiper.rtlTranslate ? -swiper.translate : swiper.translate;

                if (swiper.params.controller.by === 'slide') {
                    getInterpolateFunction(c); // i am not sure why the values have to be multiplicated this way, tried to invert the snapGrid
                    // but it did not work out

                    controlledTranslate = -swiper.controller.spline.interpolate(-translate);
                }

                if (!controlledTranslate || swiper.params.controller.by === 'container') {
                    multiplier = (c.maxTranslate() - c.minTranslate()) / (swiper.maxTranslate() - swiper.minTranslate());
                    controlledTranslate = (translate - swiper.minTranslate()) * multiplier + c.minTranslate();
                }

                if (swiper.params.controller.inverse) {
                    controlledTranslate = c.maxTranslate() - controlledTranslate;
                }

                c.updateProgress(controlledTranslate);
                c.setTranslate(controlledTranslate, swiper);
                c.updateActiveIndex();
                c.updateSlidesClasses();
            }

            if (Array.isArray(controlled)) {
                for (let i = 0; i < controlled.length; i += 1) {
                    if (controlled[i] !== byController && controlled[i] instanceof Swiper) {
                        setControlledTranslate(controlled[i]);
                    }
                }
            } else if (controlled instanceof Swiper && byController !== controlled) {
                setControlledTranslate(controlled);
            }
        }

        function setTransition(duration, byController) {
            const Swiper = swiper.constructor;
            const controlled = swiper.controller.control;
            let i;

            function setControlledTransition(c) {
                c.setTransition(duration, swiper);

                if (duration !== 0) {
                    c.transitionStart();

                    if (c.params.autoHeight) {
                        nextTick(() => {
                            c.updateAutoHeight();
                        });
                    }

                    c.$wrapperEl.transitionEnd(() => {
                        if (!controlled) return;

                        if (c.params.loop && swiper.params.controller.by === 'slide') {
                            c.loopFix();
                        }

                        c.transitionEnd();
                    });
                }
            }

            if (Array.isArray(controlled)) {
                for (i = 0; i < controlled.length; i += 1) {
                    if (controlled[i] !== byController && controlled[i] instanceof Swiper) {
                        setControlledTransition(controlled[i]);
                    }
                }
            } else if (controlled instanceof Swiper && byController !== controlled) {
                setControlledTransition(controlled);
            }
        }

        function removeSpline() {
            if (!swiper.controller.control) return;

            if (swiper.controller.spline) {
                swiper.controller.spline = undefined;
                delete swiper.controller.spline;
            }
        }

        on('beforeInit', () => {
            swiper.controller.control = swiper.params.controller.control;
        });
        on('update', () => {
            removeSpline();
        });
        on('resize', () => {
            removeSpline();
        });
        on('observerUpdate', () => {
            removeSpline();
        });
        on('setTranslate', (_s, translate, byController) => {
            if (!swiper.controller.control) return;
            swiper.controller.setTranslate(translate, byController);
        });
        on('setTransition', (_s, duration, byController) => {
            if (!swiper.controller.control) return;
            swiper.controller.setTransition(duration, byController);
        });
        Object.assign(swiper.controller, {
            setTranslate,
            setTransition
        });
    }

    function A11y(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            a11y: {
                enabled: true,
                notificationClass: 'ax-notification',
                prevSlideMessage: '上一个',//AXUI
                nextSlideMessage: '下一个',//AXUI
                firstSlideMessage: '这是第一个',//AXUI
                lastSlideMessage: '这是最后一个',//AXUI
                paginationBulletMessage: '跳转到第{{index}}页',//AXUI
                slideLabelMessage: '第{{index}}/{{slidesLength}}页',//AXUI
                containerMessage: null,
                containerRoleDescriptionMessage: null,
                itemRoleDescriptionMessage: null,
                slideRole: 'group',
                id: null
            }
        });
        let liveRegion = null;

        function notify(message) {
            const notification = liveRegion;
            if (notification.length === 0) return;
            notification.html('');
            notification.html(message);
        }

        function getRandomNumber(size) {
            if (size === void 0) {
                size = 16;
            }

            const randomChar = () => Math.round(16 * Math.random()).toString(16);

            return 'x'.repeat(size).replace(/x/g, randomChar);
        }

        function makeElFocusable($el) {
            $el.attr('tabIndex', '0');
        }

        function makeElNotFocusable($el) {
            $el.attr('tabIndex', '-1');
        }

        function addElRole($el, role) {
            $el.attr('role', role);
        }

        function addElRoleDescription($el, description) {
            $el.attr('aria-roledescription', description);
        }

        function addElControls($el, controls) {
            $el.attr('aria-controls', controls);
        }

        function addElLabel($el, label) {
            $el.attr('aria-label', label);
        }

        function addElId($el, id) {
            $el.attr('id', id);
        }

        function addElLive($el, live) {
            $el.attr('aria-live', live);
        }

        function disableEl($el) {
            $el.attr('aria-disabled', true);
        }

        function enableEl($el) {
            $el.attr('aria-disabled', false);
        }

        function onEnterOrSpaceKey(e) {
            if (e.keyCode !== 13 && e.keyCode !== 32) return;
            const params = swiper.params.a11y;
            const $targetEl = $(e.target);

            if (swiper.navigation && swiper.navigation.$nextEl && $targetEl.is(swiper.navigation.$nextEl)) {
                if (!(swiper.isEnd && !swiper.params.loop)) {
                    swiper.slideNext();
                }

                if (swiper.isEnd) {
                    notify(params.lastSlideMessage);
                } else {
                    notify(params.nextSlideMessage);
                }
            }

            if (swiper.navigation && swiper.navigation.$prevEl && $targetEl.is(swiper.navigation.$prevEl)) {
                if (!(swiper.isBeginning && !swiper.params.loop)) {
                    swiper.slidePrev();
                }

                if (swiper.isBeginning) {
                    notify(params.firstSlideMessage);
                } else {
                    notify(params.prevSlideMessage);
                }
            }

            if (swiper.pagination && $targetEl.is(classesToSelector(swiper.params.pagination.bulletClass))) {
                $targetEl[0].click();
            }
        }

        function updateNavigation() {
            if (swiper.params.loop || swiper.params.rewind || !swiper.navigation) return;
            const {
                $nextEl,
                $prevEl
            } = swiper.navigation;

            if ($prevEl && $prevEl.length > 0) {
                if (swiper.isBeginning) {
                    disableEl($prevEl);
                    makeElNotFocusable($prevEl);
                } else {
                    enableEl($prevEl);
                    makeElFocusable($prevEl);
                }
            }

            if ($nextEl && $nextEl.length > 0) {
                if (swiper.isEnd) {
                    disableEl($nextEl);
                    makeElNotFocusable($nextEl);
                } else {
                    enableEl($nextEl);
                    makeElFocusable($nextEl);
                }
            }
        }

        function hasPagination() {
            return swiper.pagination && swiper.pagination.bullets && swiper.pagination.bullets.length;
        }

        function hasClickablePagination() {
            return hasPagination() && swiper.params.pagination.clickable;
        }

        function updatePagination() {
            const params = swiper.params.a11y;
            if (!hasPagination()) return;
            swiper.pagination.bullets.each(bulletEl => {
                const $bulletEl = $(bulletEl);

                if (swiper.params.pagination.clickable) {
                    makeElFocusable($bulletEl);

                    if (!swiper.params.pagination.renderBullet) {
                        addElRole($bulletEl, 'button');
                        addElLabel($bulletEl, params.paginationBulletMessage.replace(/\{\{index\}\}/, $bulletEl.index() + 1));
                    }
                }

                if ($bulletEl.is(`.${swiper.params.pagination.bulletActiveClass}`)) {
                    $bulletEl.attr('aria-current', 'true');
                } else {
                    $bulletEl.removeAttr('aria-current');
                }
            });
        }

        const initNavEl = ($el, wrapperId, message) => {
            makeElFocusable($el);

            if ($el[0].tagName !== 'BUTTON') {
                addElRole($el, 'button');
                $el.on('keydown', onEnterOrSpaceKey);
            }

            addElLabel($el, message);
            addElControls($el, wrapperId);
        };

        const handleFocus = e => {
            const slideEl = e.target.closest(`.${swiper.params.slideClass}`);
            if (!slideEl || !swiper.slides.includes(slideEl)) return;
            const isActive = swiper.slides.indexOf(slideEl) === swiper.activeIndex;
            const isVisible = swiper.params.watchSlidesProgress && swiper.visibleSlides && swiper.visibleSlides.includes(slideEl);
            if (isActive || isVisible) return;
            swiper.slideTo(swiper.slides.indexOf(slideEl), 0);
        };

        function init() {
            const params = swiper.params.a11y;
            swiper.$el.append(liveRegion); // Container

            const $containerEl = swiper.$el;

            if (params.containerRoleDescriptionMessage) {
                addElRoleDescription($containerEl, params.containerRoleDescriptionMessage);
            }

            if (params.containerMessage) {
                addElLabel($containerEl, params.containerMessage);
            } // Wrapper


            const $wrapperEl = swiper.$wrapperEl;
            const wrapperId = params.id || $wrapperEl.attr('id') || `ax-wrapper-${getRandomNumber(16)}`;//AXUI,原`swiper-wrapper-${getRandomNumber(16)}`
            const live = swiper.params.autoplay && swiper.params.autoplay.enabled ? 'off' : 'polite';
            addElId($wrapperEl, wrapperId);
            addElLive($wrapperEl, live); // Slide

            if (params.itemRoleDescriptionMessage) {
                addElRoleDescription($(swiper.slides), params.itemRoleDescriptionMessage);
            }

            addElRole($(swiper.slides), params.slideRole);
            const slidesLength = swiper.params.loop ? swiper.slides.filter(el => !el.classList.contains(swiper.params.slideDuplicateClass)).length : swiper.slides.length;
            swiper.slides.each((slideEl, index) => {
                const $slideEl = $(slideEl);
                const slideIndex = swiper.params.loop ? parseInt($slideEl.attr('data-swiper-slide-index'), 10) : index;
                const ariaLabelMessage = params.slideLabelMessage.replace(/\{\{index\}\}/, slideIndex + 1).replace(/\{\{slidesLength\}\}/, slidesLength);
                addElLabel($slideEl, ariaLabelMessage);
            }); // Navigation

            let $nextEl;
            let $prevEl;

            if (swiper.navigation && swiper.navigation.$nextEl) {
                $nextEl = swiper.navigation.$nextEl;
            }

            if (swiper.navigation && swiper.navigation.$prevEl) {
                $prevEl = swiper.navigation.$prevEl;
            }

            if ($nextEl && $nextEl.length) {
                initNavEl($nextEl, wrapperId, params.nextSlideMessage);
            }

            if ($prevEl && $prevEl.length) {
                initNavEl($prevEl, wrapperId, params.prevSlideMessage);
            } // Pagination


            if (hasClickablePagination()) {
                swiper.pagination.$el.on('keydown', classesToSelector(swiper.params.pagination.bulletClass), onEnterOrSpaceKey);
            } // Tab focus


            swiper.$el.on('focus', handleFocus, true);
        }

        function destroy() {
            if (liveRegion && liveRegion.length > 0) liveRegion.remove();
            let $nextEl;
            let $prevEl;

            if (swiper.navigation && swiper.navigation.$nextEl) {
                $nextEl = swiper.navigation.$nextEl;
            }

            if (swiper.navigation && swiper.navigation.$prevEl) {
                $prevEl = swiper.navigation.$prevEl;
            }

            if ($nextEl) {
                $nextEl.off('keydown', onEnterOrSpaceKey);
            }

            if ($prevEl) {
                $prevEl.off('keydown', onEnterOrSpaceKey);
            } // Pagination


            if (hasClickablePagination()) {
                swiper.pagination.$el.off('keydown', classesToSelector(swiper.params.pagination.bulletClass), onEnterOrSpaceKey);
            } // Tab focus


            swiper.$el.off('focus', handleFocus, true);
        }

        on('beforeInit', () => {
            liveRegion = $(`<span class="${swiper.params.a11y.notificationClass}" aria-live="assertive" aria-atomic="true"></span>`);
        });
        on('afterInit', () => {
            if (!swiper.params.a11y.enabled) return;
            init();
        });
        on('fromEdge toEdge afterInit lock unlock', () => {
            if (!swiper.params.a11y.enabled) return;
            updateNavigation();
        });
        on('paginationUpdate', () => {
            if (!swiper.params.a11y.enabled) return;
            updatePagination();
        });
        on('destroy', () => {
            if (!swiper.params.a11y.enabled) return;
            destroy();
        });
    }

    function History(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            history: {
                enabled: false,
                root: '',
                replaceState: false,
                key: 'slides'
            }
        });
        let initialized = false;
        let paths = {};

        const slugify = text => {
            return text.toString().replace(/\s+/g, '-').replace(/[^\w-]+/g, '').replace(/--+/g, '-').replace(/^-+/, '').replace(/-+$/, '');
        };

        const getPathValues = urlOverride => {
            const window = getWindow();
            let location;

            if (urlOverride) {
                location = new URL(urlOverride);
            } else {
                location = window.location;
            }

            const pathArray = location.pathname.slice(1).split('/').filter(part => part !== '');
            const total = pathArray.length;
            const key = pathArray[total - 2];
            const value = pathArray[total - 1];
            return {
                key,
                value
            };
        };

        const setHistory = (key, index) => {
            const window = getWindow();
            if (!initialized || !swiper.params.history.enabled) return;
            let location;

            if (swiper.params.url) {
                location = new URL(swiper.params.url);
            } else {
                location = window.location;
            }

            const slide = swiper.slides.eq(index);
            let value = slugify(slide.attr('data-history'));

            if (swiper.params.history.root.length > 0) {
                let root = swiper.params.history.root;
                if (root[root.length - 1] === '/') root = root.slice(0, root.length - 1);
                value = `${root}/${key}/${value}`;
            } else if (!location.pathname.includes(key)) {
                value = `${key}/${value}`;
            }

            const currentState = window.history.state;

            if (currentState && currentState.value === value) {
                return;
            }

            if (swiper.params.history.replaceState) {
                window.history.replaceState({
                    value
                }, null, value);
            } else {
                window.history.pushState({
                    value
                }, null, value);
            }
        };

        const scrollToSlide = (speed, value, runCallbacks) => {
            if (value) {
                for (let i = 0, length = swiper.slides.length; i < length; i += 1) {
                    const slide = swiper.slides.eq(i);
                    const slideHistory = slugify(slide.attr('data-history'));

                    if (slideHistory === value && !slide.hasClass(swiper.params.slideDuplicateClass)) {
                        const index = slide.index();
                        swiper.slideTo(index, speed, runCallbacks);
                    }
                }
            } else {
                swiper.slideTo(0, speed, runCallbacks);
            }
        };

        const setHistoryPopState = () => {
            paths = getPathValues(swiper.params.url);
            scrollToSlide(swiper.params.speed, swiper.paths.value, false);
        };

        const init = () => {
            const window = getWindow();
            if (!swiper.params.history) return;

            if (!window.history || !window.history.pushState) {
                swiper.params.history.enabled = false;
                swiper.params.hashNavigation.enabled = true;
                return;
            }

            initialized = true;
            paths = getPathValues(swiper.params.url);
            if (!paths.key && !paths.value) return;
            scrollToSlide(0, paths.value, swiper.params.runCallbacksOnInit);

            if (!swiper.params.history.replaceState) {
                window.addEventListener('popstate', setHistoryPopState);
            }
        };

        const destroy = () => {
            const window = getWindow();

            if (!swiper.params.history.replaceState) {
                window.removeEventListener('popstate', setHistoryPopState);
            }
        };

        on('init', () => {
            if (swiper.params.history.enabled) {
                init();
            }
        });
        on('destroy', () => {
            if (swiper.params.history.enabled) {
                destroy();
            }
        });
        on('transitionEnd _freeModeNoMomentumRelease', () => {
            if (initialized) {
                setHistory(swiper.params.history.key, swiper.activeIndex);
            }
        });
        on('slideChange', () => {
            if (initialized && swiper.params.cssMode) {
                setHistory(swiper.params.history.key, swiper.activeIndex);
            }
        });
    }

    function HashNavigation(_ref) {
        let {
            swiper,
            extendParams,
            emit,
            on
        } = _ref;
        let initialized = false;
        const document = getDocument();
        const window = getWindow();
        extendParams({
            hashNavigation: {
                enabled: false,
                replaceState: false,
                watchState: false
            }
        });

        const onHashChange = () => {
            emit('hashChange');
            const newHash = document.location.hash.replace('#', '');
            const activeSlideHash = swiper.slides.eq(swiper.activeIndex).attr('data-hash');

            if (newHash !== activeSlideHash) {
                const newIndex = swiper.$wrapperEl.children(`.${swiper.params.slideClass}[data-hash="${newHash}"]`).index();
                if (typeof newIndex === 'undefined') return;
                swiper.slideTo(newIndex);
            }
        };

        const setHash = () => {
            if (!initialized || !swiper.params.hashNavigation.enabled) return;

            if (swiper.params.hashNavigation.replaceState && window.history && window.history.replaceState) {
                window.history.replaceState(null, null, `#${swiper.slides.eq(swiper.activeIndex).attr('data-hash')}` || '');
                emit('hashSet');
            } else {
                const slide = swiper.slides.eq(swiper.activeIndex);
                const hash = slide.attr('data-hash') || slide.attr('data-history');
                document.location.hash = hash || '';
                emit('hashSet');
            }
        };

        const init = () => {
            if (!swiper.params.hashNavigation.enabled || swiper.params.history && swiper.params.history.enabled) return;
            initialized = true;
            const hash = document.location.hash.replace('#', '');

            if (hash) {
                const speed = 0;

                for (let i = 0, length = swiper.slides.length; i < length; i += 1) {
                    const slide = swiper.slides.eq(i);
                    const slideHash = slide.attr('data-hash') || slide.attr('data-history');

                    if (slideHash === hash && !slide.hasClass(swiper.params.slideDuplicateClass)) {
                        const index = slide.index();
                        swiper.slideTo(index, speed, swiper.params.runCallbacksOnInit, true);
                    }
                }
            }

            if (swiper.params.hashNavigation.watchState) {
                $(window).on('hashchange', onHashChange);
            }
        };

        const destroy = () => {
            if (swiper.params.hashNavigation.watchState) {
                $(window).off('hashchange', onHashChange);
            }
        };

        on('init', () => {
            if (swiper.params.hashNavigation.enabled) {
                init();
            }
        });
        on('destroy', () => {
            if (swiper.params.hashNavigation.enabled) {
                destroy();
            }
        });
        on('transitionEnd _freeModeNoMomentumRelease', () => {
            if (initialized) {
                setHash();
            }
        });
        on('slideChange', () => {
            if (initialized && swiper.params.cssMode) {
                setHash();
            }
        });
    }

    /* eslint no-underscore-dangle: "off" */
    function Autoplay(_ref) {
        let {
            swiper,
            extendParams,
            on,
            emit
        } = _ref;
        let timeout;
        swiper.autoplay = {
            running: false,
            paused: false
        };
        extendParams({
            autoplay: {
                enabled: false,
                delay: 3000,
                waitForTransition: true,
                disableOnInteraction: true,
                stopOnLastSlide: false,
                reverseDirection: false,
                pauseOnMouseEnter: false
            }
        });

        function run() {
            const $activeSlideEl = swiper.slides.eq(swiper.activeIndex);
            let delay = swiper.params.autoplay.delay;

            if ($activeSlideEl.attr('data-swiper-autoplay')) {
                delay = $activeSlideEl.attr('data-swiper-autoplay') || swiper.params.autoplay.delay;
            }

            clearTimeout(timeout);
            timeout = nextTick(() => {
                let autoplayResult;

                if (swiper.params.autoplay.reverseDirection) {
                    if (swiper.params.loop) {
                        swiper.loopFix();
                        autoplayResult = swiper.slidePrev(swiper.params.speed, true, true);
                        emit('autoplay');
                    } else if (!swiper.isBeginning) {
                        autoplayResult = swiper.slidePrev(swiper.params.speed, true, true);
                        emit('autoplay');
                    } else if (!swiper.params.autoplay.stopOnLastSlide) {
                        autoplayResult = swiper.slideTo(swiper.slides.length - 1, swiper.params.speed, true, true);
                        emit('autoplay');
                    } else {
                        stop();
                    }
                } else if (swiper.params.loop) {
                    swiper.loopFix();
                    autoplayResult = swiper.slideNext(swiper.params.speed, true, true);
                    emit('autoplay');
                } else if (!swiper.isEnd) {
                    autoplayResult = swiper.slideNext(swiper.params.speed, true, true);
                    emit('autoplay');
                } else if (!swiper.params.autoplay.stopOnLastSlide) {
                    autoplayResult = swiper.slideTo(0, swiper.params.speed, true, true);
                    emit('autoplay');
                } else {
                    stop();
                }

                if (swiper.params.cssMode && swiper.autoplay.running) run(); else if (autoplayResult === false) {
                    run();
                }
            }, delay);
        }

        function start() {
            if (typeof timeout !== 'undefined') return false;
            if (swiper.autoplay.running) return false;
            swiper.autoplay.running = true;
            emit('autoplayStart');
            run();
            return true;
        }

        function stop() {
            if (!swiper.autoplay.running) return false;
            if (typeof timeout === 'undefined') return false;

            if (timeout) {
                clearTimeout(timeout);
                timeout = undefined;
            }

            swiper.autoplay.running = false;
            emit('autoplayStop');
            return true;
        }

        function pause(speed) {
            if (!swiper.autoplay.running) return;
            if (swiper.autoplay.paused) return;
            if (timeout) clearTimeout(timeout);
            swiper.autoplay.paused = true;

            if (speed === 0 || !swiper.params.autoplay.waitForTransition) {
                swiper.autoplay.paused = false;
                run();
            } else {
                ['transitionend', 'webkitTransitionEnd'].forEach(event => {
                    swiper.$wrapperEl[0].addEventListener(event, onTransitionEnd);
                });
            }
        }

        function onVisibilityChange() {
            const document = getDocument();

            if (document.visibilityState === 'hidden' && swiper.autoplay.running) {
                pause();
            }

            if (document.visibilityState === 'visible' && swiper.autoplay.paused) {
                run();
                swiper.autoplay.paused = false;
            }
        }

        function onTransitionEnd(e) {
            if (!swiper || swiper.destroyed || !swiper.$wrapperEl) return;
            if (e.target !== swiper.$wrapperEl[0]) return;
            ['transitionend', 'webkitTransitionEnd'].forEach(event => {
                swiper.$wrapperEl[0].removeEventListener(event, onTransitionEnd);
            });
            swiper.autoplay.paused = false;

            if (!swiper.autoplay.running) {
                stop();
            } else {
                run();
            }
        }

        function onMouseEnter() {
            if (swiper.params.autoplay.disableOnInteraction) {
                stop();
            } else {
                emit('autoplayPause');
                pause();
            }

            ['transitionend', 'webkitTransitionEnd'].forEach(event => {
                swiper.$wrapperEl[0].removeEventListener(event, onTransitionEnd);
            });
        }

        function onMouseLeave() {
            if (swiper.params.autoplay.disableOnInteraction) {
                return;
            }

            swiper.autoplay.paused = false;
            emit('autoplayResume');
            run();
        }

        function attachMouseEvents() {
            if (swiper.params.autoplay.pauseOnMouseEnter) {
                swiper.$el.on('mouseenter', onMouseEnter);
                swiper.$el.on('mouseleave', onMouseLeave);
            }
        }

        function detachMouseEvents() {
            swiper.$el.off('mouseenter', onMouseEnter);
            swiper.$el.off('mouseleave', onMouseLeave);
        }

        on('init', () => {
            if (swiper.params.autoplay.enabled) {
                start();
                const document = getDocument();
                document.addEventListener('visibilitychange', onVisibilityChange);
                attachMouseEvents();
            }
        });
        on('beforeTransitionStart', (_s, speed, internal) => {
            if (swiper.autoplay.running) {
                if (internal || !swiper.params.autoplay.disableOnInteraction) {
                    swiper.autoplay.pause(speed);
                } else {
                    stop();
                }
            }
        });
        on('sliderFirstMove', () => {
            if (swiper.autoplay.running) {
                if (swiper.params.autoplay.disableOnInteraction) {
                    stop();
                } else {
                    pause();
                }
            }
        });
        on('touchEnd', () => {
            if (swiper.params.cssMode && swiper.autoplay.paused && !swiper.params.autoplay.disableOnInteraction) {
                run();
            }
        });
        on('destroy', () => {
            detachMouseEvents();

            if (swiper.autoplay.running) {
                stop();
            }

            const document = getDocument();
            document.removeEventListener('visibilitychange', onVisibilityChange);
        });
        Object.assign(swiper.autoplay, {
            pause,
            run,
            start,
            stop
        });
    }

    function Thumb(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            thumbs: {
                swiper: null,
                multipleActiveThumbs: true,
                autoScrollOffset: 0,
                slideThumbActiveClass: 'ax-slide-thumb-active',
                thumbsContainerClass: 'ax-thumbs'
            }
        });
        let initialized = false;
        let swiperCreated = false;
        swiper.thumbs = {
            swiper: null
        };

        function onThumbClick() {
            const thumbsSwiper = swiper.thumbs.swiper;
            if (!thumbsSwiper || thumbsSwiper.destroyed) return;
            const clickedIndex = thumbsSwiper.clickedIndex;
            const clickedSlide = thumbsSwiper.clickedSlide;
            if (clickedSlide && $(clickedSlide).hasClass(swiper.params.thumbs.slideThumbActiveClass)) return;
            if (typeof clickedIndex === 'undefined' || clickedIndex === null) return;
            let slideToIndex;

            if (thumbsSwiper.params.loop) {
                slideToIndex = parseInt($(thumbsSwiper.clickedSlide).attr('data-swiper-slide-index'), 10);
            } else {
                slideToIndex = clickedIndex;
            }

            if (swiper.params.loop) {
                let currentIndex = swiper.activeIndex;

                if (swiper.slides.eq(currentIndex).hasClass(swiper.params.slideDuplicateClass)) {
                    swiper.loopFix(); // eslint-disable-next-line

                    swiper._clientLeft = swiper.$wrapperEl[0].clientLeft;
                    currentIndex = swiper.activeIndex;
                }

                const prevIndex = swiper.slides.eq(currentIndex).prevAll(`[data-swiper-slide-index="${slideToIndex}"]`).eq(0).index();
                const nextIndex = swiper.slides.eq(currentIndex).nextAll(`[data-swiper-slide-index="${slideToIndex}"]`).eq(0).index();
                if (typeof prevIndex === 'undefined') slideToIndex = nextIndex; else if (typeof nextIndex === 'undefined') slideToIndex = prevIndex; else if (nextIndex - currentIndex < currentIndex - prevIndex) slideToIndex = nextIndex; else slideToIndex = prevIndex;
            }

            swiper.slideTo(slideToIndex);
        }

        function init() {
            const {
                thumbs: thumbsParams
            } = swiper.params;
            if (initialized) return false;
            initialized = true;
            const SwiperClass = swiper.constructor;

            if (thumbsParams.swiper instanceof SwiperClass) {
                swiper.thumbs.swiper = thumbsParams.swiper;
                Object.assign(swiper.thumbs.swiper.originalParams, {
                    watchSlidesProgress: true,
                    slideToClickedSlide: false
                });
                Object.assign(swiper.thumbs.swiper.params, {
                    watchSlidesProgress: true,
                    slideToClickedSlide: false
                });
            } else if (isObject(thumbsParams.swiper)) {
                const thumbsSwiperParams = Object.assign({}, thumbsParams.swiper);
                Object.assign(thumbsSwiperParams, {
                    watchSlidesProgress: true,
                    slideToClickedSlide: false
                });
                swiper.thumbs.swiper = new SwiperClass(thumbsSwiperParams);
                swiperCreated = true;
            }

            swiper.thumbs.swiper.$el.addClass(swiper.params.thumbs.thumbsContainerClass);
            swiper.thumbs.swiper.on('tap', onThumbClick);
            return true;
        }

        function update(initial) {
            const thumbsSwiper = swiper.thumbs.swiper;
            if (!thumbsSwiper || thumbsSwiper.destroyed) return;
            const slidesPerView = thumbsSwiper.params.slidesPerView === 'auto' ? thumbsSwiper.slidesPerViewDynamic() : thumbsSwiper.params.slidesPerView;
            const autoScrollOffset = swiper.params.thumbs.autoScrollOffset;
            const useOffset = autoScrollOffset && !thumbsSwiper.params.loop;

            if (swiper.realIndex !== thumbsSwiper.realIndex || useOffset) {
                let currentThumbsIndex = thumbsSwiper.activeIndex;
                let newThumbsIndex;
                let direction;

                if (thumbsSwiper.params.loop) {
                    if (thumbsSwiper.slides.eq(currentThumbsIndex).hasClass(thumbsSwiper.params.slideDuplicateClass)) {
                        thumbsSwiper.loopFix(); // eslint-disable-next-line

                        thumbsSwiper._clientLeft = thumbsSwiper.$wrapperEl[0].clientLeft;
                        currentThumbsIndex = thumbsSwiper.activeIndex;
                    } // Find actual thumbs index to slide to


                    const prevThumbsIndex = thumbsSwiper.slides.eq(currentThumbsIndex).prevAll(`[data-swiper-slide-index="${swiper.realIndex}"]`).eq(0).index();
                    const nextThumbsIndex = thumbsSwiper.slides.eq(currentThumbsIndex).nextAll(`[data-swiper-slide-index="${swiper.realIndex}"]`).eq(0).index();

                    if (typeof prevThumbsIndex === 'undefined') {
                        newThumbsIndex = nextThumbsIndex;
                    } else if (typeof nextThumbsIndex === 'undefined') {
                        newThumbsIndex = prevThumbsIndex;
                    } else if (nextThumbsIndex - currentThumbsIndex === currentThumbsIndex - prevThumbsIndex) {
                        newThumbsIndex = thumbsSwiper.params.slidesPerGroup > 1 ? nextThumbsIndex : currentThumbsIndex;
                    } else if (nextThumbsIndex - currentThumbsIndex < currentThumbsIndex - prevThumbsIndex) {
                        newThumbsIndex = nextThumbsIndex;
                    } else {
                        newThumbsIndex = prevThumbsIndex;
                    }

                    direction = swiper.activeIndex > swiper.previousIndex ? 'next' : 'prev';
                } else {
                    newThumbsIndex = swiper.realIndex;
                    direction = newThumbsIndex > swiper.previousIndex ? 'next' : 'prev';
                }

                if (useOffset) {
                    newThumbsIndex += direction === 'next' ? autoScrollOffset : -1 * autoScrollOffset;
                }

                if (thumbsSwiper.visibleSlidesIndexes && thumbsSwiper.visibleSlidesIndexes.indexOf(newThumbsIndex) < 0) {
                    if (thumbsSwiper.params.centeredSlides) {
                        if (newThumbsIndex > currentThumbsIndex) {
                            newThumbsIndex = newThumbsIndex - Math.floor(slidesPerView / 2) + 1;
                        } else {
                            newThumbsIndex = newThumbsIndex + Math.floor(slidesPerView / 2) - 1;
                        }
                    } else if (newThumbsIndex > currentThumbsIndex && thumbsSwiper.params.slidesPerGroup === 1);

                    thumbsSwiper.slideTo(newThumbsIndex, initial ? 0 : undefined);
                }
            } // Activate thumbs


            let thumbsToActivate = 1;
            const thumbActiveClass = swiper.params.thumbs.slideThumbActiveClass;

            if (swiper.params.slidesPerView > 1 && !swiper.params.centeredSlides) {
                thumbsToActivate = swiper.params.slidesPerView;
            }

            if (!swiper.params.thumbs.multipleActiveThumbs) {
                thumbsToActivate = 1;
            }

            thumbsToActivate = Math.floor(thumbsToActivate);
            thumbsSwiper.slides.removeClass(thumbActiveClass);

            if (thumbsSwiper.params.loop || thumbsSwiper.params.virtual && thumbsSwiper.params.virtual.enabled) {
                for (let i = 0; i < thumbsToActivate; i += 1) {
                    thumbsSwiper.$wrapperEl.children(`[data-swiper-slide-index="${swiper.realIndex + i}"]`).addClass(thumbActiveClass);
                }
            } else {
                for (let i = 0; i < thumbsToActivate; i += 1) {
                    thumbsSwiper.slides.eq(swiper.realIndex + i).addClass(thumbActiveClass);
                }
            }
        }

        on('beforeInit', () => {
            const {
                thumbs
            } = swiper.params;
            if (!thumbs || !thumbs.swiper) return;
            init();
            update(true);
        });
        on('slideChange update resize observerUpdate', () => {
            update();
        });
        on('setTransition', (_s, duration) => {
            const thumbsSwiper = swiper.thumbs.swiper;
            if (!thumbsSwiper || thumbsSwiper.destroyed) return;
            thumbsSwiper.setTransition(duration);
        });
        on('beforeDestroy', () => {
            const thumbsSwiper = swiper.thumbs.swiper;
            if (!thumbsSwiper || thumbsSwiper.destroyed) return;

            if (swiperCreated) {
                thumbsSwiper.destroy();
            }
        });
        Object.assign(swiper.thumbs, {
            init,
            update
        });
    }

    function freeMode(_ref) {
        let {
            swiper,
            extendParams,
            emit,
            once
        } = _ref;
        extendParams({
            freeMode: {
                enabled: false,
                momentum: true,
                momentumRatio: 1,
                momentumBounce: true,
                momentumBounceRatio: 1,
                momentumVelocityRatio: 1,
                sticky: false,
                minimumVelocity: 0.02
            }
        });

        function onTouchStart() {
            const translate = swiper.getTranslate();
            swiper.setTranslate(translate);
            swiper.setTransition(0);
            swiper.touchEventsData.velocities.length = 0;
            swiper.freeMode.onTouchEnd({
                currentPos: swiper.rtl ? swiper.translate : -swiper.translate
            });
        }

        function onTouchMove() {
            const {
                touchEventsData: data,
                touches
            } = swiper; // Velocity

            if (data.velocities.length === 0) {
                data.velocities.push({
                    position: touches[swiper.isHorizontal() ? 'startX' : 'startY'],
                    time: data.touchStartTime
                });
            }

            data.velocities.push({
                position: touches[swiper.isHorizontal() ? 'currentX' : 'currentY'],
                time: now()
            });
        }

        function onTouchEnd(_ref2) {
            let {
                currentPos
            } = _ref2;
            const {
                params,
                $wrapperEl,
                rtlTranslate: rtl,
                snapGrid,
                touchEventsData: data
            } = swiper; // Time diff

            const touchEndTime = now();
            const timeDiff = touchEndTime - data.touchStartTime;

            if (currentPos < -swiper.minTranslate()) {
                swiper.slideTo(swiper.activeIndex);
                return;
            }

            if (currentPos > -swiper.maxTranslate()) {
                if (swiper.slides.length < snapGrid.length) {
                    swiper.slideTo(snapGrid.length - 1);
                } else {
                    swiper.slideTo(swiper.slides.length - 1);
                }

                return;
            }

            if (params.freeMode.momentum) {
                if (data.velocities.length > 1) {
                    const lastMoveEvent = data.velocities.pop();
                    const velocityEvent = data.velocities.pop();
                    const distance = lastMoveEvent.position - velocityEvent.position;
                    const time = lastMoveEvent.time - velocityEvent.time;
                    swiper.velocity = distance / time;
                    swiper.velocity /= 2;

                    if (Math.abs(swiper.velocity) < params.freeMode.minimumVelocity) {
                        swiper.velocity = 0;
                    } // this implies that the user stopped moving a finger then released.
                    // There would be no events with distance zero, so the last event is stale.


                    if (time > 150 || now() - lastMoveEvent.time > 300) {
                        swiper.velocity = 0;
                    }
                } else {
                    swiper.velocity = 0;
                }

                swiper.velocity *= params.freeMode.momentumVelocityRatio;
                data.velocities.length = 0;
                let momentumDuration = 1000 * params.freeMode.momentumRatio;
                const momentumDistance = swiper.velocity * momentumDuration;
                let newPosition = swiper.translate + momentumDistance;
                if (rtl) newPosition = -newPosition;
                let doBounce = false;
                let afterBouncePosition;
                const bounceAmount = Math.abs(swiper.velocity) * 20 * params.freeMode.momentumBounceRatio;
                let needsLoopFix;

                if (newPosition < swiper.maxTranslate()) {
                    if (params.freeMode.momentumBounce) {
                        if (newPosition + swiper.maxTranslate() < -bounceAmount) {
                            newPosition = swiper.maxTranslate() - bounceAmount;
                        }

                        afterBouncePosition = swiper.maxTranslate();
                        doBounce = true;
                        data.allowMomentumBounce = true;
                    } else {
                        newPosition = swiper.maxTranslate();
                    }

                    if (params.loop && params.centeredSlides) needsLoopFix = true;
                } else if (newPosition > swiper.minTranslate()) {
                    if (params.freeMode.momentumBounce) {
                        if (newPosition - swiper.minTranslate() > bounceAmount) {
                            newPosition = swiper.minTranslate() + bounceAmount;
                        }

                        afterBouncePosition = swiper.minTranslate();
                        doBounce = true;
                        data.allowMomentumBounce = true;
                    } else {
                        newPosition = swiper.minTranslate();
                    }

                    if (params.loop && params.centeredSlides) needsLoopFix = true;
                } else if (params.freeMode.sticky) {
                    let nextSlide;

                    for (let j = 0; j < snapGrid.length; j += 1) {
                        if (snapGrid[j] > -newPosition) {
                            nextSlide = j;
                            break;
                        }
                    }

                    if (Math.abs(snapGrid[nextSlide] - newPosition) < Math.abs(snapGrid[nextSlide - 1] - newPosition) || swiper.swipeDirection === 'next') {
                        newPosition = snapGrid[nextSlide];
                    } else {
                        newPosition = snapGrid[nextSlide - 1];
                    }

                    newPosition = -newPosition;
                }

                if (needsLoopFix) {
                    once('transitionEnd', () => {
                        swiper.loopFix();
                    });
                } // Fix duration


                if (swiper.velocity !== 0) {
                    if (rtl) {
                        momentumDuration = Math.abs((-newPosition - swiper.translate) / swiper.velocity);
                    } else {
                        momentumDuration = Math.abs((newPosition - swiper.translate) / swiper.velocity);
                    }

                    if (params.freeMode.sticky) {
                        // If freeMode.sticky is active and the user ends a swipe with a slow-velocity
                        // event, then durations can be 20+ seconds to slide one (or zero!) slides.
                        // It's easy to see this when simulating touch with mouse events. To fix this,
                        // limit single-slide swipes to the default slide duration. This also has the
                        // nice side effect of matching slide speed if the user stopped moving before
                        // lifting finger or mouse vs. moving slowly before lifting the finger/mouse.
                        // For faster swipes, also apply limits (albeit higher ones).
                        const moveDistance = Math.abs((rtl ? -newPosition : newPosition) - swiper.translate);
                        const currentSlideSize = swiper.slidesSizesGrid[swiper.activeIndex];

                        if (moveDistance < currentSlideSize) {
                            momentumDuration = params.speed;
                        } else if (moveDistance < 2 * currentSlideSize) {
                            momentumDuration = params.speed * 1.5;
                        } else {
                            momentumDuration = params.speed * 2.5;
                        }
                    }
                } else if (params.freeMode.sticky) {
                    swiper.slideToClosest();
                    return;
                }

                if (params.freeMode.momentumBounce && doBounce) {
                    swiper.updateProgress(afterBouncePosition);
                    swiper.setTransition(momentumDuration);
                    swiper.setTranslate(newPosition);
                    swiper.transitionStart(true, swiper.swipeDirection);
                    swiper.animating = true;
                    $wrapperEl.transitionEnd(() => {
                        if (!swiper || swiper.destroyed || !data.allowMomentumBounce) return;
                        emit('momentumBounce');
                        swiper.setTransition(params.speed);
                        setTimeout(() => {
                            swiper.setTranslate(afterBouncePosition);
                            $wrapperEl.transitionEnd(() => {
                                if (!swiper || swiper.destroyed) return;
                                swiper.transitionEnd();
                            });
                        }, 0);
                    });
                } else if (swiper.velocity) {
                    emit('_freeModeNoMomentumRelease');
                    swiper.updateProgress(newPosition);
                    swiper.setTransition(momentumDuration);
                    swiper.setTranslate(newPosition);
                    swiper.transitionStart(true, swiper.swipeDirection);

                    if (!swiper.animating) {
                        swiper.animating = true;
                        $wrapperEl.transitionEnd(() => {
                            if (!swiper || swiper.destroyed) return;
                            swiper.transitionEnd();
                        });
                    }
                } else {
                    swiper.updateProgress(newPosition);
                }

                swiper.updateActiveIndex();
                swiper.updateSlidesClasses();
            } else if (params.freeMode.sticky) {
                swiper.slideToClosest();
                return;
            } else if (params.freeMode) {
                emit('_freeModeNoMomentumRelease');
            }

            if (!params.freeMode.momentum || timeDiff >= params.longSwipesMs) {
                swiper.updateProgress();
                swiper.updateActiveIndex();
                swiper.updateSlidesClasses();
            }
        }

        Object.assign(swiper, {
            freeMode: {
                onTouchStart,
                onTouchMove,
                onTouchEnd
            }
        });
    }

    function Grid(_ref) {
        let {
            swiper,
            extendParams
        } = _ref;
        extendParams({
            grid: {
                rows: 1,
                fill: 'column'
            }
        });
        let slidesNumberEvenToRows;
        let slidesPerRow;
        let numFullColumns;

        const initSlides = slidesLength => {
            const {
                slidesPerView
            } = swiper.params;
            const {
                rows,
                fill
            } = swiper.params.grid;
            slidesPerRow = slidesNumberEvenToRows / rows;
            numFullColumns = Math.floor(slidesLength / rows);

            if (Math.floor(slidesLength / rows) === slidesLength / rows) {
                slidesNumberEvenToRows = slidesLength;
            } else {
                slidesNumberEvenToRows = Math.ceil(slidesLength / rows) * rows;
            }

            if (slidesPerView !== 'auto' && fill === 'row') {
                slidesNumberEvenToRows = Math.max(slidesNumberEvenToRows, slidesPerView * rows);
            }
        };

        const updateSlide = (i, slide, slidesLength, getDirectionLabel) => {
            const {
                slidesPerGroup,
                spaceBetween
            } = swiper.params;
            const {
                rows,
                fill
            } = swiper.params.grid; // Set slides order

            let newSlideOrderIndex;
            let column;
            let row;

            if (fill === 'row' && slidesPerGroup > 1) {
                const groupIndex = Math.floor(i / (slidesPerGroup * rows));
                const slideIndexInGroup = i - rows * slidesPerGroup * groupIndex;
                const columnsInGroup = groupIndex === 0 ? slidesPerGroup : Math.min(Math.ceil((slidesLength - groupIndex * rows * slidesPerGroup) / rows), slidesPerGroup);
                row = Math.floor(slideIndexInGroup / columnsInGroup);
                column = slideIndexInGroup - row * columnsInGroup + groupIndex * slidesPerGroup;
                newSlideOrderIndex = column + row * slidesNumberEvenToRows / rows;
                slide.css({
                    '-webkit-order': newSlideOrderIndex,
                    order: newSlideOrderIndex
                });
            } else if (fill === 'column') {
                column = Math.floor(i / rows);
                row = i - column * rows;

                if (column > numFullColumns || column === numFullColumns && row === rows - 1) {
                    row += 1;

                    if (row >= rows) {
                        row = 0;
                        column += 1;
                    }
                }
            } else {
                row = Math.floor(i / slidesPerRow);
                column = i - row * slidesPerRow;
            }

            slide.css(getDirectionLabel('margin-top'), row !== 0 ? spaceBetween && `${spaceBetween}px` : '');
        };

        const updateWrapperSize = (slideSize, snapGrid, getDirectionLabel) => {
            const {
                spaceBetween,
                centeredSlides,
                roundLengths
            } = swiper.params;
            const {
                rows
            } = swiper.params.grid;
            swiper.virtualSize = (slideSize + spaceBetween) * slidesNumberEvenToRows;
            swiper.virtualSize = Math.ceil(swiper.virtualSize / rows) - spaceBetween;
            swiper.$wrapperEl.css({
                [getDirectionLabel('width')]: `${swiper.virtualSize + spaceBetween}px`
            });

            if (centeredSlides) {
                snapGrid.splice(0, snapGrid.length);
                const newSlidesGrid = [];

                for (let i = 0; i < snapGrid.length; i += 1) {
                    let slidesGridItem = snapGrid[i];
                    if (roundLengths) slidesGridItem = Math.floor(slidesGridItem);
                    if (snapGrid[i] < swiper.virtualSize + snapGrid[0]) newSlidesGrid.push(slidesGridItem);
                }

                snapGrid.push(...newSlidesGrid);
            }
        };

        swiper.grid = {
            initSlides,
            updateSlide,
            updateWrapperSize
        };
    }

    function appendSlide(slides) {
        const swiper = this;
        const {
            $wrapperEl,
            params
        } = swiper;

        if (params.loop) {
            swiper.loopDestroy();
        }

        if (typeof slides === 'object' && 'length' in slides) {
            for (let i = 0; i < slides.length; i += 1) {
                if (slides[i]) $wrapperEl.append(slides[i]);
            }
        } else {
            $wrapperEl.append(slides);
        }

        if (params.loop) {
            swiper.loopCreate();
        }

        if (!params.observer) {
            swiper.update();
        }
    }

    function prependSlide(slides) {
        const swiper = this;
        const {
            params,
            $wrapperEl,
            activeIndex
        } = swiper;

        if (params.loop) {
            swiper.loopDestroy();
        }

        let newActiveIndex = activeIndex + 1;

        if (typeof slides === 'object' && 'length' in slides) {
            for (let i = 0; i < slides.length; i += 1) {
                if (slides[i]) $wrapperEl.prepend(slides[i]);
            }

            newActiveIndex = activeIndex + slides.length;
        } else {
            $wrapperEl.prepend(slides);
        }

        if (params.loop) {
            swiper.loopCreate();
        }

        if (!params.observer) {
            swiper.update();
        }

        swiper.slideTo(newActiveIndex, 0, false);
    }

    function addSlide(index, slides) {
        const swiper = this;
        const {
            $wrapperEl,
            params,
            activeIndex
        } = swiper;
        let activeIndexBuffer = activeIndex;

        if (params.loop) {
            activeIndexBuffer -= swiper.loopedSlides;
            swiper.loopDestroy();
            swiper.slides = $wrapperEl.children(`.${params.slideClass}`);
        }

        const baseLength = swiper.slides.length;

        if (index <= 0) {
            swiper.prependSlide(slides);
            return;
        }

        if (index >= baseLength) {
            swiper.appendSlide(slides);
            return;
        }

        let newActiveIndex = activeIndexBuffer > index ? activeIndexBuffer + 1 : activeIndexBuffer;
        const slidesBuffer = [];

        for (let i = baseLength - 1; i >= index; i -= 1) {
            const currentSlide = swiper.slides.eq(i);
            currentSlide.remove();
            slidesBuffer.unshift(currentSlide);
        }

        if (typeof slides === 'object' && 'length' in slides) {
            for (let i = 0; i < slides.length; i += 1) {
                if (slides[i]) $wrapperEl.append(slides[i]);
            }

            newActiveIndex = activeIndexBuffer > index ? activeIndexBuffer + slides.length : activeIndexBuffer;
        } else {
            $wrapperEl.append(slides);
        }

        for (let i = 0; i < slidesBuffer.length; i += 1) {
            $wrapperEl.append(slidesBuffer[i]);
        }

        if (params.loop) {
            swiper.loopCreate();
        }

        if (!params.observer) {
            swiper.update();
        }

        if (params.loop) {
            swiper.slideTo(newActiveIndex + swiper.loopedSlides, 0, false);
        } else {
            swiper.slideTo(newActiveIndex, 0, false);
        }
    }

    function removeSlide(slidesIndexes) {
        const swiper = this;
        const {
            params,
            $wrapperEl,
            activeIndex
        } = swiper;
        let activeIndexBuffer = activeIndex;

        if (params.loop) {
            activeIndexBuffer -= swiper.loopedSlides;
            swiper.loopDestroy();
            swiper.slides = $wrapperEl.children(`.${params.slideClass}`);
        }

        let newActiveIndex = activeIndexBuffer;
        let indexToRemove;

        if (typeof slidesIndexes === 'object' && 'length' in slidesIndexes) {
            for (let i = 0; i < slidesIndexes.length; i += 1) {
                indexToRemove = slidesIndexes[i];
                if (swiper.slides[indexToRemove]) swiper.slides.eq(indexToRemove).remove();
                if (indexToRemove < newActiveIndex) newActiveIndex -= 1;
            }

            newActiveIndex = Math.max(newActiveIndex, 0);
        } else {
            indexToRemove = slidesIndexes;
            if (swiper.slides[indexToRemove]) swiper.slides.eq(indexToRemove).remove();
            if (indexToRemove < newActiveIndex) newActiveIndex -= 1;
            newActiveIndex = Math.max(newActiveIndex, 0);
        }

        if (params.loop) {
            swiper.loopCreate();
        }

        if (!params.observer) {
            swiper.update();
        }

        if (params.loop) {
            swiper.slideTo(newActiveIndex + swiper.loopedSlides, 0, false);
        } else {
            swiper.slideTo(newActiveIndex, 0, false);
        }
    }

    function removeAllSlides() {
        const swiper = this;
        const slidesIndexes = [];

        for (let i = 0; i < swiper.slides.length; i += 1) {
            slidesIndexes.push(i);
        }

        swiper.removeSlide(slidesIndexes);
    }

    function Manipulation(_ref) {
        let {
            swiper
        } = _ref;
        Object.assign(swiper, {
            appendSlide: appendSlide.bind(swiper),
            prependSlide: prependSlide.bind(swiper),
            addSlide: addSlide.bind(swiper),
            removeSlide: removeSlide.bind(swiper),
            removeAllSlides: removeAllSlides.bind(swiper)
        });
    }

    function effectInit(params) {
        const {
            effect,
            swiper,
            on,
            setTranslate,
            setTransition,
            overwriteParams,
            perspective,
            recreateShadows,
            getEffectParams
        } = params;
        on('beforeInit', () => {
            if (swiper.params.effect !== effect) return;
            swiper.classNames.push(`${swiper.params.containerModifierClass}${effect}`);

            if (perspective && perspective()) {
                swiper.classNames.push(`${swiper.params.containerModifierClass}3d`);
            }

            const overwriteParamsResult = overwriteParams ? overwriteParams() : {};
            Object.assign(swiper.params, overwriteParamsResult);
            Object.assign(swiper.originalParams, overwriteParamsResult);
        });
        on('setTranslate', () => {
            if (swiper.params.effect !== effect) return;
            setTranslate();
        });
        on('setTransition', (_s, duration) => {
            if (swiper.params.effect !== effect) return;
            setTransition(duration);
        });
        on('transitionEnd', () => {
            if (swiper.params.effect !== effect) return;

            if (recreateShadows) {
                if (!getEffectParams || !getEffectParams().slideShadows) return; // remove shadows

                swiper.slides.each(slideEl => {
                    const $slideEl = swiper.$(slideEl);
                    $slideEl.find('.swiper-slide-shadow-top, .swiper-slide-shadow-right, .swiper-slide-shadow-bottom, .swiper-slide-shadow-left').remove();
                }); // create new one

                recreateShadows();
            }
        });
        let requireUpdateOnVirtual;
        on('virtualUpdate', () => {
            if (swiper.params.effect !== effect) return;

            if (!swiper.slides.length) {
                requireUpdateOnVirtual = true;
            }

            requestAnimationFrame(() => {
                if (requireUpdateOnVirtual && swiper.slides && swiper.slides.length) {
                    setTranslate();
                    requireUpdateOnVirtual = false;
                }
            });
        });
    }

    function effectTarget(effectParams, $slideEl) {
        if (effectParams.transformEl) {
            return $slideEl.find(effectParams.transformEl).css({
                'backface-visibility': 'hidden',
                '-webkit-backface-visibility': 'hidden'
            });
        }

        return $slideEl;
    }

    function effectVirtualTransitionEnd(_ref) {
        let {
            swiper,
            duration,
            transformEl,
            allSlides
        } = _ref;
        const {
            slides,
            activeIndex,
            $wrapperEl
        } = swiper;

        if (swiper.params.virtualTranslate && duration !== 0) {
            let eventTriggered = false;
            let $transitionEndTarget;

            if (allSlides) {
                $transitionEndTarget = transformEl ? slides.find(transformEl) : slides;
            } else {
                $transitionEndTarget = transformEl ? slides.eq(activeIndex).find(transformEl) : slides.eq(activeIndex);
            }

            $transitionEndTarget.transitionEnd(() => {
                if (eventTriggered) return;
                if (!swiper || swiper.destroyed) return;
                eventTriggered = true;
                swiper.animating = false;
                const triggerEvents = ['webkitTransitionEnd', 'transitionend'];

                for (let i = 0; i < triggerEvents.length; i += 1) {
                    $wrapperEl.trigger(triggerEvents[i]);
                }
            });
        }
    }

    function EffectFade(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            fadeEffect: {
                crossFade: false,
                transformEl: null
            }
        });

        const setTranslate = () => {
            const {
                slides
            } = swiper;
            const params = swiper.params.fadeEffect;

            for (let i = 0; i < slides.length; i += 1) {
                const $slideEl = swiper.slides.eq(i);
                const offset = $slideEl[0].swiperSlideOffset;
                let tx = -offset;
                if (!swiper.params.virtualTranslate) tx -= swiper.translate;
                let ty = 0;

                if (!swiper.isHorizontal()) {
                    ty = tx;
                    tx = 0;
                }

                const slideOpacity = swiper.params.fadeEffect.crossFade ? Math.max(1 - Math.abs($slideEl[0].progress), 0) : 1 + Math.min(Math.max($slideEl[0].progress, -1), 0);
                const $targetEl = effectTarget(params, $slideEl);
                $targetEl.css({
                    opacity: slideOpacity
                }).transform(`translate3d(${tx}px, ${ty}px, 0px)`);
            }
        };

        const setTransition = duration => {
            const {
                transformEl
            } = swiper.params.fadeEffect;
            const $transitionElements = transformEl ? swiper.slides.find(transformEl) : swiper.slides;
            $transitionElements.transition(duration);
            effectVirtualTransitionEnd({
                swiper,
                duration,
                transformEl,
                allSlides: true
            });
        };

        effectInit({
            effect: 'fade',
            swiper,
            on,
            setTranslate,
            setTransition,
            overwriteParams: () => ({
                slidesPerView: 1,
                slidesPerGroup: 1,
                watchSlidesProgress: true,
                spaceBetween: 0,
                virtualTranslate: !swiper.params.cssMode
            })
        });
    }

    function EffectCube(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            cubeEffect: {
                slideShadows: true,
                shadow: true,
                shadowOffset: 20,
                shadowScale: 0.94
            }
        });

        const createSlideShadows = ($slideEl, progress, isHorizontal) => {
            let shadowBefore = isHorizontal ? $slideEl.find('.swiper-slide-shadow-left') : $slideEl.find('.swiper-slide-shadow-top');
            let shadowAfter = isHorizontal ? $slideEl.find('.swiper-slide-shadow-right') : $slideEl.find('.swiper-slide-shadow-bottom');

            if (shadowBefore.length === 0) {
                shadowBefore = $(`<div class="swiper-slide-shadow-${isHorizontal ? 'left' : 'top'}"></div>`);
                $slideEl.append(shadowBefore);
            }

            if (shadowAfter.length === 0) {
                shadowAfter = $(`<div class="swiper-slide-shadow-${isHorizontal ? 'right' : 'bottom'}"></div>`);
                $slideEl.append(shadowAfter);
            }

            if (shadowBefore.length) shadowBefore[0].style.opacity = Math.max(-progress, 0);
            if (shadowAfter.length) shadowAfter[0].style.opacity = Math.max(progress, 0);
        };

        const recreateShadows = () => {
            // create new ones
            const isHorizontal = swiper.isHorizontal();
            swiper.slides.each(slideEl => {
                const progress = Math.max(Math.min(slideEl.progress, 1), -1);
                createSlideShadows($(slideEl), progress, isHorizontal);
            });
        };

        const setTranslate = () => {
            const {
                $el,
                $wrapperEl,
                slides,
                width: swiperWidth,
                height: swiperHeight,
                rtlTranslate: rtl,
                size: swiperSize,
                browser
            } = swiper;
            const params = swiper.params.cubeEffect;
            const isHorizontal = swiper.isHorizontal();
            const isVirtual = swiper.virtual && swiper.params.virtual.enabled;
            let wrapperRotate = 0;
            let $cubeShadowEl;

            if (params.shadow) {
                if (isHorizontal) {
                    $cubeShadowEl = $wrapperEl.find('.swiper-cube-shadow');

                    if ($cubeShadowEl.length === 0) {
                        $cubeShadowEl = $('<div class="swiper-cube-shadow"></div>');
                        $wrapperEl.append($cubeShadowEl);
                    }

                    $cubeShadowEl.css({
                        height: `${swiperWidth}px`
                    });
                } else {
                    $cubeShadowEl = $el.find('.swiper-cube-shadow');

                    if ($cubeShadowEl.length === 0) {
                        $cubeShadowEl = $('<div class="swiper-cube-shadow"></div>');
                        $el.append($cubeShadowEl);
                    }
                }
            }

            for (let i = 0; i < slides.length; i += 1) {
                const $slideEl = slides.eq(i);
                let slideIndex = i;

                if (isVirtual) {
                    slideIndex = parseInt($slideEl.attr('data-swiper-slide-index'), 10);
                }

                let slideAngle = slideIndex * 90;
                let round = Math.floor(slideAngle / 360);

                if (rtl) {
                    slideAngle = -slideAngle;
                    round = Math.floor(-slideAngle / 360);
                }

                const progress = Math.max(Math.min($slideEl[0].progress, 1), -1);
                let tx = 0;
                let ty = 0;
                let tz = 0;

                if (slideIndex % 4 === 0) {
                    tx = -round * 4 * swiperSize;
                    tz = 0;
                } else if ((slideIndex - 1) % 4 === 0) {
                    tx = 0;
                    tz = -round * 4 * swiperSize;
                } else if ((slideIndex - 2) % 4 === 0) {
                    tx = swiperSize + round * 4 * swiperSize;
                    tz = swiperSize;
                } else if ((slideIndex - 3) % 4 === 0) {
                    tx = -swiperSize;
                    tz = 3 * swiperSize + swiperSize * 4 * round;
                }

                if (rtl) {
                    tx = -tx;
                }

                if (!isHorizontal) {
                    ty = tx;
                    tx = 0;
                }

                const transform = `rotateX(${isHorizontal ? 0 : -slideAngle}deg) rotateY(${isHorizontal ? slideAngle : 0}deg) translate3d(${tx}px, ${ty}px, ${tz}px)`;

                if (progress <= 1 && progress > -1) {
                    wrapperRotate = slideIndex * 90 + progress * 90;
                    if (rtl) wrapperRotate = -slideIndex * 90 - progress * 90;
                }

                $slideEl.transform(transform);

                if (params.slideShadows) {
                    createSlideShadows($slideEl, progress, isHorizontal);
                }
            }

            $wrapperEl.css({
                '-webkit-transform-origin': `50% 50% -${swiperSize / 2}px`,
                'transform-origin': `50% 50% -${swiperSize / 2}px`
            });

            if (params.shadow) {
                if (isHorizontal) {
                    $cubeShadowEl.transform(`translate3d(0px, ${swiperWidth / 2 + params.shadowOffset}px, ${-swiperWidth / 2}px) rotateX(90deg) rotateZ(0deg) scale(${params.shadowScale})`);
                } else {
                    const shadowAngle = Math.abs(wrapperRotate) - Math.floor(Math.abs(wrapperRotate) / 90) * 90;
                    const multiplier = 1.5 - (Math.sin(shadowAngle * 2 * Math.PI / 360) / 2 + Math.cos(shadowAngle * 2 * Math.PI / 360) / 2);
                    const scale1 = params.shadowScale;
                    const scale2 = params.shadowScale / multiplier;
                    const offset = params.shadowOffset;
                    $cubeShadowEl.transform(`scale3d(${scale1}, 1, ${scale2}) translate3d(0px, ${swiperHeight / 2 + offset}px, ${-swiperHeight / 2 / scale2}px) rotateX(-90deg)`);
                }
            }

            const zFactor = browser.isSafari || browser.isWebView ? -swiperSize / 2 : 0;
            $wrapperEl.transform(`translate3d(0px,0,${zFactor}px) rotateX(${swiper.isHorizontal() ? 0 : wrapperRotate}deg) rotateY(${swiper.isHorizontal() ? -wrapperRotate : 0}deg)`);
            $wrapperEl[0].style.setProperty('--swiper-cube-translate-z', `${zFactor}px`);
        };

        const setTransition = duration => {
            const {
                $el,
                slides
            } = swiper;
            slides.transition(duration).find('.swiper-slide-shadow-top, .swiper-slide-shadow-right, .swiper-slide-shadow-bottom, .swiper-slide-shadow-left').transition(duration);

            if (swiper.params.cubeEffect.shadow && !swiper.isHorizontal()) {
                $el.find('.swiper-cube-shadow').transition(duration);
            }
        };

        effectInit({
            effect: 'cube',
            swiper,
            on,
            setTranslate,
            setTransition,
            recreateShadows,
            getEffectParams: () => swiper.params.cubeEffect,
            perspective: () => true,
            overwriteParams: () => ({
                slidesPerView: 1,
                slidesPerGroup: 1,
                watchSlidesProgress: true,
                resistanceRatio: 0,
                spaceBetween: 0,
                centeredSlides: false,
                virtualTranslate: true
            })
        });
    }

    function createShadow(params, $slideEl, side) {
        const shadowClass = `swiper-slide-shadow${side ? `-${side}` : ''}`;
        const $shadowContainer = params.transformEl ? $slideEl.find(params.transformEl) : $slideEl;
        let $shadowEl = $shadowContainer.children(`.${shadowClass}`);

        if (!$shadowEl.length) {
            $shadowEl = $(`<div class="swiper-slide-shadow${side ? `-${side}` : ''}"></div>`);
            $shadowContainer.append($shadowEl);
        }

        return $shadowEl;
    }

    function EffectFlip(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            flipEffect: {
                slideShadows: true,
                limitRotation: true,
                transformEl: null
            }
        });

        const createSlideShadows = ($slideEl, progress, params) => {
            let shadowBefore = swiper.isHorizontal() ? $slideEl.find('.swiper-slide-shadow-left') : $slideEl.find('.swiper-slide-shadow-top');
            let shadowAfter = swiper.isHorizontal() ? $slideEl.find('.swiper-slide-shadow-right') : $slideEl.find('.swiper-slide-shadow-bottom');

            if (shadowBefore.length === 0) {
                shadowBefore = createShadow(params, $slideEl, swiper.isHorizontal() ? 'left' : 'top');
            }

            if (shadowAfter.length === 0) {
                shadowAfter = createShadow(params, $slideEl, swiper.isHorizontal() ? 'right' : 'bottom');
            }

            if (shadowBefore.length) shadowBefore[0].style.opacity = Math.max(-progress, 0);
            if (shadowAfter.length) shadowAfter[0].style.opacity = Math.max(progress, 0);
        };

        const recreateShadows = () => {
            // Set shadows
            const params = swiper.params.flipEffect;
            swiper.slides.each(slideEl => {
                const $slideEl = $(slideEl);
                let progress = $slideEl[0].progress;

                if (swiper.params.flipEffect.limitRotation) {
                    progress = Math.max(Math.min(slideEl.progress, 1), -1);
                }

                createSlideShadows($slideEl, progress, params);
            });
        };

        const setTranslate = () => {
            const {
                slides,
                rtlTranslate: rtl
            } = swiper;
            const params = swiper.params.flipEffect;

            for (let i = 0; i < slides.length; i += 1) {
                const $slideEl = slides.eq(i);
                let progress = $slideEl[0].progress;

                if (swiper.params.flipEffect.limitRotation) {
                    progress = Math.max(Math.min($slideEl[0].progress, 1), -1);
                }

                const offset = $slideEl[0].swiperSlideOffset;
                const rotate = -180 * progress;
                let rotateY = rotate;
                let rotateX = 0;
                let tx = swiper.params.cssMode ? -offset - swiper.translate : -offset;
                let ty = 0;

                if (!swiper.isHorizontal()) {
                    ty = tx;
                    tx = 0;
                    rotateX = -rotateY;
                    rotateY = 0;
                } else if (rtl) {
                    rotateY = -rotateY;
                }

                $slideEl[0].style.zIndex = -Math.abs(Math.round(progress)) + slides.length;

                if (params.slideShadows) {
                    createSlideShadows($slideEl, progress, params);
                }

                const transform = `translate3d(${tx}px, ${ty}px, 0px) rotateX(${rotateX}deg) rotateY(${rotateY}deg)`;
                const $targetEl = effectTarget(params, $slideEl);
                $targetEl.transform(transform);
            }
        };

        const setTransition = duration => {
            const {
                transformEl
            } = swiper.params.flipEffect;
            const $transitionElements = transformEl ? swiper.slides.find(transformEl) : swiper.slides;
            $transitionElements.transition(duration).find('.swiper-slide-shadow-top, .swiper-slide-shadow-right, .swiper-slide-shadow-bottom, .swiper-slide-shadow-left').transition(duration);
            effectVirtualTransitionEnd({
                swiper,
                duration,
                transformEl
            });
        };

        effectInit({
            effect: 'flip',
            swiper,
            on,
            setTranslate,
            setTransition,
            recreateShadows,
            getEffectParams: () => swiper.params.flipEffect,
            perspective: () => true,
            overwriteParams: () => ({
                slidesPerView: 1,
                slidesPerGroup: 1,
                watchSlidesProgress: true,
                spaceBetween: 0,
                virtualTranslate: !swiper.params.cssMode
            })
        });
    }

    function EffectCoverflow(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            coverflowEffect: {
                rotate: 50,
                stretch: 0,
                depth: 100,
                scale: 1,
                modifier: 1,
                slideShadows: true,
                transformEl: null
            }
        });

        const setTranslate = () => {
            const {
                width: swiperWidth,
                height: swiperHeight,
                slides,
                slidesSizesGrid
            } = swiper;
            const params = swiper.params.coverflowEffect;
            const isHorizontal = swiper.isHorizontal();
            const transform = swiper.translate;
            const center = isHorizontal ? -transform + swiperWidth / 2 : -transform + swiperHeight / 2;
            const rotate = isHorizontal ? params.rotate : -params.rotate;
            const translate = params.depth; // Each slide offset from center

            for (let i = 0, length = slides.length; i < length; i += 1) {
                const $slideEl = slides.eq(i);
                const slideSize = slidesSizesGrid[i];
                const slideOffset = $slideEl[0].swiperSlideOffset;
                const centerOffset = (center - slideOffset - slideSize / 2) / slideSize;
                const offsetMultiplier = typeof params.modifier === 'function' ? params.modifier(centerOffset) : centerOffset * params.modifier;
                let rotateY = isHorizontal ? rotate * offsetMultiplier : 0;
                let rotateX = isHorizontal ? 0 : rotate * offsetMultiplier; // var rotateZ = 0

                let translateZ = -translate * Math.abs(offsetMultiplier);
                let stretch = params.stretch; // Allow percentage to make a relative stretch for responsive sliders

                if (typeof stretch === 'string' && stretch.indexOf('%') !== -1) {
                    stretch = parseFloat(params.stretch) / 100 * slideSize;
                }

                let translateY = isHorizontal ? 0 : stretch * offsetMultiplier;
                let translateX = isHorizontal ? stretch * offsetMultiplier : 0;
                let scale = 1 - (1 - params.scale) * Math.abs(offsetMultiplier); // Fix for ultra small values

                if (Math.abs(translateX) < 0.001) translateX = 0;
                if (Math.abs(translateY) < 0.001) translateY = 0;
                if (Math.abs(translateZ) < 0.001) translateZ = 0;
                if (Math.abs(rotateY) < 0.001) rotateY = 0;
                if (Math.abs(rotateX) < 0.001) rotateX = 0;
                if (Math.abs(scale) < 0.001) scale = 0;
                const slideTransform = `translate3d(${translateX}px,${translateY}px,${translateZ}px)  rotateX(${rotateX}deg) rotateY(${rotateY}deg) scale(${scale})`;
                const $targetEl = effectTarget(params, $slideEl);
                $targetEl.transform(slideTransform);
                $slideEl[0].style.zIndex = -Math.abs(Math.round(offsetMultiplier)) + 1;

                if (params.slideShadows) {
                    // Set shadows
                    let $shadowBeforeEl = isHorizontal ? $slideEl.find('.swiper-slide-shadow-left') : $slideEl.find('.swiper-slide-shadow-top');
                    let $shadowAfterEl = isHorizontal ? $slideEl.find('.swiper-slide-shadow-right') : $slideEl.find('.swiper-slide-shadow-bottom');

                    if ($shadowBeforeEl.length === 0) {
                        $shadowBeforeEl = createShadow(params, $slideEl, isHorizontal ? 'left' : 'top');
                    }

                    if ($shadowAfterEl.length === 0) {
                        $shadowAfterEl = createShadow(params, $slideEl, isHorizontal ? 'right' : 'bottom');
                    }

                    if ($shadowBeforeEl.length) $shadowBeforeEl[0].style.opacity = offsetMultiplier > 0 ? offsetMultiplier : 0;
                    if ($shadowAfterEl.length) $shadowAfterEl[0].style.opacity = -offsetMultiplier > 0 ? -offsetMultiplier : 0;
                }
            }
        };

        const setTransition = duration => {
            const {
                transformEl
            } = swiper.params.coverflowEffect;
            const $transitionElements = transformEl ? swiper.slides.find(transformEl) : swiper.slides;
            $transitionElements.transition(duration).find('.swiper-slide-shadow-top, .swiper-slide-shadow-right, .swiper-slide-shadow-bottom, .swiper-slide-shadow-left').transition(duration);
        };

        effectInit({
            effect: 'coverflow',
            swiper,
            on,
            setTranslate,
            setTransition,
            perspective: () => true,
            overwriteParams: () => ({
                watchSlidesProgress: true
            })
        });
    }

    function EffectCreative(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            creativeEffect: {
                transformEl: null,
                limitProgress: 1,
                shadowPerProgress: false,
                progressMultiplier: 1,
                perspective: true,
                prev: {
                    translate: [0, 0, 0],
                    rotate: [0, 0, 0],
                    opacity: 1,
                    scale: 1
                },
                next: {
                    translate: [0, 0, 0],
                    rotate: [0, 0, 0],
                    opacity: 1,
                    scale: 1
                }
            }
        });

        const getTranslateValue = value => {
            if (typeof value === 'string') return value;
            return `${value}px`;
        };

        const setTranslate = () => {
            const {
                slides,
                $wrapperEl,
                slidesSizesGrid
            } = swiper;
            const params = swiper.params.creativeEffect;
            const {
                progressMultiplier: multiplier
            } = params;
            const isCenteredSlides = swiper.params.centeredSlides;

            if (isCenteredSlides) {
                const margin = slidesSizesGrid[0] / 2 - swiper.params.slidesOffsetBefore || 0;
                $wrapperEl.transform(`translateX(calc(50% - ${margin}px))`);
            }

            for (let i = 0; i < slides.length; i += 1) {
                const $slideEl = slides.eq(i);
                const slideProgress = $slideEl[0].progress;
                const progress = Math.min(Math.max($slideEl[0].progress, -params.limitProgress), params.limitProgress);
                let originalProgress = progress;

                if (!isCenteredSlides) {
                    originalProgress = Math.min(Math.max($slideEl[0].originalProgress, -params.limitProgress), params.limitProgress);
                }

                const offset = $slideEl[0].swiperSlideOffset;
                const t = [swiper.params.cssMode ? -offset - swiper.translate : -offset, 0, 0];
                const r = [0, 0, 0];
                let custom = false;

                if (!swiper.isHorizontal()) {
                    t[1] = t[0];
                    t[0] = 0;
                }

                let data = {
                    translate: [0, 0, 0],
                    rotate: [0, 0, 0],
                    scale: 1,
                    opacity: 1
                };

                if (progress < 0) {
                    data = params.next;
                    custom = true;
                } else if (progress > 0) {
                    data = params.prev;
                    custom = true;
                } // set translate


                t.forEach((value, index) => {
                    t[index] = `calc(${value}px + (${getTranslateValue(data.translate[index])} * ${Math.abs(progress * multiplier)}))`;
                }); // set rotates

                r.forEach((value, index) => {
                    r[index] = data.rotate[index] * Math.abs(progress * multiplier);
                });
                $slideEl[0].style.zIndex = -Math.abs(Math.round(slideProgress)) + slides.length;
                const translateString = t.join(', ');
                const rotateString = `rotateX(${r[0]}deg) rotateY(${r[1]}deg) rotateZ(${r[2]}deg)`;
                const scaleString = originalProgress < 0 ? `scale(${1 + (1 - data.scale) * originalProgress * multiplier})` : `scale(${1 - (1 - data.scale) * originalProgress * multiplier})`;
                const opacityString = originalProgress < 0 ? 1 + (1 - data.opacity) * originalProgress * multiplier : 1 - (1 - data.opacity) * originalProgress * multiplier;
                const transform = `translate3d(${translateString}) ${rotateString} ${scaleString}`; // Set shadows

                if (custom && data.shadow || !custom) {
                    let $shadowEl = $slideEl.children('.swiper-slide-shadow');

                    if ($shadowEl.length === 0 && data.shadow) {
                        $shadowEl = createShadow(params, $slideEl);
                    }

                    if ($shadowEl.length) {
                        const shadowOpacity = params.shadowPerProgress ? progress * (1 / params.limitProgress) : progress;
                        $shadowEl[0].style.opacity = Math.min(Math.max(Math.abs(shadowOpacity), 0), 1);
                    }
                }

                const $targetEl = effectTarget(params, $slideEl);
                $targetEl.transform(transform).css({
                    opacity: opacityString
                });

                if (data.origin) {
                    $targetEl.css('transform-origin', data.origin);
                }
            }
        };

        const setTransition = duration => {
            const {
                transformEl
            } = swiper.params.creativeEffect;
            const $transitionElements = transformEl ? swiper.slides.find(transformEl) : swiper.slides;
            $transitionElements.transition(duration).find('.swiper-slide-shadow').transition(duration);
            effectVirtualTransitionEnd({
                swiper,
                duration,
                transformEl,
                allSlides: true
            });
        };

        effectInit({
            effect: 'creative',
            swiper,
            on,
            setTranslate,
            setTransition,
            perspective: () => swiper.params.creativeEffect.perspective,
            overwriteParams: () => ({
                watchSlidesProgress: true,
                virtualTranslate: !swiper.params.cssMode
            })
        });
    }

    function EffectCards(_ref) {
        let {
            swiper,
            extendParams,
            on
        } = _ref;
        extendParams({
            cardsEffect: {
                slideShadows: true,
                transformEl: null,
                rotate: true
            }
        });

        const setTranslate = () => {
            const {
                slides,
                activeIndex
            } = swiper;
            const params = swiper.params.cardsEffect;
            const {
                startTranslate,
                isTouched
            } = swiper.touchEventsData;
            const currentTranslate = swiper.translate;

            for (let i = 0; i < slides.length; i += 1) {
                const $slideEl = slides.eq(i);
                const slideProgress = $slideEl[0].progress;
                const progress = Math.min(Math.max(slideProgress, -4), 4);
                let offset = $slideEl[0].swiperSlideOffset;

                if (swiper.params.centeredSlides && !swiper.params.cssMode) {
                    swiper.$wrapperEl.transform(`translateX(${swiper.minTranslate()}px)`);
                }

                if (swiper.params.centeredSlides && swiper.params.cssMode) {
                    offset -= slides[0].swiperSlideOffset;
                }

                let tX = swiper.params.cssMode ? -offset - swiper.translate : -offset;
                let tY = 0;
                const tZ = -100 * Math.abs(progress);
                let scale = 1;
                let rotate = -2 * progress;
                let tXAdd = 8 - Math.abs(progress) * 0.75;
                const slideIndex = swiper.virtual && swiper.params.virtual.enabled ? swiper.virtual.from + i : i;
                const isSwipeToNext = (slideIndex === activeIndex || slideIndex === activeIndex - 1) && progress > 0 && progress < 1 && (isTouched || swiper.params.cssMode) && currentTranslate < startTranslate;
                const isSwipeToPrev = (slideIndex === activeIndex || slideIndex === activeIndex + 1) && progress < 0 && progress > -1 && (isTouched || swiper.params.cssMode) && currentTranslate > startTranslate;

                if (isSwipeToNext || isSwipeToPrev) {
                    const subProgress = (1 - Math.abs((Math.abs(progress) - 0.5) / 0.5)) ** 0.5;
                    rotate += -28 * progress * subProgress;
                    scale += -0.5 * subProgress;
                    tXAdd += 96 * subProgress;
                    tY = `${-25 * subProgress * Math.abs(progress)}%`;
                }

                if (progress < 0) {
                    // next
                    tX = `calc(${tX}px + (${tXAdd * Math.abs(progress)}%))`;
                } else if (progress > 0) {
                    // prev
                    tX = `calc(${tX}px + (-${tXAdd * Math.abs(progress)}%))`;
                } else {
                    tX = `${tX}px`;
                }

                if (!swiper.isHorizontal()) {
                    const prevY = tY;
                    tY = tX;
                    tX = prevY;
                }

                const scaleString = progress < 0 ? `${1 + (1 - scale) * progress}` : `${1 - (1 - scale) * progress}`;
                const transform = `
          translate3d(${tX}, ${tY}, ${tZ}px)
          rotateZ(${params.rotate ? rotate : 0}deg)
          scale(${scaleString})
        `;

                if (params.slideShadows) {
                    // Set shadows
                    let $shadowEl = $slideEl.find('.swiper-slide-shadow');

                    if ($shadowEl.length === 0) {
                        $shadowEl = createShadow(params, $slideEl);
                    }

                    if ($shadowEl.length) $shadowEl[0].style.opacity = Math.min(Math.max((Math.abs(progress) - 0.5) / 0.5, 0), 1);
                }

                $slideEl[0].style.zIndex = -Math.abs(Math.round(slideProgress)) + slides.length;
                const $targetEl = effectTarget(params, $slideEl);
                $targetEl.transform(transform);
            }
        };

        const setTransition = duration => {
            const {
                transformEl
            } = swiper.params.cardsEffect;
            const $transitionElements = transformEl ? swiper.slides.find(transformEl) : swiper.slides;
            $transitionElements.transition(duration).find('.swiper-slide-shadow').transition(duration);
            effectVirtualTransitionEnd({
                swiper,
                duration,
                transformEl
            });
        };

        effectInit({
            effect: 'cards',
            swiper,
            on,
            setTranslate,
            setTransition,
            perspective: () => true,
            overwriteParams: () => ({
                watchSlidesProgress: true,
                virtualTranslate: !swiper.params.cssMode
            })
        });
    }

    // Swiper Class
    const modules = [Virtual, Keyboard, Mousewheel, Navigation, Pagination, Scrollbar, Parallax, Zoom, Lazy, Controller, A11y, History, HashNavigation, Autoplay, Thumb, freeMode, Grid, Manipulation, EffectFade, EffectCube, EffectFlip, EffectCoverflow, EffectCreative, EffectCards];
    Swiper.use(modules);

    return Swiper;

}));
  //# sourceMappingURL=swiper-bundle.js.map

/*!
* 插件：滚动反弹；使用方法：new axScroll({参数})
*/
class axScroll {
    constructor(elem, options) {
        if (!elem) { return false; }
        // 始终将elem转成DOM对象
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom || this.targetDom.children.length === 0) { return false; }
        this.options = axExtend({
            wrapper: '',//内滚动容器，可填#id、.className、nodeName或节点
            axis: 'y',//滚动方向，默认横轴滚动（水平滚动），可选择'y'
            snap: {
                enable: false,//是否吸附子元素，默认为false即不吸附，可设定true自动吸附，吸附会触发吸附监听
                force: false,//即使是当前节点已经是吸附元素了是否还强制再次吸附（目的是触发吸附监听），默认false，可选择true
                offset: 0,//偏差值，默认为0即表示不使用该参数，如果需要元素必须在起始线附近才吸附则可设置该参数，请填写一个正整数
                snapped: '',//可吸附的子元素，通常为.className或nodeName等节点选择器，也可以是一个节点或节点数组，这些元素将可吸附
                unsnapped: '',//可跳过的子元素，通常为.className或nodeName等节点选择器，也可以是一个节点或节点数组，这些元素将不可吸附；如果一个元素即是snapped又是unsnapped，那么会被定义为unsnapped
            },
            unbound: {
                attr: 'unbound',//未绑定事件的元素被设置的属性，默认unbound
                selector: '',//未绑定事件元素，通常为.className或nodeName等节点选择器，也可以是一个节点或节点数组，未绑定事件元素不可拖拽滚动，可复制文本
            },
            dampingRatio: 0.5,//拖拽出空白区域时的阻尼系数，直观表现为拖拽比较迟滞
            transInit: 0,//初始滚动距离，默认为0，支持整数、节点和#id等字符串
            toInitMode: 'scroll',//初始化时滚动模式，默认scroll即平滑滚动到初始位置，可选择locate直接跳到初始位置（无动画）
            duration: 300,//执行动画基础时间，单位ms
            distMouse: 0,//鼠标滚动距离，默认为外容器1/4，,如果是一个<=1的值，那么将按外容器尺寸的百分比滚动
            //默认为0即外容器1/4
            //可填大于1的整数值（PC中滚动一次是60px）
            //可填大于0小于等于1的小树值，那么将按外容器尺寸的百分比滚动
            //填'snap'，则滚动到下一个可吸附节点
            //填'neighbor'，则滚动到下一个兄弟节点
            completedTHR: 50,//拖出首尾空白多长表示完成，单位px
            drift: {
                enable: true,//是否开启漂移模式，默认true开启，可选择false不开启
                mode: 'free',//漂移模式，
                //默认free模式即产生漂移后是否自动计算距离，
                //填neighbor模式即滚动鼠标（手指）按下的节点的下一个兄弟节点，
                //填snap模式即滚动到下一个可吸附节点，
                //填一个小于等于1的小数，滚动距离为外容器尺寸*该参数
                //填一个大于1的具体数值，滚动距离即为该参数
                timeTHR: 300,//产生甩动漂移的时间阈值，小于该值会有漂移效果（与drift.distTHR配合），单位ms
                distTHR: 20,//产生甩动漂移的抖动阈值，小于该值则不漂移（与drift.timeTHR配合），单位px，避免垂直（相对）方向滑动也漂移
                coef: 2,//free模式下的漂移系数，值越大滚动距离越远，默认为2，请填一个正整数
            },
            paddingStart: 0,//头部反弹边界距离，是正整数，默认为0即与头部贴合
            paddingEnd: 0,//尾部反弹边界距离，是正整数，默认为0即与尾部贴合
            mouseINR: 300,//鼠标滚轮节流间隔延时，避免频繁触发操作导致发生某些未知错误，默认为300，可设置一个正整数，单位ms
            startINR: 300,//开始触发节流间隔延时，避免频繁触发操作导致发生某些未知错误，默认为300，可设置一个正整数，单位ms
            barShow: false,//是否显示虚拟滚动条，默认false不显示，可选择true
            insertable: true,//动画运行中是否可以再插入执行拖拽和滚轮事件，默认true允许，可选择false禁止，false的表现为动画完成后才可进行下一次拖拽或滚轮事件
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onBeforeToStart: '',//头部拖出空白达到completedTHR准备回弹时回调，同步代码返回true将自动回弹，否则需要手动回弹，支持参数percent，两位小数值
            onBeforeToEnd: '',//尾部拖出空白达到completedTHR准备回弹时回调，同步代码返回true将自动回弹，否则需要手动回弹，支持参数percent，两位小数值
            onInit: '',//回调函数，加载完毕执行，无参数
            onAdd: '',//回调函数，添加新节点后执行，支持一个参数即新添加的子节点数组
            onRemove: '',//回调函数，删除节点后执行，支持一个参数即刚删除的子节点
            onCompleted: '',//回调函数，拖出空白长度大于等于completedTHR时执行，支持一个参数placement，即拖拽位置，为start或end
            onExceed: '',//回调函数，拖到头或尾，直至出现空白区域时执行，支持两个参数：percent和placement，percent是一个小数即：空白长度/completedTHR，支持参数percent，两位小数值；placement为start或end
            onStart: '',//回调函数，手指/鼠标按下执行，支持一个参数transValue即内容器上一次偏移值
            onMove: '',//回调函数，手指/鼠标拖动过程执行，支持一个参数transValue即内容器当前偏移值
            onEnd: '',//回调函数，松开手指/鼠标后执行，支持一个参数transValue即内容器当前偏移值
            onToStart: '',//回调函数，回弹头部后执行，无参数
            onToEnd: '',//回调函数，回弹尾部后执行，无参数
            onTranslated: '',//回调函数，内容器发生translate偏移时执行，支持一个参数transValue即内容器当前偏移值
            onRefactored: '',//回调函数，内容器发生节点变化（删除或添加子元素）时执行，无参数
            onSnap: '',//回调函数，开始吸附执行，支持一个参数即当前被吸附子元素
            onSnapped: '',//回调函数，完成吸附后执行，支持一个参数即当前被吸附子元素
            onUpdateContent: '',//回调函数，更新内容后执行，无参数
            onResized: '',//回调函数，外容器尺寸变化后执行，支持两个参数即rect值和容器节点（target或wrapper）
            onMouse: '',//回调函数，滚轮操作后执行，支持一个参数即目标translate值
            onScrolled: '',//回调函数，完成一次滚动后执行，支持两个参数即原始值（节点）和目标translate值
            onLocated: '',//回调函数，完成一次跳转后执行，支持两个参数即原始值（节点）和目标translate值
            onClick: '',//回调函数，快速触摸或点击后执行，支持一个参数即点击目标对象节点
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //根据滚动方向定义style属性映射
        //index为从style取得transform文本的位置，[1,0,0,1,0,45],[4]为translateX；[5]为translateY
        if (this.options.axis === 'x') {
            this.propsMap = { position: 'left', axis: 'X', overflow: 'overflowX', client: 'clientX', size: 'width', index: 4, offset: 'offsetLeft' }
        } else if (this.options.axis === 'y') {
            this.propsMap = { position: 'top', axis: 'Y', overflow: 'overflowY', client: 'clientY', size: 'height', index: 5, offset: 'offsetTop' }
        }
        //滚轮对象
        this.scrollObj = axGetScrollObj();
        //定义容器变量，要求内外容器都存在，而且至少有一个子元素
        this.wrapperDom = this.options.wrapper ? axIdToDom(this.options.wrapper) : this.targetDom.firstElementChild;
        if (!this.wrapperDom || this.wrapperDom.children.length === 0) { return false; }
        //存储未绑定事件元素

        this.getUnbounds();
        //存储可吸附元素和不可吸附元素
        this.getSnappeds();
        //滚动条
        if (this.options.barShow) {
            this.scrollThumb = axAddElem('span');
            this.scrollTrack = axAddElem('div', { class: 'ax-scroll-bar' }, this.scrollThumb);
            this.targetDom.appendChild(this.scrollTrack);
        }
        //终端事件数组
        this.isMobi = axIsMobi();
        //触发状态
        this.started = false;
        //整个容器运动状态
        this.scrolled = true
        //抖动距离
        this.distJitter = 0;
        //超出时间函数
        this.exceedListen = (percent, placement) => {
            //监听回调
            this.handlers.hasOwnProperty('exceed') ? this.emit('exceed', percent, placement) : null;
            this.options.onExceed && this.options.onExceed.call(this, percent, placement);
            //超出阈值表示完成
            if (percent >= 1) {
                this.handlers.hasOwnProperty('completed') ? this.emit('completed', placement) : null;
                this.options.onCompleted && this.options.onCompleted.call(this, placement);
            }
        }
        
        //start函数，防抖
        this.startFun = axThrottle((e) => {
            //阻止事件向上（父层）传播
            e.stopPropagation();

            if (this.unbounds.length > 0 && (this.unbounds.includes(e.target) || this.unbounds.find(k => k.contains(e.target))) || !this.getInsertable()) {
                //在未绑定事件项里滑动则不滚动，按默认事件执行
                //至少有一个unbound，并且当前鼠标所指节点再其中，或者所指节点是某一个unbound的子节点（contains）
                return false;
            }
            //阻止默认事件，放在unbounds判断下方，避免元素不能复制（被禁用了事件）
            e.preventDefault();

            //不能从虚拟滚动条上开始
            if (this.scrollThumb && (e.target === this.scrollThumb || e.target === this.scrollTrack)) {
                return false;
            }
            //手指触碰获取坐标值（相对浏览器左上角，超出为负值），this.startValue存起来用来屏蔽拖拽中触发click事件
            let coordValue = this.startValue = this.getCoordValue(e);
            this.distStart = coordValue - this.transLast;
            //鼠标按下时的时间戳。pc端甩鼠标使用
            this.startTime = e.timeStamp;
            //更新状态
            this.started = true;
            //监听回调
            this.handlers.hasOwnProperty('start') ? this.emit('start', this.transLast) : null;
            this.options.onStart && this.options.onStart.call(this, this.transLast);
        }, this.options.startINR, true);
        //move函数
        this.moveFun = (e) => {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            //PC端并且不是在wrapper中点击则不再执行，pc端必须有this.started
            if (!this.isMobi && !this.started) {
                return false;
            }

            if (this.unbounds.length > 0 && (this.unbounds.includes(e.target) || this.unbounds.find(k => k.contains(e.target))) || !this.getInsertable()) {
                //在未绑定事件项里滑动则不滚动，按默认事件执行
                return false;
            }

            //阻止默认事件
            e.preventDefault();


            //手指触碰获取坐标值（相对浏览器左上角，超出为负值）
            let coordValue = this.getCoordValue(e);

            if (Math.abs(coordValue - this.startValue) <= 2) {
                //如果基本一致，即单击或双击则不再执行，(0~5都可以)
                return false;
            }

            //追加moveing属性
            this.targetDom.setAttribute('moving', '');

            this.direction = 0;
            this.transNow = coordValue - this.distStart;
            //拖动方向，true表示下移，false表示上移
            let goToEnd = this.transNow >= this.transLast ? true : false;
            if (goToEnd) {
                //下移
                //target小于wrapper时diff为负值，target大于wrapper时diff为正值（有空不能滚动）
                if (this.transNow > this.options.paddingStart) {
                    //向下拖拽出空白，抛物线
                    this.transNow = this.options.paddingStart + (this.transNow - this.options.paddingStart) * Math.pow(this.options.dampingRatio, 2);
                }
            } else {
                //上移
                //如果wrapper+上下边界值区域小于target则使用阻尼
                if (this.transNow <= this.difference && this.difference - this.options.paddingStart <= 0) {
                    //向上拖拽出空白，抛物线
                    this.transNow = this.difference + (this.transNow - this.difference) * Math.pow(this.options.dampingRatio, 2);
                }
            }

            //禁用transition
            this.wrapperDom.style.transitionDuration = '0ms';
            //偏移滚动
            this.wrapperDom.style.transform = `translate${this.options.axis}(${this.transNow}px)`;

            //监听回调
            this.handlers.hasOwnProperty('move') ? this.emit('move', this.transNow) : null;
            this.options.onMove && this.options.onMove.call(this, this.transNow);
            //仅监听
            this.exceedHandle();
        }
        //end函数
        this.endFun = (e) => {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            //取消moving属性
            this.targetDom.removeAttribute('moving');
            //PC端并且不是在wrapper中点击则不再执行，pc端必须有this.started
            if ((!this.isMobi && !this.started)) {
                return false;
            }

            //阻止默认事件
            if (this.unbounds.length > 0 && (this.unbounds.includes(e.target) || this.unbounds.find(k => k.contains(e.target))) || !this.getInsertable()) {
                //在未绑定事件项里滑动则不滚动，按默认事件执行
                return false;
            }
            e.preventDefault();
            //更新状态
            this.started = false;
            // this.isMoving = false;

            //拖动方向，1表示下移，-1表示上移
            this.flagDir = this.transNow >= this.transLast ? 1 : -1;
            //抖动值
            this.distJitter = Math.abs(this.transNow - this.transLast);
            //获得历史距离
            this.transLast = this.transNow;

            //如果偏移值很小，应当认定为单击或双击事件，不再执行滚动
            if (this.distJitter <= 2) {
                this.handlers.hasOwnProperty('click') ? this.emit('click', e.target) : null;
                this.options.onClick && this.options.onClick.call(this, e.target);
                return false;
            }

            //计算漂移
            let timeOffset = e.timeStamp - this.startTime;
            if (this.options.drift.enable && (timeOffset < this.options.drift.timeTHR && this.distJitter > this.options.drift.distTHR)) {
                let dist;
                if (this.options.drift.mode === 'neighbor') {
                    //neighbor模式即漂移（滚动）到自己前后元素
                    let target = null;
                    if (this.flagDir === -1) {
                        //向后找到兄弟元素
                        target = e.target.nextElementSibling;
                    } else {
                        //向前找到兄弟元素
                        target = e.target.previousElementSibling;
                    }

                    if (this.options.snap.enable) {
                        //如果是可吸附的则使用吸附方法滚动并吸附
                        this.setSnapped({ target });
                    } else {
                        //如果没有吸附功能，仅仅滚动
                        this.scrollTo(target);
                    }
                } else if (this.options.drift.mode === 'snap') {
                    //snap模式即漂移（滚动）到下一个可吸附节点
                    if (!this.options.snap.enable) {
                        return false;
                    }
                    let index = this.snappeds.findIndex(k => k === this.snapped),
                        target = null;
                    if (this.flagDir === -1) {
                        //向后找到吸附元素
                        index !== this.snappeds.length - 1 ? target = this.snappeds[index + 1] : null;
                    } else {
                        //向前找到吸附元素
                        index !== 0 ? target = this.snappeds[index - 1] : null;
                    }
                    this.setSnapped({ target });
                } else {
                    if (!isNaN(this.options.drift.mode)) {
                        if (this.options.drift.mode <= 1) {
                            //外容器的比例值
                            dist = this.options.drift.mode * this.targetDomSize * this.flagDir;
                        } else {
                            //漂移距离为具体值，则直接等于漂移距离
                            dist = this.options.drift.mode * this.flagDir;
                        }
                    } else {
                        //自由模式则自动计算，快甩产生漂移，追加滚动：时间差的两倍（根据drift.coef参数）距离
                        dist = (this.options.drift.timeTHR - timeOffset) * this.options.drift.coef * this.flagDir;
                    }
                    //scrollTo会产生动画，动画结束后会自动吸附，this.distJitter * this.flagDir为onstart到onend滑过的距离需要减掉
                    this.scrollTo(this.transLast + dist - (this.distJitter * this.flagDir));
                }
            } else {
                //吸附
                this.options.snap.enable ? this.setSnapped() : null;
            }

            //监听回调
            this.handlers.hasOwnProperty('end') ? this.emit('end', this.transNow) : null;
            this.options.onEnd && this.options.onEnd.call(this, this.transNow);
            //限制滚动距离
            this.exceedHandle((percent) => {
                //向下拖拽，回弹头部
                if (this.options.onBeforeToStart) {
                    //返回true则执行toStart
                    this.options.onBeforeToStart.call(this, percent) ? this.toStart() : null;
                } else {
                    this.toStart();
                }
            }, (percent) => {
                //向上拖拽，回弹尾部
                if (this.options.onBeforeToEnd) {
                    //返回true则执行toEnd
                    this.options.onBeforeToEnd.call(this, percent) ? this.toEnd() : null;
                } else {
                    this.toEnd();
                }
            });

        }
        //scroll函数
        this.scrollFun = axThrottle((e) => {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            //阻止默认事件
            e.preventDefault();
            if (this.unbounds.length > 0 && (this.unbounds.includes(e.target) || this.unbounds.find(k => k.contains(e.target))) || !this.getInsertable()) {
                //在未绑定事件项里滑动则不滚动，按默认事件执行；
                return false;
            }

            let isUp = axIsScrollUp(e, this.scrollObj.detail),
                flag = isUp ? 1 : -1,
                target = null;
            if (typeof this.options.distMouse !== 'string') {
                //火狐上卷-1，下卷+1，其他相反
                let amount = Math.abs(this.distMouse) * flag,
                    nowTranslate = this.getTranslateValue(),
                    targetTranslate = nowTranslate - amount;
                //经过以上运算，amount已经各浏览器表现一致了，大于0则上卷，小于0则下卷
                target = targetTranslate;
            } else {
                if (this.options.distMouse === 'neighbor') {
                    //滚动到下一个邻居
                    if (isUp) {
                        //向后找到兄弟元素
                        target = e.target.nextElementSibling;
                    } else {
                        //向前找到兄弟元素
                        target = e.target.previousElementSibling;
                    }
                } else if (this.options.distMouse === 'snap') {
                    //滚动下一个可吸附元素
                    if (this.options.snap.enable) {
                        let index = this.snappeds.findIndex(k => k === this.snapped)
                        if (isUp) {
                            //向后找到吸附元素
                            index !== this.snappeds.length - 1 ? target = this.snappeds[index + 1] : null;
                        } else {
                            //向前找到吸附元素
                            index !== 0 ? target = this.snappeds[index - 1] : null;
                        }
                    }
                }
            }

            this.scrollTo(target, this.options.duration);
            //监听回调
            this.handlers.hasOwnProperty('mouse') ? this.emit('mouse', target) : null;
            this.options.onMouse && this.options.onMouse.call(this, target);
        }, this.options.mouseINR, true);
        //动画完成函数
        this.transitionendFun = (e) => {
            //阻止事件向上（父层）传播
            e.stopPropagation();
            //transitionend生效的前提是元素的transition-duration和transition-property存在
            if (this.transNow === 0) {
                //到头
                this.handlers.hasOwnProperty('toStart') ? this.emit('toStart', '') : null;
                this.options.onToStart && this.options.onToStart.call(this);
            }
            if (this.transNow === this.difference) {
                this.handlers.hasOwnProperty('toEnd') ? this.emit('toEnd', '') : null;
                this.options.onToEnd && this.options.onToEnd.call(this);
            }
            //偏移完成时不显示滚动条
            this.scrollTrack ? this.scrollTrack.removeAttribute('shown') : null;
            //如果最接近的吸附元素不是当前的吸附元素，则自动吸附
            this.options.snap.enable && this.snapped !== this.getLatestSnap().latest ? this.setSnapped() : null;
        }
        //初始化
        this.init();
    }
    init(callback) {
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;
        //设置风格
        this.setAttribute();
        //更新了this.difference
        this.updateDiff();
        //手指按下的X/Y轴值
        this.distStart = 0;
        //上一次移动距离
        this.transLast = 0;
        //初始距离，transInit可为整数或节点或字符串
        let initType = axType(this.options.transInit);
        if (initType.includes('HTML') || initType === 'String') {
            //是个节点或#id
            let childInit = initType === 'String' ? this.wrapperDom.querySelector(this.options.transInit) : this.options.transInit;
            this.transInit = -childInit[this.propsMap.offset] + this.options.paddingStart;
        } else {
            //是数字，初始值只能是在起点线以上
            this.transInit = this.options.paddingStart > Math.abs(this.options.transInit) ? this.options.paddingStart : this.options.transInit;
        }

        //整个元素当前的拖拽滚动距离
        this.transNow = 0;

        //移动到初始值，scroll为平滑移动，locate为直接跳到（无动画）
        if (this.options.toInitMode === 'scroll') {
            this.scrollTo(this.transInit);
        } else if (this.options.toInitMode === 'locate') {
            this.locateTo(this.transInit);
        }

        //手指放下
        this.targetDom.addEventListener(axEventMap[0], this.startFun);
        //手指拖拽
        this.targetDom.addEventListener(axEventMap[1], this.moveFun);
        //手指松开
        this.targetDom.addEventListener(axEventMap[2], this.endFun);

        //监听translate值变化
        this.StyleObs = new MutationObserver(mutations => {
            //style发生了变化
            mutations.forEach(k => {
                if (k.type === "attributes") {
                    let transNow = Number(k.target.style.transform.replace(/[^0-9.\-,]/g, ''));
                    //联动滚动条
                    if (this.options.barShow) {
                        //偏移时显示
                        this.scrollTrack.setAttribute('shown', '');
                        //拖出空白缩小滑块尺寸
                        if (transNow >= 0 || this.isBaby()) {
                            this.updateScrollBarSize(transNow);
                        } else if (transNow <= this.difference) {
                            this.updateScrollBarSize(this.difference - transNow);
                        }
                        //小于0，即滑出头部才滚动
                        if (transNow <= 0) {
                            let scrollRatio = 1 - this.scrollThumbSizeRatio,
                                translateRatio = Math.abs(transNow / (this.targetDomSize - this.wrapperDomSize)),
                                offsetStartRatio = translateRatio * scrollRatio,
                                offsetStartPercent = Math.ceil(offsetStartRatio * 100) + '%';
                            this.scrollThumb.style[this.propsMap.position] = offsetStartPercent;
                        }


                    }
                    //监听回调
                    this.handlers.hasOwnProperty('translated') ? this.emit('translated', transNow) : null;
                    this.options.onTranslated && this.options.onTranslated.call(this, transNow);
                }
                if (k.type === "childList") {
                    //节点变化后修正关系
                    this.response();
                    //更新未绑定事件元素
                    this.options.unbound ? this.getUnbounds() : null;
                    //更新可吸附
                    this.options.snap.enable ? this.getSnappeds() : null;
                    //监听回调
                    this.handlers.hasOwnProperty('refactored') ? this.emit('refactored', '') : null;
                    this.options.onRefactored && this.options.onRefactored.call(this);
                }
            });
        });
        this.StyleObs.observe(this.wrapperDom, {
            childList: true,
            attributes: true,
            attributeFilter: ['style']
        });

        //监听target尺寸变化，
        this.targetDomSizeObserver = new ResizeObserver(axDebounce((entries) => {
            //初次加载都会执行一次，display:none会执行一次，再次show会执行一次

            //target尺寸变化后修正关系
            for (let k of entries) {
                //隐藏元素（display:none）不触发
                if (!k.contentRect.width && !k.contentRect.height) {
                    return false;
                }
                this.response();
                //监听回调
                this.handlers.hasOwnProperty('resized') ? this.emit('resized', k, this.targetDom) : null;
                this.options.onResized && this.options.onResized.call(this, k, this.targetDom);

            }
        }));
        this.targetDomSizeObserver.observe(this.targetDom);

        //监听wrapper尺寸变化
        this.wrapperDomSizeObserver = new ResizeObserver(axDebounce((entries) => {
            //wrapper尺寸变化后修正关系
            for (let k of entries) {
                //隐藏元素（display:none）不触发
                if (!k.contentRect.width && !k.contentRect.height) {
                    return false;
                }
                this.response();
                //监听回调
                this.handlers.hasOwnProperty('resized') ? this.emit('resized', k, this.wrapperDom) : null;
                this.options.onResized && this.options.onResized.call(this, k, this.wrapperDom);
            }
        }));
        this.wrapperDomSizeObserver.observe(this.wrapperDom);

        //PC端的mouseleave事件
        if (!this.isMobi) {
            //离开target等同于mouseup
            this.targetDom.addEventListener('mouseleave', this.endFun);
            //鼠标滚动
            this.targetDom.addEventListener(this.scrollObj.event, this.scrollFun);
        }
        //监听到头和到尾
        this.wrapperDom.addEventListener("transitionend", this.transitionendFun);

        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        callback && callback.call(this);
        return this;
    }
    //获得可插入事件状态
    getInsertable () {
        if (this.options.insertable) {
            return true;
        } else {
            return this.scrolled ? true : false;
        }
    }
    setAttribute() {
        //强制添加ax-scroll类
        !this.targetDom.classList.contains('ax-scroll') ? this.targetDom.classList.add('ax-scroll') : null;
        !this.wrapperDom.classList.contains('ax-scroll-body') ? this.wrapperDom.classList.add('ax-scroll-body') : null;
        //外容器强制要求overflow=hidden
        axStyle(this.targetDom)[this.propsMap.overflow] !== 'hidden' && axStyle(this.targetDom).overflow !== 'hidden' ? this.targetDom.style[this.propsMap.overflow] = 'hidden' : null;
        this.targetDom.setAttribute('axis', this.options.axis);
    }
    getUnbounds() {
        this.unbounds = this.getDomsFromArg(this.options.unbound.selector);
        this.unbounds.forEach(k => k.setAttribute(this.options.unbound.attr, ''));
    }
    getDomsFromArg(arg) {
        let arr = [];
        if (arg) {
            let unboundType = axType(arg);
            if (unboundType === 'String') {
                arr = [...this.wrapperDom.querySelectorAll(arg)];
            } else if (unboundType === 'Array') {
                arr = arg;
            } else if (unboundType.includes('HTML')) {
                arr.push(arg);
            }
        }
        return arr.filter(Boolean);
    }
    getSnappeds() {
        if (!this.options.snap.enable) {
            this.snappeds = [];
        } else {
            //存储不可吸附元素
            this.unsnappeds = this.getDomsFromArg(this.options.snap.unsnapped);
            let snappeds = this.getDomsFromArg(this.options.snap.snapped);
            if (!snappeds.length) {
                snappeds = [...this.wrapperDom.children];
            }
            this.snappeds = snappeds.filter(k => !this.unsnappeds.includes(k));
        }
    }
    getLatestSnap() {
        let snappeds = this.snappeds.map(k => {
            //获得每个子元素距离start线的绝对距离
            return { dom: k, offset: Math.abs(this.transNow - this.options.paddingStart + k[this.propsMap.offset]) }
        }), latest = null;

        //差值从小到大排序
        snappeds.sort((a, b) => a.offset - b.offset);
        if (this.options.snap.offset) {
            //如果有偏移值则看哪个最近
            let temp = snappeds.find(k => k.offset <= this.options.snap.offset);
            latest = temp ? temp.dom : null;
        } else {
            //如果没有偏移值则取第一项
            latest = snappeds[0].dom;
        }
        return { latest, snappeds };
    };
    setSnapped(options) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.wrapperDom.children.length === 0) {
            return this;
        }
        if (!this.options.snap.enable) {
            return false;
        }
        //更新运动状态
        this.scrolled = false;
        //options={force:true,target:''}
        let opt = Object.assign({ force: this.options.snap.force }, options);
        if (this.options.snap.enable) {
            //排除不可吸附的元素
            let current = this.getLatestSnap(),
                snappeds = current.snappeds,
                latest = current.latest,
                target = null,
                toDom = opt.target ? axIdToDom(opt.target) : null;

            if (toDom) {
                //如果有opt.target，则直接滚动到该元素并吸附
                target = toDom;
            } else {
                //如果没有传入，则自动获得被吸附的target
                target = latest;
            }
            //可能旧吸附元素，始终滚动到起点线位置，确保边缘拖动一点点也可以顺利回弹
            if (target) {
                this.scrollTo(target, opt.duration, () => {
                    //滚动完毕，而且完成吸附
                    if (target === this.snapped) {
                        this.handlers.hasOwnProperty('snapped') ? this.emit('snapped', target) : null;
                        this.options.onSnapped && this.options.onSnapped.call(this, target);
                        opt.callback && opt.callback.call(this, target);
                    }
                });
            }

            //是新的吸附元素才执行以下代码
            if (target && (target !== this.snapped || opt.force)) {

                //设置属性
                target.setAttribute('snapped', '');
                //其他的取消吸附属性
                let other = snappeds.find(i => i !== target && i.dom.hasAttribute('snapped'));
                other ? other.dom.removeAttribute('snapped') : null;
                //保存被吸附的元素
                this.snapped = target;
                //监听回调
                this.handlers.hasOwnProperty('snap') ? this.emit('snap', target) : null;
                this.options.onSnap && this.options.onSnap.call(this, target);
            }
        }
    }
    rollback(dist, duration) {
        duration = duration || this.options.duration + 'ms';
        //更新运动状态
        //this.isMoving = true;
        //实验之后，使用异步不会跳跃
        setTimeout(() => {
            this.wrapperDom.style.transitionDuration = duration;
            this.wrapperDom.style.transform = `translate${this.options.axis}(${dist}px)`;
        }, 0);
        setTimeout(() => {
            //更新运动状态
            //this.isMoving = false;
        }, parseInt(duration));
    }
    toStart(duration) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.wrapperDom.children.length === 0) {
            return this;
        }
        //回弹头部
        this.transNow = this.transLast = this.options.paddingStart;
        this.rollback(this.transNow, duration);
    }
    toEnd(duration) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.wrapperDom.children.length === 0) {
            return this;
        }
        //回弹尾部
        this.transNow = this.transLast = this.difference;
        this.rollback(this.transNow, duration);
    }
    getDistMouse() {
        //获取滚动值，distMouse为数字类型
        this.distMouse = 0;
        //是数字或为空
        if (typeof this.options.distMouse !== 'string' || !this.options.distMouse) {
            if (this.options.distMouse) {
                //小于等于1则按比例获取滚动距离
                this.distMouse = this.options.distMouse <= 1 ? parseInt(this.targetDomSize * this.options.distMouse) : this.options.distMouse;
            } else {
                //参数为0则按外容器的1/4获取滚动距离
                this.distMouse = parseInt(this.targetDomSize * 0.25);
            }
        }
    }
    getEdgeValue(data) {
        let datatType = axType(data), value;
        if (datatType.includes('HTML') || datatType === 'String') {
            //data是子元素
            let dom = axIdToDom(data);
            dom ? value = -dom[this.propsMap.offset] + this.options.paddingStart : this.options.paddingStart;
        } else {
            //数值
            value = data;
        }
        if (value || value === 0) {
            //如果value是正值，则不应该大于偏移值
            if (value >= this.options.paddingStart || this.isBaby()) {
                value = this.options.paddingStart;
            } else if (value <= this.difference) {
                //不能超过target的尺寸+尾部偏移值
                value = this.difference;
            }

        }
        return value;
    }
    scrollTo(value, dur, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.wrapperDom.children.length === 0) {
            return this;
        }
        //更新运动状态
        this.scrolled = false;
        //没有value则不再执行，另外如果运动不可干预且还在运动中也不再执行
        let dist = this.getEdgeValue(value),
            compare = Math.abs(this.transNow - dist) / this.targetDomSize,
            //要滚动的距离大于了外围尺寸则自然增加动画时间，动画时间只要整数
            duration = dur || dur == 0 ? `${dur}ms` : compare > 1 ? Math.trunc(this.options.duration * compare) + 'ms' : `${this.options.duration}ms`;
        this.wrapperDom.style.transitionDuration = duration;
        this.wrapperDom.style.transform = `translate${this.options.axis}(${dist}px)`;
        //更新当前滚动到的位置
        this.transLast = this.transNow = dist;
        //回调
        setTimeout(() => {
            //监听回调
            this.handlers.hasOwnProperty('scrolled') ? this.emit('scrolled', value, dist) : null;
            this.options.onScrolled && this.options.onScrolled.call(this, value, dist);
            callback && callback.call(this, value, dist);
            //更新运动状态
            this.scrolled = true;
        }, parseInt(duration));
    }
    locateTo(value, snap = true, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.wrapperDom.children.length === 0) {
            return this;
        }
        let dist = this.getEdgeValue(value);
        this.wrapperDom.style.transitionDuration = '0ms';
        this.wrapperDom.style.transform = `translate${this.options.axis}(${dist}px)`;
        //更新当前滚动到的位置
        this.transLast = this.transNow = dist;
        //自动吸附
        snap ? this.setSnapped() : null;
        //监听回调
        this.handlers.hasOwnProperty('located') ? this.emit('located', value, dist) : null;
        this.options.onLocated && this.options.onLocated.call(this, value, dist);
        callback && callback.call(this, value, dist);
    }
    getCoordValue(e) {
        return this.isMobi ? e.changedTouches[0][this.propsMap.client] : e[this.propsMap.client];
    }
    exceedHandle(start, end) {
        //可滚动时difference是负值，不可滚动时是正值（target大于wrapper）
        let percent = 0;
        if (this.transNow > this.options.paddingStart || this.wrapperDomSize <= (this.targetDomSize - this.options.paddingStart - this.options.paddingEnd)) {
            percent = Math.abs(this.transNow / this.options.completedTHR);
            start && start(percent.toFixed(2));
            //监听回调
            this.exceedListen(percent.toFixed(2), 'start');
        } else if (this.transNow < this.difference) {
            percent = Math.abs(this.transNow - this.difference) / this.options.completedTHR;
            end && end(percent.toFixed(2));
            //监听回调
            this.exceedListen(percent.toFixed(2), 'end');
        }
    }
    isBaby() {
        return this.wrapperDomSize <= (this.targetDomSize - this.options.paddingStart - this.options.paddingEnd);
    }
    addTranslate(value) {
        //头插入元素，sizeAdded为其高度，且为正值
        value ? this.locateTo(this.getTranslateValue() - value) : null;
    }
    updateDiff() {
        //更新内外容器的尺寸差值，target的border、padding不作为高度一部分
        let targetStyle = axStyle(this.targetDom),
            wrapperStyle = axStyle(this.wrapperDom),
            targetHeight = parseInt(targetStyle.height),
            wrapperHeight = parseInt(wrapperStyle.height),
            targetWidth = parseInt(targetStyle.width),
            wrapperWidth = parseInt(wrapperStyle.width);
        if (this.options.axis === 'x') {
            this.difference = targetWidth - wrapperWidth - this.options.paddingEnd;
            //获得变量
            this.targetDomSize = targetWidth;
            this.wrapperDomSize = wrapperWidth;
        } else if (this.options.axis === 'y') {
            this.difference = targetHeight - wrapperHeight - this.options.paddingEnd;
            //获得变量
            this.targetDomSize = targetHeight;
            this.wrapperDomSize = wrapperHeight;
        }
        //更新滚动条尺寸
        this.updateScrollBarSize();
        //根据target尺寸获得滚轮距离
        this.getDistMouse();
    }
    response() {
        //相应wrapper内容变化和target尺寸变化
        //节点变化后修正关系
        this.updateDiff();
        //回弹头或者尾
        if (this.isBaby()) {
            this.toStart();
        } else if (this.wrapperDom[this.propsMap.offset] > this.options.paddingStart) {
            this.toStart();
        } else if (this.transNow < this.difference) {
            this.toEnd();
        }
    }
    add(data, opt) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let dataType = axType(data),
            arr = [],
            target = null,
            options = Object.assign({ placement: 'end', fixed: true }, opt);
        if (dataType === 'DocumentFragment') {
            //碎片
            arr = [...data.children];
        } else if (dataType === 'Array') {
            //数组
            arr = data;
        } else {
            //节点
            arr.push(data);
        }
        if (options.placement === 'start') {
            let target = this.wrapperDom.firstElementChild,
                //先保存添加前的第一个节点数据
                offsetTarget = target ? target[this.propsMap.offset] : 0;
            //头部添加
            //反转顺序，确保数组第一项添加到最前面
            arr = [...arr].reverse();
            for (let k of arr) {
                this.wrapperDom.insertAdjacentElement('afterbegin', k);
            }
            //更新内外容器关系
            this.updateDiff();
            //修正当前translate偏移值，不跳动
            if (target) {
                if (options.fixed) {
                    let offsetTranslate = offsetTarget - target[this.propsMap.offset];
                    this.transNow += offsetTranslate;
                    this.locateTo(this.transNow);
                } else {
                    //回弹头部
                    this.toStart();
                }
            }

        } else if (options.placement === 'end') {
            target = this.wrapperDom.lastElementChild;
            //尾部添加
            for (let k of arr) {
                this.wrapperDom.insertAdjacentElement('beforeend', k);
            }
            //更新内外容器关系
            this.updateDiff();
            //回弹尾部
            !options.fixed ? this.toEnd() : null;
        } else {
            let placementType = axType(options.placement);
            if (placementType === 'Number') {
                //索引，在该索引子节点后面添加
                target = this.wrapperDom.children[options.placement];
            } else if (placementType === 'String') {
                //#id，在该索引子节点后面添加
                target = this.wrapperDom.querySelector(options.placement);
            } else if (placementType.includes('HTML')) {
                //节点，在该节点后面添加
                target = options.placement;
            }
            if (target) {
                arr = [...arr].reverse();
                for (let k of arr) {
                    target.insertAdjacentElement('afterend', k);
                }
                //更新内外容器关系
                this.updateDiff();
            }
        }
        //监听回调
        this.handlers.hasOwnProperty('add') ? this.emit('add', arr, target) : null;
        this.options.onAdd && this.options.onAdd.call(this, arr, target);
        options.callback && options.callback.call(this, arr, target);
    }
    remove(placement = 'end', callback) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.wrapperDom.children.length === 0) {
            return this;
        }
        let target = null;

        if (placement === 'start') {
            //头部删除
            target = this.wrapperDom.firstElementChild;
        } else if (placement === 'end') {
            //尾部删除
            target = this.wrapperDom.lastElementChild;
        } else {
            //中间删除
            let placementType = axType(placement);
            if (placementType === 'Number') {
                //索引，在该索引子节点后面添加
                target = this.wrapperDom.children[placement];
            } else if (placementType === 'String') {
                //#id，在该索引子节点后面添加
                target = this.wrapperDom.querySelector(placement);
            } else if (placementType.includes('HTML')) {
                //节点，在该节点后面添加
                target = placement;
            }
        }
        if (target) {
            target.remove();
            //更新内外容器关系
            this.updateDiff();
            //监听回调
            this.handlers.hasOwnProperty('remove') ? this.emit('remove', target) : null;
            this.options.onRemove && this.options.onRemove.call(this, target);
            callback && callback.call(this, target);
        }

    }
    updateScrollBarSize(value) {
        if (this.options.barShow) {
            this.scrollThumbSizeRatio = this.targetDomSize / this.wrapperDomSize >= 1 ? 1 : (this.targetDomSize / this.wrapperDomSize);
            //value为空白尺寸（正整数），exceedRadio为拖出空白后滚动条缩小比例，value/2是希望滚动条变化小一些
            let exceedRadio = ((this.targetDomSize - Math.abs(value / 2)) / this.targetDomSize);
            //拖出空白则变短
            value ? this.scrollThumbSizeRatio = this.scrollThumbSizeRatio * exceedRadio : null;
            this.scrollThumb.style[this.propsMap.size] = Math.ceil(this.scrollThumbSizeRatio * 100) + '%';
        }
    }
    getTranslateValue() {
        //从style的transform中获得偏移值
        let value = axStyle(this.wrapperDom).transform.replace(/[^0-9.\-,]/g, '').split(',').map(k => Number(k))[this.propsMap.index]
        return value || 0;
    }
    updateContent(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let dataType = axType(data);
        this.wrapperDom.innerHTML = '';
        if (dataType === 'String') {
            //是字符串，则作为html文本写入
            if (axStrType(data)) {
                //是#id节点选择器
                let dom = axIdToDom(data);
                dom ? this.wrapperDom.innerHTML = dom.innerHTML : null;
            } else {
                //普通节点文本
                this.wrapperDom.innerHTML = data;
            }
        } else if (dataType.includes('HTML')) {
            //是节点，整体append
            this.wrapperDom.appendChild(data);
        } else if (dataType === 'Array') {
            //是节点数组，逐个append
            let fragment = document.createDocumentFragment();
            for (let k of data) {
                fragment.appendChild(k);
            }
            this.wrapperDom.appendChild(fragment);
        } else if (dataType === 'DocumentFragment') {
            //是碎片，直接append
            this.wrapperDom.appendChild(data);
        }
        //更新内外容器关系
        this.updateDiff();
        //监听回调
        this.handlers.hasOwnProperty('updateContent') ? this.emit('updateContent', '') : null;
        this.options.onUpdateContent && this.options.onUpdateContent.call(this);
        callback && callback.call(this);
    }
    destroy(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //取消监听事件
        this.targetDom.removeEventListener(axEventMap[0], this.startFun);
        this.targetDom.removeEventListener(axEventMap[1], this.moveFun);
        this.targetDom.removeEventListener(axEventMap[2], this.endFun);
        this.wrapperDom.removeEventListener("transitionend", this.transitionendFun);
        //停止观察
        this.StyleObs.disconnect();
        this.targetDomSizeObserver.disconnect();
        this.wrapperDomSizeObserver.disconnect();
        //pc端特别要求
        if (!this.isMobi) {
            this.targetDom.removeEventListener('mouseleave', this.endFun);
            this.targetDom.removeEventListener(this.scrollObj.event, this.scrollFun);
        }
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
* 插件：兼容手势库；使用方法：new axGesture({参数})
*/
class axGesture {
    constructor(elem, options) {
        if (!elem) { return false; }
        // 始终将elem转成DOM对象
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }

        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            parent: '',//父节点，离开父节点将终止事件，可填#id、.className、NODENAME、节点，默认为空即
            jitterValue: 10,//抖动值，小于该值不触发滑动事件，只触发单击或双击事件
            origin: null,//元素变换原点，不填则默认为中间，如果需要改变元素原点可以填，例如：{x:50,y:50}（单位px）
            click: {
                timeTHR: 200,//触发单击或双击事件的阈值时间，默认200即从start到end的时间需要小于200才会触发click或dblclick事件，单位ms
                delay: 200,//执行双击的等待时间，默认200，单位ms，即一次单击后在200ms内再次单击会取消单击事件并触发双击事件
                dblclickable: true,//是否可双击，默认true可双击，如果希望单击无延时可禁用双击
                holdDelay: 1000,//长按延时，默认是1000，单位ms，即1秒，按下不放将触发hold事件，等同于pc端鼠标右键事件
                holdToMenu: true,//在pc端上触发右键事件是否等效长按事件，默认false，可选择true，true将禁用默认右键事件
            },
            wheel: {
                action: '',//是否允许鼠标滚动等效事件，可填scale、translate或rotate，默认为空即不改变原鼠标滚轮事件
                axis: 'y',//如果action填了translate则需要指明方向（滚轮只有上下），默认为y轴，可填x轴
                distance: 60,//滚轮触发一次的滚动距离，默认是60，单位px，与常规浏览器滚轮距离基本一致
                scale: 0.5,//滚轮一次缩放比例的累加值，默认0.5即+0.5或-0.5，可自定义该值，请填写一个正数
                degree: 10,//如果action=rotate时的滚轮操作旋转角度，默认10，单位deg
                duration: 500,//PC端滚轮操作是否强制加上动画效果，默认500即动画执行500ms，如果不需要PC端滚轮动画请设为0
                reverse: false,//是否让滚轮逆向操作，默认false即滚轮向下拨动表示scale放大/rotate顺时针，若设为true则滚轮向下拨动表示scale缩小/rotate逆时针
            },
            translate: {
                max: 10000,//位移最大极限，默认为10000px
                min: -10000,//位移最小极限，默认为-10000px
                positionInstead: false,//发生位移的属性默认是transform（改变translateX和translateY），如果需要改变top和left属性来实现位移设设为true
            },
            drift: {
                enable: true,//是否开启漂移功能，默认true开启，可设为false关闭
                timeTHR: 300,//拖拽完成后触发drift的最大时间，小于该时间能达成触发条件
                distTHR: 20,//拖拽完成后触发drift的最少移动量，大于该值能达成触发条件
                coef: 1,//漂移系数，系数越大手指或鼠标释放后获得额外距离就越远，默认为1，请填写一个正数
                duration: 500,//漂移动画持续时间，默认500，单位ms
            },
            scale: {
                enable: false,//是否开启缩放功能，默认false关闭，可设为true开启
                centered: true,//是否居中缩放，默认无论落点在哪里都居中缩放，可选择false可将落点作为中心点进行缩放
                max: 100,//缩放最大极限，默认为100倍
                min: 0,//缩放最小极限，默认为0倍即可缩小到无限小
            },
            rotate: {
                enable: false,//是否开启旋转功能，默认false关闭，可设为true开启
                max: 3600,//旋转最大极限，默认为3600即顺时针最大可旋转10圈
                min: -3600,//旋转最小极限，默认为-3600即逆时针最大可旋转10圈
            },
            initial: null,//设置初始值，如果在css中设置了transform则需要在此设置初始值，以保持一致
            /* initial:{
                translate: { x: 0, y: 0 },//位移初始值，单位px
                scale: { x: 1, y: 1 },//缩放初始值，无单位
                rotate: 0,//旋转初始值，单位deg
            }, */
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面

            onInit: '',//回调函数，加载完毕执行，无参数

            onStart: '',//回调函数，手指/鼠标按下执行，支持一个参数即当前数据
            onMove: '',//回调函数，手指/鼠标拖动过程执行，支持一个参数即当前数据
            onEnd: '',//回调函数，松开手指/鼠标后执行，支持一个参数即当前数据

            onClick: '',//回调函数，短时间内点击后执行，支持一个参数即当前数据
            onDblclick: '',//回调函数，双击后执行，支持一个参数即当前数据
            onHold: '',//回调函数，长按后执行，支持一个参数即当前数据

            onScale: '',//回调函数，开始缩放执行，支持一个参数即当前数据
            onScaling: '',//回调函数，缩放中执行，支持一个参数即当前数据
            onScaled: '',//回调函数，缩放结束后执行，支持一个参数即当前数据

            onRotate: '',//回调函数，开始旋转执行，支持一个参数即当前数据
            onRotating: '',//回调函数，旋转中执行，支持一个参数即当前数据
            onRotated: '',//回调函数，旋转结束后执行，支持一个参数即当前数据

            onTranslate: '',//回调函数，开始位移执行，支持一个参数即当前数据
            onTranslating: '',//回调函数，位移中执行，支持一个参数即当前数据
            onTranslated: '',//回调函数，位移结束后执行，支持一个参数即当前数据

            onDrift: '',//回调函数，位移结束后产生漂移执行，支持一个参数即当前数据
            onDrifting: '',//回调函数，漂移中执行，支持一个参数即当前数据
            onDrifted: '',//回调函数，漂移结束后执行，支持一个参数即当前数据

            onCancel: '',//回调函数，失去目标节点后执行，支持一个参数即当前数据

            onReset: '',//回调函数，重置style.transform后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        let _this = this;
        //回调监听
        this.handlers = {};

        //触发状态
        this.started = false;
        this.eventState = 'end';
        //滚轮对象
        this.scrollObj = axGetScrollObj();
        //终端事件数组
        this.isMobi = axIsMobi();
        //时间戳
        this.startTime = 0;
        this.diffTime = 0;

        //抖动值，移动端触摸抖动稍微加大
        this.jitterValue = _this.options.jitterValue + (this.isMobi ? 10 : 0);
        //点击次数
        this.clickCount = 0;
        //长按定时器
        this.holdHandle = null;

        //移动值
        this.moveValues = {
            h: 0,
            x: 0,
            y: 0,
            //c:coord
            c: { id: 0, x: 0, y: 0 },
            //d:direction
            d: { x: '', y: '' }
        }
        //差值，包括缩放、旋转和拖拽滑动（translate和drift）
        this.diffValues = { scale: 0, rotate: 0, translate: { x: 0, y: 0, h: 0 } };
        //初始坐标值
        this.startCoord = { x: 0, y: 0 };
        this.triangleValues = { start: { x: 0, y: 0, h: 0 }, last: { x: 0, y: 0, h: 0 }, now: { x: 0, y: 0, h: 0 } };
        //move次数和多指move次数
        this.moveCount = 0;
        this.touchesMoveCount = 0;
        //参数格式
        this.paramsFormat = {
            originEvent: null,
            scale: {
                direction: 0,
                diff: 0,
                value: { x: 1, y: 1 },
                //动态中心点会产生transform.translate偏移值
                translate: { x: 0, y: 0 }
            },
            rotate: {
                direction: 0,
                diff: 0,
                value: 0
            },
            translate: {
                direction: { x: 0, y: 0 },
                diff: { x: 0, y: 0, h: 0 },
                value: { x: 0, y: 0 },
                isDrift: false,
            },
            drift: {
                duration: 0,
                //同translate.direction
                direction: { x: 0, y: 0 },
                //产生drift监听的额外移动距离
                diff: { x: 0, y: 0, h: 0 },
                //drift之后最终停留的tanslate值
                value: { x: 0, y: 0 }
            },
            wheel: {
                direction: { x: 0, y: 0 },
            },
            coord: { x: 0, y: 0 },
            //目标对象
            target: this.targetDom,
            //事件名称
            name: '',
            //是否是首次move
            moveStart: false,
            //多指触摸开始
            touchesStart: false,
            //是否是最后一次move
            moveEnd: false,
            //从movestart到moveend的持续时间
            moveTime: 0,
        }
        this.params = null;
        this.endParams = {
            moveStart: false,
            touchesStart: false,
            moveEnd: true,
        }
        //放下手指/鼠标开始
        this.startFun = function (e) {
            e.stopPropagation();
            e.cancelable ? e.preventDefault() : null;

            //更新参数
            _this.params = axClone(_this.paramsFormat);
            _this.params.originEvent = e;
            //禁止三个触点开始，开始changedTouches.length为0，三个触点通常会被手机或原生APP接管
            if (_this.isMobi && e.targetTouches.length > 2 && e.changedTouches.length === 0) {
                //更新状态，不让执行move和end
                _this.started = false;
                return false;
            }
            //获得transform初始值并设定this.nowValues[.scale]和this.lastValues[.scale]
            _this.getStartValues();


            let changedValues = _this.getTouchCoords(e, 'targetTouches'),
                targetValues = _this.getTouchCoords(e, 'targetTouches'),
                paramsHold = { ..._this.params },
                paramsTranslate = { ..._this.params },
                paramsScale = { ..._this.params },
                paramsRotate = { ..._this.params },
                paramsStart = { ..._this.params };
            //存储当前坐标
            _this.startCoord = _this.getCenterCoord(changedValues);
            //存储当前时间
            _this.startTime = Date.now();
            //单指长按
            if (!_this.isMobi || e.targetTouches.length === 1) {
                _this.holdHandle = setTimeout(() => {
                    //更新参数
                    paramsHold.coord = { ..._this.startCoord };
                    paramsHold.name = 'hold';

                    //长按事件
                    _this.handlers.hasOwnProperty('hold') ? _this.emit('hold', paramsHold) : null;
                    _this.options.onHold && _this.options.onHold.call(_this, paramsHold);
                    //取消定时器
                    clearTimeout(_this.holdHandle);
                }, _this.options.click.holdDelay);
            } else {
                //可能会执行多次touchstart，所以大于1个触点则取消长按
                _this.holdHandle ? clearTimeout(_this.holdHandle) : null;
            }
            //多指缩放监听（移动端才可）
            if (_this.isMobi && e.targetTouches.length > 1) {
                //从targetTouches对象获得前两个触点的中心点
                _this.startCoord = _this.getCenterCoord(targetValues);
                //缩放监听，更新参数
                paramsScale.coord = { ..._this.startCoord };
                paramsScale.scale.value = _this.lastValues.scale;
                paramsScale.name = 'scale';
                _this.handlers.hasOwnProperty('scale') ? _this.emit('scale', paramsScale) : null;
                _this.options.onScale && _this.options.onScale.call(_this, paramsScale);

                //旋转监听，更新参数
                paramsRotate.coord = { ..._this.startCoord };
                paramsRotate.rotate.value = _this.lastValues.rotate;
                paramsRotate.name = 'rotate';
                _this.handlers.hasOwnProperty('rotate') ? _this.emit('rotate', paramsRotate) : null;
                _this.options.onRotate && _this.options.onRotate.call(_this, paramsRotate);
            }
            //单指或多值拖拽开始
            paramsTranslate.coord = { ..._this.startCoord };
            paramsTranslate.translate.value = { ..._this.lastValues.translate };
            paramsTranslate.name = 'translate';
            _this.handlers.hasOwnProperty('translate') ? _this.emit('translate', paramsTranslate) : null;
            _this.options.onTranslate && _this.options.onTranslate.call(_this, paramsTranslate);
            //所有监听
            paramsStart.coord = { ..._this.startCoord };
            paramsStart.scale.value = _this.lastValues.scale;
            paramsStart.rotate.value = _this.lastValues.rotate;
            paramsStart.translate.value = { ..._this.lastValues.translate };
            paramsStart.name = 'start';
            _this.handlers.hasOwnProperty('start') ? _this.emit('start', paramsStart) : null;
            _this.options.onStart && _this.options.onStart.call(_this, paramsStart);
            //更新状态
            _this.started = true;
            _this.eventState = 'start';
        }
        //拖拽进行
        this.moveFun = function (e) {
            e.stopPropagation();
            e.cancelable ? e.preventDefault() : null;
            if (!_this.started) {
                return false;
            }
            //移动鼠标或手指则立即取消漂移动画监听，即取消drifting监听
            _this.driftListen ? (cancelAnimationFrame(_this.driftListen), _this.driftListen = null) : null;
            //累加移动次数
            _this.moveCount++;
            //更新参数
            _this.params.originEvent = e;
            //移动过程取消动画效果
            _this.targetDom.style.transitionDuration = `0ms`;

            let targetValues = _this.getTouchCoords(e, 'targetTouches'),
                moveStart = _this.moveCount === 1 ? true : false;
            //获得移动的值
            _this.moveValues = _this.getMoveValues(targetValues, _this.startCoord);
            let paramsTouches = { ..._this.params, coord: _this.moveValues.c, moveStart },
                paramsTranslate = { ...paramsTouches },
                paramsScale = { ...paramsTouches },
                paramsRotate = { ...paramsTouches },
                paramsMove = { ...paramsTouches };
            //需要保存拖拽差值，以便于在end中计算是translate还是drift
            _this.diffValues.translate = { x: ~~_this.moveValues.x, y: ~~_this.moveValues.y, h: ~~_this.moveValues.h };
            _this.updateNowTranslate(_this.diffValues.translate);
            //有位移则取消hold事件
            _this.moveValues.h > _this.jitterValue && _this.holdHandle ? clearTimeout(_this.holdHandle) : null;
            //拖拽监听
            paramsTranslate = axExtend(paramsTouches, {
                translate: {
                    diff: { ..._this.diffValues.translate },
                    value: { ..._this.nowValues.translate },
                    direction: _this.moveValues.d
                },
                name: 'translating'
            });
            _this.handlers.hasOwnProperty('translating') ? _this.emit('translating', paramsTranslate) : null;
            _this.options.onTranslating && _this.options.onTranslating.call(_this, paramsTranslate);
            //两个以及以上触点可触发缩放和旋转
            if (_this.isMobi && e.targetTouches.length > 1) {
                //累加多指移动次数
                _this.touchesMoveCount++;

                _this.triangleValues.now = _this.getTriangleValues(targetValues[0], targetValues[1]);
                //第一次多指移动则记录为triangle
                if (_this.touchesMoveCount === 1) {
                    _this.triangleValues.start = { ..._this.triangleValues.now };
                    _this.diffValues.scale = 0;
                    _this.diffValues.rotate = 0;
                    //更新参数
                    paramsTouches.touchesStart = true;
                } else {
                    //更新scale和rotate差值
                    _this.diffValues.scale = (_this.triangleValues.now.h / _this.triangleValues.last.h) - 1;
                    _this.diffValues.rotate = _this.triangleValues.now.a - _this.triangleValues.last.a;
                    //纠正,否则会产生跳跃值（_this.triangleValues.now.a和_this.triangleValues.last.a的值范围是-180~180，需要转成transform.rotate的常见deg值）
                    _this.correctDiffRotate();
                }

                //params.direction = _this.moveValues.d;
                //纠正差值，获得正确的diff.scale和this.nowValues.scale
                _this.correctRangeScale();
                //纠正差值，获得正确的diff.rotate和this.nowValues.rotate
                _this.correctRangeRotate();
                //缩放监听
                paramsScale = axExtend(paramsTouches, {
                    scale: {
                        diff: axNumber(_this.diffValues.scale),
                        value: { x: axNumber(_this.nowValues.scale.x), y: axNumber(_this.nowValues.scale.y) },
                        direction: _this.diffValues.scale > 0 ? 1 : _this.diffValues.scale < 0 ? -1 : 0
                    },
                    name: 'scaling'
                });
                //动态中心获得偏移值
                if (!_this.options.scale.centered) {
                    paramsScale.scale.translate = _this.getTranslate({
                        coord: paramsScale.coord,
                        diff: paramsScale.scale.diff,
                    });
                }
                _this.handlers.hasOwnProperty('scaling') ? _this.emit('scaling', paramsScale) : null;
                _this.options.onScaling && _this.options.onScaling.call(_this, paramsScale);
                //旋转监听
                paramsRotate = axExtend(paramsTouches, {
                    rotate: {
                        diff: axNumber(_this.diffValues.rotate),
                        value: axNumber(_this.nowValues.rotate),
                        direction: _this.diffValues.rotate > 0 ? 1 : _this.diffValues.rotate < 0 ? -1 : 0
                    },
                    name: 'rotating'
                });
                _this.handlers.hasOwnProperty('rotating') ? _this.emit('rotating', paramsRotate) : null;
                _this.options.onRotating && _this.options.onRotating.call(_this, paramsRotate);
            }
            //所有监听
            paramsMove = axExtend(paramsTouches, {
                rotate: paramsRotate.rotate,
                scale: paramsScale.scale,
                translate: paramsTranslate.translate,
                name: 'move'
            });
            _this.handlers.hasOwnProperty('move') ? _this.emit('move', paramsMove) : null;
            _this.options.onMove && _this.options.onMove.call(_this, paramsMove);
            //移动一次则保存last值
            _this.triangleValues.last = { ..._this.triangleValues.now };
            //更新状态
            _this.eventState = 'move';
        }
        //手指/鼠标抬起
        this.endFun = function (e) {
            e.stopPropagation();
            e.cancelable ? e.preventDefault() : null;
            if (!_this.started) {
                return false;
            }
            //清除累计移动次数（包含多指移动）
            _this.moveCount = _this.touchesMoveCount = 0;
            //更新参数
            _this.params.originEvent = e;
            //更新时间差值，单位ms
            _this.params.moveTime = _this.diffTime = Date.now() - _this.startTime;
            //清除长按定时器
            _this.holdHandle ? clearTimeout(_this.holdHandle) : null;

            let isDrift = _this.diffValues.translate.h > _this.options.drift.distTHR && _this.diffTime < _this.options.drift.timeTHR,
                paramsTranslate = {
                    ..._this.params,
                    coord: { ..._this.moveValues.c },
                    translate: {
                        diff: { ..._this.diffValues.translate },
                        value: { ..._this.nowValues.translate },
                        direction: { ..._this.moveValues.d },
                    },
                    ..._this.endParams,
                },
                paramsClick = { ..._this.params, ..._this.endParams },
                paramsScale = { ...paramsClick },
                paramsRotate = { ...paramsClick },
                paramsEnd = { ...paramsClick };
            if (_this.diffValues.translate.h < _this.jitterValue) {
                //小幅抖动只能触发单击和双击(确保pc端是左键)
                if (_this.diffTime < _this.options.click.timeTHR && (!_this.isMobi && e.button === 0 || _this.isMobi)) {
                    //只触发单击或双击事件
                    let coords = _this.getTouchCoords(e, 'changedTouches'),
                        endCoord = _this.getCenterCoord(coords);
                    //更新参数
                    paramsClick.coord = endCoord;
                    paramsClick.name = 'click';
                    if (_this.options.click.dblclickable) {
                        //允许双击则需要延时判断
                        _this.clickCount++;
                        if (_this.clickCount === 1) {
                            setTimeout(() => {
                                if (_this.clickCount === 1) {
                                    //单击事件
                                    _this.handlers.hasOwnProperty('click') ? _this.emit('click', paramsClick) : null;
                                    _this.options.onClick && _this.options.onClick.call(_this, paramsClick);
                                } else if (_this.clickCount > 1) {
                                    //双击事件
                                    paramsClick.name = 'dblclick';
                                    _this.handlers.hasOwnProperty('dblclick') ? _this.emit('dblclick', paramsClick) : null;
                                    _this.options.onDblclick && _this.options.onDblclick.call(_this, paramsClick);
                                }
                                //归零
                                _this.clickCount = 0;
                            }, _this.options.click.delay);
                        }
                    } else {
                        //单击事件
                        _this.handlers.hasOwnProperty('click') ? _this.emit('click', paramsClick) : null;
                        _this.options.onClick && _this.options.onClick.call(_this, paramsClick);
                    }
                }
            }
            //漂移事件
            if (_this.options.drift.enable && isDrift) {
                //快速滑动事件
                let distH = ~~((_this.options.drift.timeTHR - _this.diffTime) * _this.options.drift.coef),
                    distX = ~~(distH * (_this.diffValues.translate.x / _this.diffValues.translate.h)),
                    distY = ~~(distH * (_this.diffValues.translate.y / _this.diffValues.translate.h));
                //更新参数
                paramsTranslate.drift.direction = paramsTranslate.translate.direction;
                paramsTranslate.drift.diff = { x: distX, y: distY, h: distH };
                paramsTranslate.drift.duration = _this.options.drift.duration;
                paramsTranslate.drift.value = { x: paramsTranslate.translate.value.x + distX, y: paramsTranslate.translate.value.y + distY };
                paramsTranslate.translate.isDrift = true;
                paramsTranslate.name = 'drift';
                //使用动画
                _this.targetDom.style.transitionDuration = `${_this.options.drift.duration}ms`;
                _this.handlers.hasOwnProperty('drift') ? _this.emit('drift', paramsTranslate) : null;
                _this.options.onDrift && _this.options.onDrift.call(_this, paramsTranslate);
                //漂移完成后监听
                _this.driftListen = axListenAnimation({
                    duration: _this.options.drift.duration,
                    doing: () => {
                        //漂移中
                        _this.handlers.hasOwnProperty('drifting') ? _this.emit('drifting', paramsTranslate) : null;
                        _this.options.onDrifting && _this.options.onDrifting.call(_this, paramsTranslate);
                    },
                    done: () => {
                        //漂移完毕
                        //时间到而且偏移值一致才表示漂移完成，因为存在漂移途中再次拖拽操作导致偏移值不一致
                        let translateValues = axGetStyleTransform(_this.targetDom, ['translate'], _this.options.translate.positionInstead).translate;
                        if ((paramsTranslate.drift.value.x === translateValues.x) && (paramsTranslate.drift.value.y === translateValues.y)) {
                            _this.handlers.hasOwnProperty('drifted') ? _this.emit('drifted', paramsTranslate) : null;
                            _this.options.onDrifted && _this.options.onDrifted.call(_this, paramsTranslate);
                        }
                    },
                });

            }
            //拖拽完成事件
            paramsTranslate.name = 'translated';
            _this.handlers.hasOwnProperty('translated') ? _this.emit('translated', paramsTranslate) : null;
            _this.options.onTranslated && _this.options.onTranslated.call(_this, paramsTranslate);
            //移动端触发缩放和旋转
            if (_this.isMobi) {
                //缩放监听
                paramsScale.scale.diff = axNumber(_this.diffValues.scale);
                paramsScale.scale.value = { x: axNumber(_this.nowValues.scale.x), y: axNumber(_this.nowValues.scale.y) };
                paramsScale.name = 'scaled';
                _this.handlers.hasOwnProperty('scaled') ? _this.emit('scaled', paramsScale) : null;
                _this.options.onScaled && _this.options.onScaled.call(_this, paramsScale);
                //旋转监听
                paramsRotate.rotate.diff = axNumber(_this.diffValues.rotate);
                paramsRotate.rotate.value = axNumber(_this.nowValues.rotate);
                paramsRotate.name = 'rotated';
                _this.handlers.hasOwnProperty('rotated') ? _this.emit('rotated', paramsRotate) : null;
                _this.options.onRotated && _this.options.onRotated.call(_this, paramsRotate);
                //手指移开，当前旋转和缩放总值自动保存为下一次开始的初始值
                _this.lastValues.rotate = _this.nowValues.rotate;
                _this.lastValues.scale = { ..._this.nowValues.scale };
            }

            //所有监听
            paramsEnd.translate = paramsTranslate.translate;
            paramsEnd.drift = paramsTranslate.drift;
            paramsEnd.scale = paramsScale.scale;
            paramsEnd.rotate = paramsRotate.rotate;
            paramsEnd.name = 'end';
            _this.handlers.hasOwnProperty('end') ? _this.emit('end', paramsEnd) : null;
            _this.options.onEnd && _this.options.onEnd.call(_this, paramsEnd);
            //拖拽结束，将当前translate值赋予last，以便于下次计算
            _this.lastValues.translate = { ..._this.nowValues.translate };
            //设置完成（数据归零）
            _this.setCompleted();
        }

        //滚轮滚动
        this.wheelFun = function (e) {
            e.stopPropagation();
            e.preventDefault();
            //获得transform初始值并设定nowScale和lastScale
            _this.getStartValues();
            let isUp = _this.correctWheelDirection(axIsScrollUp(e, _this.scrollObj.detail)),
                dirValue = isUp ? 1 : -1;
            //使用动画
            _this.options.wheel.duration ? _this.targetDom.style.transitionDuration = `${_this.options.wheel.duration}ms` : null;
            //更新参数
            _this.params = axClone(_this.paramsFormat);
            _this.params = axExtend(_this.params, {
                originEvent: e,
                coord: { x: e.clientX, y: e.clientY },
                wheel: {
                    direction: { x: dirValue, y: dirValue }
                },
                translate: { value: { ..._this.nowValues.translate } },
                scale: {
                    value: { ..._this.nowValues.scale },
                    direction: dirValue
                },
                rotate: {
                    value: _this.nowValues.rotate,
                    direction: dirValue
                },
                moveTime: parseFloat(axStyle(_this.targetDom).transitionDuration) * 1000
            });

            //滚轮可能使用reverse
            let paramsAction = { ..._this.params, name: `${_this.options.wheel.action}move` },
                paramsMove = { ..._this.params, name: 'move' };

            if (_this.options.wheel.action === 'scale') {
                _this.diffValues.scale = (isUp ? _this.options.wheel.scale : -_this.options.wheel.scale);
                //纠正差值，获得正确的diff.scale和this.nowValues.scale（加上了差值）
                _this.correctRangeScale();
                //更新参数
                paramsAction.scale.diff = axNumber(_this.diffValues.scale);
                paramsAction.scale.value = { x: axNumber(_this.nowValues.scale.x), y: axNumber(_this.nowValues.scale.y) };
                //获得偏移值
                if (!_this.options.scale.centered) {
                    paramsAction.scale.translate = _this.getTranslate({
                        coord: paramsAction.coord,
                        diff: paramsAction.scale.diff,
                    });
                }
                _this.handlers.hasOwnProperty('scaling') ? _this.emit('scaling', paramsAction) : null;
                _this.options.onScaling && _this.options.onScaling.call(_this, paramsAction);
            } else if (_this.options.wheel.action === 'rotate') {
                _this.diffValues.rotate = (isUp ? _this.options.wheel.degree : -_this.options.wheel.degree);
                //纠正差值，获得正确的diff.rotate和this.nowValues.rotate（加上了差值）
                _this.correctRangeRotate();
                //更新参数
                paramsAction.rotate.diff = axNumber(_this.diffValues.rotate);
                paramsAction.rotate.value = axNumber(_this.nowValues.rotate);
                _this.handlers.hasOwnProperty('rotating') ? _this.emit('rotating', paramsAction) : null;
                _this.options.onRotating && _this.options.onRotating.call(_this, paramsAction);
            } else if (_this.options.wheel.action === 'translate') {
                //下滚，元素向下滑动，方向为正；上滚，元素向上滑动，方向为负；
                let direction = _this.options.wheel.axis === 'y' ? { x: 0, y: -dirValue, } : { x: -dirValue, y: 0 },
                    axisDistance = -dirValue * _this.options.wheel.distance,
                    x = _this.options.wheel.axis === 'y' ? 0 : axisDistance,
                    y = _this.options.wheel.axis === 'x' ? 0 : axisDistance;
                _this.diffValues.translate = { x, y, h: _this.options.wheel.distance };
                _this.updateNowTranslate(_this.diffValues.translate);
                //更新参数
                paramsAction.translate.diff = _this.diffValues.translate;
                paramsAction.translate.value = _this.nowValues.translate;
                paramsAction.translate.direction = direction;
                _this.handlers.hasOwnProperty('translating') ? _this.emit('translating', paramsAction) : null;
                _this.options.onTranslating && _this.options.onTranslating.call(_this, paramsAction);
            }
            //更新参数
            paramsMove[_this.options.wheel.action] = paramsAction[_this.options.wheel.action];
            _this.handlers.hasOwnProperty('move') ? _this.emit('move', paramsMove) : null;
            _this.options.onMove && _this.options.onMove.call(_this, paramsMove);
            //更新同步
            _this.lastValues[_this.options.wheel.action] = { ..._this.nowValues[_this.options.wheel.action] };
        }
        //失去目标节点
        this.cancelFun = function (e) {
            e.stopPropagation();
            e.preventDefault();
            //设置完成（数据归零）
            _this.setCompleted();
            let paramsCancel = Object.assign(axClone(_this.paramsFormat), {
                originEvent: e,
                coord: { x: e.clientX, y: e.clientY },
                ..._this.endParams,
                name: 'cancel'
            });
            _this.handlers.hasOwnProperty('cancel') ? _this.emit('cancel', paramsCancel) : null;
            _this.options.onCancel && _this.options.onCancel.call(_this, paramsCancel);
        }
        //右键事件
        this.menuFun = function (e) {
            e.stopPropagation();
            e.preventDefault();
            let paramsClick = Object.assign(axClone(_this.paramsFormat), {
                originEvent: e,
                coord: { x: e.clientX, y: e.clientY },
                ..._this.endParams,
                name: 'hold'
            });
            //长按事件
            _this.handlers.hasOwnProperty('hold') ? _this.emit('hold', paramsClick) : null;
            _this.options.onHold && _this.options.onHold.call(_this, paramsClick);
        }
        //初始化
        this.init();
    }
    getStartValues() {
        //transform初始值
        this.startValues = axGetStyleTransform(this.targetDom, ['translate', 'scale', 'rotate'], this.options.translate.positionInstead);
       
        this.nowValues = {
            translate: {
                ...this.startValues.translate,
                h: ~~this.calcDistance(this.startValues.translate.x, this.startValues.translate.y)
            },
            scale: { ...this.startValues.scale },
            rotate: this.startValues.rotate,
        }
        this.lastValues = axClone(this.nowValues);
    }
    init(callback) {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'gesture');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;
        //设置初始值
        this.options.initial ? axSetStyleTransform(this.targetDom, this.options.initial, this.options.translate.positionInstead) : null;
        //保存style.transform初始值
        this.initialValues = this.options.translate.positionInstead ?
            axGetStyleTransform(this.targetDom, ['translate', 'scale', 'rotate', 'skew'], this.options.translate.positionInstead) :
            this.targetDom.style.transform;
        //手指放下
        this.targetDom.addEventListener(axEventMap[0], this.startFun, { passive: false });
        //手指拖拽
        this.targetDom.addEventListener(axEventMap[1], this.moveFun, { passive: false });
        //手指松开
        this.targetDom.addEventListener(axEventMap[2], this.endFun, { passive: false });
        //目标丢失
        this.targetDom.addEventListener(axEventMap[3], this.cancelFun, { passive: false });


        //绑定鼠标缩放
        if (!this.isMobi) {
            //pc端滚轮事件
            if (this.options.wheel.action) {
                this.targetDom.addEventListener(this.scrollObj.event, this.wheelFun, { passive: false })
            }
            //pc端右键事件
            if (this.options.click.holdToMenu) {
                this.targetDom.addEventListener('contextmenu', this.menuFun);
            }
        }

        //离开父节点强制完成事件
        this.parentDom = this.options.parent ? axIdToDom(this.options.parent) : null;
        this.setCompleted();

        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        callback && callback.call(this);
        return this;
    }
    updateNowTranslate(diff) {
        let x = this.lastValues.translate.x + diff.x,
            y = this.lastValues.translate.y + diff.y;
        //限制范围
        x = axClampValue({ value: x, min: this.options.translate.min, max: this.options.translate.max });
        y = axClampValue({ value: y, min: this.options.translate.min, max: this.options.translate.max });
        this.nowValues.translate = { x: ~~x, y: ~~y };
    }
    getMoveValues(values, startCoord) {
        //c:coord,d:direction,h:hypotenuse,x:x-axis,y:y-axis
        let c = this.getCenterCoord(values),
            diff = this.getTriangleValues(c, startCoord),
            h = ~~diff.h,
            x = ~~diff.x,
            y = ~~diff.y,
            d = { x: x > 0 ? 1 : x < 0 ? -1 : 0, y: y > 0 ? 1 : y < 0 ? -1 : 0 };
        return { c, h, x, y, d };
    }
    setCompleted() {
        if (this.parentDom && !this.isMobi) {
            this.parentDom.addEventListener('mouseleave', this.endFun, { passive: false });
            this.options.wheel.action ? this.parentDom.addEventListener(this.scrollObj.event, this.wheelFun, { passive: false }) : null;
        }
        //更新状态
        this.started = false;
        this.moveCount = this.touchesMoveCount = 0;
        this.diffValues = { scale: 0, rotate: 0, translate: { x: 0, y: 0, h: 0 } };
        //清除长按定时器
        this.holdHandle ? clearTimeout(this.holdHandle) : null;
    }
    getOrign(coord, rect) {
        //获得相对elem的坐标值
        let x = coord.x - rect.left,
            y = coord.y - rect.top;
        x < 0 ? x = 0 : x > rect.width ? x = rect.width : null;
        y < 0 ? y = 0 : y > rect.height ? y = rect.height : null;
        return { x, y }
    }
    getTranslate(options) {
        //options = {coord:{x,y},diff:''}
        let style = axStyle(this.targetDom),
            width = parseInt(style.width),
            height = parseInt(style.height),
            origin = this.options.origin || { x: width / 2, y: height / 2 },
            rect = this.targetDom.getBoundingClientRect(),
            newOrigin = this.getOrign(options.coord, rect),
            transform = axGetStyleTransform(this.targetDom, ['translate'], this.options.translate.positionInstead),
            //获得相对原始尺寸的百分比
            percent = (attr) => {
                let value = newOrigin[attr] / rect[attr === 'x' ? 'width' : 'height'];
                //纠正值范围
                value < 0 ? value = 0 : value > 1 ? value = 1 : null;
                return value;
            },
            orginDiff = (attr) => {
                return (origin[attr] - (attr === 'x' ? width : height) * percent(attr)) * options.diff;
            },
            x = transform.translate.x + orginDiff('x'),
            y = transform.translate.y + orginDiff('y');
        return { x: ~~x, y: ~~y };
    }
    getTouchCoords(event, target = 'targetTouches') {
        let touches = [];
        if (this.isMobi) {
            for (let k of event[target]) {
                touches.push({
                    //id: k.identifier, 
                    x: ~~k.clientX,
                    y: ~~k.clientY
                });
            }
        } else {
            touches.push({
                //id: 0, 
                x: ~~event.clientX,
                y: ~~event.clientY
            });
        }
        return touches;
    }
    getCenterCoord(data) {
        let result,
            avg = (n, axis) => {
                let value = 0;
                for (let k = 0; k < n; k++) {
                    value += data[k][axis];
                }
                return value / n;
            };
        if (data.length > 1) {
            result = { x: ~~avg(2, 'x'), y: ~~avg(2, 'y') };
        } else if (data.length === 1) {
            result = data[0];
        } else {
            result = { x: 0, y: 0 };
        }
        return result;
    }
    calcDistance(x, y) {
        return Math.sqrt(x * x + y * y)
    }
    getTriangleValues(now, last) {
        let x = now.x - last.x,
            y = now.y - last.y,
            h = this.calcDistance(x, y),
            a = Math.atan2(y, x) * 180 / Math.PI;
        //Math.atan2()* 180 / Math.PI取值范围是0~180和0~-180
        //x:x-axis,y:y-axis,h:hypotenuse,a:angle
        return { x, y, h, a };
    }
    correctRangeRotate() {
        //先纠正总值确保在min和max范围
        this.nowValues.rotate = this.correctRangeValue('rotate', this.nowValues.rotate);
        let temp = this.nowValues.rotate + this.diffValues.rotate;
        //this.diffValues.rotate可纠正为0
        if (temp >= this.options.rotate.max) {
            this.diffValues.rotate = this.options.rotate.max - this.nowValues.rotate;
        } else if (temp <= this.options.rotate.min) {
            this.diffValues.rotate = this.options.rotate.min - this.nowValues.rotate;
        }
        //加上差值
        this.nowValues.rotate += this.diffValues.rotate;
    }
    correctDiffRotate() {
        if (this.triangleValues.now.a > 0 && this.triangleValues.last.a < 0) {
            if (this.triangleValues.now.a > 90 && this.triangleValues.last.a < -90) {
                //从第三到第二顺时针越过
                this.diffValues.rotate = 360 - this.diffValues.rotate;
            } else if (this.triangleValues.now.a < 90 && this.triangleValues.last.a > -90) {
                //从第四到第一逆时针越过
                this.diffValues.rotate = - this.diffValues.rotate;
            }
        } else if (this.triangleValues.now.a < 0 && this.triangleValues.last.a > 0) {
            if (this.triangleValues.now.a > -90 && this.triangleValues.last.a < 90) {
                //从第一到第四顺时针越过
                value = - this.diffValues.rotate;
            } else if (this.triangleValues.now.a < -90 && this.triangleValues.last.a > 90) {
                //从第二到第三逆时针越过
                this.diffValues.rotate = -(360 + this.diffValues.rotate);
            }
        }
    }
    correctRangeScale() {
        //先纠正总值确保在min和max范围
        this.nowValues.scale.x = this.correctRangeValue('scale', this.nowValues.scale.x);
        this.nowValues.scale.y = this.correctRangeValue('scale', this.nowValues.scale.y);

        let tempX = this.nowValues.scale.x + this.diffValues.scale,
            tempY = this.nowValues.scale.y + this.diffValues.scale;
        //this.diffValues.scale可纠正为0
        if (tempX >= this.options.scale.max || tempY >= this.options.scale.max) {
            this.diffValues.scale = Math.min((this.options.scale.max - this.nowValues.scale.x), (this.options.scale.max - this.nowValues.scale.y));
        } else if (tempX <= this.options.scale.min || tempY <= this.options.scale.min) {
            this.diffValues.scale = Math.min((this.options.scale.min - this.nowValues.scale.x), (this.options.scale.min - this.nowValues.scale.y));
        }
        //加上差值
        this.nowValues.scale.x += this.diffValues.scale;
        this.nowValues.scale.y += this.diffValues.scale;
    }
    correctRangeValue(prop, value) {
        let result;
        if (value < this.options[prop].min) {
            result = this.options[prop].min;
        } else if (value > this.options[prop].max) {
            result = this.options[prop].max;
        } else {
            result = value;
        }
        return result;
    }
    keepDiffZero(prop) {
        if ((this.nowValues[prop].x === this.options[prop].min && this.nowValues[prop].y === this.options[prop].min) || (this.nowValues[prop].x === this.options[prop].max && this.nowValues[prop].y === this.options[prop].max)) {
            this.diffValues[prop] = 0;
        }
    }
    correctWheelDirection(value) {
        return this.options.wheel.reverse ? !value : value;
    }
    reset(callback) {
        //恢复初始值
        this.targetDom.style.transform = this.initialValues;
        //监听回调
        this.handlers.hasOwnProperty('reset') ? this.emit('reset', '') : null;
        this.options.onReset && this.options.onReset.call(this);
        callback && callback.call(this);
        return this;
    }
    destroy(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //如果有剩余内容则先执行完毕
        this.setCompleted();
        //重置style.transform
        this.reset();
        //取消绑定事件
        this.targetDom.removeEventListener(axEventMap[0], this.startFun);
        this.targetDom.removeEventListener(axEventMap[1], this.moveFun);
        this.targetDom.removeEventListener(axEventMap[2], this.endFun);
        this.targetDom.removeEventListener(axEventMap[3], this.cancelFun);
        if (!this.isMobi) {
            if (this.options.wheel.action) {
                this.targetDom.removeEventListener(this.scrollObj.event, this.wheelFun)
            }
            if (this.options.click.holdToMenu) {
                this.targetDom.removeEventListener('contextmenu', this.menuFun);
            }
        }
        if (this.parentDom && !this.isMobi) {
            this.parentDom.removeEventListener('mouseleave', this.endFun);
            this.options.wheel.action ? this.parentDom.removeEventListener(this.scrollObj.event, this.wheelFun) : null;
        }

        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
 * 插件：轮播；使用方法：new axSwiper('#id',{参数})
 */
class axSlider {
    constructor(elem, options) {
        if (!elem) { return false; }
        // 始终将elem转成DOM对象
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }
        // 定义参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//存储名称，字符串格式；如果需要保存实例当前数据请设定唯一的值，比如：storageName:'mystorage001'
            content: '',//传入数据，需使用标准写法:[{thumb:'',image:'',label:'',text:'',callback:''},...]，
            //如果需要在分页中显示标题请填label属性
            //如果需要在分页中显示缩略图请填thumb属性或image属性
            //如果image和text都填写了，那么在滑块中会将image放上方，text放下方
            //如果填写callback，那么将需要用户手动添加滑块内容，callback支持一个参数即本项的数据
            separator: ',',//如果content为字符串，通过该参数分离出图片地址

            async: {
                enable: false,//是否启用异步获取数据，默认不启用，可选择true，那么content参数应该填写一个json或动态页面地址
                type: 'post',//异步提交的方式，默认post，可填get
                source: 'json',//异步数据源，默认json，可选择sql（使用sql数据源，通常需要设置data参数）
                data: null,//异步请求传递的数据，默认为空即不向后台传递数据，通常source为sql才需要填写该属性
            },
            lazyShow: false,//是否开启懒加载，默认false不开启内容会一次性加载完毕；可设为true对内容实施懒加载，避免加载内容过多占用太多内存
            lazyload: {
                visible: [0.2],//滑块出现在容器中，并且是自身尺寸0.2的时候才加载
                dataName: 'src',
                onAsync: '',//如果滑块从异步获取数据，数据类型是json，应该使用该监听方法生成滑块内容，详细请参考axLazyload
            },//懒加载参数，有axLazyload相同写法
            active: 0,//初始化索引，默认为0即第一个滑块
            duration: 500,//移动一个滑块的时间，也就是相对移动速度，默认500，单位ms，
            throttle: 0,//对点击前后导航操作以及键盘事件使用节流，默认600（稍大于动画时间，确保在动画完成之后触发其他事件），可选择0表示禁用
            slides: 1,//窗口每次可显示几个元素，可填一个大于等于1的数值，插件将根据窗口尺寸自动计算元素宽/高以适配，如果填auto插件将按元素默认处理
            grouped: false,//是否将窗口的多个元素成组滚动，默认false不成组，可选择true即：当slides>1时，使用导航或滑动将一次性移动多个子元素
            loop: false,//是否左右无缝滚动，默认false即不启用，可选择true启用，slides为数字类型才有效
            unsnappedAttr: 'unsnapped',//使用分组后会对子元素设置禁止吸附的属性，字符串格式，默认unsnapped
            autoHeight: false,//是否让外容器随滑块的高度自动伸缩变化，默认false即不做高度设置随css，可选择true由js控制高度
            gap: 0,//滑块间隙，默认0即无间隙，单位px，对slides>0有效
            centered: false,//是否让滑块居中，也就是平移吸附点到中间
            centeredFill: false,//使用centered参数会在起始部位产生空挡，使用改参数可避免空挡，默认false，可使用true启用，原理是自动修改当前active值
            divide: 1,//是否以类似网格形式布局，与slides参数相对，该参数设置行数（列数），参数gap也将生效
            keyboard: false,//是否能使用键盘的方向键控制滑动，默认false禁止，可选择true允许
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            autoplay: {
                enable: false,
                delay: 4000,//进入下一个滑块的延迟时间，默认2000，单位ms，可自定义延迟时间可设定一个正整数
                countdown: false,//是否显示倒计时进度条，默认false不显示，可选择true显示
            },//是否自动轮播，默认false手动轮播，如果设为true将以delay值自动轮播
            scroll: {
                axis: 'x',
                snap: {
                    enable: true,
                },
                drift: {
                    timeTHR: 200,
                }
            },//axScroll虚拟滚动的参数，详细请查看：ax-utils-scroll.php
            scale: {
                enable: false,
                min: 1,
                max: 3,
            },
            scaleGap: 50,//缩放状态拖出多少空白进入下一个滑块，默认50，单位px
            tools: [], //每一项的属性有：name(默认有download/zoom/close)、text、icon、action(url/function)、props
            total: {
                enable: false,//是否显示分页统计，默认不显示，可选择true显示
                placement: 'left-top',//显示位置，默认在左上角，可选择right-top、left-bottom、right-bottom
            },//该参数可按total:false/true设置亦可
            pagination: {
                enable: false,//是否启用，默认false不启用，可选择true禁用
                className: '',//可追加一个或多个class样式，多个用英文逗号或空格分开
                type: 'lamp',//导航类型，默认lamp指示灯，可选择thumb缩略图、index带索引的指示灯、text带标题
                selector: '',//节点选择器，为空则内部自动创建
                data: [],//分页数据
                callback: '',//如果填了callback表示手动创建分页器，支持一个参数，其为一个对象，包含属性：index（当前分页）、total（总页数）、target（目标滑块）、parent（分页容器）
            },//该参数可按pagination:false/true设置
            navigation: {
                enable: false,//是否启用，默认false不启用，可选择true禁用
                className: '',//可追加一个或多个class样式，多个用英文逗号或空格分开
                Filled: false,//是否为色块填充，默认false，可选择true
                prev: {
                    selector: '',//节点选择器
                    className: '',//可追加一个或多个class样式，多个用英文逗号或空格分开
                    text: '',//节点的文本（HTML格式）
                },
                next: {
                    selector: '',//节点选择器
                    className: '',//可追加一个或多个class样式，多个用英文逗号或空格分开
                    text: '',//节点的文本（HTML格式）
                },
            },//该参数可按navigation:false/true设置
            integrated: false,//是否将分页和导航整合在一起作为一个整体出现，默认false，可选择true

            onInit: '',//回调函数，初始化后执行，无参数
            onSlide: '',//回调函数，滑动开始前执行，支持一个参数即滑动目标bullet项
            onSlided: '',//回调函数，滑动完成后执行，支持一个参数即当前bullet项
            onStartPlay: '',//回调函数，启动自动播放后执行，无参数
            onStopPlay: '',//回调函数，停止播放后执行，无参数
            onToStart: '',//回调函数，滑到第一个滑块并完成后执行，支持一个参数即当前bullet项
            onToEnd: '',//回调函数，滑到最后一个滑块并完成后执行，支持一个参数即当前bullet项
            onZoomin: '',//回调函数，放大后执行，支持一个参数即当前slide项
            onZoomout: '',//回调函数，缩小后执行，支持一个参数即当前slide项
            onUpdated: '',//回调函数，参数更新后执行，支持一个参数即设置的参数
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
        }, options, this.targetDom, this.constructor.name);
        // 定义监听事件成员
        this.handlers = {};
        //属性对应索引
        if (this.options.scroll.axis === 'x') {
            this.propsMap = {
                position: 'left',
                gap: 'marginLeft',
                client: 'clientX',
                size: 'width',
                offset: 'offsetLeft',
                reverse: {
                    size: 'height',
                    gap: 'marginTop'
                }
            }
        } else if (this.options.scroll.axis === 'y') {
            this.propsMap = {
                position: 'top',
                gap: 'marginTop',
                client: 'clientY',
                size: 'height',
                offset: 'offsetTop',
                reverse: {
                    size: 'width',
                    gap: 'marginLeft'
                }
            }
        }
        //定时轮播函数
        this.delayScroll = () => {
            //到最后一个或者
            if (this.activeIndex >= this.bullets.at(-1).index) {
                //先取消定时
                this.stopPlay();
                //到尾了，则跳回头部第一个
                this.toStart(() => {
                    //跳到头部回调后，重启定时器
                    this.startPlay();
                });
            } else {
                //设置到计时
                this.setCountdown();
                this.toNext();
            }
        }
        //循环复制头尾节点数组
        this.startClone = [];
        this.endClone = [];
        //键盘事件
        this.keyToSwipe = axThrottle((e) => {

            if (e.code === 'ArrowLeft') {
                this.toPrev();
            }
            if (e.code === 'ArrowRight') {
                this.toNext();
            }
            /* if (this.options.scroll.axis === 'x') {
                
            } else {
                if (e.code === 'ArrowUp') {
                    this.toPrev();
                }
                if (e.code === 'ArrowDown') {
                    this.toNext();
                }
            } */
        }, this.options.throttle, true);
        //初始化
        this.init();
        return this;
    }
    begin() {
        //定义标准数据
        this.data = [];
        //列头节点数组
        this.bullets = [];
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'slider');
        //设置销毁状态
        this.destroyed = false;
        //判断是否有存储值而更新参数
        if (this.options.storageName) {
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            } else {
                //则先存下空对象
                axLocalStorage.set(this.options.storageName, {});
            }
        }
        //初始化第一个可视索引
        this.activeIndex = this.options.active;
        //修正参数

        //分页参数
        if (this.options.pagination === true) {
            this.options.pagination = {
                enable: true,
                type: 'lamp',
            }
        } else if (this.options.pagination === false) {
            this.options.pagination = {
                enable: false,
            }
        }
        //前后导航
        if (this.options.navigation === true) {
            this.options.navigation = {
                enable: true,
            }
        } else if (this.options.navigation === true) {
            this.options.navigation = {
                enable: false,
            }
        }
        //自动滚动
        if (this.options.autoplay === true) {
            this.options.autoplay = { enable: true, delay: 4000 }
        } else if (this.options.autoplay === false) {
            this.options.autoplay = { enable: false }
        }
        //播放状态，与autoplay相关
        this.pause = true;
        //滑块总计
        if (this.options.total === true) {
            this.options.total = { enable: true, placement: 'left-top' }
        } else if (this.options.total === false) {
            this.options.total = { enable: false }
        }
        //异步参数
        if (this.options.async === true) {
            this.options.async = { enable: true, type: 'post', source: 'json' }
        }
        //缩放参数
        if (this.options.scale === true) {
            this.options.scale = {
                enable: true,
                min: 1,
                max: 3,
            }
        }
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
    }
    getDistMouse() {
        let result = 0;
        if (this.options.loop) {
            //循环必须滚动到下一个吸附节点，否则因为回弹耽误时间导致locateTo不及时
            result = 'snap';
        } else if (this.options.grouped && this.options.slides > 1) {
            //分组并且一页是2个以上，则设定禁止吸附属性，确保滚轮一滚必然到可吸附节点上
            result = 'snap';
        } else if (this.options.slides >= 1) {
            if (this.options.gap) {
                //具体数量
                result = this.slideSize + this.options.gap;
            } else {
                //一个比例
                result = 1 / this.options.slides;
            }
        }
        //为0则按scroll插件的默认距离进行，即父尺寸的1/4
        return result;
    }
    getPaddingStart() {
        return this.options.slides >= 1 && this.options.centered ? (this.targetSize - this.slideSize) / 2 : 0;
    }
    getInitSlide() {
        if (this.options.centered && this.options.centeredFill) {
            if (this.options.slides > 1 && this.activeIndex < ~~(this.options.slides / 2)) {
                this.activeIndex = ~~(this.options.slides / 2);
            }
        } else if (this.activeIndex > this.slides.length - 1) {
            this.activeIndex = this.slides.length - 1;
        }
        return this.slides[this.activeIndex].dom
    }
    init() {
        this.begin();
        //创建容器
        this.createWrapper();
        //给容器设置样式，给节点设置尺寸，以确认子元素的所在位置
        this.setAttribute();
        //进入视窗侦听
        this.spyIns = new axSpy(this.targetDom, {
            visible: [1],
            onShown: () => {
                //使用键盘控制滚动（添加）
                this.options.keyboard ? window.addEventListener('keydown', this.keyToSwipe) : null;
            },
            onOut: () => {
                //使用键盘控制滚动（取消）
                this.options.keyboard ? window.removeEventListener('keydown', this.keyToSwipe) : null;
            }
        });

        //获得滑块的标准数据this.slides，[{type:'',label:'',dom:'',content:''}...]
        this.getSlidersData((data) => {
            //获得标准数据
            this.slides = data;
            //监听slides数组增减变化，this.slidesObserve.proxy 是 this.slides的代理对象
            this.setSlidesObserve();
            //根据this.slides，设置slides尺寸，以及填充分页器和导航器
            this.fillTarget();
            //设置拖拽和滚动
            this.setScroll();
            //设置笼子
            this.setCage();
            //无缝滚动
            this.setLoopPlay();
            //自动滚动
            this.setAutoPlay();
            //监听回调
            this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
            this.options.onInit && this.options.onInit.call(this);
            return this;
        });
    }
    getUnbounds() {

        return [...this.scrollIns.unbounds, this.prevDom, this.nextDom, this.toolsDom, this.countdownDom, ...this.bullets.map(k => k.dom)].filter(Boolean);
    }
    setScroll() {
        //创建滚动参数
        this.scrollOpt = axExtend({
            wrapper: this.wrapperDom,
            transInit: this.getInitSlide(),
            duration: this.options.duration,
            //如果gap>0的则需要加上间隙占比
            distMouse: this.getDistMouse(),
            paddingStart: this.getPaddingStart(),
            paddingEnd: this.getPaddingStart(),
            //循环模式必须采用false，禁止插入
            insertable: this.options.loop ? false : true,
        }, this.options.scroll);
        //合并组就需要设置禁止吸附
        if (this.options.grouped && this.options.slides > 1) {
            //querySelector('[ATTR]')
            this.scrollOpt.snap.unsnapped = `[${this.options.unsnappedAttr}]`;
        }
        //设置漂移模式，固定滑块数量的slider漂移整个父尺寸，auto则自定义不强制限制
        if (this.options.slides >= 1) {
            //快速拖拽产生漂移的移动距离为整个外容器尺寸，如果有间隙则加上间隙距离
            this.scrollOpt.drift.mode = this.options.gap ? this.targetSize + this.options.gap : 1;
        }
        //创建滚动实例
        this.scrollIns = new axScroll(this.targetDom, this.scrollOpt);
        //将导航按钮加入到非绑定元素
        this.scrollIns.unbounds = this.getUnbounds();
        //异步监听snapped，获得当前的吸附项并修改this.activeIndex
        this.scrollIns.on('snap', (item) => {
            //开始滑动状态
            this.slided = false;
            let bullet = this.bullets.find(k => k.target.dom === item);
            if (bullet) {
                //改变this.activeIndex值
                this.activeIndex = bullet.index;
                //如果有统计则更新文本
                this.numeratorDom ? this.numeratorDom.innerHTML = this.activeIndex + 1 : null;
                //修改导航active
                this.setBulletActive(bullet);
                //设置前后导航按钮disabled状态
                this.setNavState();

                //自动高度
                if (this.options.autoHeight) {
                    this.targetDom.style.height = axStyle(bullet.target.dom).height;
                }
                //分页监听
                this.handlers.hasOwnProperty('slide') ? this.emit('slide', bullet) : null;
                this.options.onSlide && this.options.onSlide.call(this, bullet);
                //保存数据
                this.save();
            };
            //取消缩放
            let scaledSlides = this.slides.filter(k => k.scaled);
            if (scaledSlides.length > 0) {
                //end事件会被中断，所以在销毁前先执行end事件
                scaledSlides.forEach(k => {
                    this.toZoomout(k);
                });
            }
        });
        this.scrollIns.on('snapped', (item) => {
            let index = this.slides.findIndex(k => k.dom === item),
                bullet = this.bullets.find(k => k.target.dom === item);
            if (index === -1) {
                //吸附的是因循环复制的节点，直接跳转到真实节点
                if (this.endClone.length > 0 && this.endClone.includes(item)) {
                    let n = this.endClone.findIndex(k => k === item);
                    this.scrollIns.locateTo(this.slides.at(n - this.totalClone).dom);
                } else if (this.startClone.length > 0 && this.startClone.includes(item)) {
                    let n = this.startClone.findIndex(k => k === item);
                    this.scrollIns.locateTo(this.slides[n].dom);
                }
            }
            //完成后监听
            if (bullet) {
                //分页监听
                this.handlers.hasOwnProperty('slided') ? this.emit('slided', bullet) : null;
                this.options.onSlided && this.options.onSlided.call(this, bullet);

                if (bullet === this.bullets[0]) {
                    //第一项，回调
                    this.handlers.hasOwnProperty('toStart') ? this.emit('toStart', bullet) : null;
                    this.options.onToStart && this.options.onToStart.call(this, bullet);

                } else if (bullet === this.bullets.at(-1)) {
                    //最后一项，回调
                    this.handlers.hasOwnProperty('toEnd') ? this.emit('toEnd', bullet) : null;
                    this.options.onToEnd && this.options.onToEnd.call(this, bullet);
                }
            }

        });
        this.scrollIns.on('move', (value) => {
            //拖拽到了头部或尾部
            if (this.options.loop) {
                if (value > 0) {
                    //无缝循环模式，立即跳转到倒数[第1、或n个]滑块
                    this.scrollIns.locateTo(this.slides.at(-this.totalClone).dom);
                } else if (value < this.scrollIns.difference) {
                    //无缝循环模式，立即跳转到第一个滑块
                    this.scrollIns.locateTo(this.slides[0].dom);
                }
            }
            //取消自动播放，true为取消倒计时
            this.stopPlay(true);
        });
        this.scrollIns.on('resized', (rect, dom) => {
            if (dom === this.targetDom) {
                //更新this.slideSize并重置slide的width/height以及margin-top/margin-right
                this.resetSlidesSize(rect);
                //重新自动吸附
                this.scrollIns.setSnapped();
            }
        });
        //鼠标滚轮监听
        this.scrollIns.on('mouse', () => {
            this.stopPlay(true);
        });
    }
    setCage() {
        if (!this.options.scale.enable) {
            return false;
        }
        for (let k of this.slides) {
            //节点有unscalable属性则跳过
            if (k.dom.hasAttribute('unscalable') || k.dom.children.length === 0) {
                continue;
            }


            k.cage = new axCage(k.dom, {
                init: false,
                drift: { enable: true },
                scale: this.options.scale,
                completedTHR: this.options.scaleGap,
                onCompleted: (dir) => {
                    if (this.options.scroll.axis === 'x') {
                        dir === 'left' ? this.toNext() : dir === 'right' ? this.toPrev() : null;
                    } else if (this.options.scroll.axis === 'y') {
                        dir === 'top' ? this.toNext() : dir === 'bottom' ? this.toPrev() : null;
                    }
                }
            });
            k.dom.ondblclick = () => {
                //或没有子节点则不再执行
                if (k.dom.children.length === 0) {
                    return false;
                }

                if (k.scaled) {
                    this.toZoomout(k);
                } else {
                    this.toZoomin(k);
                }

            }
        }
    }
    scaleToggle(item, type) {
        if (type === 'destroy') {
            item.cage.destroy();
            Reflect.deleteProperty(item, 'scaled');
        } else {
            item.cage.init(() => {
                item.cage.setValues({ value: this.options.scale.max, type: 'scale' });
            });
            item.scaled = true;
        }
    }
    findSlide(data) {
        let dataType = axType(data), slide = null;
        if (!data) {
            slide = this.slides[this.activeIndex];
        } else {
            if (dataType === 'Number') {
                //slide索引
                slide = this.slides[data];
            } else if (dataType === 'Object') {
                //slide对象
                slide = data;
            } else if (dataType.includes('HTML')) {
                //节点
                slide = this.slides.find(k => k.dom === data);
            }
        }
        return slide;
    }
    toZoomin(item, callback) {
        let slide = this.findSlide(item);
        if (slide || slide.children.length > 0) {
            this.scaleToggle(slide, 'init');
            callback && callback.call(this);
            this.handlers.hasOwnProperty('zoomin') ? this.emit('zoomin', slide) : null;
            this.options.onZoomin && this.options.onZoomin.call(this, slide);
        }
    }
    toZoomout(item, callback) {
        let slide = this.findSlide(item);
        if (slide || slide.children.length > 0) {
            this.scaleToggle(slide, 'destroy');
            callback && callback.call(this);
            this.handlers.hasOwnProperty('zoomout') ? this.emit('zoomout', slide) : null;
            this.options.onZoomout && this.options.onZoomout.call(this, slide);
        }
    }
    setAttribute() {
        //设置样式类
        !this.targetDom.classList.contains('ax-slider') ? this.targetDom.classList.add('ax-slider') : null;
        this.options.divide > 1 ? this.targetDom.setAttribute('display', 'grid') : null;
        //针对axis:x和slides:auto，需要对wrapper设置white-space:nowrap，虽然在scroll中会再次追加axis属性，但是已经在节点创建完毕之后，已经晚了
        this.targetDom.setAttribute('axis', this.options.scroll.axis);
        this.targetDom.setAttribute('slides', this.options.slides);

    }
    createSliderDom(obj) {
        let content;
        if (obj.type === 'image') {
            //image类型，使用懒加载
            props = { type: 'image' };
            if (this.options.lazyShow) {
                content = `<img src="${axImage.blank}" data-${this.options.lazyload.dataName}="${obj.content}"/>`;
            } else {
                content = `<img src="${obj.content}"/>`;
            }
            obj.dom = axAddElem('li', { type: 'image' }, content);
        } else if (obj.type === 'video') {
            //image类型，使用懒加载
            if (this.options.lazyShow) {
                content = `<video controls="controls" data-${this.options.lazyload.dataName}="${obj.content}"></video>`;
            } else {
                content = `<video controls="controls" src="${obj.content}"></video>`;
            }
            obj.dom = axAddElem('li', { type: 'video' }, content);
        } else if (obj.type === 'audio') {
            //image类型，使用懒加载
            if (this.options.lazyShow) {
                content = `<audio controls="controls" data-${this.options.lazyload.dataName}="${obj.content}"></audio>`;
            } else {
                content = `<audio controls="controls" src="${obj.content}"></audio>`;
            }
            obj.dom = axAddElem('li', { type: 'audio' }, content);
        } else if (obj.type === 'url') {
            //url类型，就是iframe标签，使用懒加载
            if (this.options.lazyShow) {
                content = `<iframe data-${this.options.lazyload.dataName}="${obj.content}" frameborder="0"></iframe>`;
            } else {
                content = `<iframe src="${obj.content}" frameborder="0"></iframe>`;
            }
            obj.dom = axAddElem('li', { type: 'url' }, content);
        } else if (obj.type === 'node') {
            //节点类型
            if (obj.content.nodeName === 'LI') {
                //content传来的本身就是li则就是dom
                obj.dom = obj.content;
                obj.dom.setAttribute('type', 'node');
            } else {
                obj.dom = axAddElem('li', { type: 'node' });
                obj.dom.appendChild(obj.content);
            }
        } else if (obj.type === 'async') {
            //异步，内容未定
            if (this.options.lazyShow) {
                let props = { type: 'async' }
                props[`data-${this.options.lazyload.dataName}`] = obj.content;
                //data={limit:4}，向sql请求数据时需要
                !axIsEmpty(obj.data) ? props['data-async'] = JSON.stringify(obj.data) : null;
                obj.dom = axAddElem('li', props);
            } else {
                obj.dom = axAddElem('li', { type: 'async' });
                axAjax({
                    url: obj.content,
                    type: this.options.async.type,
                    data: obj.data,
                    before: () => {
                        obj.dom.setAttribute('loading', '');
                    },
                    success: (response) => {
                        //手动清除loading
                        obj.dom.removeAttribute('loading');
                        //创建内容
                        if (typeof response.content === 'string') {
                            //如果获得的是字符串则直接作为HTML载入
                            obj.dom.innerHTML = response.content;
                        } else {
                            //如果是其他，数组或对象则需要使用callback
                            obj.callback && obj.callback.call(this, response.content, obj)
                        }
                    }
                });
            }
        } else if (obj.type === 'text' || !obj.type) {
            //文本类型
            obj.type = 'text';
            obj.content = obj.content.trim();
            if (axStrType(obj.content)) {
                //{type:'text',content:'#demo/.demo'}
                content = document.querySelector(obj.content).innerHTML
            } else {
                //{type:'text',content:'文字'}
                content = obj.content;
            }
            obj.dom = axAddElem('li', { type: 'text' }, content);
        }
    }
    getSlidersData(callback) {
        let data = [];
        if (this.wrapperDom.innerHTML) {
            //页面有ax-slider-wrapper节点
            data = [...this.wrapperDom.children].map(k => {
                let obj = { dom: k };
                k.getAttribute('label') ? obj.label = k.getAttribute('label') : null;
                if (k.children.length === 1) {
                    let child = k.children[0],
                        type = child.nodeName;
                    if (type === 'IMG') {
                        obj.type = 'image';
                        obj.content = child.src;
                        obj.label = child.alt || child.title || k.getAttribute('label') || '';
                        obj.poster = k.getAttribute('poster') || '';
                    } else if (type === 'IFRAME') {
                        obj.type = 'url';
                        obj.content = child.src;
                        obj.label = child.alt || child.title || k.getAttribute('label') || '';
                        obj.poster = k.getAttribute('poster') || '';
                    } else if (type === 'VIDEO') {
                        obj.type = 'video';
                        obj.content = child.src;
                        obj.label = k.getAttribute('label') || '';
                        obj.poster = child.getAttribute('poster') || k.getAttribute('poster') || '';
                    } else if (type === 'AUDIO') {
                        obj.type = 'audio';
                        obj.content = child.src;
                        obj.label = k.getAttribute('label') || '';
                        obj.poster = k.getAttribute('poster') || '';
                    }
                } else {
                    obj.type = 'text';
                    obj.content = k.innerHTML;
                    obj.label = k.getAttribute('label') || '';
                    obj.poster = k.getAttribute('poster') || '';
                }
                return obj;
            });
            callback(data);
        } else {
            let dataType = axType(this.options.content), content;
            if (dataType === 'String') {
                if (this.options.async.enable) {
                    //异步json
                    axAjax({
                        url: this.options.content,
                        type: this.options.async.type,
                        data: this.options.async.data,
                        before: () => {
                            this.targetDom.setAttribute('loading', '');
                        },
                        success: (response) => {
                            //手动清除loading
                            this.targetDom.removeAttribute('loading');
                            //创建节点
                            response.content.forEach(k => {
                                this.createSliderDom(k);
                            });
                            data = response.content;
                            callback(data);
                        }
                    });
                } else {
                    //否则判断为图片地址字符串，'xxx/1.jpg,xxx/2.jpg,xxx/3.jpg'
                    data = this.options.content.split(this.options.separator).map(k => {
                        content = this.options.lazyShow ? `<img src="${axImage.blank}" data-${this.options.lazyload.dataName}="${k}"/>` : `<img src="${k}"/>`;
                        return { type: 'image', content: k, dom: axAddElem('li', { type: 'image' }, content) };
                    });
                    callback(data);
                }
            } else if (dataType === 'Array') {
                let contentType = axType(this.options.content[0]);
                if (contentType === 'String') {
                    //图片地址数组
                    data = this.options.content.map(k => {
                        content = this.options.lazyShow ? `<img src="${axImage.blank}" data-${this.options.lazyload.dataName}="${k}"/>` : `<img src="${k}"/>`;
                        return { type: 'image', content: k, dom: axAddElem('li', { type: 'image' }, content) };
                    });
                } else if (contentType.includes('HTML')) {
                    //节点数组
                    data = this.options.content.map(k => {
                        let obj = { type: 'node', content: k };
                        if (k.nodeName === 'LI') {
                            obj.dom = k;
                            obj.dom.setAttribute('type', 'node');
                        } else {
                            obj.dom = axAddElem('li', { type: 'node' });
                            obj.dom.appendChild(k);
                        }
                        return obj;
                    });
                } else if (contentType === 'Object') {
                    //标准对象数组
                    this.options.content.forEach(k => {
                        this.createSliderDom(k);
                    });
                    data = this.options.content;
                }
                callback(data);
            }
        }

    }

    getAccuracy(value, higher = true) {
        //取小数一位
        let places = value.toString().split(".")[1] ? value.toString().split(".")[1].length : 0;
        //小数位数大于3则粗略定义为除不尽（类似10/3），除不尽会导致几个滑块无法铺满外容器，所以刻意加了0.1，并向上取整1位小数
        return places > 3 ? axNumber(value + (higher ? 0.1 : 0), { mode: higher ? 'ceil' : 'floor', places: 1 }) : value;
    }
    getSingleSize() {
        //因为间隙的存在，需要先减掉间隙之和，算出来净值
        let netValue = this.targetSize - this.options.gap * (this.options.slides - 1),
            singleValue = netValue / this.options.slides;
        //保存变量，由于gap值是准确的，true会导致计算偏大以至于不能百分百填满，所以设为false
        this.slideSize = this.getAccuracy(singleValue, this.options.gap > 0 ? false : true);

        //使用多行多列情况下获取反向值
        if (this.options.divide > 1) {
            let netValueDivide = parseInt(axStyle(this.targetDom)[this.propsMap.reverse.size]) - this.options.gap * (this.options.divide - 1),
                singleValueDivide = netValueDivide / this.options.divide;
            //由于gap值是准确的，true会导致计算偏大以至于不能百分百填满，所以设为false
            this.slideSizeDivide = this.getAccuracy(singleValueDivide, this.options.gap > 0 ? false : true);
        }

    }
    setSingleSize(item) {
        //slides是可以填auto的，auto>0 === false
        if (this.options.slides > 0) {
            item.dom.style[this.propsMap.size] = this.slideSize + 'px';
        }
    }
    setSingleGap(item, index) {
        index = index === 0 ? 0 : index || this.slides.findIndex(k => k === item);
        if (this.options.divide > 1) {
            item.dom.style[this.propsMap.reverse.size] = this.getAccuracy(this.slideSizeDivide, false) + 'px';
            if (index % this.options.divide !== 0 && this.options.gap) {
                //不被整除的，加上margin-top
                item.dom.style[this.propsMap.reverse.gap] = this.options.gap + 'px';
            }
            //第一组之后才有margin-left
            if (index >= this.options.divide) {
                item.dom.style[this.propsMap.gap] = this.options.gap + 'px';
            }
        } else {
            if (index !== 0) {
                item.dom.style[this.propsMap.gap] = this.options.gap + 'px';
            }
        }
    }
    resetSlidesSize(rect) {
        //获得父尺寸-获得滑块尺寸-获得间隙尺寸-设置width/margin属性
        let size = !axIsEmpty(rect) ? rect[this.propsMap.size] : '';
        //获得父尺寸
        this.getTargetSize(size);
        //获得this.slideSize
        this.getSingleSize();
        //添加slide的width/height以及gap（margin-top/margin-right）
        this.setSlidersSize();
    }
    resetSlidesGap() {
        //slides是可以填auto的，auto>0 === false
        if (this.options.slides > 0) {
            //设置行数
            if (this.options.divide > 1) {
                this.slides.forEach((k, i) => {
                    this.setSingleGap(k, i);
                });
            } else {
                //设置滑块间隙margin-left
                if (this.options.gap > 0) {
                    this.slides.forEach((k) => {
                        this.setSingleGap(k);
                    });
                }
            }
        }
    }
    getTargetSize(size) {
        let targetStyle = axStyle(this.targetDom),
            targetHeight = parseInt(targetStyle.height),
            targetWidth = parseInt(targetStyle.width);
        if (this.options.scroll.axis === 'x') {
            this.targetSize = size || targetWidth;
        } else if (this.options.scroll.axis === 'y') {
            this.targetSize = size || targetHeight;
        }
    }
    setSlidersSize() {
        //前提是需要使用getTargetSize方法获得target尺寸
        //slides是可以填auto的，auto>0 === false
        if (this.options.slides > 0) {
            //设置滑块尺寸
            this.slides.forEach(k => {
                this.setSingleSize(k);
            });
            //重置间隙
            this.resetSlidesGap();
        }
    }
    createWrapper() {
        //添加wrapper节点
        if (this.targetDom.querySelector('.ax-slider-wrapper')) {
            this.wrapperDom = this.targetDom.querySelector('.ax-slider-wrapper');
        } else {
            this.wrapperDom = axAddElem('ul', { class: 'ax-slider-wrapper' }, '');
            this.targetDom.appendChild(this.wrapperDom);
        }
        //获取导航节点，有两个方法，1是已经存在物理节点；2是根据integrated为true来创建
        if (this.targetDom.querySelector('.ax-slider-navigation')) {
            this.navigationDom = this.targetDom.querySelector('.ax-slider-navigation');
        } else if (this.options.integrated) {
            this.navigationDom = axAddElem('div', { class: 'ax-slider-navigation' }, '');
            this.targetDom.appendChild(this.navigationDom);
        }
    }
    fillTarget() {
        //content有值，说明wrapperDom是空的，需要从数据中创建节点
        if (this.options.content) {
            let fragment = document.createDocumentFragment();
            for (let k of this.slides) {
                fragment.appendChild(k.dom);
            }
            this.wrapperDom.appendChild(fragment);
        }
        //获得this.slideSize并设置滑块尺寸和间距
        this.resetSlidesSize();
        //创建懒加载(重设滑块尺寸之后才可使用懒加载，否则这些滑块没有高度或宽度直接显示在外容器里，那基本都会加载完毕了，没有lazy效果了)
        this.lazyIns = this.options.lazyShow ? new axLazyload(this.wrapperDom, axExtend({ parent: this.targetDom }, this.options.lazyload)) : null;
        //指示灯数组
        this.getBullets();
        //添加前后导航
        this.fillNavigation();
        //添加分页导航
        this.fillPagination();
        //添加总数统计
        this.fillTotal();
        //添加工具
        this.fillTools();
    }
    setSlidesObserve() {
        //监听数组变化
        this.slidesObserve = new axObserve(this.slides);
        this.slidesObserve.instance.on('complete', (data) => {
            //data={proxy,target,type:'complete'}
            //重置间隙
            this.resetSlidesGap();
            //重置分页
            this.getBullets();
            this.fillPagination();
            //修正prev和next状态
            this.setNavState()
            //修改scroll.unbounds
            this.scrollIns.unbounds = this.getUnbounds();
            //修正数据统计
            this.resetTotal();
            //触发吸附
            this.scrollIns.setSnapped({ force: true });
        });
    }
    getBullets() {
        //先清空
        this.bullets = [];
        this.paginationDom ? this.paginationDom.innerHTML = '' : null;

        let len = this.slides.length,
            wrapperSize = this.wrapperDom.getBoundingClientRect()[this.propsMap.size];
        if (this.options.slides === 'auto') {
            this.slides.forEach((k, i) => {
                //子元素的offsetTop（left）值小于内外容器差值
                if (k.dom[this.propsMap.offset] <= wrapperSize - this.targetSize) {
                    this.bullets.push({ target: k, dom: axAddElem('li'), index: i });
                }
            });
        } else {
            this.slides.forEach((k, i) => {
                //如果slides===1，则每个子元素都可以滚动到，否则最后的若干元素是不能滚动到的
                let obj = { target: k, dom: axAddElem('li'), index: i };
                if (this.options.centered && this.options.slides > 0) {
                    //如果滑块居中，且非auto模式，每一项都是
                    this.bullets.push(obj);
                } else {
                    if (this.options.loop) {
                        if (this.options.grouped && this.options.slides > 1) {
                            //能整除就是bullut，否则禁止吸附
                            if (i % this.options.slides === 0) {
                                this.bullets.push(obj);
                            } else {
                                //不能除尽则追加不可吸附属性
                                k.dom.setAttribute(this.options.unsnappedAttr, '');
                            }
                        } else {
                            //如果不成组，每一项都是bullet
                            this.bullets.push(obj);
                        }
                    } else {
                        if (this.options.grouped && this.options.slides > 1) {
                            //index:0,1,2,3,4,5,6,7,8,9;visible:3,那么0,3,6,7都是bullut,其他的禁止吸附
                            if ((i % this.options.slides === 0 && this.slides.length - i >= this.options.slides) || i === len - this.options.slides) {
                                //不能除尽则追加不可吸附属性
                                this.bullets.push(obj);
                            } else {
                                k.dom.setAttribute(this.options.unsnappedAttr, '');
                            }
                        } else {
                            //index:0,1,2,3,4,5,6,7,8,9;visible:3,那么0,1,2,3,4,5,6,7都是bullut
                            if (i <= len - this.options.slides) {
                                this.bullets.push(obj);
                            }
                        }

                    }
                }

            });
        }


    }
    fillPagination() {
        if (!this.options.pagination.enable) {
            return false;
        }
        //分页是另一个slider实例
        this.pagesIns = null;
        if (this.targetDom.querySelector('.ax-slider-pagination')) {
            this.paginationDom = this.targetDom.querySelector('.ax-slider-pagination');
        } else if (this.options.pagination.selector) {
            //this.paginationDom可能为null的
            let selectorType = axType(this.options.pagination.selector);
            if (selectorType.includes('HTML')) {
                this.paginationDom = this.options.pagination.selector;
            } else if (selectorType === 'String') {
                if (axStrType(this.options.pagination.selector)) {
                    //是节点选择器，页面中能找到节点
                    this.paginationDom = axIdToDom(this.options.pagination.selector);
                } else {
                    //找不到则被认为是其他slider实例的实例名
                    this.pagesIns = axInstance.find(this.options.pagination.selector, 'slider');
                    this.pagesIns ? this.paginationDom = this.pagesIns.wrapperDom : null;
                }
            } else if (selectorType === 'Object') {
                //new实例的变量
                this.pagesIns = this.options.pagination.selector;
                this.paginationDom = this.pagesIns.wrapperDom;
            }
        } else {
            this.paginationDom = axAddElem('ul', { class: 'ax-slider-pagination' }, '');
            //如果有导航节点则将分页节点加入
            this.navigationDom && this.prevDom ? this.prevDom.insertAdjacentElement('afterend', this.paginationDom) : this.targetDom.appendChild(this.paginationDom);
        }
        //添加子节点
        if (this.paginationDom) {
            this.options.pagination.type ? this.paginationDom.setAttribute('type', this.options.pagination.type) : null;
            //统一添加样式或属性
            this.options.pagination.className ? axClass.add(this.paginationDom, this.options.pagination.className) : null;

            if (this.paginationDom.children.length) {
                //如果已经有了子节点，则从中取节点
                this.bullets.forEach((k, i) => {
                    if (this.paginationDom.children[i]) {
                        k.dom = this.paginationDom.children[i];
                        if (i === this.activeIndex) {
                            this.setBulletActive(k);
                        };
                    }
                });
            } else {
                //如果是空，则创建子节点
                let len = this.bullets.length,
                    fragment = document.createDocumentFragment();
                //循环
                this.bullets.forEach((k, i) => {
                    if (i === this.activeIndex) {
                        this.setBulletActive(k);
                    };
                    let text;
                    //循环生成li内容
                    switch (this.options.pagination.type) {
                        case 'text':
                            text = k.target.label || this.options.pagination.data[i];
                            break;
                        case 'thumb':
                            text = `<img src="${k.target.poster || this.options.pagination.data[i]}"/>`;
                            break;
                        case 'index':
                            text = i + 1;
                            break;
                        default:
                            text = '';
                    }
                    if (this.options.pagination.callback) {
                        //有函数则优先执行函数
                        this.options.pagination.callback.call(this, { index: i, total: len, data: k, parent: this.paginationDom });
                    } else {
                        k.dom.innerHTML = text;
                    }
                    //加入到碎片对象中
                    fragment.appendChild(k.dom);
                });
                this.paginationDom.appendChild(fragment);
            }
            //点击触发跳转
            this.bullets.forEach((k, i) => {
                k.dom.onclick = () => {
                    this.setBulletActive(k);
                    this.scrollIns.scrollTo(k.target.dom, this.options.duration);
                    //取消自动播放，true为取消倒计时
                    this.stopPlay(true);
                }
            });
        }

    }
    setNavState() {
        if (this.slides.length === 0) {
            this.prevDom.setAttribute('disabled', '');
            this.nextDom.setAttribute('disabled', '');
            return false;
        }
        //对导航前后按钮设置disabled，如果是loop无限循环就不需要设置了
        if (!this.options.loop) {
            if (this.prevDom) {
                this.activeIndex === this.bullets[0].index ? this.prevDom.setAttribute('disabled', '') : this.prevDom.removeAttribute('disabled');
            }
            if (this.nextDom) {
                this.activeIndex >= this.bullets.at(-1).index ? this.nextDom.setAttribute('disabled', '') : this.nextDom.removeAttribute('disabled');
            }
            //取消倒计时
            //this.activeIndex === this.bullets.at(-1).index && this.countdownDom ? this.stopPlay(true) : null;
        }
    }
    fillNavigation() {
        if (!this.options.navigation.enable) {
            return false;
        }
        let addNav = (type) => {
            let dom = null,
                text;
            if (type === 'prev') {
                text = this.options.scroll.axis === 'x' ? 'left' : 'up';
            } else if (type === 'next') {
                text = this.options.scroll.axis === 'x' ? 'right' : 'down';
            } else {
                text = '';
            }

            if (this.targetDom.querySelector(`.ax-slider-${type}`)) {
                dom = this.targetDom.querySelector(`.ax-slider-${type}`);
            } else if (this.options.navigation[type] && this.options.navigation[type].selector) {
                //dom可能为null的
                dom = axIdToDom(this.options.navigation[type].selector);
            } else {
                dom = axAddElem('i', { class: `ax-slider-${type}` }, text);
                //如果有导航节点则prev和next节点加入
                this.navigationDom ? this.navigationDom.appendChild(dom) : this.targetDom.appendChild(dom);
            }
            if (dom) {
                //添加class
                this.options.navigation[type] && this.options.navigation[type].className ? axClass.add(dom, this.options.navigation[type].className) : null;
                //添加文本
                this.options.navigation[type] && this.options.navigation[type].text ? dom.innerHTML = this.options.navigation[type].text : null;
                //统一添加样式或属性
                this.options.navigation.className ? axClass.add(dom, this.options.navigation.className) : null;
                this.options.navigation.filled ? dom.setAttribute('filled', '') : null;
                //初始化设置禁用，loop无限循环则不需要设置了
                if (!this.options.loop) {
                    if ((type === 'prev' && this.activeIndex === 0) || (type === 'next' && this.activeIndex >= this.bullets.length - 1)) {
                        dom.setAttribute('disabled', '');
                    }
                }
            }
            return dom;
        }
        if (this.options.navigation.enable) {
            //添加向前节点
            this.prevDom = addNav('prev');
            //添加向后节点
            this.nextDom = addNav('next');
            //前一页操作
            if (this.prevDom) {
                this.prevDom.onclick = axThrottle(() => {
                    this.toPrev();
                    //取消自动播放，true为取消倒计时
                    this.stopPlay(true);
                }, this.options.throttle, true);
            }
            //后一页操作
            if (this.nextDom) {
                this.nextDom.onclick = axThrottle(() => {
                    this.toNext();
                    //取消自动播放，true为取消倒计时
                    this.stopPlay(true);
                }, this.options.throttle, true);

            }

        }
    }
    resetTotal(value) {
        if (!this.options.total.enable) {
            return false;
        }
        value = value || this.getTotal();
        //没有子元素了，则分子和分母都应该为0
        if (value.pages === 0) {
            this.numeratorDom.innerHTML = this.denominatorDom.innerHTML = 0;
        } else {
            this.numeratorDom.innerHTML = value.current + 1;
            this.denominatorDom.innerHTML = value.pages;
        }
    }
    fillTotal() {
        if (!this.options.total.enable) {
            return false;
        }
        let total = this.getTotal();
        this.totalDom = axAddElem('span', { class: 'ax-slider-total' }, `<i></i><s>/</s><u></u>`);
        this.numeratorDom = this.totalDom.querySelector('i');
        this.denominatorDom = this.totalDom.querySelector('u');
        //填充值
        this.resetTotal(total);
        this.options.total.placement ? this.totalDom.setAttribute('placement', this.options.total.placement) : null;
        this.targetDom.appendChild(this.totalDom);
    }
    getTotal() {
        return { snapped: this.activeIndex, slides: this.slides.length, current: this.bullets.findIndex(k => k.active), pages: this.bullets.length }
    }

    createTool(data) {
        //data={name:'',text:'',icon:'',action:''}
        let actionType = axType(data.action);
        //追加dom属性
        data.dom = axAddElem('a', !axIsEmpty(data.props) ? data.props : {}, data.text ? data.text : '');
        //追加图标
        data.icon ? data.dom.classList.add('ax-iconfont', data.icon) : null;
        //禁用
        data.disabled ? data.dom.setAttribute('inert', '') : null;
        //追加事件
        if (actionType === 'String') {
            //action是超链接
            data.dom.href = data.action;
        } else if (actionType === 'Function') {
            data.dom.onclick = () => {
                //参数1：当前滑块，参数2：按钮自身对象
                data.action.call(this, this.bullets.find(k => k.active), data);
                //追加active属性
                data.dom.setAttribute('active', '');
                this.options.tools.forEach(k => {
                    k.dom !== data.dom ? k.dom.removeAttribute('active') : null;
                });
            }
        }
        this.toolsDom.appendChild(data.dom);
    }
    fillTools() {
        if (!this.options.tools.length === 0) {
            return false;
        }
        //获得tools节点
        if (this.targetDom.querySelector('.ax-slider-tools')) {
            this.toolsDom = this.targetDom.querySelector('.ax-slider-tools');
        } else {
            this.toolsDom = axAddElem('span', { class: 'ax-slider-tools' });
            this.targetDom.appendChild(this.toolsDom);
        }
        //追加按钮
        this.options.tools.forEach(k => {
            this.createTool(k);
        });
    }
    toolDisabledToggle(name, disabled = true) {
        let btn = this.options.tools.find(k => k.name === name);
        if (btn) {
            if (disabled) {
                btn.disabled = true;
                btn.dom.setAttribute('inert', '');
            } else {
                Reflect.deleteProperty(btn, 'disabled');
                btn.dom.removeAttribute('inert', '');
            }
        }
    }
    toPrev(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let target = this.findOneSnap('prev');
        //到头了就不能再滚动了
        if (!this.options.loop && (this.scrollIns.snapped === this.bullets[0].target)) {
            return false;
        }
        //主动吸附并滚动过去
        this.scrollIns.setSnapped({
            target, callback: (item) => {
                callback && callback.call(this, this.bullets.find(k => k.target.dom === item));
            }
        });
    }
    toNext(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let target = this.findOneSnap('next');
        //到尾了就不能再滚动了
        if (!this.options.loop && (this.scrollIns.snapped === this.bullets.at(-1).target)) {
            return false;
        }
        //主动吸附并滚动过去
        this.scrollIns.setSnapped({
            target, callback: (item) => {
                callback && callback.call(this, this.bullets.find(k => k.target.dom === item));
            }
        });
    }
    toStart(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let item = this.bullets[0];
        if (item.active) {
            return false;
        }
        this.toBullet(item, {
            complete: (bullet) => {
                callback && callback.call(this, bullet);
            }
        });
    }
    toEnd(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let item = this.bullets.at(-1);
        if (item.active) {
            return false;
        }
        this.toBullet(item, {
            complete: (bullet) => {
                callback && callback.call(this, bullet);
            }
        });
    }
    findOneSnap(type = 'next') {
        let index = this.scrollIns.snappeds.findIndex(k => k === this.scrollIns.snapped),
            target = null;
        if (type === 'next') {
            //向后找到吸附元素
            index !== this.scrollIns.snappeds.length - 1 ? target = this.scrollIns.snappeds[index + 1] : null;
        } else if (type === 'prev') {
            //向前找到吸附元素
            index !== 0 ? target = this.scrollIns.snappeds[index - 1] : null;
        } else if (type === 'start') {
            //找到头
            target = this.scrollIns.snappeds[0];
        } else if (type === 'end') {
            //找到尾
            target = this.scrollIns.snappeds.at(-1);
        }
        return target;
    }
    toBullet(item, opt = { before: null, complete: null }) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        if (item !== 0 && axIsEmpty(item)) {
            return false;
        }
        let itemType = axType(item), bullet;
        if (itemType === 'Number') {
            //纠正
            bullet = (item >= this.bullets.length - 1) ? this.bullets.at(-1) : (item <= 0) ? this.bullets[0] : this.bullets[item];
        } else {
            bullet = item;
        }
        if (bullet === this.bullets.find(k => k.active)) {
            //当前是active则不再执行了
            return false;
        }
        //active的toggle
        this.setBulletActive(bullet);
        //执行前回调，监听事件放在scroll.snap中
        opt.before && opt.before.call(this, bullet);
        this.scrollIns.scrollTo(bullet.target.dom, this.options.duration, () => {
            //完成后回调，监听事件放在scroll.snapped中
            opt.complete && opt.complete.call(this, bullet);
        });
    }
    setBulletActive(item) {
        //没有启用分页或本身是active状态则不再执行
        if (item.active) {
            return false;
        }
        item.active = true;
        item.dom.setAttribute('active', '');
        //设置滑块active
        item.target.dom.setAttribute('snapped', '');
        item.target.snapped = true;
        this.activeIndex = item.index;
        //其他的清除active
        let other = this.bullets.find(k => k !== item && k.active);
        if (other) {
            Reflect.deleteProperty(other, 'active');
            other.dom.removeAttribute('active');
            //取消滑块active
            Reflect.deleteProperty(other.target, 'snapped');
            other.target.dom.removeAttribute('snapped');
        }
        //如果分页是slider实例则需要确保active节点始终出现在视野中
        if (this.pagesIns && this.pagesIns.options.slides > 0) {
            let index = this.pagesIns.activeIndex, bulletIndex = this.bullets.findIndex(k => k.active);
            if (this.pagesIns.activeIndex + this.pagesIns.options.slides <= bulletIndex) {
                index = bulletIndex - this.pagesIns.options.slides + 1;
            } else if (bulletIndex < this.pagesIns.activeIndex) {
                index = bulletIndex;
            }
            this.pagesIns.scrollIns.setSnapped({ target: this.pagesIns.slides[index].dom });
        }
    }
    setLoopPlay() {
        //如果不启用则不再执行了
        if (this.options.loop === false) {
            return false;
        }
        if (this.options.slides === 'auto') {
            //自由模式，窗口不确定滑块数量，所以头尾都复制一遍
            this.startClone = this.slides.map(k => k.dom.cloneNode(true));
            this.endClone = this.slides.map(k => k.dom.cloneNode(true));
            //获取头尾需要克隆的数量
            this.totalClone = this.slides.length;
        } else {
            //窗口显示固定数量的滑块
            //克隆头部滑块放到尾部，克隆尾部滑块放到头部
            this.startClone = this.slides.slice(0, this.options.slides).map(k => k.dom.cloneNode(true));
            this.endClone = this.slides.slice(this.slides.length - this.options.slides).map(k => k.dom.cloneNode(true));
            //获取头尾需要克隆的数量
            this.totalClone = this.options.slides;
        }
        //使用虚拟滚动的添加元素方法可保持当前状态
        this.scrollIns.add(this.endClone, { placement: 'start' });
        this.scrollIns.add(this.startClone, { placement: 'end' });
        //当滑到克隆的尾元素时候，立即跳到this.slides中对应的尾元素；滑到头元素同理，如此形成视觉上的无限循环
    }
    setCountdown() {
        if (this.countdownIns) {
            axIsVnode(this.countdownDom) ? this.targetDom.appendChild(this.countdownDom) : null;
            //先退回百分百状态，再倒计时到0
            this.countdownIns.done();
            this.countdownIns.set(0, this.options.autoplay.delay);
            //点击可播放和暂停
            this.countdownDom.onclick = () => {
                let state = this.countdownDom.getAttribute('state')
                if (state === 'pause') {
                    this.startPlay();
                } else if (state === 'play') {
                    this.stopPlay();
                }
            }
        }
    }
    setAutoPlay() {
        //如果不启用自动滑动则不再执行了
        if (!this.options.autoplay.enable) {
            return false;
        }
        //创建进度条
        if (this.options.autoplay.countdown) {
            this.countdownDom = axAddElem('div', { class: 'ax-progress' });
            this.countdownIns = new axProgress(this.countdownDom, {
                labelShow: false,
                value: 100,
                type: 'circle',
                curve: 'linear',
            });
            this.targetDom.appendChild(this.countdownDom);
        }
        //创建定时轮播
        this.startPlay();
        //触发操作停止，移出继续

    }
    startPlay(callback) {
        //如果已经销毁则不再执行,如果已经是播放状态则不再执行
        if (this.destroyed || !this.pause) {
            return this;
        }
        //如果不启用自动滑动则不再执行了
        if (!this.options.autoplay.enable) {
            return false;
        }
        //先清除旧定时
        this.interval ? clearInterval(this.interval) : null;
        //再设置定时
        this.interval = setInterval(this.delayScroll, this.options.autoplay.delay);
        //从百分百开始倒计时
        this.setCountdown();
        //更新状态
        this.countdownDom ? this.countdownDom.setAttribute('state', 'play') : null;
        this.pause = false;
        //监听回调
        this.options.onStartPlay && this.options.onStartPlay.call(this);
        this.handlers.hasOwnProperty('startPlay') ? this.emit('startPlay', '') : null;
        callback && callback.call(this);
        return this;
    }
    stopPlay(clear = false, callback) {
        //如果已经销毁则不再执行，如果已经是停止状态也不再执行
        if (this.destroyed || this.pause) {
            return this;
        }
        //如果不启用自动滑动则不再执行了
        if (!this.options.autoplay.enable) {
            return false;
        }

        //倒计时归零
        this.countdownIns ? this.countdownIns.set(0, 0) : null;
        clearInterval(this.interval);
        //从百分百开始倒计时
        clear && this.countdownDom ? this.countdownDom.remove() : null;
        //更新状态
        this.countdownDom ? this.countdownDom.setAttribute('state', 'pause') : null;
        this.pause = true;
        //监听回调 
        this.options.onStopPlay && this.options.onStopPlay.call(this);
        this.handlers.hasOwnProperty('stopPlay') ? this.emit('stopPlay', '') : null;
        callback && callback.call(this);
        return this;
    }
    add(data, placement = 'end', callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let dataType = axType(data),
            obj = dataType.includes('HTML') ? { type: 'node', content: data } : data;

        //根据data数据创建dom节点
        this.createSliderDom(obj);
        //设置尺寸
        this.setSingleSize(obj);
        this.scrollIns.add(obj.dom, {
            placement, callback: (doms, target) => {
                //加入到this.slides中
                if (placement === 'start') {
                    this.slidesObserve.proxy.unshift(obj);
                } else if (placement === 'end') {
                    this.slidesObserve.proxy.push(obj);
                } else {
                    let index = this.slides.findIndex(k => k.dom === target);
                    //插入
                    this.slidesObserve.proxy.splice(index + 1, 0, obj);
                }

            }
        });
        //监听回调
        this.handlers.hasOwnProperty('add') ? this.emit('add', obj) : null;
        this.options.onAdd && this.options.onAdd.call(this, obj);
        callback && callback.call(this, obj);
    }
    remove(placement = 'end', callback) {
        //如果已经销毁则不再执行，没有子元素了也不执行
        if (this.destroyed || this.slides.length === 0) {
            return this;
        }
        //可能placement直接传入是slide对象
        let raw = axType(placement) === 'Object' ? placement : null;
        this.scrollIns.remove(raw ? raw.dom : placement, (target) => {
            let item = raw ? raw : this.slides.find(k => k.dom === target),
                index = this.slides.findIndex(k => k.dom === target);
            this.slidesObserve.proxy.splice(index, 1);
            //监听回调
            this.handlers.hasOwnProperty('remove') ? this.emit('remove', item) : null;
            this.options.onRemove && this.options.onRemove.call(this, item);
            callback && callback.call(this, item);
        });
    }
    save(props, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }

        if (!props) {
            //保存当前焦点索引
            axLocalStorage.set(this.options.storageName, axExtend(this.options, { active: this.activeIndex }), { cover: false });
        } else {
            //保存参数,props={a:'',b:''}
            props.active = this.activeIndex;
            axLocalStorage.set(this.options.storageName, axExtend(this.options, props), { cover: false });
        }
        //回调监听
        let localValue = axLocalStorage.get(this.options.storageName);
        'save' in this.handlers ? this.emit('save', localValue) : null;
        this.options.onSave && this.options.onSave.call(this, localValue);
        callback && callback.call(this, localValue);
        return this;
    }
    destroy(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //取消笼子相关事件
        this.slides.forEach(k => {
            if (k.cage) {
                k.cage.destroy();
                k.dom.ondblclick = null;
            }
        });
        //相关实例均销毁
        this.scrollIns.destroy();
        this.spyIns ? this.spyIns.destroy() : null;
        this.lazyIns ? this.lazyIns.destroy() : null;
        this.countdownIns ? this.countdownIns.destroy() : null;
        window.removeEventListener('keydown', this.keyToSwipe);
        //取消点击事件
        this.bullets.forEach((k) => {
            k.dom.onclick = null;
        });
        this.prevDom ? this.prevDom.onclick = null : null;
        this.nextDom ? this.nextDom.onclick = null : null;
        if (this.toolsDom) {
            this.options.tools.forEach(k => {
                k.dom.onclick = null;
            });
        }
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.options = axExtend(this.options, setting);
        //保存数据
        this.save();

        this.init();
        //监听回调
        this.options.onUpdated && this.options.onUpdated.call(this, setting);
        this.handlers.hasOwnProperty('updated') ? this.emit('updated', setting) : null;
        callback && callback.call(this, setting);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
 * 设置多久后过期时间，返回过期时间（国标时间），
 */
const axExpires = (type = 'd', expires = 365) => {
    let dateExp;
    if (axIsDate(expires)) {
        //如果是日期格式，例如2021-2-6
        dateExp = new Date(expires).toUTCString()
    } else {
        let now = new Date();
        //支持按日、时、分、秒设置过期时间，使用~~按位符转为数字
        if (type == 'second' || type == 's') {
            //是秒为单位
            now.setSeconds(now.getSeconds() + ~~expires);
            //cookie的过期时间是按国标时间计算的，所以需要使用toUTCString()而不是toLocaleString()
            dateExp = now.toUTCString();
        } else if (type == 'minute' || type == 'm') {
            //以分为单位
            now.setMinutes(now.getMinutes() + ~~expires);
            dateExp = now.toUTCString();
        } else if (type == 'hour' || type == 'h') {
            //以时为单位
            now.setHours(now.getHours() + ~~expires);
            dateExp = now.toUTCString();
        } else {
            //默认以日为单位
            now.setDate(now.getDate() + ~~expires);
            dateExp = now.toUTCString();
        }
    }
    return dateExp;
}
/*!
 * 定义cookie方法
 * 支持axCookie.set、axCookie.get、axCookie.del三个方法
 * set方法支持options参数，支持追加path、domain、secure和sameSite属性
 * 默认过期时间是365天
 */
const axCookie = {
    set: function (name, value, options = { type: 'd', expires: 365 }) {
        if (!name) {
            console.error('The name of cookie can not be empty!');
            return false;
        }
        //设置cookie
        let opt = Object.assign({}, options),
            //除了type，expires之外，还支持其他的参数：path、domain、secure和sameSite
            string = opt.path ? ';path=' + opt.path : '';
        string += opt.domain ? ';domain=' + opt.domain : '';
        string += opt.secure ? ';secure=' + opt.secure : '';
        string += opt.sameSite ? ';sameSite=' + opt.sameSite : '';
        opt.expires = axExpires(options.type, options.expires);
        //axCookie.set('name',null)或axCookie.set('name','',{expires:-1})，并且cookie中有该键值对，那么删掉
        if ((options.expires < 0 || value === null) && this.get(name)) {
            this.del(name);
        } else {
            //字符串编码
            document.cookie = name + "=" + encodeURIComponent(value) + ";expires=" + opt.expires + string;
        }
        return this;
    },
    get: function (name) {
        //获取cookie
        //先去掉“;”后的空格，再以“;”划分为数组
        let cookieArr = document.cookie.replace(/;\s*/, ';').split(';');
        let value = '';
        for (let i = 0; i < cookieArr.length; i++) {
            let k = cookieArr[i].split('=')[0],
                v = cookieArr[i].split('=')[1];
            if (k == name) {
                value = v;
                break;
            }
        }
        if (value) {
            //字符串解码
            return decodeURIComponent(value);
        } else {
            return false;
        }
    },
    remove: function (name) {
        //删除cookie
        document.cookie = name + "=;expires=Thu, 01 Jan 1970 00:00:00 GMT";
        return this;
    }
}
/*!
 * 定义localStorage本地存储
 * 支持set、get、del、clear、key五个方法
 * set方法支持options参数，type和expires
 * 默认过期时间是0，即永不过期
 */
const axLocalStorage = {
    set: function (key, value = '', options = { type: 'd', expires: 0, cover: true }) {
        //为空则不再执行
        if (typeof key !== 'string' && !key) {
            return false;
        }
        let newValue = axClone(value),
            type = axType(value),
            filterProps = (obj) => {
                Object.keys(obj).forEach(k => {
                    let val = obj[k];
                    (axType(val) === 'Array' || axType(val) === 'Object') && filterProps(val);
                    //将DOM节点和函数键值对删掉
                    (axType(val).includes('HTML') || axType(val) === 'Function') && delete obj[k];
                });
                return obj;
            };
        //localStorage无法存储DOM节点和函数
        if ((type === 'Array' || type === 'Object') && !axIsEmpty(newValue)) {
            //返回去掉DOM和函数的value
            filterProps(newValue);
        }
        let expires = options.expires !== 0 ? axExpires(options.type, options.expires) : 0,
            oldValue = axLocalStorage.get(key),
            valueExp = {
                //如果已经有保存过并且不覆盖原有值，则使用extend方法
                data: (!options.cover && axLocalStorage.get(key)) ? axExtend(oldValue, newValue) : newValue,
                expires: expires,
            };
        //axLocalStorage.set('name',null)或axLocalStorage.set('name','',{expires:-1})，并且LocalStorage中有该键值对，那么删掉
        if ((options.expires < 0 || newValue === null) && this.get(key)) {
            this.del(key);
        } else {
            window.localStorage.setItem(key, JSON.stringify(valueExp));
        }
        return this;
    },
    get: function (key) {
        //为空则不再执行
        if (axIsEmpty(key)) {
            return false;
        }
        let now = new Date().toUTCString(),
            valueExp = JSON.parse(window.localStorage.getItem(key));
        if (valueExp) {
            //如果有键值对
            if (valueExp.expires === 0 || now <= valueExp.expires) {
                //如果为0，或当前时间还在过期时间以前则正常获取
                return valueExp.data;
            } else {
                //如果过期了，则删除并返回null
                window.localStorage.removeItem(key);
                return null;
            }
        } else {
            //如果没有键值对，则返回null
            return null;
        }
    },
    remove: function (key) {
        //为空则不再执行
        if (axIsEmpty(key)) {
            return false;
        }
        let arr = [];
        if (Array.isArray(key)) {
            //是数组：['','','']
            arr = key;
        } else {
            //是字符串
            arr.push(key);
        }
        arr.forEach(k => {
            window.localStorage.removeItem(k);
        });
        return this;
    },
    key: function (num) {
        //num必须是数字
        if (typeof num !== 'number') {
            return false;
        }
        return window.localStorage.key(num);
    },
    clear: function () {
        window.localStorage.clear();
        return this;
    }
}
/*!
* 插件：告示；使用方法：new axAlert('#id',{参数})
*/
class axAlert {
    constructor(elem, options) {
        if (!elem) {
            return false;
        }
        this.targetDom = axIdToDom(elem);
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            caption: '',//告示标题，html或纯文本
            content: '',//告示内容，html或纯文本，可使用p和br标签
            theme: 'warning',//主题风格，可填primary、secondary、success、danger、warning、ad、info或light，默认warning主题
            borderless: false,//是否显示边框，默认false显示，可选择true不显示
            square: false,//是否是方角，默认false非方角即圆角，可选择true方角
            className: '',//其他样式类，多个用英文逗号或空格分开
            closable: false,//是否出现关闭按钮并可关闭，默认false不可关闭，可选择true可关闭
            style: 'simple',//alert的结构风格，默认simple风格即只有一个DIV包裹；可选择complex风格，可实现带图标多行告示
            icon: '',//图标class类，比如可填ax-icon-warning-o-f
            iconShow: true,//是否显示图标，默认true显示，可选择false显示，当然如果icon参数为空也不会显示图标
            notable: false,//是否凸显图标，默认false，可选择true凸显（大图标）
            display: '',//显示的方式，相对display:none，可以填block、flex、inline-flex、inline或inline-block，如果不填将按容器自身的display方式显示
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onShow: '',//回调函数，开始显示后执行，无参数
            onHide: '',//回调函数，开始关闭后执行，无参数
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onUpdateContent: '',//回调函数，内容更新后执行，无参数
            onClear: '',//回调函数，清空后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //初始化
        this.init();
    }
    init() {
        //默认显示状态
        this.shown = true;
         //添加到实例合集
         axInstance.push(this, this.options.insName, 'alert');
         //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //更新销毁状态
        this.destroyed = false;
        //创建dom节点
        this.render();
        //设置风格
        this.setAttribute();
        //点击关闭
        if (this.close) {
            this.close.onclick = () => {
                this.hide();
            }
        }
        //监听回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    render() {
        if (!this.targetDom.innerHTML.trim()) {
            //如果没有内容，则创建节点
            let html = '';
            if (this.options.style == 'simple' || !this.options.style) {
                //简单内容，simple类型下caption无效
                //取得内容
                this.content = this.targetDom.innerHTML.trim() ? this.targetDom.innerHTML : this.options.content;
                //获得html
                if (this.options.icon && this.options.iconShow) {
                    html = `<i class="ax-iconfont ${this.options.icon}" icon></i><span content>${this.content}</span>`;
                } else {
                    html = `<span content>${this.content}</span>${this.options.closable ? '<i close></i>' : ''}`;
                }
            } else if (this.options.style == 'complex') {
                //复杂内容
                if (!this.targetDom.innerHTML.trim()) {
                    html = `${this.options.icon && this.options.iconShow ? '<i class="ax-iconfont ' + this.options.icon + '" icon></i>' : ''}
                    <div wrapper>
                        ${this.options.caption ? '<div caption>' + this.options.caption + '</div>' : ''}
                        <div content>${this.options.content}</div>
                    </div>
                    ${this.options.closable ? '<i close></i>' : ''}
                    `;
                }
            }
            //赋值内容
            this.targetDom.innerHTML = html;
        }
        //定义节点变量
        this.close = this.targetDom.querySelector('[close]');
        this.icon = this.targetDom.querySelector('[icon]');
        this.wrapper = this.targetDom.querySelector('[wrapper]');
        this.caption = this.targetDom.querySelector('[caption]');
        this.content = this.targetDom.querySelector('[content]');
    }
    setAttribute() {
        //添加class类
        !this.targetDom.classList.contains('ax-alert') ? this.targetDom.classList.add('ax-alert') : null;
        axClass.add(this.targetDom, this.options.className);
        //添加主题属性
        this.options.theme ? this.targetDom.setAttribute('theme', this.options.theme) : null;
        //添加边框属性
        this.options.borderless ? this.targetDom.setAttribute('borderless', '') : null;
        //添加尺寸属性
        this.options.notable && this.options.style === 'complex' ? this.targetDom.setAttribute('notable', '') : null;
        //添加方形属性
        this.options.square ? this.targetDom.setAttribute('square', '') : null;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //重新初始化
        this.options = axExtend(this.options, setting);
        //清除属性并清空内容
        this.targetDom.removeAttribute('class');
        this.targetDom.removeAttribute('theme');
        this.targetDom.removeAttribute('borderless');
        this.targetDom.removeAttribute('square');
        this.targetDom.removeAttribute('notable');
        this.targetDom.innerHTML = '';
        this.init();
        //监听回调
        callback && callback.call(this);
        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        return this;
    }
    updateContent(html, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //更新内容
        this.content.innerHTML = html;
        //监听回调
        callback && callback.call(this);
        this.options.onUpdateContent && this.options.onUpdateContent.call(this);
        'updateContent' in this.handlers ? this.emit('updateContent', '') : null;
        return this;
    }
    show(callback) {
        //如果已经销毁或本身就是打开状态则不再执行
        if (this.destroyed || this.shown) {
            return this;
        }
        //如果没有设置display参数，使用自身的display属性，那么去掉display: none;即可
        if (this.options.display) {
            this.targetDom.style.display = this.options.display;
        } else {
            this.targetDom.style.cssText = this.targetDom.style.cssText.replace('display: none;', '');
        }
        //更新显示状态
        this.shown = true;
        //回调
        this.options.onShow && this.options.onShow.call(this);
        'show' in this.handlers ? this.emit('show', '') : null;
        callback && callback.call(this);
        return this;
    }
    hide(callback) {
        //如果已经销毁，或者本身就是关闭状态则不再执行
        if (this.destroyed || !this.shown) {
            return this;
        }
        //删除节点
        this.targetDom.style.display = 'none';
        //更新显示状态
        this.shown = false;
        //回调
        this.options.onHide && this.options.onHide.call(this);
        'hide' in this.handlers ? this.emit('hide', '') : null;
        callback && callback.call(this);
        return this;
    }
    clear(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //清除属性并清空内容
        this.targetDom.removeAttribute('class');
        this.targetDom.removeAttribute('theme');
        this.targetDom.removeAttribute('borderless');
        this.targetDom.removeAttribute('square');
        this.targetDom.removeAttribute('notable');
        this.targetDom.innerHTML = '';
        //监听回调
        callback && callback.call(this);
        this.options.onClear && this.options.onClear.call(this);
        'clear' in this.handlers ? this.emit('clear', '') : null;
        return this;
    }
    destroy(callback) {
        //重新赋予属性和内容
        this.close ? this.close.onclick = null : null;
        this.destroyed = true;
        //监听回调
        this.options.onDestroy && this.options.onDestroy.call(this);
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/* (() => {
    //父节点下多图
    document.querySelectorAll('[axAlert]').forEach(element => {
        new axAlert(element);
    });
})(); */
/*!
* 插件：Tags标签；使用方法：new axTags('#id',{参数})
*/
class axTags {
    constructor(elem, options) {
        //必须有节点
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) {
            return false;
        }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            content: '',//数据源，支持四种格式，包括：一个字符串/对象，一个字符串数组/一个对象数组
            theme: '',//主题风格，可填primary、secondary、success、danger、warning、ad、info或light，默认为空
            borderless: false,//是否显示边框，默认false显示，可选择true不显示
            compact: false,//是否使用紧凑风格，默认false不使用，可选择true使用
            text: false,//是否使用纯文本风格，默认false不使用，可选择true使用
            corner: '',//设置单个Tag的圆角，可填round则是圆角，square则是方角，默认为空即是radiu:3px的小圆角
            className: '',//其他样式类，多个用空格分开
            removeNode: 'i',//关闭按钮的节点名，默认是i，可选择s
            removable: false,//是否出现关闭按钮并可关闭，默认false不可关闭，可选择true可关闭
            separator: ',',//字符串数据分隔符，默认为英文逗号
            unique: false,//是否要求添加同唯一值的tag，默认fase不要求，可选择true要求
            emptyShow: true,//是否显示空状态，默认true显示，可选择false不显示
            emptyTip: '<i class="ax-color-ignore">还没有创建标签！</i>',//空提示
            edit: {
                enable: false,//是否显示行内input，默认false不显示，可选择true显示
                addable: true,//是否可通过回车新增tag，默认true允许，可选择false不允许
                deletable: true,//是否可通过del键删除tag，默认true允许，可选择false不允许
                input: '',//默认行内编辑，如果需要外部编辑则需要设置该参数，可以是#id、.className、[name]或输入表单节点，通常是input或textarea
                placeholder: '输入关键字...',//input的提示文字
            },
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onBeforeAdd: '',//使用input添加之前执行，支持一个参数即添加label文本值
            onAdd: '',//回调函数，添加后执行，支持一个参数即增加的对象
            onBeforeRemove: '',//使用input删除之前执行，支持一个参数即删除的项
            onRemove: '',//回调函数，删除后执行，支持一个参数即删除的对象
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onUClear: '',//回调函数，清零新后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //当前最大索引值
        this.maxIndex = 0;
        //定义空提示
        this.emptyDom = axAddElem('span', { emptyTip: '' }, this.options.emptyTip);
        //input编辑tags
        this.inputEdit = (e) => {
            let value = this.inputDom.value,
                selected = this.content.find(k => k.dom.hasAttribute('selected'));
            //修改行内input的width，让input自动变长
            !this.options.edit.input ? this.inputDom.style.width = value.length + 'em' : null;
            if (value) {
                if (this.options.edit.addable && e.code === 'Enter') {
                    //添加一个Tag
                    if (axIsEmpty(this.options.onBeforeAdd)) {
                        this.add(this.inputDom.value.trim());
                        //添加完毕清除input值
                        this.inputDom.value = '';
                    } else {
                        this.options.onBeforeAdd.call(this, this.inputDom.value.trim());
                    }
                }
                selected ? selected.dom.removeAttribute('selected') : null;
            } else {
                if (this.options.edit.deletable && e.code === 'Backspace' && this.content.length > 0) {
                    let end = this.content.slice(-1)[0];
                    //第一次删除是选择Tag，第二次删除才是真正删除
                    if (end.dom.hasAttribute('selected')) {
                        if (axIsEmpty(this.options.onBeforeRemove)) {
                            this.remove(end.dom);
                        } else {
                            this.options.onBeforeRemove.call(this, this.content.find(k => k.dom == end.dom));
                        }
                        /* 
                                                end.dom.remove();
                                                this.content = this.content.slice(0, -1); */
                    } else {
                        end.dom.setAttribute('selected', '');
                    }
                    this.toggleEmpty();
                } else {
                    selected ? selected.dom.removeAttribute('selected') : null;
                }
            }
        }
        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'tags');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //更新销毁状态
        this.destroyed = false;
        //统一设置数据格式，获得this.content变量
        this.content = [];
        this.contentProxy = new axObserve(this.content, {
            deep: true,
        });
        this.contentProxy.proxy.push(...this.getSource(this.options.content));
        //创建dom节点
        this.render(this.content);
        //绑定input
        this.observeInput();
        //设置风格
        this.setAttribute();
        //
        this.renderFinish();
        //监听回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    renderFinish() {
        this.content.forEach(k => {
            k.dom.onclick = () => {
                k.dom.setAttribute('selected', '');
                this.content.filter(i => i !== k).forEach(i => {
                    i.dom.removeAttribute('selected');
                });
            }
        });
    }
    getSource(obj) {
        let result = [];

        if (!axIsEmpty(obj)) {
            if (Array.isArray(obj)) {
                //是数组
                if (typeof obj[0] === 'string') {
                    //['a','b']
                    result = obj.map(k => {
                        return { value: k, theme: '' };
                    });
                } else {
                    //[{value:'',theme:''}]
                    result = obj;
                }
            } else if (typeof obj === 'object') {
                //是对象
                result = [obj];
            } else if (typeof obj === 'string') {
                //是字符串
                let arr = obj.trim().split(this.options.separator);
                result = arr.map(k => {
                    return { value: k, theme: '' };
                });
            }
        }

        //清除为false的项
        result = result.filter(Boolean);
        //增加索引
        result.forEach((k, i) => {
            !k.hasOwnProperty('id') ? k.id = this.maxIndex + i : null;
            !k.hasOwnProperty('theme') ? k.theme = '' : null;
        });
        //更新最大索引
        this.maxIndex += result.length;
        return result;
    }
    render(content) {
        //归零切换
        this.toggleEmpty();

        if (axIsEmpty(content)) {
            return false;
        }
        let fragment = document.createDocumentFragment();
        content.forEach(i => {
            let tagDom = this.createTag(i);
            //追加dom属性
            i.dom = tagDom;
            fragment.appendChild(tagDom);
        });
        this.targetDom.appendChild(fragment);
        //行内input
        if (this.inputDom && !this.options.edit.input) {
            this.targetDom.appendChild(this.inputDom);
            this.inputDom.focus();
        }
    }
    createTag(obj) {
        let tagDom = axAddElem('span', { class: 'ax-tag' }, obj.value),
            tagClose = axAddElem(this.options.removeNode, { remove: '' });
        //增加主题颜色
        if (!this.options.theme && obj.theme) {
            tagDom.setAttribute('theme', obj.theme)
        }
        //显示remove按钮
        if (this.options.removable) {
            tagDom.appendChild(tagClose);
            tagClose.onclick = () => {
                if (axIsEmpty(this.options.onBeforeRemove)) {
                    this.remove(tagDom);
                } else {
                    //使用onBeforeRemove劫持，自定义remove
                    this.options.onBeforeRemove.call(this, this.content.find(k => k.dom == tagDom));
                }
            }
        }
        //更新this.content = [{value:'',theme:'',dom:''}]
        obj.dom = tagDom;
        return tagDom;
    }
    getTag(obj) {
        //为空则不再执行
        if (axIsEmpty(obj)) {
            return false;
        }
        let filter, tag,
            type = axType(obj);
        if (type === 'Object') {
            //单个对象，支持属性id、value和dom
            let str = '';
            obj.id || obj.id == 0 ? str += 'item.id == obj.id ' : null;
            obj.value ? str += '&& item.value == obj.value ' : null;
            obj.dom ? str += '&& item.dom == obj.dom' : null;
            //如果没有id属性，那么str将以&&开头，会报错，所以在前方加上true
            str.startsWith('&&') ? str = 'true ' + str : null;
            filter = new Function('data', 'obj', `return data.filter(item =>${str})`)(this.content, obj);
            filter.length > 0 ? tag = filter[0] : null;
        } else if (type.includes('HTML')) {
            //单个节点
            tag = this.content.find(k => k.dom === obj);
        } else if (type === 'String') {
            //单个字符串
            tag = this.content.find(k => k.value === obj);
        } else if (type === 'Number') {
            //单个数字
            tag = this.content.find(k => k.id == obj);
        }
        return { item: tag, id: this.content.findIndex(k => k === tag) };
    }
    setAttribute() {
        //添加class类
        !this.targetDom.classList.contains('ax-tags') ? this.targetDom.classList.add('ax-tags') : null;
        axClass.add(this.targetDom, this.options.className);
        //添加主题属性
        this.options.theme ? this.targetDom.setAttribute('theme', this.options.theme) : null;
        //添加边框属性
        this.options.borderless ? this.targetDom.setAttribute('borderless', '') : null;
        //添加尺寸属性
        this.options.compact ? this.targetDom.setAttribute('compact', '') : null;
        //添加纯文本属性
        this.options.text ? this.targetDom.setAttribute('text', '') : null;
        //添加圆角属性
        this.options.corner ? this.targetDom.setAttribute('corner', this.options.corner) : null;
    }
    observeInput() {
        //如果setting为空或已经销毁则不再执行
        if (!this.options.edit.enable || this.destroyed) {
            return false;
        }
        if (!this.options.edit.input) {
            //行内编辑，内置控件
            this.inputDom = axAddElem('input', { type: 'text', placeholder: this.options.edit.placeholder });
            this.targetDom.appendChild(this.inputDom);
        } else {
            //外部编辑，从input获得编辑控件
            this.inputDom = axIdToDom(this.options.edit.input);
            !this.inputDom.hasAttribute('placeholder') ? this.inputDom.setAttribute('placeholder', this.options.edit.placeholder) : null;
        }
        this.inputDom ? this.inputDom.addEventListener('keyup', this.inputEdit) : null;
    }
    toggleEmpty() {
        if (!this.options.emptyShow) {
            return false;
        }
        if (this.content.length === 0) {
            //并初始化maxIndex
            this.targetDom.insertAdjacentElement('afterbegin', this.emptyDom)
            this.maxIndex = 0;
        } else {
            this.emptyDom.remove();
        }
    }
    update(setting, callback) {
        //如果setting为空或已经销毁则不再执行
        if (axIsEmpty(setting) || this.destroyed) {
            return this;
        }
        //重新初始化
        this.options = axExtend(this.options, setting);
        //清空内容再次初始化
        this.targetDom.innerHTML = '';
        //如果行内编辑则首先追加inputDom
        this.inputDom && !this.options.edit.input ? this.targetDom.appendChild(this.inputDom) : null;
        this.init();
        //监听回调
        callback && callback.call(this);
        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        return this;
    }
    add(obj, callback) {
        //如果obj为空或已经销毁则不再执行
        if (axIsEmpty(obj) || this.destroyed) {
            return this;
        }
        //转成统一格式数据
        let arr = this.getSource(obj),
            oldValues = this.content.map(k => k.value);


        //如果禁止添加相同value的tag，则对arr筛选
        if (this.options.unique) {
            //先去重
            arr = axUnique(arr, 'value');
            for (let i = 0; i < arr.length; i++) {
                if (oldValues.includes(arr[i].value)) {
                    //如果已经有了，删除自己
                    arr.splice(i--, 1);
                }
            }
            //为0说明有重复的
            if (arr.length === 0) {
                new axMessage({
                    content: '不可添加重复的标签!',
                    result: 'warning',
                    iconShow: true,
                }).show();
            }
        }

        if (arr.length === 0) {
            return false;
        }

        //更新this.content
        this.contentProxy.proxy.push(...arr);
        //渲染和追加节点
        this.render(arr);
        //
        this.renderFinish();
        //回调
        this.options.onAdd && this.options.onAdd.call(this, arr);
        'add' in this.handlers ? this.emit('add', arr) : null;
        callback && callback.call(this, arr);
        return this;
    }
    remove(obj, callback) {
        //如果obj为空或已经销毁则不再执行
        if (axIsEmpty(obj) || this.destroyed) {
            return this;
        }
        let tags = [],
            ids = [],
            type = axType(obj);
        if (type === 'Array' && obj.length > 0) {
            //数组批量删除
            obj.forEach(k => {
                let tagObj = this.getTag(k);
                if (tagObj.item) {
                    tags.push(tagObj.item)
                    ids.push(tagObj.id);
                }
            });
        } else if (type === 'String') {
            //带分隔符字符串
            let arr = obj.split(this.options.separator);
            tags = this.content.filter(k => arr.includes(k.value));
            ids = tags.map(k => this.content.findIndex(i => i === k));
        } else {
            let tagObj = this.getTag(obj);
            if (tagObj.item) {
                tags.push(tagObj.item)
                ids.push(tagObj.id);
            }
        }

        //删除物理节点
        tags.forEach(k => {
            k.dom.remove();
        });
        //更新this.content
        let doms = tags.map(k => {
            return k.dom;
        });
        //删除自己
        this.contentProxy.proxy.forEach((k, i) => {
            doms.includes(k.dom) ? this.contentProxy.proxy.splice(i--, 1) : null;
        });
        //this.content = this.content.filter(k => !doms.includes(k.dom));
        //归零
        this.toggleEmpty();
        //回调
        this.options.onRemove && this.options.onRemove.call(this, tags, ids);
        'remove' in this.handlers ? this.emit('remove', tags, ids) : null;
        callback && callback.call(this, tags, ids);
        return this;
    }
    clear(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.contentProxy.proxy.splice(0);
        this.maxIndex = 0;
        this.targetDom.innerHTML = '';
        //如果行内编辑则首先追加inputDom
        this.inputDom && !this.options.edit.input ? this.targetDom.appendChild(this.inputDom) : null;
        //归零
        this.toggleEmpty();
        //回调
        this.options.onClear && this.options.onClear.call(this);
        'clear' in this.handlers ? this.emit('clear', '') : null;
        callback && callback.call(this);
        return this;
    }
    destroy(callback) {
        //取消remove点击事件
        this.content.forEach(K => {
            let remove = K.dom.querySelector('[remove]');
            remove ? remove.onclick = null : null;
        });
        //取消input绑定状态
        this.inputDom ? this.inputDom.removeEventListener('keyup', this.inputEdit, false) : null;
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.options.onDestroy && this.options.onDestroy.call(this);
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/* (() => {
    document.querySelectorAll('[axTags]').forEach(element => {
        new axTags(element);
    });
})();  */
/*!
* 插件：定义效果更好Hover操作；使用方法：new axHover('#id',{参数})
*/
class axHover {
    constructor(targetDom, options) {
        // 始终将targetDom转成DOM对象
        this.targetDom = axIdToDom(targetDom);
        // 定义参数
        this.options = axExtend({
            timeout: 50,//鼠标离开按钮后多久事件恢复初始状态，单位毫秒
            interval: 50,//在按钮上移动是检查光标坐标的间隔时间，单位毫秒。该值越小越容易触发，越大越不容易触发；因为检查间隔时间长了，光标滑动的距离自然就长了，如果大于了threshold就不会触发的。
            threshold: 7,//触发阈值，在按钮上interval时间内移动直线距离小于该值才会触发enter事件。该值越大越容易触发，越小越不容易触发；因为阈值大，所以比较容易让移动距离小于阈值，也就容易触发。
            enter: () => { },//移入按钮上的执行事件
            leave: () => { },//移出按钮后的执行事件
            hold: '',//悬停不隐藏的窗口元素
        }, options);
 
        //定义当前在按钮内的光标坐标
        this.curX;
        this.curY;
        //定义当前在按钮内的上一次坐标
        this.preX;
        this.preY;
        //定义enter和leave事件的定时器
        this.enterTimer;
        this.leaveTimer;
        //enter事件默认未触发，未激活
        this.isActive = false;
        //定义悬停元素
        this.options.hold ? this.hold = axIdToDom(this.options.hold) : this.hold = '';
        //toggle事件
        this.event = (type, elem, event) => {
            if (type === 'enter') {
                this.options.enter.call(this, elem, event);
                this.isActive = true;
            } else if (type === 'leave') {
                this.options.leave.call(this, elem, event);
                this.isActive = false;
            }
        }
        //进入按钮后每隔interval事件检查一次光标坐标，对比上一次的光标坐标，以判断是否触发enter事件
        this.enterCompare = (elem, e) => {
            let distX = this.preX - this.curX,
                distY = this.preY - this.curY,
                distance = Math.sqrt(distX ** 2 + distY ** 2);

            if (distance < this.options.threshold) {
                //触发enter事件,避免重复执行ennter事件，当为非激活状态才执行enter事件
                if(!this.isActive){
                    this.event('enter', elem, e);
                }
            } else {
                //既然没有触发，那么将当前坐标更新为上一个坐标
                this.preX = this.curX;
                this.preY = this.curY;
                //继续按interval间隔时间进行坐标比较，直到触发enter事件
                this.enterTimer = setTimeout(() => {
                    this.enterCompare(elem, e);
                }, this.options.interval);
            }
        };

        //初始化
        this.init();
    }
    init() {
        let _this = this;
        //鼠标进入按钮后跟踪鼠标位置，更新光标坐标
        this.targetDom.addEventListener('mousemove', (e) => {
            this.curX = e.pageX;
            this.curY = e.pageY;
        });
        this.targetDom.addEventListener('mouseenter', function (e) {
            //进入按钮则清除离开定时器和leave事件
            clearTimeout(_this.leaveTimer);
            //设置刚进入按钮时光标坐标
            _this.preX = e.pageX;
            _this.preY = e.pageY;
            //如果移入按钮上，每隔一段事件检查移动距离
            _this.enterTimer = setTimeout(() => {
                _this.enterCompare(this, e);
            }, _this.options.interval);
        });

        this.targetDom.addEventListener('mouseleave', function (e) {
            //进入按钮则清除进入定时器和enter事件
            clearTimeout(_this.enterTimer);
            //设置leave事件，延迟执行
            _this.leaveTimer = setTimeout(() => {
                _this.event('leave', this, e);
            }, _this.options.timeout);
        });

        //有hold项
        if (this.hold) {
            _this.hold.onmouseenter = function (e) {
                clearTimeout(_this.leaveTimer);
            }
            _this.hold.onmouseleave = function (e) {
                //离开则清除enter事件定时器
                clearTimeout(_this.enterTimer);
                //延时执行leave事件
                _this.leaveTimer = setTimeout(() => {
                    _this.event('leave', _this.targetDom, e);
                }, _this.options.timeout);
            }
        }

    };
}

/*!
 * 插件：查看更多；使用方法：new axMore('#id',{参数})
 */
const axMores = [];
class axMore {
    constructor(targetDom, options) {
        // 始终将targetDom转成DOM对象
        this.targetDom = axIdToDom(targetDom);
        // 定义参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            length: 60,//文字数量，默认60个字
            textFold: '展开',//按钮展开状态文字
            textUnfold: '收起',//按钮折叠状态文字
            classFold: 'ax-color-primary',//按钮折叠状态className
            classUnfold: 'ax-color-primary',//按钮展开状态className
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
        }, options, this.targetDom,'axMore');
        // 定义监听事件成员
        this.handlers = {};
        // 定义原内容
        this.html = this.targetDom.innerHTML;
        // 去掉收尾空格+删除所有标签+将多个空格转成1个空格
        this.pureHtml = this.targetDom.textContent;
        let _this = this,
            more = true;
        this.event = function() {
            if (more) {
                _this.targetDom.removeChild(this.previousSibling);
                this.innerHTML = _this.options.textUnfold;
                this.setAttribute('class', _this.options.classUnfold);
                this.insertAdjacentHTML('beforeBegin', _this.text);
                more = false;
                // 执行展开事件
                'unfold' in _this.handlers ? _this.emit('unfold', '') : null;
            } else {
                _this.targetDom.removeChild(this.previousSibling);
                this.innerHTML = _this.options.textFold;
                this.setAttribute('class', _this.options.classFold);
                this.insertAdjacentHTML('beforeBegin', _this.ell);
                more = true;
                // 执行折叠事件
                'fold' in _this.handlers ? _this.emit('fold', '') : null;
            }
        };
        // 获得axMore元素的id
        !!this.targetDom.getAttribute('axMore') ? this.referName = this.targetDom.getAttribute('axMore') : this.referName = '';
        //初始化
        this.init();
    }
    init() {
         //添加到实例合集
         axInstance.push(this, this.options.insName, 'more');
          //断点合并参数
          !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        this.text = this.pureHtml.trim().replace(/<[^>]+>/g, '').replace(/ +/g, ' ');
        this.ell = this.text.substring(0, this.options.length) + '...';
        let textLength = [...this.text].length,
            link = `<a href="###" class="${this.options.classFold}">${this.options.textFold}</a>`;
        // 保存内容

        // 初始化折叠
        if (textLength > this.options.length) {
            this.targetDom.innerHTML = this.ell + link;
        }
        // 核心点击事件
        this.targetDom.lastChild.onclick = this.event;
        // 将实例存入数组(只存有ID的实例)
        if (this.referName) {
            if (axIsEmpty(axMores)) {
                axMores.push({ name: this.referName, instance: this });
            } else {
                //如果相同的name已经写入了数组，则不再写入
                if (!axMores.some(item => (item.name == this.referName))) {
                    axMores.push({ name: this.referName, instance: this });
                }
                /*  for (let i = 0; i < axMores.length; i++) {
                     if (axMores[i].name == this.referName) {
                         axMores[i].instance = this;
                         break;
                     } else {
                         axMores.push({ name: this.referName, instance: this });
                     }
                 } */
            }
        }
        'load' in this.handlers ? this.emit('load', '') : null;
        return this;
    }
    destroy(callback) {
        this.targetDom.lastChild.removeEventListener('click', this.event, false);
        this.targetDom.innerHTML = this.html;
        callback && callback.call(this);
    }
    updateContent(text, callback) {
        // text可以是#id，也可以普通字符串，也可以是dom
        if ((typeof text) === 'string') {
            if (text.substr(0, 1) == "#") {
                text = document.querySelector(text).textContent;
            } else {
                text = text;
            }
        } else {
            text = text.textContent;
        }
        this.pureHtml = text;
        this.init();
        'updated' in this.handlers ? this.emit('updated', '') : null;
        callback && callback.call(this);
        return this;
    }
    update(setting, callback) {
        this.options = axExtend(this.options, setting);
        this.init();
        'updated' in this.handlers ? this.emit('updated', '') : null;
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    //对使用axMore属性的元素使用本插件
    document.querySelectorAll('[axMore]').forEach(element => {
        new axMore(element);
    });
})();
 */
/*!
 * 插件：手风琴折叠；使用方法：new axAccordion('#id',{参数})
 */
class axAccordion {
    constructor(targetDom, options) {
        // 始终将targetDom转成DOM对象
        this.targetDom = axIdToDom(targetDom);
        // 定义参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//存储名称，字符串格式；如果需要保存实例当前数据请设定唯一的值，比如：storageName:'mystorage001'
            content: '',//设置数据源，支持数组格式数据和链接字符串，如果是链接字符串则为异步获取json
            dispose: false,//是否禁止使用，默认是false即不禁用，可选择true禁用，头部将不能点击操作
            active: '',//初始化焦点项，默认为0即默认以第一页为active状态，支持多种格式：一个字符串/一个数字/一个DOM节点/一个对象
            disabled: '',//初始化禁用项，默认为空即不使用禁用状态，支持多种格式：一个字符串/一个数字/一个DOM节点/一个对象，或以上数组
            multiple: false,//是否可同时显示多个item，默认false不允许，可填true允许
            collapseAll: true,//初始化是否默认折叠所有ul，默认是true折叠，可选择false不折叠
            borderShow: false,//是否显示边框，默认false不显示，可选择true显示边框
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            trigger: 'click',//打开菜单的事件，默认是click点击显示，可选择其他原生事件例如mouseenter
            max: 0,//最多允许多少项，默认为0即不限制
            headerBreakShow: true,//头部和内容之间是否显示分割线，默认true显示，可填false不显示
            bodyBreakShow: true,//项目之间是否显示分割线，默认true显示，可填false不显示
            targetClass: '',//给目标节点即ax-accordion添加class类，一个多个均可，多个class用英文逗号或空格分开
            itemNode: 'li',//项目的节点名，默认li，可使用其他节点名，例如section或div
            arrow: {
                enable: true,//是否显示指示箭头，默认true显示，可填false不显示
                rotate: 180,//指示箭头的旋转角度，可填0、90或180；如果填0表示不旋转，为了表示打开和关闭，那么iconClose和iconOpen应该不同
                iconClose: 'ax-icon-up',
                iconOpen: 'ax-icon-up',
                position: 'end',//指示箭头所在位置，默认end即在末尾，可选择begin在标题前面
            },
            tipPosition: 'begin',//提示文字的存放位置，默认begin即放在标题后面，可选择end放在末尾
            addToActive: true,//是否在添加子项后让新子项成为焦点，默认true，可填false
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onAdd: '',//回调函数，新增项后执行，支持一个参数即新增项
            onRemove: '',//回调函数，删除项后执行，无参数
            onDisable: '',//回调函数，禁用项后执行，支持一个参数即禁用项
            onEnable: '',//回调函数，启用项后执行，支持一个参数即启用项
            onDisableALL: '',//回调函数，全部禁用项后执行，无参数
            onEnableAll: '',//回调函数，全部启用项后执行，无参数
            onExpand: '',//回调函数，项目展开前执行，支持一个参数即当前展开项
            onExpanded: '',//回调函数，项目展开后执行，支持一个参数即当前展开项
            onCollapse: '',//回调函数，项目折叠前执行，支持一个参数即当前折叠项
            onCollapsed: '',//回调函数，项目折叠后执行，支持一个参数即当前折叠项
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onUpdateContent: '',//回调函数，子项内容更新后执行，支持一个参数即当前更新项
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
            onExpandAll: '',//回调函数，全部展开后执行，无参数
            onCollapseAll: '',//回调函数，全部折叠后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        // 保存this
        let _this = this;
        // 定义监听事件成员
        this.handlers = {};
        //触发事件函数
        this.triggerShow = function () {
            _this.toggle(this);
        }
        //初始化
        this.init();
    };
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'accordion');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;
        //初始化数据
        this.data = [];
        //判断是否有存储值而更新参数
        if (this.options.storageName) {
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            } else {
                //则先存下空对象
                axLocalStorage.set(this.options.storageName, {});
            }
        }

        //给节点增加属性和class类
        this.setAttribute();
        //获得标准数据并渲染
        this.initDataStd();
        //监听回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }

    async initDataStd() {
        //将数据格式统一：[{label:'',content:'',active:true,disabled:true,labelDom:'',contentDom:'',tip:''},{label:'',content:''}]
        if (!axIsEmpty(this.options.content)) {
            //content不为空则按content生成节点
            //先初始化清空
            this.targetDom.innerHTML = '';
            if (typeof this.options.content === 'string') {
                //是字符串则表示从json地址中获取json数据
                await this.getAsyncData((content) => {
                    //将数据渲染为节点
                    this.initRender(content);
                });
            } else {
                //更新变量
                this.data = this.options.content;
                if (this.data.length > 0) {
                    //将数据渲染为节点
                    this.initRender(this.data);
                }
            }
        } else {
            //content为空的话，则判断是否有物理节点
            if (this.targetDom.innerHTML) {
                let children = [...this.targetDom.children];
                for (let k in children) {
                    if (k >= this.options.max && this.options.max !== 0) {
                        //不可超过数量
                        break;
                    }
                    let wrapperDom = children[k],
                        headerDom = wrapperDom.querySelector('.ax-accordion-header'),
                        flexDom = wrapperDom.querySelector('[flex]'),
                        bodyDom = wrapperDom.querySelector('.ax-accordion-body'),
                        contentDom = wrapperDom.querySelector('.ax-accordion-content'),
                        labelDom = headerDom.querySelector('[label]'),
                        arrowDom = headerDom.querySelector('[arrow]'),
                        tipDom = wrapperDom.querySelector('[tip]'),
                        legendDom = wrapperDom.querySelector('[legend]'),
                        headerBreakDom = headerDom.nextElementSibling && headerDom.nextElementSibling.classList.contains('ax-break-line') ? headerDom.nextElementSibling : null,
                        bodyBreakDom = contentDom.nextElementSibling && contentDom.nextElementSibling.classList.contains('ax-break-line') ? contentDom.nextElementSibling : null;

                    let obj = {
                        label: labelDom.innerHTML.trim(),
                        content: contentDom.innerHTML.trim(),
                        wrapperDom: wrapperDom,
                        flexDom: flexDom,
                        labelDom: labelDom,
                        headerDom: headerDom,
                        bodyDom: bodyDom,
                        contentDom: contentDom,
                        arrowDom: arrowDom,
                        tipDom: tipDom,
                        legendDom: legendDom,
                        headerBreakDom: headerBreakDom,
                        bodyBreakDom: bodyBreakDom,
                    }
                    //设置焦点
                    if (wrapperDom.hasAttribute('active')) {
                        obj.active = true;
                    };
                    //设置禁用
                    if (wrapperDom.hasAttribute('disabled')) {
                        obj.disabled = true;
                    }
                    //设置旋转角度

                    this.data.push(obj);
                }
            }

        }
        //生成了数据和节点之后操作
        this.renderFinish();
    }
    getAsyncData(callback) {
        return axAjax({
            type: this.options.ajaxType,
            url: this.options.content,
            opened: (response) => {
                this.contentXhr = response.xhr;
            },
            success: (response) => {
                if (axIsEmpty(response.content)) {
                    console.warn('No data obtained!');
                    return false;
                }
                //清除内容（有loading的span节点）
                this.targetDom.innerHTML = '';
                //更新变量
                this.data = response.content;
                callback && callback(this.data);
            }
        }, this.targetDom);
    }
    initRender(data) {
        for (let k in data) {
            if (k >= this.options.max && this.options.max !== 0) {
                //不可超过数量
                break;
            }
            this.appendItem(data[k]);
        }
        //设置禁用
        this.disable(this.options.disabled);
    }
    appendItem(obj, callback) {
        //先判断是否异步或url，再判断content是否是字符串
        let content;
        if (obj.type === 'async' || obj.type === 'url') {
            //异步则先为空，active之后才加载内容
            content = '';
        } else if (obj.type === 'node') {
            //如果是节点类型，那么content应该填#id、.class、[name]、DIV或节点
            content = axIdToDom(obj.content);
        } else {
            if (typeof obj.content === 'string') {
                if (axStrType(obj.content)) {
                    //#id,.className,[name],DIV,.parent .child,
                    content = axIdToDom(obj.content).innerHTML;
                } else {
                    //非选择器字符串，也就是普通字符
                    content = obj.content
                }
            } else {
                //content本身是一个节点
                content = obj.content || '';
            }

        }
        //根据data创建节点
        let wrapperDom = axAddElem(this.options.itemNode),
            flexDom = axAddElem('div', { flex: '' }),
            labelDom = axAddElem('i', { label: '' }, obj.label),
            headerDom = axAddElem('div', { class: 'ax-accordion-header' }, labelDom),
            bodyDom = axAddElem('div', { class: 'ax-accordion-body' }),
            headerBreakDom = axAddElem('hr'),
            bodyBreakDom = axAddElem('hr'),
            //是url类型，则需要src属性
            contentDom = axAddElem(obj.type === 'url' ? 'iframe' : 'div', obj.type === 'url' ? { class: 'ax-accordion-content', frameborder: 0 } : { class: 'ax-accordion-content' }, content);
        //填充基本节点
        flexDom.appendChild(labelDom);
        headerDom.appendChild(flexDom);
        wrapperDom.appendChild(headerDom);
        //header分割线
        if (this.options.headerBreakShow) {
            headerDom.insertAdjacentElement('afterend', headerBreakDom);
            obj.headerBreakDom = headerBreakDom;
        } else {
            obj.headerBreakDom = null;
        }
        bodyDom.appendChild(contentDom);
        //body分割线
        if (this.options.bodyBreakShow) {
            bodyDom.insertAdjacentElement('beforeend', bodyBreakDom);
            obj.bodyBreakDom = bodyBreakDom;
        } else {
            obj.bodyBreakDom = null;
        }
        wrapperDom.appendChild(bodyDom);
        //添加节点属性
        obj.wrapperDom = wrapperDom;
        obj.headerDom = headerDom;
        obj.bodyDom = bodyDom;
        obj.contentDom = contentDom;
        obj.labelDom = labelDom;
        obj.headerBreakDom = headerBreakDom;

        //增加disabled属性
        obj.disabled ? wrapperDom.setAttribute('disabled', '') : null;
        //增加其他属性
        if (!axIsEmpty(obj.props)) {
            for (let k in obj.props) {
                wrapperDom.setAttribute(k, obj.props[k]);
            }
        }
        //异步和iframe内容，active后才显示以提高渲染速度
        //此刻：obj:{label:'',content:'',labelDom:'',contentDom:'',wrapperDom:'',bodyDom:'',headerDom:'',legend:'',tip:'',active:true,disabled:true,type:'async'}
        //显示tip
        if (obj.tip) {
            let tipDom = axAddElem('i', { tip: '' }, obj.tip);
            if (this.options.tipPosition === 'begin') {
                labelDom.insertAdjacentElement('afterend', tipDom);
            } else if (this.options.tipPosition === 'end') {
                headerDom.insertAdjacentElement('beforeend', tipDom);
            }
            //追加入对象
            obj.tipDom = tipDom;
        }
        //显示legend
        if (obj.legend) {
            let legendDom = axAddElem('i', { legend: '', class: obj.legend ? 'ax-iconfont ' + obj.legend : 'ax-iconfont' });
            headerDom.insertAdjacentElement('afterbegin', legendDom);
            //追加入对象
            obj.legendDom = legendDom;
        }
        //显示arrow
        if (this.options.arrow.enable) {
            let arrowDom = axAddElem('i', { arrow: '', class: `ax-iconfont ${this.options.arrow.iconClose}` });
            if (this.options.arrow.position === 'begin') {
                headerDom.insertAdjacentElement('afterbegin', arrowDom);
            } else if (this.options.arrow.position === 'end') {
                headerDom.insertAdjacentElement('beforeend', arrowDom);
            }
            //追加入对象
            obj.arrowDom = arrowDom;
        }
        //追加到容器里
        this.targetDom.appendChild(wrapperDom);
        callback && callback(obj);
    }
    initActive() {
        //如果数据为空则不再执行
        if (this.data.length === 0) {
            return false;
        }

        if (this.options.multiple && !this.options.collapseAll) {
            //全部展开
            this.expandAll();
        } else {
            let activeItems = this.data.filter(k => k.active);
            if (this.options.active !== '') {
                let stringArr = axValueToArr(this.options.active),
                    itemArr = [...stringArr.map(k => axFindItem(k, this.data)), ...activeItems].filter(Boolean);
                //获得当前所有active项
                if (itemArr.length > 0) {
                    if (this.options.multiple) {
                        //允许多个
                        itemArr.forEach(k => {
                            this.expand(k);
                        });
                    } else {
                        this.expand(itemArr[0]);
                    }
                }
            } else {
                if (this.options.multiple) {
                    //允许多个
                    activeItems.forEach(k => {
                        this.expand(k);
                    });
                } else {
                    this.expand(activeItems[0]);
                }
            }
        }
    }
    asyncActive(obj, callback) {
        if (obj.contentDom.innerHTML.trim()) {
            //要求contentDom当前是空的（避免反复渲染）
            return false;
        }
        //要求type===async，content有地址路径而且是字符串格式
        if (obj.type === 'async' && obj.content && typeof obj.content === 'string') {
            axAjax({
                url: obj.content,
                type: this.options.ajaxType,
                success: function (content) {
                    obj.contentDom.innerHTML = content;
                    callback && callback(content);
                }
            });
        }
    }
    setAttribute() {
        //强行带上ax-accordion类
        !this.targetDom.classList.contains('ax-accordion') ? this.targetDom.classList.add('ax-accordion') : null;
        //给主要外框带上class类
        this.options.targetClass ? axClass.add(this.targetDom, this.options.targetClass) : null;
        //设置边框
        this.options.borderShow ? axClass.add(this.targetDom, 'ax-border') : null;
        //追加rotate属性
        this.options.arrow.enable && this.options.arrow.rotate ? this.targetDom.setAttribute('rotate', this.options.arrow.rotate) : null;
    }
    renderFinish() {
        //初始化展开
        this.initActive();
        //启用状态才可点击操作
        if (!this.options.dispose) {
            //触发头部标题
            this.data.forEach(k => {
                //未禁止才可触发
                if (!k.disabled) {
                    k.headerDom.addEventListener(this.options.trigger, this.triggerShow, false);
                }
            });
        }
        //保存数据
        this.save();
    }
    workToggle(item, type = 'disable') {
        if (type === 'enable' && item.disabled) {
            delete item.disabled;
            item.wrapperDom.removeAttribute('disabled');
            //绑定
            item.headerDom.addEventListener(this.options.trigger, this.triggerShow, false);
        } else if (type === 'disable' && !item.disabled) {
            item.disabled = true;
            item.wrapperDom.setAttribute('disabled', '');
            //解绑
            item.headerDom.removeEventListener(this.options.trigger, this.triggerShow, false);
        }
        //保存数据
        this.save();
    }
    getDisabled() {
        return this.data.filter(k => k.disabled);
    }
    getActive() {
        return this.data.filter(k => k.active);
    }
    collapseAll(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.data.filter(k => k.active).forEach(k => {
            this.collapse(k);
        });
        //监听回调
        'collapseAll' in this.handlers ? this.emit('collapseAll', '') : null;
        this.options.onCollapseAll && this.options.onCollapseAll.call(this);
        callback && callback.call(this);
    }
    expandAll(callback) {
        //如果已经销毁或不允许多个展示则不再执行
        if (this.destroyed || !this.options.multiple) {
            return this;
        }
        this.data.forEach(k => {
            this.expand(k);
        });
        //监听回调
        'expandAll' in this.handlers ? this.emit('expandAll', '') : null;
        this.options.onExpandAll && this.options.onExpandAll.call(this);
        callback && callback.call(this);
    }
    disable(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj为空则不再执行,0除外
        if (obj !== 0 && axIsEmpty(obj)) {
            return false;
        }
        let disableFun = (k) => {
            let item = axFindItem(k, this.data);
            //找到元素并设置禁用
            if (item) {
                this.workToggle(item, 'disable');
            }
        }
        //设置
        if (Array.isArray(obj)) {
            obj.forEach(k => {
                disableFun(k);
            });
        } else {
            disableFun(obj);
        }
        //保存数据
        this.save();
        //监听回调
        let disableds = this.data.filter(k => k.disabled);
        'disable' in this.handlers ? this.emit('disable', disableds) : null;
        this.options.onDisable && this.options.onDisable.call(this, disableds);
        callback && callback.call(this, disableds);
    }
    disableAll(callback) {
        this.disable(this.data);
        //监听回调
        'disableAll' in this.handlers ? this.emit('disableAll', '') : null;
        this.options.onDisableAll && this.options.onDisableAll.call(this);
        callback && callback.call(this);
    }
    enable(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj为空则不再执行,0除外
        if (obj !== 0 && axIsEmpty(obj)) {
            return false;
        }
        let enableFun = (k) => {
            let item = axFindItem(k, this.data);
            //找到元素并设置禁用
            if (item) {
                this.workToggle(item, 'enable');
            }
        }
        //设置
        if (Array.isArray(obj)) {
            obj.forEach(k => {
                enableFun(k);
            });
        } else {
            enableFun(obj);
        }
        //保存数据
        this.save();
        //监听回调
        let enableds = this.data.filter(k => !k.disabled);
        'enable' in this.handlers ? this.emit('enable', enableds) : null;
        this.options.onEnable && this.options.onEnable.call(this, enableds);
        callback && callback.call(this, enableds);
    }
    enableAll(callback) {
        this.enable(this.data);
        //监听回调
        'enableAll' in this.handlers ? this.emit('enableAll', '') : null;
        this.options.onEnableAll && this.options.onEnableAll.call(this);
        callback && callback.call(this);
    }
    updateContent(item, content, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //item可是数字、字符串、DOM和{}对象
        let me = axFindItem(item, this.data);
        if (me) {
            //先清空内容
            me.contentDom.innerHTML = '';
            //更新content
            content ? me.content = content : null;
            if (me.type === 'url') {
                //url类型
                me.contentDom.src = content;
            } else if (me.type === 'async') {
                //异步类型
                this.asyncActive(me);
            } else if (axType(me.conten).includes('HTML')) {
                //content是一个节点
                me.contentDom.appendChild(me.content);
            } else {
                //字符串
                me.contentDom.innerHTML = me.content;
            }

            //保存数据
            this.save();
        }
        //回调
        'updateContent' in this.handlers ? this.emit('updateContent', me) : null;
        this.options.onUpdateContent && this.options.onUpdateContent.call(this, me);
        callback && callback.call(this, me);
    }
    toggle(item) {
        let me = axFindItem(item, this.data);
        if (me.active) {
            this.collapse(me);
        } else {
            this.expand(me);
        }
    }
    collapse(obj, callback) {
        let type = axType(obj),
            item = type === 'Object' ? obj : axFindItem(obj, this.data);
        //如果是隐藏状态则不再执行
        if (!item || !item.shown) {
            return false;
        }
        //监听回调
        'collapse' in this.handlers ? this.emit('collapse', item) : null;
        this.options.onCollapse && this.options.onCollapse.call(this, item);
        //设置属性
        delete item.active;
        item.wrapperDom.removeAttribute('active');
        //如果不旋转则替换图标
        if (this.options.arrow.enable && !this.options.arrow.rotate) {
            item.arrowDom.classList.remove(this.options.arrow.iconOpen);
            item.arrowDom.classList.add(this.options.arrow.iconClose);
        }
        axSlideUp(item.bodyDom, '', () => {
            //监听回调
            Reflect.deleteProperty(item, 'shown');
            'collapsed' in this.handlers ? this.emit('collapsed', item) : null;
            this.options.onCollapsed && this.options.onCollapsed.call(this, item);
            callback && callback.call(this, item);
        });
        //保存数据
        this.save();
    }
    slideDown(obj, callback) {
        let type = axType(obj),
            item = type === 'Object' ? obj : axFindItem(obj, this.data);
        //如果是显示状态则不再执行
        if (!item || item.shown) {
            return false;
        }
        //设置属性
        item.active = true;
        item.wrapperDom.setAttribute('active', '');
        //如果不旋转则替换图标
        if (this.options.arrow.enable && !this.options.arrow.rotate) {
            item.arrowDom.classList.remove(this.options.arrow.iconClose);
            item.arrowDom.classList.add(this.options.arrow.iconOpen);
        }
        axSlideDown(item.bodyDom, '', () => {
            //监听回调
            item.shown = true;
            'expanded' in this.handlers ? this.emit('expanded', item) : null;
            this.options.onExpanded && this.options.onExpanded.call(this, item);
            callback && callback.call(this, item);
        });
    }
    expand(item, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj可以是对象，字符串或数字，如果是对象直接相等不再找
        let me = axFindItem(item, this.data), others;
        if (!axIsEmpty(me)) {
            //监听回调
            'expand' in this.handlers ? this.emit('expand', me) : null;
            this.options.onExpand && this.options.onExpand.call(this, me);

            //如果是异步则获取内容
            if (me.type === 'async' && !me.contentDom.innerHTML) {
                //arrow转成loading加载效果
                me.arrowDom ? me.arrowDom.setAttribute('loading', '') : null;
                this.asyncActive(me, () => {
                    this.slideDown(me, () => {
                        //回调
                        callback && callback.call(this, me);
                    });
                    //移除加载效果
                    me.arrowDom ? me.arrowDom.removeAttribute('loading') : null;
                });
            } else {
                //iframe类型则添加src属性
                if (me.type === 'url' && !me.contentDom.getAttribute('src')) {
                    me.contentDom.setAttribute('src', me.content)
                }
                this.slideDown(me, () => {
                    //回调
                    callback && callback.call(this, me);
                });
            }

            //其他兄弟节点
            if (!this.options.multiple) {
                others = this.data.filter(k => (k !== me && k.active));
                others.forEach(k => {
                    this.collapse(k);
                });
            }
            //保存数据
            this.save();
        }

    }
    add(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //data为空则不再执行
        if (axIsEmpty(data)) {
            return false;
        }
        //当前data的length大于等于允许的max值则不再允许添加了
        if (this.data.length >= this.options.max && this.options.max !== 0) {
            console.warn('The number of children exceeds the maximum!');
            return false;
        }
        let type = axType(data),
            items = [],
            appendFun = (obj) => {
                if (this.data.findIndex(k => k.label === obj.label) === -1) {
                    //如果没有相同的头部label，才append
                    this.appendItem(obj, () => {
                        this.data.push(obj);
                        items.push(obj);
                    });
                }
            };
        if (type === 'Object') {
            //data是一个对象
            appendFun(data);
        } if (type === 'Array' && data.length > 0) {
            //data是一个数组
            data.forEach(k => {
                appendFun(k);
            });
        }
        //items有至少一项才执行
        if (items.length > 0) {
            //是否允许新增后成为焦点，让新增的最后一个成为焦点
            if (this.options.addToActive) {
                if (this.options.multiple) {
                    items.forEach(k => {
                        this.expand(k);
                    });
                } else {
                    this.expand(items[items.length - 1]);
                }
            }
            //生成了数据和节点之后操作
            this.renderFinish();
        }
        //保存数据
        this.save();
        //回调监听
        'add' in this.handlers ? this.emit('add', items) : null;
        this.options.onAdd && this.options.onAdd.call(this, items);
        callback && callback.call(this, items);
        return this;
    }
    remove(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //data为空则不再执行
        if (axIsEmpty(data)) {
            return false;
        }
        //数据为空了就不能再删除了
        if (this.data.length === 0) {
            console.warn('The data is already empty!');
            return false;
        }
        let removeItem = (child) => {
            let item = axFindItem(child, this.data);
            if (item) {
                //因为close按钮在labelDom里，点击close会触发expand事件，所以先移除
                item.headerDom.removeEventListener(this.options.trigger, this.triggerShow, false);
                //再移除两个节点
                item.wrapperDom.remove();
                //更新this.data
                this.data = this.data.filter(k => k !== item);
            }
        };
        //删除
        if (Array.isArray(data)) {
            data.forEach(k => {
                removeItem(k);
            });
        } else {
            removeItem(data);
        }
        //保存数据
        this.save();
        //回调监听
        'remove' in this.handlers ? this.emit('remove', '') : null;
        this.options.onRemove && this.options.onRemove.call(this);
        callback && callback.call(this);
        return this;
    }
    save(props, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }
        setTimeout(() => {
            let idsActive = this.data.filter(k => k.active).map(k => k.id).filter(Boolean),
                idsDisabled = this.data.filter(k => k.disabled).map(k => k.id).filter(Boolean);
            if (!props) {
                //保存当前焦点索引
                axLocalStorage.set(this.options.storageName, { active: idsActive, disabled: idsDisabled, content: this.data });
            } else {
                //保存参数,props={}
                !props.hasOwnProperty('active') ? props.active = idsActive : null;
                !props.hasOwnProperty('disabled') ? props.disabled = idsDisabled : null;
                !props.hasOwnProperty('content') ? props.content = this.data : null;
                axLocalStorage.set(this.options.storageName, props);
            }
            //回调监听
            let getValue = axLocalStorage.get(this.options.storageName);
            'save' in this.handlers ? this.emit('save', getValue) : null;
            this.options.onSave && this.options.onSave.call(this, getValue);
            callback && callback.call(this, getValue);
            return this;
        }, 0)
    }
    destroy(callback) {
        //取消绑定事件
        this.data.forEach(k => {
            k.headerDom.removeEventListener(this.options.trigger, this.triggerShow);
            //k.closeDom ? k.closeDom.onclick = null : null;
            //k.updateDom ? k.updateDom.onclick = null : null;
        });
        //更新销毁状态
        this.destroyed = true;
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
        //监听回调
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.options = axExtend(this.options, setting);
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
        this.init();
        //监听回调
        'update' in this.handlers ? this.emit('update', '') : null;
        this.options.onUpdate && this.options.onUpdate.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
// 对使用axAccordion属性的元素使用本插件
/* document.querySelectorAll('[axAccordion]').forEach(element => {
    new axAccordion(element);
}); */

/*!
 * 插件：页签切换；使用方法：axTab('#id',{参数})
 */
class axTab {
    constructor(targetDom, options) {
        //始终将targetDom转成DOM对象
        this.targetDom = axIdToDom(targetDom);
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//是否保存当前实例数据，包括active状态、disabled状态、content数据以及更新的参数，如果需要保存请设定唯一的name，比如：'mystorage001'；(不应该使用update方法改变该值)
            trigger: 'click',//页签切换的动作，默认是click点击切换，可选mouseenter、mouseover等原生事件切换
            disabled: '',//页签初始化禁用项，默认为空即不使用禁用状态，支持多种格式：一个字符串/一个数字/一个DOM节点/一个对象，或以上数组
            active: 0,//页签初始化焦点项，默认为0即默认以第一页为active状态，支持多种格式：一个字符串/一个数字/一个DOM节点/一个对象
            content: '',//设置tab的数据源，支持数组格式数据和链接字符串，如果是链接字符串则为异步获取json
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            headerNode: 'ul',//tab头部的节点名，默认ul
            bodyNode: 'ul',//body部分的节点名，默认ul
            headerChildNode: 'li',//tab头部子节点名，默认li
            bodyChildNode: 'li',//body部分的子节点名，默认li
            max: 0,//最多允许多少个tab页签，默认为0即不限制
            removable: false,//是否显示页签的关闭按钮，默认false不显示，可选择true显示
            updatable: false,//是否显示页签的刷新按钮，默认false不显示，可选择true显示
            wrapperClass: '',//Tab外框节点的class类，多个用英文逗号或空格分开
            headerClass: 'ax-tab-group',//header节点的class类，可定义头部风格，可填ax-tab-group、ax-btn-group或ax-lamp-group，多个用英文逗号或空格分开
            bodyClass: '',//body节点的class类，多个用英文逗号或空格分开
            sticky: 0,//保持某页签始终不可关闭，可填一个数字即页签索引，或页签文本，默认为0即第一个页签不可关闭
            slider: {
                enable: false,//是否让header可滚动，默认false不滚动，可填true滚动（引入axSlider插件）
                arrowShow: false,//是否显示slider的前后按钮箭头
                options: {
                    slides: 'auto', 
                    scroll: {
                        drift: { mode:'free',coef:1 }
                    }
                },//如果启用了slider.enable，那么可填写slider.options参数
            },
            menu: {
                enable: false,//是否开启页签菜单，默认false不显示，可选择true开启
                options: {
                    columns: 4,//设置菜单下拉窗一行几列，如果>1将使用grid布局
                    size: 'lg',//设置菜单下拉窗的尺寸，与axDropdown的size参数使用方法一致
                }
            },//menu下拉菜单功能在使用slider之后才有效
            addToActive: true,//是否在添加Tab后让新Tab成为焦点，默认true，可填false
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onAdd: '',//回调函数，新增项后执行，支持一个参数即新增项
            onBeforeRemove: '',//回调函数，删除前执行，支持一个参数即删除项，如果返回true则将删除节点，返回false将停止操作；如果回调中有异步操作需要手动使用this.remove(item)方法删除节点
            onRemove: '',//回调函数，删除项后执行，无参数
            onDisable: '',//回调函数，禁用项后执行，支持一个参数即禁用项
            onEnable: '',//回调函数，启用项后执行，支持一个参数即启用项
            onDisableALL: '',//回调函数，全部禁用项后执行，无参数
            onEnableAll: '',//回调函数，全部启用项后执行，无参数
            onActive: '',//回调函数，激活项后执行，支持一个参数即当前激活项
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onUpdateContent: '',//回调函数，Tab内容更新后执行，支持一个参数即当前更新项
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
        }, options, this.targetDom, this.constructor.name);
        let _this = this;
        // 定义监听事件成员
        this.handlers = {};
        //头部滚动的节点
        this.sliderDom = null;
        //头部滚动的实例
        this.sliderIns = null;
        //极限索引
        this.maxIndex = 0;
        //触发头部tab
        this.triggerShow = function () {
            _this.active(this);
        }
        //头部按钮删除tab
        this.labelRemove = (obj) => {
            if (this.options.onBeforeRemove) {
                let flag = this.options.onBeforeRemove.call(this, obj);
                if (flag) {
                    this.remove(obj);
                }
            } else {
                this.remove(obj);
            }
        }
        //初始化
        this.init();
    }
    init() {
        //添加到实例合集并设置销毁状态
        axInstance.push(this, this.options.insName, 'tab');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;
        //初始化数据
        this.data = [];
        //如果有存储名，则先存下对象
        if (this.options.storageName) {
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            } else {
                //则先存下空对象
                axLocalStorage.set(this.options.storageName, {});
            }
        }

        //获得标准数据并渲染
        this.initDataStd();
        //监听回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    async initDataStd() {
        //将数据格式统一：[{label:'',content:'',active:true,disabled:true,labelDom:'',headerDom:'',removeDom:'',updateDom:'',bodyDom:''},{label:'',content:''}]
        if (!axIsEmpty(this.options.content)) {
            //content不为空则按content生成节点
            //先初始化清空
            this.targetDom.innerHTML = '';
            if (typeof this.options.content === 'string') {
                //是字符串则表示从json地址中获取json数据
                await axAjax({
                    url: this.options.content,
                    type: this.options.ajaxType,
                    success: (response) => {
                        //手动清除loading
                        this.targetDom.innerHTML = '';
                        //更新变量
                        this.data = response.content;
                    }
                }, this.targetDom);
            } else {
                //更新变量
                this.data = this.options.content;
            }
            if (this.data.length > 0) {
                //将数据渲染为节点
                this.initRender(this.data);
            }
        } else {
            //content为空的话，则判断是否有物理节点

            //强行带上ax-tab类
            !this.targetDom.classList.contains('ax-tab') ? this.targetDom.classList.add('ax-tab') : null;
            //从物理节点中获取数据属性
            if (this.targetDom.querySelector('.ax-tab-header') && this.targetDom.querySelector('.ax-tab-body')) {
                //如果targetDom本身有header和body节点，那么从现有的HTML中获取变量
                this.navDom = this.targetDom.querySelector('[nav]');
                this.prevDom = this.navDom ? this.navDom.querySelector('[prev]') : null;
                this.nextDom = this.navDom ? this.navDom.querySelector('[next]') : null;
                this.menuDom = this.navDom ? this.navDom.querySelector('[menu]') : null;
                this.sliderDom = this.targetDom.querySelector('.ax-slider');
                this.headerWrapperDom = this.targetDom.querySelector('.ax-tab-header');
                this.bodyWrapperDom = this.targetDom.querySelector('.ax-tab-body');

                let headerChildren = [...this.headerWrapperDom.children],
                    bodyChildren = [...this.bodyWrapperDom.children];


                for (let i = 0, len = headerChildren.length; i < len; i++) {
                    if (i >= this.options.max && this.options.max !== 0) {
                        //不可超过数量
                        break;
                    }
                    let k = headerChildren[i],
                        tabCaptionDom = k.querySelector('[label]'),
                        tabRemoveDom = k.querySelector('[remove]'),
                        tabUpdateDom = k.querySelector('[update]');
                    //body子元素必须等于或多于header子元素，才可形成对应关系
                    let bodyItem = bodyChildren[i];
                    if (bodyItem) {
                        let obj = {
                            label: tabCaptionDom ? tabCaptionDom.innerText : k.innerHTML,
                            content: bodyItem.innerHTML,
                            labelDom: tabCaptionDom,
                            headerDom: k,
                            bodyDom: bodyItem,
                        }
                        //设置焦点
                        if (k.hasAttribute('active')) {
                            obj.active = true;
                        };
                        //设置禁用
                        if (k.hasAttribute('disabled')) {
                            obj.disabled = true;
                        }
                        //追加索引
                        obj.index = this.maxIndex;
                        this.data.push(obj);
                        //找到更新按钮并赋予刷新事件
                        tabUpdateDom ? obj.updateDom = tabUpdateDom : null;
                        //找到关闭按钮并赋予删除事件
                        if (tabRemoveDom) {
                            !this.isSticky(obj) ? obj.removeDom = tabRemoveDom : tabRemoveDom.remove();
                        }
                    }
                    //递增
                    this.maxIndex++;
                }
                //创建slider实例
                this.createSlider();
                //必须有一项是active
                this.initActive();
            }
        }
        //生成了数据和节点之后操作
        this.renderFinish();
    }
    initRender(data) {
        this.headerWrapperDom = axAddElem(this.options.headerNode, { class: 'ax-tab-header' });
        this.bodyWrapperDom = axAddElem(this.options.bodyNode, { class: 'ax-tab-body' });

        for (let i = 0, len = data.length; i < len; i++) {
            if (i >= this.options.max && this.options.max !== 0) {
                //不可超过数量
                break;
            }
            this.appendItem(data[i]);
        }
        //必须有一项是active
        this.initActive();
        //设置禁用
        this.disable(this.options.disabled);
        //给节点增加属性和class类
        this.setAttribute();
        //使用slider
        if (this.options.slider.enable) {
            this.headerWrapperDom.classList.add('ax-slider-wrapper');
            this.sliderDom = axAddElem('div', { class: 'ax-slider' }, this.headerWrapperDom);
            this.navDom = axAddElem('div', { class: 'ax-tab-slider' }, this.sliderDom);
            //添加操作按钮
            if (this.options.slider.arrowShow) {
                this.prevDom = axAddElem('i', { prev: '' });
                this.nextDom = axAddElem('i', { next: '' });
                this.navDom.insertAdjacentElement('afterbegin', this.prevDom);
                this.navDom.insertAdjacentElement('beforeend', this.nextDom);
            }
            //添加Tab菜单
            if (this.options.menu.enable) {
                this.menuDom = axAddElem('i', { menu: '' });
                this.navDom.insertAdjacentElement('beforeend', this.menuDom);
            }
            //将slider加入到ax-tab中
            this.targetDom.appendChild(this.navDom);
            //创建slider实例
            this.createSlider();
        } else {
            //循环完毕将header加入到ax-tab中
            this.targetDom.appendChild(this.headerWrapperDom);
        }
        //循环完毕将body加入到ax-tab中
        this.targetDom.appendChild(this.bodyWrapperDom);
    }
    isSticky(obj) {
        return typeof this.options.sticky === 'string' ? this.options.sticky === obj.label : this.options.sticky === obj.index;
    }
    //根据单个属性对象追加一个Tab
    appendItem(obj, callback) {
        if (!obj.label) {
            //如果label属性不存在或为空则不再执行
            return false;
        }
        //先判断是否异步或url，再判断content是否是字符串
        let content;
        if (obj.type === 'async' || obj.type === 'url') {
            //异步则先为空，active之后才加载内容
            content = '';
        } else if (obj.type === 'node') {
            //如果是节点类型，那么content应该填#id、.class、[name]、DIV或节点
            content = axIdToDom(obj.content);
        } else {
            if (typeof obj.content === 'string') {
                if (axStrType(obj.content)) {
                    //#id,.className,[name],DIV,.parent .child,
                    content = axIdToDom(obj.content).innerHTML;
                } else {
                    //非选择器字符串，也就是普通字符
                    content = obj.content
                }
            } else {
                //content本身是一个节点
                content = obj.content || '';
            }

        }
        //ax-bounce-group需特别处理，标题用<i label>包裹
        let headerChildDom = axAddElem(this.options.headerChildNode),
            tabLabelDom = axAddElem('i', { label: '' }, obj.label),
            //是url类型，则需要src属性
            bodyChildDom = axAddElem(obj.type === 'url' ? 'iframe' : this.options.bodyChildNode, obj.type === 'url' ? { src: obj.content, frameborder: 0 } : '', content),
            tabRemoveDom = axAddElem('i', { remove: '' }),
            tabUpdateDom = axAddElem('i', { update: '' });
        //追加索引
        obj.index = this.maxIndex;
        //追加label节点
        headerChildDom.appendChild(tabLabelDom);
        //增加active属性
        obj.active ? (headerChildDom.setAttribute('active', ''), bodyChildDom.setAttribute('active', '')) : null;
        //增加disabled属性
        obj.disabled ? headerChildDom.setAttribute('disabled', '') : null;
        //增加其他属性
        if (!axIsEmpty(obj.props)) {
            for (let k in obj.props) {
                bodyChildDom.setAttribute(k, obj.props[k]);
            }
        }
        //追加tab和content节点
        obj.headerDom = headerChildDom;
        obj.labelDom = tabLabelDom;
        obj.bodyDom = bodyChildDom;
        this.headerWrapperDom.appendChild(headerChildDom);
        //body的子节点先不append，active后才append以提高渲染速度
        //this.bodyWrapperDom.appendChild(bodyChildDom);
        //此刻：obj:{label:'',content:'',labelDom:'',headerDom:'',bodyDom:'',active:true,disabled:true,type:'async'}
        //追加刷新和关闭按钮
        //显示update按钮
        if (this.options.updatable) {
            //将更新节点追加到obj里
            obj.updateDom = tabUpdateDom;
            //追加到头部
            headerChildDom.appendChild(tabUpdateDom);

        }

        //显示remove按钮，但是不是sticky
        if (this.options.removable && !this.isSticky(obj)) {
            //将更新节点追加到obj里
            obj.removeDom = tabRemoveDom;
            //追加到头部
            headerChildDom.appendChild(tabRemoveDom);
        }
        //maxIndex递增
        this.maxIndex++;
        callback && callback(obj);
    }
    initActive() {
        //如果数据为空则不再执行
        if (this.data.length === 0) {
            return false;
        }
        let activeTab = this.data.find(k => k.active) || this.data[0];
        if (this.options.active) {
            //如果参数active不等于0或者填写了其他索引值，则以参数为准
            this.active(this.options.active);
        } else {
            //如果能找到一个active项则执行，如果没有则让第一项active
            this.active(activeTab);
        }

        /* 
                if (this.options.active) {
                    //如果参数active不等于0或者填写了其他索引值，则以参数为准
                    this.active(this.options.active);
                } else {
                    let activeTab;
                    //如果active为0或者没有填则让插件自动设置active
                    if (!this.data.find(k => k.active)) {
                        //如果没有active项，将第一项设为active
                        activeTab = this.data[0];
                        activeTab.active = true;
                        //设置焦点属性
                        activeTab.labelDom.setAttribute('active', '');
                        activeTab.bodyDom.setAttribute('active', '');
                    } else {
                        //如果有多个active项，只保留第一项
                        let activeTabs = this.data.filter(k => k.active);
                        if (activeTabs.length > 1) {
                            activeTabs.slice(1).forEach(k => {
                                k.active ? delete k.active : null;
                                k.labelDom.hasAttribute('active') ? k.labelDom.removeAttribute('active') : null;
                                k.bodyDom.hasAttribute('active') ? k.bodyDom.revoveAttribute('active') : null;
                            });
                        }
                        //取active的第一项
                        activeTab = activeTabs[0];
                    }
                    //将body子节点append
                    axIsVnode(activeTab.bodyDom) ? this.bodyWrapperDom.appendChild(activeTab.bodyDom) : null;
                    //如果是异步则获取内容
                    activeTab.type === 'async' ? this.asyncActive(activeTab) : null;
                } */
    }
    asyncActive(obj, callback) {
        if (obj.bodyDom.innerHTML.trim()) {
            //要求bodyDom当前是空的（避免反复渲染）
            return false;
        }
        //要求type===async，content有地址路径而且是字符串格式
        if (obj.type === 'async' && obj.content && typeof obj.content === 'string') {
            axAjax({
                url: obj.content,
                type: this.options.ajaxType,
                success: function (response) {
                    obj.bodyDom.innerHTML = response.content;
                    callback && callback(response.content);
                }
            }, obj.bodyDom);
        }
    }
    getDropDownData() {
        return this.data.map(k => {
            return { value: k.label, label: k.label }
        })
    }
    createSlider() {
        if (!this.sliderDom) {
            return false;
        }
        let sliderOpt = this.options.slider.options;
        if (this.prevDom && this.nextDom) {
            sliderOpt.navigation = { enable: true, prev: { selector: this.prevDom }, next: { selector: this.nextDom } };
        }
        this.sliderIns = new axSlider(this.sliderDom, sliderOpt);
        //slider菜单
        if (this.menuDom) {
            setTimeout(() => {
                this.dropdown = new axDropdown(this.menuDom, {
                    arrowShow: false,
                    grid: {
                        //为0或1则不使用grid布局
                        enable: this.options.menu.options.columns === 0 || this.options.menu.options.columns === 1 ? false : true,
                        columns: this.options.menu.options.columns,
                        fenceShow: true,
                    },
                    type: 'button',
                    popup: {
                        size: this.options.menu.options.size,
                    },
                    content: this.getDropDownData(),
                });
                this.dropdown.popup.on('show', () => {
                    this.getActive() && this.getActive().label ? this.dropdown.updateValues(this.getActive().label) : null;
                });
                //监听dropdown的设置事件（设置并关闭浮窗后使用active方法）
                this.dropdown.on('set', (value) => {
                    value || value === 0 ? this.active(value) : null;
                    this.sliderIns.toBullet(this.getIndexActive());
                });
            }, 0);
        }
    }
    setAttribute() {
        //强行带上ax-tab类
        !this.targetDom.classList.contains('ax-tab') ? this.targetDom.classList.add('ax-tab') : null;
        //给主要外框带上class类
        this.options.wrapperClass ? axClass.add(this.targetDom, this.options.wrapperClass) : null;
        this.options.headerClass ? axClass.add(this.headerWrapperDom, this.options.headerClass) : null;
        this.options.bodyClass ? axClass.add(this.bodyWrapperDom, this.options.bodyClass) : null;
    }
    renderFinish() {
        //触发头部tab
        this.data.forEach(k => {
            //未禁止才可触发
            if (!k.disabled) {
                //label可能不存在
                let triggerDom = k.labelDom ? k.labelDom : k.headerDom;
                triggerDom.addEventListener(this.options.trigger, this.triggerShow, false);

                //更新按钮
                if (k.updateDom) {
                    k.updateDom.onclick = () => {
                        this.updateContent(k);
                    }
                }
                //关闭按钮
                if (k.removeDom) {
                    k.removeDom.onclick = () => { this.labelRemove(k) };
                }
            }
        });
        //保存数据
        this.save();
    }
    workToggle(item, type = 'disable') {
        let triggerDom = item.labelDom ? item.labelDom : item.headerDom;
        if (type === 'enable' && item.disabled) {
            delete item.disabled;
            item.headerDom.removeAttribute('disabled');
            item.bodyDom.removeAttribute('disabled');
            //绑定
            triggerDom.addEventListener(this.options.trigger, this.triggerShow, false);
            item.removeDom ? item.removeDom.onclick = () => { this.labelRemove(item) } : null;
            item.updateDom ? item.updateDom.onclick = () => { this.updateContent(item) } : null;
        } else if (type === 'disable' && !item.disabled) {
            item.disabled = true;
            item.headerDom.setAttribute('disabled', '');
            item.bodyDom.setAttribute('disabled', '');
            //解绑
            triggerDom.removeEventListener(this.options.trigger, this.triggerShow, false);
            item.removeDom ? item.removeDom.onclick = null : null;
            item.updateDom ? item.updateDom.onclick = null : null;
        }
        //保存数据
        this.save();
    }
    getIndexActive() {
        return this.data.findIndex(k => k.active);
    }
    getDisabled() {
        return this.data.filter(k => k.disabled);
    }
    getActive() {
        return this.data.find(k => k.active);
    }
    disable(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj为空则不再执行,0除外
        if (obj !== 0 && axIsEmpty(obj)) {
            return false;
        }
        let disableFun = (k) => {
            let item = axFindItem(k, this.data);
            //找到元素并设置禁用
            if (item) {
                this.workToggle(item, 'disable');
            }
        }
        //设置
        if (Array.isArray(obj)) {
            obj.forEach(k => {
                disableFun(k);
            });
        } else {
            disableFun(obj);
        }
        //监听回调
        let disableds = this.data.filter(k => k.disabled);
        'disable' in this.handlers ? this.emit('disable', disableds) : null;
        this.options.onDisable && this.options.onDisable.call(this, disableds);
        callback && callback.call(this, disableds);
    }
    disableAll(callback) {
        this.disable(this.data);
        //监听回调
        'disableAll' in this.handlers ? this.emit('disableAll', '') : null;
        this.options.onDisableAll && this.options.onDisableAll.call(this);
        callback && callback.call(this);
    }
    enable(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj为空则不再执行,0除外
        if (obj !== 0 && axIsEmpty(obj)) {
            return false;
        }
        let enableFun = (k) => {
            let item = axFindItem(k, this.data);
            //找到元素并设置禁用
            if (item) {
                this.workToggle(item, 'enable');
            }
        }
        //设置
        if (Array.isArray(obj)) {
            obj.forEach(k => {
                enableFun(k);
            });
        } else {
            enableFun(obj);
        }
        //监听回调
        let enableds = this.data.filter(k => !k.disabled);
        'enable' in this.handlers ? this.emit('enable', enableds) : null;
        this.options.onEnable && this.options.onEnable.call(this, enableds);
        callback && callback.call(this, enableds);
    }
    enableAll(callback) {
        this.enable(this.data);
        //监听回调
        'enableAll' in this.handlers ? this.emit('enableAll', '') : null;
        this.options.onEnableAll && this.options.onEnableAll.call(this);
        callback && callback.call(this);
    }
    updateContent(tab, content, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //tab可是数字、字符串、DOM和{}对象
        let me = axFindItem(tab, this.data);
        if (me) {
            //如果还在内存中则忽略
            // if (!axIsVnode(me.bodyDom)) {
            //先清空内容
            //me.content = '';
            me.bodyDom.innerHTML = '';
            //更新content
            content ? me.content = content : null;
            if (me.type === 'url') {
                //url类型
                me.bodyDom.src = content;
            } else if (me.type === 'async') {
                //异步类型
                this.asyncActive(me);
            } else if (axType(me.content).includes('HTML')) {
                //content是一个节点
                me.bodyDom.appendChild(me.content);
            } else {
                //字符串
                me.bodyDom.innerHTML = me.content;
            }
            // }
            //保存数据
            this.save();
        }
        //回调
        'updateContent' in this.handlers ? this.emit('updateContent', me) : null;
        this.options.onUpdateContent && this.options.onUpdateContent.call(this, me);
        callback && callback.call(this, me);
    }
    active(tab, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj可以是对象，字符串或数字
        let findMethod = axType(tab).includes('HTML') && tab.hasAttribute('label') ? axFindItem(tab, this.data, { dom: 'labelDom' }) : axFindItem(tab, this.data),
            me = findMethod, others;
        if (!axIsEmpty(me)) {
            //自己
            !me.active ? me.active = true : null;
            //追加焦点属性
            me.headerDom.setAttribute('active', '');
            me.bodyDom.setAttribute('active', '');
            //如果还是虚拟节点则append，如果不是则不用管
            axIsVnode(me.bodyDom) ? this.bodyWrapperDom.appendChild(me.bodyDom) : null;
            //如果是异步则获取内容
            me.type === 'async' ? this.asyncActive(me) : null;
            //其他兄弟节点
            others = this.data.filter(k => k !== me);
            others.forEach(k => {
                k.active ? delete k.active : null;
                k.headerDom.removeAttribute('active');
                k.bodyDom.removeAttribute('active');
            });
            //保存数据
            this.save();
        }
        //监听回调
        'active' in this.handlers ? this.emit('active', me) : null;
        this.options.onActive && this.options.onActive.call(this, me);
        callback && callback.call(this, me);
    }
    add(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //data为空则不再执行
        if (axIsEmpty(data)) {
            return false;
        }
        //当前data的length大于等于允许的max值则不再允许添加了
        if (this.data.length >= this.options.max && this.options.max !== 0) {
            console.warn('The number of tabs exceeds the maximum!');
            return false;
        }
        let type = axType(data),
            items = [],
            appendFun = (obj) => {
                if (this.data.findIndex(k => k.label === obj.label) === -1) {
                    //如果没有相同的头部label，才append
                    this.appendItem(obj, () => {
                        this.data.push(obj);
                        items.push(obj);
                        //slider实例追加
                        this.sliderIns ? this.sliderIns.add(obj.headerDom) : null;
                    });
                }
            };
        if (type === 'Object') {
            //data是一个对象
            appendFun(data);
        } if (type === 'Array' && data.length > 0) {
            //data是一个数组
            data.forEach(k => {
                appendFun(k);
            });
        }
        //items有至少一项才执行
        if (items.length > 0) {
            //是否允许新增后成为焦点，让最后一个成为焦点
            if (this.options.addToActive) {
                //激活最后一个元素
                this.active(items[items.length - 1]);
                //滚动到最后一个元素
                this.sliderIns ? this.sliderIns.toBullet(items.length - 1) : null;
            }
            //生成了数据和节点之后操作
            this.renderFinish();
            //如果有菜单则重置菜单
            this.dropdown ? this.dropdown.updateContent(this.getDropDownData()) : null;
        }
        //回调监听
        'add' in this.handlers ? this.emit('add', items) : null;
        this.options.onAdd && this.options.onAdd.call(this, items);
        callback && callback.call(this, items);
        return this;
    }
    remove(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //data为空则不再执行
        if (axIsEmpty(data)) {
            return false;
        }
        //数据为空了就不能再删除了
        if (this.data.length === 0) {
            console.warn('The data is already empty!');
            return false;
        }
        let removeItem = (tab) => {
            let item = axFindItem(tab, this.data);
            if (item) {
                //因为remove按钮在headerDom里，点击remove会触发active事件，所以先移除
                !item.labelDom ? item.headerDom.removeEventListener(this.options.trigger, this.triggerShow, false) : null;
                //再移除两个节点
                item.headerDom.remove();
                item.bodyDom.remove();
                //更新this.data
                this.data = this.data.filter(k => k !== item);
                //slider实例删除
                this.sliderIns ? this.sliderIns.remove(item.headerDom) : null;
            }
        };
        //删除
        if (Array.isArray(data)) {
            data.forEach(k => {
                removeItem(k);
            });
        } else {
            removeItem(data);
        }
        //保存数据
        this.save();
        //如果有菜单则重置菜单
        this.dropdown ? this.dropdown.updateContent(this.getDropDownData()) : null;
        //回调监听
        'remove' in this.handlers ? this.emit('remove', '') : null;
        this.options.onRemove && this.options.onRemove.call(this);
        callback && callback.call(this);
        return this;
    }
    save(props, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }
        setTimeout(() => {
            let idsActive = this.getIndexActive(),
                idsDisabled = this.data.filter(k => k.disabled).map(k => k.id).filter(Boolean);
            if (!props) {
                //保存当前焦点索引
                axLocalStorage.set(this.options.storageName, { active: idsActive, disabled: idsDisabled, content: this.data });
            } else {
                //保存参数,props={}
                !props.hasOwnProperty('active') ? props.active = idsActive : null;
                !props.hasOwnProperty('disabled') ? props.disabled = idsDisabled : null;
                !props.hasOwnProperty('content') ? props.content = this.data : null;
                axLocalStorage.set(this.options.storageName, props);
            }
            //回调监听
            let getValue = axLocalStorage.get(this.options.storageName);
            'save' in this.handlers ? this.emit('save', getValue) : null;
            this.options.onSave && this.options.onSave.call(this, getValue);
            callback && callback.call(this, getValue);
            return this;
        }, 0)
    }

    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.options = axExtend(this.options, setting);
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
        this.init();
        //监听回调
        'update' in this.handlers ? this.emit('update', '') : null;
        this.options.onUpdate && this.options.onUpdate.call(this);
        callback && callback.call(this);
        return this;
    }
    destroy(callback) {
        //取消绑定事件
        this.data.forEach(k => {
            let triggerDom = k.labelDom ? k.labelDom : k.headerDom;
            triggerDom.removeEventListener(this.options.trigger, this.triggerShow);
            k.removeDom ? k.removeDom.onclick = null : null;
            k.updateDom ? k.updateDom.onclick = null : null;
        });
        this.sliderIns ? this.sliderIns.destroy() : null;
        this.dropdown ? this.dropdown.popup.destroy() : null;
        //更新销毁状态
        this.destroyed = true;
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
        //监听回调
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
//对使用axTab属性的元素使用本插件
/* document.querySelectorAll('[axTab]').forEach(element => {
    new axTab(element);
});
 */
/*!
* 插件：抽屉侧边窗；使用方法：new axDrawer({参数},'#id')；#id是可选参数，可以从该dom获取参数值
*/
class axDrawer {
    constructor(options, elem) {
        this.options = axExtend({
            insName: '',//实例名称，对axInstance有用
            title: '',//窗口标题，为空等于不显示
            titleBreak: true,//标题分割线，默认显示
            overlay: true,//默认显示半透明遮罩
            overlayClose: true,//点击遮罩是否关闭，默认可关闭
            duration: 0,//弹出和关闭窗口动画的时间，如果为0则使用css中定义的动画时间，单位ms
            placement: 'right',//抽屉的位置，默认在右侧，可填left|right|top|bottom
            size: 'sm',//窗口尺寸，默认小尺寸，可填sm|middle|large|full
            padding: 'ax-padding',//窗口内边距，值可以是布尔值也可以是类名，默认带边距，可填false|true|ax-padding|ax-padding ax-xxl|和其他
            offset: '100px', //子层抽屉偏移的基准距离，单位只能用px
            content: '',//窗口内容，可填#id或html,
            type: 'html',//内容的类型，默认是html，可填iframe|async|html
            url: '',//内容的来源，该参数依赖type参数，除了html类型，其他几种类型均需要填入url，async的url可以是“地址+#id”
            btnClass: 'ax-opened',//触发按钮在打开抽屉后新增的样式名
            rel: '',//可以触发本窗口的元素，可以是#id，节点对象，或数组（多个按钮均触发）
            close: true,//关闭窗口的按钮，默认显示
            footerType: 'center',//底部按钮风格，可选center,line,right
            footerShow: true,//是否显示底部，默认显示
            footerBreak: false,//底部分割线，默认不显示
            note: '',//其他内容
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            zIndex: 0,//气泡层级，默认为0即不设置，如果与其他组件发生遮挡请自行设置
            onInit: '',//回调函数，初始化后执行，无参数
            onShow: '',//回调函数，开始显示后执行，无参数
            onShown: '',//回调函数，完成显示后执行，无参数
            onHide: '',//回调函数，开始关闭后执行，无参数
            onHidden: '',//回调函数，完全关闭后执行，无参数
            onUpdate: '',//回调函数，参数/内容更新前执行，无参数
            onUpdated: '',//回调函数，参数更新后执行，支持参数isUpdated(是否更新了：true或false)、before（更新前参数）、after（更新后参数）
            onUpdateContent: '',//回调函数，内容更新后执行，支持参数isUpdated(是否更新了：true或false)、before（更新前内容）、after（更新后内容）
            onDestroy: '',//回调函数，销毁后执行，无参数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            confirm: {
                text: '确定',//确认按钮文字
                classname: '',//确认按钮的类名
                icon: '',//确认按钮的图标，格式是<i class="ax-iconfont ax-icon-edit"></i>
                type: 'close',//按钮类型，可选择close关闭、async异步提交表单、submit直接提交表单、prevent阻止按钮默认事件什么事也不做
                url: '',//提交的地址，如果type为async则必须使用该参数
                callback: ''//确认按钮的回调函数，最终按函数方式使用，格式是function(button,content){}，button是按钮节点，content是返回内容
            },
            cancel: {
                text: '关闭',//取消按钮文字
                classname: '',//取消按钮的类名
                icon: '',//取消按钮的图标
                callback: ''//取消按钮的回调函数，格式是function(button,content){}，button是按钮节点，content是返回内容
            },
            diy: {
                text: '',//自定义按钮文字
                classname: '',//自定义按钮的类名
                icon: '',//自定义按钮的图标
                url: '',//自定义按钮的超链接
                target: '',//自定义按钮的链接目标，一般是_blank或者_self
                callback: ''//自定义按钮的回调函数，格式是function(button,content){}，button是按钮节点，content是返回内容
            }
        }, options, elem, this.constructor.name);
        //定义监听事件成员
        this.handlers = {};
        //如果使用表单
        this.form = null;
        //获得button变量;
        elem ? this.button = axIdToDom(elem) : null;
        //定义显示状态
        this.state = 'hidden';
        //逐个显示
        this.sequenceShow = (dom, direction) => {
            if (!dom.style[direction]) {
                //没有style属性，说明同方向是第一个打开的
                dom.style[direction] = this.options.offset
            } else {
                //有style属性，说明多个drawer层叠了
                dom.style[direction] = parseInt(dom.style[direction]) * 0.6 + parseInt(this.options.offset) + 'px';
            }
        }
        //逐个关闭
        this.sequenceHide = (dom, direction) => {
            if (!dom.style[direction]) {
                //没有style属性，说明同方向是第一个打开的
                let prop = direction === 'left' || direction === 'right' ? 'width' : 'height';
                dom.style[direction] = '-' + getComputedStyle(dom)[prop];
            } else {
               //有style属性，说明多个drawer层叠了
                if (dom.style[direction] == this.options.offset) {
                    //与偏移值相等的drawer是层叠的第二个
                    dom.removeAttribute('style');
                } else {
                    //层叠的第三个或以上
                    dom.style[direction] = (parseInt(dom.style[direction]) - parseInt(this.options.offset)) / 0.6 + 'px';
                }
            }
        }
        //初始化
        this.init();
        return this;
    }
    async init() {
        //添加到实例合集并设置销毁状态
        axInstance.push(this, this.options.insName, 'drawer');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
         //设置销毁状态
        this.destroyed = false;

        //定义关联按钮
        this.rel = axAllToDoms(this.options.rel);

        //创建窗口模板
        this.renderDom();
        //设置data属性
        this.setAttribute();
        //获取内容
        if (this.options.content.nodeType === 1) {
            //如果content是Dom节点，则整体装入节点的内容
            this.content.innerHTML = '';
            this.content.appendChild(this.options.content);
            //定义form节点，可能为null
            this.form = this.content.querySelector('form');
        } else if (this.options.type === 'form' && typeof this.options.content === 'string' && this.options.content.substr(0, 1) === '#') {
            this.form = document.querySelector(this.options.content).querySelector('form');
            if (this.form) {
                //如果类型是form节点，则整体装入节点的内容
                this.content.innerHTML = '';
                this.content.appendChild(this.form);
            } else {
                console.error('Form node is required!');
                return false;
            }
        } else {
            let response;
            if (this.options.type === 'async' && this.options.url) {
                //异步文本
                //content是文本
                response = await axCreateContent(this.options, this);
            } else {
                //同步是文本
                response = axCreateContent(this.options, this);
            }
            //填充内容
            if (typeof this.options.content === 'function') {
                //如果content是函数
                this.content.innerHTML = this.options.content.call(this,response.content);
            } else {
                this.content.innerHTML = response.content;
            }
            //定义form节点，可能为null
            this.form = this.content.querySelector('form');
        }
        //设置底部
        this.footer.innerHTML = axFooterBtns(this.options, this);
        //按钮操作
        this.targetDom.querySelectorAll('[confirm]').forEach(elem => {
            elem.onclick = (e)=> {
                if (this.options.confirm.type == 'submit') {
                    if (this.form && this.options.type == 'form') {
                        //校验通过后执行
                        axValidRepeat(e, this.form, () => {
                            this.form.submit();
                        });
                        //执行回调
                        this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                    } else {
                        //报错提示
                        console.error('Form node is required and type of options must be form string!');
                        return false;
                    }
                } else if (this.options.confirm.type == 'async') {
                    if (this.form && this.options.type == 'form' && this.options.confirm.url) {
                        //校验通过后执行
                        axValidRepeat(e, this.form, () => {
                            //使用表单异步提交函数
                            axAjaxSubmit(this.options.confirm.url, this.form, elem, {
                                type: this.options.ajaxType,
                                opened: (response) => {
                                    this.submitXhr = response.xhr;
                                },
                                success: (response) => {
                                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, response.content);
                                }
                            });
                        })
                    } else {
                        //报错提示
                        console.error('Form node is required & type of options must be form string & url of async is required!');
                        return false;
                    }
                } else if (this.options.confirm.type == 'prevent') {
                    axPreventDefault(e);
                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                } else {
                    this.hide();
                    //执行回调
                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                }

            };
        });
        this.targetDom.querySelectorAll('[diy]').forEach(elem => {
            elem.onclick =  ()=> {
                this.options.diy.callback && this.options.diy.callback.call(this, elem);
            };
        });
        this.targetDom.querySelectorAll('[cancel]').forEach(elem => {
            elem.onclick =  ()=> {
                this.hide();
                this.options.cancel.callback && this.options.cancel.callback.call(this, elem);
            };
        });
        this.targetDom.querySelectorAll('[close]').forEach(elem => {
            elem.addEventListener('click',  ()=> {
                this.hide();
            }, false);
        });

        //点击按钮打开
        if (this.button) {
            //用于绑定和解绑button
            this.button.onclick = () => {
                this.show();
            }
        }
        //其他按钮也可打开
        if (this.rel.length > 0) {
            this.rel.forEach(item => {
                item.onclick = () => {
                    this.show();
                }
            });
        }
        //回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    createTemplate() {
        return `
                    <div class="ax-drawer" data-overlay="true">
                        <# if(this.overlay){ #><div class="ax-drawer-overlay" <# if(this.overlayClose){ #>close<# } #>></div><# } #>
                        <div class="ax-drawer-wrapper">
                            <# if(this.close){ #> <a href="###" class="ax-drawer-close" close><i class="ax-iconfont ax-icon-close"></i></a><# } #>
                            <# if(this.title){ #><div class="ax-drawer-header"><div class="ax-col ax-align-center"><# this.title #></div></div><# } #>
                            <# if(this.title && this.titleBreak){ #><div class="ax-break-line"></div><# } #>
                            <div class="ax-drawer-body">
                                <# if(this.padding){ #><div class="<# this.padding #>" <# if(this.type == 'iframe'){ #>style="height:100%;"<# } #>> <# } #>
                                    <div class="ax-drawer-content" <# if(this.type == 'iframe'){ #>style="height:100%;"<# } #>></div>
                                <# if(this.padding){ #></div><# } #>
                            </div>
                            <div class="ax-drawer-footer">
                            </div>
                        </div>
                    </div>
                  `;
    }
    renderDom() {
        this.template = axTplEngine(this.createTemplate(), this.options);
        //将字符串转成DOM对象
        this.targetDom = axStrToDom(this.template);
        //定义抽屉Dom元素
        this.wrapper = this.targetDom.querySelector('.ax-drawer-wrapper');
        this.targetDom.querySelector('.ax-drawer-overlay') ? this.overlay = this.targetDom.querySelector('.ax-drawer-overlay') : null;
        this.content = this.targetDom.querySelector('.ax-drawer-content');
        this.body = this.targetDom.querySelector('.ax-drawer-body');
        this.targetDom.querySelector('.ax-drawer-footer') ? this.footer = this.targetDom.querySelector('.ax-drawer-footer') : null;
        //获得实例的css值
        this.width = window.getComputedStyle(this.wrapper).width;
        this.height = window.getComputedStyle(this.wrapper).height;
        this.left = window.getComputedStyle(this.wrapper).left;
        this.right = window.getComputedStyle(this.wrapper).right;
        this.top = window.getComputedStyle(this.wrapper).top;
        this.bottom = window.getComputedStyle(this.wrapper).bottom;

    }
    setAttribute() {
        //设置窗口位置
        if (this.options.placement == "left") {
            this.targetDom.setAttribute('data-placement', 'left');
        } else if (this.options.placement == "right") {
            this.targetDom.setAttribute('data-placement', 'right');
        } else if (this.options.placement == "top") {
            this.targetDom.setAttribute('data-placement', 'top');
        } else if (this.options.placement == "bottom") {
            this.targetDom.setAttribute('data-placement', 'bottom');
        } else {
            return;
        }
        //设置窗口尺寸
        if (this.options.size == "sm") {
            this.targetDom.setAttribute('data-size', 'sm');
        } else if (this.options.size == 'md') {
            this.targetDom.setAttribute('data-size', 'md');
        } else if (this.options.size == 'lg') {
            this.targetDom.setAttribute('data-size', 'lg');
        } else {
            return;
        }
        //设置zIndex
        this.options.zIndex ? this.targetDom.style.zIndex = this.options.zIndex : null;
    }
    show(callback) {
        //如果已经销毁或本身就是打开状态则不再执行
        if (this.destroyed || this.state !== 'hidden') {
            return this;
        }
        //更新状态
        this.state = 'ing';
        //在准备Dom和内容之后创建html节点
        document.body.appendChild(this.targetDom);
        //设置打开关闭动画时间
        if (this.options.duration) {
            this.duration = this.options.duration;
        } else {
            this.duration = parseFloat(axStyle(this.wrapper).animationDuration) * 1000
        }
      
        //根据方向打开抽屉
        axInstance.data.filter(i => i.type === 'drawer' && !i.instance.destroyed).forEach(item => {
            if (item.instance.state === 'shown' && item.instance.options.placement === this.options.placement) {
                this.sequenceShow(item.instance.wrapper, this.options.placement);
            }
        });

        this.targetDom.classList.add("ax-show");
        //如果使用按钮触发抽屉，那么给抽屉赋予样式
        this.button ? this.button.classList.add(this.options.btnClass) : null;

          //回调
          this.options.onShow && this.options.onShow.call(this);
          'show' in this.handlers ? this.emit('show', '') : null;
          
        //完全打开执行
        setTimeout(() => {
            //更新显示状态
            this.state = 'shown';
            //回调
            this.options.onShown && this.options.onShown.call(this);
            'shown' in this.handlers ? this.emit('shown', '') : null;
        }, this.duration);

        callback && callback.call(this);
        return this;
    }
    hide(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.state !== 'shown') {
            return this;
        }
        //更新状态
        this.state = 'ing';
        //回调
        this.options.onHide && this.options.onHide.call(this);
        'hide' in this.handlers ? this.emit('hide', '') : null;
        //如果使用按钮触发对话框，那么删除样式
        this.button ? this.button.classList.remove(this.options.btnClass) : null;
        //定义隐藏抽屉函数

        this.overlay ? this.overlay.style.opacity = 0 : null;
        this.sequenceHide(this.wrapper, this.options.placement);

        setTimeout(() => {
            //更新显示状态
            this.state = 'hidden';
            this.targetDom.classList.remove("ax-show");
            this.overlay ? this.overlay.removeAttribute("style") : null;
            this.wrapper.removeAttribute("style");
            //移除节点
            this.targetDom.remove();
            //回调
            this.options.onHidden && this.options.onHidden.call(this);
            'hidden' in this.handlers ? this.emit('hidden', '') : null;
        }, this.duration);


        //其他已经显示的抽屉自动同向减少固定偏移值
        axInstance.data.filter(i => i.type === 'drawer' && !i.instance.destroyed).forEach(item => {
            if (item.instance.state === 'shown' && item.instance.options.placement === this.options.placement && item.instance !== this) {
                this.sequenceHide(item.instance.wrapper, this.options.placement);
            }
        });

        this.targetDom.querySelectorAll('video,audio').forEach(elem => {
            elem.pause();
        });
        //如果使用按钮触发抽屉，那么删除样式
        this.button ? this.button.classList.remove(this.options.btnClass) : null;

        callback && callback.call(this);
        return this;
    }
    destroy(callback) {
        //删除对应的DOM
        this.targetDom.remove();
        //取消绑定事件
        this.button.onclick = null;
        this.rel.forEach(i => {
            i.onclick = null;
        });
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        this.submitXhr ? this.submitXhr.abort() : null;
        //监听回调
        this.options.onDestroy && this.options.onDestroy.call(this);
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        callback && callback.call(this);
        return this;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }

        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        let targetDom = this.targetDom,
            before = axClone(this.options),
            after = axExtend(this.options, setting),
            isUpdated = false;
        if (setting && JSON.stringify(after) != JSON.stringify(before)) {
            this.options = after;
            //更新后保持原有的打开状态
            if (this.state === 'shown') {
                this.init().show();
            } else if (this.state === 'hidden') {
                this.init();
            }
            //更新后删除原节点
            targetDom.remove();
            isUpdated = true;
        }
        callback && callback.call(this, isUpdated, before, after);
        this.options.onUpdated && this.options.onUpdated.call(this, isUpdated, before, after);
        'updated' in this.handlers ? this.emit('updated', isUpdated, before, after) : null;
        return this;
    }
    updateContent(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }

        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        let before = this.content.innerHTML,
            isUpdated = false;
        //
        if (axType(data) === 'String') {
            if (axStrType(data)) {
                //能识别出节点
                this.content.innerHTML = axIdToDom(data) ? axIdToDom(data).innerHTML : '';
            } else {
                //纯文本
                this.content.innerHTML = data;
            }
        } else if (axType(data).includes('HTML')) {
            //是节点
            this.content.innerHTML = '';
            this.content.appendChild(data);
        } else {
            this.content.innerHTML = `No data to update!`;
        }
        //
        if (this.content.innerHTML == before) {
            isUpdated = false;
        } else {
            isUpdated = true;
        }
        callback && callback.call(this, isUpdated, before, this.content.innerHTML);
        this.options.onUpdateContent && this.options.onUpdateContent.call(this, isUpdated, before, this.content.innerHTML);
        'updateContent' in this.handlers ? this.emit('updateContent', isUpdated, before, this.content.innerHTML) : null;
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    //对使用axDrawer属性的元素使用本插件，可窗口嵌套
    function eachInstance(arr, parent) {
        arr.forEach(elem => {
            let instance = new axDrawer('', elem);
            parent ? instance.targetDom.style.zIndex = parseInt(getComputedStyle(parent.targetDom).zIndex) + 1 : null;
            eachInstance(instance.targetDom.querySelectorAll('[axDrawer]'), instance);
        });
    }
    let arrInstance = document.querySelectorAll('[axDrawer]');
    eachInstance(arrInstance);
})();
 */
/*!
 * 插件：对话弹窗；使用方法：new axDialog({参数},'#id')；#id是可选参数，可以从该dom获取参数值
 */
class axDialog {
    constructor(options, elem) {
        this.options = axExtend({
            insName: '',//实例名称，对axInstance有用
            title: '',//窗口标题，为空等于不显示
            titleBreak: false,//标题分割线，默认不显示
            overlay: true,//默认显示半透明遮罩
            overlayClose: true,//点击遮罩是否关闭，默认可关闭
            duration: 0,//弹出和关闭窗口动画的时间，如果为0则使用css中定义的动画时间，单位ms
            placement: 'center',//对话框的位置，默认在居中弹出，可填left-top|center-top|right-top|left-bottom|center-bottom|right-bottom
            in: 'springIn',//可填fadeIn,fadeInDown,fadeInUp,slideDown,fadeInLeft,fadeInRight,springIn,scaleIn
            out: 'scaleOut',//可填fadeOut,fadeOutDown,fadeOutUp,slideUp,fadeOutLeft,fadeOutRight,springOut,scaleOut
            multiple: true,//是否允许打开多个窗口，默认允许
            size: 'md',//窗口尺寸，默认正常尺寸，可填sm,middle,large,full
            padding: 'ax-padding',//窗口内边距，值可以是布尔值也可以是类名，默认带边距，可填false,true,ax-padding,ax-padding ax-xxl,...
            drag: false,//是否能拖拽，默认不可拖拽；注意如果title为空时即使drag为true也不能拖拽
            content: '',//窗口内容，可填#id或文本内容,
            type: 'html',//内容的类型，默认是html，可填picture|video|audio|iframe|async|html|form
            url: '',//内容的来源，该参数依赖type参数，除了html类型，其他几种类型均需要填入url，async的url可以是“地址+#id”
            btnClass: 'ax-opened',//触发按钮在打开气泡后新增的样式名
            rel: '',//可以触发本窗口的元素，可以是#id，节点对象，或数组（多个按钮均触发）
            name: '',//内容名称，当type是picture|video|audio时有效
            width: '',//内容宽度，当type是iframe的时候，有必要设置该参数，比如800px
            height: '',//内容高度，当type是iframe的时候，有必要设置该参数，比如400px，但是不要超过600px否则会产生垂直滚动条
            close: true,//关闭窗口的按钮，默认显示
            closePlace: 'inside',//可选 inside,bottom,outside
            footerShow: true,//是否显示底部，默认显示
            footerType: 'center',//底部按钮风格，可选center,line,right
            footerBreak: false,//底部分割线，默认不显示
            note: '',//其他内容
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            zIndex: 0,//气泡层级，默认为0即不设置，如果与其他组件发生遮挡请自行设置
            onDrag: '',//回调函数，拖拽开始执行，无参数
            onDragged: '',//回调函数，拖拽完成执行，无参数
            onInit: '',//回调函数，初始化后执行，无参数
            onShow: '',//回调函数，开始显示后执行，无参数
            onShown: '',//回调函数，完成显示后执行，无参数
            onHide: '',//回调函数，开始关闭后执行，无参数
            onHidden: '',//回调函数，完全关闭后执行，无参数
            onUpdate: '',//回调函数，参数/内容更新前执行，无参数
            onUpdated: '',//回调函数，参数更新后执行，支持参数isUpdated(是否更新了：true或false)、before（更新前参数）、after（更新后参数）
            onUpdateContent: '',//回调函数，内容更新后执行，支持参数isUpdated(是否更新了：true或false)、before（更新前内容）、after（更新后内容）
            onDestroy: '',//回调函数，销毁后执行，无参数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            confirm: {
                text: '确定',//确认按钮文字
                classname: '',//确认按钮的类名
                icon: '',//确认按钮的图标，格式是<i class="ax-iconfont ax-icon-edit"></i>
                type: 'close',//按钮类型，可选择close关闭、async异步提交表单、submit直接提交表单、prevent阻止按钮默认事件什么事也不做
                url: '',//提交的地址，如果type为async则必须使用该参数
                callback: ''//确认按钮的回调函数，最终按函数方式使用，格式是function(button,content){}，button是按钮节点，content是返回内容
            },
            cancel: {
                text: '关闭',//取消按钮文字
                classname: '',//取消按钮的类名
                icon: '',//取消按钮的图标
                callback: ''//取消按钮的回调函数，格式是function(button,content){}，button是按钮节点，content是返回内容
            },
            diy: {
                text: '',//自定义按钮文字
                classname: '',//自定义按钮的类名
                icon: '',//自定义按钮的图标
                url: '',//自定义按钮的超链接
                target: '',//自定义按钮的链接目标，一般是_blank或者_self
                callback: ''//自定义按钮的回调函数，格式是function(button,content){}，button是按钮节点，content是返回内容
            }
        }, options, elem, this.constructor.name);
        //定义节点成员
        this.wrapper = '';
        this.content = '';
        this.overlay = '';
        //定义监听事件成员
        this.handlers = {};
        //定义位置成员
        this.wrapPosition = '';
        this.wrapLeft = '';
        this.wrapTop = '';
        //如果使用表单
        this.form = null;
        //获得按钮节点
        elem ? this.button = axIdToDom(elem) : null;
        //显示状态
        this.state = 'hidden';
        //绑定事件状态
        //初始化
        this.init();
        return this;
    }
    async init() {
        //添加到实例合集并设置销毁状态
        axInstance.push(this, this.options.insName, 'dialog');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
         //设置销毁状态
        this.destroyed = false;
        //定义关联按钮
        this.rel = axAllToDoms(this.options.rel);

        //手机端都是居中弹出
        axClient() === 'phone' ? this.options.placement = 'center' : null;

        //根据模板初步渲染
        this.renderDom();
        //data-*属性赋值
        this.setAttribute();

        //获取内容
        if (this.options.content.nodeType === 1) {
            //如果content是Dom节点，则整体装入节点的内容
            this.content.innerHTML = '';
            this.content.appendChild(this.options.content);
            //定义form节点，可能为null
            this.form = this.content.querySelector('form');
        } else if (this.options.type === 'form' && typeof this.options.content === 'string' && this.options.content.substr(0, 1) === '#') {
            this.form = document.querySelector(this.options.content).querySelector('form');
            if (this.form) {
                //如果类型是form节点，则整体装入节点的内容
                this.content.innerHTML = '';
                this.content.appendChild(this.form);
            } else {
                console.error('Form node is required!');
                return false;
            }
        } else {
            let response;
            if (this.options.type === 'async' && this.options.url) {
                //异步文本
                //content是文本
                response = await axCreateContent(this.options, this);
            } else {
                //同步是文本
                response = axCreateContent(this.options, this);
            }
            //填充内容
            if (typeof this.options.content === 'function') {
                //如果content是函数
                this.content.innerHTML = this.options.content.call(this, response.content);
            } else {
                this.content.innerHTML = response.content;
            }
            //定义form节点，可能为null
            this.form = this.content.querySelector('form');
        }
        //设置底部
        this.footer.innerHTML = axFooterBtns(this.options, this);
        //设置拖拽方法
        if (this.options.drag) {
            this.drag();
        }
        //按钮操作
        this.targetDom.querySelectorAll('[confirm]').forEach(elem => {
            elem.onclick = (e) => {
                if (this.options.confirm.type == 'submit') {
                    if (this.form && this.options.type == 'form') {
                        //校验通过后执行
                        axValidRepeat(e, this.form, () => {
                            this.form.submit();
                        });
                        //执行回调
                        this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                    } else {
                        //报错提示
                        console.error('Form node is required and type of options must be form string!');
                        return false;
                    }
                } else if (this.options.confirm.type == 'async') {
                    if (this.form && this.options.type == 'form' && this.options.confirm.url) {
                        //校验通过后执行
                        axValidRepeat(e, this.form, () => {
                            //使用表单异步提交函数
                            axAjaxSubmit(this.options.confirm.url, this.form, elem, {
                                type: this.options.ajaxType,
                                opened: (response) => {
                                    this.submitXhr = response.xhr;
                                },
                                success: (response) => {
                                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, response.content);
                                }
                            });
                        })
                    } else {
                        //报错提示
                        console.error('Form node is required & type of options must be form string & url of async is required!');
                        return false;
                    }
                } else if (this.options.confirm.type == 'prevent') {
                    axPreventDefault(e);
                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                } else {
                    this.hide();
                    //执行回调
                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                }

            };

        });
        this.targetDom.querySelectorAll('[diy]').forEach(elem => {
            elem.onclick =  ()=> {
                this.options.diy.callback && this.options.diy.callback.call(this, elem);
            };
        });
        this.targetDom.querySelectorAll('[cancel]').forEach(elem => {
            elem.onclick =  ()=> {
                this.hide();
                this.options.cancel.callback && this.options.cancel.callback.call(this, elem);
            };
        });
        this.targetDom.querySelectorAll('[close]').forEach(elem => {
            elem.addEventListener('click',  ()=> {
                this.hide();
            }, false);
        });

        //点击按钮打开
        if (this.button) {
            //用于绑定和解绑button
            this.button.onclick = () => {
                this.show();
            }
        }

        //其他按钮也可打开
        if (this.rel.length > 0) {
            this.rel.forEach(item => {
                item.onclick = () => {
                    this.show();
                }
            });
        }

        //回调
        this.options.onInit && this.options.onInit.call(this);
        //返回实例，以便于链式操作
        return this;
    }
    createTemplate() {
        return `
        <div class="ax-dialog" data-overlay="true">
        <# if(this.overlay){ #><div class="ax-dialog-overlay" <# if(this.overlayClose){ #>close<# } #>></div><# } #>
            <div class="ax-dialog-wrapper">
                <# if(this.close){ #><a href="###" class="ax-dialog-close" close><i class="ax-iconfont ax-icon-close"></i></a><# } #>
                <# if(this.title){ #><div class="ax-dialog-header"><# this.title #></div><# } #>
                <# if(this.titleBreak){ #><div class="ax-break-line"></div><# } #>
                <div class="ax-dialog-body">
                    <# if(typeof this.padding == 'string'){ #><div class="<# this.padding #>"> <# }else if(this.padding){ #><div class="ax-padding"> <# } #>
                    <div class="ax-dialog-content"></div>
                    <# if(this.padding){ #></div><# } #>
                </div>
                <div class="ax-dialog-footer">
                </div>
            </div>
        </div>
      `
    }
    renderDom() {
        this.template = axTplEngine(this.createTemplate(), this.options);
        //将字符串转成DOM对象
        this.targetDom = axStrToDom(this.template);
        //获取Dom树
        this.wrapper = this.targetDom.querySelector('.ax-dialog-wrapper');
        this.header = this.targetDom.querySelector('.ax-dialog-header');
        this.content = this.targetDom.querySelector('.ax-dialog-content');
        this.body = this.targetDom.querySelector('.ax-dialog-body');
        this.overlay = this.targetDom.querySelector('.ax-dialog-overlay');
        this.footer = this.targetDom.querySelector('.ax-dialog-footer');
    }
    setAttribute() {
        this.options.size ? this.targetDom.setAttribute('data-size', this.options.size) : null;
        this.options.footerType ? this.targetDom.setAttribute('data-footerType', this.options.footerType) : null;
        this.options.height ? this.content.style.height = this.options.height : null;
        this.oldStyle = this.options.width ? `width:${this.options.width};` : null;
        //根据内容创建data-type属性
        this.options.type ? this.targetDom.setAttribute('data-type', this.options.type) : null;
        //根据overlay参数给窗口创建data-overlay属性，设定为false
        !this.options.overlay ? this.targetDom.setAttribute('data-overlay', 'false') : null;
        //设定关闭按钮位置
        if (!this.options.title && this.options.close) {
            if (this.options.size == 'lg') {
                this.targetDom.setAttribute('data-closePlace', 'bottom');
            } else if (!this.options.title && this.options.closePlace == 'inside') {
                this.targetDom.setAttribute('data-closePlace', 'outside');
            } else {
                this.targetDom.setAttribute('data-closePlace', this.options.closePlace)
            }
        } else if (this.options.closePlace && this.options.close) {
            this.targetDom.setAttribute('data-closePlace', this.options.closePlace)
        }
        //设置位置属性
        this.targetDom.setAttribute('placement', this.options.placement);
        //设置zIndex
        this.options.zIndex ? this.targetDom.style.zIndex = this.options.zIndex : null;
    }
    drag(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return false;
        }
        if (this.options.title) {
            let _this = this,
                header = this.targetDom.querySelector('.ax-dialog-header');
            this.targetDom.setAttribute('data-drag', this.options.drag);
            header.onmousedown = function (e) {
                _this.options.onDrag && _this.options.onDrag.call(this);
                'drag' in _this.handlers ? _this.emit('drag', '') : null;
                this.style.cursor = 'move';
                let moveDiv = this.parentNode;
                moveDiv.style.cssText += 'position:fixed ;';
                //存储拖拽后的位置
                _this.wrapPosition = 'fixed';

                let x = e.clientX - moveDiv.offsetLeft;
                let y = e.clientY - moveDiv.offsetTop;

                document.onmousemove = function (ev) {
                    let _x = ev.clientX - x;
                    let _y = ev.clientY - y;
                    //设置边界并判断边界
                    let maxX = document.documentElement.clientWidth - moveDiv.clientWidth;
                    let maxY = document.documentElement.clientHeight - moveDiv.clientHeight;
                    if (_x < 0) {
                        _x = 0;
                    }
                    if (_x > maxX) {
                        _x = maxX;
                    }
                    if (_y < 0) {
                        _y = 0;
                    }
                    if (_y > maxY) {
                        _y = maxY;
                    }
                    moveDiv.style.left = _x + 'px';
                    moveDiv.style.top = _y + 'px';
                    //存储拖拽后的位置
                    _this.wrapLeft = moveDiv.style.left;
                    _this.wrapTop = moveDiv.style.top;
                };
            };
            header.onmouseup = function () {
                document.onmousemove = null;
                document.onmouseup = null;
                _this.options.onDragged && _this.options.onDragged.call(this);
                'dragged' in _this.handlers ? _this.emit('dragged', '') : null;
            }
            //禁止HTML5下的拖拽事件
            header.ondragstart = function (ev) {
                axPreventDefault(ev);
            };
            header.ondragend = function (ev) {
                axPreventDefault(ev);
            };
            callback && callback.call(this);
        } else {
            return false;
        }
        //返回实例，以便于链式操作
        return this;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }

        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        let targetDom = this.targetDom,
            before = axClone(this.options),
            after = axExtend(this.options, setting),
            isUpdated = false;
        if (setting && JSON.stringify(after) != JSON.stringify(before)) {
            this.options = after;
            //更新后保持原有的打开状态
            if (this.state === 'shown') {
                this.init().show();
            } else if (this.state === 'hidden') {
                this.init();
            }
            //更新后删除原节点
            targetDom.remove();
            isUpdated = true;
        }
        callback && callback.call(this, isUpdated, before, after);
        this.options.onUpdated && this.options.onUpdated.call(this, isUpdated, before, after);
        'updated' in this.handlers ? this.emit('updated', isUpdated, before, after) : null;
        return this;
    }
    updateContent(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        let before = this.content.innerHTML,
            isUpdated = false;

        //
        if (axType(data) === 'String') {
            if (axStrType(data)) {
                //能识别出节点
                this.content.innerHTML = axIdToDom(data) ? axIdToDom(data).innerHTML : '';
            } else {
                //纯文本
                this.content.innerHTML = data;
            }
        } else if (axType(data).includes('HTML')) {
            //是节点
            this.content.innerHTML = '';
            this.content.appendChild(data);
        } else {
            this.content.innerHTML = `No data to update!`;
        }
        //
        if (this.content.innerHTML == before) {
            isUpdated = false;
        } else {
            isUpdated = true;
        }
        callback && callback.call(this, isUpdated, before, this.content.innerHTML);
        this.options.onUpdateContent && this.options.onUpdateContent.call(this, isUpdated, before, this.content.innerHTML);
        'updateContent' in this.handlers ? this.emit('updateContent', isUpdated, before, this.content.innerHTML) : null;
        return this;
    }
    destroy(callback) {
        //删除对应的DOM
        this.targetDom.remove();
        //取消绑定事件
        this.button.onclick = null;
        this.rel.forEach(i => {
            i.onclick = null;
        });
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        this.submitXhr ? this.submitXhr.abort() : null;
        //监听回调
        this.options.onDestroy && this.options.onDestroy.call(this);
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        callback && callback.call(this);
        return this;
    }
    show(callback) {
        //如果已经销毁或本身就是打开状态则不再执行
        if (this.destroyed || this.state !== 'hidden') {
            return this;
        }
        //更新状态
        this.state = 'ing';
        //在准备Dom和内容之后创建html节点
        document.body.appendChild(this.targetDom);
        //获取内容部分的高度，用于slideDown和slideUp
        this.wrapperHeight = axArrHeight(this.wrapper).height;

        //设置打开关闭动画时间
        if (this.options.duration) {
            this.duration = this.options.duration;
        } else if (this.wrapperHeight) {
            this.duration = Math.floor(this.wrapperHeight / 3 + 250);
        } else {
            this.duration = parseFloat(axStyle(this.wrapper).animationDuration) * 1000
        }

        //如果参数multiple为false，则打开自己关闭其他实例
        if (!this.options.multiple) {
            let items = axInstance.data.filter(i => i.type.dialog && i.instance !== this && i.instance.isShow)
            items.forEach(i => {
                i.instance.hide();
            });
        }
        //执行窗口打开动画
        this.targetDom.classList.remove('ax-hide');
        this.targetDom.classList.add('ax-show');
        //如果使用按钮触发对话框，那么给气泡赋予样式
        this.button ? this.button.classList.add(this.options.btnClass) : null;
        
        //回调
        this.options.onShow && this.options.onShow.call(this);
        'show' in this.handlers ? this.emit('show', '') : null;
        
        //如果被拖拽过，则在关闭的地方打开
        let dragStyle = '';
        this.wrapPosition ? dragStyle = `position:${this.wrapPosition};left:${this.wrapLeft};top:${this.wrapTop};` : null;
        if (this.options.in == 'slideDown') {
            //特别针对slideDown展示
            axHeightDown(this.wrapper, { height: this.wrapperHeight }, '', this.duration);
        } else {
            //如果设置了动画时间参数则按用该值
            this.options.duration ? this.targetDom.style.animationDuration = `${this.options.duration}ms` : null;
            //如何设置了动画名则用该动画
            this.options.in ? this.wrapper.style.cssText = `animation-name: ${this.options.in};${dragStyle};${this.oldStyle}` : null;
        }


        //完全打开执行
        setTimeout(() => {
            //更新显示状态
            this.state = 'shown';
            //回调
            this.options.onShown && this.options.onShown.call(this);
            'shown' in this.handlers ? this.emit('shown', '') : null;
        }, this.duration);


        callback && callback.call(this);

        //返回实例，以便于链式操作
        return this;
    }
    hide(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed || this.state !== 'shown') {
            return this;
        }
        //更新状态
        this.state = 'ing';
        //回调
        this.options.onHide && this.options.onHide.call(this);
        'hide' in this.handlers ? this.emit('hide', '') : null;
        //如果使用按钮触发对话框，那么删除样式
        this.button ? this.button.classList.remove(this.options.btnClass) : null;
        //获得执行动画时间
        this.options.duration ? this.targetDom.style.animationDuration = this.options.duration : null;
        //执行窗口关闭动画
        this.targetDom.classList.add('ax-hide');
        //如果被拖拽过，则在关闭的地方打开
        let dragStyle = '';
        this.wrapPosition ? dragStyle = `position:${this.wrapPosition};left:${this.wrapLeft};top:${this.wrapTop};` : null;
        if (this.options.out == 'slideUp') {
            //特别针对slideDown展示
            axHeightUp(this.wrapper, { height: this.wrapperHeight }, '', this.duration);
        } else {
            this.options.out ? this.wrapper.style.cssText = `animation-name: ${this.options.out};${dragStyle};${this.oldStyle}` : null;
        }

        //关闭窗口同时暂停媒体播放
        this.targetDom.querySelectorAll('video,audio').forEach(elem => {
            elem.pause();
        });



        //延迟执行（等待关闭动画完成）
        setTimeout(() => {
            //更新显示状态
            this.state = 'hidden'
            //移除样式
            this.targetDom.classList.remove('ax-show');
            //移除节点
            this.targetDom.remove();
            //回调
            this.options.onHidden && this.options.onHidden.call(this);
            'hidden' in this.handlers ? this.emit('hidden', '') : null;
            this.wrapper.style.cssText = this.oldStyle;
            //动画结束，恢复wrapper高度
            //this.options.out == 'slideUp' ? this.wrapper.style.height = 'auto' : null;
        }, this.duration);

        callback && callback.call(this);
        //返回实例，以便于链式操作
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }

}
/* (() => {
    //对使用axDialog属性的元素使用本插件，可窗口嵌套
    function eachInstance(arr, parent) {
        arr.forEach(elem => {
            let instance = new axDialog('', elem);
            parent ? instance.targetDom.style.zIndex = parseInt(getComputedStyle(parent.targetDom).zIndex) + 1 : null;
            eachInstance(instance.targetDom.querySelectorAll('[axDialog]'), instance);
        });
    }
    let arrInstance = document.querySelectorAll('[axDialog]');
    eachInstance(arrInstance);
})(); */


/*!
* 插件：气泡定位；使用方法：new axPosition(btnDom, boxDom,{参数})
*/
class axPosition {
    constructor(btnDom, boxDom, options) {
        //没有btn或box则不再执行
        if (!btnDom || !boxDom) {
            return false;
        }
        this.btnDom = axIdToDom(btnDom);
        this.boxDom = axIdToDom(boxDom);

        this.options = axExtend({
            placement: 'top-center',//浮层所在位置，支持如下位置：
            //left(left-center)、right(right-center)、top(top-center)、bottom(bottom-center)、center
            //left-start、left-end
            //right-start、right-end
            //top-start、top-end
            //bottom-start、bottom-end
            arrowShow: true, //是否显示三角指示箭头，默认true显示，可选择false不显示
            arrowOffset: '2.8rem',//当*-start或*-end时并且btn尺寸大于box一半，三角指示箭头的边距，单位可以是px或rem，注意1rem=10px
            gap: '.8rem',//浮层与按钮的距离，单位可以是px或rem，注意1rem=10px
            parent: document.body,//观察节点，该节点中有变化将reset位置，默认是body
            btnObserve: {
                attrs: ['class', 'style', 'size', 'width', 'height'],//监听button的属性数组，这些属性的变化会触发位置更新
                size: true,//是否监听属性变化
            },
            boxObserve: {
                attrs: ['class', 'style', 'size', 'width', 'height'],//监听box的属性数组，这些属性的变化会触发位置更新
                size: true,//是否监听属性变化

            },
            delay: 20,//监听延时，即多久后更新位置，默认20，单位ms
            onInit: '',//监听函数，监听初始化，支持参数placement
            onUpdated: '',//监听函数，监听更新位置，支持参数placement
        }, options);
        //定义监听事件成员
        this.handlers = {};
        //标准化参数
        this.options.placement === 'left' ? this.options.placement = 'left-center' :
            this.options.placement === 'right' ? this.options.placement = 'right-center' :
                this.options.placement === 'top' ? this.options.placement = 'top-center' :
                    this.options.placement === 'bottom' ? this.options.placement = 'bottom-center' : null;
        //将gap从带单位的string转纯数字
        this.gap = this.strToNum(this.options.gap);
        //定义监听节点
        this.parent = axIdToDom(this.options.parent);

        //初始化位置
        this.placement = this.options.placement;

        //记录触发次数
        this.observerCount = 0;
        this.completeCount = 0;
        //触发监听函数
        this.trigger = () => {
            this.observerCount++;
            this.complete(this.observerCount);
        }
        //监听按钮属性变化，
        this.btnAttrsObserver = new MutationObserver((mutations) => {
            for (let mutation of mutations) {
                if (this.options.btnObserve.attrs.includes(mutation.attributeName)) {
                    //按钮可能被修改class或设置style.left/top，导致其位移
                    this.trigger();
                }
            }
        });
        //监听按钮尺寸变化，因内容撑开原尺寸，以及手动修改其width、height、padding会触发
        this.btnSizeObserver = new ResizeObserver(() => {
            this.trigger();
        });
        //监听气泡的属性和节点变化
        this.boxAttrsObserver = new MutationObserver((mutations) => {
            for (let mutation of mutations) {
                //需要确保只监听一次
                if (this.options.boxObserve.attrs.includes(mutation)) {
                    this.trigger();
                }
            }
        });
        this.boxSizeObserver = new ResizeObserver(() => {
            //需要确保只监听一次
            this.trigger();
        });

        //异步父节点变化，因为button可能会被其他节点撑开导致位移
        this.parentObserver = new MutationObserver(() => {
            if (!this.unsettable('btn') && this.btnData) {
                //在按钮非hidden或append前提下
                if (axOffset(this.btnDom).left !== this.btnData.left || axOffset(this.btnDom).top !== this.btnData.top) {
                    //只有相对位置发生变化才reset
                    this.trigger();
                }
            }
        });


        this.init();
    }
    init() {
        //创建三角指示箭头
        if (this.placement !== 'center') {
            this.createArrow();
        }
        //设置box属性
        this.setAttribute();
        //先预设位置并定位
        this.resetPlacement();
        //监听窗口变化，修正位置并重新定位
        window.addEventListener("resize", this.trigger);
        //监听页面滚动，修正位置并重新定位
        window.addEventListener('scroll', this.trigger);
        //按钮监听
        if (this.options.btnObserve.attrs.length > 0) {
            this.btnAttrsObserver.observe(this.btnDom, {
                //只监听属性变化，节点和内容监听借助ResizeObserver监听即可
                attributes: true,
            });
        }
        this.options.btnObserve.size ? this.btnSizeObserver.observe(this.btnDom) : null;
        //容器监听
        if (this.options.boxObserve.attrs.length > 0) {
            this.boxAttrsObserver.observe(this.boxDom, {
                attributes: true,
            });
        }
        this.options.boxObserve.size ? this.boxSizeObserver.observe(this.boxDom) : null;
        //父节点监听
        if (this.parent) {
            this.parentObserver.observe(this.parent, {
                attributes: true,
                characterData: true,
                childList: true,
                subtree: true,
                attributeOldValue: true,
                characterDataOldValue: true
            });
        }

        //监听初始化
        this.options.onInit && this.options.onInit.call(this, this.placement);
        this.handlers.hasOwnProperty('init') ? this.emit('init', this.placement) : null;
        return this;
    }
    complete(count) {
        setTimeout(() => {
            if (this.observerCount === count) {
                this.resetPlacement();
                //归零或者不归零持续累加亦可
                //this.observerCount = 0;
            }
            //因为原生observer方法是异步的，此处的时间为0的话可能导致有些异步observer还没完成，所以这里使用类似防抖的方法延迟执行
        }, this.options.delay);
    }
    strToNum(str) {
        if (!str) {
            return 0;
        }
        let num;
        if (str.endsWith('rem') || str.endsWith('REM')) {
            //'1.2rem'或'1.2REM'转成数字12（AXUI中px和rem换算方式为px=rem*10）
            num = str.replace('rem', '').replace('REM', '') * 10;
        } else if (str.endsWith('px') || str.endsWith('PX')) {
            //'12px'或'12PX'转成数字12
            num = ~~str.replace('px', '').replace('PX', '');
        } else {
            //数字字符串或者数字
            num = ~~str;
        }
        return num;
    }
    setAttribute() {
        //强行设置绝对定位
        if (getComputedStyle(this.boxDom).position !== 'absolute') {
            this.boxDom.style.position = 'absolute';
        }
    }
    createArrow() {
        if (this.options.arrowShow) {
            this.arrow = axAddElem('i', { style: 'position:absolute', arrow: '' });
            this.boxDom.appendChild(this.arrow);
        }
    }
    setArrow(placement) {
        let exceedWidth = this.btnData.width > this.boxData.width / 2,
            exceedHeight = this.btnData.height > this.boxData.height / 2,
            arrowHalf = this.strToNum(getComputedStyle(this.arrow).width) / 2;
        //添加位置属性
        this.boxDom.setAttribute('placement', placement);
        //添加top、bottom、left和right属性
        if (placement.includes('top') || placement.includes('bottom')) {
            if (placement.includes('start')) {
                if (!exceedWidth) {
                    this.arrow.style.left = Math.floor(this.btnData.width / 2 - arrowHalf) + 'px';
                    this.arrow.style.right = 'auto';
                } else {
                    this.arrow.style.left = this.options.arrowOffset;
                    this.arrow.style.right = 'auto';
                }
            } else if (placement.includes('end')) {
                if (!exceedWidth) {
                    this.arrow.style.left = 'auto';
                    this.arrow.style.right = Math.floor(this.btnData.width / 2 - arrowHalf) + 'px';
                } else {
                    this.arrow.style.left = 'auto';
                    this.arrow.style.right = this.options.arrowOffset;
                }
            } else {
                this.arrow.style.left = this.boxData.width / 2 - 1 - arrowHalf + 'px';
            }
        } else if (placement.includes('left') || placement.includes('right')) {
            if (placement.includes('start')) {
                if (!exceedHeight) {
                    this.arrow.style.top = Math.floor(this.btnData.height / 2 - arrowHalf) + 'px';
                    this.arrow.style.bottom = 'auto';
                } else {
                    this.arrow.style.top = this.options.arrowOffset;
                    this.arrow.style.bottom = 'auto';
                }
            } else if (placement.includes('end')) {
                if (!exceedHeight) {
                    this.arrow.style.top = 'auto';
                    this.arrow.style.bottom = Math.floor(this.btnData.height / 2 - arrowHalf) + 'px';
                } else {
                    this.arrow.style.top = 'auto';
                    this.arrow.style.bottom = this.options.arrowOffset;
                }
            } else {
                this.arrow.style.top = this.boxData.height / 2 - 1 - arrowHalf + 'px';
            }
        }
    }
    getRectsData() {
        //浏览器尺寸不包含滚动条宽度
        this.browserData = {
            width: document.documentElement.clientWidth,
            height: document.documentElement.clientHeight
        }
        //按钮数据
        let btnRect = this.btnDom.getBoundingClientRect();
        this.btnData = {
            width: btnRect.width,
            height: btnRect.height,
            offsetLeft: btnRect.left,
            offsetRight: btnRect.right,
            offsetTop: btnRect.top,
            offsetBottom: btnRect.bottom,
            gapRight: this.browserData.width - btnRect.right,
            gapBottom: this.browserData.height - btnRect.bottom,
            //left:'',top:''
            ...axOffset(this.btnDom)
        };
        //浮层数据
        let boxRect = this.boxDom.getBoundingClientRect();
        this.boxData = {
            width: boxRect.width,
            height: boxRect.height,
            offsetLeft: boxRect.left,
            offsetRight: boxRect.right,
            offsetTop: boxRect.top,
            offsetBottom: boxRect.bottom,
            gapRight: this.browserData.width - boxRect.right,
            gapBottom: this.browserData.height - boxRect.bottom
        };
    }
    //根据位置参数初步设定box的left和top值
    setPlacement(placement) {
        let [left, right, top, bottom] = ['auto', 'auto', 'auto', 'auto'];
        if (placement.includes('bottom')) {
            //先计算left
            left = placement === 'bottom' || placement === 'bottom-center' ? (this.btnData.left - (this.boxData.width - this.btnData.width) / 2) :
                placement === 'bottom-start' ? this.btnData.left :
                    placement === 'bottom-end' ? this.btnData.left + this.btnData.width - this.boxData.width : 0;
            //再计算top，确保box扩展内容后不改变定位
            top = `${Math.ceil(this.btnData.top + this.btnData.height + this.gap)}px`;
            left = `${Math.ceil(left)}px`;
        } else if (placement.includes('top')) {
            left = placement === 'top' || placement === 'top-center' ? (this.btnData.left - (this.boxData.width - this.btnData.width) / 2) :
                placement === 'top-start' ? this.btnData.left :
                    placement === 'top-end' ? (this.btnData.left + this.btnData.width - this.boxData.width) : 0;
            //再计算bottom，确保box扩展内容后不改变定位
            bottom = this.browserData.height - this.btnData.top + this.gap;
            //如果有间隙则对top加上间隙和单位
            bottom = `${Math.ceil(bottom)}px`;
            left = `${Math.ceil(left)}px`;
        } else if (placement.includes('left')) {
            //先计算top
            if (placement === 'left' || placement === 'left-center') {
                top = `${Math.ceil(this.btnData.top - (this.boxData.height - this.btnData.height) / 2)}px`;
            } else if (placement === 'left-start') {
                top = `${Math.ceil(this.btnData.top)}px`;
            } else if (placement === 'left-end') {
                //使用bottom值，确保box扩展内容后不改变定位
                bottom = `${Math.ceil(this.browserData.height - this.btnData.top - this.btnData.height)}px`;
            }
            //再计算right，确保box扩展内容后不改变定位
            right = `${Math.ceil(this.browserData.width - this.btnData.left + this.gap)}px`;
        } else if (placement.includes('right')) {
            //先计算top
            if (placement === 'right' || placement === 'right-center') {
                top = `${Math.ceil(this.btnData.top - (this.boxData.height - this.btnData.height) / 2)}px`;
            } else if (placement === 'right-start') {
                top = `${Math.ceil(this.btnData.top)}px`;
            } else if (placement === 'right-end') {
                //使用bottom值，确保box扩展内容后不改变定位
                bottom = `${Math.ceil(this.browserData.height - this.btnData.top - this.btnData.height)}px`;
            }
            //再计算left
            left = `${Math.ceil(this.btnData.left + this.btnData.width + this.gap)}px`;
        } else if (placement === 'center') {
            this.boxDom.style.position = 'fixed';
            left = '50%';
            top = '50%';
            this.boxDom.style.marginLeft = `-${Math.ceil(this.boxData.width / 2)}px`;
            this.boxDom.style.marginTop = `-${Math.ceil(this.boxData.height / 2)}px`;
        }
        //将坐标值赋予box
        this.boxDom.style.left = left;
        this.boxDom.style.right = right;
        this.boxDom.style.top = top;
        this.boxDom.style.bottom = bottom;
        //将位置写入left、right、top和bottom属性
        if (this.arrow) {
            this.setArrow(placement);
        }
    }
    //获得box与浏览器四周的空隙，通过正负判断进行重定位
    getSurroundGap() {
        let left = this.btnData.offsetLeft - this.boxData.width - this.gap,
            top = this.btnData.offsetTop - this.boxData.height - this.gap,
            right = this.browserData.width - this.btnData.offsetRight - this.boxData.width - this.gap,
            bottom = this.browserData.height - this.btnData.offsetBottom - this.boxData.height - this.gap,
            //按钮左右box的top和bottom
            startTop = this.btnData.offsetTop,
            startBottom = this.browserData.height - (startTop + this.boxData.height),
            centerTop = this.btnData.offsetTop - (this.boxData.height - this.btnData.height) / 2,
            centerBottom = this.browserData.height - (centerTop + this.boxData.height),
            endTop = this.btnData.offsetBottom - this.boxData.height,
            endBottom = this.browserData.height - this.btnData.offsetBottom,
            //按钮上下box的left和right
            startLeft = this.btnData.offsetLeft,
            startRight = this.browserData.width - (startLeft + this.boxData.width),
            centerLeft = this.btnData.offsetLeft - (this.boxData.width - this.btnData.width) / 2,
            centerRight = this.browserData.width - (centerLeft + this.boxData.width),
            endLeft = this.btnData.offsetRight - this.boxData.width,
            endRight = this.browserData.width - this.btnData.offsetRight;
        return {
            left, right, top, bottom,
            start: { left: startLeft, right: startRight, top: startTop, bottom: startBottom },
            center: { left: centerLeft, right: centerRight, top: centerTop, bottom: centerBottom },
            end: { left: endLeft, right: endRight, top: endTop, bottom: endBottom }
        };
    }
    fixPlacement() {
        //完全离开了浏览器，则不理会
        if (this.btnData.gapRight > this.browserData.width || this.btnData.offsetLeft > this.browserData.width) {
            return this;
        }

        let gaps = this.getSurroundGap();
        //先掉转方向
        if (this.placement.includes('top') && (gaps.top < 0 && gaps.bottom >= 0)) {
            this.placement = this.placement.replace('top', 'bottom');
        } else if (this.placement.includes('bottom') && (gaps.bottom < 0 && gaps.top >= 0)) {
            this.placement = this.placement.replace('bottom', 'top');
        } else if (this.placement.includes('left') && (gaps.left < 0 && gaps.right >= 0)) {
            this.placement = this.placement.replace('left', 'right');
        } else if (this.placement.includes('right') && (gaps.right < 0 && gaps.left >= 0)) {
            this.placement = this.placement.replace('right', 'left');
        }

        //水平阶梯平移，对于居上和居下的位置，box的宽度应该要大于浏览器宽度才需要改变
        if ((this.placement.includes('top') || this.placement.includes('bottom')) && this.boxData.width < this.browserData.width) {
            //浮层宽度大于按钮宽度
            if (this.boxData.width > this.btnData.width) {
                if (this.btnData.offsetLeft < 0 && this.btnData.gapRight < this.browserData.width) {
                    //按钮左侧已经半离开浏览器
                    this.placement = this.placement.replace('end', 'start').replace('center', 'start');
                } else if (this.btnData.gapRight < 0 && this.btnData.offsetLeft < this.browserData.width) {
                    //按钮右侧已经半离开浏览器
                    this.placement = this.placement.replace('start', 'end').replace('center', 'end');
                } else {
                    //按钮完全暴露在浏览器里
                    //左侧

                    if (gaps.end.left < 0 && gaps.center.left < 0) {
                        //进去二级
                        this.placement = this.placement.replace('end', 'start').replace('center', 'start');
                    } else if (gaps.end.left < 0 && this.placement.includes('end')) {
                        //进去一级
                        this.placement = this.placement.replace('end', 'center');
                    } else if (gaps.end.left < 0 && gaps.center.left > 0 && this.placement.includes('start') && (this.options.placement.includes('end') || this.options.placement.includes('center'))) {
                        //出来一级
                        this.placement = this.placement.replace('start', 'center');
                    } else if (gaps.end.left > 0 && this.placement.includes('center') && (this.options.placement.includes('end'))) {
                        //出来二级
                        this.placement = this.placement.replace('center', 'end');
                    }
                    //右侧
                    if (gaps.start.right < 0 && gaps.center.right < 0) {
                        this.placement = this.placement.replace('start', 'end').replace('center', 'end');
                    } else if (gaps.start.right < 0 && this.placement.includes('start')) {
                        this.placement = this.placement.replace('start', 'center');
                    } else if (gaps.start.right < 0 && gaps.center.right > 0 && this.placement.includes('end') && (this.options.placement.includes('start') || this.options.placement.includes('center'))) {
                        //出来一级
                        this.placement = this.placement.replace('end', 'center');
                    } else if (gaps.start.right > 0 && this.placement.includes('center') && (this.options.placement.includes('start'))) {
                        //出来二级
                        this.placement = this.placement.replace('center', 'start');
                    }
                }

            } else {
                //浮层宽度小于按钮宽度

                //左侧
                if (gaps.start.left < 0 && gaps.center.left < 0) {
                    //进去二级
                    this.placement = this.placement.replace('start', 'end').replace('center', 'end');
                } else if (gaps.start.left < 0 && this.placement.includes('start')) {
                    //进去一级
                    this.placement = this.placement.replace('start', 'center');
                } else if (gaps.start.left < 0 && gaps.center.left > 0 && this.placement.includes('end') && (this.options.placement.includes('start') || this.options.placement.includes('center'))) {
                    //出来一级
                    this.placement = this.placement.replace('end', 'center');
                } else if (gaps.start.left > 0 && this.placement.includes('center') && (this.options.placement.includes('start'))) {
                    //出来二级
                    this.placement = this.placement.replace('center', 'start');
                }
                //右侧
                if (gaps.end.right < 0 && gaps.center.right < 0) {
                    this.placement = this.placement.replace('end', 'start').replace('center', 'start');
                } else if (gaps.end.right < 0 && this.placement.includes('end')) {
                    this.placement = this.placement.replace('end', 'center');
                } else if (gaps.end.right < 0 && gaps.center.right > 0 && this.placement.includes('start') && (this.options.placement.includes('end') || this.options.placement.includes('center'))) {
                    //出来一级
                    this.placement = this.placement.replace('start', 'center');
                } else if (gaps.end.right > 0 && this.placement.includes('center') && (this.options.placement.includes('end'))) {
                    //出来二级
                    this.placement = this.placement.replace('center', 'end');
                }
            }

        }
        //垂直阶梯平移，对于居左和居右的位置，box的高度应该要大于浏览器高度才需要改变
        if ((this.placement.includes('left') || this.placement.includes('right')) && this.boxData.height < this.browserData.height) {
            if (this.boxData.height > this.btnData.height) {
                //浮层高度大于按钮高度
                if (this.btnData.offsetTop < 0 && this.btnData.gapBottom < this.browserData.height) {
                    //按钮顶侧已经半离开浏览器
                    this.placement = this.placement.replace('end', 'start').replace('center', 'start');
                } else if (this.btnData.gapBottom < 0 && this.btnData.offsetTop < this.browserData.height) {
                    //按钮底侧已经半离开浏览器
                    this.placement = this.placement.replace('start', 'end').replace('center', 'end');
                } else {
                    //按钮完全暴露在浏览器里
                    //顶侧
                    if (gaps.end.top < 0 && gaps.center.top < 0) {
                        this.placement = this.placement.replace('end', 'start').replace('center', 'start');
                    } else if (gaps.end.top < 0 && this.placement.includes('end')) {
                        this.placement = this.placement.replace('end', 'center');
                    } else if (gaps.end.top < 0 && gaps.center.top > 0 && this.placement.includes('start') && (this.options.placement.includes('end') || this.options.placement.includes('center'))) {
                        //出来一级
                        this.placement = this.placement.replace('start', 'center');
                    } else if (gaps.end.top > 0 && this.placement.includes('center') && (this.options.placement.includes('end'))) {
                        //出来二级
                        this.placement = this.placement.replace('center', 'end');
                    }
                    //底侧
                    if (gaps.start.bottom < 0 && gaps.center.bottom < 0) {
                        this.placement = this.placement.replace('start', 'end').replace('center', 'end');
                    } else if (gaps.start.bottom < 0 && this.placement.includes('start')) {
                        this.placement = this.placement.replace('start', 'center');
                    } else if (gaps.start.bottom < 0 && gaps.center.bottom > 0 && this.placement.includes('end') && (this.options.placement.includes('start') || this.options.placement.includes('center'))) {
                        //出来一级
                        this.placement = this.placement.replace('end', 'center');
                    } else if (gaps.start.bottom > 0 && this.placement.includes('center') && (this.options.placement.includes('start'))) {
                        //出来二级
                        this.placement = this.placement.replace('center', 'start');
                    }
                }
            } else {
                //浮层高度小于按钮高度

                //顶侧
                if (gaps.start.top < 0 && gaps.center.top < 0) {
                    this.placement = this.placement.replace('start', 'end').replace('center', 'end');
                } else if (gaps.start.top < 0 && this.placement.includes('start')) {
                    this.placement = this.placement.replace('start', 'center');
                } else if (gaps.start.top < 0 && gaps.center.top > 0 && this.placement.includes('end') && (this.options.placement.includes('start') || this.options.placement.includes('center'))) {
                    //出来一级
                    this.placement = this.placement.replace('end', 'center');
                } else if (gaps.start.top > 0 && this.placement.includes('center') && (this.options.placement.includes('start'))) {
                    //出来二级
                    this.placement = this.placement.replace('center', 'start');
                }
                //底侧
                if (gaps.end.bottom < 0 && gaps.center.bottom < 0) {
                    this.placement = this.placement.replace('end', 'start').replace('center', 'start');
                } else if (gaps.end.bottom < 0 && this.placement.includes('end')) {
                    this.placement = this.placement.replace('end', 'center');
                } else if (gaps.end.bottom < 0 && gaps.center.bottom > 0 && this.placement.includes('start') && (this.options.placement.includes('end') || this.options.placement.includes('center'))) {
                    //出来一级
                    this.placement = this.placement.replace('start', 'center');
                } else if (gaps.end.bottom > 0 && this.placement.includes('center') && (this.options.placement.includes('end'))) {
                    //出来二级
                    this.placement = this.placement.replace('center', 'end');
                }
            }

        }
        //根据原参数重构placement，让box完全暴露
        this.fullyExpose(gaps);
    }
    fullyExpose(gaps) {
        if (this.placement === this.options.placement) {
            return;
        }
        let placeArr = this.placement.split('-'), parentStr = placeArr[0], childStr = placeArr[1];
        //先添加主方位
        parentStr = (this.options.placement.includes('top') && gaps.top >= 0) ? 'top' :
            (this.options.placement.includes('bottom') && gaps.bottom >= 0) ? 'bottom' :
                (this.options.placement.includes('left') && gaps.left >= 0) ? 'left' :
                    (this.options.placement.includes('right') && gaps.right >= 0) ? 'right' : placeArr[0];
        //再添加偏移位
        if (this.options.placement.includes('top') || this.options.placement.includes('bottom')) {
            //容器与按钮尺寸关系影响要设置的属性
            if (this.options.placement.includes('start') && gaps.start[(this.boxData.width > this.btnData.width) ? 'right' : 'left'] >= 0) {
                childStr = 'start';
            } else if (this.options.placement.includes('end') && gaps.end[(this.boxData.width > this.btnData.width) ? 'left' : 'right'] >= 0) {
                childStr = 'end';
            } else if (this.options.placement.includes('center') && gaps.center.left >= 0 && gaps.center.right >= 0) {
                childStr = 'center';
            }
        } else if (this.options.placement.includes('left') || this.options.placement.includes('right')) {
            if (this.options.placement.includes('start') && gaps.start[(this.boxData.height > this.btnData.height) ? 'bottom' : 'top'] >= 0) {
                childStr = 'start';
            } else if (this.options.placement.includes('end') && gaps.end[(this.boxData.height > this.btnData.height) ? 'top' : 'bottom'] >= 0) {
                childStr = 'end';
            } else if (this.options.placement.includes('center') && gaps.center.top >= 0 && gaps.center.bottom >= 0) {
                childStr = 'center';
            }
        }
        this.placement = `${parentStr}-${childStr}`;
    }
    unsettable(type) {
        let dom = type === 'btn' ? this.btnDom : this.boxDom;
        //dom是否已经存在（在页面中或内存中），是否在页面中出现append，是否是隐藏的
        return (axIsVnode(dom) || getComputedStyle(dom).display === 'none');
    }
    resetPlacement() {
        if (this.destroyed) {
            return this;
        }
        if (this.unsettable()) {
            //如果页面还没有append节点或者是隐藏元素，则不执行（如果页面没有节点则getComputedStyle获得的所有属性均为空）
            return false;
        }
        //更新元素数据（btn、box和browser）
        this.getRectsData();
        //更新位置变量（this.placement）
        this.fixPlacement();
        //重新定位
        this.setPlacement(this.placement);
        //监听
        this.options.onUpdated && this.options.onUpdated.call(this, this.placement);
        this.handlers.hasOwnProperty('updated') ? this.emit('updated', this.placement) : null;
    }
    update(callback) {
        if (this.destroyed) {
            return this;
        }
        //重新定位
        this.resetPlacement();
        //回调
        callback && callback.call(this, this.placement);
        return this;
    }
    change(elem, callback) {
        if (this.destroyed) {
            return this;
        }
        //重新定义btn
        this.btnDom = axIdToDom(elem);
        //重新定位
        this.resetPlacement();
        //回调
        callback && callback.call(this, this.placement);
        return this;
    }
    destroy() {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //取消观察
        this.interact.disconnect();
        //取消监听
        this.options.btnObserve.attrs.length > 0 ? this.btnAttrsObserver.disconnect() : null;
        this.options.btnObserve.size ? this.btnResizeObserver.unobserve() : null;
        this.options.boxObserve.attrs.length > 0 ? this.boxAttrsObserver.disconnect() : null;
        this.options.boxObserve.size ? this.boxResizeObserver.unobserve() : null;
        this.parent ? this.parentObserver.disconnect() : null;
        window.removeEventListener('resize', this.trigger);
        window.removeEventListener('scorll', this.trigger);
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}



/*!
* 插件：气泡弹窗；使用方法：new axPopup('#id',{参数})
*/
class axPopup {
    constructor(elem, options) {
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            title: '',//窗口标题，为空等于不显示
            titleBreak: false,//标题分割线，默认不显示
            placement: 'top',//窗口相对按钮的位置，可填top|top-start|top-end|bottom|bottom-start|bottom-end|left|left-start|left-end|right|right-start|right-end|center
            arrow: true,//是否显示指示箭头，默认true显示，可填false不显示
            trigger: 'click',//显示窗口的动作，默认click点击显示，可填hover，sticky，input，none
            pageClose: true,//点击窗口之外的区域是否允许关闭窗口，默认true允许，可填false
            size: 'md',//小窗尺寸，默认为md，可填sm|md|lg|xl|xxl
            duration: 0,//弹出和关闭窗口动画的时间，如果为0则使用css中定义的动画时间，单位ms
            in: 'fadeIn',//可填fadeIn,fadeInDown,fadeInUp,slideDown,fadeInLeft,fadeInRight,springIn,scaleIn等
            out: 'fadeOut',//可填fadeOut,fadeOutDown,fadeOutUp,slideUp,fadeOutLeft,fadeOutRight,springOut,scaleOut等
            multiple: false,//是否允许打开多个窗口，默认不允许
            offsetShow: true,//是否保持弹窗与元素的间隙，默认保持，可选false取消间隙
            rel: '',//可以触发本窗口的元素，可以是#id，节点对象，或数组（多个按钮均触发）
            triggerRel: 'click',//关联按钮显示窗口的动作，默认click点击显示，可填hover，sticky，input，none，如果不填将使用trigger来代替
            theme: '',//窗口的风格，默认为空，可选crude
            stay: false,//设置打开后是否就不再关闭，默认为false，可选true
            fixDistance: '2.4rem',//如果窗口宽度或高度小于按钮，placement为*-start或*-end时箭头所处的位置，字符串格式
            name: '',//多媒体文件名称
            width: '',//内容宽度，当type是iframe的时候，有必要设置该参数，比如800px
            height: '',//内容高度，当type是iframe的时候，有必要设置该参数，比如400px，但是不要超过600px否则会产生垂直滚动条
            padding: 'ax-padding',//窗口内边距，值可以是布尔值也可以是类名，默认带边距，可填false,true,ax-padding,ax-padding ax-xxl,...
            content: '',//窗口内容，可填#id或html,
            type: 'html',//内容的类型，默认是html，可填picture|video|audio|iframe|async|html|form
            url: '',//异步请求地址，可填“url”或者“url#id”
            btnClass: 'ax-opened',//触发按钮在打开气泡后新增的样式名
            close: true,//关闭窗口的按钮，默认显示
            closePlace: 'inside',//可选 inside或outside
            triggerShowOnly: false,//通过触发是否只能显示而不能关闭，默认false，可选择true只显示不关闭
            triggerStrict: false,//触发元素是否必须是指定input或rel节点，默认false忽略，如果为true那么input或rel的子元素将不会触发
            footerShow: true,//是否显示底部，默认显示
            footerType: 'line',//底部按钮风格，可选center,line,right
            footerBreak: false,//底部分割线，默认不显示
            note: '',//其他内容
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            zIndex: 0,//气泡层级，默认为0即不设置，如果与其他组件发生遮挡请自行设置
            response: true,//是否在手机端强制以绝对居中的方式打开气泡，默认true，可选择false保持pc端的打开方式
            onInit: '',//回调函数，初始化后执行，无参数
            onShow: '',//回调函数，开始显示后执行，无参数
            onShown: '',//回调函数，完成显示后执行，无参数
            onHide: '',//回调函数，开始关闭后执行，无参数
            onHidden: '',//回调函数，完全关闭后执行，无参数
            onUpdate: '',//回调函数，参数/内容/位置更新前执行，无参数
            onUpdated: '',//回调函数，参数更新后执行，支持参数isUpdated(是否更新了：true或false)、before（更新前参数）、after（更新后参数）
            onUpdateContent: '',//回调函数，内容更新后执行，支持参数isUpdated(是否更新了：true或false)、before（更新前内容）、after（更新后内容）
            onUpdatePosition: '',//回调函数，位置更新后执行，支持参数placement(位置)
            onDestroy: '',//回调函数，销毁后执行，无参数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            position: {},//气泡定位器axPosition的参数，
            confirm: {
                text: '确定',//确认按钮文字
                classname: '',//确认按钮的类名
                icon: '',//确认按钮的图标，格式是<i class="ax-iconfont ax-icon-edit"></i>
                type: 'close',//按钮类型，可选择close关闭、async异步提交表单、submit直接提交表单、prevent阻止按钮默认事件什么事也不做
                url: '',//提交的地址，如果type为async则必须使用该参数
                callback: ''//确认按钮的回调函数，最终按函数方式使用，格式是function(button,content){}，button是按钮节点，content是返回内容
            },
            cancel: {
                text: '关闭',//取消按钮文字
                classname: '',//取消按钮的类名
                icon: '',//取消按钮的图标
                callback: ''//取消按钮的回调函数，格式是function(button,content){}，button是按钮节点，content是返回内容
            },
            diy: {
                text: '',//自定义按钮文字
                classname: '',//自定义按钮的类名
                icon: '',//自定义按钮的图标
                url: '',//自定义按钮的超链接
                target: '',//自定义按钮的链接目标，一般是_blank或者_self
                callback: ''//自定义按钮的回调函数，格式是function(button,content){}，button是按钮节点，content是返回内容
            }
        }, options, elem, this.constructor.name);
        let _this = this;
        //定义监听事件成员
        this.handlers = {};
        //如果使用表单
        this.form = null;
        //获得axPopup元素的定位按钮
        this.button = axIdToDom(elem);
        //显示状态
        this.state = 'hidden'
        //绑定事件状态
        this.bound = false;
        //绑定的事件函数变量，待解绑
        this.triggerShow = function (e) {
            if (_this.options.triggerStrict && ![..._this.rel, _this.button].includes(e.target)) {
                return false;
            }
            //rel节点会给自身绑定trigger事件
            if (_this.options.trigger === 'click' || this.trigger === 'click') {
                //点击按钮
                if (_this.state === 'hidden') {
                    _this.show();
                    if (!_this.options.multiple) {
                        let items = axInstance.data.filter(i => i.type === 'popup' && i.instance !== _this && i.instance.isShow);
                        items.forEach(i => {
                            i.instance.hide();
                        });
                    }
                } else if (_this.state === 'shown') {
                    !_this.options.triggerShowOnly ? _this.hide() : null;
                }
            } else if (_this.options.trigger === 'input') {
                //键入input
                if (!_this.button.value) {
                    !_this.options.triggerShowOnly ? _this.hide() : null;
                } else {
                    _this.show();
                }
            }
        }
        //按钮和关联按钮之外点击
        this.toggleShow = (ev) => {
            //判断是否在其他按钮之外
            let relOut = true;
            for (let i = 0; i < this.rel.length; i++) {
                if (!axIsOutside(ev, this.rel[i])) {
                    relOut = false;
                    break;
                }
            }
            //窗口之外，主按钮之外，其他按钮之外，并且窗口时显示状态，点击页面才可关闭窗口
            if (axIsOutside(ev, this.targetEl) && axIsOutside(ev, this.button) && relOut && this.state === 'shown') {
                this.hide();
            };
        }
        //初始化
        this.init();
        return this;
    }
    async init() {
        //添加到实例合集并设置销毁状态
        axInstance.push(this, this.options.insName, 'popup');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;

        //定义关联按钮
        this.rel = axAllToDoms(this.options.rel);

        //创建窗口模板
        this.renderDom();
        //写入属性
        this.setAttribute();
        //获取内容
        if (this.options.content.nodeType === 1) {
            //如果content是Dom节点，则整体装入节点的内容
            this.content.innerHTML = '';
            this.content.appendChild(this.options.content);
            //定义form节点，可能为null
            this.form = this.content.querySelector('form');
        } else if (this.options.type === 'form' && typeof this.options.content === 'string' && this.options.content.substr(0, 1) === '#') {
            this.form = document.querySelector(this.options.content).querySelector('form');
            if (this.form) {
                //如果类型是form节点，则整体装入节点的内容
                this.content.innerHTML = '';
                this.content.appendChild(this.form);
            } else {
                console.error('Form node is required!');
                return false;
            }
        } else {
            let response;
            if (this.options.type === 'async' && this.options.url) {
                //异步文本
                //content是文本
                response = await axCreateContent(this.options, this);
            } else {
                //同步是文本
                response = axCreateContent(this.options, this);
            }
            //填充内容
            if (typeof this.options.content === 'function') {
                //如果content是函数
                this.content.innerHTML = this.options.content.call(this, response.content);
            } else {
                this.content.innerHTML = response.content;
            }
            //定义form节点，可能为null
            this.form = this.content.querySelector('form');
        }
        //设置底部
        this.footer.innerHTML = axFooterBtns(this.options, this);
        //使用Popper.js定位
        this.createPopup();
        //按钮操作
        this.targetEl.querySelectorAll('[confirm]').forEach(elem => {
            elem.onclick = (e) => {
                if (this.options.confirm.type == 'submit') {
                    if (this.form && this.options.type == 'form') {
                        //校验通过后执行
                        axValidRepeat(e, this.form, () => {
                            this.form.submit();
                        });
                        //执行回调
                        this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                    } else {
                        //报错提示
                        console.error('Form node is required and type of options must be form string!');
                        return false;
                    }
                } else if (this.options.confirm.type == 'async') {
                    if (this.form && this.options.type == 'form' && this.options.confirm.url) {
                        //校验通过后执行
                        axValidRepeat(e, this.form, () => {
                            //使用表单异步提交函数
                            axAjaxSubmit(this.options.confirm.url, this.form, elem, {
                                type: this.options.ajaxType,
                                opened: (response) => {
                                    this.submitXhr = response.xhr;
                                },
                                success: (response) => {
                                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, response.content);
                                }
                            });
                        })
                    } else {
                        //报错提示
                        console.error('Form node is required & type of options must be form string & url of async is required!');
                        return false;
                    }
                } else if (this.options.confirm.type === 'prevent') {
                    axPreventDefault(e);
                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                } else {
                    this.hide();
                    //执行回调
                    this.options.confirm.callback && this.options.confirm.callback.call(this, elem, '');
                }

            };
        });
        this.targetEl.querySelectorAll('[diy]').forEach(elem => {
            elem.onclick = () => {
                this.options.diy.callback && this.options.diy.callback.call(this, elem);
            };
        });
        this.targetEl.querySelectorAll('[cancel]').forEach(elem => {
            elem.onclick = () => {
                this.hide();
                this.options.cancel.callback && this.options.cancel.callback.call(this, elem);
            };
        });
        //任意close属性按钮均关闭窗口
        this.targetEl.querySelectorAll('[close]').forEach(elem => {
            elem.addEventListener('click', () => {
                this.hide();
            }, false);
        });

        //根据事件判断打开方式
        this.showEvent();
        //回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    createTemplate() {
        return `
                            <div class="ax-popup" id="popup${new Date().getTime()}">
                                <div class="ax-popup-wrapper">
                                    <# if(this.close){ #><i close></i><# } #>
                                    <# if(this.title){ #><div class="ax-popup-header"><# this.title #></div><# } #>
                                    <# if(this.titleBreak){ #><div class="ax-break-line"></div><# } #>
                                    <div class="ax-popup-body">
                                        <# if(typeof this.padding === 'string'){ #><div class="<# this.padding #>"> <# }else if(this.padding){ #><div class="ax-padding"> <# } #>
                                        <# this.media #>
                                        <div class="ax-popup-content"></div>
                                        <# if(this.padding){ #></div><# } #>
                                    </div>
                                    <div class="ax-popup-footer">
                                    </div>
                                </div>
                            </div>
                            `
    }
    renderDom() {
        this.template = axTplEngine(this.createTemplate(), this.options);
        //将字符串转成DOM对象
        this.targetEl = axStrToDom(this.template);
        //获取Dom树
        this.wrapper = this.targetEl.querySelector('.ax-popup-wrapper');
        this.content = this.targetEl.querySelector('.ax-popup-content');
        this.body = this.targetEl.querySelector('.ax-popup-body');
        this.targetEl.querySelector('.ax-popup-header') ? this.header = this.targetEl.querySelector('.ax-popup-header') : null;
        this.footer = this.targetEl.querySelector('.ax-popup-footer');
    }
    setAttribute() {
        this.options.size ? this.targetEl.setAttribute('size', this.options.size) : null;
        this.options.type ? this.targetEl.setAttribute('type', this.options.type) : null;
        this.options.footerType ? this.targetEl.setAttribute('footerType', this.options.footerType) : null;
        this.options.height ? this.content.style.height = this.options.height : null;
        this.options.theme ? this.targetEl.setAttribute('theme', this.options.theme) : null;
        //设置zIndex
        this.options.zIndex ? this.targetEl.style.zIndex = this.options.zIndex : null;
        //设定关闭按钮位置
        if (this.options.close) {
            this.targetEl.setAttribute('closePlace', this.options.closePlace);
        }
    }
    createPopup() {
        let placement;
        //如果是center或者在移动端则直接绝对居中显示窗口
        if (this.options.placement === 'center' || (this.options.response && axIsMobi())) {
            placement = 'center';
        } else {
            placement = this.options.placement;
        }
        this.popupPosition = new axPosition(this.button, this.targetEl, axExtend({
            placement: placement,
            arrowShow: this.options.arrow,
            gap: this.options.arrow ? '.8rem' : '.4rem',
            arrowOffset: this.options.fixDistance,
            btnObserve: {
                attrs: [],
                size: false,
            },
            boxObserve: {
                attrs: [],
                size: false,
            },
            delay:0,
            parent: null,
        }, this.options.position));
        //Popup创建后会在dom上自动创建style，在此之后才添加宽度参数
        this.oldStyle = this.options.width ? `width:${this.options.width};` : '';
    }

    showEvent() {
        if (this.options.trigger === 'click' || this.options.trigger === 'input') {
            //input用于axComplete
            //绑定事件，没有绑定才绑定事件
            !this.bound ? this.button.addEventListener(this.options.trigger, this.triggerShow, false) : null;
            //其他按钮也可打开
            if (this.rel.length > 0) {
                this.rel.forEach(item => {
                    //如果没有指定事件则使用button的trigger事件
                    if (!this.bound) {
                        //将事件帮到到节点，因为如果是input事件会出问题
                        item.trigger = this.options.triggerRel || this.options.trigger;
                        item.addEventListener(this.options.triggerRel || this.options.trigger, this.triggerShow, false)
                    }
                });
            }
        } else if (this.options.trigger === 'hover') {
            this.hover = new axHover(this.button, {
                enter: () => {
                    this.show();
                },
                leave: () => {
                    !this.options.triggerShowOnly ? this.hide() : null;
                },
                hold: this.targetEl,
            });
        } else if (this.options.trigger === 'sticky') {
            this.show();
        }
        //点击小窗外区域关闭窗口，也就是页面只显示一个小窗
        if (this.options.pageClose && !this.options.multiple) {
            !this.bound ? document.addEventListener('click', this.toggleShow, false) : null;
        }
        //更新绑定状态
        this.bound = true;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }

        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        let before = axClone(this.options),
            after = axExtend(this.options, setting),
            isUpdated = false;
        if (setting && JSON.stringify(after) != JSON.stringify(before)) {
            this.options = after;
            this.targetEl.remove();
            this.popupPosition.destroy();
            this.popupPosition = null;
            //更新后保持原有的打开状态
            if (this.state === 'shown') {
                this.init().show();
            } else if (this.state === 'hidden') {
                this.init();
            }
            isUpdated = true;
        }
        callback && callback.call(this, isUpdated, before, after);
        this.options.onUpdated && this.options.onUpdated.call(this, isUpdated, before, after);
        'updated' in this.handlers ? this.emit('updated', isUpdated, before, after) : null;
        return this;
    }
    updateContent(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }

        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        let before = this.content.innerHTML,
            isUpdated = false;
        //
        if (axType(data) === 'String') {
            if (axStrType(data)) {
                //能识别出节点
                this.content.innerHTML = axIdToDom(data) ? axIdToDom(data).innerHTML : '';
            } else {
                //纯文本
                this.content.innerHTML = data;
            }
        } else if (axType(data).includes('HTML')) {
            //是节点
            this.content.innerHTML = '';
            this.content.appendChild(data);
        } else {
            this.content.innerHTML = `No data to update!`;
        }
        //
        if (this.content.innerHTML == before) {
            isUpdated = false;
        } else {
            isUpdated = true;
        }
        callback && callback.call(this, isUpdated, before, this.content.innerHTML);
        this.options.onUpdateContent && this.options.onUpdateContent.call(this, isUpdated, before, this.content.innerHTML);
        'updateContent' in this.handlers ? this.emit('updateContent', isUpdated, before, this.content.innerHTML) : null;
        return this;
    }
    updatePosition(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }

        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        let newButton = axIdToDom(obj);
        this.popupPosition.change(newButton, (placement) => {
            this.options.onUpdatePosition && this.options.onUpdatePosition.call(this, placement);
            'updatePosition' in this.handlers ? this.emit('updatePosition', placement) : null;
        });
        callback && callback.call(this);
        return this;
    }
    show(callback) {
        //如果已经销毁或本身就是打开状态则不再执行
        if (this.destroyed || this.state !== 'hidden') {
            return this;
        }
        //动画进行中
        this.state = 'ing';
        //如果页面还没有浮窗节点则创建html节点（如果页面没有生成DOM，则getComputedStyle的属性为空）
        if (!getComputedStyle(this.targetEl).display) {
            document.body.appendChild(this.targetEl);
        }

        //获取内容部分的高度，用于slideDown和slideUp
        this.wrapperHeight = axArrHeight(this.wrapper).height;

        //设置打开关闭动画时间
        if (this.options.duration) {
            this.duration = this.options.duration;
        } else if (this.wrapperHeight) {
            this.duration = Math.floor(this.wrapperHeight / 3 + 250);
        } else {
            this.duration = parseFloat(axStyle(this.targetEl).animationDuration) * 1000
        }

        //页面加载完毕会出现跳动，导致定位不准确，显示前先更新位置
        this.popupPosition.update();


        //执行窗口打开动画
        this.targetEl.classList.add('ax-show');
        //如果使用按钮触发气泡，那么给气泡赋予样式
        this.button ? this.button.classList.add(this.options.btnClass) : null;

        //回调
        this.options.onShow && this.options.onShow.call(this);
        'show' in this.handlers ? this.emit('show', '') : null;

        //一般css3动画展示
        if (this.options.in === 'slideDown') {
            //特别针对slideDown展示
            axHeightDown(this.wrapper, { height: this.wrapperHeight }, '', this.duration)
        } else {
            //如果设置了动画时间参数则按用该值
            this.options.duration ? this.targetEl.style.animationDuration = `${this.options.duration}ms` : null;
            //如何设置了动画名则用该动画
            this.options.in ? this.targetEl.style.animationName = this.options.in : null;
        }

        this.targetEl.style.cssText += `;${this.oldStyle}`;

        //如果是使用了hover触发方式则修改isActive状态
        this.hover ? this.hover.isActive = true : null;

        //完全打开执行
        setTimeout(() => {
            //更新显示状态
            this.state = 'shown'
            //回调
            this.options.onShown && this.options.onShown.call(this);
            'shown' in this.handlers ? this.emit('shown', '') : null;
        }, this.duration);
        callback && callback.call(this);

        //返回实例，以便于链式操作
        return this;
    }
    hide(callback) {
        //如果已经销毁，或者本身就是关闭状态则不再执行
        if (this.destroyed || this.state !== 'shown') {
            return this;
        }
        //动画进行中
        this.state = 'ing';
        //只有stay为false才能关闭
        if (!this.options.stay) {
            //回调
            this.options.onHide && this.options.onHide.call(this);
            'hide' in this.handlers ? this.emit('hide', '') : null;

            //如果使用按钮触发气泡，那么删除样式
            this.button ? this.button.classList.remove(this.options.btnClass) : null;

            if (this.options.out === 'slideUp') {
                //特别针对slideDown展示
                axHeightUp(this.wrapper, { height: this.wrapperHeight }, '', this.duration)
            } else {
                this.options.out ? this.targetEl.style.animationName = this.options.out : null;
            }

            this.targetEl.style.cssText += `;${this.oldStyle}`;

            //关闭窗口同时暂停媒体播放
            this.targetEl.querySelectorAll('video,audio').forEach(elem => {
                elem.pause();
            });

            //如果是使用了hover触发方式则修改isActive状态
            this.hover ? this.hover.isActive = false : null;

            //延迟执行（等待关闭动画完成）
            setTimeout(() => {
                //更新显示状态
                this.state = 'hidden';
                //移除样式，取消visible状态
                this.targetEl.classList.remove('ax-show');
                //移除节点
                this.targetEl.remove();
                //回调
                this.options.onHidden && this.options.onHidden.call(this);
                'hidden' in this.handlers ? this.emit('hidden', '') : null;
                //动画结束，恢复wrapper高度
                this.options.out === 'slideUp' ? this.wrapper.style.height = 'auto' : null;
            }, this.duration);
            callback && callback.call(this);
        }
        //返回实例，以便于链式操作
        return this;
    }
    destroy(callback) {
        //删除对应的DOM
        this.targetEl.remove();
        //清空popper实例
        this.popupPosition.destroy();
        this.popupPosition = null;
        //清空hover实例
        this.hover = null;
        //取消绑定事件
        this.button.removeEventListener(this.options.trigger, this.triggerShow);
        this.rel.forEach(i => {
            i.removeEventListener(this.options.triggerRel || this.options.trigger, this.triggerShow);
        });
        this.toggleShow ? document.removeEventListener('click', this.toggleShow) : null;
        this.bound = false;
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        this.submitXhr ? this.submitXhr.abort() : null;
        //监听回调
        this.options.onDestroy && this.options.onDestroy.call(this);
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    //对使用axPopup属性的元素使用本插件，可窗口嵌套
    function eachInstance(arr, parent) {
        arr.forEach(elem => {
            let instance = new axPopup(elem, '');
            parent ? instance.targetEl.style.zIndex = parseInt(getComputedStyle(parent.targetEl).zIndex) + 1 : null;
            eachInstance(instance.targetEl.querySelectorAll('[axPopup]'), instance);
        });
    }
    let arrInstance = document.querySelectorAll('[axPopup]');
    eachInstance(arrInstance);
})(); */
/*!
 * 插件：小窗文字提示；使用方法：axTooltip('#id',{参数})
 */
class axTooltip {
    constructor(elem, options) {
        //列表Dom
        this.button = axIdToDom(elem);
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            placement: 'top', //窗口相对按钮的位置，可填top|top-start|top-end|bottom|bottom-start|bottom-end|left|left-start|left-end|right|right-start|right-end|center
            arrow: true, //是否显示指示箭头，默认true显示，可填false不显示
            trigger: 'hover', //显示窗口的动作，默认hover显示，可填click点击打开或初始化即展开sticky，或者不显示none
            size: '', //窗口尺寸，默认正常尺寸，可填sm,md,lg
            multiple: true, //是否允许打开多个窗口，默认允许
            theme: 'text', //提示窗风格，可选择primary、success、error、warning、info、question、text（默认）、brief、ignore和light（或为空）
            content: '', //提示内容，纯文本格式
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
        }, options, this.button, this.constructor.name);
        //初始化
        this.init();
    }
    init() {
         //添加到实例合集
         axInstance.push(this, this.options.insName, 'tooltip');
          //断点合并参数
          !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //axTooltip="伟大的中国"，这样也可以取得提示文字
        let btnAttr = this.button.getAttribute('axTooltip');
        //一顿操作后content依然为为空并且有ax*属性
        if (!this.options.content && btnAttr) {
            //如果属性是有键值对的冒号，则不从其中取值
            !btnAttr.includes(':') ? this.options.content = btnAttr : null;
        }

        let popOptions = Object.assign({
            type: 'html',
            close: false,
            footerShow: false,
            padding: false,
        }, this.options);
        this.popup = new axPopup(this.button, popOptions);
        //内容Dom节点
        this.content = this.popup.content;
        this.setAttribute();
    }
    setAttribute() {
        this.popup.targetDom.classList.add('ax-tooltip');
        if (this.options.theme) {
            this.popup.targetDom.setAttribute('theme', this.options.theme)
        }
    }
    updateContent(str, callback) {
        this.content.innerHTML = str;
        callback && callback.call(this, update, before, this.content.innerHTML, this.options.content);
        'updated' in this.handlers ? this.emit('updated', '') : null;
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/*!
 * 插件：文件上传；使用方法：new axUpload('#id',{参数})
 */
class axUpload {
    constructor(elem, options) {
        //列表Dom
        this.targetDom = elem ? axIdToDom(elem) : '';
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            type: 'beautify', //文件回显方式，beautify原始文件域，text普通文件列表，info信息列表，pictrue图片列表，gallery多图画廊
            accept: 'jpg,jpeg,gif,png,svg,doc,docx,xls,xlsx,txt,mp4,mp3,pdf', //支持的文件类型（后缀），多个用英文逗号隔开
            mime: '',//打开窗口后所能显示的文件类型，比如填"image/*"在窗口只会显示图片类型的文件
            multiple: true, //是否支持多选，默认true支持，可选择false仅单选
            name: '', //隐藏的input控件的name
            separator: ',',//填入input隐藏域值的分割符，默认是英文逗号
            columns: 0,//在gallery模式下设定列数，默认为0表示不设置
            min: 0, //至少上传多少个文件，0表示不限制
            max: 0, //最多能上传多少个文件，0表示不限制
            size: 0, //尺寸，单位MB，0表示不限制
            url: '', //提交的动态页面地址
            fileType: [], //文件对象的图标样式，[{name:'',suffix:'',icon:''}]
            action: 'auto', //选好文件后的提交方式，默认是auto选好之后就自动上传，可选择manual手动点击某按钮上传
            tipsShow: true, //是否显示上传要求文字信息，默认true显示，可选择false不显示
            summaryShow: false,//是否显示上传的统计信息，默认fale不显示，可选择true显示
            cloud: '',//是否直传到对象存储服务器，默认为空即上传到本地服务器，目前可填upcloud，其他云平台用户自行扩展，
            //cloud:{name:'upcloud',field:'file',code:200,server:'https://src.axui.cn/',append:{policy:'',signature:''},keys:{name:'',url:'url',size:'file_size',time:'time'}}
            pastable: {
                before: '请点击使用ctrl+v粘贴上传',
                ing: '使用ctrl+v粘贴进来',
                after: 'ctrl+v粘贴上传成功',
            },//是否允许粘贴上传，默认允许，如果需要禁止，可以让该参数为空或者为false
            icons: {
                file: 'ax-iconfont ax-icon-file-text',//默认文件图标
                attach: 'ax-iconfont ax-icon-attach',//默认附件图标
                image: 'ax-iconfont ax-icon-image',//图片图标
                video: 'ax-iconfont ax-icon-video',//视频图标
                audio: 'ax-iconfont ax-icon-music'//音频图标
            },//常用图标
            beautify: {
                placeholder: '请选择文件...', //美化file的input提示文字
                text: '选择文件', //美化file的上传按钮文字
                className: 'ax-file',//美化file的样式名
                display: '',//选择的图片存放的预览位置（适合单个文件选择），#id、className、node均可
                callback: '',//选择文件后的回调函数，支持参数files
            },
            message: {
                passed: '通过文件校验!',
                exceeded: '超过了文件数量限制!',
                tooLarge: '文件体积太大了!',
                errorSuffix: '不支持该文件格式!',
                success: '上传成功!',
                failed: '提交地址可能错误!',
            },//定义实例中的提示文字
            chooseBtn: {
                icon: '<i class="ax-iconfont ax-icon-plus"></i>',
                text: '选择文件',
                className: '',
                props: {},
                show: true,
            },//定义选择文件按钮属性
            uploadBtn: {
                icon: '<i class="ax-iconfont ax-icon-upload"></i>',
                text: '立即上传',
                className: '',
                props: {},
                show: false,
            },//定义上传文件按钮属性
            clearBtn: {
                icon: '<i class="ax-iconfont ax-icon-trash"></i>',
                text: '删除全部',
                className: '',
                props: {},
                show: false,
            },//定义清除文件按钮属性
            cubeBtn: {
                icon: '<i class="ax-iconfont ax-icon-plus-t"></i>',
                text: '',
                className: '',
                props: {},
                show: true,
            },//定义picture添加文件按钮属性
            boxBtn: {
                icon: '<i class="ax-iconfont ax-icon-addpic"></i>',
                text: '点击或拖拽上传',
                className: '',
                props: {},
                show: true,
            },//定义gallery选择文件按钮属性
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onRendered: '',//文件DOM节点渲染完成执行，支持参数obj，即该文件详细
            onPasted: '',// 对上传区域使用粘贴后执行，支持参数file，即剪切板的文件
            onDropped: '',// 拖放上传区域后执行，支持参数files，即拖来的文件
            onUploading: '', //文件列表开始上传执行，无参数
            onUploaded: '', //上传成功后的回调函数，支持参数extract（当前文件详细{file:'',dom:'',state:''}）和支持参数raw即后台返回的原始数据
            onAllUploaded: '', //所有有效文件都上传后执行，支持参数extract（当前文件详细{file:'',dom:'',state:''}）和支持参数raw即后台返回的原始数据
            onError: '', //发生错误时回调函数，支持参数extract（当前文件详细{file:'',dom:'',state:''}）和支持参数raw即后台返回的原始数据
            onCompleted: '', //文件100%上传后执行，支持参数obj，即该文件详细，注意即使返回错误也可以执行该监听事件
            onAllCompleted: '', //所有有效文件100%上传后执行，无参数，注意即使返回错误也可以执行该监听事件
            onChanged: '', //文件改变时的回调函数，支持参数files（所有通过校验的文件）
            onShow: '', //初始化回显时的回调函数，支持参数details（当前正在回显的文件详细）
            onRemove: '', //文件删除后的回调函数，支持参数file（当前删除的文件）
            onClear: '', //文件全部清空时的回调函数，支持参数files（所有通过校验的文件）
            onBeforeRemove: '', //文件删除前的回调函数，支持参数obj（当前删除的文件详细{file:'',dom:'',state:''}）
            onBeforeClear: '', //文件清空前的回调函数，无参数
            onBeforeUpload: '',//上传前确认是否上传，如果返回true则上传，如果返回false则停止上传，支持参数files（所有通过校验待上传的文件）
        }, options, this.targetDom, this.constructor.name);
        //修正name值
        this.name = this.targetDom && this.targetDom.nodeName == 'INPUT' ? this.targetDom.getAttribute('name') : this.options.name;
        // 定义监听事件成员
        this.handlers = {};

        this.sizeLimit = this.options.size * 1024 * 1024; //从MB换算成b

        //定义可用文件类型
        let dftTypes = [{
            name: 'image',
            icon: this.options.icons.image,
            suffix: ['png', 'jpg', 'jpeg', 'bmp', 'gif', 'webp', 'svg', 'tiff']
        }, {
            name: 'video',
            icon: this.options.icons.video,
            suffix: ['mp4', 'mpeg', 'avi', 'wmv', 'mov', 'm4v', 'flv', 'rmvb', '3gp']
        }, {
            name: 'audio',
            icon: this.options.icons.audio,
            suffix: ['mp3', 'wma', 'm4a', 'ogg']
        }];
        if (!axIsEmpty(this.options.fileType)) {
            this.fileTypes = [...this.options.fileType, ...dftTypes];
        } else {
            this.fileTypes = [...dftTypes];
        }

        //参与列表的数据(只有文件)
        this.files = [];
        //参与列表的详细数据，（包括文件、dom和状态）
        this.details = [];
        //选择过的历史文件，包括报错的文件
        this.history = [];
        //最终获得的文件路径值
        this.value = [];
        this.inputValue = '';
        //汇总节点
        this.summary = axAddElem('div', { summary: '' });
        //初始化
        this.init();

    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'upload');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //创建上传要求，获得this.tips
        this.createTips();
        //创建上传按钮，获得this.chooseBtn
        this.createButton();
        //创建主体结构，获得this.parent，this.inputFile，this.inputText，this.list
        this.createSkeleton();
        //设置属性
        this.setAttribute();
        //如果是file类型，到此就结束了
        if (this.options.type == 'beautify') {
            return false;
        }
        //点击按钮选择文件
        this.chooseBtn.onclick = () => {
            this.inputFile.click();
        }
        //监听file的change事件，渲染list并填充this.details
        this.readyUpload();

        return this;
    }
    createTips() {
        this.tipsText = '';
        this.options.size ? this.tipsText += `要求大小不超过${this.options.size}MB,` : null;
        this.options.min ? this.tipsText += `至少上传${this.options.min}个文件,` : null;
        this.options.max ? this.tipsText += `最多上传${this.options.max}个文件` : null;
        this.tips = axAddElem('div', { tips: '' }, this.tipsText);
    }
    createButton() {
        if (this.options.type == 'gallery') {
            this.chooseBtn = axAddElem('div', Object.assign({ choose: '' }, this.options.boxBtn.props), this.options.boxBtn.icon + '<div text>' + this.options.boxBtn.text + '</div>');
        } else if (this.options.type == 'picture') {
            this.chooseBtn = axAddElem('li', Object.assign(this.options.cubeBtn.className ? { class: this.options.cubeBtn.className } : { choose: '' }, this.options.cubeBtn.props), this.options.cubeBtn.icon + this.options.cubeBtn.text);
        } else {
            this.chooseBtn = axAddElem('a', Object.assign({ href: '###', class: this.options.chooseBtn.className ? 'ax-btn ' + this.options.chooseBtn.className : 'ax-btn', choose: '' }, this.options.chooseBtn.props), this.options.chooseBtn.icon + this.options.chooseBtn.text);
        }
    }
    createSkeleton() {
        if (this.targetDom) {
            //target分两种情况，一种是input一种是div
            if (this.targetDom.nodeName == 'INPUT' && this.targetDom.type == 'file') {
                //beautify类型
                this.inputFile = this.targetDom;
                this.inputFile.name ? this.inputFile.setAttribute('name', this.name) : null;
                //创建父节点并插入到页面
                this.parent = axAddElem('div', { class: 'ax-upload' });
                this.inputFile.insertAdjacentElement('beforeBegin', this.parent);
            } else if (this.targetDom.nodeName == 'INPUT' && (this.targetDom.type == 'hidden' || this.targetDom.type == 'text')) {
                //如果目标节点是隐藏的文本域
                this.inputText = this.targetDom;
                this.inputText.name ? this.inputText.setAttribute('name', this.name) : null;
                this.inputText.setAttribute('readonly', '');
                //创建选择input
                this.inputFile = axAddElem('input', { type: 'file' });
                //创建父节点并插入到页面
                this.parent = axAddElem('div', { class: 'ax-upload' });
                this.inputText.insertAdjacentElement('beforeBegin', this.parent);
                this.options.multiple ? this.inputFile.setAttribute('multiple', '') : null;
            } else {
                //如果目标节点是一个非input元素
                this.parent = this.targetDom;
                this.inputFile = axAddElem('input', { type: 'file' });
                this.inputText = axAddElem('input', { type: 'hidden', readonly: '', name: this.name });
                this.options.multiple ? this.inputFile.setAttribute('multiple', '') : null;
            }
        } else {
            //在内存中创建
            this.parent = axAddElem('div', { class: 'ax-upload' });
            this.inputFile = axAddElem('input', { type: 'file' });
            this.options.multiple ? this.inputFile.setAttribute('multiple', '') : null;
            this.inputText = axAddElem('input', { type: 'hidden', readonly: '', name: this.options.name });
        }
        //添加mime类型
        this.options.mime ? this.inputFile.setAttribute('accept', this.options.mime) : null;
        //包裹选择域
        this.parent.appendChild(this.inputFile);

        //根据类型
        if (this.options.type == 'beautify') {
            //美化类型
            this.parent.appendChild(this.inputFile);
            this.parent.dataset.placeholder = this.options.beautify.placeholder;
            this.parent.dataset.text = this.options.beautify.text;
            this.parent.classList.add(this.options.beautify.className);
            //专门针对此类型
            this.fileChange(this.parent, this.inputFile);
        } else {
            //其他类型
            this.list = axAddElem('ul');
            this.parent.appendChild(this.list);
            this.parent.appendChild(this.inputText);
            let info = axAddElem('div', { info: '' }),
                footer = axAddElem('div', { footer: '' }, info);
            //显示限制提示
            if (this.options.tipsShow && this.tipsText) {
                info.appendChild(this.tips);
                this.parent.setAttribute('withtips', '');
            }
            //显示统计
            if (this.options.summaryShow) {
                info.appendChild(this.summary);
                this.summary.innerHTML = this.getSummary();
                this.parent.setAttribute('withtips', '');
            }
            if (this.options.type == 'gallery') {
                //gallery可使用ctrlv
                this.createCtrlv();
                this.parent.appendChild(footer);
                this.parent.insertAdjacentElement('afterBegin', this.chooseBtn);
                //gallery模式的列表使用grid布局
                this.list.classList.add('ax-grid', 'ax-space-xs');
                this.options.columns ? this.list.classList.add('ax-grid-' + this.options.columns) : null;
                //清除按钮
                if (this.options.clearBtn.show) {
                    this.clearBtn = axAddElem('a', Object.assign({ href: '###', class: this.options.clearBtn.className ? 'ax-btn ' + this.options.clearBtn.className : 'ax-btn', clear: '' }, this.options.clearBtn.props), this.options.clearBtn.icon + this.options.clearBtn.text);
                    footer.appendChild(this.clearBtn);
                    this.parent.setAttribute('withtips', '');
                }
                //上传按钮
                if (this.options.uploadBtn.show) {
                    this.uploadBtn = axAddElem('a', Object.assign({ href: '###', class: this.options.uploadBtn.className ? 'ax-btn ' + this.options.uploadBtn.className : 'ax-btn', upload: '' }, this.options.uploadBtn.props), this.options.uploadBtn.icon + this.options.uploadBtn.text);
                    footer.appendChild(this.uploadBtn);
                    this.parent.setAttribute('withtips', '');
                }
            } else if (this.options.type == 'picture') {
                this.parent.appendChild(footer);
                this.list.appendChild(this.chooseBtn);
                if (this.options.clearBtn.show) {
                    this.clearBtn = axAddElem('li', Object.assign({ class: this.options.clearBtn.className, clear: '' }, this.options.clearBtn.props), this.options.clearBtn.icon + this.options.clearBtn.text);
                    this.chooseBtn.insertAdjacentElement('afterEnd', this.clearBtn);
                }
                if (this.options.uploadBtn.show) {
                    this.uploadBtn = axAddElem('li', Object.assign({ class: this.options.uploadBtn.className, upload: '' }, this.options.uploadBtn.props), this.options.uploadBtn.icon + this.options.uploadBtn.text);
                    this.chooseBtn.insertAdjacentElement('afterEnd', this.uploadBtn);
                }

            } else {
                let header = axAddElem('div', { header: '' });
                this.list.insertAdjacentElement('beforeBegin', footer);
                this.parent.insertAdjacentElement('afterBegin', header);
                header.appendChild(this.chooseBtn);
                if (this.options.uploadBtn.show) {
                    this.uploadBtn = axAddElem('a', Object.assign({ href: '###', class: this.options.uploadBtn.className ? 'ax-btn ' + this.options.uploadBtn.className : 'ax-btn', upload: '' }, this.options.uploadBtn.props), this.options.uploadBtn.icon + this.options.uploadBtn.text);
                    header.appendChild(this.uploadBtn);
                }
                if (this.options.clearBtn.show) {
                    this.clearBtn = axAddElem('a', Object.assign({ href: '###', class: this.options.clearBtn.className ? 'ax-btn ' + this.options.clearBtn.className : 'ax-btn', clear: '' }, this.options.clearBtn.props), this.options.clearBtn.icon + this.options.clearBtn.text);
                    header.appendChild(this.clearBtn);
                }
            }

        }
    }
    createCtrlv() {
        if (this.options.pastable) {
            this.ctrlv = axAddElem('i', { ctrlv: '' }, this.options.pastable.before);
            this.parent.appendChild(this.ctrlv);
            this.parent.onclick = () => {
                this.ctrlv.innerHTML = this.options.pastable.ing;
            }
            //监听paste事件在readyUpload方法里
            this.parent.onmouseleave = () => {
                this.ctrlv.innerHTML = this.options.pastable.before;
            }
        }
    }
    sizeStr(data) {
        let str = ''
        if (data / 1024 >= 1024) {
            str = '<i>' + (data / (1024 * 1024)).toFixed(2) + '</i>MB'
        } else if (data / 1024 >= 1) {
            str = '<i>' + (data / 1024).toFixed(2) + '</i>KB'
        } else {
            str = '<i>' + data + '</i>B'
        }
        return str;
    }
    getSummary() {
        let originalSize = 0,
            realSize,
            totalNum = this.details.length,
            uploadedItems = this.details.filter(i => i.progress == 'uploaded'),
            okNum = uploadedItems.length;
        uploadedItems.forEach(i => {
            originalSize += i.file.size;
        });
        realSize = this.sizeStr(originalSize);
        return `提交了<i>${totalNum}</i>个文件,成功上传了<i>${okNum}</i>个,共${realSize}`;
    }
    createItem(file, state) {
        let tpl = '';
        if (this.options.type == 'gallery') {
            tpl = `
            <li valid=${(state && !state.valid) ? "error" : 'success'}>
                <div file>
                    <div overlay>
                        <i view class="ax-iconfont ax-icon-zoomin"></i>
                        <a down href="###" class="ax-iconfont ax-icon-download"></a>
                        <i remove class="ax-iconfont ax-icon-trash"></i>
                    </div>
                    <div track><div bar></div></div>
                    <i percent></i>
                    <div message>${(state && !state.valid) ? state.message : ''}</div>
                </div>
                <div info>
                    <div title>${file.name}</div>
                    <div size>${this.sizeStr(file.size)}</div>
                    <div url></div>
                </div>
            </li>
            `;
        } else if (this.options.type == 'picture') {
            tpl = `
            <li valid=${(state && !state.valid) ? "error" : 'success'}>
                <div file></div>
                <div overlay>
                    <i view class="ax-iconfont ax-icon-zoomin"></i>
                    <a down href="###" class="ax-iconfont ax-icon-download"></a>
                    <i remove class="ax-iconfont ax-icon-trash"></i>
                </div>
                <div track><div bar></div></div>
                <i percent></i>
                <div message>${(state && !state.valid) ? state.message : ''}</div>
            </li>
            `;
        } else if (this.options.type == 'info') {
            tpl = `
            <li valid=${(state && !state.valid) ? "error" : 'success'}>
                <a href="###" file></a>
                <div info>
                    <a href="###" title>${file.name}</a>
                </div>
                <div track><div bar></div></div>
                <i percent></i>
                <i message>${(state && !state.valid) ? state.message : ''}</i>
                <i remove class="ax-iconfont ax-icon-trash"></i>
            </li>
            `;
        } else {
            tpl = `
            <li valid=${(state && !state.valid) ? "error" : 'success'}>
                <i icon class="${this.options.icons.attach}"></i>
                <div info>
                    <a href="###" title>${file.name}</a>
                </div>
                <div track><div bar></div></div>
                <i percent></i>
                <i message>${(state && !state.valid) ? state.message : ''}</i>
                <i remove class="ax-iconfont ax-icon-trash"></i>
            </li>
            `;
        }
        let li = axStrToDom(tpl),
            fileDom = li.querySelector('[file]');
        //填充文件头
        fileDom ? this.fillHeader(file, fileDom) : null;
        //追加进列表
        if (this.options.type == 'picture') {
            //图片列表的添加按钮始终在列表最后面
            this.chooseBtn.insertAdjacentElement('beforeBegin', li);
        } else {
            this.list.appendChild(li);
        }

        return li;
    }
    fileChange(item, file) {
        let _this = this;
        let text = item.getAttribute('data-text'),
            placeholder = item.getAttribute('data-placeholder');
        if (!text) {
            item.classList.add('ax-simple');
        }
        if (file.classList.contains("ax-xs")) {
            item.classList.add('ax-xs');
        } else if (file.classList.contains('ax-sm')) {
            item.classList.add('ax-sm');
        } else if (file.classList.contains('ax-lg')) {
            item.classList.add('ax-lg');
        }
        //获得图片存放位置
        let imgWrapper;
        if (this.options.beautify.display) {
            imgWrapper = axIdToDom(this.options.beautify.display);
        }
        //监听input值变化
        file.addEventListener('change', function () {
            item.classList.add('ax-active');
            item.setAttribute('data-placeholder', this.value.replace(/.*(\/|\\)/, ''));
            if (!this.value) {
                item.setAttribute('data-placeholder', placeholder);
                item.classList.remove('ax-active');
            }
            //base64回显图片
            if ([...this.files].length > 0) {
                //获得返回值
                [...this.files].forEach(i => {
                    _this.value.push({ name: i.name, size: i.size, time: i.lastModified, url: '' })
                });
                if (imgWrapper) {
                    //先清空
                    imgWrapper.innerHTML = '';
                    [...this.files].forEach(i => {
                        if (i.type.startsWith('image')) {
                            _this.readFile(i, (data) => {
                                imgWrapper.appendChild(axAddElem('img', { src: data }));
                            });
                        }
                    });
                }

            }
            //回调
            _this.options.beautify.callback && _this.options.beautify.callback.call(_this, this.files);
            _this.options.onChange && _this.options.onChange.call(_this, this.files);
            //监听
            'changed' in _this.handlers ? _this.emit('changed', this.files) : null;
            _this.options.onChanged && _this.options.onChanged.call(_this, this.files);
        }, false);
        file.onfocus = () => {
            item.classList.add("ax-focus");
        };
        file.onblur = () => {
            item.classList.remove("ax-focus");
        };
    }
    setAttribute() {
        //设置类别属性
        this.options.type ? this.parent.setAttribute('type', this.options.type) : null;
        this.options.pastable ? this.parent.setAttribute('pastable', '') : null;
        //设置插件状态
        this.parent.setAttribute('state', 'ready');
    }
    readyUpload() {
        //监听input取值
        this.inputFile.addEventListener('change', () => {
            let files = [...this.inputFile.files];
            //渲染ul列表
            this.renderList(files);
            //获取了文件则清空input，等待下次触发change事件
            this.inputFile.value = '';
            //监听
            'changed' in this.handlers ? this.emit('changed', files) : null;
            this.options.changed && this.options.changed.call(this, files);
        }, false);
        //监听粘贴取值
        if (this.options.pastable) {
            this.parent.addEventListener('paste', (e) => {
                //获得剪切板数据
                let data = e.clipboardData ? e.clipboardData : window.clipboardData;
                if (data) {
                    //只取第一个
                    let item = data.items[0];
                    if (item && item.kind === 'file' && item.type.includes('image')) {
                        //只对图片有效
                        this.renderList([item.getAsFile()]);
                        //更新ctrv提示
                        this.ctrlv ? this.ctrlv.innerHTML = this.options.pastable.after : null;
                    }
                    //监听
                    'pasted' in this.handlers ? this.emit('pasted', item) : null;
                    this.options.onPasted && this.options.onPasted.call(this, item);
                }
            }, false);
        }
        //监听拖拽取值
        this.parent.addEventListener("dragleave", (e) => {
            e.preventDefault();
            this.parent.removeAttribute('drag');
        });
        this.parent.addEventListener("dragover", (e) => {
            e.preventDefault();
        });
        this.parent.addEventListener("dragenter", (e) => {
            e.preventDefault();
            //追加drag属性以响应拖入
            this.parent.setAttribute('drag', '');
        });
        this.parent.addEventListener("drop", (e) => {
            e.preventDefault();
            //移除drag属性
            this.parent.removeAttribute('drag');
            //获取文件对象 
            //let files = e.dataTransfer.files;
            let items = e.dataTransfer.items;
            for (let i = 0; i < items.length; i++) {
                let item = items[i];
                if (item.kind === "file") {
                    //FileSystemFileEntry或FileSystemDirectoryEntry对象
                    let entry = item.webkitGetAsEntry();
                    this.getEntries(entry);
                }
            }
            //监听
            'dropped' in this.handlers ? this.emit('dropped', items) : null;
            this.options.onDropped && this.options.onDropped.call(this, items);
        }, false);
        //清空操作
        if (this.clearBtn) {
            this.clearBtn.onclick = () => {
                //如果列表是空的就停止执行
                if (this.details.length == 0) {
                    return false;
                }
                if (this.options.onBeforeClear) {
                    this.options.onBeforeClear.call(this);
                } else {
                    this.clear.call(this);
                }
            }
        }
    }
    renderItem(file) {
        //在列表追加节点
        let elem,
            fileValid = this.fileValid(file),
            obj = { file: file, state: fileValid },
            render = (obj, elem) => {
                obj.dom = elem;
                obj.progress = 'rendered';
                elem.setAttribute('progress', 'rendered');
            };
        //填充可用的文件并报错
        if (fileValid.valid) {
            //如果校验通过
            //填充可用文件
            this.files.push(file);
            //追加成功节点
            elem = this.createItem(file);
        } else {
            //如果校验不通过则创建error节点
            elem = this.createItem(file, fileValid);
        }
        //更新属性
        render(obj, elem);
        //追加更新文件详细
        this.details.push(obj);
        //删除项目
        let removeDom = elem.querySelector('[remove]');
        //点击删除按钮删除项目并执行回调
        if (removeDom) {
            removeDom.onclick = () => {
                if (this.options.onBeforeRemove) {
                    this.options.onBeforeRemove.call(this, obj);
                } else {
                    this.remove(obj.file);
                }
            }
        }
        //this.details= [{dom:'',file:'',progress:'',state:{valid:'',message:''}}]
        //监听
        'rendered' in this.handlers ? this.emit('rendered', obj) : null;
        this.options.onRendered && this.options.onRendered.call(this, obj);
        return elem;
    }
    getEntries(entry) {
        //entry.file方法是异步的
        if (entry.isFile) {
            //是一个文件
            entry.file(i => this.renderList([i]));
        } else {
            //是一个文件夹
            let reader = entry.createReader();
            reader.readEntries(k => {
                k.forEach(i => this.getEntries(i));
            }
            );
        }
    }
    renderList(files, callback) {
        if (files.length == 0) {
            return false;
        }
        for (let x = 0, len = files.length; x < len; x++) {
            let item = files[x];
            //填充历史文件
            if (this.history.length == 0) {
                this.history.push(item);
                this.renderItem(item);
            } else {
                //通过文件名和文件大小判断是否是同一个文件
                if (this.history.every(i => (i.name != item.name && i.size != item.size))) {
                    this.history.push(item);
                    this.renderItem(item);
                } else {
                    continue;
                }
            }
        }
        //如果是自动上传模式
        if (this.options.action == 'auto') {
            this.upload();
        } else {
            //点击提交按钮上传
            if (this.uploadBtn) {
                this.uploadBtn.onclick = () => {
                    this.upload();
                }
            }
        }
        //设置插件状态
        this.parent.setAttribute('state', 'processing');
        //添加文件回调
        callback && callback.call(this);
    }
    upload(callback) {
        //检测是否填了url值，type=beautify由于没有url所以不会执行以下代码
        if (!this.options.url) {
            console.error('The url is required!');
            return false;
        }
        //全局校验可上传文件的数量
        let globalValid = this.globalValid();
        if (!globalValid.valid) {
            new axMessage({
                content: globalValid.message,
                result: 'error',
            }).show();
            return false;
        }
        //拦截上传
        //progress属性不等于uploaded才上传，如果是在上传中或已经上传的则取消上传，如此是避免用户反复上传
        //还需要校验成功的才可上传
        let uploadingFiles = this.getUploadingFiles();
        if (this.options.onBeforeUpload && !this.options.onBeforeUpload.call(this, uploadingFiles)) {
            return false;
        }
        //this.details=this.history+dom+校验
        //this.details=[{file:'',dom:'',state:{valid:true,message:'通过校验'}]
        uploadingFiles.forEach(i => {
            this.uploadFile(i);
        });
        /* for (let i = 0, len = this.details.length; i < len; i++) {
            let item = this.details[i];
            if (!item.state.valid) {
                continue;
            }
            //progress属性不等于uploaded才上传，如果是在上传中或已经上传的则取消上传，如此是避免用户反复上传
            item.progress != 'uploaded' ? this.uploadFile(item) : null;
        } */
        //监听
        'uploading' in this.handlers ? this.emit('uploading', '') : null;
        this.options.onUploading && this.options.onUploading.call(this);
        //上传回调
        callback && callback.call(this);
    }
    getUploadingFiles() {
        return this.details.filter(i => i.state.valid && i.progress != 'uploaded');
    }
    activeOperate(file, dom, url, name) {
        //将dom设为success状态
        dom.setAttribute('valid', 'success');
        dom.setAttribute('progress', 'uploaded');
        //异步成功后会执行多次，所以在此判断是否存在state，不存在才添加
        if (!dom.querySelector('[state]')) {
            dom.querySelector('[percent]').insertAdjacentHTML('afterend', `<i state class="ax-iconfont ax-icon-check"></i>`);
        }
        //使用云存储直传
        if (this.options.cloud) {
            url = this.options.cloud.server + url;
        }
        let downBtn = dom.querySelector('[down]'),
            viewBtn = dom.querySelector('[view]');
        //下载按钮
        if (downBtn) {
            //使用云存储直传
            //如果url是跨域的，那么download失效
            if (this.options.cloud) {
                downBtn.onclick = () => {
                    this.downloadFile(url, file.name);
                }
            } else {
                downBtn.setAttribute('download', '');
                downBtn.setAttribute('href', url);
            }
        }
        //预览按钮
        if (viewBtn) {
            let mediaSrc = { media: url, caption: name, type: '' };
            if (this.h5Type(file) == 'image') {
                mediaSrc.type = 'image';
            } else if (this.h5Type(file) == 'video') {
                mediaSrc.type = 'video';
            } else if (this.h5Type(file) == 'audio') {
                mediaSrc.type = 'audio';
            }
            if (mediaSrc.type) {
                //打开灯箱
                viewBtn.onclick = function () {
                    new axLightbox({
                        src: mediaSrc,
                    }).init().show();
                }
            }
        }
        //给普通列表添加链接
        if (this.options.type == 'text' || this.options.type == 'info') {
            let titleDom = dom.querySelector('[title]'),
                fileDom = dom.querySelector('[file]');
            titleDom.setAttribute('href', url);
            titleDom.setAttribute('target', '_blank');
            if (fileDom) {
                fileDom.setAttribute('href', url);
                fileDom.setAttribute('target', '_blank');
            }
        }
    }
    downloadFile(url, fileName) {
        let xhr = new XMLHttpRequest();
        xhr.open("GET", url, true);
        xhr.responseType = 'blob';
        xhr.onload = function () {
            let url = window.URL.createObjectURL(xhr.response),
                dom = axAddElem('a');
            dom.href = url;
            dom.download = fileName;
            dom.click();
        }
        xhr.send();
    }
    uploadFile(obj) {
        let data = new FormData(),
            file = obj.file,
            dom = obj.dom,
            valid = obj.state.valid,
            fixDom = (flag, url, message, dom) => {
                if (flag) {
                    //激活下载和查看
                    this.activeOperate(file, dom, url, file.name);
                } else {
                    //将dom设为error状态
                    dom.setAttribute('valid', 'error');
                    dom.querySelector('[message]').innerText = message;
                }
                //更新汇总
                this.options.summaryShow ? this.summary.innerHTML = this.getSummary() : null;
            }
        //前端校验失败的或已经上传过的禁止上传
        if (!valid || obj.progress == 'uploaded') {
            return false;
        }
        //追加progress属性为uploading，标记正在上传状态
        obj.progress = 'uploading';
        dom.setAttribute('progress', 'uploading');
        //使用云存储直传
        if (this.options.cloud) {
            //必须使用云存储服务指定的字段名，通常为file
            data.append(this.options.cloud.field ? this.options.cloud.field : 'file', file);
            for (let k in this.options.cloud.append) {
                data.append(k, this.options.cloud.append[k]);
            }
        } else {
            data.append(this.name ? this.name : 'file', file);
        }
        axAjax({
            url: this.options.url,
            data: data,
            uploading: (response) => {
                dom.querySelector('[percent]').innerText = response.result;
                dom.querySelector('[bar]').style.width = response.result;
                //上传完毕，修改dom属性方便控制样式
                if (response.percent >= 100) {
                    dom.setAttribute('completed', '');
                    obj.completed = true;
                    //监听
                    this.handlers.hasOwnProperty('completed') ? this.emit('completed', obj) : null;
                    this.options.onCompleted && this.options.onCompleted.call(this, obj);
                    if (this.details.every(i => i.state.valid && i.completed)) {
                        this.handlers.hasOwnProperty('allCompleted') ? this.emit('allCompleted', '') : null;
                        this.options.onAllCompleted && this.options.onAllCompleted.call(this);
                    }
                }
            },
            success: (response) => {
                let msg = '', url = '', size = 0, time = 0, valid,
                    //考虑返回对象可能是二维或三维对象
                    getKeys = (obj, str) => {
                        if (!str) {
                            return '';
                        }
                        if (!str.includes('.')) {
                            return obj[str];
                        }
                        let arr = str.split("."),
                            value;
                        //支持两层对象
                        if (arr.length == 1) {
                            value = obj[arr[0]]
                        } else if (arr.length == 2) {
                            value = obj[arr[0]][arr[1]]
                        } else {
                            value = '';
                        }
                        return value;
                    };

                if (this.options.cloud) {
                    //使用云存储直传
                    let keys = this.options.cloud.keys;
                    url = getKeys(response.content, keys.url);
                    size = getKeys(response.content, keys.size);
                    time = getKeys(response.content, keys.time);
                    msg = getKeys(response.content, keys.message);
                    //通过状态码判断成功与否
                    valid = this.options.cloud.code == response.content.code ? true : false;
                } else {
                    //上传到服务器
                    url = response.content.url;
                    size = file.size;
                    time = response.content.time;
                    msg = response.content.message;
                    valid = response.content.valid === true ? true : false;
                }
                //从this.details调整校验状态
                obj.state.valid = valid;
                obj.state.time = time;
                obj.state.url = url;
                obj.state.message = msg;

                //修正this.details
                if (!valid) {
                    //从this.files删除
                    this.files = this.files.filter(i => i != file);
                    //监听
                    this.handlers.hasOwnProperty('error') ? this.emit('error', obj, response.content) : null;
                    this.options.onError && this.options.onError.call(this, obj, response.content);
                } else {
                    //生成实例返回值
                    this.value.push({ name: file.name, url: url, size: size, time: time });
                    this.valueToValue(this.value);
                    //标记完成上传状态
                    obj.progress = 'uploaded';
                    //监听
                    this.handlers.hasOwnProperty('uploaded') ? this.emit('uploaded', obj, response.content) : null;
                    this.options.onUploaded && this.options.onUploaded.call(this, obj, response.content);
                    if (this.details.every(i => i.state.valid && i.progress == 'uploaded')) {
                        this.handlers.hasOwnProperty('allUploaded') ? this.emit('allUploaded', response.content) : null;
                        this.options.onAllUploaded && this.options.onAllUploaded.call(this, obj, response.content);
                    }
                }
                //修正节点
                fixDom(valid, url, msg, dom);
            },
            error: () => {
                fixDom(false, '', this.options.message.failed, dom);
            }
        });
    }
    valueToValue(value) {
        let arr = [];
        value.forEach(i => {
            arr.push(i.url);
        });
        this.inputValue = arr.join(this.options.separator);
        //对于隐藏域
        if (this.inputText) {
            this.inputText.value = this.inputValue;
            //手动执行change事件
            this.inputText.onchange ? this.inputText.onchange() : null;
        }

    }
    getSuffix(file) {
        return file.name.substring(file.name.lastIndexOf('.') + 1);
    }
    h5Type(file) {
        let mediaType;
        //if (this.fileTypes.find(i => i.name == 'image').suffix.includes(this.getSuffix(file))) {
        if (file.type.startsWith('image')) {
            mediaType = 'image';
        } else if (['mp4', 'ogg', 'webm'].includes(this.getSuffix(file))) {
            mediaType = 'video';
        } else if (['mp3', 'ogg', 'wav'].includes(this.getSuffix(file))) {
            mediaType = 'audio';
        }
        return mediaType;
    }
    readFile(file, callback) {
        let reader = new FileReader();
        reader.readAsDataURL(file);
        reader.onload = (e) => {
            callback(e.target.result);
        }
    };
    fillHeader(file, dom) {
        //上传前获得预览文件
        let elem,
            type = this.h5Type(file),
            iconType = this.fileTypes.filter(i => i.suffix.includes(this.getSuffix(file)))[0];
        if (type == 'image') {
            //如果是图片
            this.readFile(file, (data) => {
                elem = axAddElem('img', { src: data });
                dom.appendChild(elem);
            });
        } else if (type == 'video') {
            //如果是视频
            this.readFile(file, (data) => {
                elem = axAddElem('video', { src: data, controls: '' });
                dom.appendChild(elem);
            });
        } else if (type == 'audio') {
            //如果是音频
            this.readFile(file, (data) => {
                elem = axAddElem('audio', { src: data, controls: '' });
                dom.appendChild(elem);
            });
        } else {
            //其他
            elem = axAddElem('i', { class: iconType ? iconType.icon : this.options.icons.file, icon: '' });
            dom.appendChild(elem);
        }

    }
    fileValid(file) {
        let msg = this.options.message.passed,
            pass = true,
            suffix = this.getSuffix(file);
        //this.sizeLimit=0表示不限制
        if (this.history.length > this.options.max && this.options.max) {
            msg = this.options.message.exceeded;
            pass = false;
        } else if (this.history.length < this.options.min && this.options.min) {
            msg = this.options.message.exceeded;
            pass = false;
        } else if (this.sizeLimit && (file.size > this.sizeLimit)) {
            msg = this.options.message.tooLarge;
            pass = false;
        } else if (!this.options.accept.includes(suffix)) {
            msg = this.options.message.errorSuffix;
            pass = false;
        }
        return { valid: pass, message: msg }
    }
    globalValid() {
        let msg = this.options.message.passed,
            pass = true;
        if (this.history.length < this.options.min) {
            msg = `请至少上传${this.options.min}个文件!`;
            pass = false;
        } else if (this.history.length < this.options.min) {
            msg = `最多只能上传${this.options.max}个文件!`;
            pass = false;
        }
        return { valid: pass, message: msg }
    }
    remove(file, callback) {
        //从hitstory和files中删除
        this.history = this.history.filter(i => i != file);
        this.files = this.files.filter(i => i != file);
        //先删除dom
        let item = this.details.find(i => i.file == file),
            name = file.name,
            size = file.size,
            url = item.state.url,
            time = item.state.time;
        item.dom.remove();
        //再从details中删除
        this.details = this.details.filter(i => i.file != file);
        if (this.details.length == 0) {
            //初始化状态
            this.parent.setAttribute('state', 'ready');
        }
        //更新value和inputValue，time是唯一不会重复的的，所以据此进行筛选
        this.value = this.value.filter(i => i.time != time);
        this.valueToValue(this.value);
        //更新汇总
        this.options.summaryShow ? this.summary.innerHTML = this.getSummary() : null;
        //可以使用回调函数，异步请求服务器删除已经上传的文件
        this.options.onRemove ? this.options.onRemove.call(this, url, name, size) : null;
        'remove' in this.handlers ? this.emit('remove', url, name, size) : null;
        callback && callback.call(this, url, name, size);
    }
    clear(callback) {
        //先清除所有dom
        this.details.forEach(i => {
            i.dom.remove();
        });
        //清空数组
        this.details = [];
        this.history = [];
        this.files = [];
        //值清零
        this.value = [];
        this.valueToValue(this.value);
        //初始化状态
        this.parent.setAttribute('state', 'ready');
        //更新汇总
        this.options.summaryShow ? this.summary.innerHTML = this.getSummary() : null;
        //回调
        this.options.onClear ? this.options.onClear.call(this) : null;
        'clear' in this.handlers ? this.emit('clear', '') : null;
        callback && callback.call(this);
        return this;
    }
    getUrlFile(url, fileName, callback) {
        axAjax({
            url: url,
            responseType: 'blob',
            headers: { 'Accept': '' },
            success: (response) => {
                let file = new File([response.content], fileName, { type: response.content.type });
                callback && callback(file);
            }
        });
    }
    show(data) {
        let newData = [];
        //转成统一格式[{url:'',name:'',time:''}]
        //time可作为文件唯一标识
        if (typeof data == 'string') {
            //',,,'格式
            let arr = data.trim().split(',');
            arr.forEach(i => {
                let obj = { url: i, name: i }
                newData.push(obj);
            });
        } else if (Array.isArray(data) && typeof data[0] == 'string') {
            //['','','']格式
            data.forEach(i => {
                let obj = { url: i, name: i }
                newData.push(obj);
            });
        } else {
            //[{url:'',name:'',time:''},]格式
            newData = data;
        }
        //如果没有数据便不再执行了
        if (newData.length == 0) {
            return false;
        }
        //初始化状态
        this.parent.setAttribute('state', 'processing');
        //生成文件并渲染
        newData.forEach((k,i) => {
            this.getUrlFile(k.url, k.name, (file) => {
                this.history.push(file);
                //this.renderItem中会pushfiles和details
                let elem = this.renderItem(file),
                    item = this.details.find(o => o.file == file);
                //补充属性，确保item是标准的数据格式
                item.progress = 'uploaded';
                item.state.message = this.options.message.success;
                item.state.url = k.url;
                //如果没有传来time，则拿文件的lastModified作为时间戳
                item.state.time = k.time || file.lastModified;
                //激活操作
                this.activeOperate(file, elem, k.url, k.name || k.url);
                //标记完成属性
                elem.setAttribute('completed', '');
                //更新汇总
                this.options.summaryShow ? this.summary.innerHTML = this.getSummary() : null;
                //回调
                if(i === newData.length - 1){
                    this.options.onShow && this.options.onShow.call(this, this.details);
                    this.handlers.hasOwnProperty('show')  ? this.emit('show', this.details) : null;
                }
            })
        });
        //生成实例返回值
        this.details.forEach(i => {
            let obj = { name: i.file.name, url: i.state.url, size: i.file.size, time: i.state.time }
            this.value.push(obj);
        });
        this.valueToValue(this.value);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    document.querySelectorAll('[axUpload]').forEach(element => {
        new axUpload(element);
    });
})(); */
/*!
* 插件：即时消息/警告小窗；使用方法：new axMessage({参数})
*/
const axMessages = [];
class axMessage {
    constructor(options) {
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            className: '',//自定义类名，默认为空，注意不需要使用‘.’
            placement: 'center-top',//弹出位置，默认中间顶部，可选择left-top|center-top|right-top|left-center|center-center|right-center|left-bottom|center-bottom|right-bottom
            content: '',//窗口内容，如果为空则插件根据窗口类型自动填入内容，可自定义窗口内容
            result: 'info',//操作状态，默认提示info，可选择success|error|question|warning
            delay: 3000,//延迟关闭时间
            progress: true,//是否显示关闭进度条，仅在complex风格中使用，默认为true
            title: {//设置标题名称
                warning: '操作警告！',
                success: '操作成功！',
                error: '操作失败！',
                question: '操作疑问！',
                info: '信息提示！',
            },
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            close: false,//是否显示关闭按钮，默认不显示将自动关闭窗口，可选择true，使用手动点击按钮关闭窗口
            iconShow: false,//是否显示状态图标，默认不显示，可选择true
            style: 'simple',//窗口风格，默认简单风格simple，可选择复杂风格complex
            zIndex: 0,//气泡层级，默认为0即不设置，如果与其他组件发生遮挡请自行设置
            icon: '',//图标代码，不需要另行设置
            onShown: '',//回调函数，打开小窗后执行，无参数
            onHidden: '',//回调函数，关闭小窗后执行，无参数
            onUpdated: '',//回调函数，更新参数后执行，无参数
        }, options);
        // 定义监听事件成员
        this.handlers = {};
        //延迟关闭方法
        this.timeout = null;
        this.setOutFun = () => {
            this.timeout = setTimeout(() => {
                this.hide();
            }, this.options.delay);
        }
        this.clearOutFun = () => {
            this.timeout && clearTimeout(this.timeout);
        }
        //初始化执行
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'message');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //根据风格定义图标
        if (this.options.style == 'simple' || !this.options.style) {
            //简单的图表
            switch (this.options.result) {
                case 'success':
                    this.options.icon = axIcon.font.success;
                    !this.options.content ? this.options.content = this.options.title.success : null;
                    break;
                case 'error':
                    this.options.icon = axIcon.font.error;
                    !this.options.content ? this.options.content = this.options.title.error : null;
                    break;
                case 'question':
                    this.options.icon = axIcon.font.question;
                    !this.options.content ? this.options.content = this.options.title.question : null;
                    break;
                case 'warning':
                    this.options.icon = axIcon.font.warning;
                    !this.options.content ? this.options.content = this.options.title.warning : null;
                    break;
                default:
                    this.options.icon = axIcon.font.info;
                    !this.options.content ? this.options.content = this.options.title.info : null;
            }
        } else if (this.options.style == 'complex') {
            //动画图标
            switch (this.options.result) {
                case 'success':
                    this.options.icon = axIcon.svg.success;
                    break;
                case 'error':
                    this.options.icon = axIcon.svg.error;
                    break;
                case 'question':
                    this.options.icon = axIcon.svg.question;
                    break;
                case 'warning':
                    this.options.icon = axIcon.svg.warning;
                    break;
                default:
                    this.options.icon = axIcon.svg.info;
            }
        }
        //定义模板
        if (this.options.style == 'simple' || !this.options.style) {
            //定义简单风格的内容模板
            this.template = axTplEngine(`
            <section data-style="simple">
                <div class="ax-message-content">
                    <# if(this.iconShow){ #><span legend><# this.icon #></span><# } #>
                    <div class="ax-message-body"><# this.content #></div>
                    <# if(this.close){ #><i close></i> <# } #>
                </div>
            </section>
            `, this.options);
        } else if (this.options.style == 'complex') {
            //定义复杂风格的内容模板
            this.template = axTplEngine(`
            <section data-style="complex">
                <div class="ax-message-content">
                    <div class="ax-message-header">
                    <# if(this.iconShow){ #><span class="ax-result ax-result-show" legend><# this.icon #></span><# } #>
                    <span class="ax-message-title"><# if(this.result == 'success'){ #> <# this.title.success #> <# }else if(this.result == 'error'){ #> <# this.title.error #> <# }else if(this.result == 'question'){#> <# this.title.question #> <# }else if(this.result == 'warning'){#> <# this.title.warning #> <# } else {#> <# this.title.info #> <# } #></span></div>
                    <div class="ax-message-body"><# this.content #></div>
                </div>
                <# if(this.close){ #><i close></i> <# } #>
                <# if(this.progress){ #><div class="ax-message-progress" style="transition-duration:<# this.delay #>ms;"></div><# } #>
            </section>
            `, this.options);
        }
        //创建Dom
        this.targetDom = axStrToDom(this.template);
        this.contentDom = this.targetDom.querySelector('.ax-message-body');
        this.titleDom = this.targetDom.querySelector('.ax-message-title');
        this.closeDom = this.targetDom.querySelector('[close]');
        this.progressDom = this.targetDom.querySelector('.ax-message-progress');
        this.legendDom = this.targetDom.querySelector('[legend]');
        //根据结果设定属性
        switch (this.options.result) {
            case 'success':
                this.targetDom.setAttribute('data-result', 'success');
                break;
            case 'error':
                this.targetDom.setAttribute('data-result', 'error');
                break;
            case 'question':
                this.targetDom.setAttribute('data-result', 'question');
                break;
            case 'warning':
                this.targetDom.setAttribute('data-result', 'warning');
                break;
            default:
                this.targetDom.setAttribute('data-result', 'info');
        }
        //如果设置了类名，则自动添加
        let className = '';
        !this.options.className ? className = ' ' + this.options.className : className = '';
        //如果有close按钮则插入data-close属性
        this.options.close ? this.targetDom.setAttribute('data-close', 'true') : null;
        //将Message内容转入Child的Dom中
        this.contentDom.innerHTML = this.options.content;
        //定义容器
        this.wrapperDom = ''
        //状态
        this.shown = false;
        //定义创建Dom事件
        this.event = (position) => {
            let wrapper = axMessages.find(k => k.position === position);
            if (!wrapper) {
                this.wrapperDom = axAddElem('div', { 'class': 'ax-message' + className, 'data-placement': position });
                axMessages.push({ position: position, parent: this.wrapperDom });
            } else {
                this.wrapperDom = wrapper.parent;
            }
        }



        //根据位置不同，创建不同的Dom
        switch (this.options.placement) {
            case 'left-top':
                this.event('left-top');
                break;
            case 'center-top':
                this.event('center-top');
                break;
            case 'right-top':
                this.event('right-top');
                break;
            case 'left-center':
                this.event('left-center');
                break;
            case 'center-center':
                this.event('center-center');
                break;
            case 'right-center':
                this.event('right-center');
                break;
            case 'left-bottom':
                this.event('left-bottom');
                break;
            case 'center-bottom':
                this.event('center-bottom');
                break;
            case 'right-bottom':
                this.event('right-bottom');
                break;
            case 'left':
                this.event('left-top');
                break;
            case 'top':
                this.event('center-top');
                break;
            case 'right':
                this.event('right-top');
                break;
            case 'center':
                this.event('center-center');
                break;
            default:
                this.event('center-top');
        }

        //点击关闭按钮关闭提示小窗
        if (this.closeDom) {
            this.closeDom.onclick = () => {
                this.hide();
            };
        }

        //鼠标移入取消timeout
        new axHover(this.targetDom, {
            enter: () => {
                this.clearOutFun();
            },
            leave: () => {
                this.setOutFun();
            }
        });

        //设置zIndex
        this.options.zIndex ? this.wrapperDom.style.zIndex = this.options.zIndex : null;

        //在页面创建信息窗容器节点，虚拟节点才创建
        axIsVnode(this.wrapperDom) ? document.body.appendChild(this.wrapperDom) : null;

        //返回实例，以便于链式操作
        return this;
    }
    show(callback) {
        //已经是打开状态则不再执行
        if (this.shown) {
            return false;
        }
        //先插入节点再展开
        this.wrapperDom.appendChild(this.targetDom);
        axSlideDown(this.targetDom, '', () => {
            this.shown = true;
            //监听回调
            this.handlers.hasOwnProperty('shown') ? this.emit('shown', '') : null;
            this.options.onShown && this.options.onShown.call(this);
            callback && callback.call(this);
            //显示关闭进度
            if (!this.options.close) {
                this.options.progress && this.progressDom ? this.progressDom.style.cssText += `width:0;` : null;
                this.setOutFun();
            }
        });
        return this;
    }
    hide(callback) {
        //已经是关闭状态则不再执行
        if (!this.shown) {
            return false;
        }
        axSlideUp(this.targetDom, '', () => {
            this.targetDom.remove();
            this.shown = false;
            //监听回调
            this.handlers.hasOwnProperty('hidden') ? this.emit('hidden', '') : null;
            this.options.onHidden && this.options.onHidden.call(this);
            callback && callback.call(this);
        });
        return this;
    }
    update(setting, callback) {
        //先移除节点
        this.targetDom.remove();
        //更新参数
        this.options = axExtend(this.options, setting);
        //根据参数初始化实例
        this.init();
        //监听回调
        this.handlers.hasOwnProperty('updated') ? this.emit('updated', '') : null;
        this.options.onUpdated && this.options.onUpdated.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}


/*!
 * 插件：表单校验；使用方法：new axValid('#id',{参数})
 */
const axValids = [];
class axValid {
    constructor(targetDom, options) {
        this.targetDom = axIdToDom(targetDom);
        if (!this.targetDom.getAttribute('name')) {
            console.error('The element to be verified must have a name attribute!');
            return false;
        }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            label: '',//表单元素的替代名称
            mode: 'newline', //设置文字展示方式，默认放在input下方，另可选择note|popup|message|#ID|dom元素
            url: '', //异步校验地址
            iconShow: true, //是否显示提示图标，默认显示，可填写false不显示
            type: '', //校验类型，用英文逗号隔开，如果使用extend追加了类型则可写入
            parent: '',//定义父form
            error: '', //自定义错误信息格式
            success: '通过验证！', //自定义成功信息格式
            rule: {}, //根据类型自定义提示文字，格式举例：{email:['错误！','正确！'],telephone:['错误！','正确！']}
            number: '', //将要比较的数字，比较大于小于或等于
            max: '', //要求大于的数值
            min: '', //要求小于的数值
            length: '', //将要比较的字符数量
            maxLength: '', //可填写最多字符数量
            minLength: '', //至少要填写的字符数量
            include: [], //包含的内容，用逗号隔开的数组
            exclude: [], //排除的内容，用逗号隔开的数组
            date: '', //比较日期
            earliest: '', //最早的日期
            latest: '', //最晚的日期
            compare: '', //对比对象，可以填input的name值或dom对象
            compareLabel: '', //对比对象的替换名称
            least: 1, //至少选择多少个，如果是checkbox、select的checked校验则需要
            most: 2, //最多选择多少个，如果是checkbox、select的checked校验则需要
            separator: '',//内容分隔符，在判断文本框length值时用到
            successShow: true,//校验通过，是否需要显示提示，默认true显示，可选择false不显示
            message: {},//自定义axMessage插件的参数
            popup: {},//自定义axPopup插件的参数
            format: '', //强度匹配格式
            inputSuccess: 'ax-check-primary', //input通过验证的类名
            inputError: 'ax-check-error', //input未通过验证的类名
            boxSuccess: 'ax-color-primary', //信息容器通过验证的类名
            boxError: 'ax-color-error', //信息容器未通过验证的类名
            chars: '~!@#', //定义特殊字符，可自行选择特殊字符，注意某些字符需要转意处理。比较完整的特殊字符有：~!@#$%^&*()_+{}|:"<>?\\-=\\[\\]\;\',./￥——：”《》？【】；’，。、
            charsShow: false, //在信息中是否显示特殊字符，由于不确定会有多少个，所以默认不显示，用户可选择true显示
            trigger: 'blur', //如何触发校验，默认是blur失去焦点时，可选change失去焦点并改变值，input键入是实时触发，listen监听隐藏域和submit提交时触发
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            immediately: true,//页面加载后是否立即检测并进行init监听，默认true，可选择false
            language: {
                required: '这是必填项！',
                requiredWithLabel: '<# this.label #>是必填项！',
                email: '<# this.label #>请填写正确的邮箱！',
                telephone: '<# this.label #>请填写11位手机号！',
                landline: '<# this.label #>请填写正确的座机号码！',
                ip: '<# this.label #>请填写正确的IP地址！',
                idCard: '<# this.label #>请填写正确的身份证号！',
                zip: '<# this.label #>只能填写6位数字邮编！',
                url: '<# this.label #>请填写正确的域名！',
                plate: '<# this.label #>请填写正确的车牌号！',
                chinese: '<# this.label #>只能填写中文！',
                letter: '<# this.label #>只能填写大小写英文字母！',
                string: '<# this.label #>只能填写大小写英文字母和数字！',
                password: '<# this.label #>只能填写大小写英文字母、数字以及特殊字符！',
                ymdhms: '<# this.label #>只能填写类似2022-11-13 2:56:12的日期格式！',
                ymd: '<# this.label #>只能填写类似2022-11-13的日期格式！',
                hms: '<# this.label #>只能填写类似2:56:12的日期格式！',
                ym: '<# this.label #>只能填写类似2022-11的日期格式！',
                y: '<# this.label #>只能填写4位数字年份！',
                m: '<# this.label #>只能填写1~12月份！',
                d: '<# this.label #>只能填写1~31日！',
                date: '<# this.label #>请填写有效的日期！',
                earlier: '<# this.label #>不可晚于<# this.value #>',
                later: '<# this.label #>不可早于<# this.value #>',
                dateRange: '<# this.label #>不可早于<# this.max #>，且不可晚于<# this.min #>',
                integer: '<# this.label #>只能填写非0开头的正整数！',
                number: '<# this.label #>只能填写数字，包括正数、负数、整数、小数！',
                moreThan: '<# this.label #>需要大于<# this.value #>',
                moreEqual: '<# this.label #>需要大于或等于<# this.value #>',
                lessThan: '<# this.label #>需要小于<# this.value #>',
                lessEqual: '<# this.label #>需要小于或等于<# this.value #>',
                equal: '<# this.label #>需要等于<# this.value #>',
                range: '<# this.label #>需要大于等于<# this.min #>并小于<# this.max #>',
                length: '<# this.label #>只能填写<# this.value #>个字符！',
                lengthMore: '<# this.label #>字符数量不可少于<# this.value #>个！',
                lengthLess: '<# this.label #>字符数量不可多于<# this.value #>个！',
                lengthRange: '<# this.label #>字符数量不可少于<# this.min #>个，且不可多余<# this.max #>个！',
                include: '<# this.label #>填写的值在指定内容中不存在！',
                exclude: '<# this.label #>不可填写该值！',
                oneChecked: '<# this.label #>必须且只能选择1项！',
                leastChecked: '<# this.label #>至少选择<# this.value #>项！',
                mostChecked: '<# this.label #>最多选择<# this.value #>项！',
                rangeChecked: '<# this.label #>至少选择<# this.min #>项，且不能多于<# this.max #>项！',
                same: '两个字段值不一致！',
                sameWithLabel: '<# this.label #>与<# this.value #>不一致！',
                strong:{
                    az:'小写字母',
                    AZ:'大写字母',
                    number:'数字',
                    char:'特殊字符',
                    chinese:'中文',
                    and:'和',
                    require:'必须是',
                    requireMore:'并且要求',
                    least:'至少',
                    unit:'个字符，',
                    select:'中至少选择',
                    group:'种字符。',
                    total:'总共只能是<# this.value #>个字符。',
                    totalMore:'总共至少<# this.value #>个字符。',
                    totalRange:'总共至少<# this.min #>个字符，且不能超过<# this.max #>个字符。',
                    },
                message:{
                    strongError:'随机字符类型必须有1种，并且不能超过允许字符类型范围！',
                }
            },
            onTrigger: '', //事件触发化后的回调函数，支持三个参数：验证状态(通过与错误)verified、验证文本信息text、验证强度strong
            onChange: '',//状态变化后的回调函数，支持三个参数：验证状态(通过与错误)verified、验证文本信息text、验证强度strong
            onInit: '',//校验完成后的回调函数，支持三个参数：验证状态(通过与错误)verified、验证文本信息text、验证强度strong
        }, options, this.targetDom, this.constructor.name);
        //定义parent，默认是最近的父form
        if (this.options.parent) {
            this.parent = axIdToDom(this.options.parent);
        } else {
            this.parent = this.targetDom.closest('form') ? this.targetDom.closest('form') : document;
        }
        //定义compare对象，如果是字符串则认为是字段的name值
        if (this.options.compare) {
            if (typeof this.options.compare == 'string') {
                this.compare = this.parent.querySelector('[name=' + this.options.compare + ']');
            } else {
                this.compare = axIdToDom(this.options.compare);
            }
        }
        //定义compare对象的替代名称
        if (this.options.compareLabel) {
            this.compareLabel = this.options.compareLabel;
        } else {
            if (this.compare && this.compare.getAttribute('data-label')) {
                this.compareLabel = this.compare.getAttribute('data-label');
            }
        }
        //监听事件合集
        this.handlers = {};
        //定义默认强度
        this.strong = 0;
        //校验结果需要被div包裹，mode=message除外
        this.msgBox = axAddElem('div', { 'valid': '' });
        //字段值
        this.value = null;
        //字段替代名称
        this.label = this.options.label;
        //字段名称
        this.name = this.targetDom.getAttribute('name');
        //字段类型,select没有type属性，需要通过.type获取（select-one或select-multiple）
        this.type = this.targetDom.getAttribute('type') ? this.targetDom.getAttribute('type') : this.targetDom.type;
        //根据条件获得判断文本
        this.text = '';
        //校验状态
        this.verified = null;
        //记录最后两次触发事件产生的verified值
        this.verifieds = [];
        //根据已知条件创建初始化规则
        this.getRules();
        //获得强度解析（根据this.options.format）
        this.strongParse = this.parseFormat();
        //内置校验类型
        this.checks = {
            /**
             * 为空校验，不填则直接返回结果
             */
            required: function (value) {
                if (value) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: this.label ? axTplEngine(this.options.language.requiredWithLabel, { label: this.label }) : this.options.language.required };
                }
            },
            /**
             * 特定字符串校验
             */
            //校验邮箱，有“@”和“.”，并且“.”在“@”之后
            email: function (value) {
                let pattern = /^([a-zA-Z0-9_-])+@([a-zA-Z0-9_-])+(\.[a-zA-Z0-9_-])+/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.email, { label: this.label }));
            },
            //校验手机号，必须是11位数字，以“1”开头，第二位必须是“3456789”中其中一个，第三位必须是“0~9”其中一个，剩下9位随意
            telephone: function (value) {
                let pattern = /^[1][3456789][0-9]{9}$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.telephone, { label: this.label }));
            },
            //固话座机号码校验，0755-12345678,010-12345678，12345678，1234567可通过校验
            landline: function (value) {
                let pattern = /^([0-9]{3,4}-)?[0-9]{7,8}$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.landline, { label: this.label }));
            },
            //IP地址校验
            ip: function (value) {
                let pattern = /^\d+\.\d+\.\d+\.\d+$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.ip, { label: this.label }));
            },
            //中国车牌号校验，支持8位新能源汽车
            plate: function (value) {
                let pattern = '';
                if (value.length == 8) {
                    //新能源汽车规则，8位
                    pattern = /^[京津沪渝冀豫云辽黑湘皖鲁新苏浙赣鄂桂甘晋蒙陕吉闽贵粤青藏川宁琼使领A-Z]{1}[A-Z]{1}(([0-9]{5}[DF]$)|([DF][A-HJ-NP-Z0-9][0-9]{4}$))/;
                } else {
                    //传统汽车规则
                    pattern = /^[京津沪渝冀豫云辽黑湘皖鲁新苏浙赣鄂桂甘晋蒙陕吉闽贵粤青藏川宁琼使领A-Z]{1}[A-Z]{1}[A-HJ-NP-Z0-9]{4}[A-HJ-NP-Z0-9挂学警港澳]{1}$/;
                }
                return this.doTest(pattern, value, axTplEngine(this.options.language.plate, { label: this.label }));
            },
            //中国18位身份证号校验
            idCard: function (value) {
                let pattern = /^(([1][1-5])|([2][1-3])|([3][1-7])|([4][1-6])|([5][0-4])|([6][1-5])|([7][1])|([8][1-2]))\d{4}(([1][9]\d{2})|([2]\d{3}))(([0][1-9])|([1][0-2]))(([0][1-9])|([1-2][0-9])|([3][0-1]))\d{3}[0-9xX]$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.idCard, { label: this.label }));
            },
            //6位邮编校验
            zip: function (value) {
                let pattern = /^\d{6}$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.zip, { label: this.label }));
            },
            //域名校验，http://www.axui.cn,https://www.axui.cn,https://www.axui.cn:8080均可通过校验
            url: function (value) {
                let pattern = /^(?=^.{3,255}$)(http(s)?:\/\/)?(www\.)?[a-zA-Z0-9][-a-zA-Z0-9]{0,62}(\.[a-zA-Z0-9][-a-zA-Z0-9]{0,62})+(:\d+)*(\/\w+\.\w+)*$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.url, { label: this.label }));
            },
            /**
             * 字符类型校验
             */
            //纯中文校验
            chinese: function (value) {
                let pattern = /^[\u0391-\uFFE5]+$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.chinese, { label: this.label }));
            },
            //纯英文字母校验
            letter: function (value) {
                let pattern = /^[a-zA-Z]+$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.letter, { label: this.label }));
            },
            //英文字母+数字校验
            string: function (value) {
                let pattern = /^[a-zA-Z0-9]+$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.string, { label: this.label }));
            },
            //英文字母+数字+特殊字符校验（密码校验）
            password: function (value) {
                let pattern = new RegExp('^[a-zA-Z0-9' + this.options.chars + ']+$');
                return this.doTest(pattern, value, axTplEngine(this.options.language.password, { label: this.label }));
            },
            /** 
             *日期格式校验
             */
            //YYYY-MM-DD hh:mm:ss格式校验
            ymdhms: function (value) {
                let pattern = /^(\d{4})(-|\/)(\d{1,2})\2(\d{1,2}) (\d{1,2}):(\d{1,2}):(\d{1,2})$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.ymdhms, { label: this.label }));
            },
            //YYYY-MM-DD格式校验
            ymd: function (value) {
                let pattern = /^(\d{4})(-|\/)(\d{1,2})\2(\d{1,2})$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.ymd, { label: this.label }));
            },
            //hh:mm:ss格式校验
            hms: function (value) {
                let pattern = /^((20|21|22|23|[0-1]\d)\:[0-5][0-9])(\:[0-5][0-9])?$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.hms, { label: this.label }));
            },
            //YYYY-MM格式校验
            ym: function (value) {
                let pattern = /^(\d{4})(-|\/)(\d{1,2})$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.ym, { label: this.label }));
            },
            //YYYY格式校验
            y: function (value) {
                let pattern = /^(\d{4})$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.y, { label: this.label }));
            },
            //MM格式校验，01~12或1~12能通过校验
            m: function (value) {
                let pattern = /^(0?[1-9]|1[0-2])$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.m, { label: this.label }));
            },
            //DD格式校验，01~31能通过校验
            d: function (value) {
                let pattern = /^((0?[1-9])|((1|2)[0-9])|30|31)$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.d, { label: this.label }));
            },
            //是一个日期校验，年、年+月、年+月+日、年+月+日+时间均可通过校验
            date: function (value) {
                let ymdhms = /^(\d{4})(-|\/)(\d{1,2})\2(\d{1,2}) (\d{1,2}):(\d{1,2}):(\d{1,2})$/,
                    ymd = /^(\d{1,4})(-|\/)(\d{1,2})\2(\d{1,2})$/,
                    ym = /^(\d{1,4})(-|\/)(\d{1,2})$/,
                    y = /^(\d{4})$/;
                if (ymdhms.test(value) || ymd.test(value) || ym.test(value) || y.test(value)) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.date, { label: this.label }) };
                }
            },
            //早于日期校验
            earlier: function (value) {
                if (!this.options.date || isNaN(Date.parse(this.options.date))) {
                    console.error('Please enter a valid target date!');
                    return;
                }
                if (isNaN(Date.parse(value))) {
                    console.error('The value is not a valid date!');
                    return;
                }
                let timestamp = axLocalTime(value),
                    target = axLocalTime(this.options.date);
                if (timestamp < target) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.earlier, { label: this.label, value: this.options.date }) };
                }
            },
            //晚于日期校验
            later: function (value) {
                if (!this.options.date || isNaN(Date.parse(this.options.date))) {
                    console.error('Please enter a valid target date!');
                    return;
                }
                if (isNaN(Date.parse(value))) {
                    console.error('The value is not a valid date!');
                    return;
                }
                let timestamp = axLocalTime(value),
                    target = axLocalTime(this.options.date);
                if (timestamp > target) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.later, { label: this.label, value: this.options.date }) };
                }
            },
            //日期区间校验
            dateRange: function (value) {
                if (!this.options.earliest || isNaN(Date.parse(this.options.earliest))) {
                    console.error('Please enter a valid earliest date!');
                    return;
                }
                if (!this.options.latest || isNaN(Date.parse(this.options.latest))) {
                    console.error('Please enter a valid latest date!');
                    return;
                }
                if (isNaN(Date.parse(value))) {
                    console.error('The value is not a valid date!');
                    return;
                }
                let timestamp = axLocalTime(value),
                    earliest = axLocalTime(this.options.earliest),
                    latest = axLocalTime(this.options.latest);
                if (timestamp > earliest && timestamp < latest) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.dateRange, { label: this.label, min: this.options.earliest, max: this.options.latest }) };
                }
            },
            /** 
             *数字校验
             */
            //非0开头正整数校验
            integer: function (value) {
                let pattern = /^[1-9]d*$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.integer, { label: this.label }));
            },
            //数字校验，正数、负数、整数、浮点数值均能通过校验
            number: function (value) {
                let pattern = /^(-?\d+)\.?(\d*)$/;
                return this.doTest(pattern, value, axTplEngine(this.options.language.number, { label: this.label }));
            },
            //大于校验，配合this.options.number参数使用
            moreThan: function (value) {
                let pattern = /^(-?\d+)\.?(\d*)$/;
                if (pattern.test(value) && value > this.options.number) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.moreThan, { label: this.label, value: this.options.number }) };
                }
            },
            //大于等于校验，配合this.options.number参数使用
            moreEqual: function (value) {
                let pattern = /^(-?\d+)\.?(\d*)$/;
                if (pattern.test(value) && value >= this.options.number) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.moreEqual, { label: this.label, value: this.options.number }) };
                }
            },
            //小于校验，配合this.options.number参数使用
            lessThan: function (value) {
                let pattern = /^(-?\d+)\.?(\d*)$/;
                if (pattern.test(value) && value < this.options.number) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.lessThan, { label: this.label, value: this.options.number }) };
                }
            },
            //大于等于校验，配合this.options.number参数使用
            lessEqual: function (value) {
                let pattern = /^(-?\d+)\.?(\d*)$/;
                if (pattern.test(value) && value <= this.options.number) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.lessEqual, { label: this.label, value: this.options.number }) };
                }
            },
            //等于校验，配合this.options.number参数使用
            equal: function (value) {
                let pattern = /^(-?\d+)\.?(\d*)$/;
                if (pattern.test(value) && value == this.options.number) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.equal, { label: this.label, value: this.options.number }) };
                }
            },
            //范围校验，配合this.options.min和max参数使用
            range: function (value) {
                let pattern = /^(-?\d+)\.?(\d*)$/;
                if (pattern.test(value) && value >= this.options.min && value < this.options.max) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.range, { label: this.label, min: this.options.min, max: this.options.max }) };
                }
            },
            /*
             *字符长度校验
             */
            //字符长度等于校验，配合this.options.length参数使用
            length: function (value) {
                let pattern = new RegExp('^.{' + this.options.length + '}$');
                return this.doTest(pattern, value, axTplEngine(this.options.language.length, { label: this.label, value: this.options.length }));
            },
            //字符长度不少于校验，配合this.options.length参数使用
            lengthMore: function (value) {
                let pattern = new RegExp('^.{' + this.options.length + ',}$');
                return this.doTest(pattern, value, axTplEngine(this.options.language.lengthMore, { label: this.label, value: this.options.length }));
            },
            //字符长度不多于校验，配合this.options.length参数使用
            lengthLess: function (value) {
                let pattern = new RegExp('^.{0,' + this.options.length + '}$');
                return this.doTest(pattern, value, axTplEngine(this.options.language.lengthLess, { label: this.label, value: this.options.length }));
            },
            //字符长度范围校验，配合this.options.minLength和this.options.maxLength参数使用
            lengthRange: function (value) {
                let pattern = new RegExp('^.{' + this.options.minLength + ',' + this.options.maxLength + '}$');
                return this.doTest(pattern, value, axTplEngine(this.options.language.lengthRange, { label: this.label, min: this.options.minLength, max: this.options.maxLength }));
            },
            /**
             * 包含与否检验
             */
            //是否在指定数组中，如果不在则报错
            include: function (value) {
                if (axIsEmpty(this.options.include)) {
                    console.error('The array of include can not be empty!');
                    return;
                }
                if (this.options.include.includes(value)) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.include, { label: this.label }) };
                }
            },
            //是否不在指定数组中，如果在则报错
            exclude: function (value) {
                if (axIsEmpty(this.options.exclude)) {
                    console.error('The array of exclude can not be empty!');
                    return;
                }
                if (!this.options.exclude.includes(value)) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.exclude, { label: this.label }) };
                }
            },
            /**
             * 强度校验
             */
            strong: function (value) {
                if (!this.options.format) {
                    console.error(axError.formatError);
                    return;
                }
                //更新this.strong
                this.getStrong(value);
                let pattern = new RegExp(this.strongParse.pattern);
                return this.doTest(pattern, value, this.strongParse.message);
            },
            /**
             * 异步校验，需要配合this.options.url参数
             */
            async: function (value) {
                if (!this.options.format) {
                    console.error(axError.formatError);
                    return;
                }
                //更新this.strong
                this.getStrong(value);
                let pattern = new RegExp(this.strongParse.pattern);
                return this.doTest(pattern, value, this.strongParse.message);
            },
            /**
             * 选择数量校验，radio、checkbox、select使用
             */
            //必须选择一个校验
            oneChecked: function (value) {
                if (!Array.isArray(value)) {
                    console.error(axError.arrayRequired);
                    return;
                }
                if (value.length == 1) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.oneChecked, { label: this.label }) };
                }
            },
            //至少选择多少项校验，需要this.options.least参数配合使用
            leastChecked: function (value) {
                if (!Array.isArray(value)) {
                    console.error(axError.arrayRequired);
                    return;
                }
                //least有值且value个数大于least值才为真
                if (this.options.least && value.length >= this.options.least) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.leastChecked, { label: this.label, value: this.options.least }) };
                }
            },
            //最多选择多少项校验，需要this.options.most参数配合使用
            mostChecked: function (value) {
                if (!Array.isArray(value)) {
                    console.error(axError.arrayRequired);
                    return;
                }
                //most为空，或者value个数小于most值才为真
                if (!this.options.most || (this.options.most && value.length <= this.options.most)) {
                    return { verified: true, message: '' };
                } else {
                    return { verified: false, message: axTplEngine(this.options.language.mostChecked, { label: this.label, value: this.options.most }) };
                }
            },
            //选择数量范围校验，需要this.options.least和this.options.most参数配合使用
            rangeChecked: function (value) {
                if (!Array.isArray(value)) {
                    console.error(axError.arrayRequired);
                    return;
                }
                if (this.options.least && this.options.most && this.options.most > this.options.least) {
                    if (value.length >= this.options.least && value.length <= this.options.most) {
                        return { verified: true, message: '' };
                    } else {
                        return { verified: false, message: axTplEngine(this.options.language.rangeChecked, { label: this.label, min: this.options.least, max: this.options.most }) };
                    }
                } else {
                    console.error(axError.paraError);
                    return;
                }
            },
            /**
             * 对比校验
            */
            //重复值一致校验
            same: function (value) {
                if (!this.compare) {
                    console.error(axError.noNode);
                    return;
                }
                if (value == this.compare.value) {
                    return { verified: true, message: '' };
                } else {
                    let info = this.compareLabel ? axTplEngine(this.options.language.samedWithLabel, { label: this.label, value: this.compareLabel }) : this.options.language.samed;
                    return { verified: false, message: info };
                }
            },


        }
        //扩展校验类型
        Object.assign(this.checks, this.options.extend);

        //校验执行
        this.init();
        return this;
    }
    init(callback) {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'valid');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //如果是异步校验，那么需要有校验后的回调，所以init增加了回调函数；当然该回调函数在其他校验类型下也可以使用

        //将校验结果存起来，用于提交表单的时候再次校验
        let info = { dom: this.targetDom, name: this.name, instance: this };
        if (axIsEmpty(axValids)) {
            //如果axValids还是空的，则直接追加
            if (this.parent) {
                axValids.push({ parent: this.parent, children: [info] });
            }
        } else {
            if (axValids.some(item => (item.parent == this.parent))) {
                //如果axValid中已经有了parent，那么直接将info追加到该parent的children中
                let index = axValids.findIndex(i => (i.parent == this.parent)),
                    infoArr = axValids[index].children;
                //在判断children数组中是否有重复的name，radio和checkbox会有多各name，只取一个
                if (infoArr.some(i => (i.name == this.name))) {
                    //如果数组中有重复的则替换
                    for (let k = 0; k < infoArr.length; k++) {
                        if (infoArr[k].name == this.name) {
                            //如果已经存在了，则直接覆盖，并中断循环
                            infoArr[k] = info;
                            break;
                        }
                    }
                } else {
                    //如果没有重复的则追加
                    infoArr.push(info);
                }
            } else {
                axValids.push({ parent: this.parent, children: [info] });
            }
        }
        /* 
        axValids的最终格式如下：
        [
            {
                parent:'', children: [
                    { dom: '', name: '', instance: '', verified: '' },
                    { dom: '', name: '', instance: '', verified: '' },
                ]
            },
            {
                parent:'', children: [
                    { dom: '', name: '', instance: '', verified: '' },
                    { dom: '', name: '', instance: '', verified: '' },
                ]
            },
            ...
        ] 
        以parent循环监听submit事件，以dom和instance执行校验，以name去重，以verfied确认表单验证是否全部通过
        */

        /**
         * 触发校验
         * */
        //定义通用校验函数
        let checkFun = async () => {
            //如果有扩展校验类型，则给rules补上相应的规则
            if (!axIsEmpty(this.options.extend)) {
                for (let i in this.options.extend) {
                    this.rules[i] = ['', this.options.success]
                }
            }

            let value = '';
            if (this.type == 'radio' || this.type == 'checkbox' || this.type == 'select-multiple' || this.type == 'file') {
                //以上属性的字段值均转成数组，单选、复选、下拉多选以及多文件选择
                //value = axValuesToArr(this.name, this.type);
                value = axMultiValues(this.name, this.type, this.parent);
            } else {
                //一般文本字段或者没有multiple的字段
                value = this.targetDom.value.trim();
                if (this.options.separator) {
                    //如果填写了分隔符，则将文本值转为数组
                    value = value.split(this.options.separator);
                }
            }

            //len和index是用来判断是否循环到最后一个，最后一个rule验证通过才算真正通过
            let len = Object.keys(this.rules).length,
                index = 0;

            //更新当前值，回调函数中可用
            this.value = value;

            for (let i in this.rules) {
                index++;
                let msgFalse = this.rules[i][0],
                    msgTrue = this.rules[i][1];
                //执行校验
                if (axIsEmpty(value)) {
                    //没有值而且有必填校验
                    if (i == 'required') {
                        this.text = !msgFalse ? this.checks[i].call(this, '').message : msgFalse;
                        this.verified = false;
                        //required类型下，值为空则跳出校验
                        break;
                    } else if (i == 'oneChecked' || i == 'leastChecked' || i == 'rangeChecked') {
                        //radio、checkbox、selet、file会返回空数组：[]或['']
                        this.text = !msgFalse ? this.checks[i].call(this, []).message : msgFalse;
                        this.verified = false;
                    } else {
                        //如果没有值而且没有必填校验则直接通过验证
                        this.text = msgTrue;
                        this.verified = true;
                    }
                } else {
                    //有值校验，由于异步校验不需要type，所以此处只针对有type的情况下
                    if (this.options.type) {
                        let result = this.checks[i].call(this, value);

                        if (!result.verified) {
                            //如果校验不通过，且没有填error参数，则使用默认错误提示
                            !msgFalse ? this.text = result.message : this.text = msgFalse;
                            this.verified = false;
                            break;
                        } else {
                            //全部校验成功，在最后一次校验成功的时候输出
                            if (index == len) {
                                this.text = msgTrue;
                                this.verified = true;
                            }

                        }
                    }

                }

            }

            //根据this.options.mode，this.verified和this.text生成dom
            if (this.options.url && !this.options.type) {
                //填写了url即使用异步校验，使用异步校验就不能在使用同步校验，也就是type为空
                await axAjax({
                    url: this.options.url,
                    data: { name: this.name, value: this.value, parent: this.parent },
                    type: this.options.ajaxType,
                    success: (response) => {
                        this.verified = response.content.verified;
                        this.text = response.content.text;
                    }
                });
            }
            this.refreshDom(this.options.mode);
            //触发后获得最新校验结果写入axValids中，info已经在前面定义并事先写入了axValids
            info.verified = this.verified;
            let resultArr = [this.verified, this.text, this.strong];
            //监听回调
            callback && callback.call(this, ...resultArr);
            //事件触发回调函数
            this.options.onTrigger && this.options.onTrigger.call(this, ...resultArr);
            this.handlers.hasOwnProperty('trigger') ? this.emit('trigger', ...resultArr) : null;
            //监听状态更新事件
            this.verifieds.unshift(this.verified);
            if (this.verifieds.length >= 2) {
                if (this.verifieds.length > 2) {
                    this.verifieds.splice(2);
                }
                if (this.verifieds[0] != this.verifieds[1]) {
                    this.options.onChange && this.options.onChange.call(this, ...resultArr);
                    'change' in this.handlers ? this.emit('change', ...resultArr) : null;
                }
            }
            return resultArr;
        }


        if (this.options.trigger == 'submit') {
            //form提交的时候，改变表单元素的触发方式
            checkFun();
        } else if (this.options.trigger == 'listen') {
            //type=hidden的表单域值变化时触发
            this.targetDom.onchange = () => {
                checkFun();
            }
        } else {
            //注意：radio、checkbox适合change和input；select只适合使用change
            this.targetDom.addEventListener(this.options.trigger, checkFun, false);
        }
        //初始化检测
        if (this.options.immediately) {
            checkFun().then((response) => {
                this.options.onInit && this.options.onInit.call(this, ...response);
                this.handlers.hasOwnProperty('init') ? this.emit('init', ...response) : null;
            });
        }
        return this;
    }
    parseFormat() {
        //解析强度格式
        if (this.options.format) {
            //定义变量
            let upperLen = '', //大写字母个数
                lowerLen = '', //小写字母个数
                numberLen = '', //数字个数
                charLen = '', //特殊字符个数
                cnLen = '', //中文字符个数
                formatMin = '', //字符数量不少于
                formatMax = '', //字符数量不多于
                onlyName = {},
                scaleNameGroup = [],
                scaleName = {},
                scale = '',
                regExp = '',
                regText = [],
                message = '',
                lengthOnly = null,
                //获得校验字符
                formatType = '',
                formatLength = '',
                formatScale = '';

            //不管有没有#，[0]始终是formatType
            formatType = this.options.format.split("#")[0];

            //定义字符长度要求
            if (this.options.format.split("#")[1]) {

                formatLength = this.options.format.split("#")[1];
                //不管有没有','，[0]始终是formatMin
                formatMin = formatLength.split(",")[0];
                if (formatLength.indexOf(',') == -1) {
                    //{6}
                    lengthOnly = true;
                    formatMax = '';
                } else if (formatLength.indexOf(',') !== -1 && formatLength.split(",")[1]) {
                    //{6,10}
                    formatMax = formatLength.split(",")[1];
                } else {
                    //{6,}
                    formatMax = '';
                }
            } else {
                //'aA9#'或者'aA9##aA9~中'
                formatMin = '';
                formatMax = '';
            }
            //定义范围要求
            if (this.options.format.split("#")[2]) {
                formatScale = this.options.format.split("#")[2];
                if (formatScale.indexOf('a') !== -1) {
                    scaleName.lower = this.options.language.strong.az;
                    scale += 'a-z';
                    regText.push('a-z');
                    scaleNameGroup.push(scaleName.lower);
                }
                if (formatScale.indexOf('A') !== -1) {
                    scaleName.upper = this.options.language.strong.AZ;
                    scale += 'A-Z';
                    regText.push('A-Z');
                    scaleNameGroup.push(scaleName.upper);
                }
                if (formatScale.indexOf('9') !== -1) {
                    scaleName.number = this.options.language.strong.number;
                    scale += '0-9';
                    regText.push('0-9');
                    scaleNameGroup.push(scaleName.number);
                }
                if (formatScale.indexOf('~') !== -1) {
                    scaleName.char = this.options.charsShow ? `${this.options.language.strong.char}"${this.options.chars}"` : this.options.language.strong.char;
                    scale += this.options.chars;
                    regText.push(this.options.chars);
                    scaleNameGroup.push(scaleName.char);
                }
                if (formatScale.indexOf('中') !== -1) {
                    scaleName.cn = this.options.language.strong.chinese;
                    scale += '\u4e00-\u9fa5';
                    regText.push('\u4e00-\u9fa5');
                    scaleNameGroup.push(scaleName.cn);
                }
                if (scaleNameGroup.length > 2) {
                    scaleNameGroup = axRangeReplace(scaleNameGroup.join(), this.options.language.strong.and, scaleNameGroup.join().lastIndexOf(','));
                } else {
                    scaleNameGroup = scaleNameGroup.join()
                }
                scaleNameGroup = scaleNameGroup.replaceAll(',', '、');
                !axIsEmpty(scaleNameGroup) ? message += `${this.options.language.strong.require}${scaleNameGroup}，${this.options.language.strong.requireMore}` : null;
            }
            //随机几种字符，只有当没有类型限制时有效
            if (axIsEmpty(this.options.format.split("#")[0]) && this.options.format.split("#")[3]) {
                if (this.options.format.split("#")[3] < formatScale.length && this.options.format.split("#")[3] > 0) {
                    //必须包含两种字符
                    if (this.options.format.split("#")[3] == 2) {
                        regText.forEach(item => {
                            regExp += '(?![' + item + ']+$)';
                        });
                    } else if (this.options.format.split("#")[3] == 3) {
                        axArrange(regText, 2).forEach(item => {
                            regExp += '(?![' + item.join().replaceAll(',', '') + ']+$)';
                        });
                    } else if (this.options.format.split("#")[3] == 4) {
                        axArrange(regExp, 3).forEach(item => {
                            regExp += '(?![' + item.join().replaceAll(',', '') + ']+$)';
                        });
                    }
                } else {
                    new axMessage({
                        result: 'error',
                        content: this.options.language.strongError,
                    }).show();
                }
                //message = scaleNameGroup + '中至少选择' + this.options.format.split("#")[3] + '种字符。';
                message = `${scaleNameGroup}${this.options.language.strong.select}${this.options.format.split("#")[3]}${this.options.language.strong.group}`;
            }
            //获取字符频次
            if (formatType.indexOf('a') !== -1) {
                let index = formatType.lastIndexOf("a");
                onlyName.lower = this.options.language.strong.az;
                lowerLen = parseInt(formatType.substring(index + 1, index + 2));
                // message += onlyName.lower + '至少' + lowerLen + '个，';
                message += `${onlyName.lower}${this.options.language.strong.least}${lowerLen}${this.options.language.strong.unit}`;
            }
            if (formatType.indexOf('A') !== -1) {
                let index = formatType.lastIndexOf("A");
                onlyName.upper = this.options.language.strong.AZ;
                upperLen = parseInt(formatType.substring(index + 1, index + 2));
                //message += onlyName.upper + '至少' + upperLen + '个，';
                message += `${onlyName.upper}${this.options.language.strong.least}${upperLen}${this.options.language.strong.unit}`;
            }
            if (formatType.indexOf('9') !== -1) {
                let index = formatType.lastIndexOf("9");
                onlyName.number = this.options.language.strong.number;
                numberLen = parseInt(formatType.substring(index + 1, index + 2));
                //message += onlyName.number + '至少' + numberLen + '个，';
                message += `${onlyName.number}${this.options.language.strong.least}${numberLen}${this.options.language.strong.unit}`;
            }
            if (formatType.indexOf('~') !== -1) {
                let index = formatType.lastIndexOf("~");
                onlyName.char = this.options.charsShow ? `${this.options.language.strong.char}"${this.options.chars}"` : this.options.language.strong.char;
                charLen = parseInt(formatType.substring(index + 1, index + 2));
                //message += onlyName.char + '至少' + charLen + '个。';
                message += `${onlyName.char}${this.options.language.strong.least}${charLen}${this.options.language.strong.unit}`;
            }
            if (formatType.indexOf('中') !== -1) {
                let index = formatType.lastIndexOf("中");
                onlyName.cn = this.options.language.strong.chinese;
                cnLen = parseInt(formatType.substring(index + 1, index + 2));
                //message += onlyName.cn + '至少' + cnLen + '个。';
                message += `${onlyName.cn}${this.options.language.strong.least}${cnLen}${this.options.language.strong.unit}`;
            }
            //必须至少一个或多个
            lowerLen ? regExp += '(?=' + Array(lowerLen + 1).join('.*?[a-z]') + ')' : null;
            upperLen ? regExp += '(?=' + Array(upperLen + 1).join('.*?[A-Z]') + ')' : null;
            numberLen ? regExp += '(?=' + Array(numberLen + 1).join('.*?[0-9]') + ')' : null;
            charLen ? regExp += '(?=' + Array(charLen + 1).join('.*?[' + this.options.chars + ']') + ')' : null;
            cnLen ? regExp += '(?=' + Array(cnLen + 1).join('.*?[\u4e00-\u9fa5]') + ')' : null;
            //追加范围限制
            if (!axIsEmpty(scale)) {
                if (formatMin) {
                    //如果有字符限制，直接连接{min,max}，结果类似：^(?=.*?[A-Z].*?[A-Z])[a-zA-Z0-9]{4,}$
                    regExp += '[' + scale + ']';
                } else {
                    //如果没有字符限制，则使用‘+’结尾，结果类似：^(?=.*?[A-Z].*?[A-Z])[a-zA-Z0-9]+$
                    regExp += '[' + scale + ']+';
                }
            } else if (axIsEmpty(scale) && formatMin) {
                //如果没有字符范围，则使用‘.’连接，结果类似：^(?=.*?[A-Z].*?[A-Z]).{4,}$
                regExp += '.'
            } else {
                //如果字符范围和长度限制都没有，结果类似：^(?=.*?[A-Z].*?[A-Z]).{0,}$
                regExp += '.{0,}'
            }
            //追加字符限制
            if (formatMin && lengthOnly && !formatMax) {
                regExp += '{' + formatMin + '}';
                // message += '总共只能是' + formatMin + '个字符。';
                message += axTplEngine(this.options.language.strong.total, { value: formatMin});
            } else if (formatMin && !formatMax) {
                regExp += '{' + formatMin + ',}';
                //message += '总共至少' + formatMin + '个字符。';
                message += axTplEngine(this.options.language.strong.totalMore, { value: formatMin});
            } else if (formatMin && formatMax) {
                regExp += '{' + formatMin + ',' + formatMax + '}';
                //message += '总共至少' + formatMin + '个字符，且不能超过' + formatMax + '个字符。';
                message += axTplEngine(this.options.language.strong.totalRange, { min: formatMin,max: formatMax});;
            }
            regExp = '^' + regExp + '$';
            return { pattern: regExp, message: message }
        } else {
            return {}
        }
    }
    getStrong(value) {
        //根据填写值判断当前强度
        let upVal = 0,
            lowVal = 0,
            numVal = 0,
            charVal = 0,
            cnVal = 0,
            lenVal = 0;

        if (this.options.format) {
            let minLength = this.options.format.split("#")[1].split(",")[0];
            //有一个小写+1，有一个大写+1，有一个数字+1，有一个特殊字符+1，有一个中文+1，有最少字符数量限制+1
            if (value && value.match(new RegExp('[a-z]'))) { lowVal = 1; };
            if (value && value.match(new RegExp('[A-Z]'))) { upVal = 1; };
            if (value && value.match(new RegExp('[0-9]'))) { numVal = 1; }
            if (value && value.match(new RegExp('[' + this.options.chars + ']'))) { charVal = 1; };
            if (value && value.match(new RegExp('[\u4e00-\u9fa5]'))) { cnVal = 1; }
            if (value && minLength && value.length >= minLength) { lenVal = 1; };
        }
        this.strong = upVal + lowVal + numVal + charVal + cnVal + lenVal;
        return this.strong;
    }
    doTest(pattern, value, message) {
        if (pattern.test(value)) {
            return { verified: true, message: '' };
        } else {
            return { verified: false, message: message };
        }
    }
    getRules() {
        //将options.error,options.success,options.rule合并输出this.rules
        this.rules = {};
        let types = this.options.type.trim().split(",");
        //如果有required类型并且不在第一位则提前
        if (types[0] != 'required' && types.includes('required')) {
            for (let i = 0; i < types.length; i++) {
                if (types[i] === 'required') {
                    types.splice(i, 1);
                    break;
                }
            }
            types.unshift('required');
        }

        types.forEach(item => {
            //定义规则标准格式
            this.rules[item] = ['', ''];
            if (!axIsEmpty(this.options.rule)) {
                //  this.options.rule格式为{isMail:['填写错误','填写正确！'],telephone:['填写错误！','填写正确！']}
                this.options.rule[item] ? this.rules[item] = this.options.rule[item] : null;
            }
            //如果都为空，那么直接使用默认值
            !this.rules[item][0] ? this.rules[item][0] = this.options.error : null;
            !this.rules[item][1] ? this.rules[item][1] = this.options.success : null;
        });
        return this.rules;

    }
    refreshText() {
        //定义图标
        let icon = this.verified ? axIcon.font.successFill : axIcon.font.errorFill;
        //填充内容
        if (this.options.iconShow) {
            this.msgBox.innerHTML = icon + ' ' + this.text;
        } else {
            this.msgBox.innerHTML = this.text;
        }
    }
    refreshDom(mode) {
        let classToggle = () => {
            //根据最终校验结果给input和信息dom添加样式
            if (this.verified) {
                this.msgBox.classList.remove(this.options.boxError)
                this.msgBox.classList.add(this.options.boxSuccess)
                this.targetDom.classList.remove(this.options.inputError);
                this.targetDom.classList.add(this.options.inputSuccess);
            } else {
                this.msgBox.classList.remove(this.options.boxSuccess)
                this.msgBox.classList.add(this.options.boxError)
                this.targetDom.classList.remove(this.options.inputSuccess);
                this.targetDom.classList.add(this.options.inputError);
            }
        };
        //更新校验内容
        this.refreshText();

        if (this.options.mode.nodeType == 1 || mode.substr(0, 1) == "#") {
            //如果是#id或者是一个dom对象，则将内容追加进该dom中
            let dom = axIdToDom(mode);

            if (!this.msgBox.parentNode) {
                dom.appendChild(this.msgBox);
            }
            //根据校验通过与否添加风格样式
            classToggle();

        } else if (mode == 'note') {
            //如果以input右侧显示
            this.msgBox.classList.add('ax-form-txt');
            let parent = this.targetDom.closest('.ax-form-con'),
                brother = parent.nextElementSibling;
            if (this.type == 'raido' || this.type == 'checkbox') {
                //这两类特别处理，生成dom前删掉之前的dom
                if (brother && brother.hasAttribute('valid')) {
                    brother.remove();
                }
                parent.insertAdjacentElement('afterend', this.msgBox);
            } else {
                if (brother != this.msgBox) {
                    parent.insertAdjacentElement('afterend', this.msgBox);
                }
            }

            //根据校验通过与否添加风格样式
            classToggle();

        } else if (mode == 'popup') {
            //如果以悬浮窗显示
            if (!this.popup) {
                this.popup = new axPopup(this.targetDom.closest('.ax-form-input'), axExtend({
                    content: this.msgBox,
                    padding: false,
                    placement: 'bottom-start',
                    offsetShow: false,
                    arrow: false,
                    close: false,
                    footerShow: false,
                    theme: 'crude',
                    stay: true,
                }, this.options.popup)).show();
            } else {
                this.popup.content.innerHTML = '';
                this.popup.content.appendChild(this.msgBox);
            }
            //根据校验通过与否添加风格样式
            classToggle();

        } else if (mode == 'message') {
            //如果以消息窗显示
            if (!this.verified) {
                new axMessage(axExtend({
                    content: this.text,
                    result: 'error',
                    iconShow: this.options.iconShow,
                }, this.options.message)).show();
            } else {
                if (this.options.successShow) {
                    new axMessage(axExtend({
                        content: this.text,
                        result: 'success',
                        iconShow: this.options.iconShow,
                    }, this.options.message)).show();
                }
            }
            //根据校验通过与否添加风格样式
            classToggle();

        } else {
            //默认紧跟着input显示（input下方另起一行）
            let brother = this.targetDom.closest('.ax-form-input');
            if (this.type == 'radio' || this.type == 'checkbox') {
                //这两类特别处理，生成dom前删掉之前的dom
                if (brother.nextElementSibling && brother.nextElementSibling.hasAttribute('valid')) {
                    brother.nextElementSibling.remove();
                }
                brother.insertAdjacentElement('afterend', this.msgBox);
            } else {
                if (brother != this.msgBox) {
                    brother.insertAdjacentElement('afterend', this.msgBox);
                }
            }

            this.msgBox.classList.add('ax-valid');
            //根据校验通过与否添加风格样式
            classToggle();
        }

    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }

}
/**
 * 点击提交表单校验，通常是对button使用js提交的时候使用
 * e：通常是window
 * data：数组数据，格式为：[{parent:'',children:{name:'',verified:'',text:''}},{}]
 * outer:表单项的父层，可以是form节点，也可以是某div节点，但是必须与表单项的父节点一致，所以建议是form节点
 * callback：校验成功后执行回调函数
 * interval：表单项校验的间隔时间，纯是为了显示逐个校验效果，默认100毫秒
*/
const axValidRepeat = (e, outer, callback, options = { data: axValids, interval: 100 }) => {
    if (!outer) {
        console.error('A form node must be filled in!');
        return false;
    }
    if (axIsEmpty(options.data)) {
        console.error(axError.required);
        return false;
    }
    if (axIsEmpty(options.data.filter(i => i.parent == outer))) {
        console.error('No submittable form found!');
        return false;
    }

    //复制数据源
    let formValids = [...options.data],
        children = formValids.filter(item => item.parent == outer)[0].children,
        k = 0,//k用于判断是否循环到最后一个
        judge = () => {
            //检验后会生成新的axValids，如果子项有一个校验失败则阻止提交
            let form = options.data.filter(i => i.parent == outer)[0];
            if (form.children.some(i => !i.verified)) {
                axPreventDefault(e);
            } else {
                //提交表单校验通过执行部分
                callback && callback();
            }

        };

    //将每个校验项的触发方式改成submit
    children.forEach((i, index) => {
        i.instance.options.trigger = 'submit';
        //根据新的options重新挨个校验,加载定时器可看到按顺序执行效果
        setTimeout(() => {
            k++;
            i.instance.init(function () {
                //在循环最后一个的时候执行判断
                if (k == children.length) {
                    judge();
                }
            });
        }, options.interval * index);
    });
}

(() => {
    [...document.querySelectorAll('[axValid]')].filter(item => {
        return item.getAttribute('name');
    }).forEach(item => {
        new axValid(item);
    });
    //点击submit按钮提交表单再次检验
    //先复制
    if (!axIsEmpty(axValids)) {
        //复制axValids
        let formValids = [...axValids];
        formValids.forEach(item => {
            //监听表单type=submit的按钮提交事件，注意如果提交按钮不是type=submit，那么不会执行submit事件
            if (item.parent.nodeName == 'FORM') {
                //只监听form表单的submit事件，因为其他Dom也可能触发submit事件
                item.parent.addEventListener("submit", function (e) {
                    //先阻止提交
                    axPreventDefault(e);
                    axValidRepeat(e, item.parent, () => {
                        item.parent.submit();
                    });
                }, false);
            }

        })
    }

})();
/*!
 * 插件：折叠菜单；使用方法：new axMenu('#id',{参数})
 */
class axMenu {
    constructor(targetDom, options) {
        // 始终将targetDom转成DOM对象
        this.targetDom = axIdToDom(targetDom);
        // 定义参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//存储名称，字符串格式；如果需要保存实例当前数据请设定唯一的值，比如：storageName:'mystorage001'
            theme: 'light',//主体颜色风格，默认light白底黑子，可选择dark深色风格
            content: '',//设置数据源，支持数组格式数据和链接字符串，如果是链接字符串则为异步获取json
            dispose: false,//是否禁止使用，默认是false即不禁用，可选择true禁用，头部将不能点击操作
            active: '',//初始化展开项，支持多种格式：一个字符串/一个数字/一个DOM节点/一个对象，或以上数组，默认为空即不展开
            disabled: '',//初始化禁用项，支持多种格式：一个字符串/一个数字/一个DOM节点/一个对象，或以上数组，默认为空即不使用禁用状态
            multiple: false,//是否允许同时展开多个父节点，默认false不允许，可填true允许
            collapseAll: true,//初始化是否默认折叠所有父节点，默认是true折叠，可选择false不折叠
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            zIndex: 0,//气泡层级，默认为0即不设置，如果与其他组件发生遮挡请自行设置
            trigger: 'click',//打开菜单的事件，默认是click点击显示，可选择hover即鼠标划入展开移除折叠
            rootStart: -1, //顶层的编号,数字类型，menu本质是一个tree，每一项都需要pId和id
            idStart: 0,//id开始编号，数字类型；如果从物理节点创建菜单，则需要自动创建项目id，默认第一项id为0
            arrowShow: true,//是否显示指示箭头，默认显示
            max: 0,//最多允许多少项，默认为0即不限制
            targetClass: '',//给目标节点即ax-menu添加class类，一个多个均可，多个class用英文逗号或空格分开
            addToActive: true,//是否在添加子项后让新子项成为焦点，默认true，可填false
            gapless: false,//菜单是否是无间隙的，默认false即有间隙，可设为true则两侧的padding将取消
            lamp: '',//展开状态和selected状态的风格，默认为空即在展开的li左侧显示一道竖线；可选择dot即在selected项左侧显示一个小点指示；也可选择line即在selected项左侧显示一条短竖线
            full: false,//是否铺满整个父层，默认false即按220px宽度展示，可设为true则按100%铺满父层，对nav模式无效
            nav: {//nav参数用于水平导航，见ax-navigation.php
                enable: false,//是否将菜单横向展示，也就是变成导航菜单，默认不启用
                headerWidth: '',//设定水平菜单中一级菜单宽度，填写举例：300px、30rem或auto
                bodyWidth: '',//设定水平菜单中下拉菜单宽度，填写举例：300px、30rem或auto
                marginLeft: '',//设定水平菜单中一级菜单之间的左间隔，填写举例：'28px'
                marginRight: '',//设定水平菜单中一级菜单之间的右间隔，填写举例：'28px'
                align: 'left',//导航文字排列方式，默认left居左对齐，可选择center或right
            },
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onTreeCompleted: '',//回调函数，创建完整的treeData后执行，支持一个参数即treeData
            onPlanted: '',//回调函数，节点渲染完成后执行，无参数
            onTrigger: '',//回调函数，节点触发后执行（click或hover），支持一个参数即当前触发项
            onAdd: '',//回调函数，新增项后执行，支持一个参数即新增项
            onRemove: '',//回调函数，删除项后执行，无参数
            onDisable: '',//回调函数，禁用项后执行，支持一个参数即禁用项
            onEnable: '',//回调函数，启用项后执行，支持一个参数即启用项
            onDisableALL: '',//回调函数，全部禁用项后执行，无参数
            onEnableAll: '',//回调函数，全部启用项后执行，无参数
            onExpand: '',//回调函数，项目展开前执行，支持一个参数即当前展开项
            onExpanded: '',//回调函数，项目展开后执行，支持一个参数即当前展开项
            onCollapse: '',//回调函数，项目折叠前执行，支持一个参数即当前折叠项
            onCollapsed: '',//回调函数，项目折叠后执行，支持一个参数即当前折叠项
            onSelected: '',//回调函数，项目点中后执行，支持一个参数即当前selected项
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onUpdateContent: '',//回调函数，子项内容更新后执行，支持一个参数即当前更新项
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
            onExpandAll: '',//回调函数，全部展开后执行，无参数
            onCollapseAll: '',//回调函数，全部折叠后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        // 保存this
        let _this = this;
        // 定义监听事件成员
        this.handlers = {};
        //触发事件函数
        this.triggerShow = function () {
            _this.toggle(this);
        }
        //selected事件
        this.triggerSelected = (item) => {
            item.headerDom.setAttribute('selected', '');
            item.selected = true;
            //其他终极节点取消selected
            let others = this.flatData.filter(i => i !== item && i.selected && !i.children);
            others.forEach(i => {
                i.headerDom.removeAttribute('selected');
                delete i.selected;
            });
            //回调
            this.options.onSelected && this.options.onSelected.call(this, item);
            'selected' in this.handlers ? this.emit('selected', item) : null;
            //监听回调
            'trigger' in this.handlers ? this.emit('trigger', item) : null;
            this.options.onTrigger && this.options.onTrigger.call(this, item);
            //保存数据
            this.save();
        }
        //从物理节点获取的原始的html内容
        this.rawHTML = ''
        //初始化
        this.init();
    };
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'menu');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;
        //初始化数据
        this.flatData = [];
        this.treeData = [];
        //判断是否有存储值而更新参数
        if (this.options.storageName) {
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            } else {
                //则先存下空对象
                axLocalStorage.set(this.options.storageName, {});
            }
        }
        //获得标准数据并渲染
        this.resolveSource();
        return this;
    }
    async resolveSource() {
        if (!axIsEmpty(this.options.content)) {
            //content不为空则按content生成节点
            //先初始化清空
            this.targetDom.innerHTML = '';
            //源tree数据
            if (axType(this.options.content) === 'String') {
                //异步同域获取json
                await axTreeMethod.getFullData({
                    source: this.options.content,
                    ajaxType: this.options.ajaxType,
                    async: true,
                    rootStart: this.options.rootStart,
                    idStart: this.options.idStart,
                    opened: (response) => {
                        this.contentXhr = response.xhr;
                    },
                    before: (response) => {
                        this.targetDom.innerHTML = response.content;
                    },
                    success: () => {
                        this.targetDom.innerHTML = '';
                    }
                }).then((result) => {
                    this.treeData = result.data;
                });
            } else if (axType(this.options.content) === 'Array') {
                //同步数据
                if (this.options.content.length === 0) {
                    console.warn('Array is empty!');
                    return false;
                }
                await axTreeMethod.getFullData({
                    source: this.options.content,
                    rootStart: this.options.rootStart,
                    idStart: this.options.idStart,
                }).then((result) => {
                    this.treeData = result.data;
                });
            }
            //根据树数据创建树节点
            this.arrayToDom(this.treeData);
        } else {
            //content为空的话，则判断是否有物理节点
            if (this.targetDom.innerHTML) {
                //保存原始html
                this.rawHTML = this.targetDom.innerHTML;
                //将节点转为数据
                this.treeData = this.ulToArr();
                //更新treeData，加上pId、id和path
                this.treeData = axTreeMethod.addPath(this.treeData, this.options.rootStart, this.options.idStart).data;
            }
        }
        //监听回调
        this.options.onTreeCompleted && this.options.onTreeCompleted.call(this, this.treeData);
        'treeCompleted' in this.handlers ? this.emit('treeCompleted', this.treeData) : null;
        //根据数据执行剩余代码
        this.processDataFormat(this.treeData);
    }
    processDataFormat(data) {
        //data本身就是treeData

        //拍平后的一纬数组
        this.flatData = axArrToFlat(data);
        //初始化参数
        this.initParms();

        //渲染完毕执行
        this.renderFinish();
        //监听回调
        this.options.onInit && this.options.onInit.call(this);
        'init' in this.handlers ? this.emit('init', '') : null;
    }
    initParms() {
        //给dom设置系列属性
        this.initAttribute();
        //获得expanded和selected项并设置
        this.initActive();
        //获得disabled项并设置
        this.initDisabled();
    }
    ulToArr() {
        let data = [],
            start = this.options.idStart,
            getJson = (elem, index) => {
                let obj = {}, ul = elem.querySelector('ul');
                obj.id = index;
                obj.wrapperDom = elem;
                obj.headerDom = elem.firstElementChild;
                obj.labelDom = obj.headerDom.querySelector('[label]');
                obj.label = obj.labelDom.textContent.trim();
                obj.tipDom = obj.headerDom.querySelector('[tip]');
                obj.tip = obj.tipDom ? obj.tipDom.textContent.trim() : '';
                obj.url = obj.labelDom.href && !obj.labelDom.href.includes('##') ? obj.labelDom.href : '';
                //url链接的target属性
                obj.target = obj.labelDom.target;
                obj.arrowDom = obj.headerDom.querySelector('[arrow]');
                obj.legendDom = obj.headerDom.querySelector('[legend]');
                obj.legendDom ? obj.legend = obj.legendDom.getAttribute('class') : null;
                elem.hasAttribute('expanded') ? obj.expanded = true : null;
                obj.headerDom.hasAttribute('selected') ? obj.selected = true : null;
                obj.headerDom.hasAttribute('disabled') ? obj.disabled = true : null;

                if (ul) {
                    obj.children = [];
                    obj.bodyDom = ul;
                    [...ul.children].forEach(item => {
                        start++;
                        obj.children.push(getJson(item, start));
                    });
                }
                return obj;
            }

        let getData = () => {
            [...this.targetDom.children].forEach(item => {
                data.push(getJson(item, start));
                start++;
            });
        }
        getData();
        return data;
    }
    childToParent(item) {
        //取消onclick选择项目
        item.headerDom.onclick = null;
        item.expanded = true;
        item.wrapperDom.setAttribute('expanded', '');
        item.arrowDom = this.options.arrowShow && item.children ? axAddElem('i', { arrow: '', class: 'ax-iconfont ax-icon-right' }) : null;
        item.headerDom.appendChild(item.arrowDom);
        item.bodyDom = axAddElem('ul', { style: 'display:block' });
        return item;
    }
    createHeaderDom(item) {
        //创建节点
        item.labelDom = axAddElem(item.url ? 'a' : 'i', { label: '' }, item.label);
        item.url ? item.labelDom.href = item.url : null;
        item.target ? item.labelDom.target = item.target : null;
        item.legendDom = item.legend ? axAddElem('i', { legend: '', class: `ax-iconfont ${item.legend}` }) : null;
        item.arrowDom = this.options.arrowShow && item.children ? axAddElem('i', { arrow: '', class: 'ax-iconfont ax-icon-right' }) : null;
        //判断是否填了url或target
        item.headerDom = axAddElem('div');
        item.tipDom = item.tip ? axAddElem('i', { tip: '' }, item.tip) : null;
        item.wrapperDom = axAddElem('li');
        //加入节点
        item.wrapperDom.appendChild(item.headerDom);
        item.legendDom ? item.headerDom.appendChild(item.legendDom) : null;
        item.headerDom.appendChild(item.labelDom);
        item.tipDom ? item.headerDom.appendChild(item.tipDom) : null;
        item.arrowDom ? item.headerDom.appendChild(item.arrowDom) : null;
    }
    arrayToDom(data) {

        let outer = axAddElem('ul'),
            fragment = document.createDocumentFragment();
        //obj = { id: '', label: '', selected: '', expanded: '', disabled: '',tip:'',url:'',target:'',floor:2 ,children: ''}
        let plantTree = (parent, data) => {
            let ul = axAddElem('ul');
            data.forEach(item => {
                //创建header节点
                this.createHeaderDom(item);
                //追加children属性
                if (item.hasOwnProperty('children')) {
                    //继续循环并追加bodyDom属性
                    item.bodyDom = plantTree(item.wrapperDom, item.children);
                }
                ul.appendChild(item.wrapperDom);
            });
            //最外围的UL写入fragment
            parent.appendChild(ul);
            return ul;
        }
        plantTree(outer, data);
        //不要ul外标签，只取一级li标签
        let list = outer.childNodes[0].childNodes;
        [...list].forEach(item => {
            fragment.appendChild(item);
        });
        //最终生成页面节点
        this.targetDom.appendChild(fragment);
        //监听回调
        this.options.onPlanted && this.options.onPlanted.call(this);
        'planted' in this.handlers ? this.emit('planted', '') : null;
    }
    initAttribute() {
        //强行带上ax-menu类
        !this.targetDom.classList.contains('ax-menu') ? this.targetDom.classList.add('ax-menu') : null;
        //给主要外框带上class类
        this.options.targetClass ? axClass.add(this.targetDom, this.options.targetClass) : null;
        //颜色风格
        this.targetDom.setAttribute('theme', this.options.theme);
        //无间隙设置
        this.options.gapless ? this.targetDom.setAttribute('gapless', '') : null;
        //expanded和selected风格
        this.options.lamp ? this.targetDom.setAttribute('lamp', this.options.lamp) : null;
        //设置铺满
        this.options.full && !this.options.nav.enable ? this.targetDom.setAttribute('full', '') : null;
        //设置zIndex
        this.options.zIndex ? this.targetDom.style.zIndex = this.options.zIndex : null;
        //启用横向导航
        if (this.options.nav.enable) {
            //追加nav属性
            this.targetDom.setAttribute('nav', '');
            //追加align属性
            this.options.nav.align ? this.targetDom.setAttribute('align', this.options.nav.align) : null;
            //一级菜单使用width
            if (this.options.nav.headerWidth) {
                let firstObj = this.flatData.filter(k => k.floor === 1);
                firstObj.forEach(k => {
                    k.wrapperDom.style.width = this.options.nav.headerWidth;
                });
            }
            //ul使用width
            if (this.options.nav.bodyWidth) {
                this.flatData.forEach(k => {
                    k.bodyDom ? k.bodyDom.style.width = this.options.nav.bodyWidth : null;
                })
            }
            //一级li使用左右边距
            if (this.options.nav.marginLeft) {
                this.flatData.filter(k => k.floor === 1).forEach(k => {
                    k.wrapperDom.style.marginLeft = this.options.nav.marginLeft;
                })
            }
            if (this.options.nav.marginRight) {
                this.flatData.filter(k => k.floor === 1).forEach(k => {
                    k.wrapperDom.style.marginRight = this.options.nav.marginRight;
                })
            }
        }
    }
    getPathIds(data) {
        if (data.length === 0) {
            return [];
        }
        //id去重
        let idsArr = [],
            //如果multiple为false则只取第一个
            idsItems = !this.options.multiple ? data.slice(0, 1) : data;
        idsItems.forEach(k => {
            let item = axFindItem(k, this.flatData), idsPath;
            if (item) {
                //节点path加入ids合集
                idsPath = item.path.split('>');
                idsArr.push(...idsPath);
            }
        });
        //id去重，并转成数字
        return [...new Set(idsArr)].map(k => parseInt(k));
    }
    initActive() {
        //如果数据为空则不再执行
        if (this.flatData.length === 0) {
            return false;
        }

        if (this.options.multiple && !this.options.collapseAll) {
            //全部展开
            this.expandAll();
        } else {
            let itemsActive = this.options.active !== '' ? axValueToArr(this.options.active) : [],
                idsActive;
            //再从数据中获取expanded项和selected项
            itemsActive.push(...this.flatData.filter(k => k.selected), ...this.flatData.filter(k => k.expanded));
            //获得所有expanded和selected项的id
            idsActive = this.getPathIds(itemsActive);

            this.flatData.forEach(k => {
                if (idsActive.includes(k.id)) {
                    if (k.children) {
                        //父节点则展开
                        this.expand(k);
                    } else {
                        //终极节点则selected
                        this.triggerSelected(k);
                    }
                }
            });
        }
    }
    initDisabled() {
        //如果数据为空则不再执行
        if (this.flatData.length === 0) {
            return false;
        }
        let items = this.flatData.filter(k => k.disabled),
            itemArr = this.options.disabled !== '' ? axValueToArr(this.options.disabled) : [];
        itemArr.push(...items);
        itemArr.forEach(k => {
            let item = axFindItem(k, this.flatData);
            if (item) {
                item.disabled = true;
                item.headerDom.setAttribute('disabled', '');
            }
        });
    }
    renderFinish() {
        //启用状态才可点击操作
        if (!this.options.dispose) {
            //触发头部标题
            this.flatData.forEach(k => {
                //有子元素则是展开关闭
                if (k.children) {
                    //未禁止才可触发
                    if (!k.disabled) {
                        //如果是一个链接，那么触发的节点应该是arrowDom
                        let triggerDom = k.url && k.arrowDom ? k.arrowDom : k.headerDom;

                        if (this.options.trigger === 'click') {
                            triggerDom.addEventListener(this.options.trigger, this.triggerShow, false);
                        } else if (this.options.trigger === 'hover') {
                            this.hover = new axHover(k.wrapperDom, {
                                enter: () => {
                                    this.expand(k);
                                    //监听回调
                                    'trigger' in this.handlers ? this.emit('trigger', k) : null;
                                    this.options.onTrigger && this.options.onTrigger.call(this, k);
                                },
                                leave: () => {
                                    this.collapse(k);
                                }
                            });
                        }
                    }
                } else {
                    if (!k.disabled) {
                        //如果没有子元素则是selected
                        k.headerDom.onclick = () => {
                            this.triggerSelected(k);
                        }
                    }
                }
                //回调
                k.callback && k.callback.call(this, k);
            });
        }
        //保存数据
        this.save();
    }
    workToggle(item, type = 'disable') {
        let triggerDom = item.url && item.arrowDom ? item.arrowDom : item.headerDom;
        if (type === 'enable') {
            item.disabled ? delete item.disabled : null;
            item.headerDom.removeAttribute('disabled');
            if (!item.children) {
                //无ul则恢复点击，
                item.headerDom.onclick = () => {
                    this.triggerSelected(item);
                }
            } else {
                //有ul则绑定
                triggerDom.addEventListener(this.options.trigger, this.triggerShow, false);
            }
        } else if (type === 'disable') {
            item.disabled = true;
            item.headerDom.setAttribute('disabled', '');
            if (!item.children) {
                //无ul则不能点击，不能selected
                item.headerDom.onclick = null;
            } else {
                //有ul则解绑
                triggerDom.removeEventListener(this.options.trigger, this.triggerShow, false);
            }
        }
        //保存数据
        this.save();
    }
    getDisabled() {
        return this.flatData.filter(k => k.disabled);
    }
    getExpanded() {
        return this.flatData.filter(k => k.expanded);
    }
    getActive() {
        return this.flatData.filter(k => k.expanded || k.selected);
    }
    getSelected() {
        return this.flatData.find(k => k.selected);
    }
    collapseAll(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.flatData.filter(k => k.expanded).forEach(k => {
            this.collapse(k);
        });
        //监听回调
        'collapseAll' in this.handlers ? this.emit('collapseAll', '') : null;
        this.options.onCollapseAll && this.options.onCollapseAll.call(this);
        callback && callback.call(this);
    }
    expandAll(callback) {
        //如果已经销毁或不允许多个展示则不再执行
        if (this.destroyed || !this.options.multiple) {
            return this;
        }
        this.flatData.forEach(k => {
            this.expand(k);
        });
        //监听回调
        'expandAll' in this.handlers ? this.emit('expandAll', '') : null;
        this.options.onExpandAll && this.options.onExpandAll.call(this);
        callback && callback.call(this);
    }
    disable(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj为空则不再执行,0除外
        if (obj !== 0 && axIsEmpty(obj)) {
            return false;
        }
        let disableFun = (k) => {
            let item = axFindItem(k, this.flatData);
            //找到元素并设置禁用
            if (item) {
                this.workToggle(item, 'disable');
            }
        }
        //设置
        if (Array.isArray(obj)) {
            obj.forEach(k => {
                disableFun(k);
            });
        } else {
            disableFun(obj);
        }
        //保存数据
        this.save();
        //监听回调
        let disableds = this.flatData.filter(k => k.disabled);
        'disable' in this.handlers ? this.emit('disable', disableds) : null;
        this.options.onDisable && this.options.onDisable.call(this, disableds);
        callback && callback.call(this, disableds);
    }
    disableAll(callback) {
        this.disable(this.flatData);
        //监听回调
        'disableAll' in this.handlers ? this.emit('disableAll', '') : null;
        this.options.onDisableAll && this.options.onDisableAll.call(this);
        callback && callback.call(this);
    }
    enable(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj为空则不再执行,0除外
        if (obj !== 0 && axIsEmpty(obj)) {
            return false;
        }
        let enableFun = (k) => {
            let item = axFindItem(k, this.flatData);
            //找到元素并设置禁用
            if (item) {
                this.workToggle(item, 'enable');
            }
        }
        //设置
        if (Array.isArray(obj)) {
            obj.forEach(k => {
                enableFun(k);
            });
        } else {
            enableFun(obj);
        }
        //保存数据
        this.save();
        //监听回调
        let enableds = this.flatData.filter(k => !k.disabled);
        'enable' in this.handlers ? this.emit('enable', enableds) : null;
        this.options.onEnable && this.options.onEnable.call(this, enableds);
        callback && callback.call(this, enableds);
    }
    enableAll(callback) {
        this.enable(this.flatData);
        //监听回调
        'enableAll' in this.handlers ? this.emit('enableAll', '') : null;
        this.options.onEnableAll && this.options.onEnableAll.call(this);
        callback && callback.call(this);
    }
    toggle(obj) {
        //如果label是A链接，那么点击指示箭头才能展开，此时的obj是指示箭头的DOM
        let item = axFindItem(obj, this.flatData) || axFindItem(obj, this.flatData, { dom: 'arrowDom' });
        if (item.expanded) {
            this.collapse(item);
        } else {
            this.expand(item);
        }
        //监听回调
        'trigger' in this.handlers ? this.emit('trigger', item) : null;
        this.options.onTrigger && this.options.onTrigger.call(this, item);
    }
    collapse(obj, callback) {
        let type = axType(obj),
            item = type === 'Object' ? obj : axFindItem(obj, this.flatData);
             //如果是隐藏状态则不再执行
        if (!item || !item.children || !item.shown) {
            return false;
        }

        //监听回调
        'collapse' in this.handlers ? this.emit('collapse', item) : null;
        this.options.onCollapse && this.options.onCollapse.call(this, item);
        //设置属性
        delete item.expanded;
        item.wrapperDom.removeAttribute('expanded');
        //关闭ul
        axSlideUp(item.bodyDom, '', () => {
            //监听回调
            Reflect.deleteProperty(item,'shown');
            'collapsed' in this.handlers ? this.emit('collapsed', item) : null;
            this.options.onCollapsed && this.options.onCollapsed.call(this, item);
            callback && callback.call(this, item);
        });
        //保存数据
        this.save();
    }
    expand(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj可以是对象，字符串或数字，如果是对象直接相等不再找
        let item = axFindItem(obj, this.flatData), others;

        //如果是显示状态则不再执行
        if(item.shown){
            return false;
        }

        if (!axIsEmpty(item) && item.children) {
            //监听回调
            'expand' in this.handlers ? this.emit('expand', item) : null;
            this.options.onExpand && this.options.onExpand.call(this, item);
            //设置属性
            item.expanded = true;
            item.wrapperDom.setAttribute('expanded', '');
            //展开ul
            axSlideDown(item.bodyDom, '', () => {
                //监听回调
                item.shown = true;
                'expanded' in this.handlers ? this.emit('expanded', item) : null;
                this.options.onExpanded && this.options.onExpanded.call(this, item);
                callback && callback.call(this, item);
            });

            //其他兄弟节点
            if (!this.options.multiple) {
                //不是自己，expanded状态，同时层级相同
                others = this.flatData.filter(k => (k !== item && k.expanded && k.floor === item.floor && k.children));
                others.forEach(k => {
                    this.collapse(k);
                });
            }
            //保存数据
            this.save();
        }
    }
    active(obj, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //obj可以是对象，字符串或数字，如果是对象直接相等不再找
        let activeArr = [],
            activeItems = [],
            activeFun = (k) => {
                let item = axFindItem(k, this.flatData);
                //找到元素并设置禁用
                if (item) {
                    let ids = item.path.split('>').map(k => parseInt(k));
                    ids.forEach(k => {
                        this.expand(k);
                    });
                    //如果是终极节点则设为selected状态
                    if (!item.children) {
                        this.triggerSelected(item);
                    }
                }
            }
        //根据obj类型和multiple参数确认active数组
        if (Array.isArray(obj)) {
            if (!this.options.multiple) {
                //取第一项
                activeArr = obj.slice(0, 1);
            } else {
                activeArr = obj
            }
        } else {
            activeArr = [obj]
        }
        activeItems = activeArr.map(k => axFindItem(k, this.flatData)).filter(Boolean);
        activeArr.forEach(k => {
            activeFun(k);
        });
        //监听回调
        'active' in this.handlers ? this.emit('active', activeItems) : null;
        this.options.onActive && this.options.onActive.call(this, activeItems);
        callback && callback.call(this, activeItems);
        //保存数据
        this.save();
    }
    add(data, target, isChild = false, isFront = false, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //当前data的length大于等于允许的max值则不再允许添加了
        if (this.flatData.length >= this.options.max && this.options.max !== 0) {
            console.warn('The number of children exceeds the maximum!');
            return false;
        }
        let type = axType(data),
            items = [],
            appendFun = (source, target, isChild, isFront) => {
                let targetData;
                if (source.hasOwnProperty('pId') && source.pId !== this.options.rootStart) {
                    //传来的source直接带了pId，则target从pId找，pId可能是字符串，所以强制从id匹配
                    targetData = axFindItem(source.pId, this.flatData);
                    isChild = true;
                } else {
                    //传来的source为其他
                    if (target === undefined || target === 'end') {
                        //tree节点最末尾
                        targetData = this.flatData.filter(k => k.floor === 1).slice(-1)[0];
                    } else if (target === 'start') {
                        //tree节点最前面
                        targetData = this.flatData.filter(k => k.floor === 1)[0];
                        isFront = true;
                    } else {
                        //id、字符串或对象
                        targetData = axFindItem(target, this.flatData);
                    }
                }

                //获得{}数据，但是还没有节点属性
                let sourceData = axTreeMethod.createData(source, targetData, this.flatData, isChild);
                //生成节点属性
                this.createHeaderDom(sourceData);

                if (isChild) {
                    //类型是添加子节点
                    if (targetData.children) {
                        //已经有子节点
                        axTreeMethod.existChildPush(sourceData, targetData);
                    } else {
                        //并没有子节点
                        axTreeMethod.noExistChildPush(sourceData, targetData, () => {
                            //将target转成有ul的节点
                            return this.childToParent(targetData);
                        });
                    }
                } else {
                    //类型是添加兄弟节点
                    let parent = this.flatData.find(i => i.id == targetData.pId),
                        //顶级节点没有父节点了，所以children直接等于flatData
                        children = parent ? parent.children : this.treeData,
                        index = children.indexOf(targetData);
                    //插入
                    if (isFront) {
                        //如果是第一项则直接头部追加
                        index == 0 ? children.unshift(sourceData) : children.splice(index, 0, sourceData);
                        targetData.wrapperDom.insertAdjacentElement('beforeBegin', sourceData.wrapperDom);
                    } else {
                        children.splice(index + 1, 0, sourceData);
                        targetData.wrapperDom.insertAdjacentElement('afterEnd', sourceData.wrapperDom);
                    }
                }
                //追加到数组中
                this.flatData.push(sourceData);
                items.push(sourceData);
            };

        if (type === 'Array' && data.length > 0) {
            //data是一个数组

            data.forEach(k => {
                appendFun(k, target, isChild, isFront);
            });
        } else {
            //data是一个对象或文本
            appendFun(data, target, isChild, isFront);
        }
        //items有至少一项才执行
        if (items.length > 0) {


            //是否允许新增后成为焦点，让新增的最后一个成为焦点
            if (this.options.addToActive) {
                if (this.options.multiple) {
                    items.forEach(k => {
                        //有children才展开
                        k.children ? this.expand(k) : null;
                    });
                } else {
                    //只取最后一项
                    let lastItem = items[items.length - 1];
                    lastItem.children ? this.expand(lastItem) : null;
                }
            }
            //生成了数据和节点之后操作
            this.renderFinish();
        }
        //回调监听
        'add' in this.handlers ? this.emit('add', items) : null;
        this.options.onAdd && this.options.onAdd.call(this, items);
        callback && callback.call(this, items);
        return this;
    }
    remove(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //data为空则不再执行
        if (axIsEmpty(data)) {
            return false;
        }
        //数据为空了就不能再删除了
        if (this.flatData.length === 0) {
            console.warn('The data is already empty!');
            return false;
        }
        let removeItem = (child) => {
            let item = axFindItem(child, this.flatData),
                data;
            if (item) {
                //从flat和tree中删除数据并清除节点
                data = axTreeMethod.remove(item, this.flatData, this.treeData);
                this.flatData = data.flat;
                this.treeData = data.tree;
            }
        };
        //删除
        if (Array.isArray(data)) {
            data.forEach(k => {
                removeItem(k);
            });
        } else {
            removeItem(data);
        }
        //保存数据
        this.save();
        //回调监听
        'remove' in this.handlers ? this.emit('remove', '') : null;
        this.options.onRemove && this.options.onRemove.call(this);
        callback && callback.call(this);
        return this;
    }
    save(props, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }
        setTimeout(() => {
            //使用异步保存，避免拥堵
            let idsActive = [...this.flatData.filter(k => k.selected), ...this.flatData.filter(k => k.expanded)].filter(Boolean).map(k => k.id),
                idsDisabled = this.flatData.filter(k => k.disabled).map(k => k.id);
            if (!props) {
                //保存当前焦点索引
                axLocalStorage.set(this.options.storageName, { active: idsActive, disabled: idsDisabled, content: this.flatData });
            } else {
                //保存参数,props={}
                !props.hasOwnProperty('active') ? props.active = idsActive : null;
                !props.hasOwnProperty('disabled') ? props.disabled = idsDisabled : null;
                !props.hasOwnProperty('content') ? props.content = this.flatData : null;
                axLocalStorage.set(this.options.storageName, props);
            }
            //回调监听
            let getValue = axLocalStorage.get(this.options.storageName);
            'save' in this.handlers ? this.emit('save', getValue) : null;
            this.options.onSave && this.options.onSave.call(this, getValue);
            callback && callback.call(this, getValue);
            return this;
        }, 0)
    }
    destroy(callback) {
        //取消绑定事件
        this.flatData.forEach(k => {
            let triggerDom = k.url && k.arrowDom ? k.arrowDom : k.headerDom;
            triggerDom.removeEventListener(this.options.trigger, this.triggerShow);
            triggerDom.onclick = null;
            k.headerDom.onclick = null;
        });
        //更新销毁状态
        this.destroyed = true;
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
        //监听回调
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    updateContent(item, content, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //item可是数字、字符串、DOM和{}对象
        let me = axFindItem(item, this.flatData);
        if (me) {
            if (axType(content) === 'String') {
                //content是字符串
                me.label = content;
                me.labelDom.innerHTML = content;
            } else if (axType(content) === 'Object') {
                //content是{}对象
                if (content.label) {
                    me.label = content.label;
                    me.labelDom.innerHTML = content.label;
                }
                if (content.url && me.url) {
                    me.url = content.url;
                    me.labelDom.href = content.url;
                }
                if (content.target && me.url) {
                    me.target = content.target;
                    me.labelDom.target = content.target;
                }
            }
            //监听回调
            'updateContent' in this.handlers ? this.emit('updateContent', me) : null;
            this.options.onUpdateContent && this.options.onUpdateContent.call(this, me);
            callback && callback.call(this, me);
            //保存数据
            this.save();
        }
        return this;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.options = axExtend(this.options, setting);
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
        //初始化html
        this.targetDom.innerHTML = this.rawHTML ? this.rawHTML : '';
        this.init();
        //监听回调
        'update' in this.handlers ? this.emit('update', '') : null;
        this.options.onUpdate && this.options.onUpdate.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/* // 对使用axMenu属性的元素使用本插件
document.querySelectorAll('[axMenu]').forEach(element => {
    new axMenu(element);
}); */

/*!
* 插件：黏住顶部或盒子；使用方法：new axStick('#id',{参数})
*/
const axSticks = [];
class axStick {
    constructor(targetDom, options) {
        //始终将targetDom转成DOM对象
        this.targetDom = axIdToDom(targetDom);
        // 定义参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            refer: 'body',//参照元素。默认参照body黏住顶部，可选择'#ID'：参照指定元素
            space: '0px',//黏住后距离顶部或底部的距离，单位px
            width: '',//默认为空即为原宽，可选择具体宽度值'1200px'，或者'100%'
            zIndex: '',//默认不设定z-index值
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
        }, options, this.targetDom,this.constructor.name);
        // 定义监听事件成员
        this.handlers = {};
        //定义黏住元素固定值
        this.height = this.targetDom.offsetHeight;
        this.width = this.targetDom.offsetWidth;
        this.left = axOffset(this.targetDom).left;
        this.top = axOffset(this.targetDom).top;
        this.bottom = axOffset(this.targetDom).top + this.height;
        this.style = this.targetDom.getAttribute('style');
        this.placeholder = axAddElem('DIV', { style: `width:${this.width}px;height:${this.height}px;display:none;` });
        //插入占位
        this.targetDom.insertAdjacentElement('beforebegin', this.placeholder);
        //初始化
        this.init();
        return this;
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'stick');
          //断点合并参数
          !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //定义参照元素固定值
        this.refer = this.targetDom.closest(this.options.refer);
        this.referHeight = this.refer.offsetHeight;
        this.referWidth = this.refer.offsetWidth;
        this.referLeft = axOffset(this.refer).left;
        this.referTop = axOffset(this.refer).top;
        this.referBottom = axOffset(this.refer).top + this.referHeight;
        //给参照元素添加postion属性，否则不能定位
        if (this.options.refer !== 'body' && (getComputedStyle(this.refer).position !== 'fixed' || getComputedStyle(this.refer).position !== 'absolute')) {
            this.refer.style.position = 'relative';
        }
        //设定黏住元素的z-index
        this.options.zIndex ? this.targetDom.style.zIndex = this.options.zIndex : null;
        //如果定位元素有浮动则赋予占位元素
        getComputedStyle(this.targetDom).float !== 'none' ? this.placeholder.style.float = getComputedStyle(this.targetDom).float : null;

        window.addEventListener("scroll", () => {
            if (this.height > window.innerHeight) {
                //定位元素高度大于视窗高度
                if (this.bottom < document.documentElement.scrollTop + window.innerHeight && this.referBottom > document.documentElement.scrollTop + window.innerHeight) {
                    //参照元素还没完全进入视窗，并且定位元素已经进入视窗，绝对定位黏住顶部
                    this.targetDom.style.cssText += `width:${this.targetDom.offsetWidth}px;`;
                    this.targetDom.style.cssText += `left:${axOffset(this.targetDom).left}px;`
                    this.targetDom.style.cssText += `position:fixed;bottom:${this.options.space};margin:0;`;
                    this.placeholder.style.display = 'block';
                    'sticked' in this.handlers ? this.emit('sticked', 'fixed') : null;// 执行监听事件
                } else if (this.referBottom < document.documentElement.scrollTop + window.innerHeight) {
                    //参照元素完全进入视窗，定位元素相对定位黏住参照元素底部
                    this.targetDom.style.position = 'absolute';
                    this.targetDom.style.left = this.left - this.referLeft + 'px';
                    this.targetDom.style.cssText += `position:absolute;bottom:${this.options.space};margin:0;`;
                    this.placeholder.style.display = 'none';
                    'sticked' in this.handlers ? this.emit('sticked', 'absolute') : null;// 执行监听事件
                } else if (axOffset(this.refer).top > document.documentElement.scrollTop) {
                    //往上滚动，参照元素实时的距顶的距离大于滚动距离时
                    this.targetDom.setAttribute('style', this.style);
                    this.placeholder.style.display = 'none';
                    'unsticked' in this.handlers ? this.emit('unsticked', '') : null;// 执行监听事件
                }

            } else {
                //定位元素高度小于视窗高度
                if (this.top < document.documentElement.scrollTop) {
                    //往下滚动，滚动距离超过定位元素距顶的距离
                    if (!this.options.width) {
                        this.targetDom.style.cssText += `width:${this.targetDom.offsetWidth}px;`;
                        this.targetDom.style.cssText += `left:${this.left}px;`
                    } else {
                        this.targetDom.style.cssText += `width:${this.options.width};`;
                        this.targetDom.style.cssText += `left:calc(50% - ${this.options.width}/2);`;
                    }
                    this.targetDom.style.cssText += `position:fixed;top:${this.options.space};bottom:auto;margin:0;`;
                    this.placeholder.style.display = 'block';
                    'sticked' in this.handlers ? this.emit('sticked', 'fixed') : null;// 执行监听事件
                    //往下滚动，如果参照元素底边距离顶部的距离小于定位元素的高度，则粘在参照元素底部
                    if (this.height > this.referBottom - document.documentElement.scrollTop) {
                        this.targetDom.style.position = 'absolute';
                        this.targetDom.style.left = this.left - this.referLeft + 'px';
                        this.targetDom.style.top = 'auto';
                        this.targetDom.style.bottom = 0;
                        'sticked' in this.handlers ? this.emit('sticked', 'absolute') : null;// 执行监听事件
                    }
                } else {
                    //往上滚动，滚动距离小于定位元素距顶的距离
                    this.targetDom.setAttribute('style', this.style);
                    this.placeholder.style.display = 'none';
                    'unsticked' in this.handlers ? this.emit('unsticked', '') : null;// 执行监听事件
                }
            }
        }, false);
        return this;
    }
    destroy(callback) {
        this.targetDom.lastChild.removeEventListener('click', this.event, false);
        this.targetDom.innerHTML = this.html;
        callback && callback.call(this);
    }
    update(setting, callback) {
        this.targetDom.style = this.style;
        this.options = axExtend(this.options, setting);
        this.init();
        'updated' in this.handlers ? this.emit('updated', '') : null;
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    // 对使用axStick属性的元素使用本插件
    document.querySelectorAll('[axStick]').forEach(element => {
        new axStick(element);
    });
})();
 */
/*!
 * 插件：下拉菜单单选和多选；使用方法：axSelect('#id',{参数})
 */
class axSelect {
    constructor(inputDom, options) {
        //转为节点
        this.inputDom = axIdToDom(inputDom);
        //无节点则不再执行
        if (!this.inputDom) return false;
        //合并参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//存储名称，字符串格式；如果需要保存实例当前数据请设定唯一的值，比如：storageName:'mystorage001'
            least: 0,//至少选择几项，默认为0即不限制，可自定义
            most: 0,//最多可选择几项，默认为0即不限制，可自定义
            content: '',//实例内容，支持多种格式，可以是字符串（包括#id/.className/[name]/DIV等选择器）、节点、数组
            async: false,//如果content是字符串，那么可能是异步请求数据，该参数默认false，可填true
            trigger: 'click',//触发动作，默认是click，另可选hover事件（同axPopup插件的trigger参数）
            selected: '',//设置初始值，如果是一个数字则从id中查找；如果是字符串（包括以英文逗号分隔的字符串）则从id、label和value中查找；如果是对象例如{id:0,label:'北京'}则逐个对比属性值；也可是以上值的数组。
            disabled: '',//设置禁用项，使用方法同selected
            clearShow: false,//是否在右侧显示清除按钮，默认false不显示，可使用true显示
            removable: false,//是否显示标签删除按钮，默认false不显示，可使用true显示
            separator: ',',//字符串分隔符，默认英文逗号，input的value应该与此相同，输出的value值将以此分隔值
            dropdownShow: true,//是否显示下拉菜单，默认true显示，如果只需要行内编辑+回车添加模式则可使用false不显示
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            absentAddable: false,//使用select方法时，如果下拉列表中没有指定项是否自动添加新项并作为selected项
            idStart: 0,//id开始编号，数字类型；如果从物理节点创建菜单，则需要自动创建项目id，默认第一项id为0
            width: '',//设置input和下拉菜单的宽度，默认为空则与.ax-select等宽，可使用px和rem单位
            size: 'md',//设置尺寸，以配合多样化的表单使用，默认md，可设置xs(18px)、sm(28px)和lg(48px)
            readonly: false,//是否禁用下拉菜单，默认false不禁用，可选择true禁用，不能使用触发展开菜单
            enterSelect: true,//是否允许在下拉列表上使用回车键添加selected值，默认true允许，可选择false不允许
            displayProp: '',//回显属性，如果为空则显示label值，如果需要显示value值可对将该参数设为value
            singleClose: true,//如果是单选，单选完成后是否立即自动关闭菜单，默认true，可选择false不关闭
            className: '',//一个或多个className，多个用英文逗号或空格隔开
            language: {
                inputPlaceholder: '请选择...',//input的选择提示
                searchPlaceholder: '输入关键字...',//下拉列表搜索提示
                editPlaceholder: '输入关键字...',//绑定input行内编辑的提示
                absent: `列表中没有"<# this.value #>"可供选择！`,//select操作时不存在提示
                absentAddable: `列表中没有"<# this.value #>"，但是作为新项加入了列表！`,//select操作不存在也能添加提示
                empty: '没有找到符合"<# this.value #>"的选项！',//检索为空时提示
                emptyAddable: '没有找到符合"<# this.value #>"的选项，请使用回车键添加！',//检索为空但是可添加提示
                most: `只能最多选择"<# this.value #>"项！`,//超过most参数提示
                addFail: `添加失败！列表中已经有了"<# this.value #>"这一项。`,//添加失败提示
            },
            search: {
                enable: false,//是否显示检索控件，默认false不显示，可选择true显示
                fuzzy: true, //是否使用模糊检索，默认true，可选择false精确检索
                limit: 0,//筛选结果显示几项，默认为0表示不限制
                reorder: false,//对检索结果是否排序，默认不重排，可选择true按权重高低重排
                editable: true,//绑定input情况下是否可使用删除键和回车键进行删除和添加操作
            },
            dialog: {},//对话实例dialog的自定义参数
            popup: {},//气泡实例popup的自定义参数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onBeforeAdd: '',//回调函数，添加前执行，需手动添加，支持一个参数即将要添加的项
            onAdd: '',//回调函数，新增列表项后执行，支持一个参数即要添加的项
            onBeforeRemove: '',//回调函数，移除前执行，需手动移除，支持一个参数即将要移除的项
            onRemove: '',//回调函数，移除列表项后执行，支持一个参数即要移除的项
            onSelect: '',//回调函数，使用selected后执行，支持一个参数要selected的项
            onUnselect: '',//回调函数，取消selected后执行，支持一个参数要取消selected的项
            onDisabled: '',//回调函数，禁用后执行，支持一个参数要禁用的项
            onEnabled: '',//回调函数，取消禁用后执行，支持一个参数要启用的项
            onClear: '',//回调函数，清除所有selected值后执行，无参数
            onAbsentSelect: '',//回调函数，使用select方法时，列表不存在依然添加select值后执行，支持一个参数即当前selected项
            onToggleSelect: '',//回调函数，点击选项发生selected状态改变时执行，支持一个参数即当前的点击对象
            onMost: '',//回调函数，selected项超出most值后执行，支持一个参数即options.most值
            onLeast: '',//回调函数，selected小于least值后执行，支持一个参数即options.least值
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
        }, options, this.inputDom, this.constructor.name);
        //定义监听事件成员
        this.handlers = {};
        //定义输入类型
        if (this.inputDom.nodeName === 'SELECT') {
            if (this.inputDom.hasAttribute('multiple')) {
                this.inputType = 'select-multiple';
            } else {
                this.inputType = 'select-single';
            }
        } else if (this.inputDom.nodeName === 'INPUT') {
            this.inputType = 'input';
        }
        //点击选择
        this.clickSelect = (obj) => {
            //禁止的或父节点则不再执行
            if (~~obj.disabled || obj.children) {
                return false;
            }
            let value = this.getDisplayValue(obj),
                tagObj = { id: obj.id, value };
            if (this.inputType === 'select-single') {
                //已经是selected状态则不再执行
                if (~~obj.selected) {
                    return false;
                } else {
                    this.toggleList(obj, this.data);
                    //更新tags
                    this.tagsIns.clear();
                    this.tagsIns.add(tagObj);
                }
                //选择后关闭
                this.options.singleClose ? this.popupIns.hide() : null;
            } else {
                this.selectedAttrToggle(obj, !~~obj.selected);
                if (~~obj.selected) {
                    //添加tag
                    this.tagsIns.add(tagObj);
                } else {
                    //删除tag
                    this.tagsIns.remove(tagObj);
                }
            }
            //切换active
            //this.activeAttrToggle(obj)
        }
        //在列表中切换select
        this.keydownSelect = (data) => {
            if (axIsEmpty(data)) {
                document.onkeydown = null;
                return false;
            }

            document.onkeydown = (e) => {
                let usableItems = data.filter(k => !k.children && !~~k.disabled),
                    curItem = usableItems.find(k => k.active),
                    curIndex = usableItems.findIndex(k => k === curItem) || 0;
                if (e.code === 'ArrowDown') {
                    axPreventDefault(e);
                    //下移
                    this.keyToggleActive('down', curItem, curIndex, usableItems);
                } else if (e.code === 'ArrowUp') {
                    axPreventDefault(e);
                    //上移
                    this.keyToggleActive('up', curItem, curIndex, usableItems);
                } else if (e.code === 'Enter') {
                    //回车选中
                    if (curItem && this.options.enterSelect) {
                        if (this.inputType === 'select-single') {
                            this.select(curItem);
                        } else {
                            !~~curItem.selected ? this.select(curItem) : this.unselect(curItem);
                        }
                    }
                }
            }
        }
        //点击外部失焦
        this.bodyClickBlur = (e) => {
            if (!this.selectDom.contains(e.target)) {
                this.selectDom.classList.remove('ax-opened');
                //既然已经取消focus状态，也取消绑定事件，释放内存
                document.removeEventListener('click', this.bodyClickBlur);
            }
        }
        //初始化
        this.init();
        return this;
    }
    beforeInit() {
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //再次初始化时,清除受参数控制的节点、实例和变量
        if (this.selectDom) {
            //销毁实例
            this.tagsIns.destroy();
            this.tagsIns = null;
            if (this.popupIns) {
                this.popupIns.destroy();
                this.popupIns = null;
            }
            //清空节点
            this.tagsDom.innerHTML = '';
            this.listDom.innerHTML = '';
            //移除searchDom
            this.searchDom ? this.searchDom.remove() : null;
        }
    }
    async init() {
        //先清空
        this.beforeInit();
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'select');
        //设置销毁状态
        this.destroyed = false;
        //自动索引最大值
        this.maxIndex = this.options.idStart;
        //输出值
        this.value = '';
        this.selected = [];
        this.searchResult = [];
        //初始化数据
        this.data = [];
        //判断是否有存储值而更新参数
        if (this.options.storageName) {
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            } else {
                //则先存下空对象
                axLocalStorage.set(this.options.storageName, {});
            }
        }
        //修正width
        this.inputDom.style.width && !this.options.width ? this.options.width = this.inputDom.style.width : null;

        //创建骨架节点
        this.createSkeleton();

        //判断节点是input还是select，select则从options中取得数据
        let contentType = axType(this.options.content);

        if (this.inputType === 'input' && contentType === 'String' && this.options.async) {
            //异步json或sql
            await this.getAsyncData((content) => {
                //获得的数据可能是树形数据，转为扁平一维数据
                this.data = axArrToFlat(content);
            });
        } else {
            //同步数据
            if (this.inputType.includes('select')) {
                //select-single和select-multiple类型输入
                this.getDataFromNode(this.inputDom);
            } else if (this.inputType === 'input') {
                //input类型输入
                if (!axIsEmpty(this.options.content)) {
                    //从content参数中获得数据
                    if (contentType === 'String') {
                        //是#id字符串
                        if (this.options.content.trim().startsWith('#')) {
                            this.getDataFromNode(this.options.content);
                        } else {
                            this.getDataFromStr(this.options.content);
                        }
                    } else if (contentType === 'Array') {
                        //content是数组,转一维数组
                        this.data = axArrToFlat(this.options.content);
                        //如果id是数字，那么自动获得maxIndex
                        !isNaN(this.data[0].id) ? this.maxIndex = Math.max(...this.data.map(k => k.id)) + 1 : null;


                    } else if (contentType.includes('HTML')) {
                        //content是节点
                        this.getDataFromNode(this.options.content);
                    }

                } else if (this.inputDom.value) {
                    //从value值中获得数据
                    //input的value有值：'北京,上海,广州'
                    this.getDataFromStr(this.inputDom.value);
                }
            }
            //此处的this.data是一维数组：[{id:'',label:'',value:'',selected:'',disabled:'',tip:'',legend:'',image:''},...] 
        }
        //初始化渲染列表
        this.initRender();
        //此处的this.data会带上dom属性
        this.renderFinish();
        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    getDisplayValue(data) {
        return this.options.displayProp ? data[this.options.displayProp] : data.label
    }
    getAsyncData(callback) {
        return axAjax({
            type: this.options.ajaxType,
            url: this.options.content,
            opened: (response) => {
                this.contentXhr = response.xhr;
            },
            success: (response) => {
                if (axIsEmpty(response.content) && !Array.isArray(response.content)) {
                    console.warn('No data obtained from database!');
                    return false;
                }
                //获取成功清除tagsDom中的loading和列表
                this.tagsDom.innerHTML = '';
                this.listDom.innerHTML = '';
                callback && callback(response.content);
            }
        }, this.tagsDom);
    }
    //从ul+li和datalist取值
    getDataFromNode(elememt) {
        let elem = axIdToDom(elememt),
            listData;
        if (elem.nodeName === 'UL') {
            listData = axUlToArr(elem, 0);
        } else if (elem.nodeName === 'SELECT' || elem.nodeName === 'DATALIST') {
            listData = axSelectToArr(elem, 0);
        } else {
            return false;
        }
        this.data = axArrToFlat(listData.data);
        this.maxIndex = listData.index;
    }
    getDataFromStr(str) {
        let strArr = str.split(this.options.separator);
        this.data = strArr.map((k, i) => {
            return { id: i, label: k, value: k }
        });
        this.maxIndex = strArr.length + 1;
    }
    //设置节点属性
    setAttribute() {
        //设置className
        this.options.className ? axClass.add(this.selectDom, this.options.className) : null;
        //添加单选多选模式属性
        this.selectDom.setAttribute('type', this.inputType);
        //设置宽度
        if (this.options.width) {
            this.selectDom.style.width = this.options.width;
        }
        //popup宽度随selectDom
        !axIsMobi() && this.selectDom.getBoundingClientRect().width ? this.options.popup.width = this.selectDom.getBoundingClientRect().width + 'px' : null;

        //设置尺寸
        this.selectDom.setAttribute('size', this.options.size);
        //禁用，使之只读
        if (this.options.readonly) {
            this.readonly();
        }

    }

    //初始化数据并标准化
    initDataStd() {
        //追加selected项
        let selectedType = axType(this.options.selected),
            disabledType = axType(this.options.disabled),
            handleFun = (obj, type) => {
                let objType = axType(obj), item;
                if (objType === 'Number' || objType === 'Object') {
                    item = axFindItem(obj, this.data);
                } else if (objType === 'String') {
                    //先从id和label获取，如果找不到则从value获取
                    item = axFindItem(obj, this.data) || axFindItem(obj, this.data, { string: 'value' });
                }
                //设置属性
                if (item) {
                    item[type] = true;
                    item.dom.setAttribute(type, '');
                }
            };
        if (selectedType === 'Array') {
            for (let k of this.options.selected) {
                handleFun(k, 'selected');
            }
        } else if (selectedType === 'String') {
            this.options.selected.split(',').forEach(k => {
                handleFun(k, 'selected');
            });
        } else {
            handleFun(this.options.selected, 'selected');
        }
        //追加disabled项
        if (disabledType === 'Array') {
            for (let k of this.options.disabled) {
                handleFun(k, 'disabled');
            }
        } else if (disabledType === 'String') {
            //可能是带英文逗号的字符串
            this.options.disabled.split(',').forEach(k => {
                handleFun(k, 'disabled');
            });
        } else {
            handleFun(this.options.disabled, 'disabled');
        }
    }
    //创建骨架节点
    createSkeleton() {
        //输入输出控件，会有init初始化操作，避免创建新的selectDom节点

        /* !this.selectDom ? this.selectDom = axAddElem('div', { class: 'ax-select' }) : null;
        this.inputDom.style.display = 'none';
        !this.selectDom.contains(this.inputDom) ? this.inputDom.insertAdjacentElement('beforebegin', this.selectDom) : null;
        this.tagsDom = axAddElem('div', { class: 'ax-tags' });
        this.clearDom = axAddElem('i', { clear: '' });
        
        this.selectDom.appendChild(this.inputDom);
        this.selectDom.appendChild(this.tagsDom);
        this.options.clearShow ? this.selectDom.appendChild(this.clearDom) : null;
        //没有下拉内容，
        if (this.options.dropdownShow) {
            this.arrowDom = axAddElem('i', { arrow: '' });
            this.selectDom.appendChild(this.arrowDom);
            //下拉菜单主体框架
            this.wrapperDom = axAddElem('div', { class: 'ax-select-wrapper' });
            this.listDom = axAddElem('ul');
            this.wrapperDom.appendChild(this.listDom);
        } */

        if (!this.selectDom) {
            //首次初始化
            this.selectDom = axAddElem('div', { class: 'ax-select' });
            this.inputDom.style.display = 'none';
            this.inputDom.insertAdjacentElement('beforebegin', this.selectDom);
            this.selectDom.appendChild(this.inputDom);

            this.tagsDom = axAddElem('div', { class: 'ax-tags' });
            this.selectDom.appendChild(this.tagsDom);


            //下拉菜单主节点
            if (this.options.dropdownShow) {
                this.arrowDom = axAddElem('i', { arrow: '' });
                this.selectDom.appendChild(this.arrowDom);
                //下拉菜单主体框架
                this.wrapperDom = axAddElem('div', { class: 'ax-select-wrapper' });
                this.listDom = axAddElem('ul');
                this.wrapperDom.appendChild(this.listDom);
            }
        }
        //清除按钮，根据参数显示或不显示
        if (!this.selectDom.contains(this.clearDom)) {
            this.clearDom = axAddElem('i', { clear: '' });
            this.options.clearShow ? this.arrowDom.insertAdjacentElement('beforebegin', this.clearDom) : null;
        }
        //设置属性
        this.setAttribute();
    }
    //根据标准数据初步渲染列表
    initRender() {
        //补充content数据
        //将content彻底转为tree格式
        this.data.forEach(k => {
            this.fillItem(k, this.listDom);
        });
    }
    fillItem(obj, parent, placement = 'end') {
        obj.dom = axAddElem('li', '', `${obj.legend ? '<i class="ax-iconfont ' + obj.legend + '" legend></i>' : ''}${obj.image ? '<i image><img src="' + obj.image + '"/></i>' : ''}<i label>${obj.label}</i>${obj.tip ? '<i tip>' + obj.tip + '</i>' : ''}`);

        if (obj.children) {
            obj.dom.setAttribute('parent', '');
        } else {
            //从sql过来的数据selected和disabled肯能是'0'
            ~~obj.selected ? obj.dom.setAttribute('selected', '') : null;
            ~~obj.disabled ? obj.dom.setAttribute('disabled', '') : null;
        }
        //如果dropdownShow为false，那么parent是不存在的
        if (parent) {
            placement !== 'end' ? parent.insertAdjacentElement('afterbegin', obj.dom) : parent.appendChild(obj.dom);
        }
    }
    reRenderList(data) {
        //根据selected和disabled参数修正数据
        this.initDataStd();
        //先清空列表
        this.listDom.innerHTML = '';
        if (data[0].hasOwnProperty('weight')) {
            //有weight属性说明是搜索结果
            data.forEach(k => {
                //高亮关键字
                let pattern = new RegExp(k.keys.join('|'), 'g'),
                    labelDom = k.source.dom.querySelector('[label]');
                //先去掉s标签复原在重设s标签
                labelDom.innerHTML = labelDom.innerHTML.replace(/<[^>]+>/g, '').replace(pattern, match => `<s>${match}</s>`);
                this.listDom.appendChild(k.source.dom);
            });
        } else {
            //无weight属性说明来自content的数据
            data.forEach(k => {
                k.dom.querySelector('[label]').innerHTML = k.dom.querySelector('[label]').innerHTML.replace(/<[^>]+>/g, '');
                this.listDom.appendChild(k.dom);
            });
        }
    }
    selectedAttrToggle(obj, state) {
        if (state) {
            //设置为selected状态
            if (!~~obj.selected) {
                //select多选限制数量
                if (this.inputType !== 'single-multiple' && this.options.most !== 0) {
                    if (this.data.filter(k => ~~k.selected).length >= this.options.most) {
                        new axMessage({
                            content: axTplEngine(this.options.language.most, { value: this.options.most }),
                            result: 'error',
                            iconShow: true,
                        }).show();
                        //监听回调
                        this.handlers.hasOwnProperty('most') ? this.emit('most', this.options.most) : null;
                        this.options.onMost && this.options.onMost.call(this, this.options.most);
                        //拦截，不再执行
                        return false;
                    }
                }
                Reflect.set(obj, 'selected', true);
                //下拉列表节点
                obj.dom.setAttribute('selected', '');
                //数据源节点
                if (this.inputType.includes('select')) {
                    //数据源的node可能不存在
                    obj.node ? obj.node.setAttribute('selected', '') : null;
                }
            }
        } else {
            if (~~obj.selected) {
                Reflect.deleteProperty(obj, 'selected');
                //下拉列表节点
                obj.dom.removeAttribute('selected');
                //数据源节点
                this.inputType.includes('select') && obj.node ? obj.node.removeAttribute('selected') : null;
            }
        }
        //监听回调
        'toggleSelect' in this.handlers ? this.emit('toggleSelect', obj) : null;
        this.options.onToggleSelect && this.options.onToggleSelect.call(this, obj);
    }
    toggleList(item, list) {
        this.selectedAttrToggle(item, true);
        let selectedItem = list.find(k => k != item && ~~k.selected);
        if (selectedItem) {
            this.selectedAttrToggle(selectedItem, false);
        }
    }
    /*     activeAttrToggle(obj) {
            //当前是active则不再执行
            if (!obj.active) {
                //其他的active项取消
                let oldActive = this.data.find(k => k.active);console.log(oldActive)
                if (oldActive) {
                    Reflect.deleteProperty(oldActive, 'active');
                    oldActive.dom.removeAttribute('active');
                }
                //自身变为active
                obj.active = true;
                obj.dom.setAttribute('active', ''); 
            }
        } */
    keyToggleActive(direction, item, index, arr) {
        if (this.popupIns.state === 'hidden') {
            this.popupIns.show();
            if (!item) {
                arr[0].active = true;
                arr[0].dom.setAttribute('active', '');
            }
        } else if (this.popupIns.state === 'shown') {
            let nextItem, offset = 0;
            if (direction === 'up') {
                if (index > 0) {
                    if (item) {
                        //移除自身的属性
                        Reflect.deleteProperty(item, 'active');
                        item.dom.removeAttribute('active');
                    }
                    nextItem = arr[index - 1];
                    //下一个添加active属性
                    nextItem.active = true;
                    nextItem.dom.setAttribute('active', '');
                    index--;
                    //获得滚动偏移量
                    if (nextItem.dom.offsetTop > this.wrapperDom.clientHeight) {
                        offset = nextItem.dom.offsetTop + nextItem.dom.getBoundingClientRect().height - this.wrapperDom.clientHeight;
                    }
                    //滚动条自动滚动
                    axScrollTo(offset, { parent: this.wrapperDom, duration: 200 });
                }
            } else if (direction === 'down') {
                if (index + 1 < arr.length) {
                    if (item) {
                        //移除自身的属性
                        Reflect.deleteProperty(item, 'active');
                        item.dom.removeAttribute('active');
                    }
                    nextItem = arr[index + 1];
                    //下一个添加active属性
                    nextItem.active = true;
                    nextItem.dom.setAttribute('active', '');
                    index++;
                    //获得滚动偏移量
                    if (!axChildVisible(nextItem.dom, this.wrapperDom)) {
                        offset = (nextItem.dom.offsetTop + nextItem.dom.getBoundingClientRect().height) - this.wrapperDom.clientHeight;
                    }
                    //滚动条自动滚动
                    axScrollTo(offset, { parent: this.wrapperDom, duration: 200 });
                }
            }
        }
    }
    //渲染完成之后执行某些事件
    renderFinish() {
        //保存原始数据
        this.rawData = [...this.data];
        let inputTip = '';
        if (this.options.most > 0 && this.options.least > 0) {
            inputTip = `最多选择${this.options.most}项，至少选择${this.options.least}项`;
        } else if (this.options.most > 0) {
            inputTip = `最多选择${this.options.most}项`;
        } else if (this.options.most > 0) {
            inputTip = `至少选择${this.options.most}项`;
        } else {
            inputTip = this.options.language.inputPlaceholder;
        }
        //tags实例
        this.tagsIns = new axTags(this.tagsDom, {
            content: this.data.filter(k => ~~k.selected).map(k => { return { id: k.id, value: this.getDisplayValue(k) } }),
            removable: this.inputType === 'select-single' ? false : this.options.removable,
            compact: this.inputType === 'select-single' ? false : true,
            text: this.inputType === 'select-single' ? true : false,
            theme: this.inputType === 'select-single' ? 'auto' : 'light',
            emptyShow: this.inputType.includes('select') || (this.inputType === 'input' && !this.options.search.enable) ? true : false,
            emptyTip: `<i class="ax-color-ignore">${inputTip}</i>`,
            edit: {
                enable: this.inputType.includes('select') ? false : this.options.search.enable,
                addable: this.options.search.editable,
                deletable: this.options.search.editable,
                placeholder: this.options.language.editPlaceholder,
            },
        });
        //定义searchDom，select下的searchDom在下拉菜单里，input的searchDom即为tags的input
        if (this.inputType.includes('select')) {
            this.searchDom = axAddElem('input', { type: 'text', placeholder: this.options.language.searchPlaceholder });
        } else if (this.inputType === 'input') {
            this.searchDom = this.tagsIns.inputDom;
        }

        //显示下拉菜单才执行
        if (this.options.dropdownShow) {
            //追加searchDom
            this.options.search.enable && this.inputType.includes('select') ? this.wrapperDom.insertAdjacentElement('afterbegin', this.searchDom) : null;

            //popup实例
            this.popupIns = new axPopup(this.selectDom, axExtend({
                content: this.wrapperDom,
                trigger: this.options.trigger,
                footerShow: false,
                close: false,
                padding: false,
                triggerStrict: true,
                arrow: false,
                theme: 'select',
                placement: 'bottom-start',
            }, this.options.popup));
            //监听气泡
            this.popupIns.on('show', () => {
                //打开下拉菜单让search成为焦点
                this.searchDom && this.inputType === 'input' ? this.searchDom.focus() : null;
            });
            this.popupIns.on('shown', () => {
                //打开下拉菜单让search成为焦点
                this.searchDom && this.inputType.includes('select') ? this.searchDom.focus() : null;
            });
            this.popupIns.on('hide', () => {
                //取消onkeydown
                this.keydownSelect();
            });
            this.popupIns.on('hidden', () => {

            });
            //点击之后可触发方向选择
            this.selectDom.onclick = () => {
                //使用方向键和回车进行选择
                this.keydownSelect(this.searchResult.length > 0 ? this.searchResult.map(k => k.source) : this.data);
            }

            //清除tag
            if (this.clearDom) {
                this.clearDom.onclick = () => {
                    this.clear();
                }
            }
            //input聚焦后打开下拉菜单
            if (this.inputType === 'input' && this.searchDom) {
                this.searchDom.onclick = () => {
                    this.popupIns.state === 'hidden' ? this.popupIns.show() : null;
                }
            }
            //点击下拉菜单更新tags
            for (let k of this.data) {
                if (~~k.disabled || k.children) {
                    //禁止项和父项不能点击，所以跳过
                    continue;
                } else {
                    k.dom.onclick = () => {
                        this.clickSelect(k);
                        //最少选择数量判断
                        if (this.inputType === 'select-multiple' && this.options.least !== 0) {
                            if (this.data.filter(k => ~~k.selected).length < this.options.least) {
                                new axMessage({
                                    content: `至少选择${this.options.least}项!`,
                                    result: 'warning',
                                    iconShow: true,
                                }).show();
                                //监听回调
                                this.handlers.hasOwnProperty('least') ? this.emit('least', this.options.least) : null;
                                this.options.onLeast && this.options.onLeast.call(this, this.options.least);
                            }
                        }
                    }
                }
            }
            //监听tag更新下拉列表的selected状态和原数据的selected状态，并更新输出值
            this.tagsIns.on('remove', (tags) => {
                //删除一项tag则找到对应content项设为unselected
                tags.forEach(k => {
                    let item = axFindItem({ id: k.id, value: k.value }, this.data);
                    item ? this.selectedAttrToggle(item, false) : null;
                });
            });
            //模糊搜索
            if (this.searchDom) {
                this.searchDom.oninput = axDebounce(() => {
                    let value = this.searchDom.value;
                    if (!value) {
                        this.reRenderList(this.data);
                        //恢复onkeydown
                        this.keydownSelect(this.data);
                    } else {
                        this.searchResult = axArrSearch(value, 'label', this.data, { fuzzy: this.options.search.fuzzy, reorder: this.options.search.reorder, limit: this.options.search.limit });
                        if (this.searchResult.length === 0) {
                            if (this.inputType === 'input') {
                                //重新添加回车和删除功能
                                this.options.enterSelect ? this.searchDom.addEventListener('keyup', this.tagsIns.inputEdit) : null;
                                this.listDom.innerHTML = `<li message><i label>${axTplEngine(this.options.language.emptyAddable, { value: value })}</i></li>`;
                            } else {
                                this.listDom.innerHTML = `<li message><i label>${axTplEngine(this.options.language.empty, { value: value })}</i></li>`;
                            }
                            //取消onkeydown
                            this.keydownSelect();
                        } else {
                            //有检索结果，因为可使用enter从list中添加项目，也能从input创建新项目，此时需要禁用input回车创建功能
                            this.inputType === 'input' && this.options.enterSelect ? this.searchDom.removeEventListener('keyup', this.tagsIns.inputEdit) : null;
                            this.reRenderList(this.searchResult);
                            //恢复onkeydown
                            this.keydownSelect(this.searchResult.map(k => k.source));
                        }
                    }
                })

            }


            //监听tag的数据变化
            this.tagsIns.contentProxy.instance.on('complete', (data) => {
                this.selected = data.target.map(k => {
                    return { id: k.id, value: k.value }
                });
                this.value = data.target.map(k => k.value).join(this.options.separator);
                //将值更新到input
                this.inputType === 'input' ? this.inputDom.value = this.value : null
                //因为会出现两行或多行情况导致定位不准，所以需要重新定位
                this.inputType !== 'select-single' && this.popupIns ? this.popupIns.popupPosition.update() : null;
                //保存数据
                this.save();
            });

            /* 
            this.searchDom.addEventListener('input', axDebounce(() => {
                let value = this.searchDom.value;
                if (!value) {
                    this.reRenderList(this.data);
                    //恢复onkeydown
                    this.keydownSelect(this.data);
                } else {
                    this.searchResult = axArrSearch(value, 'label', this.data, { fuzzy: this.options.search.fuzzy, reorder: this.options.reorder, limit: this.options.search.limit });
                    if (this.searchResult.length === 0) {
                        if (this.inputType === 'input') {
                            //重新添加回车和删除功能
                            this.options.enterSelect ? this.searchDom.addEventListener('keyup', this.tagsIns.inputEdit) : null;
                            this.listDom.innerHTML = `<li message><i label>${axTplEngine(this.options.language.emptyAddable, { value: value })}</i></li>`;
                        } else {
                            this.listDom.innerHTML = `<li message><i label>${axTplEngine(this.options.language.empty, { value: value })}</i></li>`;
                        }
                        //取消onkeydown
                        this.keydownSelect();
                    } else {
                        //有检索结果，因为可使用enter从list中添加项目，也能从input创建新项目，此时需要禁用input回车创建功能
                        this.inputType === 'input' && this.options.enterSelect ? this.searchDom.removeEventListener('keyup', this.tagsIns.inputEdit) : null;
                        this.reRenderList(this.searchResult);
                        //恢复onkeydown
                        this.keydownSelect(this.searchResult.map(k => k.source));
                    }
                }
            })
            ); */
        } else {
            //没有下拉菜单，无法通过检测popup的show，所以需要手动设置焦点状态
            this.selectDom.onclick = () => {
                if (!this.selectDom.classList.contains('ax-opened')) {
                    this.selectDom.classList.add('ax-opened');
                }
                this.searchDom && this.inputType === 'input' ? this.searchDom.focus() : null;
                //绑定一个事件，区域外点击可取消focus状态
                document.addEventListener('click', this.bodyClickBlur, false);
            }
        }

    }
    select(data, callback) {
        //如果已经销毁则返回实例
        if (axIsEmpty(data) || this.destroyed) {
            return this;
        }

        let dataType = axType(data),
            oldItems = [],
            absentItems = [],
            dataFun = (obj) => {
                let item = axFindItem(obj, this.data);
                //以上找出来了在this.data的对应项，以下则处理对应项
                if (item) {
                    //追加
                    oldItems.push(item);
                    let tagObj = { id: item.id, value: this.getDisplayValue(item) };
                    if (!~~item.selected) {
                        //已经是selected状态则不再执行
                        if (this.inputType === 'select-single') {
                            this.toggleList(item, this.data);
                            //更新tags
                            this.tagsIns.clear();
                            this.tagsIns.add(tagObj);
                        } else {
                            this.selectedAttrToggle(item, !~~item.selected);
                            //添加tag
                            this.tagsIns.add(tagObj);
                        }
                    }
                } else {
                    absentItems.push(obj);


                }
            };
        //循环
        if (dataType === 'Array') {
            data.forEach(k => {
                dataFun(k);
            });
        } else if (dataType === 'String') {
            data.split(',').forEach(k => {
                dataFun(k);
            });
        } else {
            dataFun(data);
        }
        //添加不存在的
        if (absentItems.length > 0) {
            let absentStr = absentItems.map(k => k.label || k).join(',');
            if (this.options.absentAddable) {
                new axDialog(axExtend({
                    content: axTplEngine(this.options.language.absentAddable, { value: absentStr }),
                    confirm: {
                        callback: () => {
                            //追加新项并添加到tags
                            absentItems.forEach(k => {
                                this.add(k, 'end', (data) => {
                                    data.forEach(k => {
                                        //设置selected状态
                                        k.dom.setAttribute('selected', '');
                                        k.selected = true;
                                        //新增tags
                                        this.tagsIns.add({ id: k.id, value: this.getDisplayValue(k) });
                                        //追加
                                        oldItems.push(...data);
                                    });
                                });
                            });
                            //监听回调
                            this.handlers.hasOwnProperty('absentSelect') ? this.emit('absentSelect', oldItems) : null;
                            this.options.onAbsentSelect && this.options.onAbsentSelect.call(this, oldItems);
                        }
                    }
                }, this.options.dialog)).show();
            } else {
                new axMessage({
                    content: axTplEngine(this.options.language.absent, { value: absentStr }),
                    iconShow: true,
                }).show();
            }
        }
        //监听回调
        this.handlers.hasOwnProperty('select') ? this.emit('select', oldItems) : null;
        this.options.onSave && this.options.onSave.call(this, oldItems);
        callback && callback.call(this, oldItems);
        return this;
    }
    unselect(data, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        let dataType = axType(data),
            oldItems = [],
            dataFun = (obj) => {
                let item = axFindItem(obj, this.data);
                //以上找出来了在this.data的对应项，以下则处理对应项
                if (item) {
                    //追加
                    oldItems.push(item);
                    let tagObj = { id: item.id, value: item.value };
                    //已经是unselected状态则不再执行
                    if (~~item.selected) {
                        this.selectedAttrToggle(item, false);
                        //删除tag
                        this.tagsIns.remove(tagObj);
                    }
                } else {
                    //删除tag
                    this.tagsIns.remove(obj);
                }
            };
        //循环
        if (dataType === 'Array') {
            data.forEach(k => {
                dataFun(k);
            });
        } else if (dataType === 'String') {
            data.split(',').forEach(k => {
                dataFun(k);
            });
        } else {
            dataFun(data);
        }

        //监听回调
        this.handlers.hasOwnProperty('unselect') ? this.emit('unselect', oldItems) : null;
        this.options.onSave && this.options.onSave.call(this, oldItems);
        callback && callback.call(this, oldItems);
        return this;
    }
    arrAdd(data, placement) {
        data.forEach(k => {
            //追加节点到列表中
            this.fillItem(k, this.listDom, placement);
            //追加数据
            placement !== 'end' ? this.data.unshift(k) : this.data.push(k);
            //绑定事件
            k.dom.onclick = () => {
                this.clickSelect(k);
            }
        });
    }
    add(data, placement = 'end', callback) {
        //如果已经销毁则返回实例
        if (axIsEmpty(data) || this.destroyed) {
            return this;
        }

        let dataType = axType(data),
            newItems = [],
            dataFun = (obj) => {
                let objType = axType(obj);

                if (objType === 'String') {
                    //是字符串
                    let strArr = obj.split(this.options.separator),
                        msgIns = new axMessage({ content: '', result: 'error', iconShow: true });
                    for (let k of strArr) {
                        if (this.data.find(i => i.label === k)) {
                            //已有了则不再添加了
                            msgIns.update({ content: axTplEngine(this.options.language.addFail, { value: k }) }).show();
                            continue;
                        } else {
                            newItems.push({ id: this.maxIndex, label: k, value: k })
                            //索引递增
                            this.maxIndex++;
                        }
                    }
                } else if (objType === 'Object') {
                    if (axFindItem(obj, this.data)) {
                        //已有了则不再添加了
                        msgIns.update({ content: axTplEngine(this.options.language.addFail, { value: obj.label }) }).show();
                    } else {
                        //没有id则给加上id
                        if (!obj.hasOwnProperty('id')) {
                            obj.id = this.maxIndex;
                            this.maxIndex++;
                        }
                        newItems.push(obj);
                    }
                }
            }
        //获得新增加的对象
        if (dataType === 'Array') {
            data.forEach(k => {
                dataFun(k);
            });
        } else if (dataType === 'String') {
            dataFun(data);
        }

        //根据新对象追加节点和数据
        if (this.options.onBeforeAdd) {
            //手动添加
            this.options.onBeforeAdd.call(newItems);
        } else {
            this.arrAdd(newItems, placement);
        }
        //监听回调
        this.handlers.hasOwnProperty('add') ? this.emit('add', newItems) : null;
        this.options.onAdd && this.options.onAdd.call(this, newItems);
        callback && callback.call(this, newItems);
        //保存数据
        this.save();
        return this;
    }
    arrRemove(data) {
        if (data.length > 0) {
            data.forEach(k => {
                //删除节点
                k.dom.remove();
            });
            //从数据中清除自己
            this.data = this.data.filter(k => !data.includes(k));
        }
    }
    remove(data, callback) {
        //如果已经销毁则返回实例
        if (axIsEmpty(data) || this.destroyed) {
            return this;
        }
        let dataType = axType(data),
            oldItems = [],
            dataFun = (obj) => {
                let item = axFindItem(obj, this.data);
                item ? oldItems.push(item) : null;
            }
        //获得新增加的对象
        if (dataType === 'Array') {
            data.forEach(k => {
                dataFun(k);
            });
        } else if (dataType === 'String') {
            data.split(',').forEach(k => {
                dataFun(k);
            });
        } else {
            dataFun(data);
        }

        if (this.options.onBeforeRemove) {
            //手动删除
            this.options.onBeforeRemove.call(oldItems);
        } else {
            //根据新对象追加节点和数据
            this.arrRemove(oldItems);
        }


        //监听回调
        this.handlers.hasOwnProperty('remove') ? this.emit('remove', oldItems) : null;
        this.options.onRemove && this.options.onRemove.call(this, oldItems);
        callback && callback.call(this, oldItems);
        //保存数据
        this.save();
        return this;
    }
    clear(callback) {
        //清空tags
        this.tagsIns.clear();
        //更新数据中的selected状态
        for (let k of this.data) {
            if (~~k.selected) {
                Reflect.deleteProperty(k, 'selected');
                k.dom.removeAttribute('selected');
                k.node ? k.node.removeAttribute('selected') : null;
            }
        }
        this.handlers.hasOwnProperty('clear') ? this.emit('clear', '') : null;
        this.options.onClear && this.options.onClear.call(this);
        callback && callback.call(this);
        return this;
    }
    readonly(callback) {
        this.inputDom.setAttribute('readonly', '');
        this.selectDom.setAttribute('readonly', '');
        callback && callback.call(this);
        return this;
    }
    readable(callback) {
        this.inputDom.removeAttribute('readonly');
        this.selectDom.removeAttribute('readonly');
        callback && callback.call(this);
        return this;
    }
    disable(data, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        let dataType = axType(data),
            oldItems = [],
            dataFun = (obj) => {
                let item = axFindItem(obj, this.data);
                item ? oldItems.push(item) : null;
            }
        //获得新增加的对象
        if (dataType === 'Array') {
            data.forEach(k => {
                dataFun(k);
            });
        } else if (dataType === 'String') {
            data.split(',').forEach(k => {
                dataFun(k);
            });
        } else {
            dataFun(data);
        }
        //根据新对象追加节点和数据
        oldItems.forEach(k => {
            //设置disabled状态
            if (!~~k.disabled) {
                k.dom.setAttribute('disabled', '');
                k.disabled = true;
            }
        });
        //监听回调
        this.handlers.hasOwnProperty('disabled') ? this.emit('disabled', oldItems) : null;
        this.options.onDisabled && this.options.onDisabled.call(this, oldItems);
        callback && callback.call(this, oldItems);
        //保存数据
        this.save();
        return this;
    }
    enable(data, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        let dataType = axType(data),
            oldItems = [],
            dataFun = (obj) => {
                let item = axFindItem(obj, this.data);
                item ? oldItems.push(item) : null;
            }
        //获得新增加的对象
        if (dataType === 'Array') {
            data.forEach(k => {
                dataFun(k);
            });
        } else if (dataType === 'String') {
            data.split(',').forEach(k => {
                dataFun(k);
            });
        } else {
            dataFun(data);
        }
        //根据新对象追加节点和数据
        oldItems.forEach(k => {
            //设置disabled状态
            if (~~k.disabled) {
                k.dom.removeAttribute('disabled');
                Reflect.deleteProperty(k, 'disabled');
            }
        });
        //监听回调
        this.handlers.hasOwnProperty('enabled') ? this.emit('enabled', oldItems) : null;
        this.options.onEnabled && this.options.onEnabled.call(this, oldItems);
        callback && callback.call(this, oldItems);
        //保存数据
        this.save();
        return this;
    }
    //保存当前数据
    save(props, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }

        if (!props) {
            this.options.content = this.data;
            axLocalStorage.set(this.options.storageName, this.options);
        } else {
            axLocalStorage.set(this.options.storageName, props);
        }
        //回调监听
        let localValue = axLocalStorage.get(this.options.storageName);
        this.handlers.hasOwnProperty('save') ? this.emit('save', localValue) : null;
        this.options.onSave && this.options.onSave.call(this, localValue);
        callback && callback.call(this, localValue);

        return this;
    }
    //更新实例参数
    async update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //更新参数
        this.options = axExtend(this.options, setting);
        //保存数据
        this.save();
        //根据参数初始化实例
        await this.init();
        //监听回调
        this.handlers.hasOwnProperty('update') ? this.emit('update', '') : null;
        this.options.onUpdate && this.options.onUpdate.call(this);
        callback && callback.call(this);

        return this;
    }
    //销毁实例
    destroy(callback) {
        //取消绑定事件
        this.selectDom.onclick = null;
        this.clearDom ? this.clearDom.onclick = null : null;
        this.searchDom ? this.searchDom.onclick = null : null;
        this.searchDom ? this.searchDom.removeEventListener('keyup', this.tagsIns.inputEdit) : null;
        document.removeEventListener('click', this.bodyClickBlur);
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        //销毁内部实例
        this.popupIns ? this.popupIns.destroy() : null;
        this.tagsIns.destroy();
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    //以下三个方法为监听
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/* //对有本属性的元素使用本插件
document.querySelectorAll('[axSelect]').forEach(element => {
    new axSelect(element);
}); */
/*!
 * 插件：查看更多；使用方法：new axDropdown('#id',{参数})
 */
class axDropdown {
    constructor(elem, options) {
        //列表Dom
        this.targetDom = axIdToDom(elem);
        this.options = axExtend({
            type: 'list',//气泡内容类型，可选择checkbox、radio、botton、list、text、textarea、html
            content: '',//气泡内容，可以是字符串，#id、className、节点或数组
            node: 'a',//列表项目节点名，默认是a标签，可选择div等
            nowrap: true,//超出是否省略显示（一行显示不完以...结尾），默认true省略，可选择false多行显示
            breakShow: true,//是否显示分割线，默认true显示，可选择false不显示
            arrowShow: true,//是否给按钮右方显示指示小箭头，默认为true显示，可选择false不显示
            grid: {
                enable: false,//是否将展示方式设为grid布局，默认false不使用grid，可选择true开启
                columns: 0,//设置一行几列，默认为0即不设置，如设为2表示一行两列
                fenceShow: false,//是否启用项目之间的1px边距，默认false不启用，可选择true启用
            },//设置grid布局，对使用content参数类型有效
            onInit: '',//回调函数，初始化后执行，无参数
            onBeforeSet: '',//回调函数，赋值给input之前执行（target是一个input），支持参数value；可赋值类型中有效
            onAfterSet: '',//回调函数，赋值给input之后执行（target是一个input），支持参数value；可赋值类型中有效
            onConfirm: '',//回调函数，点击确认按钮后的回调，支持参数value；有窗口底部按钮类型有效
            onUpdate: '',//回调函数，更新参数后的回调，无参数
            onUpdateContent: '',//回调函数，更新内容后的回调，无参数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            plugin: {
                value: '',//插件初始值
            },//使用子插件自身的参数，包括axCheckbox、axInput、axTree等
            popup: {
                in: 'slideDown',//默认下拉动画显示
                out: 'slideUp',//默认上收动画关闭
                size: 'sm',//默认小尺寸气泡
                placement: 'bottom-start',//默认气泡在目标下方居左
                close: false,//默认不使用气泡右上角关闭按钮
            },//popup插件的参数
        }, options, this.targetDom, this.constructor.name);

        //回调监听
        this.handlers = {};

        //初始化
        this.init();
        return this;
    }
    init() {
        axInstance.push(this, this.options.insName, 'dropdown');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //插件实例
        this.plugin = null;
        this.popup = null;
        //字符串值
        this.value = '';
        //数组值
        this.valueArr = [];

        //定义列表变量
        this.list = [];

        //设置初始值
        if (this.options.plugin.value) {
            //如果插件有值则赋予初始值
            this.initValue = this.options.plugin.value;
        } else {
            this.initValue = '';
        }

        //输入表单域节点特别处理
        if (this.targetDom.nodeName == 'INPUT' || this.targetDom.nodeName == 'TEXTAREA') {
            //如果是input则只读
            this.targetDom.setAttribute('readonly', '');
            //如果有值应该被视为初始值
            if (this.targetDom.value) {
                this.options.plugin.value = this.targetDom.value;
                this.initValue = this.targetDom.value;
            } else {
                if (this.initValue) {
                    this.targetDom.value = this.initValue;
                }
            }
        }


        //不同类型生成不同内容
        if (this.options.type === 'checkbox' || this.options.type === 'radio') {
            //如果填写了content，那么赋值给插件的data参数
            if (!axIsEmpty(this.options.content)) {
                this.options.plugin.children = this.options.content;
            }
            //赋值类型
            this.options.plugin.type = this.options.type;
            //都使用block展示方式
            if (this.options.grid.enable) {
                this.options.plugin.display = 'grid';
                this.options.grid.columns ? this.options.plugin.columns = this.options.grid.columns : null;
                if (this.options.grid.fenceShow) {
                    this.options.plugin.parentClass = 'ax-fence';
                    this.options.plugin.childClass = 'ax-padding-left'
                    this.options.popup.padding = false;
                }
            } else {
                this.options.plugin.display = 'block';
            }
            //创建插件实例
            this.plugin = new axCheckbox('', this.options.plugin);
            //创建窗口实例
            this.popup = new axPopup(this.targetDom, Object.assign(this.options.popup, {
                content: this.plugin.parent,
                cancel: '',
                diy: {
                    text: '重置',
                    callback: () => {
                        this.plugin.reset();
                    }
                },
                confirm: {
                    type: 'prevent',
                    callback: () => {
                        //是数组
                        this.value = this.plugin.getValues('string');
                        this.valueArr = this.plugin.getValues('array');
                        //如果目标时input则点击赋值
                        if (this.options.onBeforeSet) {
                            //有则执行
                            this.options.onBeforeSet.call(this, this.value);
                        } else {
                            //没有则关闭窗口
                            this.setToHide(this.value);
                        }
                        //回调
                        this.options.onConfirm && this.options.onConfirm.call(this, this.value);
                        'confirm' in this.handlers ? this.emit('confirm', this.value) : null;
                    }
                }
            }));
            this.popup.on('show', () => {
                //如果目标是input则从中取值，如果input值与插件值不同则以input为准
                if ((this.targetDom.nodeName === 'INPUT' || this.targetDom.nodeName === 'TEXTAREA') && this.targetDom.value) {
                    if (this.targetDom.value != this.plugin.getValues('string')) {
                        this.updateValues(this.targetDom.value);
                    }
                }
            });
        } else if (this.options.type === 'list') {
            //列表
            this.popup = new axPopup(this.targetDom, Object.assign(this.options.popup, { footerShow: false, padding: false }));

            let listDom = null;
            if (axType(this.options.content) === 'String') {
                //content=''
                if (axStrType(this.options.content)) {
                    //id、className、name或者nodeName
                    listDom = axIdToDom(this.options.content);
                    //先清空再追加节点
                    this.popup.content.innerHTML = axIdToDom(this.options.content).innerHTML;
                } else {
                    this.popup.content.innerHTML = this.options.content;
                }
            } else if (axType(this.options.content).includes('HTML')) {
                //content=物理节点
                this.popup.content.innerHTML = this.options.content.innerHTML;
            } else if (axType(this.options.content) === 'Array') {
                //content=[{label:'',target:'',url:'',prefix:{type:'',data:''},suffix:'',props:{},callback:''}]
                listDom = axAddElem('div', { class: 'ax-list-wrapper' });
                //循环生成子节点
                this.renderList(this.options.content, listDom);
                //先清空再追加节点
                this.popup.content.innerHTML = '';
                this.popup.content.appendChild(listDom);
            }

        } else if (this.options.type === 'button') {
            //data=[{label:'',value:'',callback:''}]
            this.popup = new axPopup(this.targetDom, Object.assign(this.options.popup, { footerShow: false, padding: false }));
            let listDom = axAddElem('div', { class: 'ax-list-wrapper' });
            //循环生成子节点
            this.renderList(this.options.content, listDom);
            //点击操作

            this.list.forEach(k => {
                if (!k.props.disabled) {
                    k.dom.onclick = (e) => {
                        axPreventDefault(e);
                        //修改value
                        this.value = k.props.value;
                        this.valueArr = [k.props.value];
                        //追加样式
                        k.dom.setAttribute('checked','');
                        this.list.filter(i => i.dom != k.dom).forEach(x => {
                            x.dom.removeAttribute('checked');
                        });

                        //如果目标时input则点击赋值
                        if (this.options.onBeforeSet) {
                            //有则执行
                            this.options.onBeforeSet.call(this, k.props.value)
                        } else {
                            //没有则关闭窗口
                            this.setToHide(k.props.value);
                        }
                    }
                }
            });
            //先清空再追加节点
            this.popup.content.innerHTML = '';
            this.popup.content.appendChild(listDom);
            //初始化
            this.initValue ? this.updateValues(this.initValue) : null;
            this.popup.on('show', () => {
                //如果目标是input则从中取值，如果input值与插件值不同则以input为准
                if ((this.targetDom.nodeName === 'INPUT' || this.targetDom.nodeName === 'TEXTAREA') && this.targetDom.value) {
                    this.updateValues(this.targetDom.value);
                }
            });
        } else if (this.options.type === 'text' || this.options.type === 'textarea') {
            //多行文本
            //赋值类型
            this.options.plugin.type = this.options.type;
            //如果填写了content，那么赋值给value
            this.options.content ? this.options.plugin.value = this.options.content : null;
            //创建插件实例
            this.plugin = new axInput('', this.options.plugin);
            //创建窗口实例
            this.popup = new axPopup(this.targetDom, Object.assign(this.options.popup, {
                content: this.plugin.parent,
                cancel: '',
                diy: {
                    text: '重置',
                    callback: () => {
                        this.plugin.set(this.initValue);
                    }
                },
                confirm: {
                    type: 'prevent',
                    callback: () => {
                        //是数组
                        this.value = this.plugin.getValues('string');
                        this.valueArr = this.plugin.getValues('array');
                        //如果目标时input则点击赋值
                        if (this.options.onBeforeSet) {
                            //有则执行
                            this.options.onBeforeSet.call(this, this.value);

                        } else {
                            //没有则关闭窗口
                            this.setToHide(this.value);
                        }
                        //回调
                        this.options.onConfirm && this.options.onConfirm.call(this, this.value);
                        'confirm' in this.handlers ? this.emit('confirm', this.value) : null;
                    }
                }
            }));
            this.popup.on('show', () => {
                //如果目标是input则从中取值，如果input值与插件值不同则以input为准
                if ((this.targetDom.nodeName === 'INPUT' || this.targetDom.nodeName === 'TEXTAREA') && this.targetDom.value) {
                    if (this.targetDom.value != this.plugin.getValues()) {
                        this.updateValues(this.targetDom.value);
                    }

                }
                //将光标置于最末，需要稍微延迟,否则focus无效
                setTimeout(() => {
                    let value = this.plugin.field.value;
                    this.plugin.field.value = '';
                    this.plugin.field.focus();
                    this.plugin.field.value = value;
                }, 100);

            });
            //input可使用回车键
            if (this.options.type === 'text') {
                this.plugin.field.onkeyup = (e) => {
                    //使用回车键
                    if (e.code === 'Enter') {
                        this.popup.options.confirm.callback();
                    }
                }
            }
        } else if (this.options.type === 'number') {
            //设置amount宽度
            this.options.plugin.width = '100%';
            //创建插件实例
            this.plugin = new axAmount('', this.options.plugin);
            //创建窗口实例
            this.popup = new axPopup(this.targetDom, Object.assign(this.options.popup, {
                content: this.plugin.parent,
                cancel: '',
                diy: {
                    text: '重置',
                    callback: () => {
                        this.plugin.reset();
                    }
                },
                confirm: {
                    type: 'prevent',
                    callback: () => {
                        this.value = this.plugin.value;
                        this.valueArr.push(this.plugin.value);
                        //如果目标时input则点击赋值
                        if (this.options.onBeforeSet) {
                            //有则执行
                            this.options.onBeforeSet.call(this, this.value);
                        } else {
                            //没有则关闭窗口
                            this.setToHide(this.value);
                        }
                        //回调
                        this.options.onConfirm && this.options.onConfirm.call(this, this.value);
                        'confirm' in this.handlers ? this.emit('confirm', this.value) : null;
                    }
                }
            }));
            this.popup.on('show', () => {
                //如果目标是input则从中取值，如果input值与插件值不同则以input为准
                if ((this.targetDom.nodeName === 'INPUT' || this.targetDom.nodeName === 'TEXTAREA') && this.targetDom.value) {
                    if (this.targetDom.value != this.plugin.value) {
                        this.updateValues(this.targetDom.value);
                    }
                }
            });
            //将光标置于最末，需要稍微延迟,否则focus无效
            setTimeout(() => {
                let value = this.plugin.field.value;
                this.plugin.field.value = '';
                this.plugin.field.focus();
                this.plugin.field.value = value;
            }, 100);
            //支持使用回车键赋值
            this.plugin.field.onkeyup = (e) => {
                //使用回车键
                if (e.code === 'Enter') {
                    this.popup.options.confirm.callback();
                }
            }
        } else if (this.options.type === 'switch') {


        } else if (this.options.type === 'html') {
            //html文本
            this.popup = new axPopup(this.targetDom, Object.assign(this.options.popup, {
                footerShow: false,
                content: this.options.content,
            }));
        }
        //target增加属性或内容
        this.setAttribute();
        //回调
        this.options.onInit && this.options.onInit.call(this);
    }
    setAttribute() {
        //显示指示箭头
        if (this.options.arrowShow) {
            if (this.targetDom && !this.targetDom.querySelector('.ax-at-right')) {
                let arrow = axAddElem('i', { class: 'ax-dropdown-arrow' });
                this.targetDom.appendChild(arrow);
            }
        }
        //popup加上dropdown样式
        this.popup.targetDom.classList.add('ax-dropdown');

    }
    setToHide(value, callback) {
        value = value || this.value;
        if (this.targetDom.nodeName === 'INPUT') {
            //赋值
            this.targetDom.value = value;
        }
        //回调监听
        this.options.onAfterSet && this.options.onAfterSet.call(this, value);
        'set' in this.handlers ? this.emit('set', value) : null;
        callback && callback.call(this, value);
        //关闭窗口
        this.popup.hide();
    }
    updateToShow(value) {
        this.popup.show();
        //此处不能使用this.popup.on监听
        setTimeout(() => {
            this.updateValues(value);
        }, 100);
    }
    updateValues(value) {
        if (['checkbox', 'radio', 'text', 'textarea', 'number'].includes(this.options.type)) {
            this.plugin.set(value);
        } else if (this.options.type === 'button') {
            //只能单选一个
            let str = ''
            if (Array.isArray(value)) {
                str = value[0];
            } else {
                str = value;
            }
            this.list.forEach(k => {
                if (k.props.value === str) {
                    k.dom.setAttribute('checked','');
                } else {
                    k.dom.removeAttribute('checked');
                }
            });
        }
    }
    getValues(type = 'string',) {
        if (type === 'string') {
            return this.value;
        } else if (type === 'array') {
            return this.valueArr;
        }
    }
    renderList(arr, parent) {
        //设定grid布局
        if (this.options.grid.enable) {
            let classNames = 'ax-grid';
            this.options.grid.columns ? classNames += ` ax-grid-${this.options.grid.columns}` : null;
            this.options.grid.fenceShow ? classNames += ' ax-fence' : null;
            axClass.add(parent, classNames);
        }

        arr.forEach((k, i) => {
            let outer = axAddElem(this.options.node, { class: 'ax-bulletin ax-hover', ...k.props }, `<div class="ax-bulletin-body"></div>`),
                inner = outer.querySelector('.ax-bulletin-body'),
                caption = null;
                //禁用
            k.disabled ? outer.setAttribute('disabled', '') : null;
            //创建项目节点
            if (this.options.node === 'a') {
                //list类型，才需要href和target属性
                if (this.options.type === 'list') {
                    outer.target = k.target ? k.target : '_blank';
                    outer.href = k.url ? k.url : '###';
                } else {
                    outer.href = '###';
                }
            }
            //创建标题
            if (this.options.nowrap) {
                if (this.options.node != 'a' && k.url) {
                    caption = axAddElem('a', { caption: '' }, k.label)
                    //list类型，才需要href和target属性
                    if (this.options.type === 'list') {
                        caption.target = k.target ? k.target : '_blank';
                        caption.href = k.url ? k.url : '###';
                    } else {
                        caption.href = '###';
                    }
                } else {
                    caption = axAddElem('i', { caption: '' }, k.label);
                }
                inner.appendChild(caption);
            } else {
                inner.textContent = k.label;
            }
            //创建图标或头像前缀
            if (!axIsEmpty(k.prefix)) {
                if (k.prefix.type === 'legend') {
                    inner.insertAdjacentElement('beforebegin', axAddElem('i', { class: 'ax-iconfont ' + k.prefix.data,legend:'' }));
                } else if (k.prefix.type === 'avatar') {
                    inner.insertAdjacentElement('beforebegin', axAddElem('i', { avatar: '' },`<img src="${k.prefix.data}" />`));
                }
                else if (k.prefix.type === 'image') {
                    inner.insertAdjacentElement('beforebegin', axAddElem('i', { image: '' }, `<img src="${k.prefix.data}"/>`));
                }
            }
            //创建后缀
            if (k.suffix) {
                inner.insertAdjacentElement('afterend', axAddElem('i', { arrow: '' }, k.suffix));
            }

            parent.appendChild(outer);
            //添加分割线，最后一项不需要分割线(grid布局不可使用分割线)
            if (!this.options.grid.enable && this.options.breakShow && i + 1 != this.options.content.length) {
                parent.insertAdjacentHTML('beforeend', '<hr/>');
            }
            //追加列表数组
            this.list.push({ props: k, dom: outer })

            if (!k.disabled) {
                //项目回调
                k.callback && k.callback(outer, k);
            }
        });
    }
    update(setting, callback) {
        this.popup.targetDom.remove();
        this.popup.destroy();
        this.options.content = axExtend(this.options.content, setting);
        this.init();
        //回调监听
        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update') : null;
        callback && callback.call(this);
        return this;
    }
    updateContent(data, callback) {
        this.popup.targetDom.remove();
        this.popup.destroy();
        this.options.content = data;
        this.init();
        //回调监听
        this.options.onUpdateContent && this.options.onUpdateContent.call(this);
        'updateContent' in this.handlers ? this.emit('updateContent') : null;
        callback && callback.call(this);
        return this;
    }
    show(callback) {
        this.popup.show();
        //回调
        callback && callback.call(this);
        return this;
    }
    hide(callback) {
        this.popup.hide();
        //回调
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/* (() => {
    //对使用axDropdown属性的元素使用本插件
    document.querySelectorAll('[axDropdown]').forEach(element => {
        new axDropdown(element);
    });
})(); */
/*!
* 插件：表单数量步进增减；使用方法：new axAmount('#id',{参数})
*/
const axAmounts = [];
class axAmount {
    constructor(elem, options) {
        //定义节点DOM
        this.targetDom = elem ? axIdToDom(elem) : '';
        //定义参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            width: '',//父节点长度，字符串类型，默认为空即152px，可设置“100%”或其他具体值
            placeholder: '输入数量',//没有初始值时候input的提示文字
            value: '',//设置默认值，数值类型，默认为空，如果为空则自动使用input的值
            min: '',//设置最小值，数值类型，默认为空（可无限小）
            max: '',//设置最大值，数值类型，默认为空（可无限大）
            step: 1,//步进，数值类型，默认是1
            precision: 0,//设置小数点精确位数，数值类型，默认是整数，可填1或2
            rule: '',//步进格则，默认为空。可填：currency|quantity|percent|month|day|hour|minute|second
            disabled: false,//设置是否禁止使用，默认为false不禁用，可选择true禁用
            increase: true,//设置是否启用增加，默认true启用，可选择false禁用
            decrease: true,//设置是否启用减少，默认true启用，可选择false禁用
            cookie: '',//设置保存cookie，默认为空不保存；如果需要刷新页面保持原数值，请填写当页唯一的cookie名称
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onMined: '',//回调函数，超过最小值后执行，无参数
            onMaxed: '',//回调函数，超过最大值执行，无参数
            onUpdated: '',//回调函数，更新参数后执行，无参数
            onChanged: '',//回调函数，值发生变化后执行，支持两个参数即旧值和新值
        }, options, this.targetDom, this.constructor.name);// 最后一个参数可省略
        //定义监听事件成员
        this.handlers = {};
        //初始值变量
        this.value = '';
        //增减按钮状态
        this.activeIn = true;
        this.activeDe = true;
        //初始化
        this.init();
    }
    init() {
         //添加到实例合集
         axInstance.push(this, this.options.insName, 'amount');
        if (this.options.min !== '' && this.options.mam !== '' && this.options.min > this.options.mam) {
            console.error('The minimum value is greater than the maximum value!');
            return false;
        }
//断点合并参数
!axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;

        //定义父节点和构成元素
        this.createSkeleton();

        //根据规则定义min、max、precision和step
        this.ruleToSet();

        //修正value
        this.initValue();

        //判断是否有cookie，再次修正value值
        this.getCookie();

        //获取初始值
        this.rawValue = this.value;

        //设置样式和属性
        this.setAttribute();

        //键盘事件
        let eventKey = (e) => {
            if (e.code === 'ArrowUp' || e.code === 'ArrowRight') {
                e.preventDefault();
                eventIncrease();
            } else if (e.code === 'ArrowLeft' || e.code === 'ArrowDown') {
                e.preventDefault();
                eventDecrease();
            }
        }

        //定义增加函数
        let eventIncrease = () => {
            //激活减少按钮
            if (!this.activeDe) {
                this.options.decrease = true;
                this.decrease.removeAttribute('disabled');
            }

            let val = Number(this.field.value),
                old = Number(this.field.value);
            if (this.options.max === '') {
                val = val + this.options.step;
            } else if (this.options.max !== '' && val < this.options.max) {
                val = val + this.options.step;
            }
            if (this.options.max !== '' && ((val + this.options.step) > this.options.max)) {
                //再步进一次大于最大值则禁止
                this.options.increae = false;
                this.increase.setAttribute('disabled', 'true');
                this.activeIn = false;
                //监听回调
                this.options.onMaxed && this.options.onMaxed.call(this);
                'maxed' in this.handlers ? this.emit('maxed', '') : null;
            }
            this.field.value = val.toFixed(this.options.precision);
            this.value = this.field.value;
            //设置cookie
            this.setCookie();
            //监听
            if (this.field.value > old) {
                //监听回调
                this.options.onChanged && this.options.onChanged.call(this, old, this.field.value);
                'changed' in this.handlers ? this.emit('changed', old, this.field.value) : null;
            }

        }
        //定义减少函数
        let eventDecrease = () => {
            //激活增加按钮
            if (!this.activeIn) {
                this.options.increae = true;
                this.increase.removeAttribute('disabled');
            }

            let val = Number(this.field.value),
                old = Number(this.field.value);
            if (this.options.min === '') {
                val = val - this.options.step;
            } else if (this.options.min !== '' && val > this.options.min) {
                val = val - this.options.step;
            }
            if (this.options.min !== '' && ((val - this.options.step) < this.options.min)) {
                //再步进一次小于最小值则禁止
                this.options.decreae = false;
                this.decrease.setAttribute('disabled', 'true');
                this.activeDe = false;
                //监听回调
                this.options.onMined && this.options.onMined.call(this);
                'mined' in this.handlers ? this.emit('mined', '') : null;
            }
            this.field.value = val.toFixed(this.options.precision);
            this.value = this.field.value;
            //设置cookie
            this.setCookie();
            //监听
            if (this.field.value < old) {
                //监听回调
                this.options.onChanged && this.options.onChanged.call(this, old, this.field.value);
                'changed' in this.handlers ? this.emit('changed', old, this.field.value) : null;
            }
        }

        //根据disabled参数判断是否可用
        if (this.options.disabled || this.field.hasAttribute('disabled')) {
            //禁止点击
            this.increase.removeEventListener('click', eventIncrease, false);
            this.decrease.removeEventListener('click', eventDecrease, false);
            this.field.setAttribute('disabled', '');
        } else {
            //增加按钮是否可用
            if (!this.options.increase || this.increase.hasAttribute('disabled')) {
                this.increase.removeEventListener('click', eventIncrease, false);
                this.increase.setAttribute('disabled', '');
            } else {
                this.increase.addEventListener('click', eventIncrease, false);
                //支持键盘操作，向上和向右键表示增加，向下和向左键表示减少
                this.field.addEventListener('keydown', eventKey, false);
            }
            //减少按钮是否可用
            if (!this.options.decrease || this.decrease.hasAttribute('disabled')) {
                this.decrease.removeEventListener('click', eventDecrease, false);
                this.decrease.setAttribute('disabled', '');
            } else {
                this.decrease.addEventListener('click', eventDecrease, false);
                //支持键盘操作，向上和向右键表示增加，向下和向左键表示减少
                this.field.addEventListener('keydown', eventKey, false);
            }
            //input失焦重新检验input的值
            this.field.onblur = () => {
                //修正值
                this.value =  this.correctValue(Number(this.field.value));
                this.field.value = this.value;
                //设置cookie
                this.setCookie();
            }
            //支持对input使用enter按键
            this.field.addEventListener('keyup',(e)=>{
                if (e.code === 'Enter') {
                    this.field.blur();
                }
            },false)
           
        }
        //回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }

    ruleToSet() {
        //根据参数rule设定不同的规则
        switch (this.options.rule) {
            case 'currency':
                this.options.min = 0.00; this.options.max = ''; this.options.step = 0.01; this.options.precision = 2;
                break;
            case 'quantity':
                this.options.min = 1; this.options.max = 999; this.options.step = 1; this.options.precision = 0;
                break;
            case 'percent':
                this.options.min = 0; this.options.max = 100; this.options.step = 1; this.options.precision = 0;
                break;
            case 'month':
                this.options.min = 1; this.options.max = 12; this.options.step = 1; this.options.precision = 0;
                break;
            case 'day':
                this.options.min = 1; this.options.max = 31; this.options.step = 1; this.options.precision = 0;
                break;
            case 'hour':
                this.options.min = 0; this.options.max = 24; this.options.step = 1; this.options.precision = 0;
                break;
            case 'minute':
                this.options.min = 0; this.options.max = 59; this.options.step = 1; this.options.precision = 0;
                break;
            case 'second':
                this.options.min = 0; this.options.max = 59; this.options.step = 1; this.options.precision = 0;
                break;
        }
    }
    initValue() {
        //如果参数设置了value则赋予input
        if (this.options.value !== '') {
            //如果提供弄个了value值，则input也填入该value值
            if (this.options.min !== '' && this.options.value < this.options.min) {
                this.value = this.options.min;
            } else if (this.options.max !== '' && this.options.value > this.options.max) {
                this.value = this.options.max;
            } else {
                this.value = this.options.value;
            }
            this.field.value = this.value;
        } else {
            //如果没有提供value初始值，则纠正input自身的值
            if (this.field.value) {
                if (this.options.min !== '' && this.options.min > Number(this.field.value)) {
                    //如果有设置最小值，那么自动向input自动填充
                    this.field.value = this.options.min;
                }
                if (this.options.max !== '' && this.options.max < Number(this.field.value)) {
                    //如果有设置最小值，那么自动向input自动填充
                    this.field.value = this.options.max;
                }
                this.value = this.field.value;
            }
        }
    }
    getCookie() {
        //如果有cookie，则将cookie值赋予参数value
        let cookieValue = axCookie.get(this.options.cookie);
        if (this.options.cookie && cookieValue) {
            this.value = cookieValue;
            this.field.value = cookieValue;
        }
    }
    setCookie() {
        //设置cookie
        if (this.options.cookie) {
            axCookie.set(this.options.cookie, Number(this.field.value));
        }
    }
    setAttribute() {
        //如果设置disabled参数为true，那么给input添加disabled属性
        if (this.options.disabled) {
            this.field.setAttribute('disabled', 'true');
            this.increase.setAttribute('diabled', 'true');
            this.decrease.setAttribute('diabled', 'true');
        }
        //如果不能增加减少则添加disabled属性
        if (!this.options.increase) {
            this.increase.setAttribute('diabled', 'true');
        }
        if (!this.options.decrease) {
            this.decrease.setAttribute('diabled', 'true');
        }
        //添加提示文字
        !this.field.getAttribute('placeholder') ? this.field.setAttribute('placeholder', this.options.placeholder) : null;
        //设置长度
        this.options.width ? this.parent.style.width = this.options.width : null;
    }
    createSkeleton() {
        if (!this.targetDom || this.targetDom.nodeName === 'INPUT') {
            //如果是input节点
            this.parent = axAddElem('div', { class: 'ax-amount' });
            this.field = this.targetDom ? this.targetDom : axAddElem('input', { type: 'text' });
            this.decrease = axAddElem('a', { href: '###', decrease: '' }, '<i class="ax-iconfont ax-icon-minus"></i>');
            this.increase = axAddElem('a', { href: '###', increase: '' }, '<i class="ax-iconfont ax-icon-plus"></i>');
            //先生成父节点
            this.field.insertAdjacentElement('beforeBegin', this.parent);
            //再依次追加进来
            this.parent.appendChild(this.field);
            this.parent.appendChild(this.decrease);
            this.parent.appendChild(this.increase);
        } else {
            //如果是div节点
            this.parent = this.targetDom;
            this.field = this.parent.querySelector('input');
            this.increase = this.parent.querySelector('[increase]');
            this.decrease = this.parent.querySelector('[decrease]');
        }
    }
    correctValue(value){
        if (this.options.min !== '' && value < this.options.min) {
            value = this.options.min;
        } else if (this.options.max !== '' && value > this.options.max) {
            value = this.options.max;
        }
        return value;
    }
    set(value, callback) {
        //先修正值
        value =  this.correctValue(value);
        //赋值
        this.field.value = value;
        this.value = value;
        //cookie
        if (this.options.cookie) {
            this.value = axCookie.set(this.options.cookie, value);
        }
        callback && callback.call(this, value);
        return this;
    }
    reset(callback) {
        this.set(this.rawValue);
        callback && callback.call(this, this.rawValue);
        //cookie
        if (this.options.cookie) {
            this.value = axCookie.set(this.options.cookie, this.rawValue);
        }
        return this;
    }
    update(setting, callback) {
        this.options = axExtend(this.options, setting);
        //监听回调
        this.options.onUpdated && this.options.onUpdated.call(this);
        'updated' in this.handlers ? this.emit('updated', '') : null;
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/* // 对使用axAmount属性的元素使用本插件
document.querySelectorAll('[axAmount]').forEach(element => {
    new axAmount(element);
}); */


/*!
* 插件：页面社交分享；使用方法：new axShare('#id',{参数})
*/
class axShare {
    constructor(targetDom, options) {
        // 始终将targetDom转成DOM对象
        this.targetDom = axIdToDom(targetDom);
        //定义参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            source: '',//类似分享页面的标题
            title: '',//分享页面的标题
            des: '',//分享页面的简介
            url: window.location.href,//分享的页面，默认是当前页面的网址，比如：http://www.axui.cn/examples/index.html
            image: '',//分享的图片，数量一张
            key: '',//微博的key，仅微博分享使用
            origin: window.location.origin,//根域名，默认是当前页面的根域名,比如：http://www.axui.cn
            qrTitle: '<p>在微信里点击“发现”，再点击“扫一扫，”<br />便可将本文分享至朋友圈。</p>',//微信分享中的文字
            qrSize: 180,//QR系统中尺寸单位是px
            newPage: true,//点击后是否在新页面打开分享页面，默认打开新页面
            dialogText: '确定分享本页面？',//跳转提示窗口的文字
            onShare: '',//回调函数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
        }, options, this.targetDom,'axShare');
        // 保存this
        // 定义监听事件成员
        this.handlers = {};
        //定义source默认值
        if (!this.options.source) {
            this.options.source = document.querySelector('meta[name="site"]') || document.title;
        }
        //定义title默认值
        if (!this.options.title) {
            this.options.title = document.querySelector('meta[name="title"]') || document.title;
        }
        //定义des默认值
        if (!this.options.des) {
            this.options.des = document.querySelector('meta[name="description"]') || '';
        }
        //定义image默认值
        if (!this.options.image) {
            this.options.image = (document.images[0] || 0).src || '';
        }
        // 点击核心事件
        //初始化

        this.init();
    }
    init() {
          //添加到实例合集
          axInstance.push(this, this.options.insName, 'share');
           //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //定义不同平台的链接格式
        this.shareUrl = {
            qzone: 'http://sns.qzone.qq.com/cgi-bin/qzshare/cgi_qzshare_onekey?url=' + this.options.url + '&title=' + this.options.title + '&desc=' + this.options.des + '&summary=' + this.options.summary + '&site=' + this.options.source + '&pics=' + this.options.image,
            youdao: 'http://note.youdao.com/memory/?title=' + this.options.title + '&pic=' + this.options.image + '&summary=' + this.options.des + '&url=' + this.options.url,
            qq: 'http://connect.qq.com/widget/shareqq/index.html?url=' + this.options.url + '&title=' + this.options.title + '&source=' + this.options.source + '&desc=' + this.options.des + '&pics=' + this.options.image + '&summary=' + this.options.summary,
            weibo: 'https://service.weibo.com/share/share.php?url=' + this.options.url + '&title=' + this.options.title + '&pic=' + this.options.image + '&appkey=' + this.options.key,
            daoban: 'http://shuo.douban.com/!service/share?href=' + this.options.url + '&name=' + this.options.title + '&text=' + this.options.des + '&image=' + this.options.image + '&starid=0&aid=0&style=11',
            linkedin: 'http://www.linkedin.com/shareArticle?mini=true&ro=true&title=' + this.options.title + '&url=' + this.options.url + '&summary=' + this.options.summary + '&source=' + this.options.source + '&armin=armin',
            facebook: 'https://www.facebook.com/sharer/sharer.php?u=' + this.options.url,
            twitter: 'https://twitter.com/intent/tweet?text=' + this.options.title + '&url=' + this.options.url + '&via=' + this.options.origin,
            google: 'https://plus.google.com/share?url=' + this.options.url,
            pinterest: 'https://www.pinterest.com/pin/create/button/?url=' + this.options.url + '&description=' + this.options.des + '&media=' + this.options.image,
            tumblr: 'https://www.tumblr.com/widgets/share/tool?shareSource=legacy&canonicalUrl=&url=' + this.options.url + '&title=' + this.options.title,
        }
        //遍历并创造新连接
        for (let k in this.shareUrl) {
            this.changeHref(k);
        }
        // 创建二维码
        if (this.targetDom.querySelector('[share="wechat"]')) {
            //如果没有引用QRCode插件则不可使用
            if(typeof QRCode == 'undefined'){
                console.error('Please import the QRCode plugin first!');
                return false;
            }else{
                this.createQrcode()
            }
        }
    }
    changeHref(name) {
        let a = this.targetDom.querySelector('[share=' + name + ']');
        if (a) {
            if (a.hasAttribute('confirm')) {
                //如果a需要通过窗口确认跳转
                a.addEventListener('click', (e)=>{
                    //阻止a标签的跳转行为
                    axPreventDefault(e);
                    //创建弹窗，通过点击确认跳转
                    new axDialog({
                        content: this.options.dialogText,
                        footerType: 'right',
                        confirm: {
                            callback: ()=> {
                                //是否打开新页面跳转
                                if (this.options.newPage) {
                                    window.open(this.shareUrl[name]);
                                    this.hide();
                                } else {
                                    window.location.href = this.shareUrl[name];
                                }
                                this.options.onShare && this.options.onShare.call(this, name);
                            }
                        }
                    }).show();
                }, false);
            } else {
                //如果不需要窗口确认，则直接使用a自身的跳转方式
                //是否打开新页面跳转
                if (this.options.newPage) {
                    a.setAttribute('target', '_blank');
                } else {
                    a.removeAttribute('target')
                }
                a.setAttribute('href', this.shareUrl[name]);
                a.addEventListener('click', (e) => {
                    this.options.onShare && this.options.onShare.call(this, name);
                }, false);
            }

        }
    }
    createQrcode() {
        let wechat = this.targetDom.querySelector('[share="wechat"]')
        if (wechat.length === 0) {
            return false;
        }
        let qr = axAddElem('div', { class: 'ax-qrcode' });
        let qrBody = axAddElem('div', { class: 'ax-qrcode-body' });
        let qrTitle = axAddElem('div', { class: 'ax-qrcode-title' }, this.options.qrTitle);
        new QRCode(qrBody, { text: this.options.url, width: this.options.qrSize, height: this.options.qrSize });

        qr.appendChild(qrBody);
        qr.appendChild(qrTitle);
        //以popup实例
        new axPopup(wechat, {
            content: qr,
            padding: true,
            placement: 'bottom',
            close: false,
            footerShow: false,
        });
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/*!
* 插件：自动完成插件；使用方法：new axComplete('#id',{参数})
*/
class axComplete {
    constructor(elem, options) {
        // 获得检索表单
        this.inputDom = axIdToDom(elem);
        if (!this.inputDom) {
            //如果不存在则不再执行
            return false
        }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            content: '',//检索数据源
            async: '',//是否为异步获取数据，可填json或sql
            ajaxType: 'post',//异步类型，默认post，可填get
            fuzzy: true,//是否模糊检索，默认为true，以空格区分多个关键字，只要满足匹配到一个关键字便可；如果需要精确检索可设为false，此时input的value文本将作为一个关键字
            reorder: false,//检索结果是否需要排序，默认false不重排，可设为true，此时将按权重进行先后排序
            limit: 20,//检索结果显示条数，默认为20，可自定义展示结果条数，为0表示不限制
            ignore: true,//是否忽略英文大小写，默认true忽略，可选择false不忽略
            keys: 'label',//检索字段（属性），默认从label文本检索，多个字段（属性）用英文逗号隔开
            template: '',//生成检索结果列表的模板，变量使用<# this.xxxx #>表示
            className: '',//生成检索结果列表的类名
            breakShow: true,//检索结果是否使用分割线
            msgShow: true,//是否显示检索结果，默认true显示，可选择false不显示
            msgPlace: 'inside',//结果提示存放在哪里，默认放在结果列表顶部，可以使用#id，dom或newline
            highlight: true,//是否让关键字高亮，默认高亮
            fill: 'label',//填入到input的值，默认为label，可以填id、value或者自定义的属性
            idStart: 0,//生成的标准数据数组的id从那个值开始，默认为0
            language: {
                message: '共有<# this.value #>个结果符合要求',
            },
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            popup: { rel: '', triggerRel: 'click' },//自定义下拉菜单popup的参数
            onInit: '',//回调函数，初始化后执行，无参数
            onSearch: '',//回调函数，检索后执行，支持一个参数即获得的检索结果
            onSelected: '',//回调函数，点选项目后执行，支持一个参数即当前点击项目
            onUpdate: '',//回调函数，更新内容后执行，支持一个参数即当前列表数据
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.inputDom, this.constructor.name);
        /* data:[{id:0,label:'张三',value:'张三',age:''},
          {id:1,label:'李四',value:'李四',age:''},
          {id:2,label:'王五',value:'王五',age:''},] 
          keys:['label','age']
          */
        // 定义监听事件成员
        this.handlers = {};
        //input编辑事件
        this.inputEdit = () => {
            let value = this.inputDom.value;
            this.search(value);
        }
        //初始化
        this.init();
        return this;
    }
    beforeInit() {
        //避免重复init，先取消再绑定
        this.inputDom.removeEventListener('input', this.inputEdit);
        //销毁气泡
        this.popupIns ? (this.popupIns.destroy(), this.popupIns = null) : null;
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
    }
    async init() {
        //销毁气泡，取消绑定事件
        this.beforeInit();
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'complete');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;

        //创建列表骨架
        this.createSkeleton();

        //设置样式属性
        this.setAttribute();

        //先生成popup
        this.popupIns = new axPopup(this.inputDom, axExtend({
            content: this.wrapperDom,
            trigger: 'input',
            padding: false,
            placement: 'bottom-start',
            close: false,
            footerShow: false,
        }, this.options.popup));
        //监听
        this.popupIns.on('show', () => {
            //显示气泡后将input的光标移动到最后面
            let value = this.inputDom.value;
            this.inputDom.value = '';
            this.inputDom.focus();
            this.inputDom.value = value;
        });


        this.popupIns.on('hidden', () => {
            //隐藏用原始数据渲染列表
            !this.inputDom.value ? this.reset() : null;
        });

        //根据条件创建模糊检索实例
        this.inputDom.addEventListener('input', this.inputEdit, false);

        //定义数据成员
        this.rawData = [];
        this.sliceData = [];
        this.filterData = [];

        let contentType = axType(this.options.content);
        if (contentType === 'String' && this.options.async) {
            //异步
            await this.getAsyncData(this.options.async, '', (data) => {
                this.getRawData(data);
            });
        } else {
            //同步
            //从文本、datalist、select以及ul获取数据源
            this.getRawData(this.options.content);
        }

        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    createSkeleton() {
        //定义检索结构列表容器
        this.wrapperDom = axAddElem('div', { class: 'ax-complete' });
        //定义检索结构列表容器
        this.listDom = axAddElem('ul');
        this.wrapperDom.appendChild(this.listDom);
        //创建结果提示容器
        this.msgDom = axAddElem('div', { class: 'ax-complete-msg' }, axTplEngine(this.options.language.message, { value: 0 }));
        //根据位置参数存放结果提示
        //newline放在input下方
        if (this.options.msgShow) {
            if (this.options.msgPlace === 'newline') {
                this.inputDom.insertAdjacentElement('afterend', this.msgDom);
            } else if (axStrType(this.options.msgPlace)) {
                //外部指定节点
                axIdToDom(this.options.msgPlace).appendChild(this.msgDom);
            } else {
                //内部列表上方
                this.wrapperDom.insertAdjacentElement('afterbegin', this.msgDom);
            }
        }
    }
    //设置节点属性
    setAttribute() {
        //添加类名
        if (this.options.className) {
            axClass.add(this.wrapperDom, this.options.className);
        }
        //input禁用历史记录
        this.inputDom.setAttribute('autocomplete', 'off');
    }
    //获得初始数据
    getRawData(content) {
        let contentType = axType(content);
        if (contentType === 'String') {
            if (axStrType(content)) {
                //从datalist标签获取数据[{id:'',label:'',value:''},{}]
                this.rawData = axSelectToArr(content, this.options.idStart).data;
            } else {
                //从字符串以英文逗号分割获得数据，并转成标准数据[{id:'',label:'',value:''},{id:'',label:'',value:''}]
                let splitArr = axSplitToArr(content);
                this.rawData = splitArr.map((k, i) => {
                    return { id: this.options.idStart + i, label: k, value: k }
                });
            }
        } else if (contentType === 'Array') {
            if (typeof content[0] == 'object') {
                //如果是对象数组，判断是否有id和value，如果没有则补充完整，转成标准数据[{id:'',label:'',value:''},{id:'',label:'',value:''}]
                content.forEach((k, i) => {
                    !k.hasOwnProperty('id') ? k.id = this.options.idStart + i : null;
                    !k.hasOwnProperty('value') ? k.value = k.label : null;
                });
                this.rawData = content;
            } else if (typeof content[0] == 'string') {
                //如果是字符串数组，判断是否有id和value，如果没有则补充完整，转成标准数据[{id:'',label:'',value:''},{id:'',label:'',value:''}]
                this.rawData = content.map((k, i) => {
                    return { id: this.options.idStart + i, label: k, value: k }
                });
            }
        }
        //根据模板追加dom
        this.rawData.forEach(k => {
            k.dom = axStrToDom(axTplEngine(this.getTemplate(), k));

            //点击赋值给input
            k.dom.onclick = () => {
                //如果有填入参数，并且该参数存在于检索结果中
                this.inputDom.value = k[this.options.fill];
                //关闭气泡
                this.popupIns.hide();
                //监听回调
                this.handlers.hasOwnProperty('selected') ? this.emit('selected', k) : null;
                this.options.onSelected && this.options.onSelected.call(this, k);
            }
        });

        //获得slice数据
        this.sliceData = this.getSliceData();
        //渲染节点
        this.reRenderList(this.sliceData);
    }
    getSliceData() {
        return this.options.limit > 0 ? this.rawData.slice(0, this.options.limit) : this.rawData;

    }
    //获得检索数据
    getFilterData(value) {
        this.filterData = axArrSearch(value, this.options.keys, this.rawData, { fuzzy: this.options.fuzzy, reorder: this.options.reorder, limit: this.options.limit, ignore: this.options.ignore });
    }
    //从数据库中获得数据
    getAsyncData(type = 'json', value, callback) {
        return axAjax({
            //根据参数进行数据扩展
            data: type === 'sql' ? { value: axSplitToArr(value, ' ', 'one').join(','), keys: this.options.keys, fuzzy: this.options.fuzzy, limit: this.options.limit, reorder: this.options.reorder, ignore: this.options.ignore } : '',
            type: this.options.ajaxType,
            url: this.options.content,
            opened: (response) => {
                this.contentXhr = response.xhr;
            },
            success: (response) => {
                if (axIsEmpty(response.content)) {
                    console.warn(`No data obtained from the ${type === 'sql' ? 'database' : 'json'}!`);
                    return false;
                }
                this.listDom.innerHTML = '';
                callback && callback(response.content);
            }
        }, this.listDom);
    }
    //渲染列表
    reRenderList(data) {
        //先清空
        this.listDom.innerHTML = '';
        if (axIsEmpty(data)) {
            return false;
        }
        //再追加
        let fragment = document.createDocumentFragment();
        if (data[0].hasOwnProperty('weight')) {
            //有weight属性说明是搜索结果
            data.forEach(k => {
                if (this.options.highlight) {
                    //高亮关键字
                    let pattern = new RegExp(k.keys.join('|'), 'g'),
                        labelDom = k.source.dom.querySelector('[label]') || k.source.dom.querySelector('[caption]');
                    //先去掉s标签复原在重设s标签
                    labelDom.innerHTML = labelDom.innerHTML.replace(/<[^>]+>/g, '').replace(pattern, match => `<s>${match}</s>`);
                }
                fragment.appendChild(k.source.dom);
            });
            //输出
            this.output = data;
        } else {
            //无weight属性说明来自content的数据
            data.forEach(k => {
                if (this.options.highlight) {
                    //清理HTML标签元素
                    let labelDom = k.dom.querySelector('[label]') || k.dom.querySelector('[caption]');
                    labelDom.innerHTML = labelDom.innerHTML.replace(/<[^>]+>/g, '');
                }
                fragment.appendChild(k.dom);
            });
        }
        //追加到父节点
        this.listDom.appendChild(fragment);

    }
    //定义模板
    getTemplate() {
        let tpl = `<li>
                        ${this.options.breakShow ? '<hr/>' : ''}
                        <div class="ax-bulletin ax-hover">
                            <div class="ax-bulletin-body"><i caption><# this.label #></i></div>
                        </div>
                    </li>
                        `;
        return this.options.template || tpl;
    }
    async search(value, callback) {
        //如果已经销毁，或者本身就是关闭状态则不再执行
        if (this.destroyed) {
            return this;
        }
        let arr = [];
        if (value) {
            if (typeof this.options.content === 'String' && this.options.async === 'sql') {
                //sql异步
                await this.getAsyncData('sql', value, (data) => {
                    //找到了则修正数据并渲染列表
                    this.getRawData(data);
                    this.popupIns.show();
                    arr = this.sliceData;
                    //显示结果提示
                    this.msgDom.innerHTML = axTplEngine(this.options.language.message, { value: arr.length });
                });
            } else {
                //同步
                this.getFilterData(value);
                if (this.filterData.length === 0) {
                    this.listDom.innerHTML = '';
                } else {
                    this.reRenderList(this.filterData);
                    this.popupIns.show();
                    arr = this.filterData;
                }
            }
            //如果结果放在外面或者不显示，结果为空应该隐藏popup
            arr.length === 0 && (this.options.msgPlace !== 'inside' || !this.options.msgShow) ? this.popupIns.hide() : null;
            //显示结果提示
            this.msgDom.innerHTML = axTplEngine(this.options.language.message, { value: arr.length });
        } else {
            this.popupIns.hide();
        }
        //监听回调
        this.handlers.hasOwnProperty('search') ? this.emit('search', arr) : null;
        this.options.onSearch && this.options.onSearch.call(this, arr);
        callback && callback.call(this, data);
        return this;
    }
    async update(content, callback) {
        //如果已经销毁，或者本身就是关闭状态则不再执行
        if (this.destroyed) {
            return this;
        }
        this.options.content = content;
        //重置列表
        if (typeof this.options.content === 'String' && this.options.async === 'sql') {
            //sql异步
            await this.getAsyncData('sql', '', (data) => {
                this.getRawData(data);
            });
        } else {
            //同步
            this.getRawData(this.options.content);
        }
        //监听回调
        this.handlers.hasOwnProperty('update') ? this.emit('update', this.sliceData) : null;
        this.options.onUpdate && this.options.onUpdate.call(this, this.sliceData);
        callback && callback.call(this, this.sliceData);
        return this;
    }
    async reset(callback) {
        //如果已经销毁，或者本身就是关闭状态则不再执行
        if (this.destroyed) {
            return this;
        }
        //input清零
        this.inputDom.value = '';
        //重置列表
        if (typeof this.options.content === 'String' && this.options.async === 'sql') {
            //sql异步
            await this.getAsyncData('sql', '', (data) => {
                this.getRawData(data);
            });
        } else {
            //同步
            this.getRawData(this.sliceData);
        }
        //监听回调
        this.handlers.hasOwnProperty('reset') ? this.emit('reset', this.sliceData) : null;
        this.options.onReset && this.options.onReset.call(this, this.sliceData);
        callback && callback.call(this, this.sliceData);
        return this;
    }
    //销毁实例
    destroy(callback) {
        //销毁气泡，取消绑定事件
        this.beforeInit();
        //更新销毁状态
        this.destroyed = true;
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
// 对使用axComplete属性的元素使用本插件
/* document.querySelectorAll('[axComplete]').forEach(element => {
    new axComplete(element);
});
 */
/*!
 * 插件：页签切换；使用方法：axDate('#id',{参数})
 */
class axDate {
    constructor(elem, options) {
        //获得定位按钮
        this.inputDom = axIdToDom(elem);
        //合并参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//存储名称，字符串格式；如果需要保存实例当前数据请设定唯一的值，比如：storageName:'mystorage001'
            content: '',//日期字符串或对象，支持写法如：2022-2-3,2022/2/3，new Date(2022,2,3)，如果为空则取当前日期
            format: 'YYYY/MM/DD',//面板类型，默认YYYY/MM/DD，可选择YYYY、YYYY/MM（或-）、YYYY/MM/DD（或-）、YYYY/MM/DD hh:mm（或-）、YYYY/MM/DD hh:mm:ss（或-）、hh:mm:ss、hh:mm、hh
            mode: 'single',//selected模式，默认single单选模式，可选择multiple多选模式或range区间模式
            maxSelection: 100,//最大选择数量，对multiple:true有效，默认可选中100项
            rows: 1,//面板网格显示几行，默认1行，注意非inline展示方式建议rows<=2，否则会产生滚动条
            cols: 1,//面板网格显示几列，默认1列
            minDate: '',//设置可选的最小日期，小于将禁止操作，日期字符串类型或Date类型，字符串类型则与format格式一致
            maxDate: '',//设置可选的最大日期，大于将禁止操作，日期字符串格式或Date格式，字符串类型则与format格式一致
            flexible: false,//是否让整个日期选择控件100%铺满容器，默认false，可选择true允许
            footerShow: true,//是否显示日期选择控件的底部按钮部分，默认true显示，可选择false不显示
            nowToHide: false,//非inline显示状态，点击现在按钮是否关闭窗口，默认false，可选择true
            nowShow: false,//底部是否显示“现在”按钮，默认false不显示，可选择true显示
            clearShow: true,//底部是否显示“清除”按钮，默认false不显示
            closeShow: true,//底部是否显示“关闭”按钮，默认true显示
            cancelShow: false,//是否显示“取消”按钮，默认false不显示，该按钮将实现clear+close两者的功能
            resultShow: false,//是否在头部显示选中结果，默认不显示，可选择true显示选中结果
            confirmToHide: true,//点击确认按钮是否关闭窗口，默认true直接关闭，可选择false手动关闭
            autoFill: false,//点击面板网格，面板中的selected值是否自动填充到input，默认false不自动而是通过“确认”按钮填充
            autoCorrect: true,//对于非inline展示方式是否需要重新打开时从input/div中读取值作为当前的selected值，默认true自动校正，可选择false不处理
            forceSelect: false,//点击确认按钮是否需要强制选择至少一个日期，默认false，可选择true
            language: {
                month: ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12'],//月份数字表示
                week: ['一', '二', '三', '四', '五', '六', '日'],//星期文字表示
                year: {//年份下拉框的文字
                    prev: '上一页',
                    next: '下一页',
                    placeholder: '输入年份',
                },
                range: '至',//区间日期选择的连接文字
                unit: {//日期单位表示
                    Y: '年',
                    M: '月',
                    D: '日',
                    h: '时',
                    m: '分',
                    s: '秒',
                },
                time: {//时分秒选择面板文字
                    select: '时间选择',
                    start: '开始时间',
                    end: '结束时间'
                },
                button: {//底部按钮文字
                    clear: '清除',
                    close: '关闭',
                    cancel: '取消',
                    now: '现在',
                    confirm: '确定',
                },
                toolTip: {
                    restore: '还原',
                    reset: '归零',
                    now: '现在',
                    close: '关闭',
                },
                empty: '<i class="ax-color-ignore">还未选择日期!</i>',//结果展示框中为空时默认文字
                message: {
                    requireTwoValue: '区间模式至少需要选择两个日期！',
                    requireYearFormat: '请填入3或4位年份数字!',
                    requireOneSelected: '请至少选择一个日期!',
                },
                noEvent: '<i class="ax-color-ignore">今天没有需要安排的事项!</i>',
            },
            weekStart: 'monday',//一周以周几开始，默认monday即以周一开始，可选择sunday即周日
            separator: ',',//输出或输入多个日期的分隔符，默认是英文逗号
            buttonDom: '',//打开日期popup或dialog的按钮节点，#id/.classname/DIV或DOM节点
            positionDom: '',//定位日期的节点，#id/.classname/DIV或DOM节点
            display: 'popup',//显示日期的方式，默认popup气泡小窗显示，可选择inline行内显示、dialog对话框或drawer显示
            toDrawer: true,//display如果是气泡，在手机端是否都转为drawer显示（手机端drawer显示日期体验最好）
            bubble: {},//popup或dialog的参数
            fillNow: false,//如果input没有值，而且content为空，是否以今天作为初始值，默认false不允许，可填true自动填入
            next: '',//关闭气泡后将要打开的其他气泡，该参数填写当页的其他气泡的实例名（insName）
            lunar: {
                enable: false,//是否启用农历日期，默认false不启用，可选择true启用
                handler: '',//公历转农历函数，最终返回JSON对象
                map: {},//属性映射，常用属性包括animal、month、date、gzYear、gzMonth、gzDate、term、lunarFestival、solarFestival
                template: `<div>农历：<#this.animal#>年 <#this.month#><#this.date#></div>
                        <div>干支：<#this.gzYear#>年 <#this.gzMonth#>月 <#this.gzDate#>日</div>
                        <div>节日/节气：<#this.term#> <#this.lunarFestival#> <#this.solarFestival#></div>
                        `,//事件容器中显示的内容模板，模板中的变量与map中的属性对应
                arrange: ['solarFestival', 'lunarFestival', 'term', 'date'],//单元格中农历位置显示内容的顺序，默认顺序是：如果当天是公历节日则显示公历节日，否则判断是否有农历节日，以此类推
                tags: [],//日期单元格右上角显示一个字符，比如“休”或“班”，每一项的属性包括date、content、props
            },
            event: {
                enable: false,//是否显示事件指示功能，默认false不显示，可选择true显示，开启后events.length>0时在日期中以小点表示
                display: 'popup',//显示事件的方式，默认popup即气泡方式展示事件，可选择inline的方式以固定面板的方式展示
                template: `<div label><#this.label#></div><div content><#this.content#></div>`,
                list: [],//事件列表数组，每一项的属性有date、label和content
            },
            menu: [],//是否显示左侧工具面板，其length==0表示不显示，每一项的格式为{label:'',action:()=>{}}
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onClear: '',//回调函数，清除所有selected项后执行，无参数
            onSet: '',//回调函数，设置selected后执行，支持一个参数即设置日期对象（数组）
            onSetRange: '',//回调函数，设置日期区间后执行，支持一个参数即设置日期对象（数组）
            onCreatePanels: '',//回调函数，创新面板（刷新面板）后执行，无参数
            onCancel: '',//回调函数，移除selected项并关闭气泡后执行，支持一个参数即设置日期对象（数组）
            onConfirm: '',//回调函数，合并日期时间填入input后执行，支持一个参数即填入文本
            onSetTime: '',//回调函数，设置时间后执行，支持一个参数即当前时间对象
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
        }, options, this.inputDom, this.constructor.name);
        //定义监听事件成员
        this.handlers = {};
        //触发事件函数
        this.triggerShow = () => {
            this.active(this);
        }
        //修正content参数
        this.content = [];
        //保存初始值
        this.raw = [];
        //固定数量的月份字符串
        this.monthStr = `
            <li><span content><i label>01${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>02${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>03${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>04${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>05${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>06${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>07${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>08${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>09${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>10${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>11${this.options.language.unit.M}</i></span></li>
            <li><span content><i label>12${this.options.language.unit.M}</i></span></li>
        `;
        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'date');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        this.destroyed = false;
        //初始化数据
        this.data = {
            startDate: '',
            selectedDates: [],
            type: '',//year/month/day/time/daytime
            panels: [
                /* {
                    yearDom: '',
                    monthDom: '',//month类型没有该属性
                    year:'',
                    month:'',//month类型没有该属性
                    listDom: '',
                    list: [
                        {
                            value: {},//代表的日期
                            wrapperDom:'',//li节点
                            label: '',//日期名
                            labelDom: '',//日期节点
                            contentDom:'',//content节点
                            busy:'',//是否有事件
                            tag:'',//右上角显示一个标记，比如“休”或“班”
                            lunar:'',//农历名
                            lunarDom:'',//农历节点
                            term:'',//是否是节气农历
                            selected: '',//是否是选择状态
                            disabled: '',//是否禁用状态
                            exceed: '',//是否是非本月日期
                            workDay:'',//是否是工作日
                            restDay:'',//是否是休息日
                            range:'',//start区间起点、end区间终点或contain区间

                        }
                    ],

                } */
            ],
            menu: [],
            year: {
                panelDom: '',
                listDom: '',
                list: []
            },
            /*
            year:{
                panelDom:'',
                prevDom:'',
                nextDom:'',
                inputDom:'',
                goDom:'',
                listDom:'',
                list:[
                    {
                        value:'',
                        wrapperDom:'',
                        label:'',
                        labelDom:'',
                        contentDom:'',
                        disabled:'',
                    },
                ]
            }
            */
            month: {
                panelDom: '',
                listDom: '',
                list: []
            },
            /*
            month:{
                panelDom:'',
                listDom:'',
                list:[
                    {
                        value:'',
                        wrapperDom:'',
                        label:'',
                        labelDom:'',
                        contentDom:'',
                        disabled:'',
                    },
                ]
            }
            */
            time: [],
            /*  time: [
                 {
                    value:{},
                    panelDom:'',
                    closeDom:'',
                    resetDom:'',
                    restoreDom:'',
                     hh: [
                         {
                             value:'',
                             wrapperDom:'',
                             contentDom: '',
                             selected: '',
                             disabled: '',
                         }
                     ],
                     mm: [],
                     ss: []
                 },
                 {
                    value:{}
                 }
             ] */

        };
        //判断是否有存储值而更新参数
        if (this.options.storageName) {
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            } else {
                //则先存下空对象
                axLocalStorage.set(this.options.storageName, {});
            }
        }

        //修正format，获得标准的this.options.format
        this.options.format = axTrim(this.options.format);
        //判断panel类型，获得this.data.type变量
        this.data.type = axDateMethod.getType(this.options.format);

        //修正列数，如果是区间模式那么cols应该大于等于2
        this.options.mode === 'range' && this.options.cols === 1 ? this.options.cols = 2 : null;

        //移动端使用使用drawer模式，并再修正cols和rows
        if (axClient() === 'phone') {
            this.options.display !== 'inline' && this.options.toDrawer ? this.options.display = 'drawer' : null;
            //农历则只能显示一个子面板
            if (this.options.lunar.enable) {
                this.options.cols = 1;
                this.options.rows = 1;
            } else {
                //其他情况
                if (this.options.cols > 1) {
                    this.options.cols = 1;
                    this.options.rows = 2;
                }
            }
            //底部按钮两侧分布false
            this.bilateralCondition = false;
        } else {
            //定义footer部分的显示条件，为true则两侧分布，为false则线条按钮分布
            this.bilateralCondition = (this.data.type !== 'time' && this.options.cols > 1) || (this.data.type === 'time' && this.options.mode === 'range');
        }

        //定义separator
        this.separator = this.options.mode === 'range' && this.options.language.range ? this.options.language.range : this.options.separator;

        //定位按钮和触发按钮，如果为空则默认等于inputDom
        this.positionDom = this.options.positionDom ? axIdToDom(this.options.positionDom) : this.inputDom;
        this.buttonDom = this.options.buttonDom ? axIdToDom(this.options.buttonDom) : this.inputDom;

        //创建selectedDate监听
        this.selectedDatesProxy = new axObserve(this.data.selectedDates, {
            deep: true,
        });
        this.selectedDatesProxy.instance.on('complete', (data) => {
            if (this.resultDom) {
                let newContent = data.target.map(k => axDateMethod.fillFormat(k, this.options.format)),
                    oldContent = this.resultIns.content.map(k => k.value);
                //有tag实例存在，而且要传递的值与tag当前的值不一致
                if (this.options.resultShow && this.resultIns && JSON.stringify(newContent) != JSON.stringify(oldContent)) {
                    this.resultIns.update({
                        content: newContent,
                    });
                }
            }

            //自动合并并填入input
            this.options.autoFill ? this.confirm(false) : null;

        });


        //因为在handleDayContent中修改了startDate和selectedDate，而createSkeleton中创建日期面板的时候需要用到，所以放在上方
        if (this.data.type !== 'time') {
            //修正content，获得this.content数组变量，但是数组中每一项并不一定都是日期对象
            this.handleDayContent();
        }

        //渲染主骨架节点
        this.createSkeleton();

        //根据不同行列和面板类型创建面板网格
        if (this.data.type !== 'time') {
            //纯时间类型不需要创建面板
            this.createPanels();
        } else {
            //时间类型处理content初始值，因为是在createSkeleton中创建了时间面板，所以放在下方
            this.handleTimeContent();
        }

        //展示方式
        if (this.options.display === 'inline') {
            this.positionDom.insertAdjacentElement('afterend', this.targetDom);
        } else if (this.options.display === 'popup') {
            this.bubbleIns = this.bubbleIns || new axPopup(this.positionDom, axExtend({
                size: '',
                padding: false,
                placement: 'bottom-start',
                close: false,
                footerShow: false,
                rel: this.options.buttonDom,
                content: this.targetDom,
            }, this.options.bubble));
        } else if (this.options.display === 'dialog') {
            this.bubbleIns = this.bubbleIns || new axDialog(axExtend({
                size: '',
                padding: false,
                footerShow: false,
                content: this.targetDom,
                rel: this.options.buttonDom,
            }, this.options.bubble), this.buttonDom);
        } else if (this.options.display === 'drawer') {
            //drawer显示方式将自动设置flexible为true
            this.options.flexible = true;
            this.bubbleIns = this.bubbleIns || new axDrawer(axExtend({
                size: '',
                padding: false,
                footerShow: false,
                content: this.targetDom,
                placement: 'bottom',
                rel: this.options.buttonDom,
            }, this.options.bubble), this.buttonDom);
        }
        if (this.bubbleIns) {
            this.bubbleIns.on('show', () => {
                //显示前校验input的值
                this.checkBeforeShow();
            })
        }

        //追加样式或属性
        this.setAttribute();


        this.renderFinish();

        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    //根据类型创建一个panel
    createPanel(index, dateObj) {
        let curDate = axDateMethod.toDate(dateObj, this.data.type),
            panelObj = {},
            getWeek = () => {
                if (this.options.weekStart === 'monday') {
                    return `<li>${this.options.language.week[0]}</li>
                            <li>${this.options.language.week[1]}</li>
                            <li>${this.options.language.week[2]}</li>
                            <li>${this.options.language.week[3]}</li>
                            <li>${this.options.language.week[4]}</li>
                            <li>${this.options.language.week[5]}</li>
                            <li>${this.options.language.week[6]}</li>`;
                } else {
                    return `<li>${this.options.language.week[6]}</li>
                            <li>${this.options.language.week[0]}</li>
                            <li>${this.options.language.week[1]}</li>
                            <li>${this.options.language.week[2]}</li>
                            <li>${this.options.language.week[3]}</li>
                            <li>${this.options.language.week[4]}</li>
                            <li>${this.options.language.week[5]}</li>`;
                }
            };

        if (this.data.type.includes('date')) {
            //有年月日，包含了day和daytime类型
            //本次panel月份+1
            curDate.setMonth(curDate.getMonth() + index);
            panelObj.month = curDate.getMonth();
            panelObj.year = curDate.getFullYear();
            //添加节点属性
            panelObj.panelDom = axStrToDom(`
                            <li panel='${this.options.lunar.enable ? 'lunar' : 'date'}'>
                                <div header>
                                    <span year><i>${panelObj.year}</i>${this.options.language.unit.Y}</span><span month><i>${panelObj.month + 1}</i>${this.options.language.unit.M}</span>
                                </div>
                                <div body days>
                                    <ul column>
                                        ${getWeek()}
                                    </ul>
                                    <ul list></ul>
                                </div>
                            </li>
                            `);
            panelObj.type = 'date';
            panelObj.headerDom = panelObj.panelDom.querySelector('[header]');
            panelObj.yearDom = panelObj.headerDom.querySelector('[year]');
            panelObj.monthDom = panelObj.headerDom.querySelector('[month]');
            //panelObj.value = ;
            //生成列表并追加list数组     
            panelObj.listDom = panelObj.panelDom.querySelector('[list]');
            panelObj.list = this.createPanelDayList(panelObj.listDom, curDate);
        } else if (this.data.type === 'month') {
            //只有月份，单指month类型
            //本次panel年份+1
            curDate.setFullYear(curDate.getFullYear() + index);
            panelObj.year = curDate.getFullYear();
            //添加节点属性
            panelObj.panelDom = axStrToDom(`
                        <li panel='month'>
                            <div header>
                                <span year><i>${panelObj.year}</i>${this.options.language.unit.Y}</span>
                            </div>
                            <div body>
                                <ul list>${this.monthStr}</ul>
                            </div>
                        </li>
                        `);
            panelObj.type = 'month';
            panelObj.headerDom = panelObj.panelDom.querySelector('[header]');
            panelObj.yearDom = panelObj.headerDom.querySelector('[year]');
            //panelObj.value = ;
            //生成列表并追加list数组     
            panelObj.listDom = panelObj.panelDom.querySelector('[list]');
            panelObj.list = this.createPanelMonthList(panelObj.listDom, curDate);
        } else if (this.data.type === 'year') {
            //只有年份，单指year类型
            //本次panel年份+18，-7是向前倒退7年让当年处于面板中间
            curDate.setFullYear(curDate.getFullYear() + 18 * index - 7);
            panelObj.year = curDate.getFullYear();
            panelObj.panelDom = axStrToDom(`
            <li panel='year'>
                <div header>
                    <span year><i>${panelObj.year}</i>${this.options.language.unit.Y}~<i>${panelObj.year + 17}</i>${this.options.language.unit.Y}</span>
                </div>
                <div body>
                    <ul list></ul>
                </div>
            </li>
            `);
            panelObj.type = 'year';
            panelObj.headerDom = panelObj.panelDom.querySelector('[header]');
            //生成列表并追加list数组     
            panelObj.listDom = panelObj.panelDom.querySelector('[list]');
            panelObj.list = this.createPanelYearList(panelObj.listDom, curDate);
        }

        //追加其他属性
        panelObj.bodyDom = panelObj.panelDom.querySelector('[body]');
        //更新panels
        this.data.panels.push(panelObj);
        //加入物理节点
        this.panelsDom.appendChild(panelObj.panelDom);
    }
    //更新或创建panel列表
    createPanels(dateObj, force = false, callback) {
        //销毁了就不再执行
        if (this.destroyed) {
            return false;
        }
        //计算子面板数量
        let index = this.options.rows * this.options.cols,
            startDate;

        //转换日期
        if (dateObj) {
            startDate = axDateMethod.toDate(dateObj, this.data.type);
        } else {
            startDate = this.data.startDate;
        }
        //日期在当前面板则不刷新面板
        if (!force && this.inCurPanels(startDate)) {
            return false;
        };


        //先清空
        this.data.panels = [];
        this.panelsDom.innerHTML = '';

        //根据行列数量创建对应数量的子面板
        for (let k = 0; k < index; k++) {
            this.createPanel(k, startDate);
        }

        //更新startDate
        this.data.startDate = startDate;

        //保持selected状态
        let curItems = this.getCurGrids();
        curItems.forEach(k => {
            if (this.inSelectedDate(k) && !k.selected) {
                this.selectedAttrToggle(k, true);
            }
        });
        //保持range状态
        this.options.mode === 'range' ? this.setRangeAttr(curItems) : null;

        //激活今天日期
        let nowObj = axDateMethod.toObj(new Date(), false, this.data.type),
            nowItem = curItems.find(k => JSON.stringify(k.value) == JSON.stringify(nowObj));
        if (nowItem) {
            nowItem.wrapperDom.setAttribute('now', '');
        }

        //渲染完成后执行
        this.renderFinish();

        //监听回调
        this.handlers.hasOwnProperty('createPanels') ? this.emit('createPanels', '') : null;
        this.options.onCreatePanels && this.options.onCreatePanels.call(this);
        callback && callback.call(this);
        return this;
    }
    //判断选中项中是否包含某日期
    inSelectedDate(dateObj) {
        //dateObj可以是{value:'',label:''}或{YYYY:'',MM:''}
        let value = this.getDateValue(dateObj, this.data.type);
        //value返回的是{YYYY:'',MM:'',DD:''}，没有hh、mm和ss
        return this.selectedDatesProxy.proxy.some(k => {
            let { hh, mm, ss, ...rest } = k;
            return JSON.stringify(rest) == JSON.stringify(value);
        });
    }
    //显示事件内容
    showEvent(obj) {
        //显示今日信息
        this.todayDom.innerHTML = axDateMethod.fillFormat(obj.value, 'YYYY年MM月DD日 周WW');
        //显示农历详细
        if (this.options.lunar.enable && this.options.lunar.handler) {
            let lunarObj = { ...this.options.lunar.map },
                //要求传入真实的年月日，插件中MM是js的规则（0~11月）
                lunarValue = this.options.lunar.handler(obj.value.YYYY, obj.value.MM + 1, obj.value.DD);
            for (let k in lunarObj) {
                if (lunarObj.hasOwnProperty(k)) {
                    lunarObj[k] = lunarValue[lunarObj[k]];
                }
            }
            this.lunarDom.innerHTML = axTplEngine(this.options.lunar.template, lunarObj)
        }
        //显示事件
        let item = this.options.event.list.find(k => JSON.stringify(axDateMethod.toObj(k.date, false, 'date')) == JSON.stringify(axDateMethod.toObj(obj.value, false, 'date')));
        this.eventDom.innerHTML = item && item.content ? axTplEngine(this.options.event.template, item) : this.options.language.noEvent;
    }

    //判断日期在当前面板中
    inCurPanels(dateObj, curGrids) {
        let type = this.data.type === 'datetime' ? 'date' : this.data.type,
            date = axType(dateObj) === 'Object' ? { ...dateObj } : dateObj,
            obj = axDateMethod.toObj(date, false, type),
            grids = curGrids || this.getCurGrids(),
            result;
        result = grids.some(k => JSON.stringify(k.value) == JSON.stringify(obj));
        return result;
    }
    setAttribute() {
        //改变底部风格
        !this.bilateralCondition ? this.targetDom.setAttribute('footerType', 'line') : null;
        //设置网格面板模式属性
        this.targetDom.setAttribute('mode', this.options.mode);
        //设置格式
        this.data.type ? this.targetDom.setAttribute('type', this.data.type) : null;
        //设置展示方式
        this.targetDom.setAttribute('display', this.options.display)
        //对input设置只读
        this.inputDom.nodeName === 'INPUT' ? this.inputDom.setAttribute('readonly', '') : null;
        //设置100%铺满
        this.options.flexible ? this.targetDom.setAttribute('flexible', '') : null;
        //没有底部设置属性
        this.targetDom.setAttribute('footerShow', this.options.footerShow);
        //设置农历属性
        this.options.lunar.enable ? this.targetDom.setAttribute('lunar', '') : null;
    }
    inputValueToDates(toDate = true) {
        //如果有button并且值不为空，button可为input或div
        let str, result = [];
        if (this.inputDom.nodeName === 'INPUT') {
            str = this.inputDom.value.trim();
        } else {
            str = this.inputDom.innerText.trim();
        }
        //如果没有内容，或者字符串前两位不是数字
        if (!str || isNaN(parseFloat(str.substr(0, 2)))) {
            return result;
        }
        if (str) {
            if (toDate) {
                //转成Date格式
                result = str.split(this.separator).map(k => new Date(k));
            } else {
                //保持文本格式
                result = axTrim(str);
            }
        }
        return result;
    }
    contentToArr() {
        let type = axType(this.options.content),
            result = [];
        if (this.data.type === 'time') {
            //保持字符串格式
            if (this.inputDom && (this.inputDom.value || this.inputDom.innerText)) {
                result = this.inputValueToDates(false);
            } else {
                if (axIsEmpty(this.options.content)) {
                    result = axDateMethod.fillFormat(new Date(), this.options.format);
                } else {
                    if (type === 'String') {
                        result = this.options.content.split(this.separator);
                    } else if (type === 'Array') {
                        result = this.options.content;
                    }
                }
            }
        } else {
            //转Date日期格式，优先判断从content取值，如果为空则从input中取值
            if (!axIsEmpty(this.options.content)) {
                let type = axType(this.options.content);
                if (type === 'Array') {
                    result = this.options.content.map(k => typeof k === 'string' ? new Date(k) : k);
                } else if (type === 'Date') {
                    result.push(this.options.content);
                } else if (type === 'String') {
                    //this.separator以range分隔符优先，其次普通分隔符
                    result = this.options.content.split(this.separator).map(k => new Date(k));
                }
            } else if (this.inputDom && (this.inputDom.value || this.inputDom.innerText)) {
                //如果有button并且值不为空，button可为input或div
                result = this.inputValueToDates(true);
            } else {
                //如果为空则取今天日期
                result = this.options.fillNow ? [new Date()] : [];
            }

        }
        return result;
    }
    handleDayContent() {
        //button值>content参数>今天
        this.content = this.contentToArr();

        //如果有多个日期，则由近及远排序
        if (this.content.length > 1 && this.options.mode === 'range') {
            this.content.sort(function (a, b) { return a - b });
            //range模式只需要前两个
            this.content.length > 2 ? this.content = this.content.slice(0, 2) : null;
        }

        //更新startDate，创建panel以startDate为基准
        this.data.startDate = this.content.length === 0 ? new Date() : this.content[0];
        //更新data.selectedDate
        let selectedContent = this.datesToObjs(this.content);
        //修改this.selectedDatesProxy.proxy
        this.selectedDatesProxy.proxy.splice(0, this.data.selectedDates.length, ...selectedContent);

        //保存初始值
        this.raw = [...selectedContent];

        //将值填入inputDom
        if (this.inputDom) {
            let value = this.selectedDatesProxy.proxy.map(k => axDateMethod.fillFormat(k, this.options.format)).join(this.separator)
            this.fillInput(value);
        }
    }
    //获得标准三位时间对象数组
    timeStrSTD(data, type = 'object') {
        let dataType = axType(data),
            objArr = [],
            strArr = [],
            dataFun = (obj) => {
                let objType = axType(obj), arr, result;
                if (objType === 'String') {
                    //转数字
                    arr = axTrim(obj).split(':').map(k => ~~k);
                    //补足3位
                    if (arr.length === 1) {
                        arr.push(0, 0);
                    } else if (arr.length === 2) {
                        arr.push(0);
                    }
                    result = { hh: arr[0], mm: arr[1], ss: arr[2] }
                } else if (objType === 'Object') {
                    for (let k in obj) {
                        if (obj.hasOwnProperty(k)) {
                            obj[k] = ~~obj[k];
                        }
                    }
                    //补足3位
                    !obj.hasOwnProperty('mm') ? obj.mm = 0 : null;
                    !obj.hasOwnProperty('ss') ? obj.ss = 0 : null;
                    result = obj;
                }
                //标准的三位时间对象{hh:'',mm:'',ss:''}
                return result;
            }

        //数组循环
        if (dataType === 'Array') {
            data.forEach(k => {
                objArr.push(dataFun(k));
            });
        } else {
            objArr.push(dataFun(data));
        }
        //输出字符串数组
        if (type === 'string') {
            strArr = objArr.map(k => `${axDateMethod.fillZero(k.hh)}:${axDateMethod.fillZero(k.mm)}:${axDateMethod.fillZero(k.ss)}`);
            return strArr;
        } else {
            return objArr;
        }
    }
    handleTimeContent() {
        //this.contentToArr获得值数组['12:02:45','09:03:21']
        //this.timeStrSTD获得标准三位时间对象数组[{hh:'',mm:'',ss:''}]
        this.content = this.timeStrSTD(this.contentToArr());

        //初始化时间面板
        if (this.options.mode === 'range') {
            if (this.content.length === 0) {
                this.setTime([{ hh: 0, mm: 0, ss: 0 }, { hh: 0, mm: 0, ss: 0 }]);
            } else if (this.content.length === 1) {
                this.setTime([this.content[0], { hh: 0, mm: 0, ss: 0 }]);
            } else {
                this.setTime(this.content);
            }
        } else {
            if (this.content.length === 0) {
                this.setTime({ hh: 0, mm: 0, ss: 0 });
            } else {
                this.setTime(this.content[0]);
            }
        }


        //保存初始值
        this.raw = [...this.content];

        //将值填入inputDom
        if (this.inputDom && !axIsEmpty(this.options.content)) {
            let value = this.timeStrSTD(this.content, 'string').join(this.separator)
            if (this.inputDom.nodeName === 'INPUT') {
                this.inputDom.value = value;
            } else {
                this.inputDom.innerText = value;
            }
        }
    }
    datesToObjs(dates) {
        let selectedContent = [];
        if (this.data.type === 'year') {
            selectedContent = dates.map(k => {
                return { YYYY: k.getFullYear() }
            });
        } else if (this.data.type === 'month') {
            selectedContent = dates.map(k => {
                return { YYYY: k.getFullYear(), MM: k.getMonth() }
            });
        } else if (this.data.type === 'time') {
            selectedContent = dates.map(k => {
                return { hh: k.getHours(), mm: k.getMinutes(), ss: k.getSeconds() }
            });
        } else if (this.data.type === 'date') {
            selectedContent = dates.map(k => {
                return { YYYY: k.getFullYear(), MM: k.getMonth(), DD: k.getDate() }
            });
        } else if (this.data.type === 'datetime') {
            selectedContent = dates.map(k => {
                return { YYYY: k.getFullYear(), MM: k.getMonth(), DD: k.getDate(), hh: k.getHours(), mm: k.getMinutes(), ss: k.getSeconds() }
            });
        }
        return selectedContent;
    }
    //打开前校验input值
    checkBeforeShow() {
        if (!this.options.autoCorrect) {
            return false;
        }
        if (this.data.type !== 'time') {
            //非时间类型
            let value = this.inputValueToDates(),
                selectedDates = this.datesToObjs(value),
                isSame = JSON.stringify(selectedDates) == JSON.stringify(this.data.selectedDates);
            if (!isSame) {
                //更新开始时间和面板
                this.data.startDate = value.length > 0 ? value[0] : axDateMethod.toDate(this.data.selectedDates[0], this.data.type);
                this.createPanels();
                //清理selected状态
                this.clear();
                //设置selected
                this.set(selectedDates);
            }
        } else {
            //时间类型
            let timeObj = this.timeStrSTD(this.inputValueToDates(false));
            this.setTime(timeObj);
        }

    }
    createSkeleton() {
        //将骨架转为DOM对象
        let outerHTML = '', footerHTML = '';
        if (this.data.type !== 'time') {
            outerHTML = `<div outer>
                            <div prev></div>
                                <div next></div>
                                <ul panels ${this.bilateralCondition ? 'style="grid-template-columns:repeat(' + this.options.cols + ', 1fr)"' : ''}></ul>
                            </div>`;
        }
        if (this.options.footerShow) {
            footerHTML = `<div footer>
                            ${this.bilateralCondition ? '<div left></div><div right></div>' : ''}
                            </div>`;
        }

        this.targetDom = axStrToDom(`
        <div class="ax-date">
            <div wrapper>
                ${this.options.menu.length > 0 ? '<ul menu></ul>' : ''}
                <div main>
                    ${this.options.resultShow && this.data.type !== 'time' ? '<div result></div>' : ''}
                    ${outerHTML}
                    ${footerHTML}
                </div>
            </div>
        </div>
    `);
        //获取Dom树
        this.wrapperDom = this.targetDom.querySelector('[wrapper]');
        this.menuDom = this.targetDom.querySelector('[menu]');
        this.resultDom = this.targetDom.querySelector('[result]');
        this.outerDom = this.targetDom.querySelector('[outer]');
        this.prevDom = this.targetDom.querySelector('[prev]');
        this.nextDom = this.targetDom.querySelector('[next]');
        this.mainDom = this.targetDom.querySelector('[main]');
        this.panelsDom = this.targetDom.querySelector('[panels]');

        //定义按钮(前提是footer存在)
        if (this.options.footerShow) {
            this.footerDom = this.targetDom.querySelector('[footer]');
            this.timeBtnWrapperDom = this.bilateralCondition ? this.footerDom.querySelector('[left]') : this.footerDom;
            this.handleBtnsWrapperDom = this.bilateralCondition ? this.footerDom.querySelector('[right]') : this.footerDom;

            this.confirmDom = axAddElem('i', { class: `ax-btn${!this.bilateralCondition ? ' ax-plain' : ''} `, confirm: '' }, this.options.language.button.confirm);
            this.clearDom = this.options.clearShow ? axAddElem('i', { class: 'ax-btn ax-plain', clear: '' }, this.options.language.button.clear) : '';
            this.closeDom = this.options.closeShow ? axAddElem('i', { class: 'ax-btn ax-plain', close: '' }, this.options.language.button.close) : '';
            this.cancelDom = this.options.cancelShow ? axAddElem('i', { class: 'ax-btn ax-plain', cancel: '' }, this.options.language.button.cancel) : '';
            this.nowDom = this.options.nowShow ? axAddElem('i', { class: 'ax-btn ax-plain', now: '' }, this.options.language.button.now) : '';
            this.timeBtnDom = this.data.type.includes('time') ? axAddElem('i', { class: 'ax-btn ax-plain', time: '' }, this.options.language.time.select) : '';
            this.yearSearchDom = this.data.type === 'year' ? axAddElem('span', { search: '' }, `<input type="text" placeholder="${this.options.language.year.placeholder}" /><i go>arrow-right</i>`) : '';
            //按钮插入节点
            this.confirmDom ? this.handleBtnsWrapperDom.appendChild(this.confirmDom) : null;
            this.nowDom ? this.handleBtnsWrapperDom.insertAdjacentElement('afterbegin', this.nowDom) : null;
            this.clearDom ? this.handleBtnsWrapperDom.insertAdjacentElement('afterbegin', this.clearDom) : null;
            this.closeDom ? this.handleBtnsWrapperDom.insertAdjacentElement('afterbegin', this.closeDom) : null;
            this.cancelDom ? this.handleBtnsWrapperDom.insertAdjacentElement('afterbegin', this.cancelDom) : null;
            this.timeBtnDom ? this.timeBtnWrapperDom.insertAdjacentElement('afterbegin', this.timeBtnDom) : null;
            this.yearSearchDom ? this.timeBtnWrapperDom.insertAdjacentElement('afterbegin', this.yearSearchDom) : null;
        }


        //详细的两个板块今天时间和今日事件
        if (this.options.event.enable) {
            //构建detail节点
            this.detailDom = axAddElem('div', { detail: '' });
            this.todayDom = axAddElem('div', { today: '' });
            this.eventDom = axAddElem('div', { event: '' }, this.options.language.noEvent);
            this.detailDom.appendChild(this.todayDom);
            this.detailDom.appendChild(this.eventDom);

            //如果开启了农历
            if (this.options.lunar.enable && this.options.lunar.handler) {
                this.lunarDom = axAddElem('div', { lunar: '' });
                this.todayDom.insertAdjacentElement('afterend', this.lunarDom);
            }

            if (this.options.event.display === 'inline') {
                //行内模式
                this.wrapperDom.appendChild(this.detailDom);
            } else if (this.options.event.display === 'popup') {
                //气泡模式
                this.detailIns = this.detailIns || new axPopup(this.panelsDom, {
                    content: this.detailDom,
                    trigger: 'hovor',
                    placement: 'bottom',
                    size: 'sm',
                    footerShow: false,
                });
            }

        }

        //创建时间选择panels（浮窗），time类型并且有footer
        if (this.data.type.includes('time') && this.options.footerShow) {
            //有年月日时间或者纯时间的面板类型才有时间panels
            this.createTimeSelectPanel();
        }

        //创建左侧菜单
        this.createMenu();

        //time类型不需要某些东西
        if (this.data.type !== 'time') {
            //显示结果的tags实例
            if (this.resultDom) {
                this.resultIns = this.resultIns || new axTags(this.resultDom, {
                    content: '',
                    closable: true,
                    text: true,
                    emptyTip: this.options.language.empty,
                });
            }
            //填充next和prev
            if (this.data.type === 'year' || this.data.type === 'month') {
                //年或月面板，值需要nextyear和prevyear
                this.prevDom.innerHTML = `<i prevYear class= "ax-iconfont ax-icon-left-double"></i> `;
                this.nextDom.innerHTML = `<i nextYear class= "ax-iconfont ax-icon-right-double"></i> `;
            } else if (this.data.type.includes('date')) {
                //有日期的面板，则包含四个按钮
                this.prevDom.innerHTML = `<i prevYear class= "ax-iconfont ax-icon-left-double"></i> <i prevMonth class="ax-iconfont ax-icon-left"></i>`;
                this.nextDom.innerHTML = `<i nextYear class= "ax-iconfont ax-icon-right-double"></i> <i nextMonth class="ax-iconfont ax-icon-right"></i>`;
            }
            this.prevYearDom = this.prevDom.querySelector('[prevYear]');
            this.prevMonthDom = this.prevDom.querySelector('[prevMonth]');
            this.nextYearDom = this.targetDom.querySelector('[nextYear]');
            this.nextMonthDom = this.targetDom.querySelector('[nextMonth]');

            //创建年份选择panel（浮窗）
            if (this.data.type === 'month' || this.data.type.includes('date')) {
                //YYYY-MM和YYYY-MM-DD类型有年份选择
                this.createYearPanel();
            }

            //创建月份选择panel（浮窗）
            if (this.data.type.includes('date')) {
                //YYYY-MM-DD类型有月份选择
                this.createMonthPanel();
            }
        }

        //监听回调



    }
    //创建左侧菜单
    createMenu() {
        if (this.options.menu.length > 0) {
            //data.meunu和options.menu的length相同
            this.data.menu = this.options.menu;
            let fragment = document.createDocumentFragment();
            this.data.menu.forEach(k => {
                let elem = axAddElem('li', '', k.label);
                k.dom = elem;
                fragment.appendChild(elem);
            });
            //加入物理节点
            this.menuDom.appendChild(fragment);
            //执行动作并toggle选中状态
            this.data.menu.forEach(k => {
                k.dom.onclick = () => {
                    k.action && k.action.call(this);
                    //点击后设置selected状态
                    k.dom.setAttribute('selected', '');
                    k.selected = true;
                    this.data.menu.filter(i => i !== k).forEach(o => {
                        Reflect.deleteProperty(o, 'selected');
                        o.dom.removeAttribute('selected');
                    });
                }
            });
        }
    }
    createMonthPanel(callback) {
        this.data.month.panelDom = axAddElem('div', { panel: 'month' }, `<div body><ul list>${this.monthStr}</ul></div>`);
        this.data.month.listDom = this.data.month.panelDom.querySelector('[list]');
        [...this.data.month.listDom.children].forEach((k, i) => {
            let obj = {
                wrapperDom: k,
                contentDom: k.querySelector('[content]'),
                value: i,
                //labelDom: k.querySelector('[label]'),
                //label: i
            };
            this.data.month.list.push(obj);
        });
        //追加到页面
        this.panelsDom.insertAdjacentElement('afterend', this.data.month.panelDom);
        callback && callback();
    }
    createYearPanel(callback) {
        this.data.year.panelDom = axAddElem('div', { panel: 'year' }, `
    <div body>
                        <ul list></ul>
                        <ul pages>
                            <li><i prev>${this.options.language.year.prev}</i></li>
                            <li><input type="text" placeholder=${this.options.language.year.placeholder} class="ax-sm"/><i go>arrow-right</i></li>
                            <li><i next>${this.options.language.year.next}</i></li>
                        </ul>
                    </div>
    `);
        this.data.year.listDom = this.data.year.panelDom.querySelector('[list]');
        this.data.year.prevDom = this.data.year.panelDom.querySelector('[prev]');
        this.data.year.nextDom = this.data.year.panelDom.querySelector('[next]');
        this.data.year.inputDom = this.data.year.panelDom.querySelector('input');
        this.data.year.goDom = this.data.year.panelDom.querySelector('[go]');
        //追加到页面
        this.panelsDom.insertAdjacentElement('afterend', this.data.year.panelDom);
        callback && callback();
    }
    createYearList(panel, index, offset) {
        //填充年份列表，先清空
        this.data.year.listDom.innerHTML = '';
        this.data.year.list = [];
        //创建15条年份

        for (let k = 0; k < 15; k++) {
            let year = index || index === 0 ? index + k : panel.year - 7 + k,
                obj = {
                    //-7是让选中的年份处在面板中间
                    wrapperDom: axAddElem('li', '', `<span content> <i label>${year}年</i></span> `),
                    value: year,
                }
            obj.contentDom = obj.wrapperDom.querySelector('[content]');
            obj.labelDom = obj.wrapperDom.querySelector('[label]');

            this.data.year.listDom.appendChild(obj.wrapperDom);
            this.data.year.list.push(obj);
        }
        //设置selected状态
        this.data.year.list.forEach(k => {
            if (panel.year === k.value) {
                this.selectedAttrToggle(k, true);
            } else {
                this.selectedAttrToggle(k, false);
            }
            //点击翻页
            k.contentDom.onclick = () => {
                if (k.disabled) {
                    return false;
                }
                if (panel.year !== k.value) {
                    //该面板的月份等于点击的月份则不翻页
                    let startValue = k.value;
                    if (this.data.type === 'month') {
                        //month模式需要减掉索引差值
                        startValue = k.value - offset;
                    } else if (this.data.type.includes('date')) {
                        //偏移值=第一个子面板的year值与当前面板year相减
                        startValue = k.value - (this.data.panels[offset].year - this.data.panels[0].year);
                    }
                    this.data.startDate.setFullYear(startValue);
                    this.createPanels('', true);
                    //关闭
                    axHide(this.data.year.panelDom);
                }
            }
        });

    }
    //跳转日期
    goto(dateObj, setSelected = false, callback) {
        //销毁了就不再执行
        if (this.destroyed) {
            return false;
        }

        let date = axDateMethod.toDate(dateObj, this.data.type);
        if (setSelected) {
            //set会自动判断是否需要更新面板
            this.set(date)
        } else {
            this.createPanels(date);
        }
        //回调
        callback && callback.call(this);
    }
    //偏离时间
    offset(dateObj, callback) {
        //销毁了就不再执行
        if (this.destroyed) {
            return false;
        }
        //dateObj格式为{YYYY:'',MM:''...}
        if (this.data.selectedDates.length > 0) {
            //有selected项

            let dates = this.data.selectedDates.map(k => axDateMethod.toObj(axDateMethod.getChain(k, dateObj), false, this.data.type));
            this.set(dates);
        } else {
            //没有selected项，以今天为startDate
            this.data.startDate = axDateMethod.getChain(axDateMethod.toObj(new Date(), false, this.data.type), dateObj);
            this.set(this.data.startDate);
        }
        //回调
        callback && callback.call(this);
    }
    //对日期对象进行selected和unselected切换
    selectedAttrToggle(obj, state) {
        if (state) {
            //设置为selected状态
            if (!obj.selected) {
                Reflect.set(obj, 'selected', true);
                obj.wrapperDom.setAttribute('selected', '');
            }
        } else {
            if (obj.selected) {
                Reflect.deleteProperty(obj, 'selected');
                obj.wrapperDom.removeAttribute('selected');
            }
        }
    }
    //设置日期
    setSelected(dateObj, curGrid, curGrids) {
        //dateObj = {YYYY:'',MM:''...}或{value:{},wrapperDom:''}
        let curObj = this.findFromPanels(dateObj, curGrid, curGrids),
            curItems = curObj.list,
            curItem = curObj.target,
            fillTime = (dateObj) => {
                let obj = {};
                //如果data.time有值，那么selectedDate带上时间属性
                if (this.options.mode === 'range') {
                    //range模式有点特别
                    if (this.selectedDatesProxy.proxy.length === 1) {
                        //当前selectedDate有一项，则合并time中第二项value
                        obj = this.data.type.includes('time') ? { ...dateObj, ...this.data.time[1].value } : dateObj;
                    } else {
                        obj = this.data.type.includes('time') ? { ...dateObj, ...this.data.time[0].value } : dateObj;
                    }
                } else {
                    if (this.data.time.length > 0) {
                        obj = this.data.type.includes('time') ? { ...dateObj, ...this.data.time[0].value } : dateObj;
                    } else {
                        obj = dateObj;
                    }
                }
                return obj;
            },
            selectedFun = (obj) => {
                //value={YYYY:'',MM:'',DD:''}或{YYYY:'',MM:'',DD:'',hh:'',mm:'',ss:''}（daytime格式）
                let value;
                if (this.data.type.includes('time')) {
                    //daytime或time类型需要补齐时间属性
                    if (dateObj.hasOwnProperty('hh') && dateObj.hasOwnProperty('mm') && dateObj.hasOwnProperty('ss')) {
                        //传入的dateObj已经是完整的日期对象
                        value = dateObj;
                    } else {
                        //缺hh、mm和ss才从data.time中取来时间
                        value = fillTime(obj.value);
                    }
                } else {
                    //year和month类型不需要补齐时间
                    value = dateObj.value || dateObj;
                }


                if (!obj.disabled) {
                    if (this.options.mode === 'range') {
                        //区间
                        let length = this.selectedDatesProxy.proxy.length;
                        if (length < 2) {
                            //selectedDate为空或有一项则追加
                            this.selectedDatesProxy.proxy.push(value);
                            if (length === 1) {
                                //此时length已经是2了，所以形成了区间
                                //大小排序，如果逆序的则对调date部分，time部分保持不变
                                let rangeStart = { ...this.data.selectedDates[0] },
                                    rangeEnd = { ...this.data.selectedDates[1] };
                                if (axDateMethod.toDate(rangeStart) > axDateMethod.toDate(rangeEnd)) {
                                    //日期颠倒了，需要重排
                                    this.selectedDatesProxy.proxy.sort((a, b) => axDateMethod.toDate(a) - axDateMethod.toDate(b));
                                    //重新合并时间
                                    this.assignDaytime();
                                }
                                this.setRange(this.data.selectedDates);
                            }
                        } else {
                            //先清理panel中的selected项
                            this.clear(false);
                            //selectedDate有两项以上，则清空后追加
                            this.selectedDatesProxy.proxy.splice(0, this.selectedDatesProxy.proxy.length, value);
                        }
                    } else {
                        if (this.options.mode === 'single') {
                            //单选
                            //如果日期已经是selected状态则不再执行
                            if (this.inSelectedDate(value)) {
                                return false;
                            };
                            //先清理panel中的selected项
                            this.clear(false);
                            this.selectedDatesProxy.proxy.splice(0, this.selectedDatesProxy.proxy.length, value);
                        } else if (this.options.mode === 'multiple') {
                            //多选
                            !this.inSelectedDate(value) ? this.selectedDatesProxy.proxy.push(value) : null;
                        }
                    }
                    //追加属性
                    this.selectedAttrToggle(obj, true);
                }
            };
        if (!curItems.includes(dateObj) && !curItem) {
            //不是本月份日期也没有展示则更新所有面板
            //重新生成面板
            this.createPanels(dateObj.hasOwnProperty('value') ? dateObj.value : dateObj);
            //更新当前的grid
            curItem = this.findFromPanels(dateObj).target;
        }
        //让grid变成selected状态
        !curItem.selected ? selectedFun(curItem) : null;
        //显示事件
        this.options.event.enable && curItem ? this.showEvent(curItem) : null;
    }
    setUnselected(dateObj, curGrid, curGrids) {
        if (axIsEmpty(dateObj)) {
            return false;
        }
        let curItems = curGrids || this.getCurGrids(),
            curItem = this.findFromPanels(dateObj, curGrid, curItems).target,
            selectedDates = [];
        if (curItem) {
            this.selectedAttrToggle(curItem, false);
        }

        //取消selected则更新curDate
        selectedDates = this.data.selectedDates.filter(k => !axDateMethod.isSameDay(k, dateObj.hasOwnProperty('value') ? dateObj.value : dateObj));
        this.selectedDatesProxy.proxy.splice(0, this.data.selectedDates.length, ...selectedDates);

        //区间模式则删除start、end和contain属性
        if (this.options.mode === 'range' && this.data.selectedDates.length < 2) {
            curItems.forEach(k => {
                this.removeRangeAttr(k)
            });
        }
    }
    /*     setDisabled(dateObj) {
            if (axIsEmpty(dateObj)) {
                return false;
            }
    
        } */
    setRangeAttr(grids) {
        //要求选中项为2个
        if (this.data.selectedDates.length !== 2) {
            return false;
        }
        let curGrids = grids || this.getCurGrids(),
            //datetime类型转为date类型
            type = this.data.type === 'datetime' ? 'date' : this.data.type;
        //设置区间
        curGrids.forEach(k => {
            //先循环全部取消start，end和contain样式
            this.removeRangeAttr(k);
            //取消selected状态
            this.selectedAttrToggle(k, false);
            if (axDateMethod.toDate(k.value, type) > axDateMethod.toDate(this.data.selectedDates[0], type) && axDateMethod.toDate(k.value, type) < axDateMethod.toDate(this.data.selectedDates[1], type)) {
                k.wrapperDom.setAttribute('contain', '');
            }
        });

        //设置起点和终点
        let gridStart = this.findFromPanels(this.data.selectedDates[0], '', curGrids).target,
            gridEnd = this.findFromPanels(this.data.selectedDates[1], '', curGrids).target;
        if (gridStart) {
            gridStart.wrapperDom.setAttribute('start', '');
            //追加selected状态
            this.selectedAttrToggle(gridStart, true);
        }
        if (gridEnd) {
            gridEnd.wrapperDom.setAttribute('end', '');
            //追加selected状态
            this.selectedAttrToggle(gridEnd, true);
        }
    }
    removeRangeAttr(dateObj) {
        if (dateObj.wrapperDom.hasAttribute('start')) {
            dateObj.wrapperDom.removeAttribute('start');
        }
        if (dateObj.wrapperDom.hasAttribute('end')) {
            dateObj.wrapperDom.removeAttribute('end');
        }
        if (dateObj.wrapperDom.hasAttribute('contain')) {
            dateObj.wrapperDom.removeAttribute('contain');
        }
    }
    //根据极限日期限制范围
    limitRange(dateObj, type, prop) {
        if (!this.options.minDate || !this.options.maxDate || dateObj.disabled) {
            return false;
        }
        let valueDate = axDateMethod.toDate(dateObj.value, type),
            minDate = axDateMethod.toDate(this.options.minDate, type),
            maxDate = axDateMethod.toDate(this.options.maxDate, type),
            condition,
            setFun = (obj) => {
                if (condition) {
                    obj.disabled = true;
                    obj.wrapperDom.setAttribute('disabled', '');
                }
            };
        //获得判断结果
        if (type === 'date' || type === 'month') {
            condition = (this.options.minDate && valueDate < minDate) || (this.options.maxDate && valueDate > maxDate);
        } else if (type === 'year') {
            condition = (this.options.minDate && dateObj.value.YYYY < ~~this.options.minDate) || (this.options.maxDate && dateObj.value.YYYY > ~~this.options.maxDate);
        } else if (type === 'time') {
            minDate = axDateMethod.toObj(this.options.minDate, false, type);
            maxDate = axDateMethod.toObj(this.options.maxDate, false, type);
            condition = dateObj.value < minDate[prop] || dateObj.value > maxDate[prop];
        }
        //设置disabled
        setFun(dateObj);
    }
    //获得日期值{YYYY:'',MM:'',DD:''}，无hh、mm和ss
    getDateValue(dateObj, type) {
        //dateObj可以是{value:'',label:''}，{YYYY:'',MM:''}或数字索引，最终返回{YYYY:'',MM:''}
        let objType = axType(dateObj),
            value = {};
        if (objType === 'Object') {
            //{value:'',label:''}或{YYYY:'',MM:''}
            if (dateObj.hasOwnProperty('value')) {
                //是列表的单元格对象{value:'',label:''}
                value = dateObj.value;
            } else {
                //是{YYYY:'',MM:''}则直接返回
                value = dateObj;
            }
        } else if (objType === 'Number') {
            //是索引则从选中项中获得
            value = this.data.selectedDates[dateObj];
        } else if (objType === 'String' || objType === 'Date') {
            //如果是日期字符串或Date类型，则转成{YYYY:'',MM:''}对象
            value = axDateMethod.toObj(dateObj, false, this.data.type);
        }
        //复制对象并抛弃hh、mm和ss，因为面板中的单元格值没有这三个属性
        let { hh, mm, ss, ...rest } = value;

        //year和month模式值有不同
        if (type === 'year') {
            Reflect.deleteProperty(rest, 'DD');
            Reflect.deleteProperty(rest, 'MM');
        } else if (type === 'month') {
            Reflect.deleteProperty(rest, 'DD');
        }
        return rest;
    }
    //获得当前网格的所有单元格对象
    getCurGrids() {
        return this.data.panels.map(k => k.list).flat(Infinity).filter(k => !k.exceed);
    }
    //从对象或索引找到日期
    findFromPanels(dateObj, grid, grids) {
        let value = this.getDateValue(dateObj, this.data.type),
            curItems = grids || this.getCurGrids(),
            curItem = grid || curItems.find(k => JSON.stringify(k.value) == JSON.stringify(value));
        return { target: curItem, list: curItems };
    }
    //渲染完成之后执行某些事件
    renderFinish() {
        //time类型不操作以下内容
        if (this.data.type !== 'time') {
            //当前展示的网格
            let curGrids = this.getCurGrids();
            this.data.panels.forEach(o => {
                //年月日面板

                o.list.forEach(k => {
                    k.contentDom.onclick = () => {
                        if (k.disabled) {
                            return false;
                        }
                        if (this.data.selectedDates.length > this.options.maxSelection) {
                            //数量大于max则不能再selected了
                            console.warn(`The selected items have exceeded the ${this.options.maxSelection} limit!`);
                            return false;
                        }
                        //从网格面板中找到对应项
                        let curItem = curGrids.find(i => JSON.stringify(i.value) == JSON.stringify(k.value));
                        if (this.options.mode === 'multiple') {
                            //多选模式
                            if (curItem) {
                                //可toggle切换selected
                                curItem.selected ? this.setUnselected(k, curItem, curGrids) : this.setSelected(k, curItem, curGrids);
                            } else {
                                //不在当前面板，则切换面板并selected
                                this.setSelected(k, curItem, curGrids);
                            }
                        } else {
                            //单选模式和区间模式
                            this.setSelected(k, curItem, curGrids);
                        }
                        //监听回调
                        this.handlers.hasOwnProperty('set') ? this.emit('set', k) : null;
                        this.options.onSet && this.options.onSet.call(this, k);
                    }
                    //如果启用了event并且使用popup方式展示event则
                    if (this.detailIns) {
                        k.contentDom.onmouseenter = () => {
                            //更新事件内容
                            this.showEvent(k);
                            //更新事件popup定位
                            this.detailIns.updatePosition(k.contentDom).show();
                        }
                    }


                })

            });


            //根据data.selectedDate输出到result盒子里
            //删除tag后同步取消日期selected状态
            if (this.resultIns) {
                this.resultIns.on('remove', (removeItems, indexs) => {
                    //删除，根据索引找对日期项
                    let curGrids = this.getCurGrids(),
                        curGrid = this.findFromPanels(indexs[0], '', curGrids).target;
                    //取消selected状态并且从selectedDate中删掉
                    if (curGrid) {
                        this.selectedAttrToggle(curGrid, false);
                    }
                    //删除对应的selectedDate
                    this.selectedDatesProxy.proxy.splice(indexs[0], 1);

                    //区间模式下，取消range属性
                    curGrids.forEach(k => {
                        this.removeRangeAttr(k);
                    });

                });
                //监听update
                this.resultIns.on('update', () => {
                    //点击tag更新面板
                    this.resultIns.content.forEach(k => {
                        k.dom.onclick = (e) => {
                            //点击区域不是close标签才更新面板
                            if (!e.target.hasAttribute('close')) {
                                //当前面板没有该日期才更新
                                if (!this.findFromPanels(k.value).target) {
                                    this.createPanels(k.value);
                                }
                            }
                        }
                    });
                });

            }

            //年月前后翻页导航
            if (this.prevYearDom) {
                this.prevYearDom.onclick = () => {
                    let years = -1;
                    if (this.data.type === 'year') {
                        years = -18 * this.options.cols * this.options.rows;
                    } else if (this.data.type === 'month') {
                        years = -this.options.cols * this.options.rows;
                    }
                    this.createPanels(axDateMethod.getChain(this.data.startDate, { YYYY: years }), true);
                }
            }
            if (this.nextYearDom) {
                this.nextYearDom.onclick = () => {
                    let years = 1;
                    if (this.data.type === 'year') {
                        years = 18 * this.options.cols * this.options.rows;
                    } else if (this.data.type === 'month') {
                        years = this.options.cols * this.options.rows;
                    }
                    this.createPanels(axDateMethod.getChain(this.data.startDate, { YYYY: years }), true);
                }
            }

            if (this.data.type.includes('date')) {
                //只有day和daytime模式才能使用月导航
                this.prevMonthDom.onclick = () => {
                    this.createPanels(axDateMethod.getChain(this.data.startDate, { MM: -1 }), true);
                }
                this.nextMonthDom.onclick = () => {
                    this.createPanels(axDateMethod.getChain(this.data.startDate, { MM: 1 }), true);
                }


            }

            //打开年和月选择面板
            this.data.panels.forEach((k, i) => {
                //年
                if (k.yearDom) {
                    k.yearDom.onclick = () => {
                        this.data.year.panelDom.style.top = k.panelDom.offsetTop + k.headerDom.clientHeight + 'px';
                        this.data.year.panelDom.style.left = k.panelDom.offsetLeft + 'px';

                        //month模式下没有month.panelDom
                        if (this.data.month.panelDom) {
                            //如果月选择面板可见则折叠
                            if (axIsVisible(this.data.month.panelDom)) {
                                axHide(this.data.month.panelDom);
                            }
                        }

                        //填充年份列表并设置selected状态
                        this.createYearList(k, '', i);

                        //折叠和展开
                        axToggle(this.data.year.panelDom);
                    }
                    //列表翻页和检索
                    this.data.year.prevDom.onclick = () => {
                        this.createYearList(k, this.data.year.list[0].value - 15, i);
                    }
                    this.data.year.nextDom.onclick = () => {
                        this.createYearList(k, this.data.year.list[14].value + 1, i);
                    }
                    this.data.year.goDom.onclick = () => {
                        this.searchYearToRefresh(this.data.year.inputDom);
                    }
                    //使用回车等于click
                    this.data.year.inputDom.onkeyup = (e) => {
                        if (e.code === 'Enter') {
                            this.searchYearToRefresh(this.data.year.inputDom);
                        }
                    }

                }
                //月
                if (k.monthDom) {
                    k.monthDom.onclick = () => {
                        this.data.month.panelDom.style.top = k.panelDom.offsetTop + k.headerDom.clientHeight + 'px';
                        this.data.month.panelDom.style.left = k.panelDom.offsetLeft + 'px';
                        //如果年选择面板可见则折叠
                        if (axIsVisible(this.data.year.panelDom)) {
                            axHide(this.data.year.panelDom);
                        }
                        //设置selected状态
                        this.data.month.list.forEach(o => {
                            if (k.month === o.value) {
                                this.selectedAttrToggle(o, true);
                            } else {
                                this.selectedAttrToggle(o, false);
                            }
                            //点击翻页
                            o.contentDom.onclick = () => {
                                if (k.month !== o.value) {
                                    //该面板的月份等于点击的月份则不翻页
                                    this.data.startDate.setMonth(o.value - i);
                                    this.createPanels('', true);
                                    //关闭
                                    axHide(this.data.month.panelDom);
                                }
                            }
                        });
                        //折叠和展开
                        axToggle(this.data.month.panelDom);
                    }

                }
            });

            //初始化日期selected
            this.data.selectedDates.forEach((k, i) => {
                let item = this.findFromPanels(k, '', curGrids).target;
                if (item) {
                    this.selectedAttrToggle(item, true);
                    //显示事件，多个selected只选第一个

                    if (i === 0 && this.options.event.enable) {
                        this.showEvent(item);
                    }
                }
            });

            //年份检索
            if (this.yearSearchDom) {
                let yearSearch = this.yearSearchDom.querySelector('[go]'),
                    yearInput = this.yearSearchDom.querySelector('input')

                yearSearch.onclick = () => {
                    this.searchYearToRefresh(yearInput, false);
                }
                //使用回车等于click
                yearInput.onkeyup = (e) => {
                    if (e.code === 'Enter') {
                        this.searchYearToRefresh(yearInput, false);
                    }
                }
            }

            //初始化时间selected
            if (this.timePanelsDom) {
                //根据this.data.selectedDates设置时间
                if (this.options.mode === 'range') {
                    if (this.data.selectedDates.length === 0) {
                        this.setTime([{ hh: 0, mm: 0, ss: 0 }, { hh: 0, mm: 0, ss: 0 }]);
                    } else if (this.data.selectedDates.length === 1) {
                        this.setTime([this.data.selectedDates[0], { hh: 0, mm: 0, ss: 0 }]);
                    } else {
                        this.setTime(this.data.selectedDates);
                    }
                } else {
                    if (this.data.selectedDates.length === 0) {
                        this.setTime({ hh: 0, mm: 0, ss: 0 });
                    } else {
                        this.setTime(this.data.selectedDates[0]);
                    }
                }
                //打开时间选择面板(非time类型才打开)
                if (this.timeBtnDom) {
                    this.timeBtnDom.onclick = () => {
                        axSlideToggle(this.timePanelsDom, 'inline-grid', () => {
                            //selected项在最前面，并且prevent滚动
                            this.openTimeToPrevent();
                        });
                    }
                }
            }
        }

        //前提是有footer才可使用按钮点击功能
        if (this.options.footerShow) {
            //点击确定按钮
            this.confirmDom.onclick = () => {
                this.confirm();
                //
            }
            //点击现在按钮
            if (this.nowDom) {
                this.nowDom.onclick = () => {
                    this.now();
                }
            }

            //点击清除按钮
            if (this.clearDom) {
                this.clearDom.onclick = () => {
                    this.clear(true);
                }
            }
            //点击清除并关闭按钮，附带清空input值
            if (this.cancelDom) {
                this.cancelDom.onclick = () => {
                    this.bubbleIns.hide();
                    this.fillInput('');
                    this.clear(true);
                }
            }
        }


        //time类型分inline和气泡展示方式，气泡展示方式需要在shown之后才执行prevent
        if (this.data.type.includes('time') && this.data.time.length > 0) {
            let timeCols = this.data.time.map(k => k.colsDom).flat(Infinity);
            if (this.options.display === 'inline') {
                this.openTimeToPrevent(timeCols);
            } else {
                if (this.bubbleIns) {
                    this.bubbleIns.on('shown', () => {
                        this.openTimeToPrevent(timeCols);
                    });
                }
            }
        }

        //整个容器没有定义宽度，会导致result盒子内容不断向右延伸，所以根据panels宽度设定result的maxwidth
        if (this.bubbleIns) {
            this.bubbleIns.on('shown', () => {
                this.setResultWidth();
            });
        } else {
            this.setResultWidth();
        }

        //保存数据
        this.save();
    }
    setResultWidth() {
        //整个容器没有定义宽度，会导致result盒子内容不断向右延伸，所以根据panels宽度设定result的maxwidth
        if (this.resultDom) {
            this.resultDom.style.maxWidth = this.panelsDom.getBoundingClientRect().width + 'px';
        }
    }
    openTimeToPrevent(cols) {
        let timeCols = cols || this.data.time.map(k => k.colsDom).flat(Infinity);
        timeCols.forEach(k => {
            this.showForemost(k);
        });
    }
    searchYearToRefresh(input, hide = true) {
        //要求input不为空且不大于4位数字
        let value = input.value;
        if (value && !isNaN(value) && (value.length === 3 || value.length === 4)) {
            //更新年份并重新生成网格面板
            this.data.startDate.setFullYear(~~value);
            this.createPanels();
            //关闭
            hide ? axHide(this.data.year.panelDom) : null;
        } else {
            new axMessage({
                result: 'error',
                iconShow: true,
                content: this.options.language.message.requireYearFormat,
            }).show();
        }
    }
    createPanelDayList(elem, dateObj) {
        //先初始化清空
        elem.innerHTML = '';
        let curDate = axDateMethod.toDate(dateObj, this.data.type);
        //设为月份的1日
        curDate.setDate(1);
        let list = [],
            startIndex,
            curMonth = curDate.getMonth(),
            curYear = curDate.getFullYear(),
            //获得当前月份的天数，因为月份是0~11，new Date(2022, 11, 0).getDate()显示10月份最后一天也就是10月份的天数
            monthDays = axDateMethod.getDays(curDate),
            preDate = axDateMethod.getChain(curDate, { MM: -1 }),
            prevMonth = preDate.getMonth(),
            prevYear = preDate.getFullYear(),
            nextDate = axDateMethod.getChain(curDate, { MM: 1 }),
            nextMonth = nextDate.getMonth(),
            nextYear = nextDate.getFullYear(),
            fragment = document.createDocumentFragment();
        //确定startIndex
        if (this.options.weekStart === 'monday') {
            //星期排序：周一~周日
            if (curDate.getDay() === 0) {
                //周日
                startIndex = 7;
            } else {
                startIndex = curDate.getDay();
            }
        } else {
            //星期排序：周日~周六
            if (curDate.getDay() === 0) {
                //周日
                startIndex = 1;
            } else {
                startIndex = curDate.getDay() + 1;
            }
        }

        let day = 1;
        //42=6*7，网格是六行七列
        for (let gridIndex = 1; gridIndex <= 42;) {
            let dayObj = {
                wrapperDom: axAddElem('li'),
                contentDom: axAddElem('span', { content: '' }),
                labelDom: axAddElem('i', { label: '' }),
            }, label;
            dayObj.wrapperDom.appendChild(dayObj.contentDom);
            dayObj.contentDom.appendChild(dayObj.labelDom);
            if (gridIndex < startIndex) {
                //填充上一个月
                label = axDateMethod.getDays({ YYYY: prevYear, MM: prevMonth }) - startIndex + gridIndex + 1;
                dayObj.value = { YYYY: prevYear, MM: prevMonth, DD: label };
                //修改day属性
                dayObj.exceed = true;
                dayObj.wrapperDom.setAttribute('exceed', '');
            } else if (gridIndex >= startIndex + monthDays) {
                //填充下一个月
                label = gridIndex - startIndex - monthDays + 1;
                dayObj.value = { YYYY: nextYear, MM: nextMonth, DD: label };
                //修改day属性
                dayObj.exceed = true;
                dayObj.wrapperDom.setAttribute('exceed', '');
            } else {
                //填充本月
                label = day;
                dayObj.value = { YYYY: curYear, MM: curMonth, DD: label };
                //递增月天数
                day++;
            }
            //追加label和labelDom属性
            dayObj.label = label;
            dayObj.labelDom.innerText = label;

            //判断minDate和maxDate
            this.limitRange(dayObj, 'date');

            //追加农历
            if (this.options.lunar && this.options.lunar.handler) {
                //追加农历日期和节气，农历插件要求是真实的年月日，月份要+1
                let lunarObj = this.options.lunar.handler(dayObj.value.YYYY, dayObj.value.MM + 1, dayObj.value.DD);
                if (lunarObj && lunarObj != -1) {
                    let lunarContent = '', attr = '';
                    for (let v of this.options.lunar.arrange) {
                        //按顺序显示农历内容，通常是节日优先
                        if (this.options.lunar.map.hasOwnProperty(v) && lunarObj[this.options.lunar.map[v]]) {
                            lunarContent = lunarObj[this.options.lunar.map[v]];
                            //lunarDom追加属性
                            if (v.includes('Festival')) {
                                attr = 'festival';
                            } else if (v === 'term') {
                                attr = v;
                            }
                            break;
                        }
                    }
                    dayObj.lunarDom = axAddElem('u', '', lunarContent);
                    attr ? dayObj.lunarDom.setAttribute(attr, '') : null;
                    dayObj.labelDom.insertAdjacentElement('afterend', dayObj.lunarDom);
                }
                //显示标记
                if (this.options.lunar.tags.length > 0) {
                    for (let v of this.options.lunar.tags) {
                        let date = axDateMethod.toObj(v.date, false, this.data.type),
                            tagDom = axAddElem('s', v.props, v.content);
                        if (JSON.stringify(date) == JSON.stringify(dayObj.value)) {
                            //日期上追加文字标记
                            dayObj.contentDom.appendChild(tagDom);
                            break;
                        }
                    }

                }
            }
            //追加事件指示
            if (this.options.event.enable) {
                for (let v of this.options.event.list) {
                    let date = axDateMethod.toObj(v.date, false, this.data.type);
                    if (JSON.stringify(date) == JSON.stringify(dayObj.value)) {
                        //日期上追加小点示意
                        dayObj.busy = true;
                        dayObj.wrapperDom.setAttribute('busy', '');
                        break;
                    }
                }
            }


            //递增网格索引
            gridIndex++;
            //添加到碎片
            fragment.appendChild(dayObj.wrapperDom);
            //追加到数组
            list.push(dayObj);
        }
        //批量渲染
        elem.appendChild(fragment);
        //返回数组
        return list;
    }
    createPanelMonthList(elem, dateObj) {
        let list = [];
        //根据子li节点创建月份对象
        [...elem.children].forEach((k, i) => {
            let obj = {
                wrapperDom: k,
                contentDom: k.querySelector('[content]'),
                value: { YYYY: dateObj.getFullYear(), MM: i }
            };
            //判断minDate和maxDate
            this.limitRange(obj, 'month');
            list.push(obj);
        });
        //返回数组
        return list;
    }
    createPanelYearList(elem, dateObj) {
        let list = [],
            fragment = document.createDocumentFragment();
        //一个子面板显示18个年份
        for (let k = 0; k < 18; k++) {
            let year = dateObj.getFullYear() + k,
                wrapper = axAddElem('li', '', `<span content><i label>${year}年</i></span>`),
                obj = {
                    wrapperDom: wrapper,
                    value: { YYYY: year },
                }
            obj.contentDom = obj.wrapperDom.querySelector('[content]');
            //判断minDate和maxDate
            this.limitRange(obj, 'year');
            list.push(obj);
            fragment.appendChild(wrapper);
        }
        elem.appendChild(fragment);
        return list;
    }
    createLoopStr(max) {
        let str = '<li>';
        for (let n = 0; n < max; n++) {
            str += `<div> <span content><i label>${axDateMethod.fillZero(n)}</i></span></div> `
        }
        str += '</li>'
        return str;
    }
    createTimeList(label) {
        return `
    <li panel = "time">
                        <span tools><i restore title='${this.options.language.toolTip.restore}'></i><i reset title='${this.options.language.toolTip.reset}'></i><i now title='${this.options.language.toolTip.now}'></i><i hide title='${this.options.language.toolTip.close}'></i></span>
                        <div header>
                            <span time>${label}</span>
                        </div>
                        <div body>
                            <ul column>
                                <li>${this.options.language.unit.h}</li>
                                <li>${this.options.language.unit.m}</li>
                                <li>${this.options.language.unit.s}</li>
                            </ul>
                            <ul list>
                            ${this.createLoopStr(24)}
                            ${this.createLoopStr(60)}
                            ${this.createLoopStr(60)}
                            </ul>
                        </div>
                    </li>
    `;
    }
    toggleTimeList(item, list) {
        this.selectedAttrToggle(item, true);
        let selectedItem = list.find(k => k != item && k.selected);
        if (selectedItem) {
            this.selectedAttrToggle(selectedItem, false);
        }
    }
    getTimeListArr(parent, type) {
        let children = parent.children,
            arr = [];
        [...children].forEach((k, i) => {
            let obj = {
                wrapperDom: k,
                contentDom: k.querySelector('[content]'),
                value: i
            };
            //通过格式设置禁用整列
            if (!this.options.format.includes(type)) {
                obj.disabled = true;
                obj.wrapperDom.setAttribute('disabled', '');
            }
            //根据minDate和maxDate设置禁用
            this.limitRange(obj, 'time', type);

            arr.push(obj);
        });
        return arr;
    }
    //让时间的selected项在最前面显示
    showForemost(parent) {
        let type = axType(parent),
            fun = (elem) => {
                let selectedItem = elem.querySelector('[selected]'),
                    toTop = selectedItem.getBoundingClientRect().height * Math.floor((~~selectedItem.innerText) / 2);
                //不再父层窗口内才滚动
                if (selectedItem && !axChildVisible(selectedItem, elem)) {
                    elem.scrollTop = toTop;
                }
            };
        if (type === 'Array') {
            parent.forEach(k => {
                fun(k);
            });
        } else {
            fun(parent);
        }
    }
    createTimeSelectPanel() {
        this.timePanelsDom = axAddElem('ul', { panels: 'hms' });
        if (this.options.mode === 'range') {
            this.timePanelsDom.innerHTML = this.createTimeList(this.options.language.time.start) + this.createTimeList(this.options.language.time.end);
        } else {
            this.timePanelsDom.innerHTML = this.createTimeList(this.options.language.time.select);
        }
        //存入time属性
        [...this.timePanelsDom.querySelectorAll('[panel]')].forEach(k => {
            let liArr = [...k.querySelectorAll('[list] li')],
                panelObj = {
                    panelDom: k,
                    colsDom: liArr,
                    restoreDom: k.querySelector('[restore]'),
                    resetDom: k.querySelector('[reset]'),
                    nowDom: k.querySelector('[now]'),
                    closeDom: k.querySelector('[hide]'),
                    value: {},
                    hh: this.getTimeListArr(liArr[0], 'hh'),
                    mm: this.getTimeListArr(liArr[1], 'mm'),
                    ss: this.getTimeListArr(liArr[2], 'ss'),
                };
            this.data.time.push(panelObj);
        });
        //写入页面节点
        this.footerDom.insertAdjacentElement('beforebegin', this.timePanelsDom);

        //面板操作
        this.data.time.forEach((k, n) => {
            //关闭面板
            k.closeDom.onclick = () => {
                axSlideUp(this.timePanelsDom, 'inline-grid');
            }
            //复原
            k.restoreDom.onclick = () => {
                //没有点确定按钮，selectedDate值没有变
                if (this.raw[n]) {
                    //可能此时data.selectedDate为空或者不能与data.time的length对应
                    this.setTime(this.raw[n], n, true);

                }
            }
            //归零
            k.resetDom.onclick = () => {
                this.setTime({ hh: 0, mm: 0, ss: 0 }, n, true);
            }
            //现在时间
            k.nowDom.onclick = () => {
                let time = axDateMethod.toObj(new Date(), false, this.data.type);
                this.setTime({ hh: time.hh, mm: time.mm, ss: time.ss }, n, true);
            }
            //点击网格hh、mm、ss列表单选
            k.hh.forEach(i => {
                i.contentDom.onclick = () => {
                    if (i.selected || i.disabled) {
                        return false;
                    }
                    this.toggleTimeList(i, k.hh);
                    k.value.hh = i.value;
                    //更新time按钮并联动更新data.selectedDate
                    this.timeToUpdate(k, n, true);
                    //监听回调
                    this.handlers.hasOwnProperty('setTime') ? this.emit('setTime', i) : null;
                    this.options.onSetTime && this.options.onSetTime.call(this, i);
                }
            });
            k.mm.forEach(i => {
                i.contentDom.onclick = () => {
                    if (i.selected || i.disabled) {
                        return false;
                    }
                    this.toggleTimeList(i, k.mm);
                    k.value.mm = i.value;
                    //更新time按钮并联动更新data.selectedDate
                    this.timeToUpdate(k, n, true);
                    //监听回调
                    this.handlers.hasOwnProperty('setTime') ? this.emit('setTime', i) : null;
                    this.options.onSetTime && this.options.onSetTime.call(this, i);
                }
            });
            k.ss.forEach(i => {
                i.contentDom.onclick = () => {
                    if (i.selected || i.disabled) {
                        return false;
                    }
                    this.toggleTimeList(i, k.ss);
                    k.value.ss = i.value;
                    //更新time按钮并联动更新data.selectedDate
                    this.timeToUpdate(k, n, true);
                    //监听回调
                    this.handlers.hasOwnProperty('setTime') ? this.emit('setTime', i) : null;
                    this.options.onSetTime && this.options.onSetTime.call(this, i);
                }
            });
        });

    }
    timeToUpdate(parent, index, updateHTML = false) {
        //更新timeselect按钮的html
        updateHTML ? this.setTimeBtnHTML() : null;
        //如果是单选模式和区间模式，可修改值
        if (this.data.type !== 'time' && (this.options.mode === 'single' || this.options.mode === 'range')) {
            if (this.data.selectedDates.length >= index + 1) {
                Object.assign(this.selectedDatesProxy.proxy[index], parent.value);
            }
        }
        //如果是time类型，自动填充
        if (this.data.type === 'time' && this.options.autoFill) {
            this.confirm(false);
        }
    }
    setTimeBtnHTML() {
        //不是datetime或time类型，或没有时间按钮则不执行
        if (!this.data.type.includes('time') || !this.timeBtnDom) {
            return false;
        }
        let timeStr = this.data.time.map(k => {
            if (axIsEmpty(k.value)) {
                return this.options.language.time.select;
            } else {
                return `${axDateMethod.fillZero(k.value.hh)}:${axDateMethod.fillZero(k.value.mm)}:${axDateMethod.fillZero(k.value.ss)}`;
            }
        });
        this.timeBtnDom.innerHTML = timeStr.join('~');
    }
    //根据selectedDate或{hh:'',mm:'',ss:''}来设置时间窗
    setTime(timeObj, index = 0, update = false, callback) {
        //为空则不再执行，不是datetime或time类型则不执行
        if (axIsEmpty(timeObj) || !this.data.type.includes('time') || this.destroyed) {
            return false;
        }
        let type = axType(timeObj);
        //仅用于mode=signle或range模式下；timeObj=[{hh:'',mm:'',ss:''}]或{hh:'',mm:'',ss:''}，当为对象的时候需要使用index
        let setFun = (k, i) => {
            let time;
            if (typeof k === 'string') {
                //字符串转对象
                time = axDateMethod.strToTimeObj(timeObj);
            } else {
                time = k;
            }

            //hh、mm、ss的列表项索引等于value中对应的hh、mm、ss值
            this.data.time[i].value.hh = time.hh;
            this.toggleTimeList(this.data.time[i].hh[time.hh], this.data.time[i].hh);
            //分
            this.data.time[i].value.mm = time.mm;
            this.toggleTimeList(this.data.time[i].mm[time.mm], this.data.time[i].mm);
            //秒
            this.data.time[i].value.ss = time.ss;
            this.toggleTimeList(this.data.time[i].ss[time.ss], this.data.time[i].ss);
            //更新time按钮并联动更新data.selectedDate，time类型没有selectedDate值
            update && this.data.type === 'datetime' ? this.timeToUpdate(this.data.time[i], i) : null;
        }
        if (type === 'Array') {
            //是数组
            timeObj.forEach((k, i) => {
                //如果mode=multiple，同时是daytime格式，会导致selectedDate项大于data.time的length，所以进行限制
                if (this.data.time.length > i) {
                    setFun(k, i);
                }
            });
        } else {
            //是对象或字符串
            setFun(timeObj, index);
        }


        //根据this.data.time.value值修改时间按钮的HTML
        this.setTimeBtnHTML();
        //监听回调
        this.handlers.hasOwnProperty('setTime') ? this.emit('setTime', timeObj) : null;
        this.options.onSetTime && this.options.onSetTime.call(this, timeObj);
        callback && callback.call(this);

    }
    //将面板日期和气泡时间合并
    assignDaytime() {
        this.selectedDatesProxy.proxy.forEach((k, i) => {
            if (this.data.time[i]) {
                Object.assign(k, this.data.time[i].value);
            }
        });
    }
    confirm(hide = this.options.confirmToHide, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //合并并填入input
        let result = '';

        if (this.data.type !== 'time') {
            //非time类型

            //if (this.data.selectedDates.length !== 0 || this.data.time.length !== 0) {
            //既没选择日期也没选择时间则不再执行

            //合并day和time，multiple模式不合并（不会将气泡日期合并到日期当中）
            //if (this.options.mode !== 'multiple') {
            // this.assignDaytime();
            //}


            //}
            if (this.options.mode === 'range') {
                if (this.data.selectedDates.length === 0) {
                    //清空了
                    result = '';
                } else if (this.data.selectedDates.length === 2) {
                    //如果是区间则取两个
                    let startStr = axDateMethod.fillFormat(this.data.selectedDates[0], this.options.format),
                        endStr = axDateMethod.fillFormat(this.data.selectedDates[1], this.options.format);
                    result = `${startStr}${this.options.language.range}${endStr} `;
                } else {
                    //不为0和不为2情况，报错并不再执行
                    new axMessage({
                        result: 'error',
                        iconShow: true,
                        content: this.options.language.message.requireTwoValue,
                    }).show();
                    return false;
                }
            } else {
                result = this.data.selectedDates.map(k => axDateMethod.fillFormat(k, this.options.format)).join(this.options.separator);

                //如果启用了强制选择一个日期，result为空则不再执行
                if (this.options.forceSelect && !result) {
                    new axMessage({
                        result: 'error',
                        iconShow: true,
                        content: this.options.language.message.requireOneSelected,
                    }).show();
                    return false;
                }
            }
        } else {
            //time类型
            result = this.data.time.map(k => axDateMethod.fillFormat(k.value, this.options.format)).join(this.separator);
        }
        //将文本传入input节点
        this.fillInput(result);

        //点击确认按钮关闭气泡
        if (hide && this.bubbleIns) {
            this.bubbleIns.hide();
            //关闭自己，再打开其他日期实例
            let nextIns = this.options.next ? axInstance.find(this.options.next, 'date') : null;
            nextIns && nextIns.bubbleIns ? nextIns.bubbleIns.show() : null;
        }

        //监听回调
        this.handlers.hasOwnProperty('confirm') ? this.emit('confirm', result) : null;
        this.options.onConfirm && this.options.onConfirm.call(this, result);
        callback && callback.call(this, result);
    }
    //将值填入input
    fillInput(text) {
        if (this.inputDom) {
            if (this.inputDom.nodeName === 'INPUT') {
                this.inputDom.value = text;
            } else {
                this.inputDom.innerHTML = text;
            }
        }
        //保存数据
        this.save();
    }
    now(hide = this.options.nowToHide) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        let date = axDateMethod.toObj(new Date(), false, this.data.type);
        if (this.data.type === 'time') {
            this.setTime(date);
        } else {
            //非range模式更新时间
            if (this.timePanelsDom && this.options.mode !== 'range') {
                this.setTime(date)
            }
            this.setSelected(date);
        }
        //如果启用点击现在关闭
        if (hide) {
            //赋值input
            this.confirm();
            //关闭气泡
            this.bubbleIns ? this.bubbleIns.hide() : null;
        }
    }
    //保存当前数据
    save(props, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }
        setTimeout(() => {
            if (!props) {
                //保存当前焦点索引
                axLocalStorage.set(this.options.storageName, Object.assign(this.options, { content: this.inputValueToDates(false) }));
            } else {
                //保存参数,prop={}
                !props.hasOwnProperty('content') ? props.content = this.inputValueToDates(false) : null;
                axLocalStorage.set(this.options.storageName, props);
            }
            //回调监听
            let getValue = axLocalStorage.get(this.options.storageName);
            this.handlers.hasOwnProperty('save') ? this.emit('save', getValue) : null;
            this.options.onSave && this.options.onSave.call(this, getValue);
            callback && callback.call(this, getValue);
        }, 0);
        return this;
    }
    removeInstances() {
        //删除节点和销毁实例
        this.targetDom.remove();
        this.bubbleIns ? (this.bubbleIns.destroy(), this.bubbleIns = null) : null;
        this.detailIns ? (this.detailIns.destroy(), this.detailIns = null) : null;
        this.resultIns ? (this.resultIns.destroy(), this.resultIns = null) : null;
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
    }
    //更新实例参数
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //删除节点和销毁实例
        this.removeInstances();
        //更新参数
        this.options = axExtend(this.options, setting);

        //根据参数初始化实例
        this.init();
        //监听回调
        this.handlers.hasOwnProperty('update') ? this.emit('update', '') : null;
        this.options.onUpdate && this.options.onUpdate.call(this);
        callback && callback.call(this);
        return this;
    }
    //销毁实例
    destroy(callback) {
        //删除节点和销毁实例
        this.removeInstances();
        //取消绑定事件
        this.inputDom.onclick = null;
        this.positionDom ? this.positionDom.onclick = null : null;
        this.buttonDom ? this.buttonDom.onclick = null : null;

        //更新销毁状态
        this.destroyed = true;

        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    setValuesToObjs(data) {
        let dataType = axType(data),
            arr = [];
        if (dataType === 'Array') {
            arr = data;
        } if (dataType === 'Date') {
            //data是new Date()格式
            arr = [data];
        } else if (dataType === 'Object') {
            //data是{YYY:'',MM:''}格式
            arr = [data];
        } else if (dataType === 'String') {
            //data是'2012-2-3'/'2012-2-3,2012-2-4'格式/'2012-2-3至2012-2-4'格式
            let separator;
            if (data.includes(this.options.separator)) {
                separator = this.options.separator;
            } else {
                separator = this.separator;
            }
            arr = data.split(separator);
        }

        if (arr.length === 0) {
            return [];
        } else {
            //转成标准的{YYYY:'',MM:''}格式
            return arr.map(k => axDateMethod.toObj(k, false, this.data.type));
        }
    }
    //设置一个或多个日期
    set(data, callback) {
        //为空则不再执行
        if (axIsEmpty(data) || this.destroyed) {
            return false;
        }
        let objArr = this.setValuesToObjs(data);

        //需要至少一项
        if (objArr.length === 0) {
            return false;
        }

        if (this.options.mode === 'range') {
            //面板中已经选中了1个，那么只能从objArr中取一个，如果面板中未选中则取2个，如果面板中
            if (objArr.length >= 2) {
                objArr.splice(2, objArr.length - 2);
                this.setRange(objArr);
            } else {
                this.setSelected(objArr[0]);
            }
            //设置时间
            this.setTime(objArr, '', true);
        } else {
            if (this.options.mode === 'single') {
                //单选，只取第一项
                this.setSelected(objArr[0]);
                //设置时间
                this.setTime(objArr[0], 0, true);
            } else if (this.options.mode === 'multiple') {
                //多选
                //循环selected
                objArr.forEach(k => {
                    //k可能是date，string或obj，都会转成{}来设置selected状态
                    this.setSelected(k);
                });
            }

        }

        //监听回调
        this.handlers.hasOwnProperty('set') ? this.emit('set', objArr) : null;
        this.options.onSet && this.options.onSet.call(this, objArr);
        callback && callback.call(this, objArr);
        return this;
    }
    setRange(data, callback) {
        //为空或length不为2则不再执行
        if (axIsEmpty(data) || data.length !== 2 || this.destroyed) {
            return false;
        }

        //转{YYYY:'',MM:''}对象数组
        let objArr = data.map(k => axDateMethod.toObj(k, true, this.data.type)),
            curGrids = this.getCurGrids();
        //从小到大排序
        axDateMethod.sort(objArr);
        //修改data.selectedDate
        this.selectedDatesProxy.proxy.splice(0, this.selectedDatesProxy.proxy.length, ...objArr);


        //如果起止日期都不在当前面板则刷新面板，自动重设startDate
        if (!this.inCurPanels(objArr[0], curGrids) && !this.inCurPanels(objArr[1], curGrids)) {
            //createPanels会执行renderFinish，其中会setTime
            this.createPanels(objArr[0]);
        } else {
            //设置时间
            this.setTime(objArr);
        }

        //设置区间start、end和contain属性
        this.setRangeAttr();

        //监听回调
        this.handlers.hasOwnProperty('ranged') ? this.emit('ranged', objArr) : null;
        this.options.onRanged && this.options.onRanged.call(this, objArr);
        callback && callback.call(this, objArr);

        return this;
    }
    /*     disable(data, callback) {
            //为空则不再执行
            if (axIsEmpty(data)) {
                return false;
            }
            let objArr = this.setValuesToObjs(data);
    
            //需要至少一项
            if (objArr.length === 0) {
                return false;
            }
    
            objArr.forEach(k => {
                this.setDisabled(k);
            });
    
            //监听回调
            'disabled' in this.handlers ? this.emit('disabled', objArr) : null;
            this.options.onDisabled && this.options.onDisabled.call(this, objArr);
            callback && callback.call(this, objArr);
            return this;
        } */
    cancel(data, callback) {
        //为空则不再执行
        if (axIsEmpty(data) || this.destroyed) {
            return false;
        }
        let objArr = this.setValuesToObjs(data);

        //需要至少一项
        if (objArr.length === 0) {
            return false;
        }

        objArr.forEach(k => {
            this.setUnselected(k);
        });

        //监听回调
        this.handlers.hasOwnProperty('cancel') ? this.emit('cancel', objArr) : null;
        this.options.onCancel && this.options.onCancel.call(this, objArr);
        callback && callback.call(this, objArr);
        return this;
    }
    clear(resetTime = true, callback) {
        //resetTime是指清除selected时是否同时让time归零，默认归零
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //time类型没有selectedDate，所以需要排除
        if (this.data.type !== 'time') {
            //清空选中项数组
            this.selectedDatesProxy.proxy.splice(0);
            //清空标签结果盒子
            this.resultIns ? this.resultIns.clear() : null;
            //取消selected状态
            let curGrids = this.getCurGrids();
            curGrids.forEach(k => {
                if (k.selected) {
                    this.selectedAttrToggle(k, false);
                }
                this.removeRangeAttr(k);
            });
        }
        //时间归零
        if (resetTime && this.timePanelsDom) {
            for (let k in this.data.time) {
                this.setTime({ hh: 0, mm: 0, ss: 0 }, k);
            }
        }
        //监听回调
        this.handlers.hasOwnProperty('clear') ? this.emit('clear', '') : null;
        this.options.onClear && this.options.onClear.call(this);
        callback && callback.call(this);
        return this;
    }
    //以下三个方法为监听
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
//对有本属性的元素使用本插件
/* document.querySelectorAll('[axDate]').forEach(element => {
    new axDate(element);
}); */
/*!
* 插件：轮滑选择；改造自ion.rangeSlider | https://github.com/IonDen/ion.rangeSlider；使用方法：new axRange('#id',{参数})
*/
class axRange {
    constructor(elem, options) {
        let dom = axIdToDom(elem);
        //无节点则不再执行
        if (!dom) return false;
        if (dom.nodeName === 'INPUT') {
            //input节点，则新建target
            this.inputDom = dom;
            this.inputDom.type = 'hidden';
            this.targetDom = axAddElem('div', { class: 'ax-range' });
            dom.insertAdjacentElement('beforebegin', this.targetDom);
            this.targetDom.appendChild(dom);
        } else {
            //div等节点则将之设为target
            this.targetDom = dom;
            this.targetDom.classList.add('ax-range');
            this.inputDom = axAddElem('input', { type: 'hidden' });
            this.targetDom.appendChild(this.inputDom);
        }
        //合并参数
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//存储名称，字符串格式；如果需要保存实例当前数据请设定唯一的值，比如：storageName:'mystorage001'
            name: '',//input控件的name，如果需要真实提交表单请务必填写
            step: 1,//步进，滑块值变化的最小单位
            max: 100,//最大值，默认为100，可为任意大于min的数值
            min: 0,//最小值，默认为0，可为任意小于max的数值
            value: 0,//初始值，默认为0，如果multiple为true则需要填写数组，例如[0,60]
            multiple: false,//是否允许两个滑块自由滑动来设定区间，默认false即一个滑块
            locked: false,//如果multiple为true，是否锁定from和to间距，默认false不锁定
            limitShow: true,//是否头尾显示极限值气泡（min和max），默认true显示
            tipShow: true,//是否显示滑块气泡，默认true显示
            result: {
                enable: false,//是否回显结果，默认false不显示，可选择true显示
                format: `<#this.value#>`,//展示内容格式，<#this.value#>用于显示值，可自由组合文本内容
            },
            fence: [0, 0],//滑块的移动范围，数组的两个值相同表示不作限制
            button: {
                enable: false,
                step: 1,
                decrease: '<i class="ax-iconfont ax-icon-minus-o-f"></i>',
                increase: '<i class="ax-iconfont ax-icon-plus-o-f"></i>',
            },
            classes: '',//追加样式，多个样式类可用英文逗号或空格隔开
            keyboard: false,//是否可以使用键盘方向键代替增减按钮，默认false
            separator: ',',//input值的分隔符，默认英文逗号
            hyphen: '~',//当multiple为ture且两个滑块靠近时的气泡值连接符
            ruler: {
                enable: false,//是否显示标尺，默认false不显示
                majorEqual: 4,//大刻度等分数量，默认4等分，可自由填写某一正整数
                minorEqual: 'auto',//小刻度等分数量，默认为auto即自动计算获得小刻度，可自由填写某一正整数，填0或1则不显示
                labels: [],//大刻度的下标，数组形式，为空则自动计算数值，可自由填写，数组的length需要等于majorEqual+1
            },
            axis: 'x',//滑块移动坐标轴，默认x即横向移动，可选择y纵向移动
            disabled: false,//是否禁用，默认false不禁用
            onInit: '',//回调函数，初始化后执行，无参数
            onSet: '',//回调函数，设置新值后执行，支持一个参数即当前的输出值（包含value和ratio）
            onToMin: '',//回调函数，返回最小值后执行，支持一个参数即回调函数，回调有一个参数即当前值
            onToMax: '',//回调函数，返回最大值后执行，支持一个参数即回调函数，回调有一个参数即当前值
            onReset: '',//回调函数，重置后执行，支持一个参数即回调函数，回调无参数
            onEnable: '',//回调函数，启用后执行，支持一个参数即回调函数，回调无参数
            onDisable: '',//回调函数，禁用后执行，支持一个参数即回调函数，回调无参数
            onUpdate: '',//回调函数，参数更新后执行，支持一个参数即赋予的新参数
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
        }, options, dom.nodeName === 'INPUT' ? this.inputDom : this.targetDom, this.constructor.name);
        //定义监听事件成员
        this.handlers = {};
        //终端事件数组
        this.isMobi = axIsMobi();
        //设定input的name
        this.inputDom.name ? this.inputDom.name = this.options.name : null;
        //手柄
        this.handles = {};
        //反转系数,向下滑动鼠标，在坐标系里是取正值，而本插件向上拖拽时正值，也就是与坐标系值相反，所以需要*-1
        this.yCoef = this.options.axis === 'y' ? -1 : 1;
        //拖行实例
        this.tows = { from: null, to: null, single: null, thumb: null };
        //属性映射
        this.attrMap = this.options.axis === 'x' ? { size: 'width', offset: 'left' } : { size: 'height', offset: 'bottom' };
        //键盘事件函数，监听键盘方向键可控制增减
        this.keyFun = (e) => {
            e.stopPropagation();
            e.cancelable ? e.preventDefault() : null;
            if (['ArrowLeft', 'ArrowDown'].includes(e.code)) {
                this.decrease();
            } else if (['ArrowRight', 'ArrowUp'].includes(e.code)) {
                this.increase();
            }
            //及时移除
            if (document.activeElement !== this.targetDom) {
                document.removeEventListener('keydown', this.keyFun);
            }
        }
        //页面的点击事件
        this.clickFun = () => {
            if (document.activeElement === this.targetDom) {
                document.removeEventListener('keydown', this.keyFun);
                document.addEventListener('keydown', this.keyFun);
            } else {
                document.removeEventListener('keydown', this.keyFun);
            }
        }
        //初始化，因为需要获得物理节点尺寸，所以确保绘制完毕才初始化
        setTimeout(() => {
            this.init()
        }, 0);
    }

    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'range');
        //设置销毁状态
        this.destroyed = false;
        //判断是否有存储值而更新参数
        if (this.options.storageName) {
            //则先存下空对象
            axLocalStorage.set(this.options.storageName, {}, false);
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            }
        }
        //保存初始值
        this.rawValue = axClone(this.options.value);
        //获得输出值位数，与min和max相同位数
        this.places = Math.max(
            this.options.min.toString().split('.')[1]?this.options.min.toString().split('.')[1].length : 0,
            this.options.max.toString().split('.')[1]?this.options.max.toString().split('.')[1].length : 0,
        );
        //输出值，multiple则格式为{value:[],ratio:[]}，否则格式为{value:0,ratio:0}
        this.output = { value: null, ratio: null };
        //获得范围值
        this.getRangeValues()
        //设定滑块的初始值，获得this.initialValue变量
        this.getInitialValue();
        //初始化节点
        this.initRender();
        //设置属性
        this.setAttribute();
        //获得滑块尺寸
        this.getSize();
        //设置滑块初始位置
        this.setValues(this.initialValue);
        //设置刻度,this.ticks只保留大刻度
        this.ticks = [];
        this.setTicks();
        //操作
        let _this = this;
        if (!this.options.disabled) {
            //点击按钮增减
            if (this.options.button.enable) {
                this.decreaseDom.onclick = () => this.decrease();
                this.increaseDom.onclick = () => this.increase();
            }
            //对手柄和滑轨增加滑动事件
            if (this.options.multiple) {
                this.setHandleTow('from');
                this.setHandleTow('to');
                //滑块拖行要特别些
                this.tempRatios = [...this.output.ratio];
                this.thumbTow = new axTow(this.thumbDom, {
                    onStart: function () {

                    },
                    onMove: function (e) {
                        e.originEvent.stopPropagation();
                        e.originEvent.cancelable ? e.originEvent.preventDefault() : null;
                        _this.slideSet(_this.thumbDom, e.diff[_this.options.axis] * _this.yCoef);
                    },
                    onEnd: () => {
                        this.tempRatios = [...this.output.ratio];
                    }
                });
            } else {
                this.setHandleTow('single');
            }
            //点击地基
            this.baseDom.onclick = (e) => {
                let clientObj = axGetClientObj(e), offset;
                if (!clientObj) { return false; }
                offset = (clientObj[this.options.axis] - this.baseDom.getBoundingClientRect()[this.attrMap.offset]) * this.yCoef;
                if (this.options.multiple) {
                    //from和to同时移动
                    let fromOffset = this.output.ratio[0] * this.size.track,
                        toOffset = this.output.ratio[1] * this.size.track;
                    if (offset > toOffset) {
                        this.slideSet(_this.thumbDom, offset - toOffset - this.size.handle / 2);
                    } else if (offset < toOffset) {
                        this.slideSet(_this.thumbDom, offset - fromOffset - this.size.handle / 2);
                    }
                    this.tempRatios = [...this.output.ratio];
                } else {
                    //
                    this.slideSet(this.handles.single, offset - this.size.handle / 2);
                }
            }
            //single可点击导轨
            if (!this.options.multiple) {
                this.thumbDom.onclick = (e) => {
                    let clientObj = axGetClientObj(e), offset;
                    if (clientObj) {
                        offset = (clientObj[this.options.axis] - this.baseDom.getBoundingClientRect()[this.attrMap.offset]) * this.yCoef;
                        this.slideSet(this.handles.single, offset - this.size.handle / 2);
                    }
                }
            }
            //点击min和max气泡
            if (this.bubbles.min) {
                this.bubbles.min.onclick = () => {
                    this.toMin();
                }
            }
            if (this.bubbles.max) {
                this.bubbles.max.onclick = () => {
                    this.toMax();
                }
            }
            //监听键盘方向事件
            if (this.options.keyboard) {
                document.removeEventListener('click', this.clickFun);
                document.addEventListener('click', this.clickFun);
            }






        }
        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    setHandleTow(type) {
        let _this = this;
        this.tows[type] = new axTow(this.handles[type], {
            onStart: function () {
                this.offset = axStyle(_this.handles[type])[_this.attrMap.offset];
            },
            onMove: function (e) {
                e.originEvent.stopPropagation();
                e.originEvent.cancelable ? e.originEvent.preventDefault() : null;

                _this.slideSet(_this.handles[type], parseFloat(this.offset) + e.diff[_this.options.axis] * _this.yCoef);
            },
            onEnd: () => {
                type !== 'single' ? this.tempRatios = [...this.output.ratio] : null;
            }
        });
    }
    setHandleActive(dom) {
        dom.setAttribute('active', '');
        for (let k in this.handles) {
            this.handles[k] !== dom ? this.handles[k].removeAttribute('active') : null;
        }
    }
    getRangeValues() {
        //强制排序
        let max, min;
        this.options.fence.sort();
        if (this.options.fence[0] !== this.options.fence[1]) {
            //采取限制值
            if (this.options.fence[1] < this.options.min) {
                max = min = this.options.min;
            } else if (this.options.fence[0] > this.options.max) {
                max = min = this.options.max;
            } else if (this.options.fence[0] < this.options.min && this.options.fence[1] > this.options.max) {
                min = this.options.min;
                max = this.options.max;
            } else if (this.options.fence[0] < this.options.min) {
                min = this.options.min;
                max = this.options.fence[1];
            } else if (this.options.fence[1] > this.options.max) {
                min = this.options.fence[0];
                max = this.options.max;
            } else {
                min = this.options.fence[0];
                max = this.options.fence[1];
            }
        } else {
            min = this.options.min;
            max = this.options.max;
        }
        this.rangeValues = { min, max };

    }
    getRatioFromValue(value) {
        return ((value - this.options.min) / (this.options.max - this.options.min)) * this.size.ratio.full;
    }
    getValueFromRatio(ratio) {
        return ratio * (this.options.max - this.options.min) / this.size.ratio.full + this.options.min;
    }
    decrease(handleDiff = (this.output.value[1] - this.output.value[0])) {
        let value = (this.output.value[0] - this.options.step);
        if (this.options.multiple) {
            //不可小于极限范围值
            value < this.rangeValues.min ? value = this.rangeValues.min : null;
            if (this.options.locked) {
                this.setValues([value, value + handleDiff]);
            } else {
                this.setValues([value, this.output.value[1]]);
            }
        } else {
            this.setValues(this.output.value - this.options.step);
        }
        //保存ratio
        this.options.multiple ? this.tempRatios = [...this.output.ratio] : null;
    }
    increase(handleDiff = (this.output.value[1] - this.output.value[0])) {
        let value = (this.output.value[1] + this.options.step);
        if (this.options.multiple) {
            //不可大于极限范围值
            value > this.rangeValues.max ? value = this.rangeValues.max : null;
            if (this.options.locked) {
                this.setValues([value - handleDiff, value]);
            } else {
                this.setValues([this.output.value[0], value]);
            }
        } else {
            this.setValues(this.output.value + this.options.step);
        }
        //保存ratio
        this.options.multiple ? this.tempRatios = [...this.output.ratio] : null;
    }
    //设置节点属性
    setAttribute() {
        this.targetDom.setAttribute('axis', this.options.axis);
        //可获得焦点
        this.targetDom.setAttribute('tabindex', "0");
        axClass.add(this.targetDom, this.options.classes);
        this.options.multiple ? this.targetDom.setAttribute('multiple', '') : null;
        this.options.disabled ? this.targetDom.setAttribute('disabled', '') : null;
    }
    getSize() {
        let handle = (this.handles.single || this.handles.from).getBoundingClientRect()[this.attrMap.size],
            track = this.trackDom.getBoundingClientRect()[this.attrMap.size],
            min = this.bubbles.min ? this.bubbles.min.getBoundingClientRect()[this.attrMap.size] : 0,
            max = this.bubbles.max ? this.bubbles.max.getBoundingClientRect()[this.attrMap.size] : 0,
            fullRatio = (track - handle) / track,
            radiusRadio = (handle / 2) / track,
            minRatio = ((this.rangeValues.min - this.options.min) / (this.options.max - this.options.min)) * fullRatio,
            maxRatio = ((this.rangeValues.max - this.options.min) / (this.options.max - this.options.min)) * fullRatio;
        this.size = {
            handle, track, ruler: 0,
            bubble: { min, max, from: 0, to: 0, single: 0 },
            ratio: { radius: radiusRadio, min: minRatio, max: maxRatio, full: fullRatio }
        };
    }
    getInitialValue() {
        //设定初始值，以input的value优先
        if (this.inputDom.value) {
            if (this.inputDom.value.includes(this.options.separator)) {
                //既然有两个值则强制multiple转为true
                this.options.multiple = true;
                //input的value是文本，需要转为数值并按小-大排序
                this.initialValue = this.inputDom.value.split(this.options.separator).map(k => parseFloat(k)).sort();
            } else {
                //不包含则是单值
                this.initialValue = this.options.multiple ? [parseFloat(this.inputDom.value), this.options.max] : parseFloat(this.inputDom.value);
            }
        } else {
            //multiple参数优先
            if (this.options.multiple && !Array.isArray(this.options.value)) {
                //对第二个值设置最大值
                this.initialValue = [this.options.value, this.options.max];
            } else if (!this.options.multiple && Array.isArray(this.options.value)) {
                //强制转双滑块
                this.options.multiple = true;
                this.initialValue = [...this.options.value].sort();
            } else {
                //正确匹配则与options.value相同
                this.initialValue = this.options.value;
                //排序
                Array.isArray(this.initialValue) ? this.initialValue.sort() : null;
            }
        }
        //multiple与value类型已经一致
        //再次修正
        this.initialValue = this.correctValue(this.initialValue);
    }
    correctValue(value) {
        let result;
        if (Array.isArray(value)) {
            //数组
            result = [axNumber(axClampValue({ value: value[0], ...this.rangeValues }), { places: this.places }), axNumber(axClampValue({ value: value[1], ...this.rangeValues }), { places: this.places })]
            //禁止to值大于from值
            result[0] > result[1] ? result[0] = result[1] : null;
        } else {
            //单值
            result = axNumber(axClampValue({ value, ...this.rangeValues }), { places: this.places });
        }
        //根据multiple修正value
        if (this.options.multiple) {
            return Array.isArray(result) ? result : [result, this.rangeValues.max];
        } else {
            return Array.isArray(result) ? result[0] : result;
        }
    }
    //根据标准数据初步渲染节点
    initRender() {
        //先清空
        if (this.wrapperDom) {
            this.wrapperDom.innerHTML = '';
        } else {
            this.wrapperDom = axAddElem('div', { class: 'ax-range-wrapper' });
            this.targetDom.appendChild(this.wrapperDom);
        }
        //初始化bubbles节点
        this.bubbles = {};
        //填充wrapper
        this.wrapperDom.innerHTML = `
        ${this.options.button.enable ? '<span class="ax-range-btn" decrease>' + this.options.button.decrease + '</span>' : ''}
        <div class="ax-range-main">
            <div class="ax-range-track">
                <div base></div>
                <div thumb></div>
            </div>
            <div class="ax-range-bubble">
                ${this.options.limitShow ? '<span min>' + this.options.min + '</span><span max>' + this.options.max + '</span>' : ''}
            </div>
            <div class="ax-range-handle"></div>
            <div class="ax-range-ruler"><ul></ul></div>
        </div>
        ${this.options.button.enable ? '<span class="ax-range-btn"  increase>' + this.options.button.increase + '</span>' : ''}
        ${this.options.result.enable ? '<span class="ax-range-result"></span>' : ''}
        `;
        //主节点
        this.decreaseDom = this.wrapperDom.querySelector('[decrease]');
        this.increaseDom = this.wrapperDom.querySelector('[increase]');
        this.resultDom = this.wrapperDom.querySelector('.ax-range-result');
        this.mainDom = this.wrapperDom.querySelector('.ax-range-main');
        this.trackDom = this.wrapperDom.querySelector('.ax-range-track');
        this.thumbDom = this.wrapperDom.querySelector('[thumb]');
        this.baseDom = this.wrapperDom.querySelector('[base]');
        this.bubbleDom = this.wrapperDom.querySelector('.ax-range-bubble');
        this.bubbles.min = this.bubbleDom.querySelector('[min]');
        this.bubbles.max = this.bubbleDom.querySelector('[max]');
        this.handleDom = this.wrapperDom.querySelector('.ax-range-handle');
        this.rulerDom = this.wrapperDom.querySelector('.ax-range-ruler');
        this.tickDom = this.rulerDom.querySelector('ul');

        //滑块和对应气泡
        if (this.options.multiple) {
            //滑块节点
            this.handles.from = axAddElem('span', { from: '' });
            this.handles.to = axAddElem('span', { to: '' });
            this.handleDom.append(this.handles.from, this.handles.to);
            //气泡节点
            if (this.options.tipShow) {
                this.bubbles.from = axAddElem('span', { from: '' });
                this.bubbles.to = axAddElem('span', { to: '' });
                this.bubbles.merge = axAddElem('span', { merge: '' });
                this.bubbleDom.append(this.bubbles.from, this.bubbles.to, this.bubbles.merge);
            }
        } else {
            //滑块节点
            this.handles.single = axAddElem('span', { single: '' });
            this.handleDom.append(this.handles.single);
            //气泡节点
            if (this.options.tipShow) {
                this.bubbles.single = axAddElem('span', { single: '' });
                this.bubbleDom.append(this.bubbles.single);
            }
        }
        //

    }
    setValues(value, callback) {
        if (this.destroyed) {
            return this;
        }
        //作为输出值
        this.output.value = this.correctValue(value);

        let radiusPercent = this.size.ratio.radius * 100,
            getTipHTML = (val) => axNumber(val, { places: this.places });
        if (Array.isArray(this.output.value)) {
            //是multiple，两个滑块
            this.output.ratio = [this.getRatioFromValue(this.output.value[0]), this.getRatioFromValue(this.output.value[1])];

            let fromPercent = this.output.ratio[0] * 100,
                toPercent = this.output.ratio[1] * 100;
            this.handles.from.style[this.attrMap.offset] = `${fromPercent}%`;
            this.handles.to.style[this.attrMap.offset] = `${toPercent}%`;
            //追加size，size.ratio={min,max,center,full,radius}
            this.size.ratio.center = this.output.ratio[1];
            this.thumbDom.style[this.attrMap.size] = `${toPercent - fromPercent}%`;
            this.thumbDom.style[this.attrMap.offset] = `${fromPercent + radiusPercent}%`;
            this.size.thumb = this.thumbDom.getBoundingClientRect()[this.attrMap.size];
            //气泡
            if (this.options.tipShow) {

                this.bubbles.from.style[this.attrMap.offset] = `${fromPercent}%`;
                //使用整数
                this.bubbles.from.innerHTML = getTipHTML(this.output.value[0]);

                this.bubbles.to.style[this.attrMap.offset] = `${toPercent}%`;
                this.bubbles.to.innerHTML = getTipHTML(this.output.value[1]);

                this.bubbles.merge.style[this.attrMap.offset] = `${fromPercent + (toPercent - fromPercent) / 2}%`;
                this.bubbles.merge.innerHTML = `${getTipHTML(this.output.value[0])}<i>${this.options.hyphen}</i>${getTipHTML(this.output.value[1])}`;

                //追加from和to气泡值
                this.size.bubble.from = this.bubbles.from.getBoundingClientRect()[this.attrMap.size];
                this.size.bubble.to = this.bubbles.to.getBoundingClientRect()[this.attrMap.size];

                //判断气泡合并
                this.mergeTips();
            }
        } else {
            //是single，一个滑块
            this.output.ratio = this.getRatioFromValue(this.output.value);

            let singlePercent = this.output.ratio * 100;
            this.handles.single.style[this.attrMap.offset] = `${singlePercent}%`;
            this.thumbDom.style[this.attrMap.size] = `${singlePercent + radiusPercent}%`;
            this.size.thumb = this.thumbDom.getBoundingClientRect()[this.attrMap.size];
            //气泡
            if (this.options.tipShow) {
                this.bubbles.single.style[this.attrMap.offset] = `${singlePercent}%`;
                this.bubbles.single.innerHTML = getTipHTML(this.output.value);
                //追加single气泡值
                this.size.bubble.single = this.bubbles.single.getBoundingClientRect()[this.attrMap.size];
            }

        }
        //判断是否隐藏头尾
        this.options.limitShow ? this.hideEnds() : null;
        //输出文本
        this.inputDom.value = this.output.value + '';
        this.options.result.enable ? this.resultDom.innerHTML = axTplEngine(this.options.result.format, { value: this.output.value + '' }) : null;
        //保存数据
        this.save();

        this.handlers.hasOwnProperty('set') ? this.emit('set', this.output) : null;
        this.options.onSet && this.options.onSet.call(this, this.output);
        callback && callback.call(this, this.output);
        return this;
    }
    getValues() {
        return this.output;
    }
    toMax(callback) {
        if (this.destroyed) {
            return this;
        }
        let value, diff;
        if (this.options.multiple) {
            diff = this.rangeValues.max - this.output.value[1];
            value = [this.output.value[0] + diff, this.output.value[1] + diff];
        } else {
            value = this.rangeValues.max;
        }
        this.setValues(value);
        this.handlers.hasOwnProperty('toMax') ? this.emit('toMax', value) : null;
        this.options.onToMax && this.options.onToMax.call(this, value);
        callback && callback.call(this, value);
        return this;
    }
    toMin(callback) {
        if (this.destroyed) {
            return this;
        }
        let value, diff;
        if (this.options.multiple) {
            diff = this.rangeValues.min - this.output.value[0];
            value = [this.output.value[0] + diff, this.output.value[1] + diff];
        } else {
            value = this.rangeValues.min;
        }
        this.setValues(value);
        this.handlers.hasOwnProperty('toMin') ? this.emit('toMin', value) : null;
        this.options.onToMin && this.options.onToMin.call(this, value);
        callback && callback.call(this, value);
        return this;
    }
    slideSet(dom, value) {
        if (this.destroyed) {
            return this;
        }
        let result;
        if (dom === this.thumbDom && this.options.multiple) {
            //拖拽滑轨，value是偏移差值（与上次相减）
            let fromRatio = this.tempRatios[0] + value / this.size.track,
                diffRatio = this.tempRatios[1] - this.tempRatios[0],
                toRatio = fromRatio + diffRatio,
                from = this.getValueFromRatio(fromRatio),
                to = this.getValueFromRatio(toRatio),
                diff = this.getValueFromRatio(toRatio) - this.getValueFromRatio(fromRatio);

            if (this.options.locked) {
                if (from < this.rangeValues.min) {
                    result = [this.rangeValues.min, this.rangeValues.min + diff];
                } else if (to > this.rangeValues.max) {
                    result = [this.rangeValues.max - diff, this.rangeValues.max];
                } else {
                    result = [from, to];
                }
            } else {
                result = [from, to];
            }
        } else {
            //拖拽手柄，value是偏移值
            let ratio = (value / this.size.track);
            //修正范围
            if (dom === this.handles.from) {
                //不可大于to滑块的比例
                if (ratio > this.output.ratio[1]) {
                    ratio = this.output.ratio[1];
                } else if (ratio < this.size.ratio.min) {
                    ratio = this.size.ratio.min;
                }
                result = [this.getValueFromRatio(ratio), this.output.value[1]];
            } else if (dom === this.handles.to) {
                //不可大于max比例
                if (ratio > this.size.ratio.max) {
                    ratio = this.size.ratio.max;
                } else if (ratio < this.output.ratio[0]) {
                    ratio = this.output.ratio[0];
                }
                result = [this.output.value[0], this.getValueFromRatio(ratio)];
            } else if (dom === this.handles.single) {
                //不可大于full比例
                if (ratio > this.size.ratio.max) {
                    ratio = this.size.ratio.max;
                } else if (ratio < this.size.ratio.min) {
                    ratio = this.size.ratio.min;
                }
                result = this.getValueFromRatio(ratio);
            }
        }
        //设置值
        this.setValues(result);
    }
    mergeTips() {
        if (this.size.thumb < (this.size.bubble.from + this.size.bubble.to) / 2) {
            this.targetDom.setAttribute('merge', '')
        } else {
            this.targetDom.removeAttribute('merge')
        }
    }
    hideEnds() {
        if (this.options.multiple) {
            let offsetFrom = this.output.ratio[0] * this.size.track,
                offsetTo = this.output.ratio[1] * this.size.track;
            //头隐藏
            if (offsetFrom < this.size.bubble.min + this.size.bubble.from / 2 - this.size.handle / 2) {
                this.targetDom.setAttribute('tostart', '');
            } else {
                this.targetDom.removeAttribute('tostart');
            }
            //尾隐藏
            if (this.size.track - offsetTo < this.size.bubble.max + this.size.bubble.from / 2 + this.size.handle / 2) {
                this.targetDom.setAttribute('toend', '');
            } else {
                this.targetDom.removeAttribute('toend');
            }
        } else {
            let offsetSingle = this.output.ratio * this.size.track;
            if (offsetSingle < this.size.bubble.min + this.size.bubble.single / 2 - this.size.handle / 2) {
                this.targetDom.setAttribute('tostart', '');
            } else {
                this.targetDom.removeAttribute('tostart');
            }
            if (this.size.track - offsetSingle < this.size.bubble.max + this.size.bubble.single / 2 + this.size.handle / 2) {
                this.targetDom.setAttribute('toend', '');
            } else {
                this.targetDom.removeAttribute('toend');
            }
        }
    }
    setTicks() {
        if (this.options.ruler.enable) {
            this.ceilValue = (this.options.max - this.options.min) / this.options.ruler.majorEqual;
            this.size.ruler = this.rulerDom.getBoundingClientRect()[this.attrMap.size];
            for (let k = 0; k <= this.options.ruler.majorEqual; k++) {
                let value = axNumber((this.options.min + this.ceilValue * k),{ places: this.places }),
                    content = this.options.ruler.labels.length > 0 ? this.options.ruler.labels[k] : value,
                    offsetRatio = k / this.options.ruler.majorEqual,
                    majorDom = axAddElem('li', { style: `${this.attrMap.offset}:${offsetRatio * 100}%`, 'data-content': content });
                this.tickDom.appendChild(majorDom);
                //点击可设置值并跳转
                majorDom.onclick = () => {
                    this.setValues(value);
                }
                //追加到数组
                this.ticks.push({
                    index: k,
                    dom: majorDom,
                    ratio: offsetRatio,
                    content: content,
                    value,
                });
                //添加子刻度
                if (k !== this.options.ruler.majorEqual) {
                    this.setMinorTicks(offsetRatio)
                }
            }
            //设置滑动区间
            if (this.options.fence[0] !== this.options.fence[1]) {
                let fullSize = (this.options.max - this.options.min),
                    sizeRatio = (this.options.fence[1] - this.options.fence[0]) / fullSize,
                    offsetRatio = (this.options.fence[0] - this.options.min) / fullSize;
                this.fenceDom = axAddElem('div', { style: `${this.attrMap.size}:${sizeRatio * 100}%;${this.attrMap.offset}:${offsetRatio * 100}%` });
                this.rulerDom.appendChild(this.fenceDom);
            }
        }
    }
    setMinorTicks(offsetRatio) {
        let count = 0, unitRatio = 0, ceilSize = this.size.ruler / this.options.ruler.majorEqual;
        if (this.options.ruler.minorEqual === 'auto') {
            //自动
            count = ceilSize < 40 ? 5 :
                (ceilSize >= 40 && ceilSize < 80) ? 6 :
                    (ceilSize >= 80 && ceilSize < 160) ? 7 :
                        (ceilSize >= 160 && ceilSize < 320) ? 8 :
                            (ceilSize >= 320 && ceilSize >= 640) ? 9 : 10;
        } else if (this.options.ruler.minorEqual > 0) {
            //手动
            count = this.options.ruler.minorEqual;
        }
        unitRatio = 1 / (count * this.options.ruler.majorEqual);
        for (let k = 0; k < count; k++) {
            this.tickDom.insertAdjacentHTML('beforeend', `<li style="${this.attrMap.offset}:${(offsetRatio + unitRatio * k) * 100}%"></li>`);
        }
    }
    disable(callback) {
        this.targetDom.setAttribute('disabled', '');
        this.handlers.hasOwnProperty('disable') ? this.emit('disable', '') : null;
        this.options.onDisable && this.options.onDisable.call(this);
        callback && callback.call(this);
        return this;
    }
    enable(callback) {
        this.targetDom.removeAttribute('disabled');
        this.handlers.hasOwnProperty('enable') ? this.emit('enable', '') : null;
        this.options.onEnable && this.options.onEnable.call(this);
        callback && callback.call(this);
        return this;
    }
    //保存当前数据
    save(props, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }
        if (!props) {
            //保存当前焦点索引
            axLocalStorage.set(this.options.storageName, { value: this.output.value }, { cover: false });
        } else {
            //保存参数,props={a:'',b:''}
            props.value = this.output.value;
            axLocalStorage.set(this.options.storageName, props, { cover: false });
        }
        //回调监听
        let localValue = axLocalStorage.get(this.options.storageName);
        this.handlers.hasOwnProperty('save') ? this.emit('save', localValue) : null;
        this.options.onSave && this.options.onSave.call(this, localValue);
        callback && callback.call(this, localValue);
        return this;
    }
    reset(callback) {
        if (this.destroyed) {
            return this;
        }
        this.setValues(this.rawValue);
        this.handlers.hasOwnProperty('reset') ? this.emit('reset', '') : null;
        this.options.onReset && this.options.onReset.call(this);
        callback && callback.call(this);
        return this;
    }
    //更新实例参数
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //保存数据
        this.save(setting);
        //更新参数
        this.options = axExtend(this.options, setting);
        //根据参数初始化实例
        this.init();
        //监听回调
        this.handlers.hasOwnProperty('update') ? this.emit('update', setting) : null;
        this.options.onUpdate && this.options.onUpdate.call(this, setting);
        callback && callback.call(this);
        return this;
    }
    //销毁实例
    destroy(callback) {
        //移除事件
        if (this.options.keyboard) {
            document.removeEventListener('keydown', this.keyFun);
            document.removeEventListener('click', this.clickFun);
        }
        if (this.bubbles.min) {
            this.bubbles.min.onclick = null;
        }
        if (this.bubbles.max) {
            this.bubbles.max.onclick = null;
        }
        this.ticks.forEach(k => {
            k.onclick = null;
        });
        if (this.options.button.enable) {
            this.decreaseDom.onclick = null;
            this.increaseDom.onclick = null;
        }
        this.baseDom.onclick = null;
        if (!this.options.multiple) {
            this.thumbDom.onclick = null
        }
        if (this.bubbles.min) {
            this.bubbles.min.onclick = null;
        }
        if (this.bubbles.max) {
            this.bubbles.max.onclick = null;
        }
        //销毁实例
        for (let k in this.tows) {
            if (this.tows.hasOwnProperty(k) && this.tows[k]) {
                this.tows[k].destroy();
            }
        }

        //更新销毁状态
        this.destroyed = true;
        //监听回调
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    //以下三个方法为监听
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }


}
/*!
 * 插件：分页；使用方法：new axPagination('#id',{参数})
 * 从现有的list中自动创建分页
 * 从已知的数组创建列表并分页
 * 可追加列表item数量并重新渲染分页
 * 分页节点能自动在列表下方创建，也可以在指定的dom创建
 */
class axPagination {
    constructor(targetDom, options) {
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            listDom: '',//存放列表dom的盒子，如果需要展示筛选出来的列表则需要填写
            count: 10, //每页item数量，整数类型
            current: 1, //当前页码，整数类型
            total: 0,//参与分页的信息条数，整数类型，默认为0表示不启用，如果启用表明将不从data获取数据，而是手动渲染数据；总页数=total/count
            firstText: '首页', //首页展示文字
            prevText: '上一页', //上一页展示文字
            nextText: '下一页', //下一页展示文字
            lastText: '尾页', //尾页展示文字
            allPages: false,//是否显示所有页码，默认否，可选择true显示全部页码
            data: [], //用于创建列表Dom的数组，据此创建分页
            clickDelay: 200,//点击按钮时的延时时间，避免快速连续多次点击导致系统负荷
            template: '', //数组转Dom的模板
            type: 'regular', //分页导航风格，可选填regular,simple,full,pagesRight,pagesLeft,pure
            //type:'regular'//类似：首页,上一页,1,2,3,...9,10下一页尾页
            //type:'simple'//只保留上一页和下一页按钮
            //type:'full'//类似：在regular基础上在加total和前往分页
            //type:'pagesRight'//内容跟full一样，只是使用弹性布局将分页至于右侧
            //type:'pagesLeft'//内容跟full一样，只是使用弹性布局将分页至于左侧
            //type:'pure'//纯页码风格，在regular基础上减掉按钮
            loadingMode: 'replace',//使用异步分页的加载效果，默认是replace即清除列表显示转圈，可选择overlay即浮在原列表上
            className: '',//给分页Dom追加class样式,比如：className: 'ax-align-right'
            lastDel: '', //删除当页最后一条某个节点，可以是className，id或node
            async: '',//异步方式，选择从json文件获取数组数据，或选择sql从数据库获得数据
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            delay: 0,//异步分页内容时延时加载
            dataExtend: '',//异步传递的数据对象扩展，除了current、count和offset可以传递更多参数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onBeforeRendered: function (listArr) {
                //渲染list节点前回调，支持listArr参数，即渲染当页的节点数组
            },
            onRendered: function (current, pagesNum,offset) {
                //渲染完list节点后回调，支持current、pagesNum和offset参数
            },
            onUpdated: function (current, pagesNum,offset) {
                //更新后回调，支持current、pagesNum和offset参数
            },
            onTurned: function (current, pagesNum,offset) {
                //初次加载和翻页后回调，支持current、pagesNum和offset参数
            }
        }, options, targetDom, this.constructor.name);
        //分页Dom
        this.targetExist = targetDom ? true : false;
        this.targetDom = targetDom ? axIdToDom(targetDom) : axAddElem('div', { class: this.options.className ? 'ax-pagination ' + this.options.className : 'ax-pagination' });
        //定义列表list盒子和分页盒子
        this.listDom = '';
        //如果使用total，则手动定义data
        this.options.total ? this.options.data = { length: this.options.total } : null;
        //监听事件
        this.handlers = {};
        //中转数组，从data创建列表Dom，并创建分页
        this.preData = [];
        //异步创建的数据，[index:'',dom:'',children:'']，包含了preData中的节点
        this.asyncData = []
        //当前页的列表Dom数组，截取自this.preData
        this.selectData = [];
        //总项目数量分页总数
        this.itemsNum = 0;
        this.pagesNum = 0;
        this.current = this.options.current;
        //异步分页加载容器
        this.loadingDom = axAddElem('div', { class: 'ax-pageTurning' })
        //可以从页面节点获取项目模板
        if (!this.options.template) {
            this.template = '';
        } else {
            if (this.options.template.trim().substr(0, 1) === '#') {
                this.template = document.querySelector(this.options.template).innerHTML
            } else {
                this.template = this.options.template;
            }
        }
        //创建分页容器其他节点
        this.pagesDom = axAddElem('div', { class: 'ax-core' });
        this.pageFirst = axAddElem('a', { href: '###', class: 'ax-first' }, this.options.firstText);
        this.pageLast = axAddElem('a', { href: '###', class: 'ax-last' }, this.options.lastText);
        this.pagePrev = axAddElem('a', { href: '###', class: 'ax-prev' }, this.options.prevText);
        this.pageNext = axAddElem('a', { href: '###', class: 'ax-next' }, this.options.nextText);
        this.pageTotal = axAddElem('span', { class: 'ax-total' });
        this.pageGo = axAddElem('span', { class: 'ax-gopage' });
        this.group = axAddElem('div', { class: 'ax-group' });

        //生成列表次数
        this.renderedIndex = 0;

        //初始化
        this.init(this.options.data);
        return this;
    }
    async init(data) {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'pagination');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //获得数据类型
        this.dataType = this.getDataType(data);
        //获得list节点
        if (this.dataType === 'pageDom') {
            //如果是从页面现存节点创建分页，则列表Dom就是现有节点
            this.listDom = axIdToDom(this.options.data);
        } else if (this.options.listDom) {
            this.listDom = axIdToDom(this.options.listDom);
        }

        //如果数据是空数组则清除所有Dom并设置相关值为0以便于回调使用，不再往下执行
        if (axIsEmpty(data)) {
            this.listDom ? this.listDom.innerHTML = '' : null;
            //既然没有数据，则清空分页，但是不删除节点
            this.targetDom.innerHTML = '';
            this.current = 1;
            this.pagesNum = 1;
            this.itemsNum = 0;
            return false;
        }
        //定义初次加载渲染的页码
        let initCurrent;
        //如果是异步数据模式
        if (this.dataType === 'json' || this.dataType === 'sql') {
            //如果是json数据，则根据地址获取数组（需要模板解析）
            if (!this.template && this.listDom.nodeName != 'TBODY') {
                console.error('The template is required!');
                return false;
            }
            if (this.options.async === 'json') {
                //如果data为json地址的异步模式
                await this.getJsonData({ url: data });
            } else if (this.options.async === 'sql') {
                //如果有异步地址和模板则使用异步方式
                await this.getSqlData({ current: this.current });
            }
            //不同数据类型初次渲染页码不同，异步即渲染1页
            initCurrent = 1;
        } else {
            //同步数据模式
            if (this.dataType === 'sortList') {
                //数据格式是[{name:'',country:'',dom:''},{...}...]，通常是来自axList的数据
                data.forEach(i => {
                    this.preData.push(i.dom);
                });
            } else if (this.dataType === 'pureNode') {
                //节点数组：[dom,dom...]，直接复制即可
                this.preData = [...data];

            } else if (this.dataType === 'tplArray') {
                //数据格式是非dom的数据，需要template进行解析生成dom
                if (!this.template) {
                    console.error('The template is required!');
                    return false;
                }
                //根据数据和模板生成虚拟节点
                this.preData = this.tplToDoms(data);
            } else if (this.dataType === 'manual') {
                //手动渲染
                this.preData = { ...data };
            } else {
                //如果是页面上的节点，则直接选择子节点
                this.preData = [...this.listDom.childNodes].filter(k => k.nodeType == 1);
            }
            //通过以上步骤已经获得preData
            //更新pagesNum，itemsNum和current
            this.getCurrent(this.preData.length);
            //不同数据类型初次渲染页码不同
            initCurrent = this.current;

        }
        //初次渲染列表，手动模式不能立即渲染
        this.listDom && this.dataType !== 'manual' ? this.renderList(initCurrent) : null;
        //完成渲染后执行
        this.renderFinish();
    }
    getDataType(data) {
        let type = '',
            dataType = axType(data);
        if (dataType === 'Object' && data.hasOwnProperty('length')) {
            //手动渲染数据
            type = 'manual';
        } else {
            //自动渲染数据
            //根据data参数判断数据类型
            if (dataType === 'Array') {
                if (data.length > 0 && data[0].hasOwnProperty('dom')) {
                    //axList用的数组：[{country:'',click:'',dom:''},{...}...]
                    type = 'sortList';
                } else if (data.length > 0 && axType(data[0]).includes('Element')) {
                    //节点数组：[dom,dom...]
                    type = 'pureNode';
                } else {
                    //需要模板渲染的数组:[{title:'',click:'',href:''}]
                    type = 'tplArray';
                }
            } else if (dataType === 'String') {
                if (this.options.async === 'json') {
                    //异步获取json
                    type = 'json';
                } else if (this.options.async === 'sql') {
                    //只从数据库取得一页数据
                    type = 'sql';
                } else {
                    //#id、class或节点，从现存的节点获取节点数组
                    type = 'pageDom';
                }
            }
        }

        return type;
    }
    //根据标准数据+模板，生成虚拟节点数组
    tplToDoms(data, tpl = this.template) {
        let domArr = [];
        data.forEach(i => {
            let item = axStrToDom(axTplEngine(tpl, i));
            domArr.push(item);
        });
        return domArr;
    }
    renderTotal() {
        //根据数据添加总计文本
        this.pageTotal.innerHTML = '共有' + this.pagesNum + '页, ' + this.itemsNum + '条数据';
    }
    renderFinish() {
        //清除正在分页属性
        this.targetDom.removeAttribute('paging');
        //根据data渲染生成分页
        this.renderPagesTotal(this.current, () => {
            //生成总数统计
            this.renderTotal();
        });
        this.getPagination();
        //如果没有分页Dom的节点,但是有列表节点，则直接在列表后面添加分页Dom
        if (this.listDom && !this.targetExist) {
            if (this.listDom.nodeName == 'TBODY') {
                //如果是表格，则应该在table之后追加分页
                this.listDom.parentElement.insertAdjacentElement('afterend', this.targetDom)
            } else {
                this.listDom.insertAdjacentElement('afterend', this.targetDom)
            }
        }

        //点击首页、尾页、上一页和下一页
        this.pageFirst.onclick = axDebounce((e) => {
            //当前页小于等于1,便取消点击事件
            if (this.current <= 1) {
                axPreventDefault(e);
            } else {
                this.current = 1;
                //重新渲染列表节点和分页节点
                this.currentToRenderList(this.current);
            }
        }, this.options.clickDelay);
        this.pageLast.onclick = axDebounce((e) => {
            //当前页大于等于总页数,便取消点击事件
            if (this.current >= this.pagesNum) {
                axPreventDefault(e);
            } else {
                this.current = this.pagesNum;
                //重新渲染列表节点和分页节点
                this.currentToRenderList(this.current);

            }
        }, this.options.clickDelay);
        this.pagePrev.onclick = axDebounce((e) => {
            //当前页小于等于1,便取消点击事件
            if (this.current <= 1) {
                axPreventDefault(e);
            } else {
                this.current = this.current - 1;
                //重新渲染列表节点和分页节点
                this.currentToRenderList(this.current);
            }
        }, this.options.clickDelay);
        this.pageNext.onclick = axDebounce((e) => {
            //当前页大于等于总页码,便取消点击事件
            if (this.current >= this.pagesNum) {
                axPreventDefault(e);
            } else {
                this.current = this.current + 1;
                //重新渲染列表节点和分页节点
                this.currentToRenderList(this.current);
            }
        }, this.options.clickDelay);

        //渲染节点次数+1，sql模式下需要会反复执行渲染
        this.renderedIndex++;
        //监听回调

        this.options.onTurned && this.options.onTurned.call(this, this.current, this.pagesNum,this.getOffset());
        this.handlers.hasOwnProperty('turned') ? this.emit('turned', this.current, this.pagesNum,this.getOffset()) : null;
    }
    //获得当前分页的信息条数
    getOffset(current = this.current, total = this.itemsNum, count = this.options.count) {
        let rest = count * current - total;
        return total && rest > 0 ? rest : count;
    }
    currentFix(current, pagesNum = this.pagesNum) {
        //根据传入的current值修正this.current
        if (!current) {
            //如果没有定义，则直接返回
            return this.current;
        }
        if (current <= 1) {
            //如果小于1那么就等于1
            this.current = 1;
        } else if (current >= pagesNum) {
            //如果当前页大于总页码，那么就等于总页码
            this.current = pagesNum;
        } else {
            this.current = current;
        }
        return this.current;
    }
    getCurrent(dataLen, current, pagesNum, itemsNum) {
        //获得项目总数
        this.itemsNum = itemsNum ? itemsNum : dataLen;
        //获得总页码数
        if (pagesNum) {
            this.pagesNum = pagesNum;
        } else {
            //向下取整，比如：5.2页则自动算成6，5则为5
            let divisor = this.itemsNum / this.options.count;
            this.pagesNum = divisor % 1 === 0 ? divisor : Math.floor(divisor) + 1;
        }
        //获取当前页码并修正当前页码
        this.currentFix(current ? current : this.current, this.pagesNum);
    }
    renderList(current, source,callback) {
        //根据实时的data渲染生成列表节点
        //使用文档碎片，提高效率
        let fragment = document.createDocumentFragment(),
            data = source || this.preData,
            //手动模式下，total数可能会少于data的length，则以total为准
            realData = this.options.total ? data.slice(0, this.options.total) : data;
        //['a','b','c','d','e','f'].slice(1,3) ==> ['b','c']
        this.selectData = realData.slice(this.options.count * (current - 1), this.options.count * current);
        this.selectData.forEach((item, index) => {
            //如果要删除当页最后一个节点中的某元素
            if (index + 1 == this.selectData.length && this.options.lastDel && item.querySelector(this.options.lastDel)) {
                item.removeChild(item.querySelector(this.options.lastDel))
            }
            fragment.appendChild(item);
        });

        //生成列表前处理
        this.options.onBeforeRendered && this.options.onBeforeRendered.call(this, this.selectData, this.asyncData);
        'beforeRendered' in this.handlers ? this.emit('beforeRendered', this.selectData, this.asyncData) : null;

        //先清空在追加节点
        this.listDom.innerHTML = '';

        //如果没有数据给出提示
        if (this.selectData.length === 0) {
            new axMessage({
                content: `有效数据只有${realData.length}条，本页没有数据，渲染数据失败！`,
                iconShow: true,
                result: 'error'
            }).show();
        } else {
            //渲染成功
            this.listDom.appendChild(fragment);
            //监听回调
            this.options.onRendered && this.options.onRendered.call(this, this.current, this.pagesNum,this.getOffset());
            this.handlers.hasOwnProperty('rendered') ? this.emit('rendered', this.current, this.pagesNum,this.getOffset()) : null;
            callback && callback.call(this,this.current, this.pagesNum,this.getOffset());
        }
    }
    renderPagesTotal(current, callback) {
        //先修正current
        current = this.currentFix(current);
        //先清空
        this.pagesDom.innerHTML = ''
        //根据实时的总页数渲染生成分页节点
        if (this.options.allPages) {
            //显示全部页码
            for (let i = 1; i <= this.pagesNum; i++) {
                let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                this.pagesDom.appendChild(item);
            }
        } else {
            //不显示全部页码
            if (this.pagesNum < 11) {
                //页面效果：1,2,3,4,5,6,7,8,9,10
                for (let i = 1; i <= this.pagesNum; i++) {
                    let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                    this.pagesDom.appendChild(item);
                }
            } else {
                if (current < 6) {
                    //页面效果：1,2,3,4,5,6,7,8...10,11
                    for (let i = 1; i <= 8; i++) {
                        let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                        this.pagesDom.appendChild(item);
                    }
                    this.pagesDom.appendChild(axAddElem('span', {}, '...'));
                    for (let i = this.pagesNum - 1; i <= this.pagesNum; i++) {
                        let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                        this.pagesDom.appendChild(item);
                    }
                } else if (this.pagesNum - current < 5) {
                    //页面效果：1,2,...21,22,23,24,25,26,27,28
                    for (let i = 1; i <= 2; i++) {
                        let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                        this.pagesDom.appendChild(item);
                    }
                    this.pagesDom.appendChild(axAddElem('span', {}, '...'));
                    for (let i = this.pagesNum - 6; i <= this.pagesNum; i++) {
                        let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                        this.pagesDom.appendChild(item);
                    }
                } else {
                    //页面效果：1,2...8,9,10,11,12...26,27
                    for (let i = 1; i <= 2; i++) {
                        let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                        this.pagesDom.appendChild(item);
                    }
                    this.pagesDom.appendChild(axAddElem('span', {}, '...'));
                    for (let i = current - 2; i <= current + 2; i++) {
                        let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                        this.pagesDom.appendChild(item);
                    }
                    this.pagesDom.appendChild(axAddElem('span', {}, '...'));
                    for (let i = this.pagesNum - 1; i <= this.pagesNum; i++) {
                        let item = axAddElem('a', { href: '###', 'data-page': i }, i);
                        this.pagesDom.appendChild(item);
                    }
                }
            }
        }
        //点击页码
        let pages = this.pagesDom.querySelectorAll('[data-page]');
        [...pages].forEach(item => {
            item.onclick = axDebounce(() => {
                //修正当前页码并获得当前分页标签的数组重新渲染
                this.current = ~~item.dataset.page; //item.dataset.page获得的数字字符串，用~~按位符转成数字
                //重新渲染列表节点和分页节点
                this.currentToRenderList(this.current);
            }, this.options.clickDelay);
        });
        //追加激活样式
        this.setClass(current);
        //回调
        callback && callback.call(this, current);
        return this.pagesDom;
    }
    setClass(current) {
        //根据当前页码给分页添加激活样式
        let pages = this.pagesDom.querySelectorAll('[data-page]'),
            page = [...pages].filter(item => ~~item.dataset.page == current)[0];
        //当前页面激活
        page.classList.add('ax-active');

        if (~~page.dataset.page == 1) {
            //如果是第一页，则首页和上一页显灰
            this.pageFirst.classList.add('ax-disabled');
            this.pagePrev.classList.add('ax-disabled');
            if (this.pagesNum == 1) {
                //总页数仅仅有一页
                this.pageLast.classList.add('ax-disabled');
                this.pageNext.classList.add('ax-disabled');
            } else {
                this.pageLast.classList.remove('ax-disabled');
                this.pageNext.classList.remove('ax-disabled');
            }
        } else if (~~page.dataset.page == this.pagesNum) {
            //如果是最后一页，则首页和上一页显灰
            this.pageLast.classList.add('ax-disabled');
            this.pageNext.classList.add('ax-disabled');
            this.pageFirst.classList.remove('ax-disabled');
            this.pagePrev.classList.remove('ax-disabled');
        } else {
            //其他正常
            this.pageFirst.classList.remove('ax-disabled');
            this.pagePrev.classList.remove('ax-disabled');
            this.pageLast.classList.remove('ax-disabled');
            this.pageNext.classList.remove('ax-disabled');
        }
    }

    getPagination() {
        //根据生成页码，追加上首页尾页等按钮
        let fragment = document.createDocumentFragment();

        //页码跳转
        this.pageGo.innerHTML = '跳到</span><input type="text"><button type="button" class="ax-btn">确定</button>';
        let goInput = this.pageGo.querySelector('input'),
            goBtn = this.pageGo.querySelector('button');
        goBtn.onclick = axDebounce(() => {
            let val = ~~goInput.value;
            if (val <= 1) {
                console.warn('The value is out of range,it has been fixed!');
                goInput.value = 1;
                this.current = 1;
            } else if (val >= this.pagesNum) {
                console.warn('The value is out of range,it has been fixed!');
                goInput.value = this.pagesNum;
                this.current = this.pagesNum;
            } else {
                this.current = val;
            }
            //重新渲染列表节点和分页节点
            this.currentToRenderList(this.current);
        }, this.options.clickDelay);
        //使用回车可执行跳转
        goInput.onkeyup = function (e) {
            if (e.code === 'Enter') {
                goBtn.click();
            }
        }
        //根据类型生成不同的分页节点
        if (this.options.type === 'regular') {
            //regular，类似：首页,上一页,1,2,3,...9,10下一页尾页
            fragment.appendChild(this.pageFirst);
            fragment.appendChild(this.pagePrev);
            fragment.appendChild(this.pagesDom);
            fragment.appendChild(this.pageNext);
            fragment.appendChild(this.pageLast);

        } else if (this.options.type === 'group') {
            //regular，基础上使用ax-group包裹
            this.group.appendChild(this.pageFirst);
            this.group.appendChild(this.pagePrev);
            this.group.appendChild(this.pagesDom);
            this.group.appendChild(this.pageNext);
            this.group.appendChild(this.pageLast);
            fragment.appendChild(this.group);
        } else if (this.options.type === 'simple') {
            //simple，只有上一页和下一页
            fragment.appendChild(this.pagePrev);
            fragment.appendChild(this.pageNext);

        } else if (this.options.type === 'full') {
            //full，类似：在regular基础上在加total
            fragment.appendChild(this.pageTotal);
            fragment.appendChild(this.pageFirst);
            fragment.appendChild(this.pagePrev);
            fragment.appendChild(this.pagesDom);
            fragment.appendChild(this.pageNext);
            fragment.appendChild(this.pageLast);
            fragment.appendChild(this.pageGo);
        } else if (this.options.type === 'pagesRight') {
            //pagesRight，弹性布局，页码放在右侧
            let left = axAddElem('div'),
                right = axAddElem('div', { class: 'ax-col ax-align-right' });
            this.targetDom.classList.add('ax-row');
            left.appendChild(this.pageTotal);

            right.appendChild(this.pageFirst);
            right.appendChild(this.pagePrev);
            right.appendChild(this.pagesDom);
            right.appendChild(this.pageNext);
            right.appendChild(this.pageLast);
            right.appendChild(this.pageGo);

            fragment.appendChild(left);
            fragment.appendChild(right);
        } else if (this.options.type === 'pagesLeft') {
            //pagesLeft，弹性布局，页码放在右侧
            let right = axAddElem('div'),
                left = axAddElem('div', { class: 'ax-col ax-align-left' });
            this.targetDom.classList.add('ax-row');
            right.appendChild(this.pageTotal);

            left.appendChild(this.pageFirst);
            left.appendChild(this.pagePrev);
            left.appendChild(this.pagesDom);
            left.appendChild(this.pageNext);
            left.appendChild(this.pageLast);
            left.appendChild(this.pageGo);

            fragment.appendChild(left);
            fragment.appendChild(right);
        } else {
            //pure，只有页码，在regular基础上减掉按钮
            fragment.appendChild(this.pagesDom);
        }
        this.targetDom.appendChild(fragment);
    }
    getJsonData(data = {}) {
        data.template = data.template || this.template;
        data.url = data.url || this.options.data;

        return axAjax({
            url: data.url,
            type: this.options.ajaxType,
            before: (response) => {
                this.listDom ? this.listDom.innerHTML = '<div class="ax-pageTurning">' + response.content + '</div>' : null;
            },
            success: (response) => {
                //先还原空数组
                let doms = [], dataLen;
                response.content.forEach((i, index) => {
                    let item, obj = { index: index };
                    if (!data.template && this.listDom.nodeName == 'TBODY') {
                        item = axAddElem('tr');
                        for (let k in i) {
                            let td = axAddElem('td', { name: k }, i[k])
                            item.appendChild(td);
                            //追加obj子元素
                            obj.children = [];
                            obj.children.push({ dom: td, name: k, label: i[k].toString() });
                        }
                    } else {
                        item = axStrToDom(axTplEngine(data.template, i));
                    }
                    //填充asyncData
                    obj.dom = item;
                    this.asyncData.push(obj);
                    //获得节点数组
                    doms.push(item);
                });
                //手动模式不需要改变preData
                this.dataType !== 'manual' ? this.preData = [...doms] : null;
                dataLen = this.preData.length;
                //通过以上步骤已经有了preData
                //更新pagesNum，itemsNum和current
                this.getCurrent(dataLen);
                data.callback && data.callback.call(this, doms, this.asyncData, response.content);
            }
        });
    }
    getSqlData(data = {}) {
        data.current = data.current || 1;
        data.count = data.count || this.options.count;
        data.total = data.total || this.options.total;
        //可自定义模板
        data.template = data.template || this.template;
        data.url = data.url || this.options.data;

        
        return axAjax({
            //根据参数进行数据扩展
            data: Object.assign({ count: data.count, current: data.current, offset:this.getOffset(data.current,data.total,data.count) }, this.options.dataExtend),
            url: data.url,
            type: this.options.ajaxType,
            before: (response) => {
                if (this.listDom) {
                    //加载容器装入loading图标
                    this.loadingDom.style = '';
                    this.loadingDom.innerHTML = response.content;

                    //模式不同，使用不同效果
                    if (this.options.loadingMode === 'overlay') {
                        this.loadingDom.classList.add('ax-overlay');
                    } else {
                        this.listDom.innerHTML = '';
                        this.listDom.appendChild(this.loadingDom)
                    }
                    this.listDom.insertAdjacentElement('afterbegin', this.loadingDom);
                    axFadeIn(this.loadingDom, 'flex');
                }
            },
            success: (response) => {
                //如果返回的不是一个{}对象，那么不再执行
                if (typeof response.content != 'object') {
                    console.error('The type of content must be object!');
                    return false;
                }
                //如果data是空数组，那么listDom和targetDom都清空
                if (axIsEmpty(response.content.data)) {
                    this.listDom ? this.listDom.innerHTML = '' : null;
                    //报错
                    new axMessage({
                        content: `没有从数据库中获得数据，已停止渲染列表！`,
                        iconShow: true,
                        result: 'error'
                    }).show();
                    //不再执行
                    return false;
                }

                //response.content的格式{data:[],current:'',pagesNum:'',itemsNum:''}
                //先清空
                this.dataType !== 'manual' ? this.preData = [] : null;
                this.asyncData = [];
                let doms = [], dataLen;
                //更新preData，当前页，总页码
                response.content.data.forEach((i, index) => {
                    let item, obj = { index: index };
                    if (!data.template && this.listDom.nodeName == 'TBODY') {
                        item = axAddElem('tr');
                        //定义children为数组
                        obj.children = [];
                        for (let k in i) {
                            let td = axAddElem('td', { name: k }, i[k].toString());
                            item.appendChild(td);
                            //追加obj子元素
                            obj.children.push({ dom: td, name: k, label: i[k].toString() });
                        }
                    } else {
                        item = axStrToDom(axTplEngine(data.template, i));
                    }
                    //填充asyncData
                    obj.dom = item;
                    this.asyncData.push(obj);
                    //填充preData
                    doms.push(item);
                });
                //手动模式不需要改变preData
                this.dataType !== 'manual' ? this.preData = [...doms] : null;
                dataLen = this.preData.length;
                //更新pagesNum、itemsNum和current
                this.getCurrent(dataLen, ~~response.content.current, ~~response.content.pagesNum, ~~response.content.itemsNum);
                data.callback && data.callback.call(this, doms, this.asyncData, response.content);
            }
        });
    }
    async currentToRenderList(current, callback) {
        //非手动模式并且有列表节点才渲染
        if (this.listDom && this.dataType !== 'manual') {
            //只重新渲染列表节点和分页节点，上一页下一页等不重新渲染
            if (this.options.async === 'sql') {
                //如果有异步地址和模板则使用异步方式
                await this.getSqlData({
                    current,
                    callback: () => {
                        //异步获取的preData只有一页，所以相当于只渲染第一页
                        //如果使用演示加载
                        if (this.options.delay) {
                            setTimeout(() => {
                                this.renderList(1)
                            }, this.options.delay)
                        } else {
                            this.renderList(1)
                        }
                    }
                });
            } else {
                this.renderList(current);
            }
        }
        this.renderPagesTotal(current, callback && callback.call(this, current));
        //监听回调
        this.options.onTurned && this.options.onTurned.call(this, this.current, this.pagesNum,this.getOffset());
        this.handlers.hasOwnProperty('turned') ? this.emit('turned', this.current, this.pagesNum,this.getOffset()) : null;
    }
    first(callback) {
        this.current = 1;
        //重新渲染列表节点和分页节点
        this.currentToRenderList(this.current);
        callback && callback.call(this, this.current, this.pagesNum);
        return this;
    }
    prev(callback) {
        if (this.current <= 1) {
            return false;
        } else {
            this.current = this.current - 1;
            //重新渲染列表节点和分页节点
            this.currentToRenderList(this.current);
            callback && callback.call(this, this.current, this.pagesNum);
        }
        return this;
    }
    next(callback) {
        if (this.current >= this.pagesNum) {
            return false;
        } else {
            this.current = this.current + 1;
            //重新渲染列表节点和分页节点
            this.currentToRenderList(this.current);
            callback && callback.call(this, this.current, this.pagesNum);
        }
        return this;
    }
    last(callback) {
        this.current = this.pagesNum;
        //重新渲染列表节点和分页节点
        this.currentToRenderList(this.current);
        callback && callback.call(this, this.current, this.pagesNum);
        return this;
    }
    go(page, callback) {
        if (this.current <= 1 || this.current >= this.pagesNum) {
            console.error('The value is out of range!');
            return false;
        } else {
            this.current = ~~page;
            //重新渲染列表节点和分页节点
            this.currentToRenderList(this.current);
            callback && callback.call(this, this.current, this.pagesNum);
        }
        return this;
    }
    info(callback) {
        let info = { current: this.current, pagesNum: this.pagesNum, count: this.options.count, itemsNum: this.itemsNum, selectData: this.selectData };
        callback && callback.call(this, info);
        return info;
    }
    updatePages(obj, type, callback) {
        //因为preData新增行或删除行则重新生成页码
        if (!obj) {
            return false;
        }
        type = type || 'remove';
        let item;
        if (typeof obj === 'number') {
            //如果obj是数字则从索引中找到目标对象
            item = this.preData[obj];
        } else {
            //如果obj是原数据中的一个对象
            item = obj;
        }
        if (!item) {
            return false;
        }
        if (type === 'remove') {
            //删除自己
            this.preData = this.preData.filter(i => i != item);
            //更新this.current，this.pagesNum和this.itemsNum
            this.getCurrent(this.preData.length, this.current, '');
        } else if (type === 'addBefore') {

        } else if (type === 'addAfter') {

        } else {

        }
        //更新列表节点、页码节点和统计节点
        this.currentToRenderList(this.current, () => {
            this.renderTotal();
        });
        //回调
        callback && callback.call(this);
    }
    async update(setting, mode, callback) {
        mode = mode || 'overwrite';
        this.options = axExtend(this.options, setting);
        //this.current并不会随着options.current变化而变化，所以需要特别指定
        this.current = this.options.current;
        //追加正在分页属性
        this.targetDom.setAttribute('paging', '');
        //整个分页，除了pagesDom会自动重写外，其他的比如上一页下一页按钮并不会重写，所以全部清零
        this.targetDom.innerHTML = '';
        //更新的数据是覆盖原数据还是在原数据后追加
        mode === 'overwrite' ? this.preData = [] : null;
        await this.init(this.options.data);
        //监听回调
        this.options.onUpdated && this.options.onUpdated.call(this, this.current, this.pagesNum,this.getOffset());
        'updated' in this.handlers ? this.emit('updated', this.current, this.pagesNum,this.getOffset()) : null;
        callback && callback.call(this);
        return this;
    }
    async updateData(data, mode, callback) {
        //data可是数据数组，#id，节点，动态页面地址，以及{url：'',sql:{}}对象（sql更新）
        mode = mode || 'overwrite';
        //追加正在分页属性
        this.targetDom.setAttribute('paging', '');
        //整个分页，除了pagesDom会自动重写外，其他的比如上一页下一页按钮并不会重写，所以全部清零
        this.targetDom.innerHTML = '';
        //更新的数据是覆盖原数据还是在原数据后追加,overwrite为覆盖则清空数据和sql扩展
        if (mode === 'overwrite') {
            this.preData = [];
            this.options.dataExtend = '';
        }
        //更新参数
        if (typeof data === 'object' && data.url && data.hasOwnProperty('sql')) {
            //如此传来的data应该是sql类型，data={url：'',sql:{sort:'order by click desc'}}
            this.options.data = data.url;
            this.options.dataExtend = data.sql;
        } else {
            this.options.data = data;
        }
        await this.init(this.options.data);
        //监听回调
        this.options.onUpdated && this.options.onUpdated.call(this, this.current, this.pagesNum,this.getOffset());
        'updated' in this.handlers ? this.emit('updated', this.current, this.pagesNum,this.getOffset()) : null;
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    // 对使用axPagination属性的元素使用本插件
    document.querySelectorAll('[axPagination]').forEach(element => {
        new axPagination(element);
    });
})(); */

/*!
* 插件：可排序列表；使用方法：new axList('#id',{参数})；#id是可选参数，可以从该dom获取参数值
*/
class axList {
    constructor(targetDom, options) {
        //列表Dom
        this.targetDom = axIdToDom(targetDom);
        //原始参数
        this.rawOptions = {
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            names: '',//列表排序筛选的字段名，多个用英文逗号隔开
            inputDelay: 200,//键入关键字检索时的时间时间，避免快速录入导致系统负荷
            clickDelay: 200,//点击按钮时的延时时间，避免快速连续多次点击导致系统负荷
            paginationDom: null,//分页的节点，id、className、节点名称或node
            maxCount: 10000,//一页的最大项目数量
            order: 'asc',//默认排序为顺序，对于有sort属性的按钮，没有指明order值那么将按该参数排序
            operator: 'or',//多条件的默认操作符，默认为and，可选择or
            paginationShow: true,//是否显示分页，默认true显示，如果为false表示不启用分页
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onUpdated: '',//渲染完list和pagination节点后回调，支持current和pagesNum参数
            onInit: '',//初始化后回调，支持current和pagesNum参数
            //以下为分页所需参数
            pagination: {
                current: 1, //当前页码
                count: 10, //每页item数量
                data: [], //用于创建列表Dom的数组，据此创建分页
                template: null, //数组转Dom的模板
                async: '',//异步方式，选择从json文件获取数组数据，或选择sql从数据库获得数据
                ajaxType: 'post',//异步提交的方式，默认post，可填get
                loadingMode: 'overlay',//使用异步分页的加载效果，默认是overlay即浮在原列表上，可选择replace即清除列表显示转圈
                //更多分页参数请参考axPagination
            }
        }
        this.options = axExtend(axClone(this.rawOptions), options, this.targetDom, this.constructor.name);
        //监听事件
        this.handlers = {};
        //原始数据
        this.rawData = [];
        //中转数组，从data创建列表Dom，并创建分页
        this.preData = [];
        //分页节点
        this.paginationDom = this.options.paginationDom ? axIdToDom(this.options.paginationDom) : this.targetDom.querySelector('.ax-pagination') || '';
        //可以从页面节点获取项目模板
        if (!this.options.pagination.template) {
            this.template = '';
        } else {
            if (this.options.pagination.template.trim().substr(0, 1) === '#') {
                this.template = document.querySelector(this.options.pagination.template).innerHTML
            } else {
                this.template = this.options.pagination.template;
            }
        }
        //去掉所有空格并转成数组
        this.names = this.options.names.replace(/\s+/g, '').split(',');
        this.listDom = this.targetDom.querySelector('[list]');
        this.searchs = [...this.targetDom.querySelectorAll('[search]')];
        this.keywords = [...this.targetDom.querySelectorAll('[keyword]')];
        this.searchVal = '';
        this.sorts = [...this.targetDom.querySelectorAll('[sort]')];
        this.resets = [...this.targetDom.querySelectorAll('[reset]')];
        this.filters = [...this.targetDom.querySelectorAll('[filter]')];
        this.extends = [...this.targetDom.querySelectorAll('[extend]')];
        //分页成员，分页参数和分页实例
        this.paginationOpt = {};
        this.pagination = null;
        //初始化
        this.init();
        return this;
    }
    async init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'list');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //将axList参数赋予axPagination
        this.setPaginationOpt();

        if (this.listDom.children.length > 0) {
            //如果页面存在列表ul+li或tbody+tr
            let data = [...this.listDom.childNodes].filter(item => item.nodeType == 1);
            //根据[dom]数据生成带排序字段的数据
            this.getDomArr(data);
            //将数据传递给分页插件data参数
            this.paginationOpt.data = this.preData;
        } else if (axType(this.options.pagination.data) === 'String' && !!this.options.pagination.template) {
            if (this.options.pagination.async === 'json') {
                //异步数据，json和sql类型有区别
                await axAjax({
                    url: this.options.pagination.data,
                    type: this.options.pagination.ajaxType,
                    success: (response) => {
                        let data = [];
                        response.content.forEach(i => {
                            let item = axStrToDom(axTplEngine(this.template, i));
                            data.push(item);
                        });
                        //根据[dom]数据生成带排序字段的数据
                        this.getDomArr(data);
                        this.paginationOpt.data = this.preData;
                    }
                });
            } else if (this.options.pagination.async === 'sql') {
                if (!['relative', 'absolute', 'fixed'].includes(axStyle(this.listDom).position)) {
                    //为了使用overlay风格的加载效果，需要给列表节点追加relative样式
                    this.listDom.style.position = 'relative';
                }
            }
        } else {
            //数据类型是[{title:'',click:'',href:''}]
            this.preData = this.options.pagination.data;
            this.rawData = [...this.preData];
            this.paginationOpt.data = this.options.pagination.data;
        }
        //根据数据执行分页和按钮操作
        this.renderFinish();
        //回调
        this.options.onInit && this.options.onInit.call(this, this.pagination.current, this.pagination.pagesNum);
    }
    getDomArr(listData) {
        //处理数据
        if (this.options.names) {
            //返回对象数组[{name:'',country:'',dom:''},{...}...]
            listData.forEach(item => {
                let obj = { dom: item };
                this.names.forEach(i => {
                    item.querySelector('[' + i + ']') ? obj[i] = item.querySelector('[' + i + ']').innerText.trim() : null;
                })
                this.preData.push(obj);
            });
            //保存原始数据
            this.rawData = [...this.preData];
            //有names才能排序、筛选和检索
            return true;
        } else {
            //直接返回只有节点的数组[dom,dom...]
            this.preData = listData;
            //保存原始数据
            this.rawData = [...listData];
            return false;
        }
    }
    renderFinish() {
        //执行分页插件
        this.pagination = new axPagination(this.paginationDom, this.paginationOpt);
        //如果不需要显示分页，那么将分页dom删除
        if (!this.options.paginationShow) {
            this.pagination.targetDom.remove();
            if (this.pagination.itemsNum > this.paginationOpt.count) {
                console.error('The maximum length cannot exceed ' + this.paginationOpt.count + '!');
                return false;
            }
        }
        let isSql = this.paginationOpt.async === 'sql' ? true : false;
        //自定义检索扩展
        this.loadExtend(isSql);
        //点击排序，重新获得preData，更新分页
        this.loadSort(isSql);
        //input输入关键字实时筛选
        this.loadKeyword(isSql);
        //搜索按钮（input和button分开的）
        this.loadSearch(isSql);
        //固定关键字筛选
        this.loadFilter(isSql);
        //点击恢复原始数据
        this.loadReset(isSql);
    }
    setPaginationOpt() {
        //先直接赋值
        this.paginationOpt = this.options.pagination;
        //再根据特点修改值
        this.paginationOpt.listDom = this.listDom;
        //如果不需要显示分页，那么将每页数量设置大一些以做到不需要做分页
        this.paginationOpt.count = !this.options.paginationShow ? this.options.maxCount : this.paginationOpt.count;
        return this.paginationOpt;
    }
    loadExtend(isSql) {
        if (this.extends.length !== 0) {
            this.extends.forEach(item => {
                let attr = item.getAttribute('extend');
                //使用节流，连续点击按钮会间隔执行
                item.onclick = axDebounce(() => {
                    this.extend(attr, isSql);
                }, this.options.clickDelay);
            });
        }
    }
    extend(str, isSql) {
        if (!isSql) {
            this.strSearch(str);
        } else {
            //跟新动态页面的sql数据
            this.updateSql({ dataExtend: { extend: str } });
            //监听分页是否有数据，如果没有则清空列表和分页
            this.pagination.on('rendered', () => {
                if (axIsEmpty(this.pagination.preData)) {
                    this.listDom.innerHTML = '';
                    this.pagination.targetDom.innerHTML = '';
                }
            });
        }
    }
    loadKeyword(isSql) {
        if (this.keywords.length !== 0) {
            this.keywords.forEach(item => {
                let source = [...this.preData],
                    attr = item.getAttribute('keyword').trim(),
                    operator = item.getAttribute('operator') ? item.getAttribute('operator').trim() : this.options.operator,
                    value = '',
                    arr = attr ? attr.split(',') : [],
                    names = arr.length !== 0 ? arr : this.names;
                //使用auto属性则会自动检索
                if (item.hasAttribute('auto')) {
                    //使用防抖，连续键入延迟处理
                    item.oninput = axDebounce(() => {
                        let inputValue = item.value;
                        if (arr.length !== 0) {
                            let obj = {};
                            arr.forEach(i => {
                                obj[i] = inputValue;
                            });
                            value = obj;
                        } else {
                            value = inputValue;
                        }
                        if (!isSql) {
                            //更新非sql数据
                            this.keySearch(value, source, operator);
                        } else {
                            //更新动态页面的sql数据
                            this.sqlSearch(names, inputValue, operator);
                        }

                    }, this.options.inputDelay);
                }
            });
        }
    }
    loadFilter(isSql) {
        if (this.filters.length !== 0) {
            this.filters.forEach(item => {
                let data,
                    source = [...this.preData],
                    operator = item.getAttribute('operator') ? item.getAttribute('operator').trim() : this.options.operator;
                if (!isSql) {
                    //从按钮属性filter={}获取检索条件
                    data = axAttrToJson(item, 'filter');
                } else {
                    data = item.getAttribute('filter');
                }
                item.onclick = axDebounce(() => {
                    this.filter(data, source, operator, isSql);
                }, this.options.clickDelay);
            });
        }
    }
    filter(data, source, operator, isSql) {
        operator = operator ? operator : this.options.operator;
        source = source ? source : [...this.preData];
        if (!isSql) {
            //更新非sql数据
            this.keySearch(data, source, operator);
        } else {
            //更新动态页面的sql数据
            this.updateSql({ dataExtend: { filter: data } });
        }
    }
    loadSearch(isSql) {
        if (this.searchs.length !== 0) {
            this.searchs.forEach(item => {
                //没有节点则不执行
                if (this.keywords.length == 0) {
                    return false;
                }
                let fieldName = item.getAttribute('field'),
                    fieldDom = this.keywords.filter(i => i.name == fieldName)[0],
                    searchNames = fieldDom.getAttribute('keyword'),
                    operator = fieldDom.getAttribute('operator') ? fieldDom.getAttribute('operator').trim() : this.options.operator,
                    source = [...this.preData];
                //如果定义了search的指定字段则使用自定义的，如果没有则对所有字段检索
                searchNames = searchNames ? searchNames.replace(/\s+/g, '').split(',') : this.names;
                //点击搜索按钮检索
                item.onclick = axDebounce(() => {
                    this.search(searchNames, fieldDom.value, source, operator, isSql);
                }, this.options.clickDelay);
                //如果input是输入状态，可用回车代替搜索按钮
                fieldDom.onkeyup = (e) => {
                    //回车键代码是13
                    if (e.code === 'Enter') {
                        item.click();
                    }
                }
            });
        }
    }
    search(names, value, source, operator, isSql) {
        operator = operator ? operator : this.options.operator;
        if (!isSql) {
            let obj = {};
            source = source ? source : [...this.preData];
            names.forEach(i => {
                obj[i] = value;
            });
            this.keySearch(obj, source, operator);
        } else {
            this.sqlSearch(names, value, operator);
        }
    }
    loadSort(isSql) {
        if (this.sorts.length !== 0) {
            for (let item of this.sorts) {
                let attr = item.getAttribute('sort').trim();
                if (!isSql) {
                    //非sql数据，如果sort=name，值不在names数组里，则跳过
                    if (this.names.every(i => i !== attr)) {
                        continue;
                    }
                }
                item.onclick = axDebounce(() => {
                    if (!isSql) {
                        //数字->英文->中文
                        if (item.getAttribute('order') === 'desc') {
                            //第一次点击默认顺序
                            item.setAttribute('order', 'asc');
                            //对preData重新排序
                            this.sort(attr, 'asc', false);
                        } else if (item.getAttribute('order') === 'asc') {
                            item.setAttribute('order', 'desc');
                            this.sort(attr, 'desc', false);
                        } else {
                            item.setAttribute('order', this.options.order);
                            this.sort(attr, this.options.order, false);
                        }
                    } else {
                        //跟新动态页面的sql数据
                        this.sort(attr, '', true);
                        if (!attr.includes(',')) {
                            if (attr.includes('desc')) {
                                item.setAttribute('order', 'desc');
                                attr = attr.replace('desc', 'asc');
                            } else if (attr.includes('asc')) {
                                item.setAttribute('order', 'asc');
                                attr = attr.replace('asc', 'desc');
                            }
                        }
                    }
                    //其他sort按钮取消order
                    this.sorts.filter(i => i != item).forEach(k => {
                        k.setAttribute('order', '');
                    });
                }, this.options.clickDelay);
            }
        }
    }
    sort(data, order, isSql) {
        if (!isSql) {
            this.preSort(data, order);
        } else {
            this.updateSql({ dataExtend: { sort: data } });
        }
    }
    preSort(name, order, callback) {
        //数字->英文->中文
        if (order === 'desc') {
            //降序排列
            this.preData.sort(function (a, b) {
                if (isNaN(a[name] - b[name])) {
                    //如果是字母、中文、特殊字符，a[name] - b[name] == NaN
                    return b[name].localeCompare(a[name], 'zh-CN-u-kf-lower');
                } else {
                    //如果是数字字符串
                    return b[name] - a[name];
                }
            })
        } else if (order === 'asc') {
            //升序排列
            this.preData.sort(function (a, b) {
                if (isNaN(a[name] - b[name])) {
                    //如果是字母、中文、特殊字符，a[name] - b[name] == NaN
                    return a[name].localeCompare(b[name], 'zh-CN-u-kf-lower');
                } else {
                    //如果是数字字符串
                    return a[name] - b[name];
                }
            })
        } else {
            return false;
        }

        this.pagination.updateData(this.preData);
        //回调
        this.handlers.hasOwnProperty('sorted') ? this.emit('sorted', this.pagination.current, this.pagination.pagesNum) : null;
        this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
        callback && callback.call(this, this.preData, this.pagination.current);
        return this;
    }
    loadReset(isSql) {
        if (this.resets.length !== 0) {
            this.resets.forEach(item => {
                item.onclick = axDebounce(() => {
                    this.reset(isSql);
                }, this.options.clickDelay);
            });
        }
    }
    reset(isSql) {
        if (!isSql) {
            //更新数据
            this.preData = [...this.rawData];
            let obj = axClone(this.rawOptions.pagination);
            obj.data = this.preData;
            this.updatePages(obj);
        } else {
            //更新动态页面的sql数据
            this.updateSql({ dataExtend: '' });
        }
        //toggle的排序按钮取消desc和asc属性
        this.sorts.forEach(i => {
            i.setAttribute('order', '');
        });
        //清空input
        this.keywords.forEach(i => {
            i.value = '';
        });
        //监听回调
        'reset' in this.handlers ? this.emit('reset', '') : null;
        this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
    }
    keySearch(value, sourceArr, operator = this.options.operator, callback) {
        let filterArr = sourceArr.filter(i => {
            let values = [];
            if (typeof value === "object") {
                //固定关键字检索，value是一个对象：{country:'china',add:''}
                let flag = [];
                for (let k in value) {
                    //数字不能使用includes方法，所以都转成字符串
                    flag.push(String(i[k]).includes(value[k]));
                }
                if (operator === 'or') {
                    //有一个true即可
                    return flag.some(k => k == true);
                } else if (operator === 'and') {
                    //都为true才行
                    return flag.every(k => k == true);
                }
            } else {
                //input关键字检索，不需要指定name
                for (let k of this.names) {
                    values.push(i[k]);
                }
                //数组转字符串，判断是否包含录入的关键字
                return values.toString().includes(value);
            }
        });
        this.preData = filterArr;
        if (axIsEmpty(this.preData)) {
            //没有搜索到数据，preData=[]，传到分页插件会出错，在此阻止
            'rendered' in this.handlers ? this.emit('rendered', this.pagination.current, this.pagination.pagesNum) : null;
            this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
            callback && callback.call(this, this.pagination.current, this.pagination.pagesNum);
            //将原始数据复制给preData，以恢复搜索能力
            this.preData = [...this.rawData];
            //清空列表和分页的Dom
            this.listDom.innerHTML = '';
            this.pagination.targetDom.innerHTML = '';
            return false;
        } else {
            this.pagination.updateData(this.preData);
            'rendered' in this.handlers ? this.emit('rendered', this.pagination.current, this.pagination.pagesNum) : null;
            this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
            callback && callback.call(this, this.pagination.current, this.pagination.pagesNum);
        }
    }
    strSearch(str, callback) {
        //字符串转filter语句，从preData获取数据
        this.preData = axStrToFilter(str, [...this.preData]);
        if (axIsEmpty(this.preData)) {
            //没有搜索到数据，preData=[]，传到分页插件会出错，在此阻止
            'rendered' in this.handlers ? this.emit('rendered', this.pagination.current, this.pagination.pagesNum) : null;
            this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
            callback && callback.call(this, this.pagination.current, this.pagination.pagesNum);
            //将原始数据复制给preData，以恢复搜索能力
            this.preData = [...this.rawData];
            //清空列表和分页的Dom
            this.listDom.innerHTML = '';
            this.pagination.targetDom.innerHTML = '';
            return false;
        } else {
            this.pagination.updateData(this.preData);
            'rendered' in this.handlers ? this.emit('rendered', this.pagination.current, this.pagination.pagesNum) : null;
            this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
            callback && callback.call(this, this.pagination.current, this.pagination.pagesNum);
        }
    }
    sqlSearch(names, value, operator) {
        let dataString = '',
            last;
        operator = operator || this.options.operator;
        //如果names.length=0，停止检索
        if (names.length == 0) {
            console.error('The length of names array cannot be greater than 0!')
            return false;
        }
        if (value) {
            names.forEach((i, index) => {
                last = names.length - 1 != index ? operator + ' ' : '';
                dataString += i + ' like "%' + value + '%" ' + last;
            });
        }
        //先清空异步扩展参数
        this.pagination.options.dataExtend = '';
        this.updatePages({ dataExtend: { search: dataString } });
        //监听分页是否有数据，如果没有则清空列表和分页
        this.pagination.on('rendered', () => {
            if (axIsEmpty(this.pagination.preData)) {
                this.listDom.innerHTML = '';
                this.pagination.targetDom.innerHTML = '';
            }
        })
    }
    updateSql(extend, callback) {
        //extend={dataExtend:{sort:''}}
        //先清空异步扩展参数
        this.pagination.options.dataExtend = '';
        this.pagination.update(extend, '', () => {
            //监听回调
            this.handlers.hasOwnProperty('updated') ? this.emit('updated', this.pagination.current, this.pagination.pagesNum) : null;
            this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
        });
        callback && callback.call(this, this.pagination.current, this.pagination.pagesNum);
        return this;
    }
    updatePages(setting, callback) {
        let opt = axExtend(this.options.pagination, setting);
        this.pagination.update(opt, '', () => {
            //监听回调
            this.handlers.hasOwnProperty('updated') ? this.emit('updated', this.pagination.current, this.pagination.pagesNum) : null;
            this.options.onUpdated && this.options.onUpdated.call(this, this.pagination.current, this.pagination.pagesNum);
        });
        callback && callback.call(this, this.pagination.current, this.pagination.pagesNum);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    // 对使用axList属性的元素使用本插件
    document.querySelectorAll('[axList]').forEach(element => {
        new axList(element);
    });
})(); */
/*!a
 * 插件：灯箱画廊；使用方法：axLightbox('#id',{参数})
 */
class axLightbox {
    constructor(options, elem) {
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            src: '',//内容来源，可以是一个图片地址，也可以多个用英文逗号“,”隔开的图片地址；也可以是[{media:'',cover:'',caption:'',type:''}]数组类型的值
            //cover是封面图片，将在小图导航显示，如果没有值则使用“暂无图片”替代
            //src.type可选择video，audio，iframe，text,image，为空或不存在按image创建Dom
            gallery: true,//是否使用多页切换的画廊模式，默认否true即为画廊模式，可选择false则只显示一个slide并不能轮播
            download: false,//是否显示图片下载按钮，默认不显示false，可选择true显示
            autoplay: false,//图片集是否自动播放，默认手动false，可选择true
            thumb: {
                show: false,//是否显示小图集，默认不显示false，可选择true
                placement: 'bottom',
                row: 1,//小图集按几行显示，默认按一行显示1，可选择2、3...
            },
            //thumb:false,表示不使用小图导航
            zoom: true,//是否显示缩放图标，默认显示true，可选择false
            share: false,//是否显示分享按钮，默认false不显示，可选择true显示
            btnClass: 'ax-lightbox-open',//触发按钮在打开灯箱后新增的样式名
            index: 0,//初始化后，显示第几页
            zIndex: 0,//气泡层级，默认为0即不设置，如果与其他组件发生遮挡请自行设置
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onShow: '',//回调函数，开始显示后执行，无参数
            onShown: '',//回调函数，完成显示后执行，无参数
            onHide: '',//回调函数，开始关闭后执行，无参数
            onHidden: '',//回调函数，完全关闭后执行，无参数
            onUpdate: '',//回调函数，参数/内容/位置更新前执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, elem, this.constructor.name);
        // 定义监听事件成员
        this.handlers = {};
        //通过src创建data
        this.data = [];
        //定义可点击关闭灯箱的遮罩区域
        this.closeArea = [];
        //定义图片是否是放大状态，默认false
        this.zoomIn = false;
        // 获得axLightbox元素的id
        this.button = elem ? axIdToDom(elem) : null;
        //显示状态
        this.shown = false;
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'lightbox');
    }
    init() {
        let _this = this;
        //设置销毁状态
        this.destroyed = false;
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;

        //将用户传过来的src值转成数组数据，以便于统一调用
        this.data = axCreateData(this.options.src, ['src', 'cover', 'caption', 'type'], ['media', 'cover', 'caption', 'type'], 'figure', ['', '', '', 'image'], function (data) {
            //如果没有cover，将image类型的media赋予cover
            data.forEach(item => {
                if (!item.cover && item.type == 'image') {
                    item.cover = item.media;
                }
                //如果没有封面图则使用默认图片
                item.cover = item.cover ? item.cover : '';
            });
        });

        //重新确认参数
        this.setAttribute();

        //创建lightbox的主体结构，并定义了主要节点Dom
        this.createBox();

        //大图集的wrapper追加slide（包括caption）
        this.createSlide();

        //如果有多条轮播数据则使用前后按钮
        if (this.data.length > 1) {
            this.prev = axStrToDom(`<span class="ax-prev ax-light"></span>`);
            this.next = axStrToDom(`<span class="ax-next ax-light"></span>`);
            this.swiper.appendChild(this.prev);
            this.swiper.appendChild(this.next);
        }

        //定义大图集和小图集的参数
        let swiperOpt = {
            zoom: this.options.zoom,
            lazy: true,
            keyboard: true,
            mousewheel: true,
            initialSlide: this.options.index || 0,
            pagination: {
                el: this.options.gallery ? this.targetDom.querySelector('.ax-pages') : null,
                type: "fraction",
            },
            thumbs: {
                swiper: '',
            },
            on: {
                touchStart:  (swiper, event)=> {
                    swiper.el.classList.add('ax-grabbing');
                },
                touchEnd: (swiper, event)=> {
                    swiper.el.classList.remove('ax-grabbing');
                },
                zoomChange:  (swiper, scale)=> {
                    if (scale > 1) {
                        this.zoomIn = true;
                    } else {
                        //避免立即this.zoomIn = false，而导致触发关闭灯箱
                        setTimeout( ()=> {
                            this.zoomIn = false;
                        }, 100)
                    }
                },
            },
        }
        let thumbOpt = {
            lazy: true,
            spaceBetween: 4,
            slidesPerView: "auto",
            mousewheel: true,
            grid: {
                rows: this.options.thumb.row,
            },
            watchSlidesProgress: true,
        }

        //轮播模式并且thumb为true才显示小图集，只有一个slide也不显示小图集
        if (this.options.gallery && this.options.thumb) {
            this.stage.insertAdjacentElement('afterend', this.createThumb());

            //初始化小图集
            this.insThumb = new axSwiper(this.thumb.querySelector('.ax-swiper'), thumbOpt);

            //根据图集参数补充大图集参数（thumbs.swiper）
            swiperOpt.thumbs.swiper = this.insThumb;

            //点击list按钮显示和隐藏thumb
            this.btnList.onclick =  ()=> {
                this.targetDom.classList.toggle('ax-thumb-hide');
            }
        }
        //初始化，如果生成了thumb，如果不显示那么追加ax-thumb-hide样式
        if (!this.options.thumb.show) {
            this.targetDom.classList.add('ax-thumb-hide');
        }

        //根据是否自动轮播修正大图集参数
        if (this.options.autoplay) {
            swiperOpt.autoplay = {
                delay: 5000,
                stopOnLastSlide: false,
                disableOnInteraction: true,
            }
        }

        //初始化大图集
        this.insSwiper = new axSwiper(this.swiper, swiperOpt);

        // 在准备Dom和内容之后创建html节点
        document.body.appendChild(this.targetDom);

        //在body添加节点后，根据小图集的行数确定小图集swiper的高度
        if (this.data.length > 1 && this.options.thumb.row && this.options.thumb.row > 1) {
            this.thumb.querySelector('.ax-swiper').style.height = axHeight(this.thumb.querySelector('.ax-slide'), 'outer') * this.options.thumb.row + thumbOpt.spaceBetween * (this.options.thumb.row - 1) + 'px';
        }
        //如果gallery为false则以单页显示
        if (!this.options.gallery) {
            this.insSwiper.disable();
        }

        //关闭灯箱
        this.targetDom.querySelectorAll('[close]').forEach(elem => {
            elem.addEventListener('click', ()=> {
                this.hide();
            }, false);
        });


        //如果zoom为true则使用该功能，在下方代码修正zoom按钮样式
        if (this.options.zoom) {
            this.btnZoom.onclick = function () {
                if (this.classList.contains('ax-icon-zoomin')) {
                    _this.insSwiper.zoom.in();
                } else {
                    _this.insSwiper.zoom.out();
                }
            }
        }

        /*     if (this.options.list && this.thumb) {
                this.btnList.onclick = function () {
                    _this.targetDom.classList.toggle('ax-thumb-hide');
                }
            } */
        //加载完成初始化
        if (this.options.download) {
            if (this.data[this.insSwiper.activeIndex].type != 'image') {
                this.btnDownload.href = '###';
                this.btnDownload.classList.add('ax-disabled');
                this.btnDownload.removeAttribute('download');
                this.btnDownload.removeAttribute('target');
            } else {
                this.btnDownload.href = this.data[this.insSwiper.activeIndex].media;
                this.btnDownload.classList.remove('ax-disabled');
                this.btnDownload.setAttribute('download', '');
                this.btnDownload.setAttribute('target', '_blank');
            }
        }
        if (this.options.zoom) {
            if (this.data[this.insSwiper.activeIndex].type != 'image') {
                this.btnZoom.classList.add('ax-disabled');
                this.insSwiper.zoom.disable();
            } else {
                this.btnZoom.classList.remove('ax-disabled');
                this.insSwiper.zoom.enable();
            }
        }
        if (this.options.share) {
            if (this.data[this.insSwiper.activeIndex].type != 'image') {
                this.btnShare.classList.add('ax-disabled');
            } else {
                this.btnShare.classList.remove('ax-disabled');
            }
        }


        //切换完成执行事件
        this.insSwiper.on('slideChangeTransitionEnd', function () {
            //根据当前的slide改变download按钮的地址,不是image类型就不使用下载按钮
            if (_this.options.download) {
                if (_this.data[this.activeIndex].type != 'image') {
                    _this.btnDownload.href = '###';
                    _this.btnDownload.classList.add('ax-disabled');
                    _this.btnDownload.removeAttribute('download');
                    _this.btnDownload.removeAttribute('target');

                } else {
                    _this.btnDownload.href = _this.data[this.activeIndex].media;
                    _this.btnDownload.classList.remove('ax-disabled');
                    _this.btnDownload.setAttribute('download', '');
                    _this.btnDownload.setAttribute('target', '_blank');
                }
            }

            //非image类型，不需要使用zoom功能
            if (_this.options.zoom) {
                if (_this.data[this.activeIndex].type != 'image') {
                    _this.btnZoom.classList.add('ax-disabled');
                    _this.insSwiper.zoom.disable();
                } else {
                    _this.btnZoom.classList.remove('ax-disabled');
                    //切换后恢复zoom按钮初始状态
                    _this.btnZoom.classList.remove('ax-icon-zoomout');
                    _this.btnZoom.classList.add('ax-icon-zoomin');
                    _this.insSwiper.zoom.enable();
                }
            }
            //非image类型，不需要使用share功能
            if (_this.options.share) {
                if (_this.data[this.activeIndex].type != 'image') {
                    _this.btnShare.classList.add('ax-disabled');
                } else {
                    _this.btnShare.classList.remove('ax-disabled');
                }
            }
            //切换后停止媒体声音
            _this.targetDom.querySelectorAll('video,audio').forEach(elem => {
                elem.pause();
            });
            //切换后所有的放大效果复原
            _this.zoomIn = false;


        })
        //当zoom完成时修正zoom按钮的样式,
        this.insSwiper.on('zoomChange',  (ins, scale)=> {
            if (scale > 1) {
                this.btnZoom.classList.remove('ax-icon-zoomin');
                this.btnZoom.classList.add('ax-icon-zoomout');
            } else {
                this.btnZoom.classList.remove('ax-icon-zoomout');
                this.btnZoom.classList.add('ax-icon-zoomin');
            }
        })

        //回调
        this.options.onInit && this.options.onInit.call(this);
        'init' in this.handlers ? this.emit('init', '') : null;
        return this;

    }
    overlayClose(ev) {
        this.closeArea = [];

        let booPrev = true,
            booNext = true,
            booThumb = true;

        if (this.prev) {
            booPrev = axIsOutside(ev, this.prev);
        }
        this.closeArea.push(booPrev);
        //
        if (this.next) {
            booNext = axIsOutside(ev, this.next)
        }
        this.closeArea.push(booNext);
        //
        if (this.options.thumb) {
            booThumb = axIsOutside(ev, this.thumb);
        } else {
            booThumb = true;
        }
        this.closeArea.push(booThumb);
        // 
        this.swiperWrapper.querySelectorAll('.ax-media').forEach(item => {
            let dom = item.firstElementChild;
            this.closeArea.push(axIsOutside(ev, dom));
        })
        //
        if (this.swiperWrapper.querySelectorAll('.ax-caption')) {
            this.swiperWrapper.querySelectorAll('.ax-caption').forEach(item => {
                this.closeArea.push(axIsOutside(ev, item));
            })
        }
        //
        this.closeArea.push(axIsOutside(ev, this.operate));
        this.closeArea.push(axIsOutside(ev, this.pages));

        return this.closeArea;

    }
    createSlide() {

        let slideTpl = (src, type) => {
            if (type && type == 'video') {
                return `
                <div class="ax-slide">
                    <div class="ax-media">
                        <video controls><source src="${src}"></video>
                    </div>
                </div>
                `
            } else if (type && type == 'audio') {
                return `
                <div class="ax-slide">
                    <div class="ax-media">
                        <audio controls><source src="${src}" ></audio>
                    </div>
                </div>
                `
            } else if (type && type == 'iframe') {
                return `
                <div class="ax-slide">
                    <div class="ax-media">
                        <iframe src="${src}" frameborder="0" width="100%" height="100%"></iframe>
                    </div>
                </div>
                `
            } else if (type && type == 'text') {
                //如果是#id，则从页面dom获取内容
                if (src.substr(0, 1) == "#") {
                    src = document.querySelector(src).innerHTML;
                }
                return `
                <div class="ax-slide">
                    <div class="ax-media">
                        <div class="ax-text">${src}</div>
                    </div>
                </div>
                `
            } else {
                return `
                <div class="ax-slide">
                    <div class="ax-media ax-zoom">
                        <img data-src="${src}" class="ax-lazy" />
                        <span class="ax-loading ax-white"><i></i></span>
                    </div>
                </div>
                `
            }
        }

        //根据data创建所有slide
        this.data.forEach(item => {
            let slide = null,
                caption = null;
            slide = axStrToDom(slideTpl(item.media, item.type));
            //有简介则追加到slide里
            caption = item.caption ? axStrToDom(`<div class="ax-caption">${item.caption}</div>`) : null;
            caption ? slide.appendChild(caption) : null;

            this.swiperWrapper.appendChild(slide);

        })
    }
    createThumb() {
        //创建thumb的Dom
        this.thumb = axStrToDom('<div class="ax-thumb"><div class="ax-swiper"><div class="ax-wrapper"></div></div></div>');
        //定义thumb的wrapper
        this.thumbWrapper = this.thumb.querySelector('.ax-wrapper');

        this.data.forEach(item => {
            let slide = axStrToDom(`
            <div class="ax-slide">
                <img data-src="${item.cover}" class="ax-lazy" />
                <span class="ax-loading ax-white"><i></i></span>
            </div>
            `);
            this.thumbWrapper.appendChild(slide);
        })
        return this.thumb;
    }
    createBox() {
        let boxTpl = `
        <div class="ax-lightbox">
        <div class="ax-overlay"></div>
        <div class="ax-tools">
            <div class="ax-pages"></div>
            <div class="ax-operate"><i class="ax-iconfont ax-icon-close" close></i></div>
        </div>
        <div class="ax-stage">
            <div class="ax-swiper">
                <div class="ax-wrapper">
                </div>
            </div>
        </div>
    </div>
      `;
        //创建lightbox的Dom
        this.targetDom = axStrToDom(boxTpl);
        this.tools = this.targetDom.querySelector('.ax-tools');
        this.pages = this.targetDom.querySelector('.ax-pages');
        this.operate = this.targetDom.querySelector('.ax-operate');
        this.stage = this.targetDom.querySelector('.ax-stage');
        this.swiper = this.stage.querySelector('.ax-swiper');
        this.swiperWrapper = this.swiper.querySelector('.ax-wrapper');
        //创建按钮
        this.createBtn();
    }
    createBtn() {
        this.btnShare = axStrToDom('<i class="ax-iconfont ax-icon-share-f" share></i>');
        this.btnDownload = axStrToDom('<a href="###" class="ax-iconfont ax-icon-download" target="_blank" download></a>');
        this.btnZoom = axStrToDom('<i class="ax-iconfont ax-icon-zoomin" zoom></i>');
        this.btnList = axStrToDom('<i class="ax-iconfont ax-icon-list" list></i>');

        //有小图集才需要显示list按钮
        this.options.thumb ? this.operate.insertAdjacentElement('afterbegin', this.btnList) : null;

        this.options.download ? this.operate.insertAdjacentElement('afterbegin', this.btnDownload) : null;

        this.options.zoom ? this.operate.insertAdjacentElement('afterbegin', this.btnZoom) : null;

        this.options.share ? this.operate.insertAdjacentElement('afterbegin', this.btnShare) : null;


    }
    setAttribute() {
        //所有slide都必须是轮播图才启用小图集、下载按钮、缩放按钮
        for (let i = 0, len = this.data.length; i < len; i++) {
            if (this.data[i].hasOwnProperty('type') && (['iframe', 'video', 'text'].includes(this.data[i].type))) {
                this.options.thumb = false;
            }
            break;
        }
        //数据长度小于2，也就是只有一页或没有，则不需要显示图集
        if (this.data.length < 2) {
            this.options.thumb == false;
        }
        //设置zIndex
        this.options.zIndex ? this.targetDom.style.zIndex = this.options.zIndex : null;
    }
    update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }

        this.options = axExtend(this.options, setting);
        this.init();
        this.options.onUpdate && this.options.onUpdate.call(this);
        'update' in this.handlers ? this.emit('update', '') : null;
        callback && callback.call(this);
        return this;
    }

    show(callback) {
        //如果已经销毁或本身就是打开状态，或还没有生成节点，或是虚拟节点则不再执行
        if (this.destroyed || this.shown || !this.targetDom || axIsVnode(this.targetDom)) {
            return this;
        }

        this.options.onShow && this.options.onShow.call(this);
        'show' in this.handlers ? this.emit('show', '') : null;
        //执行窗口打开动画
        this.targetDom.style.visibility = 'visible';
        this.targetDom.classList.add('ax-show');
        //如果使用按钮触发气泡，那么给气泡赋予样式
        this.button ? this.button.classList.add(this.options.btnClass) : null;

        callback && callback.call(this);
        this.options.onShown && this.options.onShown.call(this);
        'shown' in this.handlers ? this.emit('shown', '') : null;

        //点击遮罩区域关闭灯箱

        this.targetDom.onclick =  (ev)=> {
            //每次点击都更新可点击区域
            this.overlayClose(ev);
            let closeArea = this.closeArea.includes(false);
            //全部为true，并且是非放大状态才可关闭
            if (!closeArea && !this.zoomIn) {
                this.hide();
            }
        }

        //更新显示状态
        this.shown = true;

        //返回实例，以便于链式操作
        return this;
    }
    hide(callback) {
        //如果已经销毁，或者本身就是关闭状态则不再执行
        if (this.destroyed || !this.shown) {
            return this;
        }

        this.options.onHide && this.options.onHide.call(this);
        'hide' in this.handlers ? this.emit('hide', '') : null;
        //如果使用按钮触发气泡，那么删除样式
        this.button ? this.button.classList.remove(this.options.btnClass) : null;

        //执行窗口关闭动画
        this.targetDom.classList.remove('ax-show');

        //移除DOM，
        //*2是确保动画结束了
        /*         setTimeout(function () {
                    _this.targetDom.remove();
                }, parseFloat(axStyle(_this.targetDom).transitionDuration) * 1000 * 2)
                 */
        this.targetDom.addEventListener('transitionend',  ()=> {
            //过渡动画结束，清除Dom
            this.targetDom.remove();
        }, false)

        //更新显示状态
        this.shown = false;

        //关闭窗口同时暂停媒体播放
        this.targetDom.querySelectorAll('video,audio').forEach(elem => {
            elem.pause();
        });
        callback && callback.call(this);
        this.options.onHidden && this.options.onHidden.call(this);
        'hidden' in this.handlers ? this.emit('hidden', '') : null;
        //返回实例，以便于链式操作
        return this;
    }
    destroy() {
        //删除对应的DOM
        this.targetDom.remove();
        callback && callback.call(this);
        this.options.onDestroy && this.options.onDestroy.call(this);
        'destroy' in this.handlers ? this.emit('destroy', '') : null;
        //更新销毁状态
        this.destroyed = true;
        //返回实例，以便于链式操作
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
(() => {
    //从a标签的href获得图片地址,标题和类型并创建数组
    let getSource = (element) => {
        let elem = axIdToDom(element),
            obj = [{}],
            node = elem.nodeName;
        //判断该标签是a还是img
        if (node == 'A') {
            if (elem.hasAttribute('href') && elem.getAttribute('href').indexOf('##') == -1) {
                obj[0].media = elem.getAttribute('href');
                obj[0].cover = elem.getAttribute('href');
            }
        } else if (node == 'IMG') {
            if (elem.hasAttribute('src')) {
                obj[0].media = elem.getAttribute('src');
                obj[0].cover = elem.getAttribute('src');
            }
        }
        //获得默认type属性值
        if (elem.hasAttribute('type')) {
            obj[0].type = elem.getAttribute('type');
        } else {
            obj[0].type = 'image';
        }
        //获得标题，优先从title获取，如果没有从alt获取
        if (elem.hasAttribute('title')) {
            obj[0].caption = elem.getAttribute('title');
        } else if (elem.hasAttribute('alt')) {
            obj[0].caption = elem.getAttribute('alt');
        } else {
            obj[0].caption = '';
        }
        return obj;
    }

    let datas = [];

    //只对有axLightbox=name属性的按钮操作
    let btns = [...document.querySelectorAll('[axLightbox]')].filter(i => i.getAttribute('axLightbox').includes('insName')),
        names = {};
    for (let i = 0, len = btns.length; i < len; i++) {
        let item = btns[i],

            //去掉空格，单引号转成双引号，并从insName:""中取值
            name = axAttrToJson(item, 'axLightbox').insName;
        if (!name) {
            //如果为空则跳过
            continue;
        }
        if (!names.hasOwnProperty(name)) {
            names[name] = [item];
        } else {
            names[name].push(item);
        }
        //循环完毕，最终names = {name01:[btnDom01,btnDom02],name02:[btnDom03,btnDom04]}
    }
    //给按钮添加上slideIndex
    for (let i in names) {
        let index = 0;
        names[i].forEach(k => {
            //关键：给按钮加上slideIndex属性
            k.slideIndex = index;
            let data = [],
                src = axAttrToJson(k, 'axLightbox').src;
            //标签有axLightbox='src:[]'属性
            if (src) {
                data = axCreateData(src, ['src', 'cover', 'caption', 'type'], 'figure', ['media', 'cover', 'caption', 'type'], ['', '', '', 'image'], function (data) {
                    //如果没有cover，将image类型的media赋予cover
                    data.forEach(item => {
                        if (!item.cover && item.type == 'image') {
                            item.cover = item.media
                        }
                        //如果没有封面图则使用默认图片
                        item.cover = item.cover ? item.cover : '';
                    });
                });
            } else {
                data = getSource(k);
            }
            //更新索引
            index += data.length;

            if (axIsEmpty(datas)) {
                //为空直接加入
                datas.push({ name: i, data: data });
            } else {
                datas.forEach(item => {
                    //如果相同的name已经写入了数组，则合并src数组
                    if (item.name == i) {
                        item.data.push(...data);
                    }
                })
                //如果没有相同的name，则追加数组对象
                if (!datas.some(item => (item.name == i))) {
                    datas.push({ name: i, data: data });
                }
            }
            //循环结束后，insName相同的合并了data值并且按钮带上了slideindex属性，datas = [{name:'name01',data:[]},{name:'name02',data:[]}]
        });
    }

    //创建实例
    datas.forEach(item => {
        //创建合并数据后的新实例
        new axLightbox({
            insName: item.name,
            src: item.data,
        });
    });

    //对使用axLightbox属性的元素使用本插件
    document.querySelectorAll('[axLightbox]').forEach((element) => {
        let attr = axAttrToJson(element, 'axLightbox'),
            name = attr.insName;
        if (name) {
            let ins = axInstance.find(name, 'lightbox');
            element.onclick = (ev) => {
                axPreventDefault(ev);
                if (ins) {
                    //初始化创建dom节点和swiper
                    ins.init();
                    //swiper跳转slide
                    ins.insSwiper.slideTo(element.slideIndex, 0);
                    //显示lightbox
                    ins.show();
                }
            }
        } else {
            //标签仅仅只有axLightbox属性，没有值
            if (attr.src) {
                element.onclick = (ev) => {
                    axPreventDefault(ev);
                    new axLightbox('', element).init().show();
                }
            } else {
                element.onclick = (ev) => {
                    axPreventDefault(ev);
                    new axLightbox({
                        src: getSource(element),
                    }).init().show();
                }
            }
        }


    });
})();
/*!
* 插件：网页媒体懒加载；使用方法：new axLazyload('#id',{参数})
*/
class axLazyload {
    constructor(elem, options) {
        //列表Dom
        this.targetDom = axIdToDom(elem);
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            offset: '0px', //偏离顶部或底部多长距离触发，单位px，如果只有一个值则上下偏移值相同，如果需要单独定义请用英文逗号隔开，比如'-50px,-30px'表示上方偏移50px，下方偏移30px
            visible: [0, 0.5, 1], //可见区域占比，单位是小数，比如0.5表示目标50%的部分出现在屏幕中则触发，默认在刚显示、显示一半以及完全显示时触发
            parent: null,//以什么父节点为参照，可填#id、classname、nodename或节点，默认null表示以整个视窗为参照
            animation: 'ax-fadeIn',//懒加载动画类名，默认是ax-fadeIn渐显，可选择ax-scaleIn或ax-fadeInUp
            dataName: 'src',//懒加载元素的data-*属性，默认是src，也就是懒加载元素将从data-src属性中获取真实地址
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onBefore: '',//加载前执行函数，支持一个参数即被监听元素数组，如果该函数直接返回true则按既定程序执行加载，否则需手动添加监听（add方法）
            onShow: '',//开始进入视窗监听函数，支持两个参数，即监听节点、节点的关联按钮（数组）和监听交互对象
            onShowing: '',//正在进入视窗监听函数，支持两个参数，即监听节点、节点的关联按钮（数组）和监听交互对象
            onShown: '',//完全显示在视窗监听函数，支持两个参数，即监听节点、节点的关联按钮（数组）和监听交互对象
            onAllShown: '',//监听回调函数，全部节点至少在视窗出现一次后执行，支持一个参数即监听的节点数组
            onIn: '',//进入视窗监听函数（包括show、showing和shown），支持两个参数，即监听节点和监听交互对象
            onOut: '',//离开视窗监听函数，支持两个参数，即监听节点和监听交互对象
            onAdd: '',//新增监听节点监听函数，支持一个参数，即监听新增监听节点数组
            onRemove: '',//取消监听节点监听函数，支持一个参数，即监听取消监听节点数组
            onAsync: '',//当手动懒加载内容时监听，支持一个参数，是为一个对象，对象属性有response（请求到的内容）、dom（懒加载节点）和data（传递的参数对象）
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //懒加载的项
        this.items = [];
        //显示的项数量
        this.shownNum = 0;
        //懒加载类型
        if (this.targetDom.hasAttribute(`data-${this.options.dataName}`)) {
            this.type = 'single';
        } else {
            this.type = 'multiple';
        }
        //定义父层，null即为window
        this.parent = this.options.parent ? axIdToDom(this.options.parent) : this.options.parent;
        //初始化
        this.init();
        return this;
    }
    init() {
        axInstance.push(this, this.options.insName, 'lazyload');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //生成交互监听参数对象
        this.interactOpt();
        //激活监听
        this.active();
        return this;
    }
    active() {
        //根据不同类型获取items
        this.createInteract();
        if (this.type === 'single') {
            //单图模式
            this.items.push(this.targetDom);
        } else if (this.type === 'multiple') {
            //父节点下多图模式
            this.items = [...this.targetDom.querySelectorAll(`[data-${this.options.dataName}]`)];
        }
        //flag返回true才执行
        let flag = this.options.onBefore ? this.options.onBefore.call(this, this.items) : true;
        if (flag) {
            //将items应用到交互实例
            this.items.forEach(i => {
                i.lazyload = { observed: true };
                this.interact.observe(i);
            });
        }
        return this;
    }
    interactOpt() {
        //更新父层偏离值
        this.offset = this.options.offset.replace(/\s*/g, "").split(",");
        if (this.offset.length == 1) {
            //如果只有一个值，那么追加第二个值等于第一个值
            this.offset.push(...this.offset)
        } else if (offset.length > 2) {
            //如果多写了，则只取前两个
            this.offset.splice(0, 2);
        }
        return {
            //父层即为视窗
            root: this.parent,
            //距离窗口上沿和下沿位置执行
            rootMargin: this.offset[0] + ' 0px ' + this.offset[1] + ' 0px',
            //目标节点的高度或宽度出现在父层的比例
            threshold: this.options.visible,
        }
    }
    createInteract() {
        this.interact = new IntersectionObserver(entries => {
            entries.forEach(i => {
                let args = [i.target, i], realData = i.target.dataset[this.options.dataName];

                if (i.isIntersecting) {
                    //data-src
                    if (realData) {
                        //动画可能为空
                        this.options.animation ? i.target.classList.add(this.options.animation) : null;
                        if (i.target.nodeName === 'VIDEO' || i.target.nodeName === 'AUDIO') {
                            //对于source元素
                            i.target.src = realData;
                            i.target.querySelector('source') ? i.target.querySelector('source').setAttribute('src', realData) : null;

                        } else if (i.target.nodeName === 'IMG' || i.target.nodeName === 'IFRAME') {
                            //一般元素，比如img或iframe
                            i.target.setAttribute('src', realData);
                        } else {
                            //异步
                            let params = axAttrToJson(i.target, 'data-async');
                            axAjax({
                                url: realData,
                                data: params,
                                before: () => {
                                    i.target.setAttribute('loading', '');
                                },
                                success: (response) => {
                                    i.target.removeAttribute('loading');
                                    i.target.innerHTML = '';
                                    if (i.target.dataset['async'] || typeof response.content === 'object') {
                                        //有data-async属性，或者获得到是一个数组（对象），说明需要手动添加内容
                                        //监听回调
                                        this.handlers.hasOwnProperty('async') ? this.emit('async', { response: response.content, data: params, dom: i.target }) : null;
                                        this.options.onAsync && this.options.onAsync.call(this, { response: response.content, data: params, dom: i.target });
                                    } else {
                                        //普通ajax获取页面内容
                                        i.target.innerHTML = response.content;
                                    }
                                }
                            });
                        }
                        //i.target.removeAttribute('data-' + this.options.dataName);
                        delete i.target.dataset[this.options.dataName];

                        //监听回调
                        if (i.intersectionRatio === 0) {
                            //显示状态，为1表示完全显示了
                            i.target.lazyload.state = 'show';
                            this.handlers.hasOwnProperty('show') ? this.emit('show', ...args) : null;
                            this.options.onShow && this.options.onShow.call(this, ...args);
                        } else if (i.intersectionRatio === 1) {
                            //显示状态，为1表示完全显示了
                            i.target.lazyload.state = 'shown';
                            this.handlers.hasOwnProperty('shown') ? this.emit('shown', ...args) : null;
                            this.options.onShown && this.options.onShown.call(this, ...args);
                            this.shownNum++;
                            //完成了便取消监听
                            this.interact.unobserve(i.target);
                        } else {
                            i.target.lazyload.state = 'showing';
                            this.handlers.hasOwnProperty('showing') ? this.emit('showing', ...args) : null;
                            this.options.onShowing && this.options.onShowing.call(this, ...args);
                        }
                        //监听回调
                        this.handlers.hasOwnProperty('in') ? this.emit('in', ...args) : null;
                        this.options.onIn && this.options.onIn.call(this, ...args);
                        //全部显示
                        if (this.shownNum >= this.items.length) {
                            //监听回调
                            this.handlers.hasOwnProperty('allShown') ? this.emit('allShown', this.items) : null;
                            this.options.onAllShown && this.options.onAllShown.call(this, this.items);
                        }

                    }
                } else {
                    //监听回调
                    this.handlers.hasOwnProperty('out') ? this.emit('out', ...args) : null;
                    this.options.onOut && this.options.onOut.call(this, ...args);
                }
            });
        }, this.interactOpt());
    }
    getDoms(data) {
        let dataType = axType(data),
            doms = [];
        //获得节点数组
        if (dataType === 'Array') {
            doms = data.map(k => axIdToDom(k)).filter(Boolean);
        } else if (dataType === 'String') {
            doms = axAllToDoms(data).filter(Boolean);
        }
        return doms;
    }
    add(data, callback) {
        let doms = this.getDoms(data);
        //追加监听
        doms.forEach(i => {
            if (!i.lazyload) {
                i.lazyload = { observed: true };
                this.interact.observe(i);
                //加入items
                !this.items.includes(i) ? this.items.push(i) : null;
            } else {
                new axMessage({
                    content: `媒体文件${i.dataset.src || i.src}已经处于监听状态，不需要添加监听操作！`,
                    iconShow: true,
                    result: 'warning',
                }).show();
            }

        });

        //监听回调
        this.handlers.hasOwnProperty('add') ? this.emit('add', doms) : null;
        this.options.onAdd && this.options.onAdd.call(this, doms);
        callback && callback.call(this, doms);
    }
    remove(data, callback) {
        let doms = this.getDoms(data);
        //追加监听
        doms.forEach(i => {
            if (this.items.includes(i)) {
                if (i.lazyload) {
                    Reflect.deleteProperty(i, 'lazyload');
                    this.interact.unobserve(i);
                } else {
                    new axMessage({
                        content: `媒体文件${i.dataset.src || i.src}还未被监听，不需要取消监听操作！`,
                        iconShow: true,
                        result: 'warning',
                    }).show();
                }
            }
        });
        //监听回调
        this.handlers.hasOwnProperty('remove') ? this.emit('remove', doms) : null;
        this.options.onRemove && this.options.onRemove.call(this, doms);
        callback && callback.call(this, doms);
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
* 插件：列表异步无限滚动；使用方法：new axInfinite('#id',{参数})
*/
class axInfinite {
    constructor(elem, options) {
        //列表Dom
        this.targetDom = axIdToDom(elem);
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            type: 'list', //无限滚动的内容类型，默认是当页list列表、异步pages、异步json文件、异步sql数据
            offset: '0px', //偏离顶部或底部多长距离触发，单位px，如果只有一个值则上下偏移值相同，如果需要单独定义请用英文逗号隔开，比如'-50px,-30px'表示上方偏移50px，下方偏移30px
            visible: [0], //可见区域占比，单位是小数，比如0.1表示目标10%的部分出现在屏幕中则触发
            url: '', //加载的页面地址，可以是一个地址、地址数组、pagination或className（#id）
            urlMax: 20,//url作为数组length最大值，默认20
            data: [], //用于追加静态列表；该参数对type=list有效
            template: null, //数组转Dom的模板
            selector: '',//使用type=pages的时候需要确定从其他页面的哪个节点获取内容，默认为空则获取整个网页
            trigger: 'scroll',//无限滚动触发方式，默认是scroll，可选择click，即加载不会触发，点击某个按钮之后才会触发；可选择clicks，即没点击一次加载一个页面，直到全部加载完毕
            parent: null,//以什么父节点为参照，默认null表示以整个视窗为参照
            animation: 'ax-fadeIn',//加载动画className，默认是ax-fadeIn渐显，可选择ax-scaleIn或ax-fadeInUp
            start: 0,//从数据表获取数据时从第几条开始，默认是0，也就是从第一条开始；或者表示从url的数组第几个开始加载，0代表第一个
            count: 10, //每次加载的信息数量，默认为10；该参数对type=sql有效
            finishText: '没有更多内容了',//全部页面加载完毕之后resultDiv显示的文字内容
            nextBtn: '<button class="ax-btn ax-btn-primary ax-longer">查看更多</button>',//下一页按钮的HTML
            ajaxType: 'post',//异步提交的方式，默认post，可填get
            active: '',//是一个函数，激活无限滚动后的回调函数
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};

        //从数据表获取数据时从第几条开始
        this.start = this.options.start;

        //json数据
        this.data = this.options.data;

        //加载器
        this.loadStatus = this.targetDom.querySelector('.ax-infinite-status') ? this.targetDom.querySelector('.ax-infinite-status') : axAddElem('div', { class: 'ax-infinite-status' });
        this.loadingDiv = axAddElem('div', { class: 'ax-infinite-loading' }, '<span class="ax-loading"><i></i></span>');
        this.resultDiv = axAddElem('div', { class: 'ax-infinite-tips' });
        this.nextDiv = axAddElem('div', { class: 'ax-infinite-next' }, this.options.nextBtn);
        this.nextButton = this.nextDiv.firstElementChild;
        //loading始终在最前方
        this.loadStatus.insertAdjacentElement('afterbegin', this.loadingDiv);
        //结果提示始终在最后方
        this.loadStatus.appendChild(this.resultDiv);

        //url数组
        this.url = [];
        if (typeof this.options.url === 'string') {
            if (this.options.url.indexOf('{') > 0 && this.options.url.indexOf('}') > 0) {
                //如果自动url，如：ajax/infinite{0}.html
                let pageStart = ~~this.options.url.match(/\{(\S*)\}/)[1],
                    str = `{${pageStart}}`;
                for (let i = pageStart; i <= pageStart + this.options.urlMax; i++) {
                    this.url.push(this.options.url.replace(str, i));
                }

            } else if (this.options.url === 'pagination') {
                //如果是从隐藏的分页获取url
                let pages = this.loadStatus.querySelectorAll('[data-page]');
                [...pages].forEach(i => {
                    this.url.push(i.href);
                });
            } else {
                //如果是普通url
                this.options.url ? this.url = [this.options.url] : null;
            }

        } else if (axType(this.options.url) == 'Array') {
            this.url = this.options.url;
        }

        //从参数取模板或从dom取模板
        if (!this.options.template) {
            this.template = '';
        } else {
            if (this.options.template.trim().substr(0, 1) == '#') {
                this.template = document.querySelector(this.options.template).innerHTML
            } else {
                this.template = this.options.template;
            }
        }

        //无限滚动的项
        this.items = [];

        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'infinite');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //追加className（单图模式不添加）
        if (this.options.type != 'image' && !this.targetDom.classList.contains('ax-infinite')) {
            this.targetDom.classList.add('ax-infinite');
        }

        //生成交互监听参数对象
        this.interactOpt();

        if (this.options.trigger == 'scroll') {
            //如果只是借助滚动条触发
            this.active();
        } else if (this.options.trigger == 'clicks') {
            //如果是阶梯式一页一页点击展开
            //准备好加载器
            this.loadStatus.appendChild(this.nextDiv);
            this.loadStatus.setAttribute('preloading', '');
            this.targetDom.appendChild(this.loadStatus);
            this.nextButton.onclick = () => {
                //改变加载器状态
                this.loadStatus.removeAttribute('preloading', '');
                this.loadStatus.removeAttribute('loaded', '');
                this.loadStatus.setAttribute('loading', '');
                //支持三种异步模式，不支持当页同步列表
                if (this.options.type == 'pages') {
                    //多页面模式
                    this.getAsyncData('page', { start: this.start });
                    //this.pageAjax(this.start);
                } else if (this.options.url && this.options.template) {
                    //异步获取页面模式
                    if (this.options.type == 'json') {
                        this.getAsyncData('json', { start: this.start });
                        //this.jsonAjax(this.start);
                    } else if (this.options.type == 'sql') {
                        this.getAsyncData('sql', { start: this.start, count: this.options.count });
                        //this.sqlAjax(this.start, this.options.count);
                    }
                }
            }
        }

    }
    active(url, callback) {
        //如果使用了url参数，则更新this.url数组
        if (!axIsEmpty(url)) {
            if (axType(url) == 'Array') {
                //url可以是数组
                this.url.push(...url)
            } else {
                //url可以是字符串，比如一个网址
                this.url.push(url);
            }
        }
        //根据不同类型获取items
        if (this.options.type == 'pages') {
            //多页面模式
            this.asyncInteract('page');
        } else if (!axIsEmpty(this.url) && this.options.template) {
            //异步获取页面模式
            if (this.options.type == 'json') {
                this.asyncInteract('json');
            } else if (this.options.type == 'sql') {
                this.asyncInteract('sql');
            }
        } else if (this.options.type == 'list') {
            //静态列表的加载
            //如果有json数据则追加
            if (!axIsEmpty(this.data)) {
                this.renderList(this.data);
            }
            let children = [...this.targetDom.childNodes].filter(i => i.nodeType == 1);
            children.forEach(i => {
                if (axOffset(i).top > window.innerHeight) {
                    i.setAttribute('infinite', '');
                    this.items.push(i);
                }
            });
            this.listInteract();
            //将items应用到交互实例
            this.items.forEach(i => {
                this.listInteract().observe(i);
            });
        }
        //回调
        callback && callback.call(this);
        this.options.active && this.options.active.call(this);
        'activated' in this.handlers ? this.emit('activated', '') : null;
    }
    interactOpt() {
        //更新父层偏离值
        this.offset = this.options.offset.replace(/\s*/g, "").split(",");
        if (this.offset.length == 1) {
            //如果只有一个值，那么追加第二个值等于第一个值
            this.offset.push(...this.offset)
        } else if (offset.length > 2) {
            //如果多写了，则只取前两个
            this.offset.splice(0, 2);
        }
        return {
            //父层即为视窗
            root: this.options.parent,
            //距离窗口上沿和下沿位置执行
            rootMargin: this.offset[0] + ' 0px ' + this.offset[1] + ' 0px',
            //目标节点的高度或宽度出现在父层的比例
            threshold: this.options.visible,
        }
    }
    listInteract() {
        let interact = new IntersectionObserver(entries => {
            entries.forEach(i => {
                if (i.isIntersecting) {
                    //进入视窗追加动画显示
                    if (i.target.hasAttribute('infinite')) {
                        i.target.classList.add(this.options.animation);
                        i.target.removeAttribute('infinite');
                        //完成了便取消监听
                        interact.unobserve(i.target);
                    }
                    //监听
                    'shown' in this.handlers ? this.emit('shown', i.target) : null;
                }
            });
        }, this.interactOpt());
        return interact;
    }
    asyncInteract(type) {
        this.targetDom.appendChild(this.loadStatus);
        this.asyncObserver = new IntersectionObserver(entries => {
            if (entries[0].isIntersecting) {
                if (type == 'page' || type === 'json') {
                    this.getAsyncData(type, { start: this.start });
                } else if (type === 'sql') {
                    this.getAsyncData('sql', { start: this.start, count: this.options.count });
                }
            }
        }, this.interactOpt());
        this.asyncObserver.observe(this.loadStatus);
    }
    itemInteract() {
        let interact = new IntersectionObserver(entries => {
            entries.forEach(i => {
                if (i.isIntersecting) {
                    if (i.target.hasAttribute('infinite')) {
                        i.target.classList.add(this.options.animation);
                        i.target.removeAttribute('infinite');
                        //完成了便取消监听
                        interact.unobserve(i.target);
                        //监听
                        'shown' in this.handlers ? this.emit('shown', i.target) : null;
                    }
                }
            });
        }, this.interactOpt());
        return interact;
    }
    renderList(data) {
        let fragment = document.createDocumentFragment();
        data.forEach(i => {
            let item = axStrToDom(axTplEngine(this.template, i));
            item.setAttribute('infinite', '');
            //更新追加的节点数组
            this.items.push(item);
            fragment.appendChild(item);
        });

        this.targetDom.appendChild(fragment);
        //异步才需要加载器
        !this.options.type == 'list' ? this.targetDom.appendChild(this.loadStatus) : null;
    }
    loadFinish() {
        //如果是点一下展示一页的方式，那么便不需要监听滚动了
        this.options.trigger != 'clicks' ? this.asyncObserver.disconnect() : null;
        this.loadStatus.removeAttribute('loading');
        this.loadStatus.removeAttribute('loaded');
        this.loadStatus.setAttribute('finish', '');
        this.loadingDiv ? this.loadingDiv.remove() : null;
        this.nextDiv ? this.nextDiv.remove() : null;
        this.resultDiv.innerHTML = this.options.finishText;
        //监听
        'finished' in this.handlers ? this.emit('finished', '') : null;
    }
    //从数据库中获得数据
    getAsyncData(type = 'json', options, callback) {
        return axAjax({
            //根据参数进行数据扩展
            data: type === 'sql' ? { start: options.start, count: options.count } : '',
            type: this.options.ajaxType,
            url: type === 'sql' ? this.options.url : this.url[options.start],
            selector: this.options.selector,
            error: (response) => {
                //页面不存在会报404错误
                if(response.status == '404'){
                    new axMessage({
                        content:'已经没有更多内容了！',
                        result:'warning',
                        iconShow:true,
                    }).show();
                }else{
                    console.error(`发生错误：${response.status}，可能是地址错误！`);
                }
                //发生错误也表示加载完毕
                this.loadFinish();
            },
            opened: (response) => {
                this.contentXhr = response.xhr;
            },
            success: (response) => {
                if (type === 'page') {
                    this.targetDom.insertAdjacentHTML('beforeEnd', response.content);
                    this.targetDom.appendChild(this.loadStatus);
                    //更新数据起始位置
                    this.start++;
                    //如果页数全部加载完毕了，则取消监听并删除加载按钮
                    if (this.start >= this.url.length) {
                        this.loadFinish();
                    }
                } else {
                    //创建屏幕滚动交互实例
                    this.itemInteract();
                    if (type === 'json') {
                        //渲染
                        this.renderList(response.content);
                        this.targetDom.appendChild(this.loadStatus);
                        //更新数据起始位置
                        this.start++;
                        //如果页数全部加载完毕了，则取消监听并删除加载按钮
                        if (this.start >= this.url.length) {
                            this.loadFinish();
                        }
                    } else if (type === 'sql') {
                        //content == {data:[],total:100}
                        //渲染
                        this.renderList(response.content.data);
                        this.targetDom.appendChild(this.loadStatus);
                        //更新数据起始位置
                        this.start += options.count;
                        //如果页数全部加载完毕了，则取消监听并删除加载按钮
                        if (this.start >= response.content.total) {
                            this.loadFinish();
                        }
                    }
                }
                //因为点一下展示一页的方式需要对loading和下一页按钮来回交替显示，需要通过preloading、loading、loaded和finish几个状态进行管理
                if (this.options.trigger == 'clicks') {
                    this.loadStatus.removeAttribute('loading', '');
                    this.loadStatus.setAttribute('loaded', '');
                }
                this.items.forEach(i => {
                    this.itemInteract().observe(i);
                });
                //监听
                this.handlers.hasOwnProperty('loaded') ? this.emit('loaded', response.content) : null;
                callback && callback(response.content);
            }
        });
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/* (() => {
    //父节点下多图
    document.querySelectorAll('[axInfinite]').forEach(element => {
        new axInfinite(element);
    });
})(); */
/*!
* 插件：进度条；使用方法：new axProgress('#id',{参数})
*/
class axProgress {
    constructor(elem, options) {
        if (!elem) { return false; }
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            type: 'line', //进度条类型，可选线性line，大圆circle、半圆semicircle和仪表圆gapcircle
            theme: '',//进度条颜色风格，默认蓝色+浅灰，可选择primary、success、error、info、question或warning
            gradient: false,//是否使用渐变色，默认false不使用，可选择true使用渐变色
            linecap: 'round',//线条头尾是否是圆角，默认round，可选square
            size: '',//进度条的整体大小（宽高），默认sm，可选择md和lg
            range: [0, 100],//取值范围，value不应超出该范围
            value: 0,//初始化后进度值，值在range范围
            curve: 'easeOut',//动画曲线，默认easeOut即先快后慢，另外可选linear(线性平均)、easeIn(先慢后快)、easeInOut(先较为慢后较为快)、easeOutIn(先较为快后较为慢)
            duration: 0,//动画执行时间，数值类型，默认为0即插件根据移动距离自动生成执行时间，单位毫秒
            rotate: false,//是否旋转，建议只对type=circle使用，默认false不旋转，可使用true旋转
            trackShow: true,//是否显示轨道，默认true显示，可选择false不显示
            labelFormat: '',//定义进度结果文本HTML格式，默认`<span label><i progress></i><i unit></i></span>`
            labelShow: true,//是否显示结果文本，默认为true显示，可选择false不显示
            lablePlace: '',//结果文本所在位置，默认为空表示按插件定义的位置存放，可填#id或className自定义位置
            labelText: {
                unit: '%',
                tips: '当前进度',
                complete: '完成!',
            },//定义进度结果文本内容
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，支持两个参数：value和barValue
            onStart: '',//回调函数，归零后执行，无参数，
            onCompleted: '',//回调函数，进度百分百后执行，无参数
            onDoing: '',//回调函数，进行中执行，支持两个参数：value和barValue
            onSet: '',//回调函数，设置值之后执行，支持两个参数，value和barValue
            onPause: '',//回调函数，停止之后执行，支持两个参数，valueNow和barValueNow
            onReset: '',//回调函数，直接重置归零后执行，无参数
            onDone: '',//回调函数，直接跳转百分百后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //对range排序,从小到大
        this.options.range = this.options.range.sort((a, b) => a - b);
        //定义svg的path长度
        this.barSize = this.options.type === 'circle' ? 251.2 : this.options.type === 'semicircle' ? 125.6 : this.options.type === 'gapcircle' ? 167.6 : this.options.range[1];
        //获得初始值，this.value为range的中间值
        this.value = this.valueNow = this.fixedValue(this.options.value);
        //根据this.value获得this.barValue
        this.barValue = this.barValueNow = this.getInitBarValue();
        //定义“完成”显示节点
        this.completeDom = axAddElem('span', { complete: '' }, `${this.options.labelText.complete}`);
        //定义stop状态，使用pause方法才会转为true，true才能continue
        this.stop = false;
        //初始化
        this.init();
        return this;
    }
    init(callback) {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'progress');
        //设置销毁状态
        this.destroyed = false;
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置销毁状态
        //创建节点并根据this.barValue值初始化节点
        this.createHtmltoInit();
        //监听完成
        this.stopWatch();
        //给节点设置样式
        this.setAttribute();
        //监听开始
        this.options.onInit && this.options.onInit.call(this, this.value, this.barValue);
        this.handlers.hasOwnProperty('init') ? this.emit('init', this.value, this.barValue) : null;
        callback && callback.call(this, this.value, this.barValue);
        return this;
    }
    setAnimation(opt) {
        //已经暂停则不能再执行，除非使用continue方法
        if (this.stop) {
            return false;
        }

        let options = Object.assign({
            duration: this.options.duration,
            valTarget: this.value,
            barTarget: this.barValue,
            valLast: this.valueNow,
            barLast: this.barValueNow,
        }, opt)
        //值相同就不需要再执行了
        if (this.value === this.valueNow) {
            return false;
        }
        //先清除之前的动画请求，避免叠加动作请求
        this.aniRequest ? cancelAnimationFrame(this.aniRequest) : null;

        //从当前值进行到目标值target
        let dftTime = size => size * 3 + 200,
            valDiffer = options.valTarget - options.valLast,
            barDiffer = options.barTarget - options.barLast,
            time = options.duration || options.duration===0 ? options.duration : dftTime(Math.abs(barDiffer)),
            initTime = new Date().getTime(),
            aniRepeat = () => {
                let newTime = new Date().getTime() - initTime,
                    timestep = newTime / time,
                    timefactor = axCurves[this.options.curve](timestep);
                //获得剩余时间，暂停和继续需要用到
                this.timeRest = Math.trunc(time - newTime);
                //实时更新当前value和bar值
                this.valueNow = axNumber(options.valLast + valDiffer * timefactor, { mode: 'floor' });
                this.barValueNow = axNumber(options.barLast + barDiffer * timefactor, { mode: 'floor' });
                //获得当前进度值整数文本
                this.valueNowText = Math.trunc(this.valueNow);

                //监听进行中
                this.options.onDoing && this.options.onDoing.call(this, this.valueNow, this.barValueNow);
                this.handlers.hasOwnProperty('doing') ? this.emit('doing', this.valueNow, this.barValueNow) : null;
                //因为进度可正向可反向，所以需要用时间戳来判断结束与否
                if (newTime < time) {
                    if (this.options.type == 'line') {
                        this.barDom.style.width = this.barValueNow + '%';
                    } else {
                        this.barDom.setAttribute('stroke-dashoffset', this.barValueNow);
                    }
                    //如果labelShow=false则不存在该节点了
                    this.progressDom ? this.progressDom.innerText = Math.trunc(this.valueNowText) : null;
                    //重复请求动画
                    this.aniRequest = requestAnimationFrame(aniRepeat);
                } else {
                    //取消循环动画
                    cancelAnimationFrame(this.aniRequest);
                    //更新值
                    this.valueNow = this.value;
                    this.barValueNow = this.barValue;
                    this.timeRest = 0;
                    //bar直接填值
                    if (this.options.type == 'line') {
                        this.barDom.style.width = this.barValue + '%';
                    } else {
                        this.barDom.setAttribute('stroke-dashoffset', this.barValue);
                    }
                    //进度文本直接填值
                    this.progressDom ? this.progressDom.innerText = this.value : null;
                }
                //监听完成
                this.stopWatch();
            }
        //执行动画
        aniRepeat()

    }
    stopWatch() {
        if (this.valueNow >= this.options.range[1]) {
            //结束
            this.completed = true;
            //追加完成节点，并设置完成属性
            this.targetDom.setAttribute('completed', 'true');
            this.labelDom ? this.labelDom.appendChild(this.completeDom) : null;
            //回调监听
            this.options.onCompleted && this.options.onCompleted.call(this);
            this.handlers.hasOwnProperty('completed') ? this.emit('completed', '') : null;
        } else {
            //其他状态
            this.completed = false;
            //移除节点和属性
            this.completeDom.remove();
            this.targetDom.removeAttribute('completed');
            //开始
            if (this.valueNow <= this.options.range[0]) {
                this.options.onStart && this.options.onStart.call(this);
                this.handlers.hasOwnProperty('start') ? this.emit('start', '') : null;
            }
        }
    }
    fixedValue(value) {
        let result;
        if (value < this.options.range[0]) {
            result = this.options.range[0];
        } else if (value > this.options.range[1]) {
            result = this.options.range[1];
        } else {
            result = value;
        }
        return result;
    }
    getInitBarValue() {
        let initBarValue;
        if (this.options.type == 'line') {
            //获得条状进度条bar的百分值
            initBarValue = (this.value / (this.options.range[1] - this.options.range[0])) * this.options.range[1];
        } else {
            //获得全圆、半圆、角圆的bar的百分值
            initBarValue = (this.barSize - ((this.value / this.options.range[1]) * this.barSize));
        }
        //trueValue为bar的真实值
        return axNumber(initBarValue, { mode: 'floor' });
    }
    getBarValue(value) {
        let target;
        //获取新旧差值differ和目标值barValue
        if (this.options.type == 'line') {
            target = (value / (this.options.range[1] - this.options.range[0])) * this.options.range[1];
            //修正
            target = target > this.options.range[1] ? this.options.range[1] : axNumber(target, { mode: 'floor' });
        } else {
            target = (this.barSize - ((value / this.options.range[1]) * this.barSize));
            //修正
            target = target > this.barSize ? this.barSize : axNumber(target, { mode: 'floor' });
        }
        return target;
    }
    setAttribute() {
        //设置样式类
        !this.targetDom.classList.contains('ax-progress') ? this.targetDom.classList.add('ax-progress') : null;
        this.targetDom.setAttribute('linecap', this.options.linecap);
        this.targetDom.setAttribute('type', this.options.type);
        this.options.size ? this.targetDom.setAttribute('size', this.options.size) : null;
        this.options.theme ? this.targetDom.setAttribute('theme', this.options.theme) : null;
        this.options.gradient ? this.targetDom.setAttribute('gradient', '') : null;
        !this.options.trackShow ? this.targetDom.setAttribute('track', 'false') : null;
        //circle无限旋转
        if (this.options.rotate && this.options.type === 'circle') {
            this.wrapperDom.classList.add('ax-rotate360');
            this.wrapperDom.style.animationIterationCount = 'infinite';
        }
    }
    createHtmltoInit() {
        //放置重复初始化导致重新生成子节点，所以清空内容
        this.targetDom.innerHTML = '';
        //创建进度条核心HTML
        if (this.options.type == 'line') {
            this.wrapperDom = this.trackDom = axAddElem('div', { 'track': '' });
            this.barDom = axAddElem('div', { 'bar': '', 'style': `width:${this.barValue}%` });
            this.labelFormat = this.options.labelFormat ? this.options.labelFormat : `<span label><i progress></i><i unit></i></span>`;
        } else {
            this.labelFormat = this.options.labelFormat ? this.options.labelFormat : `<div label><div tips></div><div result><i progress></i><i unit></i></div></div>`;
            //加入渐变
            let gradient = '', svg;
            if (this.options.gradient) {
                if (this.options.theme == 'primary') {
                    gradient = `
                <linearGradient id="svgbar-primary" >
                    <stop offset="0%" style="stop-color:var(--color-primary);" ></stop>
                    <stop offset="100%" style="stop-color:var(--color-primary-aj);" ></stop>
                </linearGradient>
              `;
                } else if (this.options.theme == 'error') {
                    gradient = `
                <linearGradient id="svgbar-error" >
                    <stop offset="0%" style="stop-color:var(--color-error);" ></stop>
                    <stop offset="100%" style="stop-color:var(--color-error-aj);" ></stop>
                </linearGradient>
              `;
                } else if (this.options.theme == 'success') {
                    gradient = `
                <linearGradient id="svgbar-success" >
                    <stop offset="0%" style="stop-color:var(--color-success);" ></stop>
                    <stop offset="100%" style="stop-color:var(--color-success-aj);" ></stop>
                </linearGradient>
              `;
                } else if (this.options.theme == 'warning') {
                    gradient = `
            <linearGradient id="svgbar-warning" >
                <stop offset="0%" style="stop-color:var(--color-warning);" ></stop>
                <stop offset="100%" style="stop-color:var(--color-warning-aj);" ></stop>
            </linearGradient>
          `;
                } else if (this.options.theme == 'info') {
                    gradient = `
                <linearGradient id="svgbar-info" >
                    <stop offset="0%" style="stop-color:var(--color-info);" ></stop>
                    <stop offset="100%" style="stop-color:var(--color-info-aj);" ></stop>
                </linearGradient>
              `;
                } else if (this.options.theme == 'question') {
                    gradient = `
                <linearGradient id="svgbar-question" >
                    <stop offset="0%" style="stop-color:var(--color-question);" ></stop>
                    <stop offset="100%" style="stop-color:var(--color-question-aj);"></stop>
                </linearGradient>
              `;
                }
            }
            if (this.options.type == 'circle') {
                svg = `<svg viewBox="0 0 100 100">
                            ${gradient}
                            ${this.options.trackShow ? '<path d="M 50 50 m -40 0 a 40 40 0 1 1 80 0  a 40 40 0 1 1 -80 0" track></path>' : ''}
                            <path d="M 50 50 m -40 0 a 40 40 0 1 1 80 0  a 40 40 0 1 1 -80 0" bar stroke-dasharray=${this.barSize} stroke-linecap=${this.options.linecap}  stroke-dashoffset=${this.barValue}></path>
                       </svg>
                        `;
            } else if (this.options.type == 'semicircle') {
                svg = `<svg viewBox="0 0 100 60">
                            ${gradient}
                            ${this.options.trackShow ? `<path d="M 50 50 m -40 0 a 40 40 0 1 1 80 0" track stroke-linecap=${this.options.linecap}></path>` : ''}
                            <path d="M 50 50 m -40 0 a 40 40 0 1 1 80 0" bar stroke-dasharray=${this.barSize} stroke-linecap=${this.options.linecap} stroke-dashoffset=${this.barValue}></path>
                       </svg>
                        `;
            }
            else if (this.options.type == 'gapcircle') {
                svg = `<svg viewBox="0 0 100 80">
                            ${gradient}
                            ${this.options.trackShow ? `<path d="M15.367,70.026C11.954,64.137,10,57.296,10,49.999 C10,27.909,27.909,10,50,10c22.092,0,40,17.909,40,39.999c0,7.295-1.952,14.134-5.363,20.022" track stroke-linecap=${this.options.linecap}></path>` : ''}
                            <path d="M15.367,70.026C11.954,64.137,10,57.296,10,49.999 C10,27.909,27.909,10,50,10c22.092,0,40,17.909,40,39.999c0,7.295-1.952,14.134-5.363,20.022" bar stroke-dasharray=${this.barSize} stroke-linecap=${this.options.linecap} stroke-dashoffset=${this.barValue}></path>
                       </svg>
                        `;
            }
            this.wrapperDom = axStrToDom(svg);
        }

        //往目标加入结果文字
        if (this.options.labelShow) {
            if (!this.options.labelPlace) {
                //labelPlace为空则内置
                this.targetDom.insertAdjacentHTML('beforeEnd', this.labelFormat);
                this.labelDom = this.targetDom.querySelector('[label]');
            } else {
                //自定义位置
                let place = axIdToDom(this.options.labelPlace);
                if (place) {
                    place.insertAdjacentHTML('beforeEnd', this.labelFormat);
                    this.labelDom = place.querySelector('[label]');
                }
            }
            this.progressDom = this.labelDom.querySelector('[progress]');
            this.unitDom = this.labelDom.querySelector('[unit]');
            this.tipsDom = this.labelDom.querySelector('[tips]');
            //赋值
            this.progressDom ? this.progressDom.innerHTML = this.value : null;
            this.unitDom ? this.unitDom.innerHTML = this.options.labelText.unit : null;
            this.tipsDom ? this.tipsDom.innerHTML = this.options.labelText.tips : null;
        }

        if (this.options.type == 'line') {
            //加入进度条
            this.wrapperDom.appendChild(this.barDom);
            //完全加入
            this.targetDom.insertAdjacentElement('afterBegin', this.wrapperDom);
        } else {
            //完全加入
            this.targetDom.insertAdjacentElement('afterBegin', this.wrapperDom);
            this.barDom = this.wrapperDom.querySelector('[bar]');
            this.trackDom = this.wrapperDom.querySelector('[track]');
        }
    }
    beforeSet(value) {
        //获得修正好的目标值
        this.value = this.fixedValue(value);
        //根据目标值获得，bar目标值和与现值的差值
        this.barValue = this.getBarValue(this.value);
    }
    set(val, duration = this.options.duration, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //已经暂停则不能再执行，除非使用continue/reset/done方法解除
        if (this.stop) {
            return false;
        }
        //获得目标值和差值
        this.beforeSet(val);
        //值没有变化则不再执行
        if (this.value === this.valueNow) {
            return false;
        }
        this.setAnimation({ duration });
        //监听回调
        this.options.onSet && this.options.onSet.call(this, this.value, this.barValue);
        this.handlers.hasOwnProperty('set') ? this.emit('set', this.value, this.barValue) : null;
        callback && callback.call(this);
        return this;
    }
    reset(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.stop = false;
        this.value = this.valueNow = 0;
        this.progressDom ? this.progressDom.innerHTML = 0 : null;
        this.barValue = this.barValueNow = this.barSize;
        this.barDom.setAttribute('stroke-dashoffset', this.barValue);
        //stopWatch默认有start监听
        this.stopWatch();
        //回调监听
        this.options.onReset && this.options.onReset.call(this);
        this.handlers.hasOwnProperty('reset') ? this.emit('reset', '') : null;
        callback && callback.call(this);
        return this;
    }
    done(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.stop = false;
        this.value = this.valueNow = this.options.range[1];
        this.progressDom ? this.progressDom.innerHTML = 100 : null;
        this.barValue = this.barValueNow = 0;
        this.barDom.setAttribute('stroke-dashoffset', this.barValue);
        //stopWatch默认有complete监听
        this.stopWatch();
        //回调监听
        this.options.onDone && this.options.onDone.call(this);
        this.handlers.hasOwnProperty('done') ? this.emit('done', '') : null;
        callback && callback.call(this);
        return this;
    }
    continue(duration) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //只有使用了pause方法之后才可使用本方法
        if (!this.stop) {
            return false;
        }
        this.stop = false;
        //获得动画时间
        duration = duration || this.timeRest;
        //根据当前进度值，重新计算trueValue
        this.setAnimation({
            duration: this.timeRest,
        });
        return this;
    }
    pause(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        if (this.aniRequest) {
            cancelAnimationFrame(this.aniRequest);
            this.stop = true;
        }
        //回调监听
        this.options.onPause && this.options.onPause.call(this, this.valueNow, this.barValueNow);
        this.handlers.hasOwnProperty('pause') ? this.emit('pause', this.valueNow, this.barValueNow) : null;
        callback && callback.call(this, this.valueNow, this.barValueNow);
        return this;
    }
    get() {
        return {
            value: this.value,
            barValue: this.barValue,
            valueNow: this.valueNow,
            barValueNow: this.barValueNow,
            completed: this.completed,
            timeRest: this.timeRest,
            valueNowText: this.valueNowText,
        };
    }
    destroy(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //归零
        this.reset();

        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
 * 插件：创建单行文本域和多行文本域；使用方法：new axInput('#id',{参数})
 * text、textarea、hidden、url、tel、search、email、color、number、range、month、week、date、time、datetime、datetime-local
 * submit、reset和button特别，通过value设定名称
 * radio和checkbox请使用axCheckbox组件
 */
class axInput {
    constructor(elem, options) {
        //定义节点DOM
        this.targetDom = elem ? axIdToDom(elem) : '';
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            type: 'text', //input类型，text、range、date、password、textarea等
            name: '', //字段名
            value: '', //初始值
            limit: 0, //最多录入多少字符，默认为0不限制
            placeholder: '', //表单域输入提示
            width: '', //设置表单域宽度，比如width:'100px'
            height: '', //设置表单域高度，比如height:'100px'
            size: '', //设置表单域的尺寸,可以填xs、sm、md或lg
            className: '', //设置表单域样式名
            props: '', //设置表单域其他属性，比如props:{role:'',display:''}
            toolsShow: true, //设置是否显示右侧工具，默认显示
            extend: '', //工具扩展，可以是一个对象也可以是一个数组，格式为[{name:'',icon:'',text:'',node:'',callback:''},]
            clearShow: false, //当toolsShow: true的时候是否显示清除按钮，默认true显示，可选择false不显示
            language:{
                limit: `已输入<i have>0</i>个字符,还可以输入<i remainder><# this.value #></i>个字`,//字数限制提示
            },
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onClear: '', //清除值后的回调函数，支持参数output，即输出详细
            onSet: '', //设置值后的回调函数，支持input节点本身，input的值和剩余可输入的文字数量
            onReset: '', //重置后回调，支持参数output，即输出详细
            onInit: '', //初始化后的回调函数，支持参数output，即输出详细
            onExceed: '', //字数到极限后的回调函数，支持参数output，即输出详细
            onInput: '', //键入文字后的回调函数，支持参数output，即输出详细
        }, options, this.targetDom, this.constructor.name);
        // 定义监听事件成员
        this.handlers = {};
        //定义输出对象
        this.output = { name: this.options.name };
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //创建定义parent和input变量
        this.createElem();
        //设置属性
        this.setAttribute();
        //增加扩展
        this.addExtend(this.options.extend);
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'input');
        
    }
    createElem() {
        //定义表单域创建方式
        let getField = () => {
            if (this.options.type == 'textarea') {
                this.field = axAddElem('textarea', { name: this.options.name, placeholder: this.options.placeholder });
            } else if (this.options.type == 'button') {
                this.field = axAddElem('button', '', this.options.value);
            } else {
                this.field = axAddElem('input', { type: this.options.type, name: this.options.name, placeholder: this.options.placeholder });
            }
        }
        //表单域
        if (this.targetDom) {
            if (this.targetDom.nodeName == 'INPUT' || this.targetDom.nodeName == 'TEXTAREA' || this.targetDom.nodeName == 'BUTTON') {
                //以input的父节点作为parent
                this.parent = this.targetDom.parentElement;
                this.field = this.targetDom;
            } else {
                //父节点就是target
                this.parent = this.targetDom;
                getField();
            }
        } else {
            //在内存中创建
            this.parent = axAddElem('div', { class: 'ax-form-wrapper' });
            getField();
        }
        this.parent.appendChild(this.field);
        //右侧工具
        if (this.parent.querySelector('.ax-pos-right')) {
            this.tools = this.parent.querySelector('.ax-pos-right');
        } else {
            this.tools = axAddElem('span', { class: "ax-pos-right" });
        }
        //清空按钮
        if (this.tools.querySelector('[clear]')) {
            this.clearBtn = this.tools.querySelector('[clear]');
        } else {
            this.clearBtn = axAddElem('a', { href: '###', class: "ax-iconfont ax-icon-close", clear: '' });
            if (this.options.clearShow) {
                this.tools.appendChild(this.clearBtn);
            }
        }


        //清空
        if (this.clearBtn) {
            this.clearBtn.onclick = () => {
                this.clear();
            }
            //给节点做个标记，避免重复绑定
            this.clearBtn.clear = true;
        }
        this.options.toolsShow ? this.parent.appendChild(this.tools) : null;
        //字数限制
        if (this.options.limit) {
            this.limit = axAddElem('div', { class: "ax-form-limit" }, axTplEngine(this.options.language.limit, { value: this.options.limit }));
            let have = this.limit.querySelector('[have]'),
                remainder = this.limit.querySelector('[remainder]');
            this.parent.appendChild(this.limit);
            this.field.oninput = () => {
                let obj = this.intercept(this.field.value);
                this.field.value = obj.text;
                have ? have.innerHTML = obj.have : null;
                remainder ? remainder.innerHTML = obj.remainder : null;
                //回调
                this.options.onInput && this.options.onInput.call(this, this.output);
                'input' in this.handlers ? this.emit('input', this.output) : null;
            }
        }
        //初始化值
        this.options.value ? this.set(this.options.value) : null;
        //追加输出
        this.output.dom = this.field;
        this.output.parent = this.parent;
        this.output.type = this.options.type;
        //回调
        this.options.onInit && this.options.onInit.call(this, this.output);
    }
    setAttribute() {
        //添加类和属性
        this.options.className ? this.field.setAttribute('class', this.options.className) : null;
        this.options.width ? this.field.style.width = this.options.width : null;
        this.options.height ? this.field.style.height = this.options.height : null;
        this.options.size ? this.field.classList.add(`ax-${this.options.size}`) : null;
        //添加其他属性
        for (let i in this.options.props) {
            this.field.setAttribute(i, this.options.props[i]);
        }
    }
    addExtend(data) {
        if (!this.options.toolsShow) {
            return false;
        }
        let render = (obj) => {
            //obj={name:'',icon:'',text:'',node:'',props:{},callback:''}
            let btn = axAddElem(obj.node || 'a', { class: obj.icon }, obj.text);
            //将name所谓属性添加
            btn.setAttribute(obj.name, '');
            //追加其他属性
            if (obj.props) {
                for (let i in obj.props) {
                    this.field.setAttribute(i, obj.props[i]);
                }
            }
            this.tools.appendChild(btn);
            obj.callback && obj.callback.call(this, btn, this.field);
        }
        if (Array.isArray(data)) {
            data.forEach(i => {
                render(i);
            });
        } else {
            render(data);
        }
    }
    intercept(text) {
        let len = text.length,
            obj = {};

        if (len >= this.options.limit && this.options.limit) {
            //已录入字符数超过限制字数，并且限制字数不为0
            obj = { text: text.substring(0, this.options.limit), have: this.options.limit, remainder: 0 }
            this.limit ? this.limit.classList.add('ax-color-error') : null;
            //回调
            this.options.onExceed && this.options.onExceed.call(this, this.output);
            'exceed' in this.handlers ? this.emit('exceed', this.output) : null;
        } else {
            //未超过限制字数或限制为0
            this.limit ? this.limit.classList.remove('ax-color-error') : null;
            obj = { text: text, have: text.length, remainder: Math.abs(this.options.limit - len) }
        }

        return obj;
    }
    set(str, callback) {
        //如果str是数组、对象或数字等先转为字符串
        if(Array.isArray(str)){
            str = JSON.stringify(str[0]);
        }else if(typeof str != 'string'){
            str = JSON.stringify(str);
        }

        let text = str.replace(/<[^>]+>/g, '').trim(),
            obj = this.intercept(text);
        this.field.value = obj.text;
        //改变内容
        if (this.limit) {
            this.limit.querySelector('[have]') ? this.limit.querySelector('[have]').innerHTML = obj.have : null;
            this.limit.querySelector('[remainder]') ? this.limit.querySelector('[remainder]').innerHTML = obj.remainder : null;
        }
        //追加输出
        this.output.value = this.field.value;
        //回调
        callback && callback.call(this, this.field, obj.text, obj.remainder);
        this.options.onSet && this.options.onSet.call(this, this.field, obj.text, obj.remainder);
        'set' in this.handlers ? this.emit('set', this.field, obj.text, obj.remainder) : null;
        return this;
    }
    getValues() {
        return this.field.value;
    }
    clear(callback) {
        this.set('');
        //回调
        callback && callback.call(this, this.field);
        this.options.onClear && this.options.onClear.call(this, this.field);
        'clear' in this.handlers ? this.emit('clear', this.field) : null;

    }
    reset(callback){
        this.set(this.options.value);
        //回调
        callback && callback.call(this, this.field);
        this.options.onReset && this.options.onReset.call(this, this.field);
        'reset' in this.handlers ? this.emit('reset', this.field) : null;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    document.querySelectorAll('[axInput]').forEach(element => {
        new axInput(element);
    });
})(); */
/*!
 * 插件：单选和多选文本域；使用方法：new axCheckbox('#id',{参数})
 */
class axCheckbox {
    constructor(elem, options) {
        //定义节点DOM
        this.targetDom = elem ? axIdToDom(elem) : '';
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            type: 'checkbox', //input类型，默认checkbox，可选择radio
            style: 'beautify', //显示风格，默认是beautify美化原生，可选择chera风格或switch风格
            name: '', //字段名
            children: '', //设置子项，可以是一个对象(一个子项)，也可以是一个数组(多个子项)，可使用的参数如下：
            /* [{
                checked: false, //设置默认checked，默认为false
                disabled: false, //设置禁用，默认false不禁用，可使用true禁用
                value: '', //设置值
                label: '', //设置标签文字
                props: {}, //设置其他属性(对label设置)
                callback: '', //设置回调函数，支持参数obj：{dom:'',label:'',value:'',checked:'',disabled:''}
            }],  */
            value: '', //初始值，字符串格式，比如“traval,read”
            limit: 0, //最多选择几项，默认为0不限制
            language:{
                limit: `已选择<i have>0</i>项,还可以选择<i remainder><# this.value #></i>项`,//字数限制提示
            },
            separator: ',', //定义输出字符串值的分隔符，默认是英文逗号
            size: '', //统一设置子项尺寸,可以填xs、sm、md或lg
            parentClass: '', //设置父层样式名，一个或多个，多个以英文逗号或空格分开
            childClass: '',//每一项div的样式名，一个或多个，多个以英文逗号或空格分开
            props: '', //设置父层其他属性，比如props:{role:'',display:''}
            display: '', //子项布局方式，默认为空表示子项行内布局，label宽度随内容自由伸缩，可选择grid或block
            gap: '', //如果grid布局,那么可以使用该参数定义子项之间的，比如1.4rem
            columns: 0, //设置列数，默认为0表示子项行内布局，如果大于0将启用grid布局
            clearShow: true, //当toolsShow: true的时候是否显示清除按钮，默认true显示，可选择false不显示
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onClear: '', //清除值后的回调函数，支持参数output，即输出详细
            onSet: '', //设置值后的回调函数，支持参数output，即输出详细
            onReset: '', //重置值后的回调函数，支持参数output，即输出详细
            onAddChecked: '', //追加项后的回调函数，支持参数output，即输出详细
            onExceed: '', //checked项数超过设定值后的回调函数，支持参数output，即输出详细
            onInit: '', //初始化后的回调函数，支持参数output，即输出详细
            onChange: '', //初始化后的回调函数，支持参数output，即输出详细
        }, options, this.targetDom, this.constructor.name);
        //定义监听事件成员
        this.handlers = {};
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //定义输出对象
        //{type:'',parent:'',children:''}
        this.output = { children: [] };
        //定义this.children，统一为数组格式
        this.getChildren();
        //定义name字段名
        this.name = this.options.name;
        //修正limit，如果limit数大于了项目数量，那么limit应该等于项目数量
        this.trueLimit = this.children.length <= this.options.limit ? this.children.length : this.options.limit;
        //创建定义parent
        this.createParent();
        //设置属性
        this.setAttribute();
        //初始化选项
        if (!axIsEmpty(this.options.value)) {
            this.set(this.options.value);
        }
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'checkbox');

        return this;
    }
    getChildren() {
        if (Array.isArray(this.options.children)) {
            //是数组
            //如果没有value属性，则将label赋予value
            if (typeof this.options.children[0] === 'object') {
                //this.options.children=[{},{}]
                this.options.children.forEach(i => {
                    if (!i.hasOwnProperty('value')) {
                        i.value = i.label;
                    }
                });
                this.children = this.options.children;
            } else {
                //this.options.children=['','']
                this.children = this.options.children.map(i => {
                    return { label: i, value: i };
                });
            }
        } else {
            //非数组
            if (typeof this.options.children === 'object') {
                //this.options.children={}
                if (!this.options.children.hasOwnProperty('value')) {
                    this.options.children.value = this.options.children.label;
                }
                this.children = [this.options.children];
            } else {
                //this.options.children=''
                this.children = [{ label: this.options.children, value: this.options.children }]
            }
        }
    }
    createParent() {
        //定义parent
        if (!this.targetDom || this.targetDom.nodeName == 'INPUT') {
            this.parent = axAddElem('div', { class: 'ax-form-wrapper' });
            if (this.targetDom.nodeName == 'INPUT') {
                //如果是type=hidden/text的input
                this.hidden = this.targetDom;
                this.hidden.insertAdjacentElement('beforeBegin', this.parent);
                this.name = this.targetDom.name;
                this.hidden.type != 'hidden' ? this.hidden.style.display = 'none' : null;
            }
        } else {
            //如果有节点，且不是type=hidden的input
            this.parent = this.targetDom;
        }
        if (this.children.length > 0) {
            this.children.forEach(i => {
                this.createElem(i);
            });
        }
        //此时input都没有使用checked，对选项进行checked
        let checkedDoms = this.output.children.filter(i => i.checked);
        for (let i = 0, len = checkedDoms.length; i < len; i++) {
            let item = checkedDoms[i],
                input = item.dom.querySelector('input');
            if (this.options.type == 'radio') {
                if (i > 0) {
                    //对于radio类型只取第一个checked项
                    item.checked = false;
                    break;
                } else {
                    input.setAttribute('checked', '');
                    item.checked = true;
                }
            } else if (this.options.type == 'checkbox') {
                if (i < this.trueLimit || !this.trueLimit) {
                    //小于限制数量的才checked，或者this.trueLimit==0（不限制）
                    input.setAttribute('checked', '');
                    item.checked = true;
                } else {
                    item.checked = false;
                }
            }
        }
        //选项限制
        if (this.options.limit) {
            this.limit = axAddElem('div', { class: "ax-form-limit" }, axTplEngine(this.options.language.limit, { value: this.options.limit }));
            //更新选项数量
            this.resetLimit();
            this.parent.insertAdjacentElement('afterend', this.limit);
        }
        //追加到输出
        this.output.name = this.name;
        this.output.parent = this.parent;
        this.output.type = this.options.type;
        this.output.value = this.getValues();
        //对于隐藏域
        if (this.hidden) {
            this.hidden.value = this.getValues('string', this.options.separator);
        }
        //赋予onchange事件
        this.output.children.forEach(i => {
            let input = i.dom.querySelector('input');
            input.onchange = () => {
                this.change(i);
            }
        });
        //回调
        this.options.onInit && this.options.onInit.call(this, this.output);
    }
    createElem(obj) {
        //定义表单域创建方式
        let tpl = `<label class="ax-${this.options.style == 'beautify' ? this.options.type : this.options.style}"><input name="${this.name + '_ax_alt'}" value="${obj.value}" ${obj.disabled ? 'disabled' : ''} type="${this.options.type}"><u></u>${obj.label ? '<i>' + obj.label + '</i>' : ''}</label>`,
            elem = axStrToDom(axTplEngine(tpl, obj)),
            labelWrapper = axAddElem('div');
        //追加项目样式名
        if (this.options.childClass) {
            axClass.add(labelWrapper, this.options.childClass);
        }
        this.options.size ? elem.classList.add(`ax-${this.options.size}`) : null;
        //添加其他属性
        for (let i in obj.props) {
            elem.setAttribute(i, obj.props[i]);
        }
        //判断是否是grid布局
        if (this.options.display == 'grid' || this.options.display == 'block') {
            labelWrapper.appendChild(elem);
            this.parent.appendChild(labelWrapper)
        } else {
            this.parent.appendChild(elem);
        }
        //追加到输出
        let item = { dom: elem, label: obj.label, value: obj.value, checked: obj.checked ? obj.checked : false, disabled: obj.disabled ? obj.disabled : false }
        this.output.children.push(item);
        //回调
        obj.callback && obj.callback.call(this, item);
        return elem;
    }
    setAttribute() {
        //添加类和属性
        if (this.options.parentClass) {
            axClass.add(this.parent, this.options.parentClass);
        }
        //添加其他属性
        for (let i in this.options.props) {
            this.parent.setAttribute(i, this.options.props[i]);
        }
        //使用grid布局并添加间隔
        if (this.options.display == 'grid') {
            this.parent.classList.add('ax-grid');
            if (this.options.columns) {
                this.parent.classList.add(`ax-grid-${this.options.columns}`);
            };
            if (this.options.gap) {
                this.parent.style.gridColumnGap = this.options.gap;
            }

        }
    }
    change(item) {
        let input = item.dom.querySelector('input'),
            value = item.value;
        if (!value) {
            //值为undefined或为空则不执行
            return false;
        }

        if (input.checked) {
            if (this.options.type == 'radio') {
                //之前checked项为false
                let old = this.output.children.find(i => i.checked);
                old ? old.checked = false : null;
            }
            item.checked = true;
        } else {
            //这里只有checkbox
            item.checked = false;
        }
        //如果超过限制则将值回设为false
        let checkedItems = this.output.children.filter(i => i.checked);
        if (this.trueLimit && checkedItems.length > this.trueLimit) {
            input.checked = false;
            item.checked = false;
        }
        //修正输出值
        this.output.value = this.getValues();
        //对于隐藏域
        if (this.hidden) {
            this.hidden.value = this.getValues('string', this.options.separator);
            //手动执行change事件
            this.hidden.onchange ? this.hidden.onchange() : null;
        }
        //修正选择数量
        this.resetLimit();
        //超数量拦截
        this.intercept();
        //回调
        this.options.onChange && this.options.onChange.call(this, item);
        'change' in this.handlers ? this.emit('change', item) : null;
    }
    intercept() {
        let len = this.output.children.filter(i => i.checked).length,
            flag = false;
        if (!this.trueLimit) {
            //为0则不拦截
            return false;
        }
        if (len >= this.trueLimit) {
            flag = true;
            //回调
            this.options.onExceed && this.options.onExceed.call(this, this.output);
            'exceed' in this.handlers ? this.emit('exceed', this.output) : null;
        } else {
            flag = false;
        }
        return flag;
    }
    resetLimit() {
        if (!this.trueLimit) {
            //为0则不拦截
            return false;
        }
        let len = this.output.children.filter(i => i.checked).length,
            have = this.limit.querySelector('[have]'),
            remainder = this.limit.querySelector('[remainder]');

        if (len >= this.trueLimit) {
            this.limit ? this.limit.classList.add('ax-color-error') : null;
            //修正数量
            have.innerHTML = this.trueLimit;
            remainder.innerHTML = 0;
        } else {
            this.limit ? this.limit.classList.remove('ax-color-error') : null;
            //修正数量
            have.innerHTML = len;
            remainder.innerHTML = this.trueLimit - len;
        }
    }
    set(data, callback) {
        if (this.options.type == 'radio') {
            //radio只取第一个值
            let str = '';
            if (Array.isArray(data)) {
                str = data[0];
            } else {
                str = data;
            }
            //空值则不再执行
            let item, other;
            if (!str) {
                item = this.output.children.find(i => i.checked);
                if (item) {
                    item.checked = false;
                    item.dom.querySelector('input').checked = false;
                }
            } else {
                item = this.output.children.find(i => i.value == str);
                other = this.output.children.filter(i => i.value != str);
                //有对应项才执行
                if (item) {
                    //第一项checked
                    item.checked = true;
                    item.dom.querySelector('input').checked = true;
                }
                //其他项全部取消checked
                other.forEach(i => {
                    i.checked = false;
                    i.dom.querySelector('input').checked = false;
                });
            }

        } else if (this.options.type == 'checkbox') {
            //checkbox需要使用limit
            //检测值的有效性
            let arr;
            if (Array.isArray(data)) {
                arr = data;
            } else if (typeof data === 'string') {
                arr = data.split(this.options.separator);
            } else {
                arr = [data];
            }
            //清除空值
            arr = arr.filter(Boolean);
            //数量拦截
            if (this.trueLimit && arr.length > this.trueLimit) {
                arr = arr.slice(0, this.trueLimit);
            }
            //循环赋值
            this.output.children.forEach(i => {
                let input = i.dom.querySelector('input');
                if (arr.includes(i.value)) {
                    i.checked = true;
                    input.checked = true;
                } else {
                    i.checked = false;
                    input.checked = false;
                }
            });
        }
        //修正输出值
        this.output.value = this.getValues();
        this.hidden ? this.hidden.value = this.getValues('string', this.options.separator) : null;
        //更新数量提示
        this.resetLimit();
        //回调
        callback && callback.call(this, this.output);
        this.options.onSet && this.options.onSet.call(this, this.output);
        'set' in this.handlers ? this.emit('set', this.output) : null;
        return this;
    }
    addChecked(data, callback) {
        //radio类型禁止使用
        if (this.options.type == 'radio') {
            console.error('Do not use this method of addChecked() in type of radio!');
            return false;
        }
        let trueData = [];
        if (Array.isArray(data)) {
            trueData = data;
        } else {
            trueData = [data];
        }
        //清除空值
        trueData = trueData.filter(Boolean);
        for (let i = 0, len = trueData.length; i < len; i++) {
            if (this.intercept()) {
                //超出限制数量则不再添加checked
                break;
            }
            let item = trueData[i];
            let find = this.output.children.find(i => i.value == item);
            find.dom.querySelector('input').checked = true;
            find.checked = true;
            //修正数量提示
            this.resetLimit();
        }
        //修正输出值
        this.output.value = this.getValues();
        this.hidden ? this.hidden.value = this.getValues('string', this.options.separator) : null;
        //回调
        callback && callback.call(this, trueData, this.output);
        this.options.onAddChecked && this.options.onAddChecked.call(this, trueData, this.output);
        'addChecked' in this.handlers ? this.emit('addChecked', trueData, this.output) : null;
    }
    checkedAll(callback) {
        this.output.children.forEach(i => {
            i.dom.querySelector('input').checked = true;
            i.checked = true;
        });
        //修正数量提示
        this.resetLimit();
        //修正输出值
        this.output.value = this.getValues();
        //回调
        callback && callback.call(this, this.output);
        this.options.onCheckedAll && this.options.onCheckedAll.call(this, this.output);
        'checkedAll' in this.handlers ? this.emit('checkedAll', this.output) : null;
    }
    getValues(type, separator) {
        type = type || 'array';
        separator = separator || this.options.separator;
        let values = [];
        this.output.children.forEach(i => {
            if (i.checked) {
                if (type == 'json') {
                    values.push({ label: i.label, value: i.value });
                } else {
                    values.push(i.value);
                }
            }
        });
        //string类型，输出“book,travel”
        if (type == 'string') {
            values = values.join(separator);
        }
        return values;
    }
    reset(callback) {
        let value;
        if (this.options.value) {
            //从初始值获取
            value = this.options.value;
        } else {
            //从原children数据中获取value数组，使用set方法重置
            let checkedItem = this.options.children.filter(i => i.checked);
            value = [];
            checkedItem.forEach(i => {
                value.push(i.value);
            });
        }
        this.set(value);
        //回调
        callback && callback.call(this, this.output);
        this.options.onReset && this.options.onReset.call(this, this.output);
        'reset' in this.handlers ? this.emit('reset', this.output) : null;
    }
    clear(callback) {
        this.output.children.forEach(i => {
            if (i.checked) {
                i.checked = false;
                i.dom.querySelector('input').checked = false;
            }
        });
        //修正输出值
        this.output.value = [];
        //修正选项限制数量
        this.resetLimit();
        //回调
        callback && callback.call(this, this.output);
        this.options.onClear && this.options.onClear.call(this, this.output);
        'clear' in this.handlers ? this.emit('clear', this.output) : null;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    document.querySelectorAll('[axCheckbox]').forEach(element => {
        new axCheckbox(element);
    });
})(); */
/*!
 * 插件：复选列表全选和取消全选；使用方法：new axCheckAll('#id',{参数})
 */
class axCheckAll {
    constructor(elem, options) {
        //定义节点DOM
        this.targetDom = elem ? axIdToDom(elem) : '';
        //没有节点便不执行了
        if (!this.targetDom) {
            return false;
        }
        this.options = axExtend({
            checkDom: '', //全选列表的按钮节点，可以是id、className、nodeName或者节点
            clearDom: '', //取消全选列表的按钮节点，可以是id、className、nodeName或者节点，如果使用checkbox或radio进行全选，clearDom不填
            separator: ',', //获取值转字符串值时的分隔符，默认为英文逗号
            includeDisabled: true,//是否包含禁用项（disabled项），默认true即全部项，可选择false只使用可用项
            onInit: '', //回调函数，初始化后执行，无参数
            onChecked: '', //回调函数，部分checked后执行，支持一个参数，即当前值数组
            onCheckAll: '', //回调函数，全部checked后执行，无参数
            onClearAll: '', //回调函数，全部取消checked后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        //定义监听事件成员
        this.handlers = {};
        //定义节点
        this.checkDom = this.options.checkDom ? axIdToDom(this.options.checkDom) : '';
        this.clearDom = this.options.clearDom ? axIdToDom(this.options.clearDom) : '';

        //定义列表
        this.list = [];
        this.listAvailable = [];
        //定义类型
        this.toggleType = '';
        //列表状态
        this.state = '';
        //切换按钮
        this.toggleBtns = [];
        //定义checked组的name
        this.name = '';
        //初始化
        this.init();
    }
    init() {
        //从物理节点中获取列表
        this.getOptionsList();
        //没有label和input节点就不再执行
        if (this.list.length === 0) {
            return false;
        }
        //根据不同类型进行操作
        this.observeToggle();
        //监听复选列表checked状态
        this.observeList();
        //初始化监听回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    observeList() {
        //联动
        this.setLinkage();
        this.listAvailable.forEach(i => {
            i.field.addEventListener('change', () => {
                this.setLinkage();
            }, false);
        });
    }
    observeToggle() {
        if (this.checkDom && this.checkDom.type === 'checkbox') {
            //通过一个checkbox切换全选和非全选
            this.toggleType = 'checkbox';
            this.toggleBtns = [this.checkDom];
            //初始化
            if (this.checkDom.checked) {
                this.checkAll();
            }
            //监听
            this.checkDom.addEventListener('change', () => {
                if (this.checkDom.checked) {
                    this.checkAll();
                } else {
                    this.clearAll();
                }
            }, false);
        } else if (this.checkDom && this.checkDom.type === 'radio') {
            //通过两个radio切换全选和非全选
            this.toggleType = 'radio';
            this.toggleBtns = [...document.querySelectorAll(this.options.checkDom.replace('[', '[name='))];
            //初始化
            if (axMultiValues(this.options.checkDom, 'radio', 'string') == 1) {
                this.checkAll();
            }
            //监听
            this.toggleBtns.forEach(i => {
                i.addEventListener('change', () => {
                    let value = axMultiValues(this.options.checkDom, 'radio', 'string');
                    if (value == 1) {
                        this.checkAll();
                    } else {
                        this.clearAll();
                    }
                }, false);
            });
        } else {
            //通过两个按钮切换全选和非全选
            this.toggleType = 'button';
            if (this.checkDom) {
                this.toggleBtns.push(this.checkDom);
                this.checkDom.onclick = () => {
                    this.checkAll();
                }
            }
            if (this.clearDom) {
                this.toggleBtns.push(this.clearDom);
                this.clearDom.onclick = () => {
                    this.clearAll();
                }
            }
        }
    }
    setLinkage() {
        let value = axMultiValues(this.name, 'checkbox', 'array');
        if (value.length === 0) {
            this.state = 'cleared';
            this.toggleBtns.forEach(i => {
                i.removeAttribute('checking');
            });
            //全不选，则取消input按钮的checked状态
            if (this.checkDom && this.toggleType === 'checkbox') {
                this.checkDom.checked = false;
            } else if (this.toggleType = 'radio') {
                let radioChecked = this.toggleBtns.find(k => k.checked);
                if (radioChecked) {
                    radioChecked.checked = false;
                }
            }
            //回调
            this.options.onClearAll && this.options.onClearAll.call(this);
            'clearAll' in this.handlers ? this.emit('clearAll', '') : null;
        } else if (value.length === this.listAvailable.length) {
            this.state = 'checked';
            this.toggleBtns.forEach(i => {
                i.removeAttribute('checking');
            });
            //全选了，那么input按钮也选上
            if (this.checkDom && this.toggleType === 'checkbox') {
                this.checkDom.checked = true;
            } else if (this.toggleType = 'radio') {
                let radio = this.toggleBtns.find(k => k.value == 1);
                if (radio) {
                    radio.checked = true;
                }
            }
            //回调
            this.options.onCheckAll && this.options.onCheckAll.call(this);
            'checkAll' in this.handlers ? this.emit('checkAll', '') : null;
        } else {
            this.state = 'checking';
            //只要有一个没选，则input按钮使用第三态
            if (this.checkDom && this.toggleType === 'checkbox') {
                this.checkDom.checked = false;
                this.checkDom.setAttribute('checking', '');
            } else if (this.toggleType = 'radio') {
                //let radio = this.toggleBtns.find(k => k.checked);

                this.toggleBtns.forEach(i => {
                    //取消checked状态
                    i.checked ? i.checked = false : null;
                    //value为1的radio追加checking属性
                    i.value == 1 ? i.setAttribute('checking', '') : null;
                });
            }
        }
    }
    getOptionsList(all = this.options.includeDisabled) {
        //all为是否包含整个列表，包括disabled的项，false表示不包含禁用项
        [...this.targetDom.querySelectorAll('label')].forEach(i => {
            let field = i.querySelector('input');
            if (all) {
                this.list.push({ dom: i, field: field, name: field.name, value: field.value });
            } else {
                if (!i.field.disabled) {
                    this.list.push({ dom: i, field: field, name: field.name, value: field.value });
                }
            }
        });
        //校验是否都有name
        if (this.list.find(i => !i.name)) {
            console.error('Checkbox must have name attribute!');
        }
        //校验是否name都一致(都与第一个相比)
        if (this.list.some(i => i.name != this.list[0].name)) {
            console.error('All names need to be consistent!');
        }
        //校验是否都有值
        if (this.list.find(i => !i.value || !i.field.hasAttribute('value'))) {
            console.error('Checkbox must have a value!');
        }
        //可用的复选列表
        this.listAvailable = all ? this.list.filter(i => !i.field.disabled) : this.list;
        //获取name名
        this.name = this.listAvailable[0].name;
        return this.list;
    }
    getValues(type, separator) {
        type = type || 'array';
        separator = separator || this.options.separator;
        let arr = this.list.filter(i => i.field.checked),
            valueArr = arr.map(i => i.field.value);
        if (type === 'string') {
            return valueArr.join(separator);
        } else if (type === 'array') {
            return valueArr;
        } else {
            return '';
        }
    }
    checked(values, callback) {
        let items = [];
        if (typeof values === 'string') {
            //如果是字符串
            items = values.split(this.options.separator);
        } else if (Array.isArray(values)) {
            //如果是数组
            items = values;
        } else {
            return false;
        }

        //根据选项进行单项checked
        items.forEach(i => {
            let item = this.list.find(k => k.value == i);
            item ? item.field.checked = true : null;
        });
        //再与按钮联动
        this.setLinkage();

        //回调
        callback && callback.call(this, items);
        this.options.onChecked && this.options.onChecked.call(this, items);
        'checked' in this.handlers ? this.emit('checked', items) : null;
        return this;
    }
    checkAll(callback) {
        this.state = 'checked';

        this.list.forEach(i => {
            if (!i.field.disabled) {
                i.field.checked = true;
            }
        });
        //radios和checkbox类型，需要清除checking属性
        this.toggleBtns.forEach(i => {
            i.removeAttribute('checking');
        });
        //回调
        callback && callback.call(this);
        this.options.onCheckAll && this.options.onCheckAll.call(this);
        'checkAll' in this.handlers ? this.emit('checkAll', '') : null;
        return this;
    }
    clearAll(callback) {
        this.state = 'cleared';

        this.list.forEach(i => {
            if (i.field.checked) {
                i.field.checked = false;
            }
        });
        //radios和checkbox类型，需要清除checking属性
        this.toggleBtns.forEach(i => {
            i.removeAttribute('checking');
        });
        //回调
        callback && callback.call(this);
        this.options.onClearAll && this.options.onClearAll.call(this);
        'clearAll' in this.handlers ? this.emit('clearAll', '') : null;
        return this;
    }


    on(type, handler) {
        axAddPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        //返回实例，以便于链式操作
        return this;
    }
}
/* (() => {
    document.querySelectorAll('[axCheckAll]').forEach(element => {
        new axCheckAll(element);
    });
})(); */
/*!
 * 插件：创建表单；使用方法：new axForm([{}],'#id')
 */
class axForm {
    constructor(targetDom, obj) {
        //node,[{type = 'text', fieldName = '', title = '', placeholder = '', value = '', offset, props = {}, trigger, callback, valid},...]
        //[{type:'radio',name:'danxuan',value:'',trigger:'click',callback:()=>{},title:'选择性别：',placeholder:'请选择...',offset:'',valid:'',props:{}}]

        // 始终将targetDom转成DOM对象
        this.targetDom = targetDom ? axIdToDom(targetDom) : null;
        //定义form父层和每项对象
        this.form;
        this.output = [];
        // 定义监听事件成员
        this.handlers = {};
        //添加form以及form属性
        if (this.targetDom) {
            //存在初始节点
            if (this.targetDom.nodeName == 'FORM') {
                this.form = this.targetDom;
            } else {
                this.form = axAddElem('form');
                this.targetDom.appendChild(this.form);
            }
        } else {
            //不存在初始节点
            this.form = axAddElem('form');
        }
        this.init(obj);
        return this;
    }
    init(obj) {
        //追加样式和属性
        this.setAttribute(obj);
        //添加子项
        obj.rows.forEach(i => {
            this.createSection(i, obj.style);
        });
        //回调
        obj.onInit && obj.onInit.call(this, this.output);
        //添加到实例合集
        axInstance.push(this, obj.insName, 'form');
        return this;
    }
    createTpl(obj, style) {
        let title = '',
            separate = '',
            props = '';
        //创建标题
        if (obj.title && obj.title.show) {
            if (obj.title.html) {
                title = obj.title.html;
            } else {
                title = `<div class="ax-form-label${obj.title.className ? ' ' + obj.title.className : ''}">${obj.title.text ? obj.title.text : ''}</div>`;
            }
        }
        //创建分隔符
        if (style == 'emulate') {
            separate = `<div class="ax-break-line"></div>`;
        } else if (style == 'fit') {
            separate = `<div class="ax-break ax-hide-tel"></div><div class="ax-break-line ax-hide-pad ax-hide-pc"></div>`;
        } else {
            separate = `<div class="ax-break"></div>`;
        }
        //追加属性
        if (!axIsEmpty(obj.props)) {
            for (let i in obj.props) {
                props += ` ${i}="${obj.props[i]}" `;
            }
        }

        return `
                    <section${props}>
                    ${obj.before || ''}
                        <div class="ax-form-group">
                            <div class="ax-flex-row">
                                ${title}
                                <div class="ax-form-con">
                                    <div class="ax-form-input"></div>
                                </div>
                                ${obj.tools ? obj.tools : ''}
                            </div>
                        </div>
                        ${obj.after || ''}
                        ${separate}
                    </section >
                    `;
    }
    setAttribute(obj) {
        //追加风格
        if (obj.style == 'emulate') {
            this.form.classList.add('ax-emulate');
        } else if (obj.style == 'fit') {
            this.form.classList.add('ax-fit-tel');
        }
        //追加类名
        obj.className ? this.form.classList.add(obj.className) : null;
        //追加属性
        if (!axIsEmpty(obj.props)) {
            for (let i in obj.props) {
                this.form.setAttribute(i, obj.props[i])
            }
        }
    }
    createCols(cols, parent, display, columns, gap, callback) {
        cols.forEach(i => {
            let colDiv, field;
            if (display == 'grid') {
                parent.classList.add('ax-grid');
                colDiv = axAddElem('div');
                columns ? parent.classList.add('ax-grid-' + columns) : null;
                parent.style.gridGap = gap;
                field = colDiv;
                parent.appendChild(colDiv);
                callback(i, colDiv);
            } else if (display == 'flex') {
                parent.classList.add('ax-row');
                colDiv = axAddElem('div', { class: 'ax-col' });
                field = colDiv;
                parent.appendChild(colDiv);
                callback(i, colDiv);
            } else {
                colDiv = parent;
                field = callback(i, colDiv);
            }
            //增加前后html字符
            i.prefix ? field.insertAdjacentHTML('beforebegin', i.prefix) : null;
            i.suffix ? field.insertAdjacentHTML('afterend', i.suffix) : null;
        });
    };
    createSection(obj, style) {
        let section = axStrToDom(this.createTpl(obj, style)),
            wrapper = section.querySelector('.ax-form-input');
        //item = { type: obj.type, name: obj.name, dom: section, instance: '' }
        if (axIsEmpty(obj.cols)) {
            //一行一列模式
            this.createItem(obj, wrapper);
        } else {
            //一行多列模式
            this.createCols(obj.cols, wrapper, obj.display, obj.columns, obj.gap, (item, parent) => {
                return this.createItem(item, parent);
            });
        }
        //生成节点
        this.form.appendChild(section);
    }
    createItem(obj, parent) {
        let field,
            item = {};
        if (obj.type == 'select') {
            //下拉菜单
            field = axAddElem('select', { name: obj.name || '', ...obj.props });
            //增加样式
            if (obj.className) {
                field.setAttribute('class', obj.className);
            }
            parent.appendChild(field);
            if (obj.options && obj.options.length > 0) {
                //options=[{label:'',value:'',checked:'',disabled:'',gutter:'',callback:''},...]
                obj.options.forEach(i => {
                    let opt = axAddElem('option', { value: i.value || '' }, i.label);
                    i.checked ? opt.setAttribute('selected', '') : null;
                    i.disabled ? opt.setAttribute('disabled', '') : null;
                    field.appendChild(opt);
                    //选项回调
                    i.callback && i.callback.call(this, opt, i);
                });
            }

        } else if (obj.type == 'radio') {
            //单选
            //给父层增加属性
            for (let i in obj.props) {
                parent.setAttribute(i, obj.props[i]);
            }
            //增加样式
            if (obj.className) {
                parent.setAttribute('class', obj.className);
            }
            field = [];
            this.createCols(obj.options, parent, obj.display, obj.columns, obj.gap, (i, p) => {
                let item = axAddElem('input', { type: 'radio', name: obj.name || '', value: i.value || '' }),
                    label = axAddElem('label', { class: 'ax-radio' }, `<u></u><i> ${i.label}</i>`);
                i.checked ? item.checked = true : null;
                i.disabled ? item.disabled = true : null;
                label.insertAdjacentElement('afterBegin', item);
                p.appendChild(label);
                field.push(item);
                return label;
            });
        } else if (obj.type == 'checkbox') {
            //多选
            //给父层增加属性
            for (let i in obj.props) {
                parent.setAttribute(i, obj.props[i]);
            }
            //增加样式
            if (obj.className) {
                parent.setAttribute('class', obj.className);
            }
            field = [];
            this.createCols(obj.options, parent, obj.display, obj.columns, obj.gap, (i, p) => {
                let item = axAddElem('input', { type: 'checkbox', name: obj.name || '', value: i.value || '' }),
                    label = axAddElem('label', { class: 'ax-checkbox' }, `<u></u><i> ${i.label}</i>`);
                i.checked ? item.checked = true : null;
                i.disabled ? item.disabled = true : null;
                label.insertAdjacentElement('afterBegin', item);
                p.appendChild(label);
                field.push(item);
                return label;
            });
        } else if (obj.type == 'html') {
            //按钮
            field = axAddElem('div', obj.props, obj.value || '');
            //增加样式
            if (obj.className) {
                field.setAttribute('class', obj.className);
            }
            parent.appendChild(field);
        } else if (obj.type == 'a') {
            //按钮
            field = axAddElem('a', obj.props, obj.value || '');
            //增加样式
            if (obj.className) {
                field.setAttribute('class', obj.className);
            }
            parent.appendChild(field);
        } else if (obj.type == 'button') {
            //按钮
            field = axAddElem('button', obj.props, obj.value || '');
            //增加样式
            if (obj.className) {
                field.setAttribute('class', obj.className);
            }
            parent.appendChild(field);
        } else if (obj.type == 'textarea') {
            //多行文本
            field = axAddElem('textarea', { name: obj.name || '', value: obj.value || '', ...obj.props });
            obj.readonly ?  field.readonly = true : null;
            obj.disabled ?  field.disabled = true : null;
            //增加样式
            if (obj.className) {
                field.setAttribute('class', obj.className);
            }
            obj.placeholder ? field.setAttribute('placeholder',obj.placeholder) : null;
            parent.appendChild(field);
        } else {
            //单行文本
            field = axAddElem('input', { type: obj.type, name: obj.name || '', value: obj.value || '', ...obj.props });
            obj.readonly ?  field.readonly = true : null;
            obj.disabled ?  field.disabled = true : null;
            //增加样式
            if (obj.className) {
                field.setAttribute('class', obj.className);
            }
            obj.placeholder ? field.setAttribute('placeholder',obj.placeholder) : null;
            parent.appendChild(field);
        }
        //将本字段相关内容追加进表单字段合集
        //item={name:'',dom:'',value:'',instance:''}
        item.name = obj.name || '';
        item.value = obj.value || '';
        item.field = field;
        item.parent = parent;
        this.output.push(item);
        //添加字段回调
        obj.callback && obj.callback.call(this, item);
        return field;
    }
    get(type = 'string', separator = ',') {
        let inputs = [];
        this.output.forEach(i => {
            inputs.push(i.field);
        });
        //由于radio和checkbox的field是子项数组，所以扁平化
        inputs = inputs.flat(2).filter(i => ['INPUT', 'SELECT', 'TEXTAREA'].includes(i.nodeName) && i.name);
        return axSerialize(inputs, type, separator);
    }
    update(obj, callback) {
        //先清空内容
        this.form.innerHTML = '';
        this.output = [];
        //重新创建section
        this.init(obj);
        //回调
         //回调
         obj.onUpdate && obj.onUpdate.call(this, this.output);
        callback && callback.call(this, this.output);
        'updated' in this.handlers ? this.emit('updated', this.output) : null;
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        // 返回实例，以便于链式操作
        return this;
    }
}
/*!
* 插件：表单标题避让；使用方法：new axDodge('#id',{参数})
*/
class axDodge {
    constructor(elem, options) {
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }
        //取得表单控件
        this.inputDom = this.targetDom.querySelector("input,textarea");
        if (!this.inputDom) { return false; }

        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            placeholder: '请输入...',//input的提示
            onInit: '',//回调函数，初始化后执行，无参数
            onSet: '',//回调函数，设置值并聚焦后执行，支持回调函数，回调带一个参数即input的值
            onClear: '',//回调函数，清空失焦后执行，支持回调函数，回调无参数
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //设置状态
        this.isFocus = false;
        //焦点事件
        this.focusEvent = () => {
            this.inputDom.focus();
            this.targetDom.classList.add("ax-active");
            this.inputDom.removeAttribute('readonly');
            this.isFocus = true;
        }
        this.blurEvent = () => {
            this.inputDom.blur();
            if (!this.inputDom.value) {
                this.targetDom.classList.remove("ax-active");
            } else {
                this.targetDom.classList.add("ax-active");
            }
            this.inputDom.setAttribute('readonly', '');
            this.isFocus = false;
        }
        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'dodge');
        //设置风格
        this.setAttribute();
        //加入label节点
        this.labelDom = axAddElem('label');
        let placeholder = this.inputDom.getAttribute("placeholder");
        if (placeholder) {
            this.labelDom.innerHTML = placeholder;
            this.inputDom.removeAttribute("placeholder");
        } else {
            this.labelDom.innerHTML = this.options.placeholder;
        }
        this.targetDom.appendChild(this.labelDom);
        //绑定事件
        this.inputDom.addEventListener('focus', this.focusEvent, false);
        this.inputDom.addEventListener('blur', this.blurEvent, false);
        //点击标签聚焦
        this.labelDom.onclick = () => {
            this.focusEvent();
        };
        //初始化有值则使用聚焦事件
        this.inputDom.value ? this.focusEvent() : null;

        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    setAttribute() {
        //设为readonly避免自动填充
        this.inputDom.setAttribute('readonly', '');
        //随input尺寸
        this.inputDom.classList.contains("ax-lg") ? this.targetDom.classList.add("ax-lg") : null;
        //添加ax-dodge
        !this.targetDom.classList.contains('ax-dodge') ? this.targetDom.classList.add('ax-dodge') : null;
    }
    set(value, callback) {
        if (value) {
            this.inputDom.value = value;
            this.focusEvent();
        } else {
            this.inputDom.focus();
        }
        //监听回调
        this.options.onSet && this.options.onSet.call(this, value);
        this.handlers.hasOwnProperty('set') ? this.emit('set', value) : null;
        callback && callback.call(this, value);
        return this;
    }
    clear(callback) {
        //清除值并失焦
        this.inputDom.value = '';
        this.inputDom.blur();
        this.blurEvent();
        //监听回调
         this.options.onClear && this.options.onClear.call(this);
         this.handlers.hasOwnProperty('clear') ? this.emit('clear', '') : null;
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
* 插件：操作面板；使用方法：new axPanel('#id',{参数})
*/
class axPanel {
    constructor(elem, options) {
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            type: 'panel',//该插件是panel和box两类面板共用的，默认是panel类型，可选box
            display: '',//body部分的显示方式，默认为block，该参数需要跟随css，如果css中定义为flex，那么该参数则需要写为flex
            expanded: true,//初始化是否是展开状态，默认true是展开，可选择false初始化折叠
            headerClosable: false,//是否让整个头部可点击切换内容，默认false，可选择true
            arrow: {
                enable: true,//是否显示指示箭头，默认true显示，可填false不显示
                rotate: 180,//指示箭头的旋转角度，可填0、90或180；如果填0表示不旋转，为了表示打开和关闭，那么iconClose和iconOpen应该不同
                iconClose: 'ax-icon-down',//指示箭头折叠图标class类名
                iconOpen: 'ax-icon-down',//指示箭头展开图标class类名
                position: 'end',//指示箭头所在位置，默认end即在末尾，可选择begin在标题前面
            },
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onCollapsed: '',//回调函数，折叠后执行，支持回调函数，无参数
            onExpanded: '',//回调函数，展开后执行，支持回调函数，无参数
            onRemove: '',//回调函数，删除panel节点后执行，支持回调函数，无参数
            onRestore: '',//回调函数，恢复panel节点后执行，支持回调函数，无参数
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //保存父节点
        this.location = { parent: this.targetDom.parentElement, brother: this.targetDom.nextElementSibling };
        //设置状态
        this.expanded = true;
        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'panel');
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //设置风格
        this.headerDom = this.targetDom.querySelector(`.ax-panel-header`);
        this.bodyDom = this.targetDom.querySelector(`.ax-panel-body`);
        //取得arrowDom
        if (this.headerDom.querySelector('[arrow]')) {
            this.arrowDom = this.headerDom.querySelector('[arrow]');
        } else {
            if (this.options.arrow.enable) {
                this.arrowDom = axAddElem('i', { arrow: '', class: `ax-iconfont ${this.options.arrow.iconOpen}` });
                if (this.options.arrow.position === 'begin') {
                    this.headerDom.insertAdjacentElement('afterbegin', this.arrowDom);
                } else if (this.options.arrow.position === 'end') {
                    this.headerDom.insertAdjacentElement('beforeend', this.arrowDom);
                }
            }
        }



        //设置节点样式
        this.setAttribute();

        //默认状态
        if (this.options.expanded && !axIsVisible(this.bodyDom)) {
            this.expand();
        } else if (!this.options.expanded && axIsVisible(this.bodyDom)) {
            this.collapse();
        }
        //点击切换body状态
        if (this.options.headerClosable) {
            this.headerDom.onclick = () => {
                if (axIsVisible(this.bodyDom)) {
                    this.collapse();
                } else {
                    this.expand();
                }
            }
        } else {
            if (this.arrowDom) {
                this.arrowDom.onclick = () => {
                    if (axIsVisible(this.bodyDom)) {
                        this.collapse();
                    } else {
                        this.expand();
                    }
                };
            }
        }


        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    setAttribute() {
        //添加ax-panel
        !this.targetDom.classList.contains(`ax-panel`) ? this.targetDom.classList.add(`ax-panel`) : null;
        //添加类型属性
        if (this.targetDom.getAttribute('type') === 'box') {
            //修正type参数
            this.options.type = 'box';
        } else {
            this.options.type === 'box' ? this.targetDom.setAttribute('type', this.options.type) : null;
        }
        //显示指示
        this.options.lamp ? this.targetDom.setAttribute('lamp', this.options.lamp) : null;
        //头部可点击
        this.options.headerClosable ? this.targetDom.setAttribute('headerClosable', '') : null;
    }
    expand(callback) {
        //是展开状态则不再执行
        if (this.expanded) {
            return false;
        }
        if (this.arrowDom) {
            if (this.options.arrow.iconClose === this.options.arrow.iconOpen) {
                //翻转图标
                this.arrowDom.classList.remove(`ax-deg${this.options.arrow.rotate}`);
            } else {
                //变化图标
                this.arrowDom.classList.remove(this.options.arrow.iconClose);
                this.arrowDom.classList.add(this.options.arrow.iconOpen);
            }
        }
        axSlideDown(this.bodyDom, this.options.display, () => {
            this.expanded = true;
            //监听回调
            this.options.onExpanded && this.options.onExpanded.call(this);
            this.handlers.hasOwnProperty('expanded') ? this.emit('expanded', '') : null;
        });
        callback && callback.call(this);
    }
    collapse(callback) {
        //是折叠状态则不再执行
        if (!this.expanded) {
            return false;
        }
        if (this.arrowDom) {
            if (this.options.arrow.iconClose === this.options.arrow.iconOpen) {
                //翻转图标
                this.arrowDom.classList.add(`ax-deg${this.options.arrow.rotate}`);
            } else {
                //变化图标
                this.arrowDom.classList.remove(this.options.arrow.iconOpen);
                this.arrowDom.classList.add(this.options.arrow.iconClose);
            }
        }
        axSlideUp(this.bodyDom, this.options.display, () => {
            this.expanded = false;
            //监听回调
            this.options.onCollapsed && this.options.onCollapsed.call(this);
            this.handlers.hasOwnProperty('collapsed') ? this.emit('collapsed', '') : null;
        });
        callback && callback.call(this);
    }
    remove(callback) {
        //还是虚拟节点就不再执行
        if (axIsVnode(this.targetDom)) {
            return false;
        }
        this.targetDom.remove();
        //监听回调
        this.options.onRemove && this.options.onRemove.call(this);
        this.handlers.hasOwnProperty('remove') ? this.emit('remove', '') : null;
        callback && callback.call(this);
    }
    restore(callback) {
        //是存在的节点则不再执行
        if (!axIsVnode(this.targetDom)) {
            return false;
        }
        if (this.location.brother) {
            //优先从兄弟节点上恢复
            this.location.brother.insertAdjacentElement('beforebegin', this.targetDom);
        } else {
            this.location.parent.appendChild(this.targetDom);
        }
        //监听回调
        this.options.onRestore && this.options.onRestore.call(this);
        this.handlers.hasOwnProperty('restore') ? this.emit('restore', '') : null;
        callback && callback.call(this);
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
* 插件：光晕特效；使用方法：new axRipple('#id',{参数})
*/
class axRipple {
    constructor(elem, options) {
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            duration: 0,//光晕动画持续时间，默认为0即css中定义的600(单位ms)
            removeTime: 800,//光晕移除时间，默认800(单位ms)，一般需要光晕移除时间需要大于光晕动画持续时间
            theme: 'dark',//光晕颜色，默认dark，可填primary、danger等7种色值
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
        }, options, this.targetDom, this.constructor.name);
        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'ripple');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //添加ax-panel
        !this.targetDom.classList.contains(`ax-ripple`) ? this.targetDom.classList.add(`ax-ripple`) : null;
        //添加主题色
        this.targetDom.setAttribute('theme',this.options.theme);

        let width = axWidth(this.targetDom, 'inner'),
            height = axHeight(this.targetDom, 'inner'),
            size = Math.max(width, height);
        this.targetDom.addEventListener('click',  (e)=> {
            let x = e.clientX - (axOffset(this.targetDom).left - (document.documentElement.scrollLeft || window.pageXOffset || document.body.scrollLeft)),
                y = e.clientY - (axOffset(this.targetDom).top - (document.documentElement.scrollTop || window.pageYOffset || document.body.scrollTop)),
                ripple = axAddElem('u', {
                    style: `left: ${x}px;top: ${y}px; width: ${size}px;height: ${size}px${this.options.duration ? ';animation-duration:' + this.options.duration + 'ms' : ''}`,
                    class: this.options.theme,
                });
            this.targetDom.appendChild(ripple);
            setTimeout(() => {
                ripple.parentNode.removeChild(ripple);
            }, this.options.removeTime);
        }, false);
        return this;
    }
}
/*!
* 插件：光晕特效；使用方法：new axShortcut('#id',{参数})
*/
class axShortcut {
    constructor(elem, options) {
        this.targetDom = axIdToDom(elem);
        if (!this.targetDom) { return false; }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            shape: '',//快捷菜单子菜单外形，默认为空即是radius即3px的圆角，可选择round原型或square方形
            placement: 'right-bottom',//所处位置，默认右下角即right-bottom，可选择right-center，left-center和left-bottom
            gap: {
                side: '',//两侧边距，默认为空即是1.4rem（14px），可以填px值或rem值甚至使用calc()值
                bottom: '',//底部边距，默认为空即是1.4rem（14px），可以填px值或rem值甚至使用calc()值
                top: '',//顶部边距，默认为空即不设置，可以填px值或rem值甚至使用calc()值
            },
            size: '',//设置尺寸，默认为空即为md，可设置lg
            threshold: 400,//显示“回到顶部”按钮的滚动距离，默认400，可自定义
            scrollTo: 0,//点击返回按钮是回到哪里，默认为0即返回浏览器顶部，可以自定义一个距离顶部的数值，或者定义一个节点
            onInit: '',//回调函数，初始化后执行，无参数，但可使用内部变量，this.targetDom/this.children/this.totop
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
        }, options, this.targetDom, this.constructor.name);
        //滚动监听事件
        this.totopEvent = () => {
            let btn = this.totop.querySelector('a'),
                scrollTop = document.documentElement.scrollTop || window.pageYOffset || document.body.scrollTop;
            if (scrollTop >= this.options.threshold) {
                btn.classList.add('ax-show');
            } else {
                btn.classList.remove('ax-show');
            }
        }

        //初始化
        this.init();
    }
    init() {
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'shortcut');
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //定义不同平台的链接格式
        //设置样式和属性
        this.setAttribute();
        this.children = [...this.targetDom.children];
        //无子元素则不再执行
        if (this.children.length === 0) {
            return false;
        }
        this.totop = this.targetDom.querySelector('[totop]');

        for (let v of this.children) {
            let pop = v.querySelector('[pop]');
            new axHover(v, {
                enter: () => {
                    v.classList.add('ax-active');
                    if (pop) { axFadeIn(pop); }
                },
                leave: () => {
                    v.classList.remove('ax-active');
                    if (pop) { axFadeOut(pop); }
                }
            });
        }

        if (this.totop) {
            this.totop.addEventListener('click', () => {
                axScrollTo(this.options.scrollTo);
            }, false);
            window.addEventListener("scroll", this.totopEvent, false);
        }
        //监听回调
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    setAttribute() {
        //添加ax-shortcut
        !this.targetDom.classList.contains(`ax-shortcut`) ? this.targetDom.classList.add(`ax-shortcut`) : null;
        //添加外形
        this.options.shape ? this.targetDom.setAttribute('shape', this.options.shape) : null;
        //添加位置
        this.options.placement ? this.targetDom.setAttribute('placement', this.options.placement) : null;
        //添加尺寸
        this.options.size ? this.targetDom.setAttribute('size', this.options.size) : null;

        //添加间隙
        if (this.options.gap.side) {
            if (this.options.placement.includes('left')) {
                this.targetDom.style.left = this.options.gap.side;
                this.targetDom.style.right = 'auto';
            } else {
                this.targetDom.style.right = this.options.gap.side;
            }
        }
        if (this.options.gap.bottom) {
            if (this.options.placement.includes('bottom')) {
                this.targetDom.style.bottom = this.options.gap.bottom;
            }
        }
        if (this.options.gap.top) {
            this.targetDom.style.top = this.options.gap.top;
            this.targetDom.style.bottom = 'auto';
        }

    }
}
/*!
* 插件：元素进入视窗监听；使用方法：new axSpy('#id',{参数})
*/
class axSpy {
    constructor(elem, options) {
        if (axIsEmpty(elem)) {
            return false;
        }
        this.targetDom = axIdToDom(elem);
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            offset: '0px', //偏离顶部或底部多长距离触发，单位px，如果只有一个值则上下偏移值相同，如果需要单独定义请用英文逗号隔开，比如'-50px,-30px'表示上方偏移50px，下方偏移30px
            visible: [0, 0.5, 1], //可见区域占比，单位是小数，比如0.5表示目标50%的部分出现在屏幕中则触发，默认在刚显示、显示一半以及完全显示时触发
            parent: null,//以什么父节点为参照，可填#id、classname、nodename或节点，默认null表示以整个视窗为参照
            distance: 0,//侦听元素使用关联按钮滚动到顶部时距离顶部的高度，数值类型，默认为0即贴合顶部
            spySelector: '',//监听子元素的选择器，可以是nodeName或className（能使用querySelectorAll获取）
            navSelector: '',//菜单节点，可以是#id、nodeName、className或节点
            btnSelector: '',//监听菜单子元素的选择器，可以是nodeName或className（能使用querySelectorAll获取）
            duration: 0,//点击关联按钮页间跳转滚动动画的时间，默认为0即自动计算，可填具体数值，单位ms
            interval: 500,//防抖间隔时间，默认500ms，因为位移动画的缘故会导致交互+非交互两个状态之间频繁切换，继而导致动画抖动
            repeat: true,//是否每次出现在视窗则重复执行in动画，默认true，可选择false只执行一次（out状态不会删除in动画）
            active: '',//关联按钮激活的属性（写法：[selected]）或className（多个用英文逗号或空格分开）
            in: '',//进入视窗的动画className，多个用英文逗号或空格分开，例如in:'ax-fade ax-ani'
            out: '',//离开视窗的动画className，多个用英文逗号或空格分开，例如in:'ax-fade ax-ani'，通常不填！
            rel: '',//关联按钮，#id、className、nodeName或节点等，多个用英文逗号或空格分开，可是数组格式，如果是batch模式则不用填
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onBefore: '',//加载前执行函数，支持一个参数即被监听元素数组，如果该函数直接返回true则按既定程序执行加载，否则需手动添加监听（add方法）
            onShow: '',//开始进入视窗监听函数，支持三个参数，即监听节点、节点的关联按钮（数组），监听交互对象
            onShowing: '',//正在进入视窗监听函数，支持三个参数，即监听节点、节点的关联按钮（数组），监听交互对象
            onShown: '',//完全显示在视窗监听函数，支持三个参数，即监听节点、节点的关联按钮（数组），监听交互对象
            onIn: '',//进入视窗监听函数（包括show、showing和shown），支持三个参数，即监听节点、节点的关联按钮（数组），监听交互对象
            onOut: '',//离开视窗监听函数，支持三个参数，即监听节点、节点的关联按钮（数组），监听交互对象
            onAdd: '',//新增监听节点监听函数，支持一个参数，即监听新增监听节点数组
            onRemove: '',//取消监听节点监听函数，支持一个参数，即监听取消监听节点数组
            onInit: '',//回调函数，加载完毕执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
        }, options, this.targetDom, this.constructor.name);
        //回调监听
        this.handlers = {};
        //定义批量操作节点
        this.navDom = this.options.navSelector ? axIdToDom(this.options.navSelector) : null;
        this.btnDoms = this.navDom && this.options.btnSelector ? this.navDom.querySelectorAll(this.options.btnSelector) : [];
        
        //定义父层
        this.parentDom = this.options.parent ? axIdToDom(this.options.parent) : this.options.parent;
        //active类型
        if (this.options.active.trim().startsWith('[') && this.options.active.trim().endsWith(']')) {
            this.isAttr = true;
            this.active = this.options.active.replace('[', '').replace(']', '').replace(' ', '');
        } else {
            this.activeIsAttr = false;
            this.active = this.options.active;
        }
        //初始化
        this.init();
        return this;
    }
    init(callback) {
        axInstance.push(this, this.options.insName, 'spy');
         //设置销毁状态
         this.destroyed = false;
         //断点合并参数
         !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //需要监听的子元素合集
         this.items = [];
         //定义不同平台的链接格式
        //生成交互监听参数对象
        this.interactOpt();
        //创建监听
        this.interact = new IntersectionObserver(entries => {
            entries.forEach(i => {
                let args = [i.target, i.target.spy, i];
                if (i.isIntersecting) {
                    //开始进入视窗
                    if (i.target.spy.in) {
                        //没有才添加classname
                        axClass.add(i.target, i.target.spy.in, (name) => {
                            return i.target.classList.contains(name) ? false : true;
                        });
                    }

                    //监听回调
                    if (i.intersectionRatio === 0) {
                        //显示状态，为1表示完全显示了
                        i.target.spy.state = 'show';
                        this.handlers.hasOwnProperty('show') ? this.emit('show', ...args) : null;
                        this.options.onShow && this.options.onShow.call(this, ...args);
                    } else if (i.intersectionRatio === 1) {
                        //显示状态，为1表示完全显示了
                        i.target.spy.state = 'shown';
                        this.handlers.hasOwnProperty('shown') ? this.emit('shown', ...args) : null;
                        this.options.onShown && this.options.onShown.call(this, ...args);

                        //追加样式，只允许一个按钮处于激活状态
                        this.setActiveOnly(i.target);

                        //取消监听
                        if (!this.options.repeat) {
                            this.interact.unobserve(i.target);
                        }
                    } else {
                        i.target.spy.state = 'showing';
                        this.handlers.hasOwnProperty('showing') ? this.emit('showing', ...args) : null;
                        this.options.onShowing && this.options.onShowing.call(this, ...args);
                    }
                    this.handlers.hasOwnProperty('in') ? this.emit('in', ...args) : null;
                    this.options.onIn && this.options.onIn.call(this, ...args);

                } else {
                    //离开视窗
                    i.target.spy.state = 'hidden';
                    if (this.options.repeat) {
                        //只有重复则才删除in样式
                        axDebounce(() => {
                            if (i.target.spy.out) {
                                axClass.add(i.target, i.target.spy.out);
                            }
                            if (i.target.spy.in) {
                                axClass.del(i.target, i.target.spy.in);
                            }
                            if (this.active) {
                                i.target.spy.rel.forEach(k => {
                                    this.isAttr ? k.removeAttribute(this.active) : axClass.del(k, this.active);
                                });
                            }
                        }, this.options.interval)();
                    }

                    //监听回调
                    this.handlers.hasOwnProperty('out') ? this.emit('out', ...args) : null;
                    this.options.onOut && this.options.onOut.call(this, ...args);

                }
            });
        }, this.interactOpt());

        //将classes和rel作为属性附上
        if (this.options.spySelector) {
            //批量
            this.items = [...this.targetDom.querySelectorAll(this.options.spySelector)];
        } else {
            //单个
            this.items.push(this.targetDom);
        }

        //flag返回true才执行
        let flag = this.options.onBefore ? this.options.onBefore.call(this, this.items) : true;
        if (flag) {
            //将items应用到交互实例
            this.items.forEach(k => {
                //加入监听
                this.setSpyAttr(k);
                this.interact.observe(k);
            });
        }
         //监听回调
         this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
         this.options.onInit && this.options.onInit.call(this);
         callback && callback.call(this);
        return this;
    }
    setActiveOnly(dom) {
        if (this.active) {
            dom.spy.rel.forEach(k => {
                //没有才添加classname
                if (this.isAttr) {
                    k.setAttribute(this.active, '');
                } else {
                    axClass.add(k, this.active, (name) => {
                        return k.classList.contains(name) ? false : true;
                    });
                }

            });
            //只保留一个激活按钮
            let other = this.items.filter(k => k !== dom && k.spy),
                otherRels = other.map(k => k.spy.rel).flat(Infinity);
            otherRels.forEach(k => {
                this.isAttr ? k.removeAttribute(this.active) : axClass.del(k, this.active);
            });
        }
    }
    setSpyAttr(dom) {
        //必须是节点
        if (!axType(dom).includes('HTML')) {
            return false;
        }
        let attrs = axAttrToJson(dom, 'spy');
        dom.spy = {
            observed: true,
            rel: this.getRels(dom),
            in: attrs.in || this.options.in,
            out: attrs.out || this.options.out,
        }
    }
    getRels(dom) {
        let arrFromId = [], arrFromSpy = [], arrFromOpt = [], relArr = [];
        //从id获取
        if (dom.id) {

            if (this.btnDoms.length > 0 && this.btnDoms[0].hasAttribute('href')) {
                //导航有href=#xxx，说明已经绑定了
                let navRel = this.btnDoms.find(k => k.getAttribute('href') === `#${k.id}`);
                navRel ? arrFromId.push(navRel) : null;
            } else {
                arrFromId = [...document.querySelectorAll(`[href='#${dom.id}']`)];
            }
        } else {
            //从菜单中获得对应的按钮
            let index = this.items.findIndex(k => k == dom);
            index !== -1 && this.btnDoms[index] ? arrFromId.push(this.btnDoms[index]) : null;
        }
        //从spy.rel获取
        if (dom.hasAttribute('spy')) {
            let attrs = axAttrToJson(dom, 'spy');
            if (attrs.rel) {
                arrFromSpy = axAllToDoms(attrs.rel);
            }
        }
        //从rel参数获取
        if (this.options.rel) {
            arrFromOpt = axAllToDoms(this.options.rel);
        }
        //合并
        relArr = [...arrFromId, ...arrFromSpy, ...arrFromOpt].filter(Boolean);
        //点击rel按钮跳转到dom
        relArr.forEach(k => {
            k.onclick = (e) => {
                axPreventDefault(e);
                axScrollTo(dom, { parent: this.parentDom, duration: this.options.duration, offset: this.options.distance });
            }
        });
        return relArr;
    }
    interactOpt() {
        //更新父层偏离值
        this.offset = this.options.offset.replace(/\s*/g, "").split(",");
        if (this.offset.length == 1) {
            //如果只有一个值，那么追加第二个值等于第一个值
            this.offset.push(...this.offset)
        } else if (offset.length > 2) {
            //如果多写了，则只取前两个
            this.offset.splice(0, 2);
        }
        return {
            //父层即为视窗
            root: this.parentDom,
            //距离窗口上沿和下沿位置执行
            rootMargin: this.offset[0] + ' 0px ' + this.offset[1] + ' 0px',
            //目标节点的高度或宽度出现在父层的比例
            threshold: this.options.visible,
        }
    }
    getDoms(data) {
        let dataType = axType(data),
            doms = [];
        //获得节点数组
        if (dataType === 'Array') {
            doms = data.map(k => axIdToDom(k)).filter(Boolean);
        } else if (dataType === 'String') {
            doms = axAllToDoms(data).filter(Boolean);
        }
        return doms;
    }
    add(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let doms = this.getDoms(data);
        //追加监听
        doms.forEach(i => {
            if (!i.spy) {
                this.setSpyAttr(i);
                this.interact.observe(i);
                //加入items
                !this.items.includes(i) ? this.items.push(i) : null;
            } else {
                new axMessage({
                    content: `媒体文件${i.dataset.src || i.src}已经处于监听状态，不需要添加监听操作！`,
                    iconShow: true,
                    result: 'warning',
                }).show();
            }
        });
        //监听回调
        this.handlers.hasOwnProperty('add') ? this.emit('add', doms) : null;
        this.options.onAdd && this.options.onAdd.call(this, doms);
        callback && callback.call(this, doms);
    }
    remove(data, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        let doms = this.getDoms(data);
        //追加监听
        doms.forEach(i => {
            if (this.items.includes(i)) {
                if (i.spy) {
                    Reflect.deleteProperty(i, 'spy');
                    this.interact.unobserve(i);
                } else {
                    new axMessage({
                        content: `媒体文件${i.dataset.src || i.src}还未被监听，不需要取消监听操作！`,
                        iconShow: true,
                        result: 'warning',
                    }).show();
                }
            }
        });
        //监听回调
        this.handlers.hasOwnProperty('remove') ? this.emit('remove', doms) : null;
        this.options.onRemove && this.options.onRemove.call(this, doms);
        callback && callback.call(this, doms);
    }
    destroy(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //取消观察
        this.interact.disconnect();
        //按钮取消点击事件
        this.items.forEach(k=>{
            k.spy.rel.forEach(i=>{
                i.onclick = null;
            });
        })
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
/*!
* 插件：轮子滚动选择；使用方法：new axWheel('#id',{参数})
*/

class axWheel {
    constructor(elem, options) {
        if (!elem) { return false; }
        this.inputDom = axIdToDom(elem);
        if (!this.inputDom) { return false; }
        this.options = axExtend({
            insName: '',//实例名称，字符串格式；如果填写了实例名称，那么该实例会被添加到实例合集当中，通过axInstance方法可获取本实例，详细请查看：ax-utils-instance.php
            storageName: '',//存储名称，字符串格式；如果需要保存实例当前数据请设定唯一的值，比如：storageName:'mystorage001'
            content: '',//传入数据，支持多种写法，可传入#id或ul+li的dom节点；可传入当页的二维数组数据；可传入json文件地址和动态页面地址
            ajaxType: 'post', //异步提交的方式，默认post，可填get
            async: '', //异步方式，选择从json文件获取数组数据，或选择sql从数据库获得数据
            placeholder: '请选择',//滚轮窗的header默认文字
            selected: '',//初始值，字符串格式，文字与output.prop值一致（也就是说输出为label值，该参数也应该是label值的连接组合），分隔符需要与output.connector一致
            //举例：selected:'广东/深圳/龙华'，selected:'第一医院>临床/内科/神经科'，“>”分隔符之前是独立列，“/”连接符是联级列父子关系
            lamps: '',//滚轮列的列头，字符串格式，写法同selected，默认为空即不显示列头
            output: {
                prop: 'label',//取值属性，默认从label取值，如果数据有value属性可填value
                type: 'chain',//取值输出值类型，默认chain即取得链条数据，可选择ultimate只取终极分支的值
                connector: '/',//输出值链接符，默认为斜杠
                separator: '>',//独立列和联动列的分隔符，必须与connector不同，否则会出错
            },
            tools: {
                enable: false,
                data: [],//操作按钮数据，属性包含name、nodeName、props、callback（点击后触发的函数）
            },
            display: 'drawer',//展示方式，默认位于页面下方的drawer，可选择popup、dialog或inline
            bubble: {},//popup或dialog的参数
            buttonDom: '',//打开滚轮popup或dialog的按钮节点，#id/.classname/DIV或DOM节点
            positionDom: '',//定位滚轮的节点，#id/.classname/DIV或DOM节点
            toDrawer: true,//display如果是气泡，在手机端是否都转为drawer显示（手机端drawer显示滚轮窗体验最好）
            resultShow: true,//是否在头部显示选中结果，默认显示，可选择false不显示
            footerShow: true,//是否显示滚轮选择控件的底部按钮部分，默认true显示，可选择false不显示
            clearShow: true,//底部是否显示“清除”按钮，默认false不显示
            closeShow: true,//底部是否显示“关闭”按钮，默认true显示
            cancelShow: false,//是否显示“取消”按钮，默认false不显示，该按钮将实现clear+close两者的功能
            resetShow: false,//是否显示“重置”按钮，默认false不显示，该按钮将实现重置归零的功能
            confirmToHide: true,//点击确认按钮是否关闭窗口，默认true直接关闭，可选择false手动关闭
            autoFill: false,//进行滚轮操作，滚轮中的selected值是否自动填充到input，默认false不自动而是通过“确认”按钮填充
            autoCorrect: true,//对于非inline展示方式是否需要重新打开时从input/div中读取值作为当前的selected值，默认true自动校正，可选择false不处理
            rootStart: -1, //顶层的编号,数字类型,通常是0或-1
            idStart: 0,//id开始编号，数字类型；如果从物理节点创建菜单，则需要自动创建项目id，默认第一项id为0
            next: '',//关闭气泡后将要打开的其他气泡，该参数填写当页的其他气泡的实例名（insName）
            language: {
                button: {//底部按钮文字
                    clear: '清除',
                    close: '关闭',
                    cancel: '取消',
                    reset: '重置',
                    confirm: '确定',
                },
            },
            scroll: {},//滚轮列的参数（axScroll）
            note: '',//选择提示，置于滚轮下方，按钮上方，文本格式
            breakpoints: {},//使用断点以适配终端，详情查看ax-utils-other.php页面
            onInit: '',//回调函数，初始化后执行，无参数
            onSet: '',//回调函数，设置值后执行，支持回调函数，支持一个参数即当前设置的数据（带id、label、dom等属性）
            onConfirm: '',//回调函数，填入值后执行，支持一个参数即填入文本
            onReset: '',//回调函数，归零后执行，支持回调函数，回调无参数
            onClear: '',//回调函数，归零并清除值后执行，支持回调函数，回调无参数
            onCancel: '',//回调函数，归零+清除值+关闭气泡后执行，支持回调函数，回调无参数
            onUpdate: '',//回调函数，参数更新后执行，无参数
            onDestroy: '',//回调函数，销毁后执行，无参数
            onSave: '',//回调函数，保存数据后执行，支持一个参数即localstorage值
        }, options, this.inputDom, this.constructor.name);
        //回调监听
        this.handlers = {};

        //初始化
        this.init();
    }
    begin() {
        //断点合并参数
        !axIsEmpty(this.options.breakpoints) ? axBreakpoints(this.options, this.options.breakpoints) : null;
        //定义标准数据
        this.data = [];
        //当前滚轮的select值
        this.ultimates = [];
        //气泡显示次数
        this.shownTimes = 0;
        //当前选中的项:[{dom:'',label:''}]
        this.selecteds = [];
        //列头节点数组
        this.lamps = [];
        //添加到实例合集
        axInstance.push(this, this.options.insName, 'wheel');
        //设置销毁状态
        this.destroyed = false;
        //判断是否有存储值而更新参数
        if (this.options.storageName) {
            //如果有保存数据则更新参数
            let storageVal = axLocalStorage.get(this.options.storageName);
            if (!axIsEmpty(storageVal)) {
                //参数与存储值合并，更新了参数
                this.options = axExtend(this.options, storageVal);
            } else {
                //则先存下空对象
                axLocalStorage.set(this.options.storageName, {});
            }
        }
    }
    async init() {
        this.begin();
        //创建主节点节点
        this.createSkeleton();
        //设置风格
        this.setAttribute();

        //定位按钮和触发按钮，如果为空则默认等于inputDom
        this.positionDom = this.options.positionDom ? axIdToDom(this.options.positionDom) : this.inputDom;
        this.buttonDom = this.options.buttonDom ? axIdToDom(this.options.buttonDom) : this.inputDom;


        //移动端使用使用drawer模式
        if (axClient() === 'phone') {
            this.options.display !== 'inline' && this.options.toDrawer ? this.options.display = 'drawer' : null;
        }

        //展示方式
        if (this.options.display === 'inline') {
            this.positionDom.insertAdjacentElement('afterend', this.targetDom);
        } else if (this.options.display === 'popup') {
            this.bubbleIns = new axPopup(this.positionDom, axExtend({
                size: 'md',
                padding: false,
                placement: 'bottom-start',
                close: false,
                footerShow: false,
                rel: this.options.buttonDom,
                content: this.targetDom,
            }, this.options.bubble), this.buttonDom);
        } else if (this.options.display === 'dialog') {
            this.bubbleIns = new axDialog(axExtend({
                size: 'md',
                padding: false,
                footerShow: false,
                content: this.targetDom,
                rel: this.options.buttonDom,
            }, this.options.bubble), this.buttonDom);
        } else if (this.options.display === 'drawer') {
            this.bubbleIns = new axDrawer(axExtend({
                size: '',
                padding: false,
                footerShow: false,
                content: this.targetDom,
                placement: 'bottom',
                rel: this.options.buttonDom,
            }, this.options.bubble), this.buttonDom);
        }
        //源数据
        let dataType = axType(this.options.content);
        if (dataType == 'String' && this.options.async) {
            if (this.options.async === 'sql') {
                //从数据库获取数据
                await axTreeMethod.getFullData({
                    source: this.options.content,
                    ajaxType: this.options.ajaxType,
                    ajaxData: { pId: this.options.rootStart },
                    rootStart: this.options.rootStart,
                    idStart: this.options.idStart,
                    async: true,
                    opened: (response) => {
                        this.contentXhr = response.xhr;
                    },
                    before: (response) => {
                        this.bodyDom.innerHTML = '';
                        this.bodyDom.setAttribute('loading', '');
                    },
                    success: () => {
                        this.bodyDom.removeAttribute('loading');
                    }
                }).then((response) => {
                    //[{treeData,flatData,maxFloor}]
                    let cascade = response.data[0].hasOwnProperty('pId') ? true : false;
                    this.data = [{ id: 0, treeData: response.data, flatData: axArrToFlat(response.data), maxFloor: response.floorMax, cascade }];
                });
            }
            else if (this.options.async === 'json') {
                //从json获取数据，可能是一个tree也可能是独立列和tree的杂合体
                await axAjax({
                    //根据参数进行数据扩展
                    url: this.options.content,
                    type: this.options.ajaxType,
                    opened: (response) => {
                        this.contentXhr = response.xhr;
                    },
                    before: (response) => {
                        this.bodyDom.innerHTML = '';
                        this.bodyDom.setAttribute('loading', '');
                    },
                    success: (response) => {
                        if (axIsEmpty(response.content)) {
                            console.warn('No data obtained!');
                            return false;
                        }
                        this.bodyDom.removeAttribute('loading');

                        if (Array.isArray(response.content[0])) {
                            //返回的是混合数组[[],[],[]]
                            this.getMultiData(response.content).then((data) => {
                                this.data = data;
                            });
                        } else {
                            //返回的是树数组，则带上路径
                            let temp = axTreeMethod.addPath(response.content, this.options.rootStart, this.options.idStart);
                            this.data = [{ id: 0, treeData: temp.data, flatData: axArrToFlat(temp.data), maxFloor: temp.floorMax, cascade: true }];
                        }
                    }
                });
            }
        } else {
            //同步数据
            if (dataType === 'Array') {
                if (this.options.content.length === 0) {
                    console.warn('Array is empty!');
                    return false;
                }
                await this.getMultiData(this.options.content).then(data => {
                    this.data = data;
                });
            } else if (dataType.includes('HTML') || axStrType(this.options.content)) {
                //node节点，从ul+li获取数据，不要value值所以为false
                await axTreeMethod.getFullData({
                    source: this.options.content,
                    rootStart: this.options.rootStart,
                    idStart: this.options.idStart,
                }).then((response) => {
                    this.data = [{ id: 0, treeData: response.data, flatData: axArrToFlat(response.data), maxFloor: response.floorMax, cascade: true }];
                });
            }
        }

        //渲染完成，继续
        this.renderFinish();
        //监听回调
        this.handlers.hasOwnProperty('init') ? this.emit('init', '') : null;
        this.options.onInit && this.options.onInit.call(this);
        return this;
    }
    async getMultiData(data) {
        //转成标准数据[{treeData:'', flatData:'', maxFloor: ''},{treeData:'', flatData:'', maxFloor: ''},...]
        let result = [];
        for (let [i, k] of data.entries()) {
            let dataType = axType(k);
            //不是数组则跳过
            if (dataType === 'Object') {
                //data=[{},{}]，树数据
                await axTreeMethod.getFullData({
                    source: data,
                    rootStart: this.options.rootStart,
                    idStart: this.options.idStart,
                }).then((response) => {
                    result.push({ id: i, treeData: response.data, flatData: axArrToFlat(response.data), maxFloor: response.floorMax, cascade: true });
                });
                break;
            } else {
                //data=[[],[],[]...]，混合数据
                if (typeof k[0] === 'string' || typeof k[0] === 'number') {
                    //子项是[0,12,'5','a']这种数据，转成标准的[{id:'',label:''},{id:'',label:''}]
                    let data = k.map((v, u) => {
                        let obj = { id: u };
                        obj[this.options.output.prop] = v;
                        return obj;
                    });
                    result.push({ id: i, treeData: data, flatData: data, maxFloor: 1, cascade: false });
                } else {
                    //子项树数据
                    await axTreeMethod.getFullData({
                        source: k,
                        rootStart: this.options.rootStart,
                        idStart: this.options.idStart,
                    }).then((response) => {
                        result.push({ id: i, treeData: response.data, flatData: axArrToFlat(response.data), maxFloor: response.floorMax, cascade: true });
                    });
                }
            }
        }
        return result;
    }
    setAttribute() {
        //设置展示方式
        this.targetDom.setAttribute('display', this.options.display);
        //对input设置只读
        this.inputDom.nodeName === 'INPUT' ? this.inputDom.setAttribute('readonly', '') : null;
    }
    createSkeleton() {
        this.targetDom = axStrToDom(`
                     <div class="ax-wheel">
                        ${this.options.resultShow ? '<div class="ax-wheel-header">' + this.options.placeholder + '</div>' : ''}
                        ${this.options.tools.enable ? '<div class="ax-wheel-tools"></div>' : ''}
                        ${this.options.lamps ? '<ul class="ax-wheel-lamp"></ul>' : ''}
                        <div class="ax-wheel-body"></div>
                        ${this.options.note ? '<div class="ax-wheel-note">' + this.options.note + '</div>' : ''}
                        ${this.options.footerShow ? '<ul class="ax-wheel-footer"></ul>' : ''}
                    </div>
    `);
        //获取Dom树
        this.headerDom = this.targetDom.querySelector('.ax-wheel-header');
        this.toolsDom = this.targetDom.querySelector('.ax-wheel-tools');
        this.lampDom = this.targetDom.querySelector('.ax-wheel-lamp');
        this.bodyDom = this.targetDom.querySelector('.ax-wheel-body');
        this.noteDom = this.targetDom.querySelector('.ax-wheel-note');
        this.footerDom = this.targetDom.querySelector('.ax-wheel-footer');
        //定义按钮
        if (this.footerDom) {
            this.confirmDom = axAddElem('li', { confirm: '' }, this.options.language.button.confirm);
            this.clearDom = this.options.clearShow ? axAddElem('li', { clear: '' }, this.options.language.button.clear) : '';
            this.closeDom = this.options.closeShow ? axAddElem('li', { close: '' }, this.options.language.button.close) : '';
            this.cancelDom = this.options.cancelShow ? axAddElem('li', { cancel: '' }, this.options.language.button.cancel) : '';
            this.resetDom = this.options.resetShow ? axAddElem('i', { reset: '' }, this.options.language.button.reset) : '';

            //按钮插入节点
            this.options.cancelShow ? this.footerDom.appendChild(this.cancelDom) : null;
            this.options.closeShow ? this.footerDom.appendChild(this.closeDom) : null;
            this.options.clearShow ? this.footerDom.appendChild(this.clearDom) : null;
            this.options.resetShow ? this.footerDom.appendChild(this.resetDom) : null;
            this.footerDom.appendChild(this.confirmDom);
        }
        //插入工具按钮
        if (this.options.tools.enable) {
            this.options.tools.data.forEach(k => {
                k.dom = axAddElem(k.node || 'i', Object.assign({}, k.props), k.name);
                if (k.callback) {
                    k.dom.onclick = k.callback.bind(this, k);
                }
                this.toolsDom.appendChild(k.dom);
            });
        }
    }
    //打开前校验input值
    resetFromValue() {
        let value = this.getInputValue();
        //值相同便不需要重设了
        if (this.value === value) {
            return false;
        }
        //将input中的值转为selecteds
        this.selecteds = this.valueToSelects(value);
        //设置滚轮多列
        if (this.selecteds.flat(Infinity).length > 0) {
            for (let [i, k] of this.selecteds.entries()) {
                //空值则跳过，不需要更新滚轮
                if (!k) {
                    continue;
                }
                let cascade = this.data[i].cascade,
                    wheel = this.data[i].wheels[0].wheel,
                    dom = k[0].dom;
                if (!cascade) {
                    //独立列
                    wheel.snapped !== dom ? wheel.scrollTo(dom) : null;
                } else {
                    //联动列，强制吸附
                    wheel.snapped !== dom ? wheel.scrollTo(dom) : wheel.setSnapped({ force: true });
                }
            }
        }

    }
    createItem(obj) {
        //创建<li>节点
        let image = obj.image ? `<i image><img src="${obj.image}"/></i>` : '',
            avatar = obj.avatar ? `<i avatar><img src="${obj.avatar}"/></i>` : '',
            legend = obj.legend ? `<i legend class="${obj.legend}"></i>` : '',
            tip = obj.tip ? `<i tip>${obj.tip}</i>` : '',
            nodeTpl = `<li>${image}${avatar}${legend}<i label><# this.label #></i>${tip}</li>`;
        //写上第一组dom节点
        obj.dom = axStrToDom(axTplEngine(nodeTpl, obj));

        //回调
        obj.callback && obj.callback.call(this, obj);
        //生成dom节点
        return obj.dom;
    }
    arrayToSection(obj, data, parent) {
        let outer = obj.dom,
            inner = axAddElem('ul'),
            fragment = document.createDocumentFragment();
        outer.innerHTML = '';
        outer.appendChild(inner);
        //填充li
        data.forEach(item => {
            //创建item.dom（li节点）
            this.createItem(item);
            fragment.appendChild(item.dom);
        });
        inner.appendChild(fragment);
        if (parent) {
            //联级数据需要在紧挨着父节点之后添加容器
            parent.insertAdjacentElement('afterend', outer);
        } else {
            //初始化或者独立数据则直接加到body里
            this.bodyDom.appendChild(outer);
        }

        return outer;
    }
    createWheels() {
        let spreadChildren = (data, host) => {
            if (data.children) {
                //有子，激活滚轮列

                //生成列表
                host.wheels[data.floor].dom.innerHTML = '';
                //删除联级结束属性，即断点属性（如果只有一列独立数据则不需要设置breakpoint属性了）
                this.data.length > 1 ? host.wheels[data.floor - 1].dom.removeAttribute('breakpoint') : null;
                //给arrayToSection传入parent参数，让联级子容器跟着添加在父容器后面
                this.arrayToSection(host.wheels[data.floor], data.children, host.wheels[data.floor - 1].dom);
                //添加lamp
                host.wheels[data.floor].lamp ? this.lampDom.appendChild(host.wheels[data.floor].lamp) : null;
                //更新滚轮列
                toggleWheel(host.wheels[data.floor], host, false);
            } else {
                //无子，
                //添加联级结束属性，即断点属性（如果只有一列独立数据则不需要设置breakpoint属性了）
                this.data.length > 1 ? host.wheels[data.floor - 1].dom.setAttribute('breakpoint', '') : null;
                //之后的滚轮列隐藏
                let rest = host.wheels.slice(data.floor);
                rest.forEach(k => {
                    //隐藏滚轮列
                    toggleWheel(k, host, true);
                });
                //存下最后的数据，将从其path取值作为value填入input
                this.getUltimates(host, data, 'data');
                //保存数据
                this.save();
                //监听回调
                this.options.onSet && this.options.onSet.call(this, data);
                this.handlers.hasOwnProperty('set') ? this.emit('set', data) : null;
                //改变section宽度
                this.setWheelsWidth();
            }

        }, toggleWheel = (item, host, hidden = true) => {
            this.setWheelState(item, hidden);
            if (!hidden) {
                //创建虚拟滚动
                item.wheel = createScroll(item, host);
            }
        }, createScroll = (data, host) => {
            return new axScroll(data.dom, {
                //host.wheels的length>=this.selecteds的length
                transInit: this.selecteds[host.id] && this.selecteds[host.id][data.id] ? this.selecteds[host.id][data.id].dom : 0,
                //distMouse量大会有滑过头的bug
                distMouse: this.liHeight,
                paddingStart: this.liHeight * 2,
                paddingEnd: this.liHeight * 2,
                drift: { mode: 'snap', timeTHR: 100 },
                snap: { enable: true },
                toInitMode: 'locate',
                onInit: function () {
                    /* [...this.wrapperDom.children].forEach(i => {
                        i.querySelector('[label]').onclick = () => {
                            if (i !== this.snapped && this.distJitter < 10) {
                                this.scrollTo(i);
                            }
                        }
                    }); */
                },
                onClick: axThrottle(function (target) {
                    let li = target.closest('li');
                    if (li !== this.snapped && [...this.wrapperDom.children].includes(li)) {
                        this.scrollTo(target);
                    }
                }, 200),
                onSnap: (item) => {
                    if (host.cascade) {
                        //是联级数据
                        let curItem = host.flatData.find(k => k.dom === item);
                        curItem ? spreadChildren(curItem, host) : null;
                    } else {
                        //非联级数据，取得this.ultimates数据,item为节点
                        let data = this.getUltimates(host, item, 'snap');
                        //更新宽度
                        this.setWheelsWidth();
                        //添加断点属性（如果只有一列独立数据则不需要设置breakpoint属性了）
                        this.data.length > 1 ? host.wheels[0].dom.setAttribute('breakpoint', '') : null;
                        //保存数据
                        this.save();
                        //监听回调
                        this.options.onSet && this.options.onSet.call(this, data);
                        this.handlers.hasOwnProperty('set') ? this.emit('set', data) : null;
                    }
                    //自动合并并填入input
                    this.options.autoFill ? this.confirm(false) : null;
                }
            });
        }
        //从第一个轮子创建
        this.data.forEach((k) => {
            k.wheels[0].wheel = createScroll(k.wheels[0], k);
        });

    }
    getUltimates(host, item, from = 'data') {
        let data = from === 'snap' ? host.flatData.find(k => k.dom === item) : item,
            itemInUlt = this.ultimates.find(k => k.id === host.id);
        if (itemInUlt) {
            //如果已经添加过，则覆盖
            itemInUlt.data = data;
        } else {
            //如果还没添加过，直接push
            this.ultimates.push({ id: host.id, data });
        }
        //按id从小到大排序
        this.ultimates.sort((a, b) => a.id - b.id);
        //头更新
        this.headerDom ? this.headerDom.innerHTML = this.getValues() : null;
        return data;
    }
    setWheelState(item, hidden = true) {
        if (hidden) {
            //wheel
            Reflect.deleteProperty(item, 'wheel');
            item.hidden = true;
            item.dom.style.width = '0%';
            item.dom.innerHTML = '';
            item.dom.setAttribute('hidden', '');
            //lamp
            if (item.lamp) {
                item.lamp.style.width = '0%';
                item.lamp.setAttribute('hidden', '');
            }
        } else {
            //wheel
            Reflect.deleteProperty(item, 'hidden');
            item.dom.removeAttribute('hidden');
            //lamp
            item.lamp ? item.lamp.removeAttribute('hidden') : null;
        }
    }
    setWheelsWidth() {
        let itemsShown = this.data.map(k => k.wheels).flat(Infinity).filter(k => !k.hidden);
        itemsShown.forEach(k => {
            //保留两位小数
            let value = Math.floor((100 / itemsShown.length) * 100) / 100 + '%';
            k.dom.style.width = value;
            k.lamp ? k.lamp.style.width = value : null;
        });
    }
    renderFinish() {
        //创建滚轮section节点和lamp节点
        this.setWheelDoms();
        //此时this.data = [{},{treeData:[],flatData:[],maxFloor,wheels:[{id,dom,lamp}]},{}...]

        /*       this.data.forEach(k => {
                  //添加滚轮容器
                  this.bodyDom.appendChild(k.wheels[0].dom);
                  //创建滚轮子元素
                  this.arrayToSection(k.wheels[0], k.treeData);
              }); */

        //优先从selected参数获取初始值，再选择从input获得初始值
        if (this.options.selected) {
            this.selecteds = this.valueToSelects(this.options.selected);
            //直接填入input
            this.setInputValue(this.options.selected);
        } else if (this.options.autoCorrect && this.getInputValue()) {
            //如果允许从input取值并且有值可取
            this.selecteds = this.valueToSelects(this.getInputValue());
        }
        this.value = this.getInputValue();

        //显示的时候执行
        if (this.bubbleIns) {
            this.bubbleIns.on('show', () => {
                //先检测input
                if (this.options.autoCorrect) {
                    this.selecteds = this.valueToSelects(this.getInputValue());
                }

                //显示第一次初始化
                if (this.shownTimes === 0) {
                    //获得行高值，以便在axScroll进行相关设置
                    this.liHeight = parseInt(axStyle(this.bodyDom).height) / 5;
                    //创建滚轮多列
                    this.createWheels();
                } else {
                    //检查并从input中取值，以此重设滚轮列
                    //要求autoCorrect为true
                    this.options.autoCorrect ? this.resetFromValue() : null;
                }
                //显示次数+1
                this.shownTimes++;
            });
        } else {
            this.liHeight = parseInt(axStyle(this.bodyDom).height) / 5;
            this.createWheels();
        }



        //按钮操作
        if (this.confirmDom) {
            this.confirmDom.onclick = () => {
                this.confirm();
            }
        }
        //点击清除按钮
        if (this.clearDom) {
            this.clearDom.onclick = () => {
                this.clear();
            }
        }
        //点击清除并关闭按钮
        if (this.cancelDom) {
            this.cancelDom.onclick = () => {
                this.cancel();
            }
        }
        //点击重置归零
        if (this.resetDom) {
            this.resetDom.onclick = () => {
                this.reset();
            }
        }

    }
    getSelects() {
        //获得滚轮的吸附项
        let snappeds = this.data.map(k => k.wheels.map(i => i.wheel.snapped)),
            selects = snappeds.map((k, i) => {
                k.map(u => this.data[i].flatData.find(v => v.dom == u)).filter(Boolean);
            });
        return selects;
    }
    getValues(opt) {
        let options = Object.assign(this.options.output, opt),
            value = this.ultimates.map((k, i) => {
                if (k.data.path) {
                    //有path属性则为联级数据
                    let temp = axTreeMethod.getItemsFromPath({ path: k.data.path, data: this.data[i].flatData, connector: options.connector, prop: options.prop }).path,
                        tempArr = temp.split(options.connector);
                    //终极取值类型则取最后一个值
                    if (this.options.output.type === 'ultimate') {
                        temp = tempArr.at(-1);
                        tempArr = tempArr.slice(-1);
                    }
                    return !options.isArray ? temp : tempArr;
                } else {
                    //非联级数据
                    return [k.data[options.prop]]
                }
            });
        //输出
        return !options.isArray ? value.join(options.separator) : value;
    }
    confirm(hide = this.options.confirmToHide, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //合并并填入input
        this.value = this.getValues();
        //将文本传入input节点
        this.setInputValue(this.value);
        //点击确认按钮关闭气泡
        if (hide && this.bubbleIns) {
            this.bubbleIns.hide();
            //关闭自己，再打开其他日期实例
            let nextIns = this.options.next ? axInstance.find(this.options.next, 'wheel') : null;
            nextIns && nextIns.bubbleIns ? nextIns.bubbleIns.show() : null;
        }
        //监听回调
        this.handlers.hasOwnProperty('confirm') ? this.emit('confirm', this.value) : null;
        this.options.onConfirm && this.options.onConfirm.call(this, this.value);
        callback && callback.call(this, this.value);
    }
    //将值填入input
    setInputValue(text) {
        if (this.inputDom) {
            if (this.inputDom.nodeName === 'INPUT') {
                this.inputDom.value = text;
            } else {
                this.inputDom.innerHTML = text;
            }
        }
    }
    getInputValue() {
        let value = '';
        if (this.inputDom) {
            if (this.inputDom.nodeName === 'INPUT') {
                value = this.inputDom.value.trim();
            } else {
                value = this.inputDom.innerText.trim();
            }
        }
        return value;
    }
    valueToSelects(value) {
        //value='a/b/c'或者value='a>b/c/d>e'
        //使用/表示父子联级关系
        //使用>表示this.data的length>1，有多列独立数据
        value = value.trim()
        let itemsInChain = [];

        /*   if (value.includes(this.options.separator)) {
              //包含>，说明有多个独立列
              itemsInChain = value.split(this.options.output.separator).map((k, i) => {
                  if (k.includes(this.options.output.connector)) {
                      return k.split(this.options.output.connector).map(u => {
                          return this.data[i].flatData.find(v => v[this.options.output.prop] == u);
                      });
                  } else {
                      return this.data[i].flatData.find(u => u[this.options.output.prop] == k);
                  }
              }).filter(Boolean);
          } else {
              //不包含>，
              let times = 0;
              itemsInChain = this.data.map(k => {
                  if (times === 0) {
                      if (value.includes(this.options.output.connector) && k.cascade) {
                          times++;
                          //带/则是联级的，找到第一个树数据
                          return value.split(this.options.output.connector).map(u => {
                              return k.flatData.find(v => v[this.options.output.prop] == u);
                          }).filter(Boolean);;
                      } else if (!value.includes(this.options.output.connector) && !k.cascade) {
                          times++;
                          //无/，即非联级，找到第一个非联级数据
                          let temp = k.flatData.find(u => u[this.options.output.prop] == value);
                          return temp ? [temp] : [];
                      }
                  } else {
                      return [];
                  }
              });
          }
   */

        //从value或label找数据
        if (this.options.output.type === 'ultimate') {
            //整体作为检索值，value是个单值从data的最后一项找
            let item = this.data.at(-1).flatData.find(k => k[this.options.output.prop] == value);
            if (item) {
                itemsInChain = this.data.map((k, i) => {
                    let temp = [];
                    if (i === this.data.length - 1) {
                        //最后一项
                        if (k.flatData[0].hasOwnProperty('pId')) {
                            //联级数据，返回数组
                            temp = axTreeMethod.getItemsFromPath({ path: item.path, data: k.flatData }).items;
                        } else {
                            //独立数据，返回一个对象
                            temp = [k];
                        }
                    }
                    return temp;
                }).filter(Boolean);
                //itemsInChain =[[],[],[item]]
            }
        } else if (this.options.output.type === 'chain') {
            //分隔为检索值数组，转成['a',['x','y','z'],'c']后获取数据
            //标准的value并与data对应
            //if (value.includes(this.options.separator) && value.includes(this.options.connector) && this.data.length === 1 && this.data[0].cascade) {
            //特殊情况1：value='a',data只有一个联级树数据
            // let temp = this.data[0].flatData.find(k => k[this.options.output.prop] == value);
            // itemsInChain = axTreeMethod.getItemsFromPath({ path: temp.path, data: this.data[0].flatData }).items;
            //} else {
            //value转成['a',['x','y','z'],'c']和data是对应关系
            itemsInChain = value.split(this.options.output.separator).map((k, i) => {
                if (k.includes(this.options.output.connector)) {
                    //带/的链值vs树数组（应该且必须如此对应）
                    if (this.data[i].flatData[0].hasOwnProperty('pId')) {
                        //联级数据，链值vs树数组返回数组
                        return k.split(this.options.output.connector).map(u => {
                            return this.data[i].flatData.find(v => v[this.options.output.prop] == u);
                        });
                    } else {
                        //独立数据，链值vs独立数组返回空
                        return [];
                    }
                } else {
                    let temp = this.data[i].flatData.find(u => u[this.options.output.prop] == k),
                        result;
                    if (temp) {
                        if (this.data[i].cascade) {
                            //是联级数据，单值vs树数组则从path获取路径
                            result = axTreeMethod.getItemsFromPath({ path: temp.path, data: this.data[i].flatData }).items;
                        } else {
                            //非联级数据，单值vs独立数组则直接取值
                            result = [temp];
                        }
                    } else {
                        result = [];
                    }
                    return result;
                }
            }).filter(Boolean);
            //}

        }
        //itemsInChain和data的length需要一致
        return this.data.length !== itemsInChain.length ? this.data.map((k, i) => itemsInChain[i] ? itemsInChain[i] : []) : itemsInChain;
    }
    setWheelDoms() {
        if (this.options.lamps) {
            //将lamps字符串转成标准数组：[[''],['','',''],['']]
            this.lamps = this.options.lamps.split(this.options.output.separator).map(k => {
                return k.split(this.options.output.connector);
            });
        }

        this.data.forEach((k, i) => {
            //设置id
            k.hasOwnProperty('id') ? k.id = i : null;
            k.wheels = [];
            for (let x = 0; x < k.maxFloor; x++) {
                let wheelDom = axAddElem('section'),
                    lampDom = this.lamps[i] && this.lamps[i][x] ? axAddElem('li', '', this.lamps[i][x]) : null;
                k.cascade ? wheelDom.setAttribute('cascade', true) : wheelDom.setAttribute('cascade', false);
                k.wheels.push({ id: x, dom: wheelDom, lamp: lampDom });
            }
            //添加滚轮容器
            this.bodyDom.appendChild(k.wheels[0].dom);
            //添加列指示
            k.wheels[0].lamp ? this.lampDom.appendChild(k.wheels[0].lamp) : null;
            //创建滚轮子元素
            this.arrayToSection(k.wheels[0], k.treeData);
        });
    }
    set(value, callback) {
        value = value.trim();
        //填充input
        if (value) {
            this.setInputValue(value);
            this.selecteds = this.valueToSelects(value);
        }
        //显示气泡，滚轮列会根据input的value值进行重设
        if (this.bubbleIns && this.bubbleIns.state !== 'shown') {
            //在气泡显示，隐藏气泡的则显示他，根据show监听自动更新
            this.bubbleIns.show();
        } else {
            //在行内显示或是气泡显示状态，直接更新
            this.resetFromValue();
        }
        //监听回调
        callback && callback.call(this, this.selecteds);
        return this;
    }
    reset(callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //清空选中项，标准格式[[],[],[]]
        this.selecteds = this.data.map(k => []);
        //归位
        this.data.forEach(k => {
            if (k.cascade) {
                //联级树
                if (k.wheels[0].wheel.snapped === k.treeData[0].dom) {
                    //第一项就是吸附项，则再次强制吸附，以便联级滚轮列都归位
                    k.wheels[0].wheel.setSnapped({ force: true });
                } else {
                    //第一项不是吸附项，则直接回到头
                    k.wheels[0].wheel.toStart();
                }
            } else {
                //非联级，不是第一项吸附则直接回到头
                k.wheels[0].wheel.snapped !== k.treeData[0].dom ? k.wheels[0].wheel.toStart() : null;
            }
        });
        //监听回调
        this.handlers.hasOwnProperty('reset') ? this.emit('reset', '') : null;
        this.options.onReset && this.options.onReset.call(this);
        callback && callback.call(this);
        return this;
    }
    clear(callback) {
        //resetTime是指清除selected时是否同时让time归零，默认归零
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        //如果autoFill为true会使得input始终有值，所以就不需要清空了
        if (!this.options.autoFill) {
            //清空input
            this.setInputValue('');
        }
        //归零重置
        this.reset();

        //监听回调
        this.handlers.hasOwnProperty('clear') ? this.emit('clear', '') : null;
        this.options.onClear && this.options.onClear.call(this);
        callback && callback.call(this);
        return this;
    }
    cancel(callback) {
        if (this.destroyed) {
            return this;
        }
        //清除并关闭
        this.bubbleIns ? this.bubbleIns.hide() : null;
        this.clear();
        //监听回调
        this.handlers.hasOwnProperty('cancel') ? this.emit('cancel', '') : null;
        this.options.onCancel && this.options.onCancel.call(this);
        callback && callback.call(this);
        return this;
    }
    async update(setting, callback) {
        //如果已经销毁则不再执行
        if (this.destroyed) {
            return this;
        }
        this.options = axExtend(this.options, setting);
        //删除缓存
        this.options.storageName ? axLocalStorage.set(this.options.storageName, {}) : null;
        //清除HTML
        this.targetDom.remove();
        //保存数据
        this.save();
        await this.init();
        //监听回调
        this.handlers.hasOwnProperty('update') ? this.emit('update', '') : null;
        this.options.onUpdate && this.options.onUpdate.call(this);
        callback && callback.call(this);
        return this;
    }
    save(props, callback) {
        //如果已经销毁则返回实例
        if (this.destroyed) {
            return this;
        }
        //没有存储名则不再执行
        if (!this.options.storageName) {
            return false;
        }

        if (!props) {
            this.options.selected = this.getValues();
            axLocalStorage.set(this.options.storageName, this.options);
        } else {
            axLocalStorage.set(this.options.storageName, props);
        }
        //回调监听
        let localValue = axLocalStorage.get(this.options.storageName);
        this.handlers.hasOwnProperty('save') ? this.emit('save', localValue) : null;
        this.options.onSave && this.options.onSave.call(this, localValue);
        callback && callback.call(this, localValue);

        return this;
    }
    //销毁实例
    destroy(callback) {
        //取消绑定事件
        this.inputDom.onclick = null;
        this.positionDom ? this.positionDom.onclick = null : null;
        this.buttonDom ? this.buttonDom.onclick = null : null;
        //取消异步请求
        this.contentXhr ? this.contentXhr.abort() : null;
        //销毁内部实例
        this.bubbleIns ? this.bubbleIns.destroy() : null;
        //更新销毁状态
        this.destroyed = true;
        //监听回调
        this.handlers.hasOwnProperty('destroy') ? this.emit('destroy', '') : null;
        this.options.onDestroy && this.options.onDestroy.call(this);
        callback && callback.call(this);
        return this;
    }
    on(type, handler) {
        axAddPlan(type, handler, this);
        return this;
    }
    emit(type, ...params) {
        axExePlan(type, this, ...params);
    }
    off(type, handler) {
        axDelPlan(type, handler, this);
        return this;
    }
}
//////////////////////////////////////////// 版块分割线 ////////////////////////////////////////////
/*!
*项目所需的立即执行的代码片段，避免污染全局变量
*/
const axInit = (type, dom = document) => {
    let parent = axIdToDom(dom);
    //其他零碎操作axInit()或axInit('other')可执行
    if (axIsEmpty(type) || type.includes('other')) {
        //根据表单左标题使input或textarea增加padding-left样式
        let posLefts = [...parent.querySelectorAll('.ax-pos-left')];
        for (let k of posLefts) {
            if (k.nextElementSibling.nodeName == 'INPUT' || k.nextElementSibling.nodeName == 'TEXTAREA') {
                //使用rem单位以自适应移动端
                k.nextElementSibling.style.paddingLeft = `${axWidth(k, 'outer') / 10}rem`;
            }
        }
        //表单组右侧删除按钮，点击清除input或textarea
        let posRights = [...parent.querySelectorAll('.ax-pos-right [clear]')].filter(k => !k.clear),
            clearFun = function () {
                let elem = this.parentElement.previousElementSibling;
                if (elem.nodeName == 'INPUT' || elem.nodeName == 'TEXTAREA') {
                    elem.value = '';
                }
            };
        for (let k of posRights) {
            //做个clear=true标记，避免重复绑定事件
            k.clear = true;
            k.addEventListener('click', clearFun, false);
        }
        //







    }

    //执行插件模块
    let moduleNodeList = [],
        allModules = [
            'axAccordion', 'axAlert', 'axAmount', 'axCage','axCheckbox', 'axCheckAll', 'axComplete',
            'axDate', 'axDialog', 'axDodge', 'axDrawer', 'axDropdown','axGesture', 'axInfinite', 'axInput',
            'axLazyload', 'axList', 'axMenu', 'axMore', 'axPanel', 'axPagination', 'axPopup',
            'axProgress', 'axRange', 'axRipple', 'axSelect', 'axShare', 'axShortcut','axSpy', 'axStick', 'axSwiper',
            'axTab', 'axTags', 'axTooltip', 'axUpload', 'axTree', 'axRate','axWheel','axScroll','axSlider',
        ],
        //'axLightbox'，'axValid'较复杂，暂时未列入
        getNodeList = (arg) => {
            if (!axIsEmpty(arg)) {
                //如果不是以ax开头则拼接起来
                arg = arg.map(k => !k.startsWith('ax') ? 'ax' + k.slice(0, 1).toUpperCase() + k.slice(1) : k);
            }
            let modules = arg || allModules,
                modulesStr = modules.map(k => `[${k}]`).join(','),
                modulesDoms = [...parent.querySelectorAll(modulesStr)],
                modulesNodes = modules.map(k => {
                    //已经激活了的即带上模块属性的，则不再重新激活
                    let items = modulesDoms.filter(i => i.attributes.hasOwnProperty(k.toLowerCase()) && !i[k]);
                    return { module: k, nodeList: items }
                });
            return modulesNodes;
        },
        activeFun = (nodes) => {
            for (let v of nodes) {
                if (v.nodeList.length === 0) {
                    //页面没有对应的节点则跳过
                    continue;
                }
                if (['axDialog', 'axDrawer', 'axPopup'].includes(v.module)) {
                    //这三个气泡插件需要在内部再打开新气泡
                    higherParent(v.module, v.nodeList);
                } else {
                    let funStr = `dom`;
                    if (v.module === 'axMessage') {
                        //message不需要绑定节点
                        funStr = '';
                    } else if (v.module === 'axSwiper') {
                        funStr = `dom,axAttrToJson(dom, 'axSwiper')`;
                    }
                    v.nodeList.forEach(k => {
                        //new Function可能会有错误，直接忽略
                        let instance = new Function('dom', `try {return new ${v.module}(${funStr})} catch(e) {}`)(k);
                        //给节点做个标记，以避免重复激活，模块可能不存在
                        instance ? k[v.module] = true : null;
                    });
                }
            }
        },
        //气泡逐级打开，不被遮挡
        higherParent = (module, nodes) => {
            let eachInstance = (arr, parent) => {
                arr.forEach(k => {
                    let instance, funStr = `dom`;
                    if (module === 'axDialog' || module === 'axDrawer') {
                        //dialog和drawer特别些
                        funStr = `'',dom`;
                    }
                    //new Function可能会有错误，直接忽略
                    instance = new Function('dom', `try {return new ${module}(${funStr})} catch(e) {}`)(k); 
                    //模块可能不存在
                    instance ? k[module] = true : null;
                    parent && instance ? instance.targetDom.style.zIndex = parseInt(getComputedStyle(parent.targetDom).zIndex) + 1 : null;
                    let children = instance.content.querySelectorAll(`[${module}]`);
                    children.length > 0 ? eachInstance(children, instance) : null;
                });
            }
            eachInstance(nodes);
        }
    //激活节点使用插件
    if (!axIsEmpty(type)) {
        let moduleType = axType(type);
        if (moduleType === 'String') {
            //激活一类，比如“axPopup”，other类型不在此执行
            type !== 'other' ? moduleNodeList = getNodeList([type]) : null;
        } else if (moduleType === 'Array') {
            //激活多类，如['axDialog','axAccordion']，other类型不在此执行
            type = type.filter(k => k !== 'other');
            moduleNodeList = getNodeList(type);
        }
    } else {
        //激活所有插件
        moduleNodeList = getNodeList();
    }
    //此时moduleNodeList=[{module:'axDialog',nodeList:[dom1,dom2]}...]
    moduleNodeList.length > 0 ? activeFun(moduleNodeList) : null;

}
//加载页面则执行一次
axInit();